using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.Specialized;
using System.Collections;
using System.Windows.Threading;
using System.Reflection;
using System.Dynamic;
using System.ComponentModel;
using System.Diagnostics;
using System.Windows;

namespace AMPS.Client.Xaml
{
    /// <summary>
    /// A SOWModel is a local reflection of the SOW 
    /// </summary>
    public interface ISOWModel
    {
        int Count { get; }

        Dispatcher Dispatcher { get; set; }
        void Clear();
        void Reset();
        void SOW(string sowKey, byte[] data, int pos, int length);
        void Publish(string sowKey, byte[] data, int pos, int length);
        void OOF(string sowKey);
        void Sort(ListSortDirection? direction, string name);
    }


    public interface ISOWRecord
    {
        void Clear();
        void SetFrom(string key, string data);
        void SetFromFIX(byte[] fixdata, int position, int length);
        string GetSOWKey();
        void SetSOWKey(string key);
        string GetValue(string key);
        void Disconnect();
    }
    
    public class BaseSOWRecord : ISOWRecord
    {
        byte[] _message = new byte[256];
        int _storedLength = 0;
        protected bool _needsParse = false;
        string _sowKey;
        public string GetSOWKey()
        {
            return _sowKey;
        }
        public void SetSOWKey(string key)
        {
            _sowKey = key;
        }
        public virtual void Disconnect()
        {
        }

        public virtual void Clear()
        {
            throw new NotImplementedException();
        }
        public virtual void SetFrom(string key, string data)
        {
            throw new NotImplementedException();
        }
        public virtual string GetValue(string key)
        {
            throw new NotImplementedException();
        }
        public void SetFromFIX(byte[] fixdata, int pos, int len)
        {
            if (_message.Length < len)
                _message = new byte[len];
            Array.Copy(fixdata, pos, _message, 0, len);
            _needsParse = true;
            _storedLength = len;
        }
        protected void Parse()
        {
            Parse(_message, 0, _storedLength);
            
        }
        private void Parse(byte[] fixdata, int pos, int length)
        {
            _needsParse = false;
            int len = fixdata.Length;
            int tagbegin = 0;
            int tagend = 0;
            
            bool readingTag = true;
            for (int i = 0; i < len; i++)
            {
                if (readingTag)
                {
                    if (fixdata[i] == '=')
                    {
                        readingTag = false;
                        tagend = i;
                    }
                }
                else
                {
                    if (fixdata[i] == (char)0x01)
                    {

                        SetFrom(Encoding.ASCII.GetString(fixdata, tagbegin, tagend - tagbegin),
                                Encoding.ASCII.GetString(fixdata, tagend + 1, i - (tagend + 1)));
                        readingTag = true;
                        tagbegin = ++i;

                    }
                }
            }
        }
    }
    

    public class DynamicSOWRecord : DynamicObject, ISOWRecord, INotifyPropertyChanged
    {
        byte[] _message = new byte[256];
        int _storedLength = 0;
        bool _needsParse = false;
        private Dictionary<string, object> _data = new Dictionary<string, object>();
        string _sowKey;
        public void SetSOWKey(string key)
        {
            _sowKey = key;
        }
        public string GetSOWKey()
        {
            return _sowKey;
        }
        public void Clear()
        {
            // we might have cached a message that we never parsed.  dump it now.
            _needsParse = false;
            IList<string> keys = null;
            if(PropertyChanged != null) 
                keys = _data.Keys.ToList();
            _data.Clear();

            if (keys != null)
            {
                foreach (var key in keys)
                {
                    PropertyChanged(this, new PropertyChangedEventArgs(key));
                }

            }
            
        }
        public override IEnumerable<string> GetDynamicMemberNames()
        {
            return _data.Keys;
        }
        public override bool TryGetMember(GetMemberBinder binder, out object result)
        {
            if (_needsParse)
                Parse(this._message, 0, _storedLength);
            return _data.TryGetValue(binder.Name, out result);
        }
        public string GetValue(string key)
        {
            object result;
            if (_needsParse)
                Parse(this._message, 0, _storedLength);
            if (_data.TryGetValue(key, out result))
            {
                return (string)result;
            }
            else
            {
                return null;
            }
        }

        public override bool TrySetMember(SetMemberBinder binder, object value)
        {
            _data[binder.Name] = value;
            if (PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs(binder.Name));
            return true;
        }

        public void SetFrom(string key, string data)
        {
            _data[key] = data;
            if (PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs(key));
        }
        public void SetFromFIX(byte[] fixdata, int pos, int len)
        {
            if (_needsParse)
            {
                Parse(this._message, 0, _storedLength);
            }
            
            if (_message.Length < len)
                _message = new byte[len];
            Array.Copy(fixdata, pos, _message, 0, len);
            _needsParse = true;
            _storedLength = len;
            if (PropertyChanged != null) Parse(this._message, 0, _storedLength);
        }
        private void Parse(byte[] fixdata, int pos, int length)
        {
            _needsParse = false;
            int tagbegin = 0;
            int tagend = 0;
            
            bool readingTag = true;
            for (int i = pos; i < length; i++)
            {
                if (readingTag)
                {
                    if (fixdata[i] == '=')
                    {
                        readingTag = false;
                        tagend = i;
                    }
                }
                else
                {
                    if (fixdata[i] == (char)0x01)
                    {
                        SetFrom(Encoding.ASCII.GetString(fixdata, tagbegin, tagend-tagbegin),
                                Encoding.ASCII.GetString(fixdata, tagend + 1, i - (tagend + 1)));
                        readingTag = true;
                        tagbegin = ++i;

                    }
                }
            }
        }
        public void Disconnect()
        {
            PropertyChanged = null;
            Clear();
        }
        public event PropertyChangedEventHandler PropertyChanged;
    }

    class TPool<T> where T: ISOWRecord, new()
    {
        Queue<T> _items = new Queue<T>();
        int _initial, _increment;
        public TPool(int initial, int increment)
        {
            _initial = initial;
            _increment = increment;
            allocate(initial);
        }

        private void allocate(int items)
        {
            
            for (int i = 0; i < items; i++)
            {
                _items.Enqueue(new T());
            }
        }
        public T create()
        {
            if (_items.Count == 0)
            {
                allocate(_increment);
            }
            return _items.Dequeue();
            
        }
        public void delete(T item)
        {
            if (item != null)
            {
                _items.Enqueue(item);
                item.Disconnect();
            }
        }
    }

    
    public class SOWModel<T> : ISOWModel, INotifyCollectionChanged, IEnumerable where T: ISOWRecord, new()
    {
        private List<T> _items = new List<T>();
        private SortedList<string, T> _sortedItems = new SortedList<string, T>();
        private TPool<T> _free = new TPool<T>(1000000, 10000);
        private Dictionary<string, int> _sowKeys = new Dictionary<string, int>();
        public event NotifyCollectionChangedEventHandler CollectionChanged;
        public Dispatcher Dispatcher { get; set; }
        public Queue<int> _empties = new Queue<int>();

        public SOWModel()
        {
        }

        public IEnumerator<T> GetEnumerator()
        {
            return _items.GetEnumerator();
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return _items.GetEnumerator();
        }
        public int Count
        {
            get
            {
                return _items.Count;
            }
        }
        public void Clear()
        {
            this.Dispatcher.Invoke(new Action(() =>
            {
                int i = 0;
                while (i < this._items.Count)
                {
                    _free.delete(this._items[i]);
                    ++i;
                }

                this._items.Clear();
                this._sowKeys.Clear();
                if (this.CollectionChanged != null)
                {
                    this.CollectionChanged(this, new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
                }
            }
            ));

        }

        public void Reset()
        {
            Dispatcher.Invoke(new Action(() =>
            {
                if(this.CollectionChanged != null) this.CollectionChanged(this, new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
            }
            
            ));
        }
       
        public void SOW(string sowKey, byte[] data, int pos, int length)
        {
            int idx = _items.Count;
            T t = _free.create();
            t.SetFromFIX(data, pos, length);
            _sowKeys[sowKey] = idx;
            _items.Add(t);
            //t.SetSOWKey(sowKey);
        }
        public void Publish(string sowKey, byte[] data, int pos, int length)
        {
            // is it new?  nobody knows, yet.
            int index;
            if(_sowKeys.TryGetValue(sowKey, out index))
            {
                // nope, not new
                if (index < _items.Count)
                {
                    T t = _items[index];
                    t.SetSOWKey(sowKey);
                    Dispatcher.Invoke(new Action(() => t.SetFromFIX(data, pos, length)));
                }
            }
            else
            {
                if (_empties.Count == 0)
                {
                    T t = _free.create();
                    t.SetFromFIX(data, pos, length);
                    index = _items.Count;
                    _items.Add(t);
                    _sowKeys[sowKey] = index;
                    t.SetSOWKey(sowKey);
                    Dispatcher.Invoke(new Action(() => CollectionChanged(this, new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, t, index))));
                }
                else
                {
                    index = _empties.Dequeue();

                    _sowKeys[sowKey] = index;
                    T t = _items[index];
                    t.SetSOWKey(sowKey);
                    
                    Dispatcher.Invoke(new Action(() =>
                        t.SetFromFIX(data, pos, length) ));
                }
            }
        }
        public void OOF(string sowKey)
        {
            int index;

            if (_sowKeys.TryGetValue(sowKey, out index))
            {
                if (index < _items.Count)
                {
                    T o = _items[index];
                    _empties.Enqueue(index);
                    
                    Dispatcher.Invoke(new Action(() =>
                        o.Clear()));
                }
            }
        }

        class Comparer<Q> : IComparer<Q> where Q:ISOWRecord
        {
            string _propname;
            bool _dir;
            int _compares = 0;
            public Comparer(string propname, bool dir)
            {
                _propname = propname;
                _dir = dir;
            }
            public int Compare(Q x, Q y)
            {
                string s1 = x.GetValue(_propname);
                string s2 = y.GetValue(_propname);
                s1= s1==null?string.Empty:s1;
                s2= s2==null?string.Empty:s2;
                ++_compares;
                if (_dir)
                    return s1.CompareTo(s2);
                else
                    return s2.CompareTo(s1);
            }
            public int Compares { get { return _compares; } }

        }
        // This sort is unused, for now.
        public void Sort(ListSortDirection? dir, string name)
        {
            Stopwatch sw = new Stopwatch();
            sw.Start();
            var c = new Comparer<T>(name, dir == ListSortDirection.Ascending);
            _items.Sort(c);
            sw.Stop();
            Console.WriteLine("sorted in {0} milliseonds, {1} comparisons.", sw.ElapsedMilliseconds, c.Compares);
            for (int i = 0; i < _items.Count; i++)
            {
                string key = _items[i].GetSOWKey();
                if (key != null)
                    _sowKeys[key] = i;
            }
            if (CollectionChanged != null)
            {
                CollectionChanged.Invoke(this, new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
            }
        }
    }

    public class ExpandoSOWModel : SOWModel<DynamicSOWRecord>
    {
    }

}
