﻿////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies 
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Collections.ObjectModel;
using System.Windows.Threading;
using System.ComponentModel;
using System.Threading;

namespace AMPS.Client.Xaml
{
    public class Server : DependencyObject, ClientDisconnectHandler
    {
        public static readonly DependencyProperty HostnameProperty = DependencyProperty.Register("Hostname", typeof(string), typeof(Server));
        public static readonly DependencyProperty PortProperty = DependencyProperty.Register("Port", typeof(int), typeof(Server));
        public static readonly DependencyProperty MessageTypeProperty = DependencyProperty.Register("MessageType", typeof(string), typeof(Server));
        public static readonly DependencyProperty UsernameProperty = DependencyProperty.Register("Username", typeof(string), typeof(Server));
        public static readonly DependencyProperty PasswordProperty = DependencyProperty.Register("PasswordProperty", typeof(string), typeof(Server));

        public string Hostname { get { return (string)GetValue(HostnameProperty); } set { SetValue(HostnameProperty, value); } }
        public int Port { get { return (int)GetValue(PortProperty); } set { SetValue(PortProperty, value); } }
        public string MessageType { get { return (string)GetValue(MessageTypeProperty); } set { SetValue(MessageTypeProperty, value); } }
        public string Username { get { return (string)GetValue(UsernameProperty); } set { SetValue(UsernameProperty, value); } }
        public string Password { get { return (string)GetValue(PasswordProperty); } set { SetValue(PasswordProperty, value); } }

        ObservableCollection<String> _topics = new ObservableCollection<String>();
        HashSet<string> _topicDiscriminator = new HashSet<string>();
        bool _topicSubActive = false;

        public Server()
        {
            Port = 9004;
            MessageType = "fix";
        }
        public string Uri
        {
            get
            {
                return GetUri();
            }
        }
        Client _client;

        private string GetUri(bool withPassword = false)
        {
            if (string.IsNullOrEmpty(Username))
            {
                return string.Format("tcp://{0}:{1}/{2}", Hostname, Port, MessageType);
            }
            else if (string.IsNullOrEmpty(Password) || !withPassword)
            {
                return string.Format("tcp://{0}@{1}:{2}/{3}", Username, Hostname, Port, MessageType);
            }
            else
            {
                return string.Format("tcp://{0}:{1}@{2}:{3}/{4}", Username, Password, Hostname, Port, MessageType);
            }
        }
        private void CreateSOWStatsSubscription(Client client)
        {
            Dispatcher callingThread = Dispatcher.CurrentDispatcher;
            new Thread(new ThreadStart(new Action(() =>
            {

                try
                {

                    client.sowAndSubscribe((msg) =>
                    {

                        string fixData = msg.Data;
                        int firstCharacter = 0;
                        int end = 0;
                        while (true)
                        {
                            end = fixData.IndexOf('=', firstCharacter);
                            if (end == -1) break;
                            string fieldName = fixData.Substring(firstCharacter, end - firstCharacter);

                            firstCharacter = fixData.IndexOf((char)0x01, end);
                            string data = (firstCharacter == -1) ? fixData.Substring(end + 1) :
                                          fixData.Substring(end + 1, firstCharacter - (end + 1));

                            if (fieldName == "20066" && _topicDiscriminator.Add(data))
                            {
                                callingThread.Invoke(new Action(() =>
                                {
                                    // Must create a new Subscription here that 
                                    // belongs to the UI thread.
                                    _topics.Add(data);
                                }));
                                break;
                            }

                            if (firstCharacter == -1 || firstCharacter > fixData.Length - 4)
                            {
                                break;
                            }
                            ++firstCharacter;
                        }
                    },
                    "/AMPS/SOWStats", 0);
                }
                catch (Exception)
                {
                }
            }))).Start();
        }

        public ObservableCollection<String> Topics
        {
            get
            {

                if (!DesignerProperties.GetIsInDesignMode(this) && !_topicSubActive)
                {

                    CreateSOWStatsSubscription(GetClient(Uri, true));
                    _topicSubActive = true;

                }

                return _topics;
            }
        }

        public AMPS.Client.Client GetClient(bool cache = true)
        {
            if (this._client != null) return this._client;

            try
            {
                if (DesignerProperties.GetIsInDesignMode(this))
                {
                    return null;
                }
            }
            catch (Exception)
            {
            }
            return GetClient(GetUri(true), cache);

        }

        private AMPS.Client.Client GetClient(string uri, bool cache = true)
        {
            AMPS.Client.Client client = this._client;
            if (client == null || !cache)
            {
                client = new AMPS.Client.Client(Guid.NewGuid().ToString());
                client.setDisconnectHandler(this);
                client.connect(uri);
                client.logon();
                if (cache)
                {
                    _client = client;
                }
            }
            return client;
        }

        public void invoke(Client client)
        {
            this.ConnectionStateChanged(this, new ConnectionEventArgs { Message = "Disconnected, attempting to retry.", State = ConnectionEventArgs.States.Reconnecting });
            // ugh, have to ask the other thread to retrieve the uri for us.
            string uri = "";

            this.Dispatcher.Invoke(new Action(() =>
            {
                uri = GetUri(true);
            }
            ));

            for (int i = 0; i < 20; i++)
            {
                try
                {
                    client.connect(uri);
                    client.logon(1000);
                    // need the main thread to do this, because we may or may not be in the background processing thread.
                    this.ConnectionStateChanged(this, new ConnectionEventArgs { Message = "Succeeded", State = ConnectionEventArgs.States.Connected });
                    if (_topicSubActive)
                    {
                        CreateSOWStatsSubscription(client);
                    }

                    return;
                }
                catch (Exception e)
                {
                    if (e.GetType() != typeof(AMPS.Client.Exceptions.AlreadyConnectedException))
                    {
                        this.ConnectionStateChanged(this,
                            new ConnectionEventArgs { Message = "Reconnect failed (" + e.Message + ").  Retrying...", State = ConnectionEventArgs.States.Reconnecting });
                        Thread.Sleep(3000);
                    }
                }
            }

            this.ConnectionStateChanged(this, new ConnectionEventArgs { Message = "Reconnect failed.", State = ConnectionEventArgs.States.Disconnected });

        }

        public class ConnectionEventArgs
        {
            public string Message { get; set; }
            public enum States { Disconnected, Reconnecting, Connected };
            public States State { get; set; }
        }
        public delegate void ConnectionStateChangedEventHandler(object sender, ConnectionEventArgs e);

        public event ConnectionStateChangedEventHandler ConnectionStateChanged;


        public string GetXaml()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("<amps:Server x:Key=\"MyServer\"");
            if (!string.IsNullOrEmpty(this.Hostname))
            {
                sb.AppendFormat(" Hostname=\"{0}\"", this.Hostname);
            }
            sb.AppendFormat(" Port=\"{0}\"", this.Port);
            if (!string.IsNullOrEmpty(this.MessageType))
            {
                sb.AppendFormat(" MessageType=\"{0}\"", this.MessageType);
            }
            if (!string.IsNullOrEmpty(this.Username))
            {
                sb.AppendFormat(" Username=\"{0}\"", this.Username);
            }
            if (!string.IsNullOrEmpty(this.Password))
            {
                sb.AppendFormat(" Password=\"...\"");
            }
            sb.Append("/>");

            return sb.ToString();

        }
    }
}
