////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System.Runtime.CompilerServices;
using System.Collections.Generic;
using System.Threading;
using System.Diagnostics;
using System.IO;
using System.Text;
using System;
using AMPS.Client.Exceptions;
using AMPS.Client.Fields;

namespace AMPS.Client
{
    /// <summary>
    /// BlockPublishStore is a base class that implements the Store interface.
    /// It is designed for publish store implementations, allowing messages to be stored
    /// and managed efficiently. As messages are discarded, space in that buffer is marked
    /// "free" for future store operations. If messages are stored faster than they are
    /// published, the buffer is re-sized to create more capacity.
    /// </summary>
    public class BlockPublishStore : Store
    {
        private long UNIX_EPOCH_OFFSET = new System.DateTime(1970, 1, 1, 0, 0, 0).Ticks;

        // This field represents the number of blocks per reallocation. It determines
        // how many blocks are allocated when resizing the buffer to create more
        // capacity
        private int _blocksPerRealloc = 10000;

        // Keeps track of the next sequence number to be used when storing messages
        private ulong _nextSequence = 1;

        // boolean flag that indicates whether the buffer is currently being resized
        private volatile bool _resizing = false;
        protected bool _errorOnPublishGap = false;
        protected PublishStoreResizeHandler _resizeHandler = null;

        /// <summary>
        /// A simple wrapper object around a byte array that allows a sub-range to be specified using
        /// its offset and length properties.
        /// </summary>
        public class ByteSequence
        {
            public byte[] array;
            public long offset, len;

            /// <summary>
            /// This constructor initializes a 'ByteSequence' with null array and zero offset
            /// and length
            /// </summary>
            public ByteSequence()
            {
                this.array = null;
                this.offset = 0;
                this.len = 0;
            }

            /// <summary>
            /// Initializes a new instance of the ByteSequence class with
            /// specified array, offset, and length values.
            /// </summary>
            /// <param name="array">The byte array.</param>
            /// <param name="offset">The offset within the byte array.</param>
            /// <param name="len">The length of the byte sequence.</param>
            public ByteSequence(byte[] array, long offset, long len)
            {
                this.array = array;
                this.offset = offset;
                this.len = len;
            }
        }

        /// <summary>
        /// Interface which is used to hold the BlockPublishStore buffer data.
        /// </summary>
        public interface Buffer
        {
            /** <summary>Returns the size of the buffer</summary>*/
            long getSize();

            /** <summary>Sets the size of the buffer</summary>*/
            void setSize(long newSize);

            /** <summary>Returns the current position in the buffer</summary>*/
            long getPosition();

            /** <summary>Sets thd current postion in the buffer</summary>*/
            void setPosition(long position);

            /** <summary>Writes a byte to the buffer</summary>*/
            void putByte(byte b);

            /** <summary>Reads a byte from the buffer</summary>*/
            byte getByte();

            /** <summary>Writes an integer to the buffer</summary>*/
            void putInt(int i);

            /** <summary>Writes an integer to a specified position in the buffer</summary>*/
            void putInt(long position, int i);

            /** <summary>Reads an integer from the buffer</summary>*/
            int getInt();

            /** <summary>Reads an integer from a specified position in the buffer</summary>*/
            int getInt(long position);

            /** <summary>Writes a long to the buffer</summary>*/
            void putLong(long l);

            /** <summary>Writes a long to a specified position in the buffer</summary>*/
            void putLong(long position, long l);

            /** <summary>Reads a long from the buffer</summary>*/
            long getLong();

            /** <summary>Writes an unsigned long to the buffer</summary>*/
            void putUlong(ulong l);

            /** <summary>Writes an unsigned long to a specified position in the buffer</summary>*/
            void putUlong(long position, ulong l);

            /** <summary>Reads an unsigned long from the buffer</summary>*/
            ulong getUlong();

            /** <summary>Writes a ByteSequence to the buffer</summary>*/
            void putBytes(ByteSequence bytes);

            /** <summary>Reads a ByteSequence from the buffer</summary>*/
            void getBytes(ByteSequence outBytes);

            /** <summary>Writes a byte array to the buffer</summary>*/
            void putBytes(byte[] buffer, int offset, int length);

            /** <summary>Reads data into a Field object from the buffer</summary>*/
            void getBytes(Field outBytes, int length);

            /** <summary>Sets a range of bytes to zero in the buffer</summary>*/
            void zero(long offset, int length);
        }

        /**********************************************************************
         * Block format
         **********************************************************************
         * Field Description                                | Type  | # Bytes
         **********************************************************************
         * HEADER as detailed below                         |       | 32 TOTAL
         *                                                  |       |
         * Num used bytes in block or total length in 1st   | int   | 4
         * block in chain.                                  |       |
         * Block index (current if last in chain)           | int   | 4
         * (next block index if there are more              |       |
         * block in the chain)                              |       |
         *                                                  |       |
         * HA Message sequence (if first block in chain)    | ulong | 8
         * (else 0)                                         |       |
         *                                                  |       |
         * CRC Value                                        | long  | 8
         *                                                  |       |
         * EMPTY (ZEROED OUT)                               | long  | 8
         *
         **********************************************************************
         * DATA SECTION CHAIN HEADER (present only on first block in chain)
         * as detailed below                                |       | 32 TOTAL
         *                                                  |       |
         * operation                                        | int   | 4
         * command id length                                | int   | 4
         * correltation id length                           | int   | 4
         * expiration length                                | int   | 4
         * sow key length                                   | int   | 4
         * topic length                                     | int   | 4
         * sow delete flag                                  | int   | 4
         * ack types                                        | int   | 4
         **********************************************************************
         * DATA SECTION - can be spread across multiple blocks
         *
         * command id                                       | byte
         * correlation id                                   | byte
         * expiration                                       | byte
         * sow key                                          | byte
         * topic                                            | byte
         * data                                             | byte
         **********************************************************************/
        public class Block
        {
            // An integer representing the index of the block
            public int index;
            // A reference to the next block in the chain
            public Block nextInChain;
            // A reference to the next block in the list
            public Block nextInList;
            // A long representing the sequence number associated with the block
            public ulong sequenceNumber;
            // A constant integer indicating the size of a block
            public const int SIZE = 2048;  // finely honed.
            // A constant integer representing the size of the block header
            public const int BLOCK_HEADER_SIZE = 32;
            // A constant integer representing the size of the block data
            public const int BLOCK_DATA_SIZE = Block.SIZE - BLOCK_HEADER_SIZE;
            // Position of the CRC value
            public const int BLOCK_CRC_POSITION = 16;
            // A constant integer representing the chain header size
            public const int CHAIN_HEADER = 32;
            // Position of the ready flag, non-zero is ready
            public const int BLOCK_READY_POSITION = 24;

            /// <summary>
            /// Initializes a block with the specified Index
            /// </summary>
            public Block(int index) { this.index = index; }

            /// <summary>
            /// Returns the position of the block within the buffer
            /// </summary>
            public long getPosition() { return SIZE * index; }

            /// <summary>
            /// Provides a string representation of the block, including its index and sequence
            /// number.
            /// </summary>
            public string toString()
            {
                return string.Format("block {0} sequenceNumber {1}", index, sequenceNumber);
            }
        }

        object _internalLock = new object();

        // A buffer acts as a random access file -- mmap, memory only, and file implementations exist.
        protected Buffer _buffer;

        // This system uses intrusive lists.  _freeList is the list of free blocks in the file.
        // _usedList is the list of the used head blocks.  Used non-head blocks are only kept
        // alive by "chain" references from the head block.
        // Used list is ordered where the start is the LEAST recently used.
        volatile internal Block _freeList, _usedList, _endOfUsedList;

        // One Block is reserved for holding metadata
        // The metadata format is currently:
        // int size - unused
        // int nextInChain - set to the int of the version number, such as 4000100 for 4.0.1.0
        // ulong sequence - set to the index of the last discarded
        internal Block _metadataBlock;

        // Save this so we're not GCing stuff.
        protected Utilities.CRC32 _crc = null;

        // When a message spans multiple blocks we need to reassemble it
        // into one place before replaying. We use this buffer to do so,
        // allocating on demand.
        byte[] _readBuffer;

        // Fixed "ByteSequence" structure to pass to underlying BufferStore
        ByteSequence _argument = new ByteSequence();

        // Used for replay
        Message _message;

        const int AMPS_MIN_PUB_STORE_DISCARDED_VERSION = 4000100;
        const int METADATA_VERSION_LOCATION = 4;
        const int METADATA_LAST_DISCARDED_LOCATION = 8;

        /// <summary>
        /// Initializes a new instance of the BlockPublishStore class with the specified buffer and block allocation configuration.
        /// </summary>
        /// <param name="buffer">The buffer used for storing messages.</param>
        /// <param name="blocksPerRealloc">The number of blocks per reallocation, determining the block allocation during buffer resizing.</param>
        /// <param name="errorOnPublishGap_">Indicates if an excpetion should be thrown if discardUpTo is called with an index below the last persisted.</param>
        protected BlockPublishStore(Buffer buffer, int blocksPerRealloc,
                                    bool errorOnPublishGap_ = false)
        {
            this._blocksPerRealloc = blocksPerRealloc;
            this._buffer = buffer;
            this._errorOnPublishGap = errorOnPublishGap_;
        }


        /// <summary>
        /// Initializes a new instance of the BlockPublishStore class with the specified buffer.
        /// </summary>
        /// <param name="buffer">The buffer used for storing messages.</param>
        /// <param name="useCrc">A boolean flag indicating whether to use CRC (Cyclic Redundancy Check).</param>
        /// <param name="blocksPerRealloc">The number of blocks per reallocation, determining the block allocation during buffer resizing.</param>
        /// <param name="errorOnPublishGap_">Indicates if an excpetion should be thrown if discardUpTo is called with an index below the last persisted.</param>
        protected BlockPublishStore(Buffer buffer, bool useCrc = false,
                                    int blocksPerRealloc = 10000,
                                    bool errorOnPublishGap_ = false)
        {
            this._blocksPerRealloc = blocksPerRealloc;
            this._buffer = buffer;
            // If useCrc is true, initialize CRC (Cyclic Redundancy Check)
            if (useCrc) _crc = new Utilities.DefaultCRC32();
            this._errorOnPublishGap = errorOnPublishGap_;
        }

        /// <summary>
        /// Performs implementation-defined tasks associated with freeing, releasing, or resetting resources.
        /// This method has no effect for the BlockPublishStore class. Publish store implementations that
        /// sub-class this class, should override it as needed.
        /// </summary>
        public virtual void Dispose()
        {
            // no-op
        }

        /// <summary>
        /// Stores a message in the BlockPublishStore.
        /// </summary>
        /// <param name="m"> The message to be stored. </param>
        /// <exception cref="StoreException"> If an error occurs during message storage. </exception>
        public void store(Message m)
        {
            store(m, true);
        }

        private void flattenToFreeList(Block toRemove)
        {
            Block currentInChain = toRemove;
            lock (_internalLock)
            {
                if (toRemove == _usedList)
                {
                    _usedList = _usedList.nextInList;
                }
                else
                {
                    Block prev = _usedList;
                    for (Block b = prev.nextInList; b!=null; b = b.nextInList)
                    {
                        if (b == toRemove)
                        {
                            prev.nextInList = b.nextInList;
                            break;
                        }
                        prev = prev.nextInList;
                    }
                }
                while (currentInChain != null)
                {
                    Block nextInChain = currentInChain.nextInChain;
                    try {
                        _buffer.zero(currentInChain.getPosition(),
                            Block.BLOCK_HEADER_SIZE);
                    } catch (IOException) {
                        // Going to ignore this, since we're here due to an error
                    }
                    currentInChain.sequenceNumber = 0;
                    currentInChain.nextInChain = null;
                    currentInChain.nextInList = _freeList;
                    _freeList = currentInChain;
                    currentInChain = nextInChain;
                }
            }
        }

        /// <summary>
        /// Internal utility method for storing messages in the BlockPublishStore.
        /// </summary>
        /// <param name="m"> The messge to be stored. </param>
        /// <param name="assignSequence"> A boolean flag indicating whether to assign a sequence number to
        /// the message. </param>
        /// <exception cref="StoreException"> If an error occurs during message storage. </exception>
        public void store(Message m, Boolean assignSequence)
        {
            Message.Commands operation = m.getCommand();
            int flag = -1;
            Field data = null;
            if (operation == Message.Commands.Publish ||
                operation == Message.Commands.DeltaPublish ||
                (operation == Message.Commands.SOWDelete && !m.isDataNull()))
            {
                data = m.getDataRaw();
                if (operation == Message.Commands.SOWDelete)
                    flag = SOWDeleteTypes.ByData;
            }
            else if (!m.isFilterNull())
            {
                data = m.getFilterRaw();
                flag = SOWDeleteTypes.ByFilter;
            }
            else if (!m.isSowKeysNull())
            {
                data = m.getSowKeysRaw();
                flag = SOWDeleteTypes.ByKeys;
            }
            else if (!m.isBookmarkNull())
            {
                if (m.isOptionsNull() || !m.getOptions().Contains("cancel")) {
                    data = m.getBookmarkRaw();
                    flag = SOWDeleteTypes.ByBookmark;
                } else {
                    data = m.getBookmarkRaw();
                    flag = SOWDeleteTypes.ByBookmarkCancel;
                }
            }
            Block current, next, first = null;
            long commandIdRemaining = m.getCommandIdRaw().length;
            long correlationIdRemaining = m.getCorrelationIdRaw().length;
            long expirationRemaining = m.getExpirationRaw().length;
            long sowKeyRemaining = m.getSowKeyRaw().length;
            long topicRemaining = m.getTopicRaw().length;
            long dataRemaining = data.length;
            long totalRemaining = ((operation == Message.Commands.Unknown) ? 0 :
                                  (Block.CHAIN_HEADER + commandIdRemaining +
                                  correlationIdRemaining + expirationRemaining +
                                  sowKeyRemaining + topicRemaining + dataRemaining));
            int lastBlockLength = ((operation == Message.Commands.Unknown) ? 0 :
                                  ((int)(totalRemaining % Block.BLOCK_DATA_SIZE)));
            int blocksToWrite = ((operation == Message.Commands.Unknown) ? 0 :
                                ((int)(totalRemaining / Block.BLOCK_DATA_SIZE) +
                                       (lastBlockLength > 0 ? 1 : 0)));

            long crcVal = 0;

            try
            {
                // First block gets sequence assigned
                ulong currentSequence = (assignSequence ? 0 : m.getSequence());
                first = next = get(assignSequence, currentSequence);
                if (first == null)
                {
                    throw new StoreException("The store is full, and no blocks can be allocated.");
                }
                if (assignSequence) m.setSequence(first.sequenceNumber);
                while (blocksToWrite > 0)
                {
                    current = next;
                    int bytesRemainingInBlock = Block.BLOCK_DATA_SIZE;
                    if (blocksToWrite > 1)
                    {
                        next = get(false, 0);
                        if (next == null)
                        {
                            flattenToFreeList(first);
                            throw new StoreException("The store is full, and no additional blocks can be allocated.");
                        }
                    }
                    else
                    {
                        next = null;
                    }

                    current.nextInChain = next;
                    if (_crc != null)
                    {
                        crcVal = _crc.init();
                        if (commandIdRemaining > 0)
                            crcVal = _crc.update(crcVal,
                                               m.getCommandIdRaw().buffer,
                                               (int)m.getCommandIdRaw().position,
                                               (int)commandIdRemaining);
                        if (correlationIdRemaining > 0)
                            crcVal = _crc.update(crcVal,
                                               m.getCorrelationIdRaw().buffer,
                                               (int)m.getCorrelationIdRaw().position,
                                               (int)correlationIdRemaining);
                        if (expirationRemaining > 0)
                            crcVal = _crc.update(crcVal,
                                               m.getExpirationRaw().buffer,
                                               (int)m.getExpirationRaw().position,
                                               (int)expirationRemaining);
                        if (sowKeyRemaining > 0)
                            crcVal = _crc.update(crcVal,
                                               m.getSowKeyRaw().buffer,
                                               (int)m.getSowKeyRaw().position,
                                               (int)sowKeyRemaining);
                        if (topicRemaining > 0)
                            crcVal = _crc.update(crcVal,
                                               m.getTopicRaw().buffer,
                                               (int)m.getTopicRaw().position,
                                               (int)topicRemaining);
                        if (dataRemaining > 0)
                            crcVal = _crc.update(crcVal,
                                               data.buffer,
                                               (int)data.position,
                                               (int)dataRemaining);
                        crcVal = _crc.getValue(crcVal);
                    }

                    // write the block info
                    lock (_internalLock)
                    {
                        _buffer.setPosition(current.getPosition());
                        if (current == first)
                        {
                            _buffer.putInt((int)totalRemaining);
                        }
                        else
                        {
                            _buffer.putInt(next == null ? lastBlockLength
                                                        : Block.BLOCK_DATA_SIZE);
                        }
                        _buffer.putInt(next == null ? current.index : next.index);
                        _buffer.putUlong(current == first ? m.getSequence() : 0);
                        _buffer.putLong(crcVal);

                        // now write data and topic and stuff.
                        _buffer.setPosition(current.getPosition() +
                                            Block.BLOCK_HEADER_SIZE);
                        if (current == first)
                        {
                            _buffer.putInt((int)operation);
                            _buffer.putInt((int)commandIdRemaining);
                            _buffer.putInt((int)correlationIdRemaining);
                            _buffer.putInt((int)expirationRemaining);
                            _buffer.putInt((int)sowKeyRemaining);
                            _buffer.putInt((int)topicRemaining);
                            _buffer.putInt(flag);
                            _buffer.putInt((int)m.getAckTypeOutgoing());
                            bytesRemainingInBlock -= Block.CHAIN_HEADER;
                        }
                        int bytesToWrite = (int)((commandIdRemaining < bytesRemainingInBlock) ? commandIdRemaining : bytesRemainingInBlock);
                        if (bytesToWrite > 0)
                        {
                            Field rawField = m.getCommandIdRaw();
                            _buffer.putBytes(rawField.buffer,
                                             (int)(rawField.position + rawField.length -
                                                   commandIdRemaining),
                                             bytesToWrite);
                            bytesRemainingInBlock -= (int)bytesToWrite;
                            commandIdRemaining -= bytesToWrite;
                        }
                        bytesToWrite = (int)((correlationIdRemaining < bytesRemainingInBlock) ? correlationIdRemaining : bytesRemainingInBlock);
                        if (bytesToWrite > 0)
                        {
                            Field rawField = m.getCorrelationIdRaw();
                            _buffer.putBytes(rawField.buffer,
                                             (int)(rawField.position + rawField.length -
                                                   correlationIdRemaining),
                                             bytesToWrite);
                            bytesRemainingInBlock -= (int)bytesToWrite;
                            correlationIdRemaining -= bytesToWrite;
                        }
                        bytesToWrite = (int)((expirationRemaining < bytesRemainingInBlock) ?
                                        expirationRemaining : bytesRemainingInBlock);
                        if (bytesToWrite > 0)
                        {
                            Field rawField = m.getExpirationRaw();
                            _buffer.putBytes(rawField.buffer,
                                             (int)(rawField.position + rawField.length -
                                                   expirationRemaining),
                                             bytesToWrite);
                            bytesRemainingInBlock -= (int)bytesToWrite;
                            expirationRemaining -= bytesToWrite;
                        }
                        bytesToWrite = (int)((sowKeyRemaining < bytesRemainingInBlock) ?
                                        sowKeyRemaining : bytesRemainingInBlock);
                        if (bytesToWrite > 0)
                        {
                            Field rawField = m.getSowKeyRaw();
                            _buffer.putBytes(rawField.buffer,
                                             (int)(rawField.position + rawField.length -
                                                   sowKeyRemaining),
                                             bytesToWrite);
                            bytesRemainingInBlock -= (int)bytesToWrite;
                            sowKeyRemaining -= bytesToWrite;
                        }
                        bytesToWrite = (int)((topicRemaining < bytesRemainingInBlock) ?
                                        topicRemaining : bytesRemainingInBlock);
                        if (bytesToWrite > 0)
                        {
                            Field rawField = m.getTopicRaw();
                            _buffer.putBytes(rawField.buffer,
                                             (int)(rawField.position + rawField.length -
                                                   topicRemaining),
                                             bytesToWrite);
                            bytesRemainingInBlock -= (int)bytesToWrite;
                            topicRemaining -= bytesToWrite;
                        }
                        bytesToWrite = (int)((dataRemaining < bytesRemainingInBlock) ?
                                        dataRemaining : bytesRemainingInBlock);
                        if (bytesToWrite > 0)
                        {
                            _buffer.putBytes(data.buffer,
                                             (int)(data.position + data.length -
                                                   dataRemaining),
                                             bytesToWrite);
                            bytesRemainingInBlock -= (int)bytesToWrite;
                            dataRemaining -= bytesToWrite;
                        }
                    }
                    --blocksToWrite;
                } //while(blocksToWrite > 0)

                // now, there should be no data or topic left to write.
                Debug.Assert(totalRemaining == 0 ||
                             (dataRemaining == 0 && topicRemaining == 0));
                _buffer.putInt(first.getPosition() + Block.BLOCK_READY_POSITION, 1);
                lock (_internalLock)
                {
                    Monitor.PulseAll(_internalLock);
                }
            }
            catch (IOException e)
            {
                if (first != null) flattenToFreeList(first);
                throw new StoreException(e);
            }
        }


        /// <summary>
        /// Discard messages in the store, up to and including the message
        /// specified by the passed in sequnence number. This is called by the
        /// client when it's notified a message is persisted on the server. The
        /// space used by discarded messages is made available for reuse by
        /// later messages.
        /// </summary>
        /// <param name="sequence">The highest sequence number of the messages
        ///     that can be safely discarded. The specified message and all prior
        ///     messages are removed from the store.</param>
        /// <exception cref="StoreException">If an error occurred during discarding.</exception>
        public void discardUpTo(ulong sequence)
        {
            // If we're empty or already past this value, just return
            if (sequence == 0 || _usedList == null ||
                sequence <= _metadataBlock.sequenceNumber)
            {
                lock (_internalLock)
                {
                    if (_errorOnPublishGap && sequence < _metadataBlock.sequenceNumber
                        && sequence > 0)
                    {
                        throw new PublishGapException($"Connection to current server could cause a message gap as server is only up to {sequence} but Store is already up to {_metadataBlock.sequenceNumber}");
                    }
                    try
                    {
                        if (_getLastPersisted() < sequence)
                        {
                            _metadataBlock.sequenceNumber = sequence;
                            // Position plus size of two ints
                            _buffer.putUlong(_metadataBlock.getPosition() +
                                            METADATA_LAST_DISCARDED_LOCATION,
                                            _metadataBlock.sequenceNumber);
                        }
                        if (_nextSequence <= sequence) _nextSequence = sequence + 1;
                    }
                    catch (IOException ex)
                    {
                        throw new StoreException("Error saving last discarded: " + ex);
                    }
                }
                return;
            }

            // Set the metadata block's sequence number to the specified sequence
            _metadataBlock.sequenceNumber = sequence;
            // First, find the beginning of where we should discard.
            Block almostFreeList = detach(sequence);

            // we no longer need to keep a lock.  The only funny thing about our state is that
            // there are "missing" entries -- some on _usedList, some on _freeList, and some
            // on our (local) almostFreeList, about to be freed.  We could be fancy
            // and enqueue them for freeing, not going to do that.

            // For each one of these guys on our list, flatten onto the
            // new "zeroedOutList" and remember the last entry.  We'll use
            // that to efficiently prepend, below.
            Block zeroedOutList = null;
            Block lastOnTheList = almostFreeList;
            try
            {
                // Clear out the muck in the byte buffer
                Block current = almostFreeList;

                while (current != null)
                {
                    Block next = current.nextInList;

                    Block currentInChain = current;
                    while (currentInChain != null)
                    {
                        Block nextInChain = currentInChain.nextInChain;
                        _buffer.zero(currentInChain.getPosition(),
                                     Block.BLOCK_HEADER_SIZE +
                                     Block.CHAIN_HEADER);
                        currentInChain.sequenceNumber = 0;
                        currentInChain.nextInChain = null;
                        // and, flatten this chain onto the zeroed out list.
                        currentInChain.nextInList = zeroedOutList;
                        zeroedOutList = currentInChain;
                        currentInChain = nextInChain;
                    }
                    current = next;
                }
            }
            catch (IOException ioex)
            {
                throw new StoreException(ioex);
            }

            // now zeroedOutList is flat and free.  We can just prepend it to the free list.
            lock (_internalLock)
            {
                _metadataBlock.sequenceNumber = sequence;
                try
                {
                    // Position plus size of two ints
                    _buffer.putUlong(_metadataBlock.getPosition() +
                                    METADATA_LAST_DISCARDED_LOCATION,
                                    _metadataBlock.sequenceNumber);
                }
                catch (IOException ex)
                {
                    throw new StoreException("Error saving last discarded: " + ex.ToString());
                }
                if (lastOnTheList != null)
                {
                    lastOnTheList.nextInList = _freeList;
                    _freeList = zeroedOutList;
                }
                Monitor.PulseAll(_internalLock);
            }
        }

        /// <summary>
        /// Gets the last persisted sequence number
        /// </summary>
        /// <returns> The last persisted sequence number. </returns>
        public ulong getLastPersisted()
        {
            lock (_internalLock)
            {
                return _getLastPersisted();
            }
        }

        // Lock should already be held. Gets the last persisted sequence number, in other words
        // it gets the last persisted sequence number from the metadata block.
        // <returns>The last persisted sequence number.</returns>
        private ulong _getLastPersisted()
        {
            if (_metadataBlock.sequenceNumber != 0)
            {
                return _metadataBlock.sequenceNumber;
            }
            if (_nextSequence != 1)
            {
                ulong minSeq = _getLowestUnpersisted();
                ulong maxSeq = _getHighestUnpersisted();
                if (minSeq != UInt64.MaxValue) _metadataBlock.sequenceNumber = minSeq - 1;
                if (maxSeq != UInt64.MaxValue && _nextSequence <= maxSeq)
                    _nextSequence = maxSeq + 1;
                if (_nextSequence < _metadataBlock.sequenceNumber)
                    _metadataBlock.sequenceNumber = _nextSequence - 1;
            }
            else
            {
                _metadataBlock.sequenceNumber = (ulong)((DateTime.UtcNow.Ticks - UNIX_EPOCH_OFFSET) / TimeSpan.TicksPerMillisecond) * 1000000UL;
                _nextSequence = _metadataBlock.sequenceNumber + 1;
            }
            try
            {
                // Position plus size of two ints
                _buffer.putUlong(_metadataBlock.getPosition() +
                                METADATA_LAST_DISCARDED_LOCATION,
                                _metadataBlock.sequenceNumber);
            }
            catch (IOException ex)
            {
                throw new StoreException("Error saving last discarded: " + ex);
            }
            return _metadataBlock.sequenceNumber;
        }

        /// <summary>
        /// Gets the lowest unpersisted sequence number.
        /// </summary>
        /// <returns> The lowest unpersisted sequence number or -1 if no sequence numbers are unpersisted. </returns>
        public ulong getLowestUnpersisted()
        {
            lock (_internalLock)
            {
                return _getLowestUnpersisted();
            }
        }

        // Lock should already be held. Gets the lowest unpersisted sequence number
        private ulong _getLowestUnpersisted()
        {
            if (_usedList != null)
            {
                return _usedList.sequenceNumber;
            }
            return UInt64.MaxValue;
        }

        /// <summary>
        /// Gets the highest unpersisted sequence number.
        /// </summary>
        /// <returns>The highest unpersisted sequence number.</returns>
        public ulong getHighestUnpersisted()
        {
            lock (_internalLock)
            {
                return _getHighestUnpersisted();
            }
        }

        private ulong _getHighestUnpersisted()
        {

            if (_endOfUsedList != null)
            {
                return _endOfUsedList.sequenceNumber;
            }
            return UInt64.MaxValue;
        }

        /// <summary>
        /// Used internally by the Client to give the store a Message
        /// implementing the correct protocol for the connection. This
        /// Message instance is also used for replay operations.
        /// </summary>
        /// <param name="message">A Message of the correct protocol type.</param>
        public void setMessage(Message message)
        {
            this._message = message;
        }

        private Block detach(ulong index)
        {
            Block keepList = null, endOfDiscardList = null, discardList = null;
            lock (_internalLock)
            {
                // advance through the list, find the first one to keep.
                // we keep everything after that point, and discard everything
                // up to that point.
                keepList = _usedList;
                while (keepList != null && keepList.sequenceNumber <= index)
                {
                    if (discardList == null)
                    {
                        discardList = keepList;
                    }
                    endOfDiscardList = keepList;
                    keepList = keepList.nextInList;
                }
                // detach the remainder of the list, and make it the new usedList.
                if (endOfDiscardList != null)
                {
                    endOfDiscardList.nextInList = null;
                }
                _usedList = keepList;
                if (keepList == null)
                {
                    _endOfUsedList = null;
                }
            }
            return discardList;
        }

        private void replayOnto(Block b, StoreReplayer replayer)
        {
            // Lock should already be held.
            // Clear fields we use
            // If message isn't set by a Client, then must not matter what type
            if (_message == null)
                _message = new JSONMessage(Encoding.UTF8,
                                           Encoding.UTF8);
            _message.reset();
            _buffer.setPosition(b.getPosition());
            int totalLength = _buffer.getInt() - Block.CHAIN_HEADER;

            // Can occur if replaying during disconnect or for queue ack messages
            if (totalLength <= 0)
            {
                return; // this can occur when replaying during a disconnect scenario.
            }

            // don't care about the next in chain -- we already have that.
            _buffer.getInt();
            _message.setSequence(_buffer.getUlong());
            while (_buffer.getInt(b.getPosition() + Block.BLOCK_READY_POSITION) == 0)
            {
                Monitor.Wait(_internalLock, 200);
            }
            _buffer.setPosition(b.getPosition() + Block.BLOCK_CRC_POSITION);
            long crcVal = _buffer.getLong();
            // grab the operation, lengths, flag and ack type
            _buffer.setPosition(b.getPosition() + Block.BLOCK_HEADER_SIZE);
            Message.Commands operation = (Message.Commands)_buffer.getInt();
            if (operation == Message.Commands.Unknown) return; // queue ack
            _message.setCommand(operation);
            int commandIdLength = _buffer.getInt();
            int correlationIdLength = _buffer.getInt();
            int expirationLength = _buffer.getInt();
            int sowKeyLength = _buffer.getInt();
            int topicLength = _buffer.getInt();
            int flag = _buffer.getInt();
            _message.setAckType((Message.AckTypes)_buffer.getInt());
            _buffer.setPosition(b.getPosition() + Block.BLOCK_HEADER_SIZE +
                                Block.CHAIN_HEADER);
            byte[] array = null;
            int offset = 0;
            if (b.nextInChain == null)
            {
                // everything fit into one block.  replay right out of the block memory.
                _argument.len = totalLength;
                _buffer.getBytes(_argument);
                array = _argument.array;
                offset = (int)_argument.offset;
            }
            else
            {
                // Multi-block data.
                // allocate a buffer with enough space to hold it contiguously.
                // then, copy the data from each block into the buffer
                // then call up the replayer.
                if (_readBuffer == null || _readBuffer.Length < totalLength)
                {
                    _readBuffer = new byte[totalLength];
                }

                int position = 0;
                for (Block part = b; part != null; part = part.nextInChain)
                {
                    int maxReadable = Block.BLOCK_DATA_SIZE;
                    //if this is not the first bock, skip past the header
                    if (b != part)
                    {
                        _buffer.setPosition(part.getPosition() + Block.BLOCK_HEADER_SIZE);
                    }
                    else
                    {
                        //if this is the first block, the data will include a
                        //chain header so adjust the readable size accordingly
                        maxReadable -= Block.CHAIN_HEADER;
                    }

                    int readLen = (totalLength - position) > maxReadable ?
                                    maxReadable : (totalLength - position);
                    _argument.len = readLen;
                    try
                    {
                        _buffer.getBytes(_argument);
                    }
                    catch (System.Exception e)
                    {
                        System.Console.WriteLine(e.StackTrace);
                    }

                    System.Array.Copy(_argument.array, _argument.offset,
                                      _readBuffer, position, readLen);
                    position += readLen;
                }
                //check to make sure the bytes read matches bytes expected
                if (position != totalLength)
                {
                    throw new IOException(System.String.Format("Corrupted message found.  Expected {0} bytes, read {1} bytes.", totalLength, position));
                }
                array = _readBuffer;
                offset = 0;
            }
            if (_crc != null)
            {
                //check the crc to ensure message is not corrupted
                long calcCrc = _crc.init();
                try
                {
                    calcCrc = _crc.update(calcCrc, array, offset, totalLength);
                }
                catch (System.Exception e)
                {
                    System.Console.WriteLine(e.StackTrace);
                }
                calcCrc = _crc.getValue(calcCrc);
                if (calcCrc != crcVal)
                {
                    StringBuilder sb = new StringBuilder();
                    for (Block part = b; part != null; part = part.nextInChain)
                    {
                        sb.append(part.toString());
                        sb.Append((char)'\n');
                    }
                    throw new IOException(string.Format("Corrupted multi-block message found in store. Topic Length {0}, BlockList:\n", topicLength) + sb.toString());
                }
            }
            // grab the command id
            if (commandIdLength > 0)
            {
                _message.setCommandId(array, (int)offset,
                                      commandIdLength);
                offset += commandIdLength;
                totalLength -= commandIdLength;
            }
            // grab the correlation id
            if (correlationIdLength > 0)
            {
                _message.getCorrelationIdRaw().set(array, (int)offset,
                                                   correlationIdLength);
                offset += correlationIdLength;
                totalLength -= correlationIdLength;
            }
            // grab the expiration
            if (expirationLength > 0)
            {
                _message.getExpirationRaw().set(array, (int)offset,
                                                expirationLength);
                offset += expirationLength;
                totalLength -= expirationLength;
            }
            // grab the sow key
            if (sowKeyLength > 0)
            {
                _message.setSowKey(array, (int)offset,
                                      sowKeyLength);
                offset += sowKeyLength;
                totalLength -= sowKeyLength;
            }
            // grab the topic
            if (topicLength > 0)
            {
                _message.setTopic(array, (int)offset, topicLength);
                offset += topicLength;
                totalLength -= topicLength;
            }
            if (operation == Message.Commands.Publish ||
                operation == Message.Commands.DeltaPublish)
            {
                _message.setData(array, (int)offset, totalLength);
                replayer.execute(_message);
            }
            else if (operation == Message.Commands.SOWDelete) {
                switch (flag) {
                    case SOWDeleteTypes.ByData:
                    {
                        _message.setData(array, (int)offset, totalLength);
                        replayer.execute(_message);
                    }
                    break;
                    case SOWDeleteTypes.ByFilter:
                    {
                        _message.setFilter(array, (int)offset, totalLength);
                        replayer.execute(_message);
                    }
                    break;
                    case SOWDeleteTypes.ByKeys:
                    {
                        _message.setSowKeys(array, (int)offset, totalLength);
                        replayer.execute(_message);
                    }
                    break;
                    case SOWDeleteTypes.ByBookmark:
                    {
                        _message.setBookmark(array, (int)offset, totalLength);
                        replayer.execute(_message);
                    }
                    break;
                    case SOWDeleteTypes.ByBookmarkCancel:
                    {
                        _message.setBookmark(array, (int)offset, totalLength)
                                .setOptions("cancel");
                        replayer.execute(_message);
                    }
                    break;
                    default:
                    {
                        StringBuilder sb = new StringBuilder();
                        for (Block part = b; part != null; part = part.nextInChain)
                        {
                            sb.append(part.toString());
                            sb.Append((char)'\n');
                        }
                        string message = string.Format("Corrupted message with invalid sow_delete flag found.  Sequence {0}, command {1}, flag {2}, topic length {3}, expected data length {4}, block list:\n{5}", b.sequenceNumber, _message.getCommand(), flag, topicLength, totalLength, sb.ToString());
                        throw new IOException(message);
                    }
                }
            } else {
                StringBuilder sb = new StringBuilder();
                for (Block part = b; part != null; part = part.nextInChain)
                {
                    sb.append(part.toString());
                    sb.Append((char)'\n');
                }
                string message = string.Format("Corrupted message with invalid operation found.  Sequence {0}, command {1}, flag {2}, topic length {3}, expected data length {4}, block list:\n{5}", b.sequenceNumber, _message.getCommand(), flag, topicLength, totalLength, sb.ToString());
                throw new IOException(message);
            }
        }

        /// <summary>
        /// Implemented to replay unpersisted messages from the stored blocks onto the provided StoreReplayer.
        /// </summary>
        /// <param name="replayer">The StoreReplayer used to replay undiscarded messages.</param>
        public virtual void replay(StoreReplayer replayer)
        {
            lock(_internalLock)
            {
                // Checks if the used list is empty and returns if true.
                if (_usedList == null) return;
                bool corrupted = false;
                StoreException ex = null;
                Block lastGood = null;
                Block nextInList = null;
                // for each used block
                for (Block b = _usedList; b != null; b = nextInList)
                {
                    nextInList = b.nextInList;
                    if (!corrupted)
                    {
                        try
                        {
                            // Checks if the block is not ready for replay.
                            if (_buffer.getInt(b.getPosition() +
                                               Block.BLOCK_READY_POSITION) == 0)
                            {
                                // Replay needs to stop here
                                break;
                            }

                            // Replays the block onto the provided StoreReplayer.
                            replayOnto(b, replayer);
                            lastGood = b;
                        }
                        catch (IOException ioex)
                        {
                            corrupted = true;
                            _endOfUsedList = lastGood == null ? _usedList : lastGood;
                            ex = new StoreException("Exception during replay, ignoring all following blocks", ioex);
                        }
                    }
                    // Can't just do an else here, need to deal with first corrupted block
                    if (corrupted) // We're corrupt, let's clean up bad blocks
                    {
                        Block currentInChain = b;
                        while (currentInChain != null)
                        {
                            Block nextInChain = currentInChain.nextInChain;
                            try
                            {
                                _buffer.zero(currentInChain.getPosition(),
                                             Block.BLOCK_HEADER_SIZE);
                            }
                            catch (IOException) { } // ignore
                            currentInChain.sequenceNumber = 0;
                            currentInChain.nextInChain = null;
                            // and, flatten this chain onto the _freeList
                            currentInChain.nextInList = _freeList;
                            _freeList = currentInChain;
                            currentInChain = nextInChain;
                        }
                    }
                }
                // Throws the stored exception if corruption occurred during replay.
                if (corrupted)
                {
                    throw ex;
                }
            }
        }

        /// <summary>
        /// Implemented to find the unpersisted message with the specified <paramref name="sequence"/>
        /// among the stored blocks, and replay it onto the provided StoreReplayer.
        /// Checks if the specified <paramref name="sequence"/> is less than or equal to the stored metadata block's sequence number.
        /// </summary>
        /// <param name="replayer">The StoreReplayer onto which the block will be replayed.</param>
        /// <param name="sequence">The sequence number of the block to be replayed.</param>
        /// <returns>True if the block is successfully replayed, false otherwise.</returns>
        public bool replaySingle(StoreReplayer replayer, ulong sequence)
        {
            try
            {
                lock (_internalLock)
                {
                    if (sequence <= _metadataBlock.sequenceNumber)
                    {
                        return true;
                    }
                    // find the first one that equals this index.
                    for (Block b = this._usedList; b != null; b = b.nextInList)
                    {
                        while (b.sequenceNumber == sequence
                               && (_buffer.getInt(b.getPosition()+Block.BLOCK_READY_POSITION)==0))
                        {
                            Monitor.Wait(_internalLock, 200);
                        }
                        if (b.sequenceNumber == sequence)
                        {
                            _buffer.setPosition(b.getPosition());
                            // Checks if the block has valid content for replay.
                            if (_buffer.getInt() <= 0) return false;
                            // Replays the block onto the provided StoreReplayer.
                            replayOnto(b, replayer);
                            return true;
                        }
                        else if (b.sequenceNumber >= sequence) // too far
                        {
                            break;
                        }
                    }
                    return false;
                }
            }
            catch (IOException ioex)
            {
                throw new StoreException(ioex);
            }
        }

        /// <summary>
        /// Returns the count of unpersisted messages in the store. This count represents the number
        /// of messages that have been received but not yet persisted.
        /// </summary>
        /// <returns> The count of unpersisted messages. </returns>
        public long unpersistedCount()
        {
            lock (_internalLock)
            {
                if (_usedList == null) return 0;
                return (long)(_endOfUsedList.sequenceNumber -
                              _usedList.sequenceNumber + 1);
            }
        }

        // Finds or creates a Block with the specified <paramref name="index"/>.
        // parameter index: The index to find or create a block for.
        // parameter allBlocks: A dictionary containing all existing blocks.
        // returns the found or newly created Block with the specified <paramref name="index"/>.
        Block findOrCreate(int index, Dictionary<int, Block> allBlocks)
        {
            // Checks if the dictionary <paramref name="allBlocks"/> contains a block with the specified <paramref name="index"/>.
            if (allBlocks.ContainsKey(index))
            {
                // Returns the existing block with the specified <paramref name="index"/>.
                return allBlocks[index];
            }
            else
            {
                // Creates a new Block with the specified <paramref name="index"/>.
                Block b = new Block((int)index);
                // Adds the newly created block to the dictionary <paramref name="allBlocks"/>.
                allBlocks.Add(index, b);
                // Returns the newly created block.
                return b;
            }
        }

        /// <summary>
        /// Recovers the state of the BlockPublishStore by reconstructing the used and free lists from
        /// the stored blocks in the buffer.
        /// </summary>
        protected void recover()
        {
            lock (_internalLock)
            {
                _usedList = _freeList = null;
                // Dictionary to store blocks using file index as the key
                Dictionary<int, Block> allBlocks = new Dictionary<int, Block>();   // file index -> block
                                                                                   // sequence number --> head block of chain for this message
                                                                                   // SortedDictionary to store head blocks of message chains using sequence numbers
                SortedDictionary<ulong, Block> heads = new SortedDictionary<ulong, Block>();
                try
                {
                    long end = _buffer.getSize();
                    // If the buffer is empty, no recovery is needed
                    if (end == 0) return;
                    ulong minSeq = 0, maxSeq = 0;
                    // Iterate through each block in the buffer
                    for (long offset = 0; offset < end; offset += Block.SIZE)
                    {
                        int index = (int)(offset / Block.SIZE);
                        _buffer.setPosition(offset);
                        // Read block information from the buffer
                        int size = _buffer.getInt();
                        int nextIndex = _buffer.getInt();
                        ulong sequence = _buffer.getUlong();
                        _buffer.getLong(); //ignore crcval
                        int ready = _buffer.getInt();

                        // The first block will be the metadata block
                        if (index == 0)
                        {
                            _metadataBlock = new Block(index);
                            _metadataBlock.nextInChain = null;
                            _metadataBlock.nextInList = null;
                            _metadataBlock.sequenceNumber = sequence;
                            _nextSequence = sequence + 1;
                            minSeq = _nextSequence;
                            continue;
                        }
                        Block b = findOrCreate(index, allBlocks);
                        // if there is no data (unused block), or the block index is
                        //the current index, this is the end of the chain, so don't create
                        //the next block
                        b.nextInChain = size == 0 || nextIndex == index ? null : findOrCreate(nextIndex, allBlocks);

                        // All empty blocks are free. That's the rule.
                        if (size == 0)
                        {
                            b.nextInList = _freeList;
                            _freeList = b;
                        }
                        else if (sequence != 0)
                        {
                            // Only head blocks have their sequence number set.
                            if (ready == 0)
                            {
                                // Block was not completed, add to free list.
                                Block currentInChain = b;
                                while (currentInChain != null)
                                {
                                    Block nextInChain = currentInChain.nextInChain;
                                    _buffer.zero(currentInChain.getPosition(),
                                                         Block.BLOCK_HEADER_SIZE);
                                    currentInChain.sequenceNumber = 0;
                                    currentInChain.nextInChain = null;
                                    // and, flatten this chain onto the free list
                                    currentInChain.nextInList = _freeList;
                                    _freeList = currentInChain;
                                    currentInChain = nextInChain;
                                }
                            }
                            else
                            {
                                //add it to the message heads list
                                b.sequenceNumber = sequence;
                                if (sequence < minSeq) minSeq = sequence;
                                if (sequence > maxSeq) maxSeq = sequence;
                                heads.Add(sequence, b);
                            }
                        }
                        // else, Belongs to part of a "chain"
                    }
                    // Adjust metadata block's sequence number based on recovered data
                    if (_metadataBlock.sequenceNumber < (minSeq - 1))
                        _metadataBlock.sequenceNumber = minSeq - 1;
                    // Adjust next sequence based on recovered data
                    if (_nextSequence <= maxSeq) _nextSequence = maxSeq + 1;
                    // Need to put the heads onto the list in ascending order.
                    foreach (KeyValuePair<ulong, Block> e in heads)
                    {
                        Block b = e.Value;
                        b.nextInList = null;
                        if (_endOfUsedList != null)
                        {
                            _endOfUsedList.nextInList = b;
                            _endOfUsedList = b;
                        }
                        else
                        {
                            _usedList = _endOfUsedList = b;
                        }
                    }
                }
                catch (IOException ioex)
                {
                    throw new StoreException(ioex);
                }
            }
        }

        /// <summary>
        /// Ensures that the free list has at least one block available for use.
        /// </summary>
        protected void growFreeListIfEmpty()
        {
            lock (_internalLock)
            {
                _growFreeListIfEmpty();
            }
        }

        // Internal method to grow the free list if it is empty.
        private void _growFreeListIfEmpty()
        {
            // Wait for ongoing resizing operations to complete
            while (_resizing)
            {
                Monitor.Wait(_internalLock, 50);
            }
            // If the free list is not empty, no action needed
            if (_freeList != null) return;
            try
            {
                long oldSize = _buffer.getSize();
                long newSize = oldSize + ((long)Block.SIZE * (long)_blocksPerRealloc);
                try
                {
                    _resizing = true;
                    // Wait until the buffer size is increased or resize handler denies the change
                    while (_buffer.getSize() == oldSize)
                    {
                        bool changeSize = true;
                        // Invoke the resize handler, if available
                        if (_resizeHandler != null)
                        {
                            Monitor.Exit(_internalLock);
                            try
                            {
                                changeSize = _resizeHandler.invoke(this, newSize);
                            }
                            finally
                            {
                                Monitor.Enter(_internalLock);
                            }
                        }
                        // During unlock time, something may have freed some space.
                        if (_freeList != null) return;
                        // Unless resize handler denied us, increase size
                        if (changeSize) _buffer.setSize(newSize);
                    }
                }
                finally
                {
                    _resizing = false;
                }
                // Check if the buffer size is not increased after resizing attempt
                if (_buffer.getSize() < newSize)
                    throw new StoreException("Publish store could not resize, possibly due to resize handler.");
                int idx = (int)(oldSize / Block.SIZE);
                // Initialize metadata block if it doesn't exist
                if (idx == 0)
                {
                    _metadataBlock = new Block(idx);
                    _metadataBlock.nextInChain = null;
                    _metadataBlock.nextInList = null;
                    _metadataBlock.sequenceNumber = 0;
                    // Zero out the metadata block
                    _buffer.zero(_metadataBlock.getPosition(), Block.SIZE);
                    // Set metadata version
                    _buffer.setPosition(_metadataBlock.getPosition() +
                                        METADATA_VERSION_LOCATION);
                    // Default to lowest version where first block is last discarded
                    int version;
                    try
                    {
                        version = Client.getVersionAsInt(Client.getVersion());
                    }
                    catch (CommandException)
                    {
                        // An exception will occur for develop branch
                        version = AMPS_MIN_PUB_STORE_DISCARDED_VERSION;
                    }
                    _buffer.putInt(version);
                    // Set the last discarded location
                    _buffer.setPosition(_metadataBlock.getPosition() + METADATA_LAST_DISCARDED_LOCATION);
                    _buffer.putLong(0);
                    ++idx;
                }
                // Create and add new blocks to the free list
                while (idx < (int)(newSize / Block.SIZE))
                {
                    Block b = new Block(idx);
                    _buffer.zero(b.getPosition(), Block.SIZE);
                    b.nextInChain = null;
                    b.nextInList = _freeList;
                    _freeList = b;
                    ++idx;
                }
                // Notify waiting threads that the free list is updated
                Monitor.PulseAll(_internalLock);
            }
            catch (System.SystemException ex) // Includes OverflowException, ArgumentOutOfRangeException, IOException
            {
                throw new StoreException("Publish store resizing failed, likely due to exceeding size limitations", ex);
            }
        }

        /// <summary>
        /// Wait for the store to discard everything that has been stored up to the point in time
        /// when flush is called, then return.
        /// </summary>
        public void flush()
        {
            lock (_internalLock)
            {
                // If no used blocks, no action needed
                if (_usedList == null) return;
                // Get the current sequence number
                ulong current = _endOfUsedList.sequenceNumber;
                // Wait until all used blocks are processed
                while (_usedList != null && current >= _usedList.sequenceNumber)
                {
                    Monitor.Wait(_internalLock, 50);
                }
            }
        }

        /// <summary>
        /// Wait for the store to discard everything that has been stored up to the point in time
        /// when flush is called, then return. Throw an exception if this is not completed in the
        /// number of milliseconds specified by the timeout.
        /// </summary>
        /// <param name="timeout">The number of milliseconds to wait for the flush.</param>
        /// <exception cref="TimedOutException"></exception>
        public void flush(int timeout)
        {
            // If the timeout is set to 0, call the parameterless flush method
            if (timeout == 0) flush();
            lock (_internalLock)
            {
                // If no used blocks, no action needed
                if (_usedList == null) return;
                // Get the current sequence number
                ulong current = _endOfUsedList.sequenceNumber;
                // Calculate the end time for timeout
                System.DateTime end = System.DateTime.UtcNow.AddMilliseconds((double)timeout);
                // Wait until all used blocks are processed or timeout is reached
                while (_usedList != null && current >= _usedList.sequenceNumber)
                {
                    if (System.DateTime.UtcNow.CompareTo(end) >= 0)
                        throw new TimedOutException("Timed out waiting to flush publish store.");
                    Monitor.Wait(_internalLock, timeout);
                }
            }
        }

        /// <summary>
        /// Set the resize handler for the publish store. An implementation can call the
        /// resize handler in the event that it needs to grow the size of the publish store.
        /// </summary>
        /// <param name="handler">The resize handler to set</param>
        public void setResizeHandler(PublishStoreResizeHandler handler)
        {
            _resizeHandler = handler;
        }

        /// <summary>
        /// Set if a PublishStoreGapException should be thrown by the Store if
        /// there is an attempt to call `discardUpTo` with a sequence number
        /// lower than the last persisted sequence number. This could occur
        /// during logon after a failover event if the new server was not
        /// current on messages due to being an async replication destination.
        /// </summary>
        /// <param name="errorOnPublishGap_">Indicates if the exception should be thrown</param>
        public void setErrorOnPublishGap(bool errorOnPublishGap_)
        {
            _errorOnPublishGap = errorOnPublishGap_;
        }

        /// <summary>
        /// Get if a PublishStoreGapException can be thrown by the Store if
        /// there is an attempt to call `discardUpTo` with a sequence number
        /// lower than the last persisted sequence number. This could occur
        /// during logon after a failover event if the new server was not
        /// current on messages due to being an async replication destination.
        /// </summary>
        /// <returns>True if the exception will be thrown</returns>
        public bool getErrorOnPublishGap()
        {
            return _errorOnPublishGap;
        }

        /********* PRIVATE METHODS *************/

        private Block get(bool assignSequence, ulong sequence)
        {
            Block b = null;
            // Ensure thread-safe access to the internal state
            lock (_internalLock)
            {
                // Ensure the free list is populated
                _growFreeListIfEmpty();
                // Get a block from the free list
                b = _freeList;
                // If a block is available in the free list
                if (b != null)
                {
                    // Update the free list
                    _freeList = b.nextInList;
                    b.nextInList = null;
                    // Assign a sequence number if specified
                    if (assignSequence)
                    {
                        // Make sure we start at a good value
                        if (_nextSequence <= 1 || _nextSequence == ulong.MaxValue) _getLastPersisted();
                        b.sequenceNumber = _nextSequence++;
                    }
                    else if (sequence != 0)
                    {
                        // Assign the specified sequence number
                        b.sequenceNumber = sequence;
                    }
                    // If assigned a sequence, update the used list
                    if (assignSequence || sequence != 0)
                    {
                        if (_endOfUsedList != null)
                        {
                            _endOfUsedList.nextInList = b;
                            _endOfUsedList = b;
                        }
                        else
                        {
                            // _endOfUsedList is null if _usedList is null, as well.
                            _endOfUsedList = _usedList = b;
                        }
                    }
                }
            }
            return b;
        }

        /*//  debugging function
        public string VerifyBlock(Block block_, long crc_)
        {
            long position = _buffer.getPosition();
            _buffer.setPosition(block_.getPosition());
            int size = _buffer.getInt();
            int blockIndex = _buffer.getInt();
            long haSeq = _buffer.getLong();
            long crc = _buffer.getLong();
            _buffer.setPosition(position);
            if (block_.nextInChain == null && blockIndex != block_.index)
                return string.Format("block index doesn't match, found {0} expected {1}", blockIndex, block_.index);
            if (block_.nextInChain != null && blockIndex != block_.nextInChain.index)
                return string.Format("block index doesn't match next in chain, found {0} expected {1}", blockIndex, block_.nextInChain.index);
            //            if (haSeq == 0)
//                return "seq num is 0";
            if (crc_ != -1 && crc != crc_)
                return string.Format("crc is {0} expected{1}", crc, crc_);
            return "";
        }
        */
    }
}
