////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////


using AMPS.Client.Exceptions;
using AMPS.Client.Fields;
using System;
using System.Collections.Generic;

namespace AMPS.Client
{
    /// <summary>
    /// A ring buffer of bookmarks and activation status.
    /// Used  by all of the bookmark stores to track state of bookmarks
    /// we need to hold on to, either because they're active or because
    /// they're after an active one.
    /// </summary>
    public class BookmarkRingBuffer
    {
        /// <summary>
        /// Represents a single entry in an array of bookmarks
        /// </summary>
        public class Entry
        {
            BookmarkField _bookmark;
            long _index = 0;
            bool _active = false;

            /// <summary>
            /// Initializes a new instance of the Entry class.
            /// </summary>
            public Entry()
            {
                _bookmark = new BookmarkField();
            }

            /// <summary>
            /// Sets the bookmark for the entry.
            /// </summary>
            /// <param name="f">The field to set as the bookmark.</param>
            /// <returns>This instance (i.e., the current entry).</returns>
            public Entry setBookmark(Field f)
            {
                _bookmark.copyFrom(f);
                return this;
            }

            /// <summary>
            /// Gets the bookmark field of the entry.
            /// </summary>
            /// <returns>The bookmark field of the Entry.</returns>
            public BookmarkField getBookmark()
            {
                return _bookmark;
            }

            /// <summary>
            /// Sets the index of the Entry in the Entry array.
            /// </summary>
            /// <param name="index">The index to be set.</param>
            /// <returns>The current entry.</returns>
            public Entry setIndex(long index)
            {
                _index = index;
                return this;
            }

            /// <summary>
            /// Gets the index of the Entry in the Entry array.
            /// </summary>
            /// <returns>The index of the Entry.</returns>
            public long getIndex()
            {
                return _index;
            }

            /// <summary>
            /// Gets the active value of the Entry.
            /// </summary>
            /// <returns>The active value of the Entry, True if the entry is active,
            /// false otherwise.</returns>
            public bool isActive()
            {
                return _active;
            }

            /// <summary>
            /// Sets the activity status for the entry.
            /// </summary>
            /// <param name="active">The new value of active.</param>
            /// <returns>The current entry for method chaining.</returns>
            public Entry setActive(bool active)
            {
                _active = active;
                return this;
            }

            /// <summary>
            /// Converts the Entry to a string.
            /// Returns a string representation of the entry.
            /// </summary>
            /// <returns>The converted Entry. A string containing the index, bookmark, and
            /// activity status.</returns>
            override public string ToString()
            {
                return string.Format("{0}: {1} {2}", _index, _bookmark.ToString(), (_active ? "active" : "discarded"));
            }
        }
        // Constant representing the initial count of entries.
        const int INITIAL_ENTRY_COUNT = 1000;
        /// <summary>
        /// Constant representing an unset index value.
        /// </summary>
        public const int UNSET_INDEX = -1;
        Entry[] _entries = new Entry[INITIAL_ENTRY_COUNT];
        int _start = 1, _end = 1;
        long _index = 1, _indexOfStart = 0;
        BookmarkField _lastDiscarded = new BookmarkField();
        bool _empty = true;
        BookmarkStoreResizeHandler _resizeHandler;
        BookmarkStore _store;
        Field _subId;
        bool _canResize = false;

        /// <summary>
        /// Initializes the underlying array, and sets the "last discarded" value
        /// to something reasonable.
        /// </summary>
        public BookmarkRingBuffer()
        {
            _lastDiscarded.set("0".getBytes());
            for (int i = 0; i < _entries.Length; ++i)
            {
                _entries[i] = new Entry();
            }
        }

        /// <summary>
        /// Resets this bookmark ring buffer for reuse by clearing out its state.
        /// </summary>
        public void reset()
        {
            for (int i = 0; i < _entries.Length; ++i)
            {
                _entries[i].setActive(false).setIndex(0)
                    .getBookmark().reset();
            }
            _start = 1;
            _end = 1;
            _index = 1;
            _indexOfStart = 0;
            _lastDiscarded.set("0".getBytes());
            _empty = true;
            _resizeHandler = null;
            _store = null;
            _subId = null;
            _canResize = false;
        }

        /// <summary>
        /// Size of underlying array
        /// </summary>
        /// <returns>Current capacity of self.</returns>
        public int capacity()
        {
            return _entries.Length;
        }

        /// <summary>
        /// Retrieves the entry from the bookmark ring buffer based on the specified index.
        /// </summary>
        /// <param name="index"> An index returned by .getIndex(), getStartIndex(), getEndIndex(), log, etc.</param>
        /// <returns>The Entry at the specified index, or <c>null</c> if the index is out of range.</returns>
        public Entry getByIndex(long index)
        {
            // The index must be higher than the lowest we have stored, and
            // less than our highest.
            int offsetFromStart = 0;

            // Check if the index is within the valid range.
            if (index < _indexOfStart + _start || index > _index)
            {
                return null;
            }
            // Calculate the offset from the start based on the index.
            offsetFromStart = (int)(index - _indexOfStart) % _entries.Length;
            // Return the entry at the calculated offset.
            return _entries[offsetFromStart];
        }

        /// <summary>
        /// Returns the "last discarded" bookmark. This may be a newer bookmark than the
        /// one you last passed to discard(), if you're discarding out of order.
        /// </summary>
        /// <returns>The last bookmark to be discarded.</returns>
        public BookmarkField getLastDiscarded()
        {
            return _lastDiscarded;
        }

        /// <summary>
        /// Returns if the buffer is currently empty.
        /// </summary>
        /// <returns>True if self is empty and false otherwise.</returns>
        public bool isEmpty()
        {
            return _empty;
        }

        /// <summary>
        /// Returns the index value associated with the first valid Entry in self.
        /// </summary>
        /// <returns>The start index if the ring buffer is not empty; otherwise, <see cref="UNSET_INDEX"/>.</return>
        public long getStartIndex()
        {
            if (_empty) return UNSET_INDEX;
            return _indexOfStart + _start;
        }

        /// <summary>
        /// Gets the end index (i.e., the index value one greater than the last valid Entry) of the
        /// bookmark ring buffer.
        /// </summary>
        /// <returns>index 1 greater than last valid entry.</returns>
        public long getEndIndex()
        {
            if (!_empty)
                return _index;
            return UNSET_INDEX;
        }

        /// <summary>
        /// Logs the bookmark by allocating an Entry, setting the Entry to active, copying
        /// the bookmark value to that entry, and returning the index of that entry.
        /// </summary>
        /// <param name="bookmark">The bookmark to log.</param>
        /// <returns>The index of the new Entry.</returns>
        public long log(BookmarkField bookmark)
        {
            if (_end == _start && !_empty)
            {
                if (!_canResize) return 0;
                resize();
            }
            _empty = false;
            int writeIndex = _end++;
            _entries[writeIndex].setBookmark(bookmark).setActive(true).setIndex(_index);
            if (_end == _entries.Length) _end = 0;
            return _index++;
        }


        /**
         * WARNING: Deprecated. Do Not Use. The notion of a separate "recovered
         * bookmark" section within the ring buffer and the ability to
         * re-log/re-order those entries into the main/active ring buffer
         * section is an obsolete concept that hasn't been needed since the AMPS
         * server stopped providing persisted acks for every message delivered
         * on a replay (i.e. not needed since AMPS 4.0). The bookmark ring
         * buffer has been simplified to remove this concept since a bug was
         * discovered in it (AC-953).
         *
         * Old Docs:
         *
         * Re-logs the bookmark by copying an Entry to a new location
         * and returning the index
         * of that new location.
         * @param bookmark the index to move
         * @return the index of the new Entry.
         */
        [Obsolete("This method is no longer required")]
        public long relog(long oldIndex, BookmarkField bookmark)
        {
            return oldIndex;
        }

        /// <summary>
        /// Discards an entry by index from the bookmark ring buffer. If the discard is
        /// completed, lastDiscarded will change, otherwise the discard is cached in the
        /// entry, and the getLastDiscarded() value is unchanged.
        /// </summary>
        /// <param name="index">The index of the entry.</param>
        /// <returns>True if this caused lastDiscarded to change.</returns>
        public bool discard(long index)
        {
            bool retVal = false;
            if (index < (_indexOfStart + _start) || index >= _index)
                return retVal;

            int offsetFromStart = (int)(index - _indexOfStart);
            int indexInArray = offsetFromStart % _entries.Length;

            _entries[indexInArray].setActive(false);
            if (_start == indexInArray)
            {
                Entry lastDiscardedEntry = null;
                while (!_entries[_start].isActive()
                       && !_entries[_start].getBookmark().isNull()
                       && (_start != _end || !_empty))
                {
                    if (lastDiscardedEntry != null)
                    {
                        lastDiscardedEntry.getBookmark().reset();
                    }
                    lastDiscardedEntry = _entries[_start];
                    if (++_start == _entries.Length)
                    {
                        _start = 0;
                        _indexOfStart += _entries.Length;
                    }
                    _empty = (_start == _end);
                }
                if (lastDiscardedEntry != null)
                {
                    BookmarkField bookmark = lastDiscardedEntry.getBookmark();
                    if (bookmark != null && !bookmark.isNull())
                    {
                        _lastDiscarded.copyFrom(bookmark);
                        bookmark.reset();
                    }
                    retVal = true;
                }
            }
            return retVal;
        }

        /// <summary>
        /// Finds an entry in the bookmark ring buffer with a specified bookmark field.
        /// </summary>
        /// <param name="field"> The bookmark field to search for.</param>
        /// <returns>The entry with the specified bookmark field if found; otherwise, <c>null</c></returns>
        public Entry find(BookmarkField field)
        {
            if (_empty) return null;
            long index = _indexOfStart + _start;
            int i = _start;
            do
            {
                if (!_entries[i].getBookmark().isNull() &&
                   _entries[i].getBookmark().Equals(field))
                {
                    // hooray! we found it.
                    return _entries[i].setIndex(index);
                }
                ++index;
                i = (i + 1) % _entries.Length;
            } while (i != _end);
            return null;
        }

        /// <summary>
        /// Called to check if resizing of the BookmarkRingBuffer is possible.
        /// </summary>
        public void checkResize()
        {
            int oldLength = _entries.Length;
            _canResize = (_resizeHandler == null ||
                          _resizeHandler.invoke(_store, _subId,
                                                (int)(oldLength*1.5)));
        }

        private void resize()
        {
            _canResize = false;
            int oldLength = _entries.Length;
            // allocate a new array (1.5x resize factor)
            // copy over the data.
            Entry[] newEntries = new Entry[(int)(oldLength * 1.5)];

            Array.Copy(_entries, _start, newEntries, 0, oldLength - _start);
            if (_start > 0)
            {
                Array.Copy(_entries, 0, newEntries, oldLength - _start, _start);
            }

            _indexOfStart += _start;
            _start = 0;
            _end = oldLength;
            for (int i = oldLength; i < newEntries.Length; ++i)
            {
                newEntries[i] = new Entry();
            }
            _entries = newEntries;
        }

        /// <summary>
        /// Call this when you want to set a resize handler that is invoked when a store needs a resize.
        /// </summary>
        /// <param name="handler">The handler to invoke for the resize.</param>
        /// <param name="store">The associated BookmarkStore to be resized.</param>
        public void setResizeHandler(BookmarkStoreResizeHandler handler,
                                     BookmarkStore store)
        {
            _resizeHandler = handler;
            _store = store;
        }

        /// <summary>
        /// Sets the Subscription Identifier for the bookmark.
        /// </summary>
        /// <param name="sub">The Field object to copy as the new SubId.</param>
        public void setSubId(Field sub)
        {
            _subId = sub.copy();
        }

        /// <summary>
        /// Sets the recovery state of the BookmarkRingBuffer.
        /// </summary>
        /// <returns>The recovery index.</returns>
        public long setRecovery()
        {
            return _start + _indexOfStart;
        }

        /// <summary>
        /// Gets a list of recovery entries from the BookmarkRingBuffer.
        /// </summary>
        /// <returns>A list of recovery entries.</returns>
        public IList<Entry> getRecoveryEntries()
        {
            return null;
        }

        /// <summary>
        /// Gets a list of active entries in the BookmarkRingBuffer.
        /// </summary>
        /// <param name="entryList_">The list to which active entries are added.</param>
        public void getActiveEntries(IList<Entry> entryList_)
        {
            for (int idx = _start; idx != _end; idx = (++idx % _entries.Length))
            {
                entryList_.Add(_entries[idx]);
            }
        }
    }
}
