////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using AMPS.Client.Fields;

namespace AMPS.Client
{
    /// <summary>
    /// Defines the interface for bookmark stores, which are used by the AMPS.Client.Client
    /// to provide resumable subscriptions and client-side duplicate message handling.
    /// </summary>
    public interface BookmarkStore : IDisposable
    {
        /// <summary>
        /// Called internally by the AMPS Client to log a bookmark to the persistent
        /// log.
        /// </summary>
        /// <param name="mesage">The message containing the bookmark to log.</param>
        /// <returns>The corresponding bookmark sequence number for thisbookmark.</returns>
        long log(Message mesage);

        /// <summary>
        /// Log a discard-bookmark entry to the persistent log based on a bookmark sequence
        /// number. You can call this when you want to mark the message specified by the
        /// subscription ID and the bookmark sequence number as discarded, indicating that the
        /// application has completed processing the message. Marking a message as discarded
        /// means that the message will not be replayed when the subscription resumes. Recommended
        /// to use discard(Message) instead for better performance. It is more efficient because
        /// it operates directly on a message object, eliminating the need to specify a subscription
        /// ID and bookmark sequence number separately.
        /// </summary>
        /// <param name="subId">The subscription ID of the bookmark.</param>
        /// <param name="bookmarkSeqNo">The bookmark sequence number.</param>
        void discard(Field subId, long bookmarkSeqNo);

        /// <summary>
        /// You can call this when you want to mark the provided message as discarded, indicating
        /// that the application has completed processing the message. Marking a message as discarded
        /// means the message will not be replayed when the subscription resumes which helps in
        /// preventing redundant processing of messages. For optimum performance, it is critical to
        /// discard every message once its processing is complete. If a message is never discarded,
        /// it remains in the bookmark store. Log a discard-bookmark entry to the persistent log
        /// based on a Message.
        /// </summary>
        /// <param name="message">Message to be marked as discarded.</param>
        void discard(Message message);

        /// <summary>
        /// You can call this when you want to return the correct recovery point for resubscription
        /// of the provided subID from the BookmarkStore. This is a single bookmark or (more often) a
        /// comma-delimited list of bookmarks.  The useList parameter is a boolean flag. When set to
        /// true (which is the default), it affects how the recovery point is determined when the
        /// store hasn't received any persisted acknowledgments (acks). Specifically, if set to true,
        /// the method will build a list of bookmarks based on the last discarded message for each
        /// publisher. If set to false, it won't consider this additional information. This option
        /// can be useful when persisted acks are not available, allowing you to make more informed
        /// decisions about where to resume message processing.
        /// </summary>
        /// <param name="subId">The subscription ID of the most recent bookmark.</param>
        /// <param name="useList">In the case where the store has not received any persisted acks,
        /// it will build a list of bookmarks based on the last discarded for each publisher if
        /// true (default).</param>
        /// <returns>Returns the most recent bookmark from the log that ought to be used for
        /// (re-)subscriptions.</returns>
        Field getMostRecent(Field subId, bool useList = true);

        /// <summary>
        /// Called internally by the AMPS Client. This method is called for each arriving message
        /// to determine if the application has already processed and discarded this message.
        /// This step is crucial to avoid processing the same message multiple times.Returns 'true'
        /// if the bookmark is in the log and marked as discarded and should therefore not be
        /// reprocessed.Otherwise, returns 'false' indicating that the message is new or hasn't
        /// been marked as discarded yet. Generally, isDiscarded is called by the AMPS Client
        /// however, there might be situations where the application needs to manually check if
        /// a message has been discarded.In such cases, the application can call this method to
        /// make that determination.
        /// </summary>
        /// <param name="message">Incoming message used to determine if the application has already
        /// processed and discarded the message.</param>
        /// <returns>Whether the message is already in the store and marked as discarded.</returns>
        bool isDiscarded(Message message);

        /// <summary>
        /// You can call this when you want to purge the contents of this store for a given subId.
        /// Removes any tracking history associated with publishers and received messages, and may
        /// delete or truncate on-disk representations as well.  NOTE: This should only be called
        /// when there are no ongoing or active subscriptions(i.e.before any subscribe or after all
        /// subscriptions are unsubscribed). Purging while subscriptions are active can disrupt
        /// those processes.
        /// </summary>
        void purge(Field subId);

        /// <summary>
        /// You can call this when you want to purge the contents of this store.
        /// Removes any tracking history associated with publishers and received messages, and may
        /// delete or truncate on-disk representations as well.  NOTE: This should only be called
        /// when there are no ongoing or active subscriptions(i.e.before any subscribe or after all
        /// subscriptions are unsubscribed). Purging while subscriptions are active can disrupt
        /// those processes.
        /// </summary>
        void purge();

        /// <summary>
        /// You can call this when you want to set a resize handler that is invoked when the store
        /// needs to resize.This method allows you to set a custom function that gets called when
        /// a bookmark store needs to resize, giving you control and flexibility over how your
        /// application responds to this event.
        /// </summary>
        /// <param name="handler">The handler to invoke for the resize.</param>
        void setResizeHandler(BookmarkStoreResizeHandler handler);

        /// <summary>
        /// You can call this when you want to retrieve the sequence number of the oldest bookmark
        /// in the store.
        /// </summary>
        /// <param name="subId">The subscription ID for the oldest bookmark in the store.</param>
        /// <returns>The internal sequence number of the oldest bookmark in the store.</returns>
        long getOldestBookmarkSeq(Field subId);

        /// <summary>
        /// Called internally by the AMPS Client to record the last persisted message in the
        /// transaction log of the connected AMPS server to all of its sync replication destinations.
        /// </summary>
        /// <param name="subId">The subscription ID for the message.</param>
        /// <param name="bookmark">The bookmark containing the message.</param>
        void persisted(Field subId, BookmarkField bookmark);

        /// <summary>
        /// Old style of setting a persisted bookmark no longer used.
        /// </summary>
        /// <param name="subId">The subscription ID for the message.</param>
        /// <param name="bookmarkSeqNo">The bookmark sequence number.</param>
        void persisted(Field subId, long bookmarkSeqNo);

        /// <summary>
        /// Called internally by the AMPS Client when connected to an AMPS server to indicate what
        /// version the server is. A bookmark store can modify behavior based on the version of the
        /// server.
        /// </summary>
        /// <param name="version">The version of the AMPS server.</param>
        void setServerVersion(int version);
    }
}

