////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

/// <summary>
using AMPS.Client.Fields;
using System;
using System.Globalization;
using System.Text;
using AMPS.Client.Exceptions;

namespace AMPS.Client
{
    /// <summary>
    /// Command is an encapsulation of a single AMPS command sent by the client.
    /// Using Command you can build  valid commands to be executed either
    /// synchronously or asynchronously via the {@link Client} execute() method
    /// or asynchronously via the executeAsync() methods.
    /// </summary>
    /// Command is designed to be used as a "builder" enabling AMPS commands
    /// to be built easily, for example:
    ///
    /// <pre>
    /// <code>
    ///
    /// Client client = new Client(...);
    /// for(Message m : client.execute(new Command("sow").setTopic("topic"))) { ... }
    ///
    /// </code>
    /// </pre>
    public class Command
    {
        CommandId CommandId, QueryId, SubId;
        Message.Commands CommandType;
        Message.AckTypes AckType;
        string Filter, OrderBy, Bookmark, Options, SOWKeys, SOWKey, CorrelationId;
        internal byte[] Topic, Data;
        string TopicString, DataString;
        internal int _topicLength, _topicOffset, _dataLength, _dataOffset;
        long Timeout = 0;
        int TopN = -1, BatchSize = 10;
        int Expiration = 0;
        bool _expiration_isSet = false;
        ulong ClientSequenceNumber = 0;
        internal Message _message = null;

        private void init()
        {
            CommandId = QueryId = SubId = null;
            Filter = OrderBy = Bookmark = Options = SOWKeys = SOWKey = CorrelationId = null;
            Topic = Data = null;
            TopicString = DataString = null;
            AckType = Message.AckTypes.None;
            TopN = -1;
            Expiration = 0;
            BatchSize = 10;
            Timeout = 0;
            ClientSequenceNumber = 0;
            _expiration_isSet = false;

            if ((CommandType & Message.Commands.NoDataCommands)
                 == Message.Commands.Unknown
                || CommandType == Message.Commands.SOWDelete)
            {
                CommandId = CommandId.nextIdentifier();
                if (isSow())
                {
                    QueryId = CommandId;
                }
            }
        }

        /// <returns>The SOW Keys.</returns>
        [Obsolete("Use getSowKeys() instead")]
        public string getSOWKeys()
        {
            return SOWKeys;
        }

        /// <summary>
        /// Returns the SOW keys on this command.
        /// </summary>
        /// <returns>The SOW Keys.</returns>
        public string getSowKeys()
        {
            return SOWKeys;
        }

        /// <summary>
        /// The SowKeys for a command are a comma-separated list
        /// of the keys that AMPS assigns to SOW messages. The SOW key for a
        /// message is available through the {@link Message#getSowKeys} method
        /// on a message.
        /// </summary>
        /// <param name="sowKeys">The sow keys to set.</param>
        [Obsolete("Use setSowKeys() instead")]
        public Command setSOWKeys(string sowKeys)
        {
            SOWKeys = sowKeys;
            return this;
        }

        /// <summary>
        /// The SowKeys for a command are a comma-separated list
        /// of the keys that AMPS assigns to SOW messages. The SOW key for a
        /// message is available through the {@link Message#getSowKeys} method
        /// on a message.
        /// </summary>
        /// <param name="sowKeys">The sow keys to set.</param>
        public Command setSowKeys(string sowKeys)
        {
            SOWKeys = sowKeys;
            return this;
        }

        /// <returns>the SOW Key</returns>
        [Obsolete("Use getSowKey() instead")]
        public string getSOWKey()
        {
            return SOWKey;
        }

        /// <summary>
        /// Retrieves the State of the World (SOW) key associated with this instance.
        /// </summary>
        /// <returns>The SOW Key.</returns>
        public string getSowKey()
        {
            return SOWKey;
        }

        /// <summary>
        /// The SowKey for a command are a comma-separated list
        /// of the keys that AMPS assigns to SOW messages. The SOW key for a
        /// message is available through the {@link Message#getSowKey} method
        /// on a message.
        /// </summary>
        /// <param name="sowKey">The sow keys to set.</param>
        [Obsolete("Use setSowKey() instead")]
        public Command setSOWKey(string sowKey)
        {
            SOWKey = sowKey;
            return this;
        }
        /// <summary>
        /// The SowKey for a command are a comma-separated list
        /// of the keys that AMPS assigns to SOW messages. The SOW key for a
        /// message is available through the {@link Message#getSowKey} method
        /// on a message.
        /// </summary>
        /// <param name="sowKey">The sow keys to set.</param>
        /// <returns>Returns this instance so that various operations can be
        /// chained together.</returns>
        public Command setSowKey(string sowKey)
        {
            SOWKey = sowKey;
            return this;
        }

        /// <summary>
        /// Checks if the current command represents a subscription.
        /// </summary>
        /// <returns>True if this command is a subscription command, otherwise
        /// false.</returns>
        public bool isSubscribe()
        {
            return CommandType == Message.Commands.DeltaSubscribe ||
                   CommandType == Message.Commands.SOWAndDeltaSubscribe ||
                   CommandType == Message.Commands.SOWAndSubscribe ||
                   CommandType == Message.Commands.Subscribe;
        }

        /// <summary>
        /// Checks if the current command represents a State of the World (SOW)
        /// operation.
        /// </summary>
        /// <returns>True if this command is a sow query, otherwise false.</returns>
        public bool isSow()
        {
            return CommandType == Message.Commands.SOW || CommandType == Message.Commands.SOWAndDeltaSubscribe ||
                    CommandType == Message.Commands.SOWAndSubscribe;
        }

        /// <summary>
        /// Checks whether a sequence number is needed for this message based on the message
        /// command.
        /// </summary>
        /// <returns>True if this command can be associated with a client sequence number,
        /// otherwise false.</returns>
        public bool needsSequenceNumber()
        {
            return CommandType == Message.Commands.Publish ||
                   CommandType == Message.Commands.DeltaPublish ||
                   CommandType == Message.Commands.SOWDelete;
        }

        // This method converts the input command into bytes using ISO-8859-1 encoding,
        // checks its validity, and returns an integer representation of the command.
        // If the command is unknown, it raises an IllegalArgumentException.
        private Message.Commands translateCommand(string command_)
        {
            CommandField cf = new CommandField();
            cf.set(command_.getBytes());
            if (CommandField.encodeCommand(cf.getValue()) != command_)
            {

                throw new System.ArgumentException("unknown command type " + command_);
            }
            return cf.getValue();
        }
        /// <summary>
        /// Create a new Command with no command type or arguments set.
        /// Provides a starting point for creating 'Command' objects.
        /// </summary>
        public Command()
        {
        }

        /// <summary>
        /// Initializes a new instance of the Command class with the specified AMPS command type.
        /// </summary>
        /// <param name="command_">The <see cref="Message.Commands"/> value indicating the AMPS command.</param>
        public Command(Message.Commands command_)
        {
            CommandType = command_;
            init();
        }

        /// <summary>
        /// Create a Command with the Command field set.
        /// </summary>
        /// <param name="command_">A string indicating the AMPS command.</param>
        public Command(string command_)
        {
            CommandType = translateCommand(command_);
            init();
        }

        /// <summary>
        /// Resets this command with a new Command type and re-initializes all
        /// other fields.
        /// </summary>
        /// <param name="command_">A {@link Message.Commands} value indicating
        /// the AMPS command.</param>
        public Command reset(Message.Commands command_)
        {
            CommandType = command_;
            init();
            return this;
        }

        /// <summary>
        /// Resets this command with a new Command type and re-initializes all other fields.
        /// </summary>
        /// <param name="command_">A string value indicating the AMPS command.</param>
        public Command reset(string command_)
        {
            CommandType = translateCommand(command_);
            init();
            return this;
        }

        /// <summary>
        /// Gets the automatically generated command ID associated with this command.
        /// </summary>
        /// <returns>The commandId, which is automatically generated for some command types.</returns>
        public CommandId getCommandId()
        {
            return CommandId;
        }

        /// <summary>
        /// Sets the command ID for this command.
        /// </summary>
        /// <param name="commandId">The commandId to set.</param>
        /// <returns>This `Command` object with the specified identifier.</returns>
        public Command setCommandId(CommandId commandId)
        {
            CommandId = commandId;
            return this;
        }

        /// <summary>
        /// Retrieve the query identifier associated with the 'Command' object. This
        /// method is used to access the identifier set by the 'setQueryId' method when
        /// needed. It allows you to read or obtain the QueryId for tracking, monitoring,
        /// or correlating messages within the messaging system.
        /// </summary>
        /// <returns>The query identifier ('CommandId') previously set using the 'setQueryId'
        /// method, or 'null' if not set.</returns>
        public CommandId getQueryId()
        {
            return QueryId;
        }

        /// <summary>
        /// Set a custom QueryId for this Command. The QueryId is a unique identifier that
        /// you can assign to this command. You have the option to set this independently
        /// of the Command so customizing the QueryId provides flexibility as you can assign
        /// QueryIds based on application-specific logic. When the QueryId is independent of
        /// the Command in AMPS, it means you are not bound by the CommandId for assigning
        /// QueryIds. Instead, you have the freedom to create custom identifiers for queries
        /// and messages, which can be valuable when you have scenarios when you need to
        /// correlate responses from different types of queries based on shared attributes
        /// other than the CommandId.
        /// </summary>
        /// <param name="queryId">A unique query identifier to be associated with this
        /// `Command`.</param>
        /// <returns>This `Command` object with the specified query identifier.</returns>
        public Command setQueryId(CommandId queryId)
        {
            QueryId = queryId;
            return this;
        }

        /// <summary>
        /// Gets the AMPS command type associated with this command.
        /// </summary>
        /// <returns>The command type.</returns>
        public Message.Commands getCommand()
        {
            return CommandType;
        }

        /// <summary>
        /// Sets the AMPS command type for this command.
        /// </summary>
        /// <param name="command">The command type to set.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setCommand(Message.Commands command)
        {
            CommandType = command;
            return this;
        }

        /// <summary>
        /// Gets the topic associated with this command.
        /// </summary>
        /// <returns>The topic.</returns>
        public string getTopic()
        {
            return Topic != null ? Encoding.UTF8.GetString(Topic, _topicOffset, _topicLength) : TopicString;
        }

        /// <summary>
        /// Sets the topic for this command using a string representation.
        /// </summary>
        /// <param name="topic">The topic to set.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setTopic(string topic)
        {
            TopicString = topic;
            Topic = null;
            return this;
        }

        /// <summary>
        /// Sets the topic for this command using raw bytes, offset, and length.
        /// </summary>
        /// <param name="topic">The raw bytes to be used for the topic.</param>
        /// <param name="offset">The offset into topic where the topic data begins.</param>
        /// <param name="length">The length of the topic.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setTopic(byte[] topic, int offset, int length)
        {
            Topic = topic;
            TopicString = null;
            _topicLength = length;
            _topicOffset = offset;
            return this;
        }

        /// <summary>
        /// Gets the filter associated with this command.
        /// </summary>
        /// <returns>The filter.</returns>
        public string getFilter()
        {
            return Filter;
        }

        /// <summary>
        /// Sets the filter for this command.
        /// </summary>
        /// <param name="filter">The filter to set.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setFilter(string filter)
        {
            Filter = filter;
            return this;
        }

        /// <summary>
        /// Gets the orderBy associated with this command.
        /// </summary>
        /// <returns>The orderBy.</returns>
        public string getOrderBy()
        {
            return OrderBy;
        }

        /// <summary>
        /// Sets the orderBy for this command.
        /// </summary>
        /// <param name="orderBy">The orderBy to set.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setOrderBy(string orderBy)
        {
            OrderBy = orderBy;
            return this;
        }

        /// <summary>
        /// Get the subId set on this command.
        /// </summary>
        /// <returns>The subId.</returns>
        public CommandId getSubId()
        {
            return SubId;
        }

        /// <summary>
        /// Sets the subId for this command.
        /// </summary>
        /// <param name="subId">The subId to set.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setSubId(CommandId subId)
        {
            SubId = subId;
            return this;
        }

        /// <summary>
        /// Sets the subId for this command using a string representation.
        /// </summary>
        /// <param name="subId">The subId to set.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setSubId(string subId)
        {
            if (!String.IsNullOrEmpty(subId))
            {
                byte[] bytes = Encoding.UTF8.GetBytes(subId);
                CommandId cid = new CommandId();
                cid.set(bytes, 0, bytes.Length);
                SubId = cid;
            }
            return this;
        }

        /// <summary>
        /// Gets the bookmark associated with this command.
        /// </summary>
        /// <returns>The bookmark as a string.</returns>
        public string getBookmark()
        {
            return Bookmark;
        }

        /// <summary>
        /// Sets the bookmark for this command.
        /// </summary>
        /// <param name="bookmark">The bookmark to set.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setBookmark(string bookmark)
        {
            Bookmark = bookmark;
            return this;
        }

        /// <summary>
        /// Gets the options associated with this command.
        /// </summary>
        /// <returns>The options.</returns>
        public string getOptions()
        {
            return Options;
        }

        /// <summary>
        /// Set the options for this command. Options are a comma-delimited list.
        /// </summary>
        /// <param name="options">The options to set.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setOptions(string options)
        {
            Options = options;
            return this;
        }

        /// <summary>
        /// Gets the ackType associated with this command.
        /// </summary>
        /// <returns>The ackType.</returns>
        public Message.AckTypes getAckType()
        {
            return AckType;
        }

        /// <summary>
        /// Sets the ackType for this command.  Different ackTypes are supported
        /// for each command. See the Command Reference for details.
        /// </summary>
        /// <param name="ackType">The ackType to set.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setAckType(Message.AckTypes ackType)
        {
            if (ackType == Message.AckTypes.None
                || ackType == Message.AckTypes.Persisted)
            {
                if (CommandId != null
                    && (CommandType & Message.Commands.NoDataCommands)
                        != Message.Commands.Unknown)
                {
                    CommandId = null;
                }
            }
            else if (CommandId == null)
            {
                // ackType is one where an ack will be returned
                CommandId = CommandId.nextIdentifier();
            }
            AckType = ackType;
            return this;
        }

        /// <summary>
        /// Adds an additional ackType to the ackTypes already set
        /// for this Command.
        /// </summary>
        /// <param name="ackType"> the ackType to add </param>
        public Command addAckType(Message.AckTypes ackType)
        {
            AckType |= ackType;
            if (CommandId == null
                && AckType != Message.AckTypes.Persisted)
            {
                // ackType is one where an ack will be returned
                CommandId = CommandId.nextIdentifier();
            }
            return this;
        }

        /// <summary>
        /// Gets the data associated with this command.
        /// </summary>
        /// <returns>The data.</returns>
        public string getData()
        {
            return Data != null ? Encoding.UTF8.GetString(Data, _dataOffset, _dataLength) : DataString;
        }

        /// <summary>
        /// Sets the data for this command.
        /// </summary>
        /// <param name="data">The data to set.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setData(string data)
        {
            DataString = data;
            Data = null;
            return this;
        }

        /// <summary>
        /// Sets the raw data for this command.
        /// </summary>
        /// <param name="data">The raw bytes to be used for the data.</param>
        /// <param name="offset">The offset into data where the data begins.</param>
        /// <param name="length">The length of the data.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setData(byte[] data, int offset, int length)
        {
            Data = data;
            DataString = null;
            _dataOffset = offset;
            _dataLength = length;
            return this;
        }



        /// <summary>
        /// Set the data for this command.
        /// </summary>
        /// <param name="builder">A CompositeMessageBuilder to set self's data from.</param>
        public Command setData(CompositeMessageBuilder builder)
        {
            DataString = null;
            Data = builder._bytes;
            _dataOffset = 0;
            _dataLength = builder._position;
            return this;
        }

        /// <summary>
        /// Gets the timeout associated with this command.
        /// </summary>
        /// <returns>The timeout value.</returns>
        public long getTimeout()
        {
            return Timeout;
        }

        /// <summary>
        /// Set the timeout for this command (in milliseconds). If the client does
        /// not receive and consume a processed ack within the specified timeout,
        /// this method throws an exception. Notice that the timeout is monitored
        /// in the AMPS client, not on the server, and the client receive thread
        /// receives and consumes the ack from the server.
        /// </summary>
        /// <param name="timeout">The timeout to set.</param>
        /// <returns>The current instance of <see cref="Command"/>.</returns>
        public Command setTimeout(long timeout)
        {
            Timeout = timeout;
            return this;
        }

        /// <summary>
        /// Get the topN value for this command.
        /// <returns>The topN.</returns>
        public int getTopN()
        {
            return TopN;
        }

        /// <summary>
        /// Sets the top_n header for this command. This parameter sets a maximum number
        /// of records returned by a SOW query. This should usually be set as part of the
        /// command's options string, rather than using this method.
        /// <param name="topN"> the topN to set</param>
        /// <returns>The current instance of Command.</returns>
        public Command setTopN(int topN)
        {
            TopN = topN;
            return this;
        }

        /// <summary>
        /// Get the batchSize set on this command.
        /// </summary>
        /// <returns>The batchSize.</returns>
        public int getBatchSize()
        {
            return BatchSize;
        }

        /// <summary>
        /// Set the batchSize for this command. This sets the number of records
        /// returned in a batch during a SOW query.
        /// </summary>
        /// <param name="batchSize">The batchSize to set.</param>
        /// <returns>The current instance of Command.</returns>
        public Command setBatchSize(int batchSize)
        {
            BatchSize = batchSize;
            return this;
        }

        /// <summary>
        /// Get the expiration set on this command. This method returns 0 if
        /// no expiration is set. Use {@link Command#hasExpiration()}to check
        /// if an expiration is set.
        /// </summary>
        /// <returns>The expiration.</returns>
        public int getExpiration()
        {
            if (!_expiration_isSet) return 0;
            return Expiration;
        }

        /// <summary>
        /// Returns true if this command has an expiration set.
        /// </summary>
        /// <returns>True if an expiration is set.</returns>
        public bool hasExpiration()
        {
            return _expiration_isSet;
        }

        /// <summary>
        /// Set the expiration for this command. For a publish command to a
        /// SOW topic or a queue, the expiration sets the amount of time to
        /// retain the message in the SOW or the queue.
        /// </summary>
        /// <param name="expiration">The expiration to set.</param>
        /// <returns>The current instance of Command.</returns>
        public Command setExpiration(int expiration)
        {
            Expiration = expiration;
            _expiration_isSet = true;
            return this;
        }

        /// <summary>
        /// Clears any expiration value set on self.
        /// </summary>
        public void unsetExpiration()
        {
            _expiration_isSet = false;
        }

        /// <summary>
        /// Binds self to a given client, preparing a message from that client to be sent.
        /// </summary>
        /// <param name="client_"> The client to bind against</param>
        /// <returns> The CommandId for this command, if one is set, or null.</returns>
        internal CommandId prepare(Client client_)
        {
            if (_message == null)
            {
                _message = client_.allocateMessage();
            }
            else
            {
                _message.reset();
            }

            _message.setCommand(CommandType);
            if (CommandId != null) _message.setCommandId(CommandId);
            _message.setAckType(AckType);
            if (CorrelationId != null) _message.setCorrelationId(CorrelationId);
            if (Bookmark != null) _message.setBookmark(Bookmark);
            if (Data != null)
                _message.setData(Data, _dataOffset, _dataLength);
            else if(DataString != null)
                _message.setData(DataString);
            if (Filter != null) _message.setFilter(Filter);
            if (Options != null) _message.setOptions(Options);
            if (OrderBy != null) _message.setOrderBy(OrderBy);
            if (QueryId != null) _message.setQueryId(QueryId);

            if (isSubscribe() && SubId==null)
            {
                // user validation: if you've supplied "replace" as an option but no subid, watch out.
                if (Options != null && Options.Contains("replace"))
                {
                    throw new CommandException("Cannot issue a replacement subscription; a valid subscription id is required.");
                }
                SubId = CommandId;
            }
            if (SubId != null) _message.setSubId(SubId);
            if (Topic != null)
                _message.setTopic(Topic, _topicOffset, _topicLength);
            else if (TopicString != null)
                _message.setTopic(TopicString);
            if (_expiration_isSet) _message.setExpiration(Expiration);
            if (SOWKeys != null) _message.setSowKeys(SOWKeys);
            if (SOWKey != null) _message.setSowKey(SOWKey);
            if (TopN != -1) _message.setTopN(TopN);
            if (BatchSize > 1
                &&  (CommandType  == Message.Commands.SOW
                    || CommandType == Message.Commands.SOWAndSubscribe
                    || CommandType == Message.Commands.SOWAndDeltaSubscribe))
            {
                _message.setBatchSize(BatchSize);
            }
            if (needsSequenceNumber() && ClientSequenceNumber > 0)
                _message.setSequence(ClientSequenceNumber);
            return CommandId;
        }

        /// <summary>
        /// Sets the client sequence number for this command.
        /// </summary>
        /// <param name="seqNumber">Sequence number for this command.</param>
        /// <returns>The current instance of Command.</returns>
        public Command setClientSequenceNumber(ulong seqNumber)
        {
            ClientSequenceNumber = seqNumber;
            return this;
        }

        /// <summary>
        /// Get the client sequence number set on this command.
        /// </summary>
        /// <returns>The client sequence number generated for this command, or 0 if none
        /// was generated.</returns>
        public ulong getClientSequenceNumber()
        {
            return ClientSequenceNumber;
        }

        /// <summary>
        /// Sets the correlation Id for this command. The correlation ID
        /// is an arbitrary string that can be used for whatever purpose the
        /// application needs. It is not interpreted or used by AMPS. However,
        /// the correlation ID must only contain characters that are valid base64
        /// encoded characters.
        /// </summary>
        /// <param name="correlationId_">The CorrelationId is a user provided string included in
        /// the log message for a logon. AMPS does not interpret or use this string for any other
        /// purpose.</param>
        /// <returns>The current instance of Command.</returns>
        public Command setCorrelationId(String correlationId_)
        {
            CorrelationId = correlationId_;
            return this;
        }

        /// <summary>
        /// Gets the correlation ID set with this command.
        /// </summary>
        /// <returns>The correlation id.</returns>
        public String getCorrelationId()
        {
            return CorrelationId;
        }
    }
}
