////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2012-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System.Text;
using System;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading;

namespace AMPS.Client
{

    /// <summary>
    /// Used to give a unique identifier to a Command.
    /// This Id can be used to reference the command later,
    /// such as when unsubscribing a subscription command.
    /// </summary>
    public class CommandId
    {
        static long currentCommandId = 0;
        public byte[] id = null;
        private int hash = 0;

        /// <summary>
        /// Creates and initializes a new, unique CommandId
        /// </summary>
        /// <returns>A new CommandId</returns>
        public static CommandId nextIdentifier()
        {
            long newCommandId = Interlocked.Increment(ref currentCommandId);

            byte[] id = Encoding.ASCII.GetBytes(string.Format("auto{0}", newCommandId));

            return new CommandId(id);
        }

        public CommandId()
        {
        }

        /// <summary>
        /// Constructs a CommandId for an AMPS command, using a string as its identifier.
        /// </summary>
        /// <param name = "value"> A string used to set the new identifier for this AMPS command. </param>
        public CommandId(String value)
        {
            if(value != null)
            {
                this.id = value.getBytes();
                UpdateHash();
            }
        }

       /// <summary>
       /// Constructs a CommandId for an AMPS command, using an array of bytes as its identifier.
       /// </summary>
       /// <param name="buffer"> An array of bytes used to set the new identifier for this AMPS
       /// command. </param>
        public CommandId(byte[] buffer)
        {
            this.id = buffer;
            UpdateHash();
        }

       /// <summary>
       /// Constructs a CommandId for an AMPS command, using an array of bytes, a position number,
       /// and a length as its identifier.
       /// </summary>
       /// <param name = "buffer"> An array of bytes used to set the new identifier for this AMPS
       /// command. </param>
       /// <param name = "position"> An integer used to represent the position that the buffer should
       /// start. </param>
       /// <param name = "length"> An integer used to determine the length of the new identifier. </param>
        public CommandId(byte[] buffer, int position, int length)
        {
            this.id = new byte[length];
            Array.Copy(buffer, position, id, 0, length);
            UpdateHash();
        }
        internal static int GetHashCode(byte[] id, int pos, int len)
        {
            int hash = 0;
            for (int i = 0; i < len; i++)
            {
                if (id[pos+i] == (byte)' ') return hash;
                //hash *= 10;
                hash += id[pos+i];
            }
            Debug.Assert(hash >= 0);

            return hash;
        }
        private void UpdateHash()
        {
            hash = GetHashCode(this.id, 0, this.id.Length);
        }

        /// <summary>
        /// Replaces the current id with the supplied identifier.
        /// </summary>
        /// <param name="buffer">A set of ASCII bytes corresponding to the Base64-encoded UUID.  This should always be 24 bytes long.</param>
        /// <param name="position">The position in <paramref name="buffer"/>buffer</param> where the UUID begins.
        /// <param name="length">The length of the UUID beginning at <paramref name="position"/>.  The behavior is undefined if this value is not 24.</param>
        public void set(byte[] buffer, int position, int length)
        {
            if (this.id == null || this.id.Length != length) this.id = new byte[length];
            Array.Copy(buffer, position, this.id, 0, length);
            UpdateHash();
        }

        /// <summary>
        /// Return a copy of self with its own buffer.
        /// </summary>
        public CommandId copy()
        {
            if (this.id == null)
                return new CommandId();
            else
                return new CommandId(this.id, 0, this.id.Length);
        }

        /// <summary>
        /// Allows CommandId's equality semantics to be applied to two byte arrays.
        /// </summary>
        internal static bool ByteEquals(byte[] id1, byte[] id2, int ofs, int len)
        {
            if(id1==null || id1.Length != len) return false;

            /* This is faster for integers, but we're using GUIDs now.
            for (int i = 0; i < StorageLength; i++)
            {
                if (id1[i] != id2[i+ofs]) return false;
                if (id1[i] == ' ') return true;

            }
            return true;
            */
            int length = id1.Length;
            unsafe
            {
                fixed(byte* b1 = id1)
                {
                    fixed (byte* b2 = &id2[ofs])
                    {

                        for (int i = 0; i < length; ++i )
                        {
                            if (b1[i] != b2[i]) return false;
                        }
                        return true;
                    }
                }
            }
        }

       /// <summary>
       /// Overridden to test equality based upon object class and the CommandId value. For this to
       /// return true, the passed in instance must be an instance of CommandId and its value must be
       /// binary equal to this instance's value.
       /// </summary>
        public override bool Equals(object obj)
        {
            byte[] rhs = ((CommandId)obj).id;
            return rhs!=null && ByteEquals(this.id, rhs,0,rhs.Length);
        }

        /// <summary>
        /// Overridden to return the hash code that is based upon CommandId value.
        /// </summary>
        public override int GetHashCode()
        {
            return hash;
        }

        /// <summary>
        /// Overridden to return the CommandId value as a string where each byte is interpreted as
        /// an ISO-8859-1 character.
        /// </summary>
        public override string ToString()
        {
            return Encoding.ASCII.GetString(id);
        }

    }
}

