///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading;
using AMPS.Client.Fields;
using AMPS.Client.Exceptions;


namespace AMPS.Client
{
    /// <summary>
    /// An implementation of the {@link RecoveryPointAdapter} interface
    /// that is meant to act as a wrapper around another recovery point
    /// adapter instance, delegating periodic updates to its wrapped
    /// adapter on an asynchronous update thread (which this instance starts).
    /// The intent is to decouple recovery state updates from message discards
    /// on a bookmark replay subscriptions so that fast paced subscriptions
    /// don't cause a rate of updates that consume too much bandwidth or
    /// overwhelm recovery point adapter implementations that use slower
    /// storage (slowing message discards).
    /// </summary>
    public class ConflatingRecoveryPointAdapter : RecoveryPointAdapter
    {
        /// <summary>
        /// The wrapped recovery point adapter instance we delegate to.
        /// </summary>
        protected RecoveryPointAdapter _adapter = null;

        /// <summary>
        /// The threshold used to determine if a subscription is due for
        /// an update operation, based upon whether the subscription's
        /// update count exceeds this threshold since the last update
        /// operation.
        /// </summary>
        protected int _updateThreshold = 10;

        /// <summary>
        /// The threshold used to determine if a subscription is due for
        /// an update operation, based upon whether this many milliseconds
        /// have elapsed since the last update operation.
        /// </summary>
        protected long _timeoutMillis = 2000;

        /// <summary>
        /// The amount of time the update thread can sit idle between checking
        /// for timeouts and sending any pending updates.
        /// </summary>
        protected int _updateIntervalMillis = 2000;

        private Dictionary<Field, RecoveryPoint> _latestUpdates =
                                        new Dictionary<Field, RecoveryPoint>();
        private Dictionary<Field, Stopwatch> _timers =
                                        new Dictionary<Field, Stopwatch>();
        private Dictionary<Field, int> _counts = new Dictionary<Field, int>();
        private Object internalLock = new Object();
        private volatile bool _closed = false;
        private volatile bool _updateAll = false;
        private Thread _thread = null;

        /// <summary>
        /// A place to report exceptions in the background thread
        /// </summary>
        public delegate void ExceptionListener(Exception e);
        private ExceptionListener _exceptionListener = null;
        // Make sure updates don't arrive out of order
        private Dictionary<Field, ulong> _sequences = new Dictionary<Field, ulong>();
        private Object _sequencesLock = new Object();

        /// <summary>
        /// Creates a new ConflatingRecoveryPointAdapter.
        /// </summary>
        /// <param name="adapter_">The RecoveryPointAdapter doing the actual
        /// updates and recovery.
        public ConflatingRecoveryPointAdapter(RecoveryPointAdapter adapter_)
        {
            _adapter = adapter_;
        }

        /// <summary>
        /// Creates a new ConflatingRecoveryPointAdapter.
        /// </summary>
        /// <param name="adapter_">The RecoveryPointAdapter doing the actual
        /// updates and recovery.
        /// <param name="updateThreshold_">The maximum number of updates to a
        /// single subId before a conflated update is delivered.
        /// <param name="timeoutMillis_">The maximum amount of time in ms
        /// between conflated updates for each subId.
        /// <param name="updateIntervalMillis_">The maximum amount of time in ms
        /// the update thread can sit idle before checking for timeouts.
        public ConflatingRecoveryPointAdapter(RecoveryPointAdapter adapter_,
                                              int updateThreshold_,
                                              long timeoutMillis_,
                                              int updateIntervalMillis_)
        {
            _adapter = adapter_;
            _updateThreshold = updateThreshold_;
            _timeoutMillis = timeoutMillis_;
            _updateIntervalMillis = updateIntervalMillis_;
        }

        /// <summary>
        /// Sets the ExceptionListener instance used for communicating absorbed exceptions.
        /// </summary>
        /// <param name="exceptionListener">The exception listener instance to invoke for exceptions.</param>
        public void setExceptionListener(ExceptionListener exceptionListener)
        {
            _exceptionListener = exceptionListener;
        }

        /// <summary>
        /// Conflate an update to the newest recovery point. It ensures that updates are
        /// correctly handled for both repeat and new subscription data.
        /// </summary>
        /// <param name="recoveryPoint_">The newest recovery point to save.</param>
        public void update(RecoveryPoint recoveryPoint_)
        {
            if (recoveryPoint_ is SequencedRecoveryPoint)
            {
                ulong currentSequence = 0;
                lock(_sequencesLock)
                {
                    if (_sequences.TryGetValue(recoveryPoint_.getSubId(),
                                               out currentSequence))
                    {
                        if (currentSequence
                            > ((SequencedRecoveryPoint)recoveryPoint_).getSequence())
                        {
                            return;
                        }
                    }
                    _sequences[recoveryPoint_.getSubId()]
                        = ((SequencedRecoveryPoint)recoveryPoint_).getSequence();
                }
            }
            Field subId = recoveryPoint_.getSubId();
            lock (internalLock)
            {
                if (_thread == null)
                {
                    try
                    {
                        _thread = new Thread(new ThreadStart(this.run));
                        _thread.Start();
                    }
                    catch(Exception e)
                    {
                        absorbedException(new StoreException($"Exception creating thread {e}", e));
                        throw new StoreException($"Exception creating thread {e}", e);
                    }
                }
                if (_latestUpdates.ContainsKey(subId))
                {
                    _latestUpdates[subId] = recoveryPoint_.copy();
                    if (++_counts[subId] >= _updateThreshold)
                    {
                        Monitor.PulseAll(internalLock);
                    }
                }
                else
                {
                    _latestUpdates.Add(subId.copy(), recoveryPoint_.copy());
                    _counts.Add(subId.copy(), 1);
                    if (_timeoutMillis != 0)
                    {
                        _timers.Add(subId.copy(), Stopwatch.StartNew());
                    }
                }
            }
        }

        /// <summary>
        /// This method is responsible for removing or purging all recovery point
        /// information from the instance(self) and from the wrapped adapter.
        /// </summary>
        public void purge()
        {
            lock (internalLock)
            {
                if (_adapter != null) _adapter.purge();
                _latestUpdates.Clear();
                _counts.Clear();
                _timers.Clear();
            }
        }

        /// <summary>
        /// Remove all recovery point information for a given sub id.
        /// </summary>
        public void purge(Field subId_)
        {
            lock (internalLock)
            {
                if (_adapter != null) _adapter.purge(subId_);
                _latestUpdates.Remove(subId_);
                _counts.Remove(subId_);
                _timers.Remove(subId_);
            }
        }

        /// <summary>
        /// Force all updates to the underlying adapter, then close it.
        /// </summary>
        public void close()
        {
            lock (internalLock)
            {
                if (!_closed)
                {
                    runUpdateAll();
                    _closed = true;
                    Monitor.PulseAll(internalLock);
                }
            }
            if (_thread != null)
            {
                _thread.Join();
                _thread = null;
            }
            if (_adapter != null)
            {
                _adapter.close();
            }
        }

        /// <summary>
        /// Return an enumerator over recovered RecoveryPoints.
        /// </summary>
        public IEnumerator<RecoveryPoint> GetEnumerator()
        {
            return _adapter.GetEnumerator();
        }

        /// <summary>
        /// IEnumerable version of get enumerator.
        /// </summary>
        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return (System.Collections.IEnumerator)(_adapter.GetEnumerator());
        }

        /// <summary>
        /// IEnumerator move to next item.
        /// </summary>
        public bool MoveNext()
        {
            return _adapter.MoveNext();
        }

        /// <summary>
        /// IEnumerator reset - not implemented.
        /// </summary>
        public void Reset()
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Dispose of the adapter. Make sure everything is saved and close the
        /// client connection if requested during construction.
        /// </summary>
        void IDisposable.Dispose()
        {
            close();
            if (_adapter != null)
            {
                _adapter.Dispose();
                _adapter = null;
            }
        }

        /// <summary>
        /// The current RecoveryPoint from the underlying Adapter.
        /// </summary>
        public RecoveryPoint Current
        {
            get { return _adapter.Current; }
        }

        /// <summary>
        /// IEnumerator version of current.
        /// </summary>
        object System.Collections.IEnumerator.Current
        {
            get { return (object)Current; }
        }

        /// <summary>
        /// Force all held updates to be flushed to the underlying adapter.
        /// </summary>
        public void updateAll()
        {
            lock (internalLock)
            {
                runUpdateAll();
            }
        }

        // Already locked
        private void runUpdateAll()
        {
            _updateAll = true;
            while (_counts.Count != 0)
            {
                Monitor.PulseAll(internalLock);
                Monitor.Wait(internalLock, 250);
            }
        }

        private void run()
        {
            List<RecoveryPoint> updates = new List<RecoveryPoint>();
            bool closed = _closed;
            bool updateAll = false;
            while (!closed)
            {
                try
                {
                    lock (internalLock)
                    {
                        Monitor.Wait(internalLock, _updateIntervalMillis);
                        foreach (var timer in _timers)
                        {
                            if (timer.Value.ElapsedMilliseconds >= _timeoutMillis)
                            {
                                RecoveryPoint update = null;
                                if (_latestUpdates.TryGetValue(timer.Key,
                                                               out update)
                                    && update != null)
                                {
                                    _latestUpdates.Remove(timer.Key);
                                    updates.Add(update);
                                }
                            }
                        }
                        updateAll = _updateAll;
                        foreach (var count in _counts)
                        {
                            if (updateAll || _timeoutMillis == 0
                                || count.Value >= _updateThreshold)
                            {
                                RecoveryPoint update = null;
                                if (_latestUpdates.TryGetValue(count.Key,
                                                               out update)
                                    && update != null)
                                {
                                    _latestUpdates.Remove(count.Key);
                                    updates.Add(update);
                                }
                            }
                        }
                        foreach (var update in updates)
                        {
                            _counts.Remove(update.getSubId());
                            _timers.Remove(update.getSubId());
                        }
                        closed = _closed;
                    }
                    if (updateAll)
                    {
                        Monitor.Enter(internalLock);
                    }
                    try
                    {
                        foreach (var update in updates)
                        {
                            _adapter.update(update);
                        }
                    }
                    catch(Exception e)
                    {
                        absorbedException(new StoreException($"Exception in Adapter.update {e}", e));
                    }
                    finally
                    {
                        if (updateAll)
                        {
                            _updateAll = false;
                            Monitor.PulseAll(internalLock);
                            Monitor.Exit(internalLock);
                        }
                    }
                    updates.Clear();
                }
                catch(Exception e)
                {
                    absorbedException(e);
                }
            }
        }

        private void absorbedException(Exception e)
        {
            if (_exceptionListener != null)
            {
                try
                {
                    _exceptionListener(e);
                }
                catch(Exception) { }
            }
        }
    }
}


