////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government�s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System.Runtime.CompilerServices;
using System.Collections.Generic;
using AMPS.Client.Exceptions;
using AMPS.Client.Fields;

namespace AMPS.Client
{

    /// <summary>
    /// A no-op default implementation of the BookmarkStore interface that
    /// does nothing.
    /// </summary>
    public class DefaultBookmarkStore : BookmarkStore
    {
        private Dictionary<Field, Field> _inMemStore;

        /// <summary>
        /// Initializes a new instance of the <see cref="DefaultBookmarkStore"/> class.
        /// </summary>
        public DefaultBookmarkStore()
        {
            _inMemStore = new Dictionary<Field, Field>();
        }

        /// <summary>
        /// Disposes of any resources held by the bookmark store.
        /// </summary>
        public void Dispose()
        {
            // No-op
        }

        /// <summary>
        /// Logs the bookmark associated with the given message.
        /// </summary>
        /// <param name="message">The message containing the bookmark.</param>
        /// <returns>Returns 0.</returns>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public long log(Message message)
        {
            _inMemStore[message.getSubIdRaw()] = message.getBookmarkRaw();
            return 0;
        }

        /// <summary>
        /// Discards a bookmark associated with the specified topic and bookmark sequence number.
        /// </summary>
        /// <param name="topic">The topic associated with the bookmark.</param>
        /// <param name="bookmarkSequenceNumber">The sequence number of the bookmark to discard.</param>
        public void discard(Field topic, long bookmarkSequenceNumber)
        {
            // No-op
        }

        /// <summary>
        /// Discards the bookmark associated with the given message.
        /// </summary>
        /// <param name="message">The message containing the bookmark to discard.</param>
        public void discard(Message message)
        {
            // No-op
        }

        /// <summary>
        /// Retrieves the most recent bookmark for the specified topic.
        /// </summary>
        /// <param name="topic">The topic for which to retrieve the most recent bookmark.</param>
        /// <param name="useList">Specifies whether to use the bookmark list.</param>
        /// <returns>The most recent bookmark for the specified topic.</returns>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public Field getMostRecent(Field topic, bool useList)
        {
            // return the epoch if we haven't seen anything yet.
            return _inMemStore.ContainsKey(topic) ? _inMemStore[topic].copy() : new Field("0");
        }

        /// <summary>
        /// Determines whether the message should be discarded.
        /// </summary>
        /// <param name="message">The message to check.</param>
        /// <returns>Always returns false.</returns>
        //  all messages should be delivered.
        public bool isDiscarded(Message message)
        {
            return false;
        }

        /// <summary>
        /// Purges the bookmark associated with the specified subscription ID.
        /// </summary>
        /// <param name="subId">The subscription ID whose bookmark to purge.</param>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void purge(Field subId)
        {
            _inMemStore.Remove(subId);
        }

        /// <summary>
        /// Purges all bookmarks stored in memory.
        /// </summary>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void purge()
        {
            _inMemStore = new Dictionary<Field, Field>();
        }

        /// <summary>
        /// Marks a bookmark as persisted for the specified subscription ID.
        /// </summary>
        /// <param name="subId">The subscription ID.</param>
        /// <param name="bookmark">The bookmark to mark as persisted.</param>
        public void persisted(Field subId, BookmarkField bookmark)
        {
            // no-op
        }

        /// <summary>
        /// Marks a bookmark as persisted for the specified subscription ID.
        /// </summary>
        /// <param name="subId">The subscription ID.</param>
        /// <param name="bookmark">The bookmark to mark as persisted.</param>
        public void persisted(Field subId, long bookmark)
        {
            // no-op
        }

        /// <summary>
        /// Retrieves the oldest bookmark sequence number associated with the specified subscription ID.
        /// </summary>
        /// <param name="subId">The subscription ID.</param>
        /// <returns>The oldest bookmark sequence number.</returns
        public long getOldestBookmarkSeq(Field subId)
        {
            return 0;
        }

        /// <summary>
        /// Sets the resize handler for the bookmark store.
        /// </summary>
        /// <param name="handler">The resize handler to set.</param>
        public void setResizeHandler(BookmarkStoreResizeHandler handler)
        {
            // no-op
        }

        /// <summary>
        /// Sets the version of the server.
        /// </summary>
        public void setServerVersion(int version)
        {
            // no-op
        }
    }
}
