////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using AMPS.Client.Exceptions;
using System.Collections.Generic;


namespace AMPS.Client
{
    /// <summary>
    /// A simple ServerChooser that keeps a list of AMPS instances and
    /// Authenticators, and advances to the next one when failure occurs.
    /// <pre>
    /// <code>
    /// // Create the client
    /// HAClient client = new HAClient("CompositeSubber");
    /// try
    /// {
    ///     // Add URIs and connect the client
    ///     DefaultServerChooser sc = new DefaultServerChooser();
    ///     sc.add("tcp://127.0.0.1:9027/amps/composite-json-binary");
    ///     client.setServerChooser(sc);
    ///     client.connectAndLogon();
    /// </code>
    /// </pre>
    /// </summary>
    public class DefaultServerChooser : ServerChooser
    {
        private string[] _uris;
        private const int DEFAULT_SIZE = 5;
        private int _currentFree, _current;

        /// <summary>
        /// Constructs a DefaultServerChooser instance with default settings.
        /// </summary>
        public DefaultServerChooser()
        {
            this._uris = new string[DEFAULT_SIZE];
            this._currentFree = 0;
            this._current = 0;
        }

        /// <summary>
        /// Adds a URI to self.
        /// </summary>
        /// <param name="uri">The AMPS URI to add to the list.</param>
        public virtual DefaultServerChooser add(string uri)
        {
            if (this._currentFree >= this._uris.Length)
            {
                String[] newUris = new String[this._uris.Length * 2];
                System.Array.Copy(this._uris, 0, newUris, 0, this._uris.Length);
                this._uris = newUris;
            }

            this._uris[this._currentFree] = uri;
            ++_currentFree;
            return this;
        }

        /// <summary>
        /// Adds URIs to self.
        /// </summary>
        /// <param name="uris">The AMPS URIs to add to the list.</param>
        public virtual DefaultServerChooser addAll<T>(T uris) where T : ICollection<String>
        {
            if (this._currentFree >= this._uris.Length - uris.Count)
            {
                int newLength = this._uris.Length * 2;
                if (newLength < this._uris.Length + uris.Count)
                    newLength = this._uris.Length + uris.Count;
                String[] newUris = new String[newLength];
                System.Array.Copy(this._uris, 0, newUris, 0, this._uris.Length);
                this._uris = newUris;
            }

            uris.CopyTo(this._uris, this._currentFree);
            this._currentFree += uris.Count;
            return this;
        }

        /// <summary>
        /// Gets the current URI.
        /// </summary>
        /// <returns>The current URI.</returns>
        public virtual string getCurrentURI()
        {
            if (this._currentFree == 0)
            {
                return null;
            }
            if (this._current >= this._currentFree)
            {
                this._current = 0;
            }
            return this._uris[this._current];
        }

        /// <summary>
        /// Returns the Authenticator instance associated with the current URI.
        /// </summary>
        /// <returns>Returns an Authenticator or null if none is required for logon.</returns>
        public virtual Authenticator getCurrentAuthenticator()
        {
            return null;
        }

        /// <summary>
        /// Called by HAClient when an error occurs connecting to the current URI,
        /// and/or when an error occurs logging on. Implementors will likely advance the current
        /// URI to the next one in a list, or choose to stay with the current one, based on the
        /// exception type.
        /// </summary>
        /// <param name="exception">The exception associated with this failure.</param>
        /// <param name="ci">Information about the connection that failed.</param>
        public virtual void reportFailure(Exception exception, ConnectionInfo ci)
        {
            // In either case we'll just move on to the next
            // server whenever we have a failure connecting.
            // If we just got disconnected, though, we'll retry.
            if (!(exception is DisconnectedException))
            {
                next();
            }
        }

        /// <summary>
        /// Handles the report of a successful connection.
        /// </summary>
        /// <param name="ci">Information about the successful connection.</param>
        public virtual void reportSuccess(ConnectionInfo ci)
        {
            // This implementation does not use success.
        }

        /// <summary>
        /// Moves to the next server in the list of URIs.
        /// </summary>
        public virtual void next()
        {
            this._current = (this._current + 1) % this._currentFree;
        }

        /// <summary>
        /// Provides additional details to be included in an exception thrown when the AMPS instance(s)
        /// are not available. Called by HAClient when creating an exception.
        /// </summary>
        /// <returns>A String with information about the connection that failed and the
        /// reason for the failure. When no further information is available, returns an empty string.</returns>
        public virtual String getError()
        {
            return "";
        }
    }
}
