///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using AMPS.Client.Fields;
using AMPS.Client.Exceptions;

namespace AMPS.Client
{
    /// <summary>
    /// Interface to provide a recovery point for a bookmark subscription. Used in
    /// the RecoveryPointAdapter interface for updates and recovery.
    /// </summary>
    public class DynamicRecoveryPoint : RecoveryPoint
    {
        private Field _subId = null;
        private BookmarkStore _store = null;

        /// <summary>
        /// Initializes a new instance of the DynamicRecoveryPoint class with the specified parameters.
        /// </summary>
        /// <param name="store_">The bookmark store.</param>
        /// <param name="subId_">The subscription id.</param>
        public DynamicRecoveryPoint(BookmarkStore store_, Field subId_)
        {
            _subId = subId_;
            _store = store_;
        }

        /// <summary>
        /// Return the subscription id for this recovery point
        /// </summary>
        public virtual Field getSubId()
        {
            return _subId;
        }

        /// <summary>
        /// Return the bookmark for this recovery point.
        /// </summary>
        /// <returns>The bookmark for this recovery point.</returns>
        public virtual BookmarkField getBookmark()
        {
            return (BookmarkField)_store.getMostRecent(_subId, false);
        }

        /// <summary>
        /// Return a deep copy of self.
        /// </summary>
        /// <returns>A deep copy of self.</returns>
        public RecoveryPoint copy()
        {
            return new DynamicRecoveryPoint(_store, _subId.copy());
        }
    }

    /// <summary>
    /// Class for creating DynamicRecoveryPoint instances.
    /// </summary>
    public class DynamicRecoveryPointFactory : RecoveryPointFactory
    {
        private BookmarkStore _store;

        /// <summary>
        /// Initializes a new instance of the DynamicRecoveryPointFactory class with the specified bookmark store.
        /// </summary>
        /// <param name="store_">The bookmark store.</param>
        public DynamicRecoveryPointFactory(BookmarkStore store_)
        {
            _store = store_;
        }

        /// <summary>
        /// Creates a new recovery point.
        /// </summary>
        /// <param name="subId_">The subscription id.</param>
        /// <param name="bookmark_">The bookmark.</param>
        /// <returns>A new recovery point.</returns>
        public RecoveryPoint createRecoveryPoint(Field subId_,
                                                 BookmarkField bookmark_)
        {
            return new DynamicRecoveryPoint(_store, subId_);
        }
    }
}


