////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using AMPS.Client.Exceptions;

namespace AMPS.Client
{
    /// <summary>
    /// ExponentialDelayStrategy is an implementation that exponentially
    /// "backs off" when reconnecting to the same server, including optional
    /// jitter when backing off and a maximum amount of time to attempt
    /// reconnection before giving up entirely.
    /// </summary>
    public class ExponentialDelayStrategy : ReconnectDelayStrategy
    {
        /// <summary>
        /// A specialized exception type thrown by ExponentialDelayStrategy
        /// to indicate that the client should "give up" on attempting to
        /// reconnect to a server.
        /// </summary>
        public class MaximumRetryExceeded : AMPSException
        {

        }
        /// <summary>
        /// The default time (in milliseconds) to wait before reconnecting to a
        /// server for the first time after a failed connection.
        /// </summary>
        public const int DEFAULT_DELAY = 200;
        /// <summary>
        /// The default maximum time to wait between any reconnection attempts.
        /// Exponential backoff will not exceed this maximum.
        /// </summary>
        public const int DEFAULT_MAXIMUM_DELAY = 20 * 1000;
        /// <summary>
        /// The default exponent to use for calculating the next delay time. For
        /// example if the initial time is 200ms and the exponent is 2.0, the next
        /// delay will be 400ms, then 800ms, etc.
        /// </summary>
        public const double DEFAULT_BACKOFF_EXPONENT = 2.0;
        /// <summary>
        /// The default time (in millseconds) to allow reconnect attempts to
        /// continue without a successful connection, before "giving up" and
        /// abandoning the connection attempt. 0 means never give up.
        /// </summary>
        public const double DEFAULT_MAXIMUM_RETRY_TIME = -1.0;
        /// <summary>
        /// The amount of 'jitter' to apply when calculating a delay time,
        /// measured in multiples of the initial delay. Jitter is used to reduce
        /// the number of simultaneous reconnects that may be issued from
        /// multiple clients.
        /// </summary>
        public const double DEFAULT_JITTER = 1.0;
        /// <summary>
        /// Constructs an ExponentialDelayStrategy, the default strategy for
        /// HAClient, using the default values for the class.
        /// </summary>
        /// <p>
        /// For example, when constructed, the new object will use
        /// {@link DEFAULT_DELAY} as the initial delay,
        /// {@link DEFAULT_BACKOFF_EXPONENT} as the backoff exponent,
        /// {@link DEFAULT_MAXIMUM_RETRY_TIME} as the maximum retry time,
        /// and so on. You can use the setters provided on the class to
        /// adjust these settings.
        /// </p>
        public ExponentialDelayStrategy()
        {
            _initialDelay     = DEFAULT_DELAY;
            _maximumDelay     = DEFAULT_MAXIMUM_DELAY;
            _backoffExponent  = DEFAULT_BACKOFF_EXPONENT;
            _maximumRetryTime = DEFAULT_MAXIMUM_RETRY_TIME;
            _jitter           = DEFAULT_JITTER;
            _currentDelay     = _initialDelay;
            _endTime          = DateTime.MinValue;
        }
        /// <summary>
        /// Sets the time (in milliseconds) to wait before reconnecting to
        /// a server for the first time after a failed connection.
        /// </summary>
        /// <param name="initialDelay_">The new initial delay value,
        ///                             in milliseconds.</param>
        /// <returns>Self.</returns>
        public ExponentialDelayStrategy setInitialDelay(int initialDelay_)
        {
            _initialDelay = initialDelay_;
            _currentDelay = initialDelay_;
            return this;
        }
        /// <summary>
        /// Sets the maximum time to wait between any reconnection attempts.
        /// Exponential backoff will not exceed this maximum.
        /// </summary>
        /// <param name="maximumDelay_">The new maximum delay value,
        ///                             in milliseconds.</param>
        /// <returns>Self.</returns>
        public ExponentialDelayStrategy setMaximumDelay(int maximumDelay_)
        {
            _maximumDelay = maximumDelay_;
            return this;
        }
        /// <summary>
        /// Sets the exponent to use for calculating the next delay time. For
        /// example if the initial time is 200ms and the exponent is 2.0, the next
        /// delay will be 400ms, then 800ms, etc.
        /// </summary>
        /// <param name="exponent_">The new exponent value.</param>
        /// <returns>self</returns>
        public ExponentialDelayStrategy setBackoffExponent(double exponent_)
        {
            _backoffExponent = exponent_;
            return this;
        }
        /// <summary>
        /// Sets the time (in millseconds) to allow reconnect attempts to
        /// continue without a successful connection, before "giving up" and
        /// abandoning the connection attempt. 0 means never give up.
        /// </summary>
        /// <param name="maximumRetryTime_">The new maximum retry time,
        ///                                 in milliseconds.</param>
        /// <returns>Self.</returns>
        public ExponentialDelayStrategy setMaximumRetryTime(
                int maximumRetryTime_)
        {
            _maximumRetryTime = (double)maximumRetryTime_;
            return this;
        }

        /// <summary>
        /// Sets the jitter factor used to add randomness to the delay time.
        /// Jitter is represented as a multiple of the initial delay time;
        /// a random number from [0, (JITTER * INITIAL_DELAY) ) is added
        /// to nonzero time delays.
        /// </summary>
        /// <param name="jitter_">The jitter factor, in multiples of
        ///                       the initial delay.</param>
        /// <returns>Self.</returns>
        public ExponentialDelayStrategy setJitter(double jitter_)
        {
            _jitter = jitter_;
            return this;
        }

        /// <summary>
        /// Gets the connection wait duration based on the provided URI.
        /// </summary>
        /// <param name="uri_">The URI string.</param>
        /// <returns>The connection wait duration.</returns>
        public int getConnectWaitDuration(String uri_)
        {
            _throwIfMaximumExceeded();
            if(_lastUri != uri_)
            {
                _lastUri = uri_;
                if(_firstUri == uri_)
                {
                    // we've wrapped around the "list." Delay.
                    return _currentDurationAndIncrease();
                }
                else if(_firstUri == null)
                {
                    _firstUri = uri_;
                }
                return 0;
            }
            return _currentDurationAndIncrease();
        }

        /// <summary>
        /// Resets the exponential delay strategy.
        /// </summary>
        public void reset()
        {
            _lastUri      = null;
            _firstUri     = null;
            _currentDelay = _initialDelay;
            _endTime    = DateTime.MinValue;
        }

        /// <summary>
        /// An internal function used to check if the configured retry has been
        /// exceeded, and if so, throw MaximumRetryExceeded to cause the calling
        /// HAClient to give up on reconnecting to a server.
        /// </summary>
        protected void _throwIfMaximumExceeded()
        {
            if(_maximumRetryTime > 0.0)
            {
                if(_endTime.CompareTo(DateTime.MinValue) != 0)
                {
                    if(DateTime.Now.CompareTo(_endTime) > 0)
                    {
                        throw new MaximumRetryExceeded();
                    }

                }
                else
                {
                    _endTime = DateTime.Now.AddMilliseconds(_maximumRetryTime);
                }
            }
        }
        /// <summary>
        /// Returns the _currentDelay, and also expands _currentDelay based
        /// on the configured exponent and maximum.
        /// </summary>
        /// <returns>The delay that should be used, prior to increasing it.</returns>
        protected int _currentDurationAndIncrease()
        {
            int newDelay = (int) (_currentDelay * _backoffExponent);
            if(newDelay > _maximumDelay) newDelay = _maximumDelay;
            int currentDelay = _currentDelay;
            int maxJitter = (int)(_jitter * _initialDelay);
            if (maxJitter > _maximumDelay) maxJitter = _maximumDelay;
            if (_jitter > 0.0)
            {
                if (currentDelay > _maximumDelay - maxJitter)
                    currentDelay = (_maximumDelay - maxJitter > _initialDelay) ?
                                    _maximumDelay - maxJitter : _initialDelay;
                currentDelay +=
                     (int)(_random.NextDouble() * maxJitter);
                if (currentDelay > _maximumDelay) currentDelay = _maximumDelay;
            }

            if (_maximumRetryTime > 0.0
                && _endTime.CompareTo(DateTime.MinValue) != 0)
            {
                double timeElapsed = _maximumRetryTime
                                     - _endTime.Subtract(DateTime.Now)
                                               .TotalMilliseconds;
                if(timeElapsed + (double)currentDelay > _maximumRetryTime)
                {
                    if(timeElapsed > _maximumRetryTime)
                        throw new MaximumRetryExceeded();
                    currentDelay = (int)(_maximumRetryTime - timeElapsed);
                }
            }
            _currentDelay = newDelay;
            return currentDelay;
        }

        protected int       _initialDelay;
        protected int       _maximumDelay;
        protected double    _backoffExponent;
        protected double    _maximumRetryTime;
        protected double    _jitter;
        protected int       _currentDelay;
        protected String    _lastUri;
        protected String    _firstUri;
        protected DateTime  _endTime;
        protected Random    _random = new Random();
    }
}
