////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Text;
using AMPS.Client.Exceptions;

namespace AMPS.Client
{
    /// <summary>
    /// Used to build up message strings for FIX.
    /// </summary>
    public class FIXBuilder
    {
        private byte[] buffer = null;
        private int size = 0;
        private int capacity = 1024;

        private byte fieldSeparator;

        /// <summary>
        /// Initializes a new instance of the FIXBuilder class with
        /// the specified capacity and field separator.
        /// </summary>
        /// <param name="capacity">The initial capacity of the builder.</param>
        /// <param name="fieldSeparator">The field separator byte.</param>
        public FIXBuilder(int capacity, byte fieldSeparator)
        {
            this.fieldSeparator = fieldSeparator;
            this.capacity = capacity < 8 ? 8 : capacity;
            this.buffer = new byte[capacity];
        }

        /// <summary>
        /// Returns the number of bytes in the byte array containing the FIX message.
        /// </summary>
        /// <returns>The number of valid bytes in the byte array.<returns>
        public int getSize()
        {
            return this.size;
        }

        /// <summary>
        /// Returns the byte array containing the FIX message. The number of valid
        /// bytes within the buffer is returned from getSize().
        /// </summary>
        /// <returns>The byte array containing the FIX message.</returns>
        public byte[] getBytes()
        {
            return buffer;
        }

        private int ruLog10(int tag)
        {
            long scalar = 10;
            for (int j = 1; j < 20; ++j)
            {
                if (tag < scalar)
                {
                    return j;
                }
                scalar = scalar * 10;
            }
            return 0;
        }

        private void checkCapacity(int bytesNeeded)
        {
            if (capacity - size < bytesNeeded)
            {
                // Not enough capacity left, must resize
                while (capacity - size < bytesNeeded)
                {
                    capacity = capacity * 2;
                }

                byte[] newbuf = new byte[capacity];

                Array.Copy(buffer, 0, newbuf, 0, size);

                buffer = newbuf;
            }
        }

        /// <summary>
        /// Appends a tag and value pair contained within a byte buffer to the FIX message.
        /// </summary>
        /// <param name="tag">The integer FIX tag to append</param>
        /// <param name="value">The byte buffer containing the FIX value to append</param>
        /// <param name="offset">The starting location of the value inside the byte buffer</param>
        /// <param name="length">The length of the value inside the byte buffer</param>
        /// <returns>A reference to this object</returns>
        /// <exception cref="ArgumentException">Thrown if the tag argument is negative.</exception>
        public FIXBuilder append(int tag, byte[] value, int offset, int length)
        {
            if (tag < 0)
                throw new ArgumentException("negative tag used in FIXBuilder");

            int tagSize = ruLog10(tag);
            int sizeNeeded = tagSize + length + 2;
            checkCapacity(sizeNeeded);

            // Tag
            int writeIndex = size + tagSize;
            for (int j = 1; j <= tagSize; ++j)
            {
                buffer[writeIndex - j] = (byte)(48 + (tag % 10));
                tag = tag / 10;
            }

            // Equal sign
            buffer[writeIndex] = (byte)'=';
            ++writeIndex;

            // value

            Array.Copy(value, offset, buffer, writeIndex, length);
            writeIndex += length;

            buffer[writeIndex] = fieldSeparator;
            size += sizeNeeded;
            return this;
        }

        /// <summary>
        /// Appends a tag and value pair to the FIX message.
        /// </summary>
        /// <param name="tag">the integer FIX tag to append</param>
        /// <param name="value">the FIX value for the tag to append. The value will be converted
        /// to an ISO-8859-1 byte array for writing.</param>
        /// <returns>a reference to this object.</returns>
        /// <exception cref="ArgumentException">If the tag argument is negative.</exception>
        /// <exception cref="CommandException">If the tag or value argument is not convertible to ISO-8859-1</exception>
        public FIXBuilder append(int tag, String value)
        {
            if (tag < 0)
                throw new ArgumentException("negative tag used in FIXBuilder");

            int tagSize = ruLog10(tag);
            byte[] valueBytes;
            try
            {
                valueBytes = Encoding.GetEncoding("ISO-8859-1").GetBytes(value);
            }
            catch (EncoderFallbackException ioex)
            {
                throw new CommandException("Unable to encode value '" + value
                      + "' to byte array in FIXBuilder.\n" + ioex.ToString());
            }
            int sizeNeeded = tagSize + valueBytes.Length + 2;
            checkCapacity(sizeNeeded);

            // Tag
            int writeIndex = size + tagSize;
            for (int j = 1; j <= tagSize; ++j)
            {
                buffer[writeIndex - j] = (byte)(48 + (tag % 10));
                tag = tag / 10;
            }

            // Equal sign
            buffer[writeIndex] = (byte)'=';
            ++writeIndex;

            // Value
            Array.Copy(valueBytes, 0, buffer, writeIndex, valueBytes.Length);
            writeIndex += valueBytes.Length;

            buffer[writeIndex] = fieldSeparator;
            size += sizeNeeded;

            return this;
        }

        /// <summary>
        /// Clears the current buffer in the Builder so it can be reused. Capacity is unaffected.
        /// </summary>
        /// <returns>a reference to this object.</returns>
        public FIXBuilder clear()
        {
            size = 0;
            return this;
        }

    }
}
