////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.IO;
using System.Text;
using AMPS.Client.Exceptions;

namespace AMPS.Client
{
    /// <summary>
    /// Implements the Protocol interface. Used to parse a byte stream as a
    /// series of messages. The byte stream must use the AMPS FIX protocol.
    /// </summary>
    public class FIXProtocol : Protocol
    {
        private const String LATIN1 = "ISO-8859-1";
        public byte fieldSeparator = 0x01;
        public byte headerSeparator = 0x02;
        public byte messageSeparator = 0x03;
        public Encoding Encoding;

        /// <summary>
        /// The character set used to encode and decode messages sent via FIX.
        /// This value must be set before connecting a Client to a FIX endpoint.
        /// </summary>
        public FIXProtocol()
        {
          this.Encoding = Encoding.GetEncoding(LATIN1);
        }

        /// <summary>
        /// Initializes a new instance of the FIXProtocol class with the
        /// specified properties.
        /// </summary>
        /// <param name="props">A dictionary of properties containing field,
        /// header, and message separators.</param>
        public FIXProtocol(Properties props)
            : this()
        {
            this.fieldSeparator = props.ContainsKey("fieldSeparator") ? (Byte)props["fieldSeparator"][0] : (byte)1;
            this.headerSeparator = props.ContainsKey("headerSeparator") ? (Byte)props["headerSeparator"][0] : (byte)2;
            this.messageSeparator = props.ContainsKey("messageSeparator") ? (Byte)props["messageSeparator"][0] : (byte)3;
        }

        /// <summary>
        /// Initializes a new instance of the FIXProtocol class with the specified separators.
        /// </summary>
        /// <param name="fieldSeparator">The field separator byte value.</param>
        /// <param name="headerSeparator">The header separator byte value.</param>
        /// <param name="messageSeparator">The message separator byte value.</param>
        public FIXProtocol(byte fieldSeparator,
              byte headerSeparator,
              byte messageSeparator)
            : this()
        {

            this.fieldSeparator = fieldSeparator;
            this.headerSeparator = headerSeparator;
            this.messageSeparator = messageSeparator;
        }

        /// <summary>
        /// Allocates a new FIX message.
        /// <returns> The newly allocated FIX Message. </returns>
        public virtual Message allocateMessage()
        {
            return new FIXMessage(this.fieldSeparator,
                  this.headerSeparator,
                  this.messageSeparator,
                  Encoding,Encoding);
        }

        /// <summary>
        /// Gets a new ProtocolParser instance for parsing messages.
        /// </summary>
        /// <returns>A new ProtocolParser instance for parsing FIX
        /// messages.</returns>
        public virtual ProtocolParser getMessageStream()
        {
            return new FIXProtocolParser(this);
        }
    }
}
