////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections;
using System.Collections.Generic;

namespace AMPS.Client.Fields
{
    /// <summary>
    /// Represents field data for a <see cref="AMPS.Client.Message"/> that
    /// consists of the bookmark. This is the AMPS Bookmark for the message.
    /// </summary>
    public class BookmarkField : StringField
    {
        public const int MAX_BOOKMARK_LENGTH = 42;
        public const int MAX_TIMESTAMP_LENGTH = 24;
        public const int MIN_TIMESTAMP_LENGTH = 10;

        /// <summary>
        /// Default constructor.
        /// </summary>
        public BookmarkField()
        { }

        /// <summary>
        /// Constructor with string version of a bookmark.
        /// </summary>
        /// <param name="bookmark">The string containing the bookmark.</param>
        public BookmarkField(string bookmark)
          : base(bookmark)
        { }

        /// <summary>
        /// Constructor with specified buffer, position, and length.
        /// </summary>
        /// <param name="buffer">The byte array containing the bookmark field.</param>
        /// <param name="position">The starting position of the bookmark field in
        /// the buffer.</param>
        /// <param name="length">The length of the bookmark field.</param>
        protected BookmarkField(byte[] buffer, int position, int length)
            : base(buffer, position, length)

        {
        }

        /// <summary>
        /// Indicates if this bookmark represents a timestamp.
        /// </summary>
        /// <returns>True if the value looks like an AMPS ISO-8601 timestamp.</returns>
        public bool isTimestamp()
        {
            return this.length >= MIN_TIMESTAMP_LENGTH
                   && this.buffer[this.position + 8] == 'T';
        }

        /// <summary>
        /// Indicates whether this bookmark likely represents a range, using
        /// only a simple check for brackets or parantheses.
        /// </summary>
        /// <returns>True if the value looks like a range.</returns>
        public bool isRange()
        {
            if (this.length >= 5)
            {
                int pos = this.position;
                while (this.buffer[pos] == ' ') {
                    if (++pos >= this.position + this.length) return false;
                }
                if (this.buffer[pos] == '(' || this.buffer[pos] == '[')
                {
                    pos = this.position + this.length - 1;
                    while (this.buffer[pos] == ' ') {
                        if (--pos <= this.position) return false;
                    }
                    if (this.buffer[pos] == ')' || this.buffer[pos] == ']')
                        return true;
                }
            }
            return false;
        }

        /// <summary>
        /// Identifies if this bookmark is a list of , separated bookmarks.
        /// </summary>
        public bool isBookmarkList()
        {
            for (int comma=position; comma < position + length; ++comma)
            {
                if (buffer[comma] == (byte)',')
                {
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// Returns the bookmark as a string.
        /// </summary>
        public override string ToString()
        {
           return System.Text.Encoding.UTF8.GetString(this.buffer, this.position, this.length);
        }

        /// <summary>
        /// Gets the numeric publisher id for this bookmark.
        /// </summary>
        /// <returns>The numeric publisher id for this bookmark or 0 if it's a timestamp.</returns>
        public ulong getPublisherId()
        {
            if (isTimestamp() || isRange()) return 0;
            ulong result = 0;
            for(int i = this.position; i < this.position + this.length; i++)
            {
                if( this.buffer[i] == (int)'|' )
                {
                    break;
                }

                result *= 10;
                result += (ulong)this.buffer[i] - 48;
            }
            return result;
        }

        /// <summary>
        /// Returns the sequence number for the bookmark or 0 if it's a timestamp.
        /// </summary>
        public ulong getSequenceNumber()
        {
            if (isTimestamp() || isRange()) return 0;
            ulong result = 0;
            int i = this.position;
            for(; i < this.position + this.length; i++)
            {
                if( this.buffer[i] == (int)'|' )
                {
                    break;
                }
            }

            for(i = i+1; i < this.position + this.length; i++)
            {
                if( this.buffer[i] == (int)'|' )
                {
                    break;
                }
                result *= 10;
                result += (ulong)this.buffer[i] - 48;
            }

            return result;
        }

        /// <summary>
        /// Parses a bookmark list to individual bookmarks.
        /// </summary>
        /// <returns>A list of <see cref="BookmarkField"/>.</returns>
        public IList<BookmarkField> parseBookmarkList()
        {
            IList<BookmarkField> bookmarks = new List<BookmarkField>();
            int start = position;
            int comma = position;
            for ( ; comma < position+length; ++comma)
            {
                if (buffer[comma] == (int)',')
                {
                    bookmarks.Add(new BookmarkField(buffer, start, comma-start));
                    start = comma + 1;
                }
            }
            if (comma > start)
            {
                bookmarks.Add(new BookmarkField(buffer, start, comma-start));
            }
            return bookmarks;
        }

        /// <summary>
        /// Provides a way to duplicate this object and retain its specific type (and hashcode() implementation).
        /// </summary>
        /// <returns>A new instance of <see cref="BookmarkField"/> as an object.</returns>
        public Object clone()
        {
            byte[] copy = null;

            if (buffer != null)
            {
                copy = new byte[length];
                Array.Copy(buffer, position, copy, 0, length);
            }
            return new BookmarkField(copy, 0, length);
        }

        /// <summary>
        /// Override base class's copy() method appropriately, so we don't inadvertently mix
        /// <see cref="BookmarkField"/> and <see cref="Field"/> up -- they have different hashCode
        /// implementations.
        /// </summary>
        /// <returns>A new instance of <see cref="BookmarkField"/>.</returns>
        public new BookmarkField copy()
        {
            return (BookmarkField)clone();
        }

        /// <summary>
        /// Gets the hash code for the <see cref="BookmarkField"/>.
        /// </summary>
        /// <returns>The hash code for the <see cref="BookmarkField"/>.</returns>
        Utilities.CRC32 _crcObject = new Utilities.DefaultCRC32();
        public override int GetHashCode()
        {
            long crc32 = _crcObject.init();
            crc32 = _crcObject.update(crc32, buffer, position, length);

            return (int)crc32;
        }


    }
}
