////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
namespace AMPS.Client.Fields
{
    /// <summary>
    /// Field data for a <see cref="AMPS.Client.Message"/> consisting
    /// of a bookmark This is the AMPS Bookmark for a Message going to
    /// the server.
    /// </summary>
    public class BookmarkRangeField : BookmarkField
    {
        private int _capacity = 0;
        private BookmarkField _start = null;
        private BookmarkField _end = null;
        private int _open = -1;

        /// <summary>
        /// Constructs a new empty BookmarkRangeField.
        /// </summary>
        public BookmarkRangeField()
        { }

        /// <summary>
        /// Constructs a new BookmarkRangeField with the provided buffer,
        /// position, and length.
        /// </summary>
        /// <param name="buffer">The buffer containing the bookmark range.</param>
        /// <param name="position">The starting position of the bookmark range within
        ///                        the buffer</param>
        /// <param name="length">The length of the bookmark range.</param>
        protected BookmarkRangeField(byte[] buffer, int position, int length)
            : base(buffer, position, length)

        {
            _capacity = length;
            parse();
        }

        /// <summary>
        /// Provides a way to make this object a duplicate of f if f
        /// represents a valid bookmark range.
        /// </summary>
        public override void copyFrom(Field f)
        {
            if (f == null || f.length == 0) {
                length = 0;
                position = 0;
                notValid();
                return;
            }
            if (buffer == null || f.length > _capacity)
            {
                buffer = new byte[f.length];
                _capacity = f.length;
            }
            // Start copy at start of range
            int start = 0;
            while (f.buffer[f.position+start] != (byte)'['
                   && f.buffer[f.position+start] != (byte)'('
                   && start < f.length) {
                ++start;
            }
            if (start < f.length) {
                length = f.length - start;
                Array.Copy(f.buffer, f.position+start, buffer, 0, length);
                position = 0;
                parse();
            }
        }

        /// <summary>
        /// Checks if the bookmark range is valid.
        /// </summary>
        /// <returns>True if the bookmark range is valid, false
        /// otherwise.</returns>
        public bool isValid()
        {
            return _open >= 0;
        }

        /// <summary>
        /// Returns the start bookmark of the range.
        /// </summary>
        public BookmarkField getStart()
        {
            return _start;
        }

        /// <summary>
        /// Returns the end bookmark of the range.
        /// </summary>
        public BookmarkField getEnd()
        {
            return _end;
        }

        /// <summary>
        /// Returns if the start of the range in inclusive.
        /// </summary>
        public bool isStartInclusive()
        {
            return (isValid() ? buffer[_open] == (byte)'[' : false);
        }

        /// <summary>
        /// Returns if the end of the range in inclusive.
        /// </summary>
        public bool isEndInclusive()
        {
            return (isValid() ? buffer[length-1] == (byte)']' : false);
        }

        /// <summary>
        /// Returns if the start of the range in exclusive.
        /// </summary>
        public bool isStartExclusive()
        {
            return (isValid() ? buffer[_open] == (byte)'(' : false);
        }

        /// <summary>
        /// Returns if the end of the range in exclusive.
        /// </summary>
        public bool isEndExclusive()
        {
            return (isValid() ? buffer[length-1] == (byte)')' : false);
        }

        /// <summary>
        /// Replaces the start bookmark of the range.
        /// </summary>
        /// <param name="start_">The new start bookmark.</param>
        /// <param name="makeExclusive"/>Indicates whether the new start should be exclusive.</param>
        public void replaceStart(BookmarkField start_, bool makeExclusive = true)
        {
            if (!isValid()) return;
            // Don't go backwards
            if (start_.getPublisherId() == _start.getPublisherId()
                && start_.getSequenceNumber() <= _start.getSequenceNumber())
            {
                if (makeExclusive && start_.getSequenceNumber() == _start.getSequenceNumber())
                {
                    buffer[position+_open] = (byte)'(';
                }
                return;
            }
            int newLen = start_.length + _end.length + 3;
            // Check if new start fits
            if (_capacity >= newLen) {
                if (makeExclusive) buffer[position+_open] = (byte)'(';
                if (_open > 0) {
                    buffer[position] = buffer[position+_open];
                    _open = 0;
                }
                if (_end.position < start_.length + 2) {
                    // Move end toward the end
                    for (int pos = newLen-1, src = length - 1;
                         src > _start.length; --pos, --src)
                    {
                        while (buffer[position+src] == (byte)' ') {
                            --src;
                        }
                        buffer[position+pos] = buffer[position+src];
                        if (buffer[position+src] == (byte)':') {
                            _end.set(buffer, start_.length+2, _end.length);
                            break;
                        }
                    }
                }
                else if (_end.position > start_.length + 2) {
                    // Move :end toward the beginning
                    for (int pos = start_.length+2, src = _end.position;
                         src < length; ++pos, ++src)
                    {
                        while (buffer[position+src] == (byte)' ') {
                            ++src;
                        }
                        buffer[position+pos] = buffer[position+src];
                        if (buffer[position+src] == (byte)')'
                            || buffer[position+src] == (byte)']')
                        {
                            _end.set(buffer, position+start_.length+2, _end.length);
                            buffer[position+start_.length+1] = (byte)':';
                            break;
                        }
                    }
                }
                length = newLen;
                Array.Copy(start_.buffer, start_.position,
                           buffer, position+1, start_.length);
                _start.set(buffer, position+1, start_.length);
            }
            else {
                _capacity = System.Math.Max(4 * BookmarkField.MAX_BOOKMARK_LENGTH + 6,
                                            start_.length + _end.length + 3);
                byte[] tempBuffer = new byte[_capacity];
                if (makeExclusive) tempBuffer[0] = (byte)'(';
                else tempBuffer[0] = buffer[_open];
                _open = 0;
                Array.Copy(start_.buffer, start_.position,
                           tempBuffer, 1, start_.length);
                _start.set(tempBuffer, 1, start_.length);
                tempBuffer[start_.length + 1] = (byte)':';
                Array.Copy(_end.buffer, _end.position, tempBuffer,
                           start_.length + 2, _end.length);
                _end.set(tempBuffer, start_.length + 2, _end.length);
                tempBuffer[newLen - 1] = buffer[length - 1];
                set(tempBuffer, 0, newLen);
            }
        }

        /// <summary>
        /// Returns the entire bookmark rangs as a string.
        /// </summary>
        public override string ToString()
        {
           return System.Text.Encoding.UTF8.GetString(this.buffer, this.position, this.length);
        }

        /// <summary>
        /// Provides a way to duplicate this object and retain its specific type
        /// (and hashcode() implementation).
        /// </summary>
        /// <returns>A new instance of BookmarkRangeField with the same data.</returns>
        public new Object clone()
        {
            byte[] copy = null;

            if (buffer != null)
            {
                copy = new byte[length];
                Array.Copy(buffer, position, copy, 0, length);
            }
            BookmarkRangeField rf = new BookmarkRangeField(copy, 0, length);
            rf.parse();
            return rf;
        }

        /// <summary>
        /// Override base class's copy() method appropriately, so we don't inadvertently mix
        /// BookmarkRangeFields and Fields up -- they have different hashCode implementations.
        /// </summary>
        public new BookmarkRangeField copy()
        {
            return (BookmarkRangeField)clone();
        }

        Utilities.CRC32 _crcObject = new Utilities.DefaultCRC32();

        /// <summary>
        /// Computes the hash code for this BookmarkRangeField.
        /// </summary>
        /// <returns>The hash code value for this BookmarkRangeField.</returns>
        public override int GetHashCode()
        {
            long crc32 = _crcObject.init();
            crc32 = _crcObject.update(crc32, buffer, position, length);

            return (int)crc32;
        }

        // <summary>
        // Invalidate self completely.
        // </summary>
        private void notValid()
        {
            _start = null;
            _end = null;
            _open = -1;
        }

        // Does a full parse into start/end of range.
        private void parse()
        {
            notValid();
            bool foundClose = false;
            bool foundSeparator = false;
            int pos = this.position;
            int end = this.position + this.length;
            while (pos < end)
            {
                switch (buffer[pos])
                {
                    // Beginning of range
                    case (byte)'(':
                    case (byte)'[':
                    {
                        // Can't be passed the end or already started
                        if (foundClose || _open > -1) {
                            notValid();
                            return;
                        }
                        _open = pos;
                    }
                    break;
                    // Valid bookmark characters
                    case (byte)'0':
                    case (byte)'1':
                    case (byte)'2':
                    case (byte)'3':
                    case (byte)'4':
                    case (byte)'5':
                    case (byte)'6':
                    case (byte)'7':
                    case (byte)'8':
                    case (byte)'9':
                    case (byte)'|':
                    case (byte)',':
                    case (byte)'.':
                    case (byte)'T':
                    case (byte)'Z':
                    {
                        // Have to be between start and end
                        if (foundClose || _open < 0) {
                            notValid();
                            return;
                        }
                        else if (foundSeparator) {
                            if (_end == null) {
                                // Create the _end field, length will change
                                _end = new BookmarkField();
                                _end.set(this.buffer, pos, end - pos);
                            }
                        }
                        else if (_start == null) {
                            // Create the _start field, length will change
                            _start = new BookmarkField();
                            _start.set(this.buffer, pos, end - pos);
                        }
                    }
                    break;
                    // Start/end separator character
                    case (byte)':':
                    {
                        // Have to be between start and end and have begun
                        // the start bookmark.
                        if (foundClose || _open < 0 || _start == null || foundSeparator) {
                            notValid();
                            return;
                        }
                        foundSeparator = true;
                        // Check for setting start length
                        if (pos - _start.position < _start.length) {
                            _start.length = pos - _start.position;
                        }
                    }
                    break;
                    // Close of the range
                    case (byte)']':
                    case (byte)')':
                    {
                        // Have to be between start and end and begun
                        // the end bookmark.
                        if (foundClose || _open < 0 || _end == null) {
                            notValid();
                            return;
                        }
                        foundClose = true;
                        length = pos - this.position + 1;
                        // Check for setting end length
                        if (pos - _end.position < _end.length) {
                            _end.length = pos - _end.position;
                        }
                    }
                    break;
                    // Spaces are skipped, but allowed
                    case (byte)' ':
                    {
                        // Is this the end of _end?
                        if (_end != null && pos - _end.position < _end.length) {
                            _end.length = pos - _end.position;
                        }
                        // Is this the end of _start?
                        else if (_start != null
                                 && pos - _start.position < _start.length) {
                            _start.length = pos - _start.position;
                        }
                    }
                    break;
                    default:
                    {
                        notValid();
                        return;
                    }
                }
                ++pos;
            }
            if (_end == null || _start == null || !foundSeparator
                || !foundClose || _open < 0) {
                notValid();
            }
        }


    }
}
