////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Text;
using AMPS.Client;

namespace AMPS.Client.Fields
{
    /// <summary>
    /// Field data for a <see cref="Message"/> which consists of a Command.
    /// This Command tells AMPS how to interpret the Message.
    /// </summary>
    public class CommandField : Field
    {
        protected static string LATIN1 = "ISO-8859-1";
        protected static byte LATIN1_ZERO = 48;

        private static byte[] CT_SUBSCRIBE = null;
        private static byte[] CT_UNSUBSCRIBE = null;
        private static byte[] CT_PUBLISH = null;
        private static byte[] CT_SOW = null;
        private static byte[] CT_LOGON = null;
        private static byte[] CT_START_TIMER = null;
        private static byte[] CT_STOP_TIMER = null;
        private static byte[] CT_SOW_AND_SUBSCRIBE = null;
        private static byte[] CT_DELTA_PUBLISH = null;
        private static byte[] CT_DELTA_SUBSCRIBE = null;
        private static byte[] CT_SOW_AND_DELTA_SUBSCRIBE = null;
        private static byte[] CT_SOW_DELETE = null;
        private static byte[] CT_GROUP_BEGIN = null;
        private static byte[] CT_GROUP_END = null;
        private static byte[] CT_OOF = null;
        private static byte[] CT_ACK = null;
        private static byte[] CT_HEARTBEAT = null;
        private static byte[] CT_FLUSH = null;
        private static byte[] CT_UNKNOWN = null;

        static CommandField()
        {
            CT_SUBSCRIBE = "subscribe".getBytes(LATIN1);
            CT_UNSUBSCRIBE = "unsubscribe".getBytes(LATIN1);
            CT_PUBLISH = "publish".getBytes(LATIN1);
            CT_SOW = "sow".getBytes(LATIN1);
            CT_LOGON = "logon".getBytes(LATIN1);
            CT_START_TIMER = "start_timer".getBytes(LATIN1);
            CT_STOP_TIMER = "stop_timer".getBytes(LATIN1);
            CT_SOW_AND_SUBSCRIBE = "sow_and_subscribe".getBytes(LATIN1);
            CT_DELTA_PUBLISH = "delta_publish".getBytes(LATIN1);
            CT_DELTA_SUBSCRIBE = "delta_subscribe".getBytes(LATIN1);
            CT_SOW_AND_DELTA_SUBSCRIBE = "sow_and_delta_subscribe".getBytes(LATIN1);
            CT_SOW_DELETE = "sow_delete".getBytes(LATIN1);
            CT_GROUP_BEGIN = "group_begin".getBytes(LATIN1);
            CT_GROUP_END = "group_end".getBytes(LATIN1);
            CT_OOF = "oof".getBytes(LATIN1);
            CT_ACK = "ack".getBytes(LATIN1);
            CT_HEARTBEAT = "heartbeat".getBytes(LATIN1);
            CT_FLUSH = "flush".getBytes(LATIN1);
            CT_UNKNOWN = "unknown".getBytes(LATIN1);
        }

        /// <summary>
        /// This method gets the enum representation of the Command.
        /// </summary>
        /// <returns>The Message.Commands enum value.</returns>
        public Message.Commands getValue()
        {
            if (this.buffer != null)
            {
                return decodeCommand(this.buffer, this.position, this.length);
            }
            return Message.Commands.Unknown;
        }

        /// <summary>
        /// Sets the value of the Command.
        /// </summary>
        /// <param name="v">The Command value to set.</param>
        public void setValue(Message.Commands v)
        {
            if (v == 0)
            {
                reset();
                return;
            }
            switch (v)
            {
                case Message.Commands.Subscribe:
                    set(CT_SUBSCRIBE);
                    break;
                case Message.Commands.Unsubscribe:
                    set(CT_UNSUBSCRIBE);
                    break;
                case Message.Commands.Publish:
                    set(CT_PUBLISH);
                    break;
                case Message.Commands.SOW:
                    set(CT_SOW);
                    break;
                case Message.Commands.Logon:
                    set(CT_LOGON);
                    break;
                case Message.Commands.StartTimer:
                    set(CT_START_TIMER);
                    break;
                case Message.Commands.StopTimer:
                    set(CT_STOP_TIMER);
                    break;
                case Message.Commands.SOWAndSubscribe:
                    set(CT_SOW_AND_SUBSCRIBE);
                    break;
                case Message.Commands.DeltaPublish:
                    set(CT_DELTA_PUBLISH);
                    break;
                case Message.Commands.DeltaSubscribe:
                    set(CT_DELTA_SUBSCRIBE);
                    break;
                case Message.Commands.SOWAndDeltaSubscribe:
                    set(CT_SOW_AND_DELTA_SUBSCRIBE);
                    break;
                case Message.Commands.SOWDelete:
                    set(CT_SOW_DELETE);
                    break;
                case Message.Commands.GroupBegin:
                    set(CT_GROUP_BEGIN);
                    break;
                case Message.Commands.GroupEnd:
                    set(CT_GROUP_END);
                    break;
                case Message.Commands.OOF: set(CT_OOF);
                    break;
                case Message.Commands.Ack:
                    set(CT_ACK);
                    break;
                case Message.Commands.Heartbeat:
                    set(CT_HEARTBEAT);
                    break;
                case Message.Commands.Flush:
                    set(CT_FLUSH);
                    break;
                case Message.Commands.Unknown:
                default:
                    set(CT_UNKNOWN);
                    break;
            }
        }

        // <summary>
        // Decodes a message command from the given byte buffer to an enum.
        // </summary>
        // <param name="buffer">The byte buffer containing the command.</param>
        // <param name="pos">The position in the buffer where the command starts.</param>
        // <param name="len">The length of the command.</param>
        // <returns>The decoded <see cref="Message.Commands"/>.</returns>
        static Message.Commands decodeCommand(byte[] buffer, int pos, int len)
        {
            switch (len)
            {
                case 1:
                    switch (buffer[pos])
                    {
                        case (byte)'p': return Message.Commands.Publish;
                    }
                    break;
                case 3:
                    // sow oof ack
                    switch (buffer[pos])
                    {
                        case (byte)'s': return Message.Commands.SOW;
                        case (byte)'o': return Message.Commands.OOF;
                        case (byte)'a': return Message.Commands.Ack;
                    }
                    break;
                case 5:
                    // sow oof ack
                    switch (buffer[pos])
                    {
                        case (byte)'l': return Message.Commands.Logon;
                        case (byte)'f': return Message.Commands.Flush;
                    }
                    break;
                case 7:
                    return Message.Commands.Publish;
                case 9:
                    // subscribe heartbeat group_end
                    switch (buffer[pos])
                    {
                        case (byte)'s': return Message.Commands.Subscribe;
                        case (byte)'h': return Message.Commands.Heartbeat;
                        case (byte)'g': return Message.Commands.GroupEnd;
                    }
                    break;
                case 10:
                    // sow_delete stop_timer
                    switch (buffer[pos + 1])
                    {
                        case (byte)'o': return Message.Commands.SOWDelete;
                        case (byte)'t': return Message.Commands.StopTimer;
                    }
                    break;
                case 11:
                    // unsubscribe start_timer group_begin
                    switch (buffer[pos])
                    {
                        case (byte)'u': return Message.Commands.Unsubscribe;
                        case (byte)'s': return Message.Commands.StartTimer;
                        case (byte)'g': return Message.Commands.GroupBegin;
                    }
                    break;
                case 13:
                    return Message.Commands.DeltaPublish;
                case 15:
                    return Message.Commands.DeltaSubscribe;
                case 17:
                    // batch_publish_end sow_and_subscribe
                    switch (buffer[pos])
                    {
                        case (byte)'s': return Message.Commands.SOWAndSubscribe;
                    }
                    break;
                case 23:
                    return Message.Commands.SOWAndDeltaSubscribe;
            }
            return Message.Commands.Unknown;
        }

        /// <summary>
        /// This method encodes the Command. This method takes the
        /// enum Command and finds the corresponding string.
        /// </summary>
        /// <param name="command">The message command to encode.</param>
        /// <returns>The encoded command string.</returns>
        static public String encodeCommand(Message.Commands command)
        {
            switch (command)
            {
                case Message.Commands.Publish: return "publish";
                case Message.Commands.Subscribe: return "subscribe";
                case Message.Commands.Unsubscribe: return "unsubscribe";
                case Message.Commands.SOW: return "sow";
                case Message.Commands.Heartbeat: return "heartbeat";
                case Message.Commands.Logon: return "logon";
                case Message.Commands.StartTimer: return "start_timer";
                case Message.Commands.StopTimer: return "stop_timer";
                case Message.Commands.SOWAndSubscribe: return "sow_and_subscribe";
                case Message.Commands.DeltaPublish: return "delta_publish";
                case Message.Commands.DeltaSubscribe: return "delta_subscribe";
                case Message.Commands.SOWAndDeltaSubscribe: return "sow_and_delta_subscribe";
                case Message.Commands.SOWDelete: return "sow_delete";
                case Message.Commands.GroupBegin: return "group_begin";
                case Message.Commands.GroupEnd: return "group_end";
                case Message.Commands.OOF: return "oof";
                case Message.Commands.Ack: return "ack";
                case Message.Commands.Flush: return "flush";
                case Message.Commands.Unknown: return "unknown";
                default: return "unknown";
            }
        }

    }

}

