////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2012-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace AMPS.Client.Fields
{
    /// <summary>
    /// Represents a single field in an AMPS message.
    /// </summary>
    /// <remarks>Implemented with byte array and int offset
    /// and length primitives to maximize heap efficiency of
    /// cloning, reads, updates, and comparisons.</remarks>
    public class Field
    {
        public byte[] buffer = null;
        public int position = 0;
        public int length = 0;

        /// <summary>
        /// Default construntotor.
        /// </summary>
        public Field()
        {
        }

        /// <summary>
        /// Constructs an instance with the bytes of the specified string
        /// value. The string is interpreted as the UTF-8 representation.
        /// </summary>
        /// <param name="value">The string to be converted into UTF-8 bytes.</param>
        public Field(string value)
        {
            this.buffer = new byte[value.Length];
            this.buffer = System.Text.Encoding.ASCII.GetBytes(value);
            //System.Buffer.BlockCopy(value.ToCharArray(), 0, this.buffer, 0, value.Length);
            this.position = 0;
            this.length = value.Length;
        }

        /// <summary>
        /// Constructs an instance with the specified region of a byte
        /// buffer. Note: the bytes are not copied from the buffer. The
        /// specified buffer region is referenced by this instance.
        /// </summary>
        /// <param name="buffer">The byte buffer.</param>
        /// <param name="position">The starting position of the value in the
        /// byte buffer.</param>
        /// <param name="length">The length from the starting position of
        /// the value in the byte buffer.</param>
        public Field(byte[] buffer, int position, int length)
        {
            set(buffer, position, length);
        }

        /// <summary>
        /// Sets the value of this instance with the specified region of a
        /// byte buffer. Note: the bytes are not copied from the buffer.
        /// The specified buffer region is referenced by this instance.
        /// </summary>
        /// <param name="buffer">The byte buffer.</param>
        /// <param name="position">The starting position of the value in the byte buffer.</param>
        /// <param name="length">The length from the starting position of the value in the byte buffer.</param>
        public void set(byte[] buffer, int position, int length)
        {
            this.buffer = length>0?buffer:null;
            this.position = position;
            this.length = length;
            if (length > 0 && position > buffer.Length) System.Diagnostics.Debug.Fail("here");
        }

        /// <summary>
        /// Sets the value of this instance with the specified buffer.
        /// Note: the bytes are not copied from the buffer. The buffer
        /// is referenced by this instance.
        /// </summary>
        /// <param name="buffer">The byte buffer to be copied.</param>
        public void set(byte[] buffer)
        {
            this.buffer = buffer;
            this.position = 0;
            this.length = buffer.Length;
        }

        /// <summary>
        /// Indicates whether this instance's value is null.
        /// </summary>
        /// <returns>True if null, false if not.</returns>
        public bool isNull()
        {
            return this.buffer == null || this.length == 0;
        }

        /// <summary>
        /// Resets the value of this instance to null.
        /// </summary>
        public void reset()
        {
            this.buffer = null;
            this.length = 0;
        }

        /// <summary>
        /// Returns the byte at {@link #position} plus index i.
        /// </summary>
        /// <param name="i">The given index.</param>
        /// <returns>The specified byte.</returns>
        public byte byteAt(int i)
        {
            return this.buffer[this.position + i];
        }

        /// <summary>
        /// Default implementation for {@link Object#equals} for a
        /// Field. For two Fields to be considered equal, their values
        /// must have the same length and all bytes in the value must
        /// be equal.
        /// </summary>
        /// <param name="obj">The object to compare this with.</param>
        /// <returns>True if the objects are equal, otherwise false.</returns>
        public override bool Equals(Object obj)
        {
            if(obj == null || !(obj is Field))
            {
                return false;
            }

            Field o = (Field)obj;
            if(o.length != length)
                return false;

            for(int i=0; i< length; i++)
            {
                if(o.buffer[o.position+i] != buffer[position+i])
                    return false;
            }
            return true;
        }

        /// <summary>
        /// Gets the hash code for the <see cref="Field"/>.
        /// </summary>
        /// <returns>The hash code for the <see cref="Field"/>.</returns>
        public override int GetHashCode()
        {
            int hash = 0;
            for (int i = position; i < position + length; i++)
            {
                //hash *= 10;
                hash += buffer[i];
            }
            return hash;
        }

        /// <summary>
        /// Constructs a new Field and sets its value to a deep copy
        /// of this instance.
        /// </summary>
        /// <returns>The new deep copy of this instance.</returns>
        public Field copy()
        {
            if (length == 0)
                return new Field();
            byte[] copy = new byte[length];
            Buffer.BlockCopy(buffer, position, copy, 0, length);
            return new Field(copy, 0, length);
        }

        /// <summary>
        /// Provides a way to make this object a duplicate of f.
        /// </summary>
        /// <param name="f">The Field to copy from.</param>
        public virtual void copyFrom(Field f)
        {
            if (buffer == null || f.length > buffer.Length)
            {
                buffer = new byte[f.length];
            }
            Array.Copy(f.buffer, f.position, buffer, 0, f.length);
            position = 0;
            length = f.length;
        }

    }
}
