////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.Text;
namespace AMPS.Client.Fields
{
    /// <summary>
    /// Field data for a AMPS.Client.Message which
    /// consists of options. These options represent the possible
    /// values for message options in an AMPS command.
    /// </summary>
    public class OptionsField : StringField
    {
        /// <summary>
        /// Initializes a new instance of the OptionsField class.
        /// </summary>
        public OptionsField()
        { }

        /// <summary>
        /// Initializes a new instance of the OptionsField class
        /// with the specified options.
        /// </summary>
        /// <param name="options">The options to be set for the field.</param>
        public OptionsField(string options)
            : base(options)
        { }

        /// <summary>
        /// Initializes a new instance of the OptionsField class
        /// with the specified buffer, position, and length.
        /// </summary>
        /// <param name="buffer">The byte array buffer containing the options.</param>
        /// <param name="position">The starting position of the options within
        /// the buffer.</param>
        /// <param name="length">The length of the options.</param>
        protected OptionsField(byte[] buffer, int position, int length)
            : base(buffer, position, length)

        {
        }

        /// <summary>
        /// Gets the value of this options field using the specified decoder.
        /// </summary>
        /// <param name="decoder">The encoding decoder to use.</param>
        /// <returns>The value of the options field as a string.</returns>
        public override string getValue(Encoding decoder)
        {
            // Java code always decoded in LATIN1 and decoder is unused.
            // I'll use decoder instead
            if (!isNull())
            {
                int len = this.length;
                if (len > 0 && buffer[this.position+len-1] == ',') --len;
                try
                {
                    return decoder.GetString(this.buffer, this.position, len);
                }
                catch (DecoderFallbackException)
                {
                    return String.Empty;
                }
            }
            return String.Empty;
        }

        /// <summary>
        /// For two OptionsFields to be considered equal, their values must
        /// have the same length and all bytes in the value must be equal,
        /// other than a possible trailing comma.
        /// </summary>
        /// <param name="obj">The object to compare with this options field.</param>
        /// <returns>True if the options field is equal to the specified object;
        /// otherwise, false.</returns>
        public override bool Equals(Object obj)
        {
            if(obj == null || !(obj is Field))
            {
                return false;
            }

            Field rhs = (Field)obj;
            int len = this.length;
            if (len > 0 && buffer[this.position+len-1] == ',')
                --len;
            int rhsLen = rhs.length;
            if (rhsLen > 0 && rhs.buffer[rhs.position+rhsLen-1] == ',')
                --rhsLen;
            if (len == rhsLen)
            {
                for(int i=0; i<len; i++)
                {
                    if(rhs.buffer[rhs.position+i] != buffer[position+i])
                        return false;
                }
                return true;
            }
            return false;
        }

        /// <summary>
        /// Returns the hash code for this options field.
        /// </summary>
        /// <returns>The hash code.</returns>
        public override int GetHashCode()
        {
            int hash = 0;
            int len = this.length;
            if (len > 0 && buffer[this.position+len-1] == ',')
                --len;
            for (int i = position; i < position + len; i++)
            {
                //hash *= 10;
                hash += buffer[i];
            }
            return hash;
        }

    }
}