////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Text;
using AMPS.Client;


namespace AMPS.Client.Fields
{
    /// <summary>
    /// Field data for a {@link Message} consisting of a reason.
    /// This reason helps to provide more information about
    /// acknowledgements.
    /// </summary>
    public class ReasonField : Field
    {
        protected const String LATIN1 = "ISO-8859-1";

        /// NOTE THESE MUST MATCH the order in the Message.Reasons enum.
        private static String[] reasonStrings = {
             "",
             "duplicate",
             "bad filter",
             "bad regex topic",
             "subscription already exists",
             "deleted",
             "expired",
             "match",
             "invalid topic",
             "name in use",
             "auth failure",
             "not entitled",
             "authentication disabled",
             "invalid bookmark",
             "invalid orderby",
             "subid in use",
             "no topic",
             "logon failed",
             "logon required",
             "invalid topic or filter",
             "invalid subId",
             "no topic or filter",
             "no client name",
             "bad sow key",
             "regex topic not supported",
             "sow store failed",
             "parse error",
             "not supported",
             "tx store failure",
             "duplicate logon attempt",
             "txn replay failed",
             "sow canceled",
             "invalid options",
             "invalid message type",
             "orderby too large",
             "rejected by transport filter",
             "sow_delete command only supports one of: filter, sow_keys, bookmark, or data",
        };

        /// <summary>
        /// Retrieves the reason value from this field.
        /// </summary>
        /// <returns>The reason value.</returns>
        public Message.Reasons getValue()
        {
            if (this.buffer != null)
            {
                return decodeReason(this.buffer, this.position, this.length);
            }
            return Message.Reasons.None;
        }

        /// <summary>
        /// Returns the value of this reason field as a descriptive text string.
        /// </summary>
        /// <returns>The reason string.</returns>
        public String getText()
        {
            if (this.buffer != null)
            {
                return Encoding.ASCII.GetString(this.buffer, this.position, this.length);
            }
            return "";
        }
        private void set(String value)
        {
            set(value.getBytes(LATIN1));
        }

        /// <summary>
        /// Sets the value of this reason field using the specified reason
        /// code integer.
        /// </summary>
        /// <param name="v">The reason code integer.</param>
        public void setValue(Message.Reasons v)
        {
            if (v == 0 || (int)v > reasonStrings.Length - 1)
            {
                reset();
                return;
            }
            set(reasonStrings[(int)v]);
        }

        // <summary>
        // Decodes the reason from the given buffer, position, and length.
        // </summary>
        // <param name="buffer">The buffer containing the reason bytes.</param>
        // <param name="pos">The starting position of the reason bytes.</param>
        // <param name="len">The length of the reason bytes.</param>
        // <returns>The decoded reason.</returns>
        static Message.Reasons decodeReason(byte[] buffer, int pos, int len)
        {
            String reasonBytes = Encoding.ASCII.GetString(buffer, pos, len);
            for (int i = 0; i < reasonStrings.Length; ++i)
            {
                if (reasonStrings[i] == reasonBytes) return (Message.Reasons)i;
            }

            return Message.Reasons.Other;

        }

        /// <summary>
        /// Utility method that takes a reason code integer and returns the
        /// associated descriptive text. Returns "unknown" if the integer
        /// code is not known.
        /// </summary>
        /// <param name="r">The reason code integer to be encoded.</param>
        /// <returns>The descriptive text reason.</returns>
        static public String encodeReason(Message.Reasons r)
        {
            if ((int)r > reasonStrings.Length - 1) return "unknown";
            return reasonStrings[(int)r];
        }
    }

}
