////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;


namespace AMPS.Client.Fields
{
    /// <summary>
    /// Field data for a Message which consists of a status.
    /// This is the status of the Message. This status indicates the
    /// success or failure of the command being acknowledged.
    /// </summary>
    public class StatusField : Field
    {
        protected const String LATIN1 = "ISO-8859-1";
        protected const byte LATIN1_ZERO = 48;

        private static byte[] ST_SUCCESS = null;
        private static byte[] ST_FAILURE = null;
        private static byte[] ST_RETRY = null;

        static StatusField()
        {
            ST_SUCCESS = "success".getBytes(LATIN1);
            ST_FAILURE = "failure".getBytes(LATIN1);
            ST_RETRY = "retry".getBytes(LATIN1);
        }

        /// <summary>
        /// This returns the value of this status field as an
        /// integer status code.
        /// </summary>
        /// <returns>The value of this status field as an integer
        /// status code.</returns>
        public Message.State getValue()
        {
            if (this.buffer != null)
            {
                return decodeStatus(this.buffer, this.position, this.length);
            }
            return Message.State.None;
        }

        /// <summary>
        /// Sets the value of this status field using the specified
        /// status code integer.
        /// </summary>
        /// <param name="v">The status code integer.</param>
        public void setValue(Message.State v)
        {
            if (v == 0)
            {
                reset();
                return;
            }
            switch (v)
            {
                case Message.State.Success:
                    set(ST_SUCCESS);
                    break;
                case Message.State.Failure:
                    set(ST_FAILURE);
                    break;
                case Message.State.Retry:
                    set(ST_RETRY);
                    break;
                default:
                    reset();
                    break;
            }
        }

        // <summary>
        // Decodes the statuc from the given buffer, position, and length.
        // </summary>
        // <param name="buffer">The buffer containing the reason bytes.</param>
        // <param name="pos">The starting position of the reason bytes.</param>
        // <param name="len">The length of the reason bytes.</param>
        // <returns>The decoded status.</returns>
        static Message.State decodeStatus(byte[] buffer, int pos, int len)
        {
            if (len == 7)
            {
                if (buffer[pos] == (byte)'s') return Message.State.Success;
                return Message.State.Failure;
            }
            else if (len == 5)
            {
                return Message.State.Retry;
            }
            return Message.State.None;
        }

        /// <summary>
        /// Utility method that takes a status code integer and returns
        /// the associated descriptive text. Returns "none" if the integer
        /// code is not known.
        /// </summary>
        /// <param name="s">The status code integer to be encoded.</param>
        /// <returns>The descriptive text status.</returns>
        static public String encodeStatus(Message.State s)
        {
            switch (s)
            {
                case Message.State.Success: return "success";
                case Message.State.Failure: return "failure";
                case Message.State.Retry: return "retry";
                case Message.State.None: return "none";
                default: return "none";
            }
        }

    }

}
