////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Text;

namespace AMPS.Client.Fields
{
    /// <summary>
    /// Field data for a Message which consists of a string.
    /// </summary>
    public class StringField : Field
    {
        protected const string LATIN1 = "ISO-8859-1";
        protected const byte LATIN1_ZERO = 48;

        private byte[] bytes = new byte[256];

        /// <summary>
        /// Initializes a new instance of the StringField class with the
        /// specified buffer, position, and length.
        /// </summary>
        /// <param name="buffer">The byte array buffer containing the string
        /// data.</param>
        /// <param name="position">The starting position of the string within
        /// the buffer.</param>
        /// <param name="length">The length of the string.</param>
        protected StringField(byte[] buffer, int position, int length)
            : base(buffer, position, length)
        {
        }

        /// <summary>
        /// Initializes a new instance of the StringField class with
        /// the specified string value.
        /// </summary>
        /// <param name="value">The string value to initialize the
        /// StringField with.</param>
        public StringField(string value)
            : base(value)
        { }

        /// <summary>
        /// Default constructor.
        /// </summary>
        public StringField()
        { }

        /// <summary>
        /// Returns a string representation of the Field value, where the
        /// bytes are interpreted by using the specified decoder.
        /// </summary>
        /// <param name="decoder">The charset decoder used to interpret bytes
        /// as characters.</param>
        /// <returns>A newly-constructed String representing the value of this
        ///          field.</returns>
        public virtual string getValue(Encoding decoder)
        {
            // Java code always decoded in LATIN1 and decoder is unused.
            // I'll use decoder instead
            if (!isNull())
            {
                try
                {
                    return decoder.GetString(this.buffer, this.position, this.length);
                }
                catch (DecoderFallbackException)
                {
                    return String.Empty;
                }
            }
            return String.Empty;
        }

        /// <summary>
        /// Writes the byte value of this instance into the specified byte
        /// buffer, if the value is not null.
        /// </summary>
        /// <param name="v">The byte buffer to write in to.</param>
        /// <returns>Return true if this instance is not null, otherwise
        /// false.</returns>
        public bool getValue(byte[] v)
        {
            if (!isNull())
            {
                Array.Copy(this.buffer, this.position, v, 0, this.length);
                return true;
            }
            return false;
        }

        /// <summary>
        /// Writes the byte value of this instance into the specified
        /// command ID object, if the value is not null.
        /// </summary>
        /// <param name="v">The command ID object to write in to</param>
        /// <returns>Return true if this instance is not null, otherwise
        /// false.</returns>
        public bool getValue(CommandId v)
        {
            if (!isNull())
            {
                v.set(this.buffer, this.position, this.length);
                return true;
            }
            return false;
        }

        /// <summary>
        /// Sets the value of this instance with the specified region of
        /// a byte buffer.
        /// </summary>
        /// <remarks>
        /// Note: the bytes are not copied from the buffer. The specified
        /// buffer region is referenced by this instance.
        /// </remarks>
        /// <param name="v">The byte buffer.</param>
        /// <param name="offset">The starting position of the value in the
        ///                      byte buffer.</param>
        /// <param name="length">The length from the starting position of
        /// the value in the byte buffer.</param>
        public void setValue(byte[] v, int offset, int length)
        {
            this.buffer = v;
            this.position = offset;
            this.length = length;
        }

        /// <summary>
        /// Sets the value of this instance from the specified command ID.
        /// </summary>
        /// <remarks>
        /// Note: the bytes are not copied from the buffer. The specified
        /// buffer region is referenced by this instance. This is not a
        /// deep copy.
        /// </remarks>
        /// <param name="v">The command ID object whose value is used to set
        /// this instance.</param>
        public void setValue(CommandId v)
        {
            this.buffer = v.id;
            this.position = 0;
            this.length = v.id.Length;
        }

        /// <summary>
        /// Sets the value of this instance from the specified string.
        /// The characters of the string are interpreted using the
        /// specified encoder. This method makes its own copy of the
        /// bytes.
        /// </summary>
        /// <param name="v">The string used as the value.</param>
        /// <param name="encoder">The encoder used to interpret the
        /// string's characters as bytes.</param>
        public void setValue(string v, Encoding encoder)
        {
            if (v == null)
            {
                reset();
                return;
            }

            // Check if buffer storage space needs expansion
            int bytesNeeded = encoder.GetMaxByteCount(v.Length);
            if (bytesNeeded > bytes.Length)
            {
                // Must resize
                int newSize = bytes.Length;
                while (bytesNeeded > newSize) newSize *= 2;
                bytes = new byte[newSize];
            }

            // Encode to the byte buffer
            int bytesUsed = 0;

            try
            {
                bytesUsed = encoder.GetBytes(v, 0, v.Length, bytes, 0);
            }
            catch (Exception ex)
            {
                throw new Exceptions.AMPSException("An error occurred while encoding", ex);
            }
            set(bytes, 0, bytesUsed);

            // getBytes is faster than using a CharBuffer, however, there's a bunch of garbage
            //    collected in using this method, so we avoid it. (about a 7% bump in performance, but
            //    under heavy load it creates some GC issues.)
            //try { this.buffer = v.getBytes(LATIN1); } catch(UnsupportedEncodingException e) {;}
        }
    }

}
