////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using AMPS.Client.Exceptions;
namespace AMPS.Client
{

    /// <summary>
    /// FixedDelayStrategy is an implementation that delays for a fixed time
    /// period, as specified in the constructor, when reconnecting to the same
    /// server as we were previously connected to, or if we are invoked again
    /// for the first server we ever tried. ReconnectDelayStrategy
    /// implementations are used by the HAClient.
    /// </summary>
    public class FixedDelayStrategy : ReconnectDelayStrategy
    {
        /// <summary>
        /// A specialized exception type thrown by FixedDelayStrategy
        /// to indicate that the client should "give up" on attempting to
        /// reconnect to a server.
        /// </summary>
        public class MaximumRetryExceeded : AMPSException
        {

        }

        /// <summary>
        /// The default delay when reconnecting to a server, in milliseconds.
        /// </summary>
        public const int DEFAULT_DELAY = 200;

        /// <summary>
        /// Construct a FixedDelayStrategy with a specified delay and maximum
        /// timeout. The delay defaults to the <code>DEFAULT_DELAY</code> and
        /// the maximum timeout defaults to <code>-1</code>, which indicates
        /// that the delay strategy has no maximum, and will continue allowing
        /// the client to connect indefinitely.
        /// </summary>
        /// <param name="delay_">The delay (milliseconds) to be used between
        /// reconnect attempts to the same server.</param>
        /// <param name="maximum_">The maximum time (milliseconds) to attempt
        /// retry before giving up.</param>
        public FixedDelayStrategy(int delay_ = DEFAULT_DELAY, int maximum_ = -1)
        {
            if(delay_ < 0)
            {
                this._delay = DEFAULT_DELAY;
            }
            else
            {
                this._delay = delay_;
            }
            this._maximum = (double)maximum_;
        }

        /// <summary>
        /// Returns the time (in milliseconds) that the client should wait
        /// before connecting to the given server URI.
        /// </summary>
        /// <param name="uri_">The URI to which the client intends to connect.</param>
        /// <returns>Returns the time in milliseconds the client should delay before
        /// connecting to uri_.</returns>
        /// <exception cref="MaximumRetryExceeded">Any exception thrown indicates no
        /// connection should be attempted.</exception>
        public int getConnectWaitDuration(String uri_)
        {
            int delay = 0;
            if(_lastUri != uri_)
            {
                if (_lastUri == null)
                {
                    _endTime = DateTime.Now.AddMilliseconds(_maximum);
                }

                // Do not wait to fail over; reconnect immediately.
                _lastUri = uri_;
                if(_firstUri == uri_)
                {
                    // We have wrapped around the "list" of servers. Delay.
                    delay = _delay;
                }
                else if(_firstUri == null)
                {
                    // This is the first time through. Mark uri_ as the first
                    // of the assumed "list" of servers.
                    _firstUri = uri_;
                }
            }
            else
            {
                delay = _delay;
            }
            if(_maximum > 0.0 && DateTime.Now.CompareTo(_endTime) > 0)
            {
                throw new MaximumRetryExceeded();
            }

            return delay;
        }

        /// <summary>
        /// Resets the state of this delay. AMPS calls this method when a
        /// connection is successfully established.
        /// </summary>
        public void reset()
        {
            _lastUri  = null;
            _firstUri = null;
        }

        protected int      _delay;
        protected double   _maximum;
        protected DateTime _endTime;
        protected String   _lastUri;
        protected String   _firstUri;
    }
}

