///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using AMPS.Client.Fields;
using AMPS.Client.Exceptions;

namespace AMPS.Client
{
    /// <summary>
    /// Interface to provide a recovery point for a bookmark subscription.
    /// Used in the RecoveryPointAdapter interface for updates and recovery.
    /// </summary>
    public class FixedRecoveryPoint : SequencedRecoveryPoint
    {
        private Field _subId = null;
        private BookmarkField _bookmark = null;

        /// <summary>
        /// Initializes a new instance of the FixedRecoveryPoint class.
        /// </summary>
        public FixedRecoveryPoint() : base()
        {
            _subId = new Field();
            _bookmark = new BookmarkField();
        }

        /// <summary>
        /// Initializes a new instance of the FixedRecoveryPoint class with
        /// the specified subscription ID and bookmark.
        /// </summary>
        /// <param name="subId_">The subscription ID.</param>
        /// <param name="bookmark_">The bookmark.</param>
        public FixedRecoveryPoint(Field subId_, BookmarkField bookmark_) : base()
        {
            _subId = subId_;
            _bookmark = bookmark_;
        }

        private FixedRecoveryPoint(Field subId_, BookmarkField bookmark_,
                                   ulong sequence_)
          : base(sequence_)
        {
            _subId = subId_;
            _bookmark = bookmark_;
        }

        /// <summary>
        /// Gets the subscription ID associated with this recovery point.
        /// </summary>
        /// <returns>The subscription ID.</returns>
        public override Field getSubId()
        {
            return _subId;
        }

        /// <summary>
        /// Gets the bookmark associated with this recovery point.
        /// </summary>
        /// <returns>The bookmark.</returns>
        public override BookmarkField getBookmark()
        {
            return _bookmark;
        }

        /// <summary>
        /// Creates a deep copy of this recovery point.
        /// </summary>
        /// <returns>A deep copy of this recovery point.</returns>
        public override RecoveryPoint copy()
        {
            return new FixedRecoveryPoint(_subId.copy(), _bookmark.copy(),
                                          getSequence());
        }
    }

    /// <summary>
    /// class for creating instances of FixedRecoveryPoint.
    /// </summary>
    public class FixedRecoveryPointFactory : RecoveryPointFactory
    {
        /// <summary>
        /// Initializes a new instance of the FixedRecoveryPointFactory class.
        /// </summary>
        public FixedRecoveryPointFactory()
        {
        }

        /// <summary>
        /// Creates a new instance of FixedRecoveryPoint with the specified subscription ID and bookmark.
        /// </summary>
        /// <param name="subId_">The subscription ID.</param>
        /// <param name="bookmark_">The bookmark.</param>
        /// <returns>A new instance of FixedRecoveryPoint.</returns>
        public RecoveryPoint createRecoveryPoint(Field subId_,
                                                 BookmarkField bookmark_)
        {
            return new FixedRecoveryPoint(subId_, bookmark_);
        }
    }
}


