////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
using AMPS.Client.Exceptions;
using AMPS.Client;
using System.Threading;

namespace AMPS.Client
{

    /// <summary>
    /// A highly-available AMPS Client object that automatically reconnects and
    /// re-subscribes to AMPS instances upon disconnect.
    /// </summary>
    /// <p>An HAClient object provides failover and resubscription functionality
    /// by default. For reliable publish, set a <see cref="PublishStore"/> for the HAClient
    /// (and, in most cases, set a <see cref="FailedWriteHandler"/> as well so the application can detect
    /// any failed publishes). For managing transaction log replays, set a
    /// <see cref="BookmarkStore"/>.
    /// </p>
    /// <p>To connect to AMPS, you must provide a <see cref="ServerChooser"/> implementation
    /// to the HAClient. By default, the HAClient provides a <see cref="ExponentialDelayStrategy"/>
    /// with the default delay and backoff behavior for that class.
    /// </p>
    /// <p> By default, an HAClient object has the reconnect and resubscribe functionality
    /// provided by the <see cref="MemorySubscriptionManager"/> and the <see cref="HADisconnectHandler"/>
    /// classes. It is typically not necessary to replace the disconnect handler or subscription manager
    /// for this class. Notice that replacing either of the default settings will change the
    /// failover, resubscription, and republish behavior of the HAClient.
    /// </p>
    /// <p>See the <em>Developer Guide</em> for more information on using the HAClient.</p>

    public class HAClient : Client
    {
        private bool _disconnected = true;
        private const int DefaultTimeout = 10000;
        private int _timeout = DefaultTimeout;
        private ReconnectDelayStrategy _delayStrategy;
        private string _options = null;
        private object _haInternalLock = new object();

        /// <summary>Sets the time delay between reconnection events.</summary>
        /// <param name="reconnectInterval">delay between reconnection events.</param>
        /// When called, this method creates and installs a new <typeparamref name="FixedDelayStrategy"/>
        /// in the client.
        public void setReconnectDelay(int reconnectInterval)
        {
            lock(internalLock)
            {
                _delayStrategy = new FixedDelayStrategy(reconnectInterval);
            }
        }

        /// <summary>Sets the <typeparamref name="ReconnectDelayStrategy"/> for
        ///         this client, replacing any previous strategy.</summary>
        /// <param name="strategy_">the strategy to install in this client</param>
        public HAClient setReconnectDelayStrategy(
            ReconnectDelayStrategy strategy_)
        {
            lock(internalLock)
            {
                if (strategy_ != null)
                {
                    _delayStrategy = strategy_;
                }
                else
                {
                    _delayStrategy = new ExponentialDelayStrategy();
                }
                return this;
            }
        }

        /// <summary>
        /// Sets the options string passed in during logon.
        /// </summary>
        /// <param name="options_">logon options</param>
        public void setLogonOptions(string options_)
        {
            lock(internalLock)
            {
                this._options = options_;
            }
        }

        /// <summary>
        /// Gets the options string passed in during logon.
        /// </summary>
        /// <returns>logon options</returns>
        public string getLogonOptions()
        {
            lock(internalLock)
            {
                return this._options;
            }
        }

        /// <summary>
        /// Sets the timeout for logon and re-subscribe sequence during a reconnection event.
        /// </summary>
        /// <param name="timeout_">logon and re-subscribe timeout, where 0 indicates no timeout</param>
        public void setTimeout(int timeout_)
        {
            lock(internalLock)
            {
                this._timeout = timeout_;
            }
        }

        /// <summary>
        /// Gets the timeout for logon and re-subscribe sequence during a reconnection event.
        /// </summary>
        /// <returns>logon and re-subscribe timeout</returns>
        public int getTimeout()
        {
            lock(internalLock)
            {
                return this._timeout;
            }
        }

        /// <summary>
        /// The disconnect handler implementation used by HAClient.
        /// </summary>
        /// <returns>logon and re-subscribe timeout</returns>
        public class HADisconnectHandler : ClientDisconnectHandler
        {
            public virtual void invoke(Client client)
            {
                HAClient haClient = (HAClient) client;
                haClient.getServerChooser().reportFailure(
                new DisconnectedException("Disconnected"), client.getConnectionInfo());

                try
                {
                    haClient.connectAndLogonInternal();
                }
                catch (Exception e)
                {
                    if (client.getExceptionListener() != null)
                    {
                        client.getExceptionListener()(e);
                    }
                    return;
                }

            }

            public HADisconnectHandler()
            {
            }
        }

        /// <summary>
        /// Returns the current <typeparamref name="ServerChooser"/> for this client
        /// </summary>
        /// <returns></returns>
        public ServerChooser getServerChooser()
        {
            lock(internalLock)
            {
                return this._serverChooser;
            }
        }

        // A ServerChooser that fails to connect or do anything useful. Can be
        // used instead of setting to null so we don't have to check for null.
        private class UnsetServerChooser : ServerChooser
        {
            public UnsetServerChooser() { }
            public string getCurrentURI()
            {
                throw new ConnectionException(
                    "No server chooser registered with this HAClient.");
            }
            public Authenticator getCurrentAuthenticator()
            {
                return null;
            }
            public void reportFailure(Exception exception, ConnectionInfo ci)
            {
                throw new ConnectionException(
                    "No server chooser registered with this HAClient.");
            }
            public String getError()
            {
                return "No server chooser registered with this HAClient.";
            }
            public void reportSuccess(ConnectionInfo ci)
            {
                throw new ConnectionException(
                    "No server chooser registered with this HAClient.");
            }
        }

        /// <summary>
        /// Sets the <typeparamref name="ServerChooser"/> for this client
        /// </summary>
        /// <param name="serverChooser_"><typeparamref name="ServerChooser"/> to set</param>
        public void setServerChooser(ServerChooser serverChooser_)
        {
            lock(internalLock)
            {
                if (serverChooser_ != null)
                {
                    this._serverChooser = serverChooser_;
                }
                else
                {
                    this._serverChooser = new UnsetServerChooser();
                }
            }
        }

        ServerChooser _serverChooser = new UnsetServerChooser();

        /// <summary>
        /// Constructs a new "highly-available" <typeparamref name="HAClient"/>
        /// This <typeparamref name="HAClient"/> will automatically reconnect and re-subscribe
        /// in the event of a disconnect.
        /// </summary>
        /// <param name="name">A name to identify this client.
        /// This name is used for duplicate message detection and should be
        /// unique. AMPS does not enforce specific restrictions on the
        /// character set used, however some protocols (for example, XML) may
        /// not allow specific characters. 60East recommends that the client
        /// name be meaningful, short, human readable, and avoid using control
        /// characters, newline characters, or square brackets.</param>
        public HAClient(string name) : base(name)
        {
            base.setDisconnectHandler(new HADisconnectHandler());
            base.setSubscriptionManager(new MemorySubscriptionManager());
            _delayStrategy = new ExponentialDelayStrategy();
        }

        /// <summary>
        /// Creates a new memory-backed highly-available <typeparamref name="HAClient"/>.
        /// This <typeparamref name="HAClient"/> will automatically reconnect and re-subscribe
        /// in the event of a disconnect, and uses memory-based stores to ensure
        /// messages are not lost and duplicates are processed appropriately.
        /// This is equivalent to creating an HAClient and setting a
        /// MemoryBookmarkStore and a MemoryPublishStore on the client,
        /// where the MemoryPublishStore will have a capacity of
        /// 10000 blocks. If your application does not both publish to AMPS and
        /// use a bookmark replay, or if a capacity of 10000 blocks isn't a
        /// useful size for your <typeparamref name="MemoryPublishStore"/>,
        /// constructing an <typeparamref name="HAClient"/> and setting the
        /// stores explicitly is recommended.
        /// </summary>
        /// <param name="name">A name to identify this client.
        /// This name is used for duplicate message detection and should be
        /// unique. AMPS does not enforce specific restrictions on the
        /// character set used, however some protocols (for example, XML) may
        /// not allow specific characters. 60East recommends that the client
        /// name be meaningful, short, human readable, and avoid using control
        /// characters, newline characters, or square brackets.</param>
        /// <returns>A new memory backed <typeparamref name="HAClient"/> instance</returns>
        public static HAClient createMemoryBacked(string name)
        {
            HAClient client = new HAClient(name);
            client.setBookmarkStore(new MemoryBookmarkStore());
            client.setPublishStore(new MemoryPublishStore(10000));

            return client;
        }

        /// <summary>
        /// Creates a new "highly-available" <typeparamref name="HAClient"/>
        /// backed by disk. This <typeparamref name="HAClient"/> will
        /// automatically reconnect and re-subscribe in the event of a disconnect.
        /// It uses disk to ensure messages are not lost and duplicates are
        /// processed appropriately. This is equivalent to creating an
        /// HAClient and setting a LoggedBookmarkStore and a PublishStore
        /// on the client. If your application does not both publish to AMPS and
        /// use a bookmark replay, constructing an <typeparamref name="HAClient"/>
        /// and setting the stores explicitly is recommended.
        /// </summary>
        /// <param name="name">A name to identify this client.
        /// This name is used for duplicate message detection and should be
        /// unique. AMPS does not enforce specific restrictions on the
        /// character set used, however some protocols (for example, XML) may
        /// not allow specific characters. 60East recommends that the client
        /// name be meaningful, short, human readable, and avoid using control
        /// characters, newline characters, or square brackets.</param>
        /// <param name="publishLog">File path to store the publisher log</param>
        /// <param name="subscribeLog">File path to the subscriber log</param>
        public static HAClient createFileBacked(string name, string publishLog,
                string subscribeLog)
        {
            HAClient client = new HAClient(name);
            client.setBookmarkStore(new LoggedBookmarkStore(subscribeLog));
            client.setPublishStore(new PublishStore(publishLog));
            return client;
        }



        /// <summary>
        /// Connects to the next server chosen for us by our {@link ServerChooser}.
        /// Will continue attempting to connect and logon to each successive URI
        /// returned by the ServerChooser until the connection succeeds or the
        /// ServerChooser returns an empty URI.
        /// </summary>
        /// <exception cref="ConnectionException">Connection to AMPS instance failed</exception>
        public void connectAndLogon()
        {
            lock(_haInternalLock)
            {
                if (!this._disconnected)
                {
                    throw new AlreadyConnectedException("HAClient is already connected or actively trying to reconnect.");
                }
                this._disconnected = false;
                _delayStrategy.reset();
                while(true)
                {
                    connectAndLogonInternal();
                    // When manually called, we must re-subscribe manually as well.
                    try
                    {
                        this.getSubscriptionManager().resubscribe(this);
                        broadcastConnectionStateChangedEvent(new ConnectionStateChangedEventArgs(ConnectionStateChangedEventArgs.ConnectionState.Resubscribed));
                        return;
                    } catch (AMPSException subException)
                    {
                        base._disconnect(true);
                        this._serverChooser.reportFailure(subException,getConnectionInfo());
                    }
                }
            }
        }

        // The real implementation of connectAndLogon. Clears the disconnect
        // handler and ack timouts. Gets net URI, get delay, sleeps for delay
        // attempts to connect, attempts to logon (which will take care of
        // republish and starting heartbeats). Reports success or failure to
        // the chooser. Retries if there is a failure. Resets the disconnect
        // handler and ack timeout if success. The loop ends when:
        // successfully connected and logged on, server chooser returns null
        // for the current URI, or an exception is thrown by the reconnect
        // delay strategy.
        private void connectAndLogonInternal()
        {
            lock(_haInternalLock)
            {
                ClientDisconnectHandler currentHandler = base.getDisconnectHandler();
                base.setDisconnectHandler(new DefaultDisconnectHandler());
                long ackTimeout = base.getAckTimeout();
                base.setAckTimeout(0);
                try
                {
                    while (!this._disconnected)
                    {
                        string uri = _serverChooser.getCurrentURI();

                        if (uri == null)
                        {
                            this._disconnected = true;
                            throw new ConnectionException(
                                "No AMPS instances available for connection. " + _serverChooser.getError());
                        }
                        int sleepTime = 0;
                        try
                        {
                            sleepTime = _delayStrategy.getConnectWaitDuration(uri);
                        }
                        catch (Exception e)
                        {
                            this._disconnected = true;
                            throw new ConnectionException(
                            "The reconnect delay strategy indicated that no more " +
                            "reconnect attempts should be made. The last connection "+
                            "error was: " + _serverChooser.getError(), e);
                        }
                        try
                        {
                            Thread.Sleep(sleepTime);
                            Authenticator authenticator = _serverChooser
                                .getCurrentAuthenticator();
                            lock(internalLock) {
                                try
                                {
                                    if (_disconnected)
                                    {
                                        return;
                                    }
                                    base.connect(uri);
                                    base.logon(_timeout, authenticator, _options);
                                }
                                catch (ConnectionException connEx)
                                {
                                    Exception exc = connEx.InnerException;
                                    if (exc != null
                                        && (exc is AuthenticationException
                                            || exc is DuplicateLogonException
                                            || exc is NameInUseException
                                            || exc is NotEntitledException
                                            || exc is StoreException
                                            || exc is TimedOutException))
                                    {
                                        try
                                        {
                                            base._disconnect(true);
                                        }
                                        catch (Exception ignoredExc)
                                        {
                                            absorbedException(ignoredExc);
                                        }
                                    }
                                    throw;
                                }
                            }
                            _serverChooser.reportSuccess(getConnectionInfo());
                            _delayStrategy.reset();
                            return;
                        }
                        catch (StoreException se)
                        {
                            try
                            {
                                base._disconnect(true);
                            }
                            catch (Exception e)
                            {
                                absorbedException(e);
                            }
                            absorbedException(se);
                            return;
                        }
                        catch (Exception exception)
                        {
                            ConnectionInfo ci = getConnectionInfo();
                            ci["client.uri"] = uri;
                            _serverChooser.reportFailure(exception, ci);
                            // Try disconnecting -- we might be partially connected?
                            try
                            {
                                base._disconnect(true);
                            }
                            catch (Exception e)
                            {
                                absorbedException(e);
                            }
                        }
                    }
                }
                finally
                {
                    base.setDisconnectHandler(currentHandler);
                    base.setAckTimeout(ackTimeout);
                }
            }
        }

        /// <summary> Synonym for connectAndLogon, ignoring the URI. </summary>
        /// <param> uri - ignored.</param>
        override public void connect(string uri)
        {
            connectAndLogon();
        }

        /// <summary> Disaalowed for HAClient.</summary>
        /// <exception>DisconnectedException - client is not connected</exception>
        override public CommandId logon()
        {
            if (this._disconnected)
                throw new DisconnectedException("Attempt to logon a disconnected HAClient. Call connectAndLogon() instead.");
            throw new AlreadyConnectedException("Attempt to logon an HAClient. Call connectAndLogon() instead.");
        }

        /// <summary> Disaalowed for HAClient.</summary>
        /// <param name="timeout">Ignored</param>
        /// <exception>DisconnectedException - client is not connected</exception>
        /// <exception>AlreadyConnectedException - client is already connected</exception>
        override public CommandId logon(long timeout)
        {
            if (this._disconnected)
                throw new DisconnectedException("Attempt to logon a disconnected HAClient. Call connectAndLogon() instead.");
            throw new AlreadyConnectedException("Attempt to logon an HAClient. Call connectAndLogon() instead.");
        }

        /// <summary> Disaalowed for HAClient.</summary>
        /// <param name="timeout">Ignored</param>
        /// <param name="authenticator">Ignored</param>
        /// <exception>DisconnectedException - client is not connected</exception>
        /// <exception>AlreadyConnectedException - client is already connected</exception>
        override public CommandId logon(long timeout, Authenticator authenticator)
        {
            if (this._disconnected)
                throw new DisconnectedException("Attempt to logon a disconnected HAClient. Call connectAndLogon() instead.");
            throw new AlreadyConnectedException("Attempt to logon an HAClient. Call connectAndLogon() instead.");
        }

        /// <summary> Disaalowed for HAClient.</summary>
        /// <param name="options">Ignored</param>
        /// <exception>DisconnectedException - client is not connected</exception>
        /// <exception>AlreadyConnectedException - client is already connected</exception>
        override public CommandId logon(string options)
        {
            if (this._disconnected)
                throw new DisconnectedException("Attempt to logon a disconnected HAClient. Call connectAndLogon() instead.");
            throw new AlreadyConnectedException("Attempt to logon an HAClient. Call connectAndLogon() instead.");
        }

        /// <summary> Disaalowed for HAClient.</summary>
        /// <param name="timeout">Ignored</param>
        /// <param name="options">Ignored</param>
        /// <exception>DisconnectedException - client is not connected</exception>
        /// <exception>AlreadyConnectedException - client is already connected</exception>
        override public CommandId logon(long timeout, string options)
        {
            if (this._disconnected)
                throw new DisconnectedException("Attempt to logon a disconnected HAClient. Call connectAndLogon() instead.");
            throw new AlreadyConnectedException("Attempt to logon an HAClient. Call connectAndLogon() instead.");
        }

        /// <summary> Disaalowed for HAClient.</summary>
        /// <param name="timeout">Ignored</param>
        /// <param name="options">Ignored</param>
        /// <param name="authenticator">Ignored</param>
        /// <exception>DisconnectedException - client is not connected</exception>
        /// <exception>AlreadyConnectedException - client is already connected</exception>
        override public CommandId logon(long timeout, Authenticator authenticator, string options)
        {
            if (this._disconnected)
                throw new DisconnectedException("Attempt to logon a disconnected HAClient. Call connectAndLogon() instead.");
            throw new AlreadyConnectedException("Attempt to logon an HAClient. Call connectAndLogon() instead.");
        }

        /// <summary>
        /// Disconnect the HAClient from any current connection and don't
        /// attempt any further reconnects.
        /// </summary>
        override public void disconnect()
        {
            lock(internalLock)
            {
                this._disconnected = true;
            }
            base.disconnect();
        }
        ///
        /// <summary>
        /// Assembles a new ConnectionInfo with the state of this client and
        /// associated classes.
        /// </summary>
        /// <returns>A new ConnectionInfo object.</returns>
        ///
        public override ConnectionInfo getConnectionInfo()
        {
            lock(internalLock)
            {
                ConnectionInfo ci = base.getConnectionInfo();
                ci["haClient.timeout"] = this._timeout;
                return ci;
            }
        }
    }
}
