﻿using System;
using System.IO;
using System.Net;
using System.Net.Sockets;

namespace AMPS.Client
{
    /// <summary>A Stream implementation that wraps a NetworkStream to keep timeout
    /// exceptions from escaping Read and causing the connection to close </summary>
    public class HandleTimeoutStream : Stream
    {
        private NetworkStream _stream;
        private TCPTransportImpl _transport;

        /// <summary>Create a wrapper around stream_ that will notify transport_
        /// of any timeouts and swallow those exceptions.</summary>
        /// <param name="stream_"> The NetworkStream to wrap and prevent timeout
        /// exceptions from escaping</param>
        /// <param name="transport_">The TCPTransportImpl to notify of timeout
        /// events when they occur.</param>
        public HandleTimeoutStream(NetworkStream stream_, TCPTransportImpl transport_)
        {
            _stream = stream_;
            _transport = transport_;
        }

        public override int Read(byte[] buffer, int offset, int count)
        {
            while (true)
            {
                try
                {
                    return _stream.Read(buffer, offset, count);
                }
                catch (IOException ex)
                {
                    SocketException sockEx = ex.InnerException as SocketException;
                    if (sockEx != null
                        && sockEx.SocketErrorCode == SocketError.TimedOut)
                    {
                        // If we throw here, SSLStream gets corrupted, so return
                        // our signal value.
                        _transport.handleExternalTimeoutEvent();
                        continue;
                    }
                    throw;
                }
                catch (Exception)
                {
                    throw;
                }
            }
        }


        public override bool CanRead => _stream.CanRead;
        public override bool CanSeek => _stream.CanSeek;
        public override bool CanTimeout => _stream.CanTimeout;
        public override bool CanWrite => _stream.CanWrite;
        public virtual bool DataAvailable => _stream.DataAvailable;
        public override long Length => _stream.Length;
        public override IAsyncResult BeginRead(byte[] buffer, int offset, int size, AsyncCallback callback, object state) => _stream.BeginRead(buffer, offset, size, callback, state);
        public override IAsyncResult BeginWrite(byte[] buffer, int offset, int size, AsyncCallback callback, object state) => _stream.BeginWrite(buffer, offset, size, callback, state);
        public void Close(int timeout) => _stream.Close(timeout);
        public override int EndRead(IAsyncResult asyncResult) => _stream.EndRead(asyncResult);
        public override void EndWrite(IAsyncResult asyncResult) => _stream.EndWrite(asyncResult);
        public override void Flush() => _stream.Flush();
#if (NETSTANDARD2_0 || NETSTANDARD2_1)
        public override System.Threading.Tasks.Task FlushAsync(System.Threading.CancellationToken cancellationToken) => _stream.FlushAsync(cancellationToken);
#endif
        public override long Seek(long offset, SeekOrigin origin) => _stream.Seek(offset, origin);
        public override void SetLength(long value) => _stream.SetLength(value);
        public override void Write(byte[] buffer, int offset, int count) => _stream.Write(buffer, offset, count);

        public override long Position
        {
            get { return _stream.Position; }
            set { _stream.Position = value; }
        }

        public override int ReadTimeout
        {
            get { return _stream.ReadTimeout; }
            set { _stream.ReadTimeout = value; }
        }

        public override int WriteTimeout
        {
            get { return _stream.WriteTimeout; }
            set { _stream.WriteTimeout = value; }
        }
    }
}
