////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.Runtime.CompilerServices;
using System.Diagnostics;
using System.Threading;
using AMPS.Client.Exceptions;

/**
 * PublishStore that stores first in memory, and swaps excess out to disk.
 *
 */
namespace AMPS.Client
{
    /// <summary>
    /// A Store implementation that uses memory to hold the newest
    /// messages and rolls older messages over to a file.
    /// </summary>
    public class HybridPublishStore : Store
    {
        protected PublishStore _fileStore;
        protected MemoryPublishStore _memoryStore;
        protected int _cap;
        protected string _path;
        protected int _lowWatermark = 0;
        private bool _holdSwapping = false;
        private Object _internalLock = new Object();

        // Used internally to move messages from memory to file
        private class SwappingOutReplayer : StoreReplayer
        {
            HybridPublishStore _store = null;
            int _numToSwap = 0;
            int _errorCount = 0;
            StoreException _exception = null;
            ulong _lastHASequence = 0;
            public SwappingOutReplayer(HybridPublishStore store_, int numToSwap_)
            {
                _store = store_;
                _numToSwap = numToSwap_;
            }
            public int getErrors()
            {
                if (_exception != null)
                {
                    throw _exception;
                }
                return _errorCount;
            }
            public ulong lastHASequence()
            {
                return _lastHASequence;
            }

            public void execute(Message message)
            {
                if (_numToSwap > 0 && _errorCount == 0)
                {
                    try
                    {
                        // Sequence numbers are assigned by memory store
                        _store._fileStore.store(message, false);
                        _lastHASequence = message.getSequence();
                    }
                    catch (StoreException e)
                    {
                        ++_errorCount;
                        _exception = e;
                    }
                    --_numToSwap;
                }
            }
        }

        /// <summary>
        /// Create a new HybridPublishStore using a <@link PublishStore> file
        /// named path, a max capacity of cap for its <@link MemoryPublishStore>.
        /// </summary>
        /// <param name="path">The file location for the file-based portion of the store.</param>
        /// <param name="cap">The capacity on in-memory messages before offloading to file.</param>
        /// <param name="errorOnPublishGap_">Indicates if an excpetion should be thrown if discardUpTo is called with an idex below the last persisted.</param>
        public HybridPublishStore(String path, int cap, bool errorOnPublishGap = false)
        {
            _fileStore = new PublishStore(path, errorOnPublishGap);
            _memoryStore = new MemoryPublishStore(cap);
            // Synchronize stores' last persisted values if file
            // store is empty.
            if (_fileStore.getLastPersisted() % 1000000 == 0)
                _fileStore.discardUpTo(_memoryStore.getLastPersisted());
            _cap = cap;
            _lowWatermark = (int)(_cap * 0.5);
            _path = path;
            // Synchronize the last persisted value if file store is empty
            if (_fileStore.getLastPersisted() % 1000000 == 0)
            {
                _fileStore.discardUpTo(_memoryStore.getLastPersisted());
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting resources.
        /// This method has no effect for the BlockPublishStore class.
        /// </summary>
        public void Dispose()
        {
            _fileStore.Dispose();
            _memoryStore.Dispose();
        }

        /// <summary>
        /// Set if a PublishStoreGapException should be thrown by the Store if
        /// there is an attempt to call `discardUpTo` with a sequence number
        /// lower than the last persisted sequence number. This could occur
        /// during logon after a failover event if the new server was not
        /// current on messages due to being an async replication destination.
        /// </summary>
        /// <param name="errorOnPublishGap_">Indicates if the exception should be thrown</param>
        ///
        public void setErrorOnPublishGap(bool errorOnPublishGap_)
        {
            // Don't set this on the memory store, as it always has only the most recent items
            // _memoryStore.setErrorOnPublishGap(errorOnPublishGap_);
            _fileStore.setErrorOnPublishGap(errorOnPublishGap_);
        }

        /// <summary>
        /// Get if a PublishStoreGapException can be thrown by the Store if
        /// there is an attempt to call `discardUpTo` with a sequence number
        /// lower than the last persisted sequence number. This could occur
        /// during logon after a failover event if the new server was not
        /// current on messages due to being an async replication destination.
        /// </summary>
        /// <returns>true if the exception will be thrown</returns>
        ///
        public bool getErrorOnPublishGap()
        {
            return _fileStore.getErrorOnPublishGap();
        }

        /// <summary>
        /// Sets the low watermark; once we start swapping out to disk, we
        /// keep going until the number of entries in memory is lower than this.
        /// </summary>
        /// <param name="lowWatermark_">The number of entries to serve as a low watermark.</param>
        public void setLowWatermark(int lowWatermark_)
        {
            lock (_internalLock)
            {
                if (lowWatermark_ < _cap)
                {
                    _lowWatermark = lowWatermark_;
                }
            }
        }

        /// <summary>
        /// Gets the lowest unpersisted sequence number
        /// </summary>
        /// <returns> The lowest unpersisted sequence number or -1 if no sequence numbers are unpersisted. </returns>
        public ulong getLowestUnpersisted()
        {
            ulong count = _fileStore.getLowestUnpersisted();
            if (count == ulong.MaxValue)
                return _memoryStore.getLowestUnpersisted();
            return count;
        }


        /// <summary>
        /// This method is repsonsible for managing the discarding of blocks in a store based on the
        /// specified index. It handles locking, updating metadata, and organizng the blocks for efficient reuse.
        /// </summary>
        /// <param name="index"></param>
        /// <exception cref="StoreException"></exception>
        public void discardUpTo(ulong index)
        {
            lock (_internalLock)
            {
                while (_holdSwapping)
                {
                    Monitor.Wait(_internalLock, 10);
                }
                _holdSwapping = true;
            }
            try
            {
                if (index >= _memoryStore.getLowestUnpersisted())
                    _memoryStore.discardUpTo(index);
                _fileStore.discardUpTo(index);
            }
            finally
            {
                lock (_internalLock)
                {
                    _holdSwapping = false;
                    Monitor.PulseAll(_internalLock);
                }
            }
        }

        /// <summary>
        /// Replays the stored blocks onto the provided StoreReplayer.
        /// </summary>
        /// <param name="replayer">The StoreReplayer onto which the blocks will be replayed.</param>
        public virtual void replay(StoreReplayer replayer)
        {
            // First replay everything on disk, then replay everything in memory.  Simple!
            lock (_internalLock)
            {
                while (_holdSwapping)
                {
                    Monitor.Wait(_internalLock, 10);
                }
                _holdSwapping = true;
            }
            try
            {
                _fileStore.replay(replayer);
                _memoryStore.replay(replayer);
            }
            finally
            {
                lock (_internalLock)
                {
                    _holdSwapping = false;
                    Monitor.PulseAll(_internalLock);
                }
            }
        }

        /// <summary>
        /// Replays a single stored block with the specified <paramref name="index"/> onto the provided StoreReplayer.
        /// Checks if the specified <paramref name="index"/> is less than or equal to the stored metadata block's sequence number.
        /// </summary>
        /// <param name="replayer">The StoreReplayer onto which the block will be replayed.</param>
        /// <param name="index">The sequence number of the block to be replayed.</param>
        /// <returns>True if the block is successfully replayed, false otherwise.</returns>
        public bool replaySingle(StoreReplayer replayer, ulong index)
        {
            if (index < _memoryStore.getLowestUnpersisted())
            {
                return _fileStore.replaySingle(replayer, index);
            }
            else
            {
                return _memoryStore.replaySingle(replayer, index);
            }
        }

        /// <summary>
        /// Returns the count of unpersisted messages in the store. This count represents the number
        /// of messages that have been received but not yet persisted.
        /// </summary>
        /// <returns> The count of unpersisted messages. </returns>
        public long unpersistedCount()
        {
            return _fileStore.unpersistedCount() + _memoryStore.unpersistedCount();
        }

        /// <summary>
        /// Stores a message in the BlockPublishStore.
        /// </summary>
        /// <param name="m"> The message to be stored. </param>
        /// <exception cref="StoreException"> If an error occurs during message storage. </exception>
        public void store(Message message)
        {
            SwappingOutReplayer replayer = null;
            lock (_internalLock)
            {
                while (_holdSwapping) { Monitor.Wait(_internalLock, 10); }
                if(_memoryStore.unpersistedCount() >= _cap)
                {
                    _holdSwapping = true;
                    // ok, swap some out to disk now.
                    replayer = new SwappingOutReplayer(this, (int)_memoryStore.unpersistedCount() - _lowWatermark);
                }
            }
            try
            {
                if (replayer != null)
                {
                    _memoryStore.replay(replayer);
                    if(replayer.getErrors() == 0)
                    {
                        _memoryStore.discardUpTo(replayer.lastHASequence());
                    }
                }
                _memoryStore.store(message);
            }
            catch(DisconnectedException)
            {
                // This "cannot" happen, as our replayer does not throw DisconnectedException
                Debug.Assert(false);
            }
            finally
            {
                lock(_internalLock)
                {
                    if (replayer != null) _holdSwapping = false;
                }
            }
        }

        /// <summary>
        /// Used internally by the Client to give the store a Message
        /// implementing the correct protocol for the connection.
        /// </summary>
        /// <param name="m">A Message of the correct protocol type.</param>
        public void setMessage(Message m)
        {
            _memoryStore.setMessage(m);
            _fileStore.setMessage(m.copy());
        }

        /// <summary>
        /// Gets the last persisted sequence number
        /// </summary>
        /// <returns> The last persisted sequence number. </returns>
        public ulong getLastPersisted()
        {
            return _fileStore.getLastPersisted();
        }

        /// <summary>
        /// Wait for the store to discard everything that has been stored up to the point in time
        /// when flush is called, then return.
        /// </summary>
        public void flush()
        {
            // Flush memory first
            _memoryStore.flush();
            // In case flush only passed from swapping to file, flush file
            _fileStore.flush();
        }

        /// <summary>
        /// Wait for the store to discard everything that has been stored up to the point in time
        /// when flush is called, then return. Throw an exception if this is not completed in the
        /// number of milliseconds specified by the timeout.
        /// </summary>
        /// <param name="timeout">The number of milliseconds to wait for the flush.</param>
        /// <exception cref="TimedOutException"></exception>
        public void flush(int timeout)
        {
            int start = Environment.TickCount;
            // Flush memory first
            _memoryStore.flush(timeout);
            // In case flush only passed from swapping to file, flush file
            int remaining = timeout - (int)((Environment.TickCount - start)/TimeSpan.TicksPerMillisecond);
            if (remaining > 0) _fileStore.flush(remaining);
            else
                throw new TimedOutException("Timed out waiting to flush publish store.");
        }

        /// <summary>
        /// Set the resize handler for the publish store. An implementation can call the
        /// resize handler in the event that it needs to grow the size of the publish store.
        /// </summary>
        /// <param name="handler">The resize handler to set</param>
        public void setResizeHandler(PublishStoreResizeHandler handler)
        {
            _fileStore.setResizeHandler(handler);
            _memoryStore.setResizeHandler(handler);
        }

        /// <summary>
        /// Close the file access, making this store unusable.
        /// </summary>
        public void close()
        {
            _fileStore.close();
        }

    }
}
