////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using AMPS.Client.Exceptions;
using AMPS.Client.Fields;
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace AMPS.Client
{

    /// <summary>
    /// A message subclass used by JSONProtocolParser to aid in
    /// building Message instances parsed from the {@link JSONProtoco}
    /// (i.e. the "amps" transport protocol).
    /// Generally not used directly by the end user.
    /// </summary>
    public class JSONMessage : Message
    {
        private static String LATIN1 = "ISO-8859-1";
        private static byte[] TG_COMMAND = null;
        private static byte[] TG_COMMANDID = null;
        private static byte[] TG_CORRELATIONID = null;
        private static byte[] TG_CLIENTNAME = null;
        private static byte[] TG_USERID = null;
        private static byte[] TG_TIMESTAMP = null;
        private static byte[] TG_TOPIC = null;
        private static byte[] TG_FILTER = null;
        private static byte[] TG_ORDERBY = null;
        private static byte[] TG_ACKTYPE = null;
        private static byte[] TG_SUBID = null;
        private static byte[] TG_VERSION = null;
        private static byte[] TG_EXPIRATION = null;
        private static byte[] TG_STATUS = null;
        private static byte[] TG_QUERYID = null;
        private static byte[] TG_BATCHSIZE = null;
        private static byte[] TG_TOPN = null;
        private static byte[] TG_SOWKEYS = null;
        private static byte[] TG_SOWKEY = null;
        private static byte[] TG_SEQUENCE = null;
        private static byte[] TG_BOOKMARK = null;
        private static byte[] TG_PASSWORD = null;
        private static byte[] TG_MESSAGETYPE = null;

        private static byte[] TG_OPTIONS = null;

        private static byte[] COMMA = null;
        private static byte[] QUOTE = null;
        private static byte[] QUOTE_COLON = null;
        private static byte[] OPEN_BRACE = null;
        private static byte[] CLOSE_BRACE = null;
        static JSONMessage()
        {
            TG_COMMAND = "c".getBytes(LATIN1);
            TG_COMMANDID = "cid".getBytes(LATIN1);
            TG_CORRELATIONID = "x".getBytes(LATIN1);
            TG_CLIENTNAME = "client_name".getBytes(LATIN1);
            TG_USERID = "user_id".getBytes(LATIN1);
            TG_TIMESTAMP = "ts".getBytes(LATIN1);
            TG_TOPIC = "t".getBytes(LATIN1);
            TG_FILTER = "filter".getBytes(LATIN1);
            TG_ORDERBY = "orderby".getBytes(LATIN1);
            TG_ACKTYPE = "a".getBytes(LATIN1);
            TG_SUBID = "sub_id".getBytes(LATIN1);
            TG_VERSION = "v".getBytes(LATIN1);
            TG_EXPIRATION = "e".getBytes(LATIN1);
            TG_STATUS = "status".getBytes(LATIN1);
            TG_QUERYID = "query_id".getBytes(LATIN1);
            TG_BATCHSIZE = "batch_size".getBytes(LATIN1);
            TG_TOPN = "top_n".getBytes(LATIN1);
            TG_SOWKEYS = "sow_keys".getBytes(LATIN1);
            TG_SOWKEY = "k".getBytes(LATIN1);
            TG_SEQUENCE = "s".getBytes(LATIN1);
            TG_BOOKMARK = "bookmark".getBytes(LATIN1);
            TG_PASSWORD = "pw".getBytes(LATIN1);
            TG_OPTIONS = "opts".getBytes(LATIN1);
            TG_MESSAGETYPE = "mt".getBytes(LATIN1);

            COMMA = ",".getBytes(LATIN1);
            QUOTE = "\"".getBytes(LATIN1);
            QUOTE_COLON = "\":".getBytes(LATIN1);
            OPEN_BRACE = "{".getBytes(LATIN1);
            CLOSE_BRACE = "}".getBytes(LATIN1);

            QUOTED_CLASSES.Add(typeof(StringField));
            QUOTED_CLASSES.Add(typeof(CommandField));
            QUOTED_CLASSES.Add(typeof(AckTypeField));
            QUOTED_CLASSES.Add(typeof(BookmarkField));
            QUOTED_CLASSES.Add(typeof(StatusField));
            QUOTED_CLASSES.Add(typeof(OptionsField));
        }

        /// <summary>
        /// Initializes a new instance of the JSONMessage class.
        /// </summary>
        /// <param name="encoder">The encoding used for the message.</param>
        /// <param name="decoder">The decoding used for the message.</param>
        public JSONMessage(Encoding encoder, Encoding decoder)
            : base(encoder, decoder)
        { ; }

        // Determines if a field needs to be excaped.
        // returns True if the field needs to be excaped, otherwise false.
        private bool needsEscape(byte[] fieldName)
        {
            switch ((byte)fieldName[0])
            {

                case (byte)'c':                 //client, cid, not c
                    return (fieldName.GetLength(0) > 1);
                case (byte)'t':                 //topic, not top_n etc.
                    return (fieldName.GetLength(0) == 1);
                case (byte)'f':                 //filter
                case (byte)'o':                 //opts, orderby
                case (byte)'p':                 //pw
                case (byte)'q':                 //qid
                case (byte)'u':                 //user_id
                    return true;
                case (byte)'s':
                    {
                        switch (fieldName.GetLength(0))
                        {
                            case 6:                //sub_id
                            case 7:                //sub_ids
                                return true;
                            default:
                                return false;
                        }
                    }
                default:
                    return false;
            }
        }

        // Write field using excaping.
        private void escapeField(MemoryStream b, Field field)
        {
            for (int i = 0; i < field.length; ++i)
            {
                if (field.byteAt(i) == '\"' || field.byteAt(i) == '\\')
                {
                    b.put((byte)'\\');
                    b.put(field.byteAt(i));
                }
                else
                {
                    b.put(field.byteAt(i));
                }
            }
        }

        private static HashSet<Type> QUOTED_CLASSES = new HashSet<Type>();

        private int serializeMessageProperty(MemoryStream b, byte[] tag, Field field, int count)
        {
            //
            // ,"field":"value"
            // "field":100
            // etc
            //
            if (field.buffer == null) return 0;
            if (count > 0) b.put(COMMA);
            b.put(QUOTE);
            b.put(tag);
            b.put(QUOTE_COLON);
            bool quoted = QUOTED_CLASSES.Contains(field.GetType());
            if (quoted) b.put(QUOTE);
            if (needsEscape(tag))
                escapeField(b, field);
            else
                b.put(field.buffer, field.position, field.length);
            if (quoted) b.put(QUOTE);
            return 1;
        }

        /// <summary>
        /// Serializes the JSON message into the specified memory stream.
        /// </summary>
        /// <param name="b">The memory stream to write the serialized message to.</param>
        /// <returns>The serialization result.</returns>
        public override SerializationResult serialize(MemoryStream b)
        {
            try
            {
                b.put(OPEN_BRACE);
                int count = 0;
                count += serializeMessageProperty(b, TG_COMMAND, _Command, count);
                count += serializeMessageProperty(b, TG_TOPIC, _Topic, count);
                count += serializeMessageProperty(b, TG_COMMANDID, _CommandId, count);
                count += serializeMessageProperty(b, TG_CORRELATIONID, _CorrelationId, count);
                count += serializeMessageProperty(b, TG_CLIENTNAME, _ClientName, count);
                count += serializeMessageProperty(b, TG_USERID, _UserId, count);
                count += serializeMessageProperty(b, TG_PASSWORD, _Password, count);
                count += serializeMessageProperty(b, TG_TIMESTAMP, _Timestamp, count);
                count += serializeMessageProperty(b, TG_FILTER, _Filter, count);
                count += serializeMessageProperty(b, TG_ORDERBY, _OrderBy, count);
                count += serializeMessageProperty(b, TG_ACKTYPE, _AckType, count);
                count += serializeMessageProperty(b, TG_OPTIONS, _Options, count);
                count += serializeMessageProperty(b, TG_SUBID, _SubId, count);
                count += serializeMessageProperty(b, TG_VERSION, _Version, count);
                count += serializeMessageProperty(b, TG_EXPIRATION, _Expiration, count);
                count += serializeMessageProperty(b, TG_STATUS, _Status, count);
                count += serializeMessageProperty(b, TG_QUERYID, _QueryId, count);
                count += serializeMessageProperty(b, TG_BATCHSIZE, _BatchSize, count);
                count += serializeMessageProperty(b, TG_BOOKMARK, _Bookmark, count);
                count += serializeMessageProperty(b, TG_SEQUENCE, _Sequence, count);
                count += serializeMessageProperty(b, TG_TOPN, _TopN, count);
                count += serializeMessageProperty(b, TG_SOWKEYS, _SowKeys, count);
                count += serializeMessageProperty(b, TG_SOWKEY, _SowKey, count);
                count += serializeMessageProperty(b, TG_MESSAGETYPE, _MessageType, count);
                b.put(CLOSE_BRACE);
                if (_Data.buffer != null)
                {
                    b.put(_Data.buffer, _Data.position, _Data.length);
                }
            }
            catch (BufferOverflowException)
            {
                return SerializationResult.BufferTooSmall;
            }
            return SerializationResult.OK;
        }

        /// <summary>
        /// Makes and returns a deep copy of this JSONMessage.
        /// </summary>
        /// <returns>A copy of the JSON message.</returns>
        public override Message copy()
        {
            JSONMessage m = new JSONMessage(this.encoder, this.decoder);
            _copyTo(m);
            return m;
        }
    }
}
