////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using AMPS.Client.Exceptions;
using System;
using System.Diagnostics;

namespace AMPS.Client
{
    /// <summary>
    /// Implements the ProtocolParser interface. Parses a byte stream using
    /// JSON Protocol (AMPS Protocol).
    /// </summary>
    public class JSONProtocolParser : ProtocolParser
    {
        static byte OPEN_BRACE = Convert.ToByte('{');
        static byte CLOSE_BRACE = Convert.ToByte('}');
        static byte QUOTE = Convert.ToByte('"');
        static byte BACKSLASH = Convert.ToByte('\\');
        static byte COLON = Convert.ToByte(':');

        private JSONMessage message;
        private SimpleReadBuffer buffer = null;
        private int remainingBytes = 0;

        private enum StreamState { start, in_sow, end }

        private StreamState state;

        /// <summary>
        /// Initializes a new instance of the JSONProtocolParser class with
        /// the specified JSONProtocol.
        /// </summary>
        /// <param name="messageType">The JSONProtocol instance.</param>
        public JSONProtocolParser(JSONProtocol messageType)
        {
            this.message = (JSONMessage)messageType.allocateMessage();
        }

        /// <summary>
        /// Processes the byte stream using the JSONProtocolParser.
        /// </summary>
        /// <param name="buffer">The buffer containing the byte stream.</param>
        /// <param name="remainingBytes">The number of remaining bytes in the buffer.</param>
        /// <param name="listener">The message handler listener.</param>
        public void process(SimpleReadBuffer buffer,
                            int remainingBytes,
                            MessageHandler listener)
        {
            this.buffer = buffer;
            this.remainingBytes = remainingBytes;
            this.state = StreamState.start;

            message.reset();
            message.setBuffer(buffer.Buffer);
            unsafe
            {
                fixed (byte* b = buffer.Buffer)
                {
                    while (read(message))
                    {
                        listener.invoke(message);
                    }
                }
            }
        }

        // Enumerates the header fields for the JSONProtocolParser.
        enum HeaderField
        {
            AckTyp,
            BkMrk,
            BtchSz,
            ClntName,
            Cmd,
            CmdId,
            Expn,
            Fltr,
            GrpSqNum,
            Hrtbt,
            Matches,
            MsgLn,
            Options,
            Password,
            QId,
            Reason,
            RecordsDeleted,
            RecordsReturned,
            RecordsUpdated,
            RecordsInserted,
            Seq,
            SubIds,
            SowKey,
            SowKeys,
            Status,
            SubId,
            TxmTm,
            Tpc,
            TopicMatches,
            TopN,
            UsrId,
            Version,
            CrlId,
            LeasePeriod,
            UNKNOWN
        }

        // starts with buffer position on open quote of field name
        // advances through field name, through close quote and colon
        // leaves buffer position on index after colon
        unsafe HeaderField extractHeaderField()
        {
            assertByte(QUOTE);
            // "":x
            if (remainingBytes < 4)
            {
                throw new StreamException("stream corruption: premature end of header");
            }

            getByte(); // eat open quote
            int name_start = buffer.Position;
            scan_or_throw();
            assertByte(QUOTE);
            int name_length = buffer.Position - name_start;

            if (remainingBytes < 2)
            {
                throw new StreamException("stream corruption: premature end of header");
            }
            getByte(); // eat close quote
            assertByte(COLON);
            getByte();

            byte[] pTag_ = buffer.Buffer;

            switch (name_length)
            {
                case 1:
                    switch (pTag_[name_start + 0])
                    {
                        case (byte)'c':
                            return HeaderField.Cmd;
                        case (byte)'t':
                            return HeaderField.Tpc;
                        case (byte)'s':
                            return HeaderField.Seq;
                        case (byte)'a':
                            return HeaderField.AckTyp;
                        case (byte)'e':
                            return HeaderField.Expn;
                        case (byte)'o':
                            return HeaderField.Options;
                        case (byte)'l':
                            return HeaderField.MsgLn;
                        case (byte)'k':
                            return HeaderField.SowKey;
                        case (byte)'v':
                            return HeaderField.Version;
                        case (byte)'x':
                            return HeaderField.CrlId;

                    }
                    break;
                case 2:
                    switch (pTag_[name_start + 1])
                    {
                        case (byte)'w':
                            return HeaderField.Password;   // pw (same as auth_key)
                        case (byte)'s':
                            return (pTag_[name_start + 0] == 'b') ? HeaderField.BtchSz : HeaderField.TxmTm; // bs (same as batch_size) or ts
                        case (byte)'m':
                            return HeaderField.BkMrk;   // bw (same as bookmark)
                        case (byte)'p':
                            return HeaderField.LeasePeriod;
                    }
                    break;
                case 3:
                    switch (pTag_[name_start + 1])
                    {
                        case (byte)'m':
                            return HeaderField.Cmd;    // cmd
                        case (byte)'i':
                            return HeaderField.CmdId;  // cid
                        case (byte)'c':
                            return HeaderField.AckTyp; // ack
                        case (byte)'e':
                            return HeaderField.Seq;    // seq
                    }
                    break;
                case 4:
                    switch (pTag_[name_start + 0])
                    {
                        case (byte)'o':
                            return HeaderField.Options;  // opts
                        case (byte)'s':
                            return HeaderField.SubIds;   // sids - same as sub_ids
                        case (byte)'g':
                            return HeaderField.GrpSqNum; // gseq - same as group_seq_num
                    }
                    break;
                case 5:
                    switch (pTag_[name_start + 3])
                    {
                        case (byte)'i':
                            return HeaderField.Tpc;  // topic
                        case (byte)'_':
                            return HeaderField.TopN; // top_n
                    }
                    break;
                case 6:
                    switch (pTag_[name_start + 1])
                    {
                        case (byte)'m':
                            return HeaderField.CmdId;   // cmd_id
                        case (byte)'i':
                            return HeaderField.Fltr;    // filter
                        case (byte)'t':
                            return HeaderField.Status;  // status
                        case (byte)'e':
                            return HeaderField.Reason;  // reason
                        case (byte)'u':
                            return HeaderField.SubId;   // sub_id
                    }
                    break;
                case 7:
                    switch (pTag_[name_start + 0])
                    {
                        case (byte)'m':
                            return HeaderField.Matches;  // matches
                        case (byte)'u':
                            return HeaderField.UsrId;    // user_id
                        case (byte)'s':
                            return HeaderField.SubIds;   // sub_ids
                        case (byte)'v':
                            return HeaderField.Version;  // version
                    }
                    break;
                case 8:
                    switch (pTag_[name_start + 6])
                    {
                        case (byte)'p':
                            return HeaderField.AckTyp;    // ack_type
                        case (byte)'e':
                            return HeaderField.Password;  // auth_key
                        case (byte)'r':
                            return HeaderField.BkMrk;     // bookmark
                        case (byte)'i':
                            return HeaderField.QId;       // query_id
                        case (byte)'y':
                            return HeaderField.SowKeys;   // sow_keys
                    }
                    break;
                case 9:
                    switch (pTag_[name_start + 0])
                    {
                        case (byte)'h':
                            return HeaderField.Hrtbt; // heartbeat
                    }
                    break;
                case 10:
                    switch (pTag_[name_start + 0])
                    {
                        case (byte)'e':
                            return HeaderField.Expn;    // expiration
                        case (byte)'b':
                            return HeaderField.BtchSz;  // batch_size
                    }
                    break;
                case 11:
                    switch (pTag_[name_start + 0])
                    {
                        case (byte)'c':
                            return HeaderField.ClntName; // client_name
                    }
                    break;
                case 13:
                    switch (pTag_[name_start + 0])
                    {
                        case (byte)'t':
                            return HeaderField.TopicMatches;  // topic_matches
                        case (byte)'g':
                            return HeaderField.GrpSqNum;      // group_seq_num
                    }
                    break;
                case 15:
                    switch (pTag_[name_start + 8])
                    {
                        case (byte)'d':
                            return HeaderField.RecordsDeleted; // records_deleted
                        case (byte)'u':
                            return HeaderField.RecordsUpdated; // records_updated
                    }
                    break;
                case 16:
                    switch (pTag_[name_start + 8])
                    {
                        case (byte)'i':
                            return HeaderField.RecordsInserted; // records_inserted
                        case (byte)'r':
                            return HeaderField.RecordsReturned; // records_returned
                    }
                    break;
            }
            return HeaderField.UNKNOWN;
        }

        private byte getUnescapedChar()
        {
            switch (peekByte())
            {
                case (byte)'\"':
                    return (byte)'\"';
                case (byte)'\\':
                    return (byte)'\\';
                case (byte)'/':
                    return (byte)'/';
                case (byte)'b':
                    return (byte)'\b';
                case (byte)'f':
                    return (byte)'\f';
                case (byte)'n':
                    return (byte)'\n';
                case (byte)'r':
                    return (byte)'\r';
                case (byte)'t':
                    return (byte)'\t';
                default:
                    return 0x00;
            }
        }

        int unescapeField()
        {
            int b = buffer.Position;

            while (remainingBytes > 0 && peekByte() != QUOTE)
            {
                if (peekByte() == BACKSLASH)
                {
                    getByte();
                    byte unescapedChar = getUnescapedChar();
                    //consume unescaped char without toggling in_quote, results in buffer corruption
                    --remainingBytes;
                    buffer.ReadByte();
                    this.buffer.put(b++, unescapedChar);
                }
                else
                {
                    this.buffer.put(b++, getByte());
                }
            }
            return b;
        }

        // starts with buffer position at value
        // advances through buffer to the position following the value
        // returns the index where the value ended
        unsafe int findValueEnd(bool needsUnescape)
        {
            int end_index;
            switch (peekByte())
            {
                case (byte)'"':
                    getByte();
                    if (needsUnescape)
                    {
                        end_index = unescapeField();
                    }
                    else
                    {
                        scan_or_throw();
                        end_index = buffer.Position;
                    }
                    assertByte(QUOTE);
                    if (remainingBytes <= 0)
                    {
                        throw new StreamException("stream corruption: premature end of header");
                    }
                    getByte(); // eat quote
                    break;
                case (byte)'0':
                case (byte)'1':
                case (byte)'2':
                case (byte)'3':
                case (byte)'4':
                case (byte)'5':
                case (byte)'6':
                case (byte)'7':
                case (byte)'8':
                case (byte)'9':
                case (byte)'-':
                    while (remainingBytes > 0 && peekByte() >= '0' && peekByte() <= '9')
                    {
                        getByte();
                    }
                    end_index = buffer.Position;
                    break;
                case (byte)'t':
                    // true
                    if (remainingBytes < 4)
                    {
                        throw new StreamException("stream corruption: premature end of header");
                    }
                    getByte();
                    assertByte((byte)'r');
                    getByte();
                    assertByte((byte)'u');
                    getByte();
                    assertByte((byte)'e');
                    getByte();
                    end_index = buffer.Position;
                    break;
                case (byte)'f':
                    // false
                    if (remainingBytes < 5)
                    {
                        throw new StreamException("stream corruption: premature end of header");
                    }
                    getByte();
                    assertByte((byte)'a');
                    getByte();
                    assertByte((byte)'l');
                    getByte();
                    assertByte((byte)'s');
                    getByte();
                    assertByte((byte)'e');
                    getByte();
                    end_index = buffer.Position;
                    break;
                case (byte)'n':
                    // null
                    if (remainingBytes < 4)
                    {
                        throw new StreamException("stream corruption: premature end of header");
                    }
                    getByte();
                    assertByte((byte)'u');
                    getByte();
                    assertByte((byte)'l');
                    getByte();
                    assertByte((byte)'l');
                    getByte();
                    end_index = buffer.Position;
                    break;
                default:
                    throw new StreamException("stream corruption: premature end of header");
            }
            return end_index;
        }

        unsafe void extractFieldValue(JSONMessage message)
        {
            HeaderField field = extractHeaderField();
            bool needsUnEscape = (field == HeaderField.Tpc || field == HeaderField.UsrId
                                  || field == HeaderField.Password || field == HeaderField.ClntName
                                  || field == HeaderField.CmdId || field == HeaderField.QId
                                  || field == HeaderField.SubId || field == HeaderField.SubIds);
            int valueStart = peekByte() == QUOTE ? buffer.Position + 1 : buffer.Position;
            int valueLength = findValueEnd(needsUnEscape) - valueStart;

            switch (field)
            {
                case HeaderField.AckTyp:
                    message._AckType.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.BtchSz:
                    message._BatchSize.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.BkMrk:
                    message._Bookmark.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.ClntName:
                    message._ClientName.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.Cmd:
                    message._Command.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.CmdId:
                    message._CommandId.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.CrlId:
                    message._CorrelationId.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.Expn:
                    message._Expiration.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.Fltr:
                    message._Filter.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.GrpSqNum:
                    message._GroupSeqNo.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                //          case Hrtbt:           message._Heartbeat.set(this.buffer.Buffer,valueStart,valueLength); break;
                case HeaderField.Matches:
                    message._Matches.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.MsgLn:
                    message._Length.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.Options:
                    message._Options.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.LeasePeriod:
                    message._LeasePeriod.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.Password:
                    message._Password.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.QId:
                    message._QueryId.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.Reason:
                    message._Reason.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.RecordsInserted:
                    message._RecordsInserted.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.RecordsUpdated:
                    message._RecordsUpdated.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.RecordsReturned:
                    message._RecordsReturned.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.RecordsDeleted:
                    message._RecordsDeleted.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.Seq:
                    message._Sequence.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.SubIds:
                    message._SubIds.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.SowKey:
                    message._SowKey.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.SowKeys:
                    message._SowKeys.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.Status:
                    message._Status.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.SubId:
                    message._SubId.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.TxmTm:
                    message._Timestamp.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.Tpc:
                    message._Topic.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.TopicMatches:
                    message._TopicMatches.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.TopN:
                    message._TopN.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.UsrId:
                    message._UserId.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                case HeaderField.Version:
                    message._Version.set(this.buffer.Buffer, valueStart, valueLength);
                    break;
                default:
                    break;
            }
            while (remainingBytes > 0 && peekByte() != QUOTE && peekByte() != CLOSE_BRACE)
            {
                getByte();
            }
        }

        unsafe private bool read(JSONMessage m)
        {
            if (remainingBytes <= 0)
            {
                // No more messages in this stream
                return false;
            }

            if (state != StreamState.in_sow)
            {
                m.setRawBufferOffset((int)buffer.Position);
            }

            scan();
            if (remainingBytes <= 0)
            {
                return false;
            }
            assertByte(OPEN_BRACE);
            getByte();
            while (remainingBytes > 0 && peekByte() != CLOSE_BRACE)
            {
                extractFieldValue(m);
            }
            assertByte(CLOSE_BRACE);
            getByte();

            int messageLength;
            Message.Commands command = m.getCommand();
            if (state == StreamState.start && command == Message.Commands.SOW)
            {

                scan();
                if (remainingBytes <= 0)
                {
                    return false;
                }

                assertByte(OPEN_BRACE);
                getByte();

                while (remainingBytes > 0 && peekByte() != CLOSE_BRACE)
                {
                    extractFieldValue(m);
                }

                assertByte(CLOSE_BRACE);
                getByte();

                if (command == Message.Commands.Publish || command == Message.Commands.OOF)
                {
                    messageLength = remainingBytes;
                }
                else
                {
                    state = StreamState.in_sow;
                    messageLength = m._Length.getValue();
                }

            }
            else if (state == StreamState.in_sow)
            {
                messageLength = m._Length.getValue();
            }
            else
            {
                messageLength = remainingBytes;
            }

            // Now, we have to deal with the data
            m._Data.set(this.buffer.Buffer, buffer.Position, messageLength);
            remainingBytes -= messageLength;
            buffer.Position += messageLength;

            m.setRawBufferLength(buffer.Position - m.getRawBufferOffset());

            return true;
        }

        private unsafe void assertByte(byte b)
        {
            if (peekByte() != b)
            {
                throw new StreamException("stream corruption: expected '" + (char)b + "'");
            }
        }

        private bool in_quote = false;

        byte getByte()
        {
            --remainingBytes;
            byte b = buffer.ReadByte();
            if (b == QUOTE)
            {
                in_quote = !in_quote;
            }
            return b;
        }

        byte peekByte()
        {
            return buffer.PeekByte(buffer.Position);
        }

        unsafe bool scan()
        {
            // if we're in a quoted-string ignore braces and colons
            while (remainingBytes > 0 &&
                    (
                       (in_quote
                         && peekByte() != QUOTE) ||
                        (!in_quote
                            && peekByte() != QUOTE
                            && peekByte() != OPEN_BRACE
                            && peekByte() != COLON
                            && peekByte() != CLOSE_BRACE)
                    ))
            {
                getByte();
            }
            return remainingBytes > 0;
        }

        unsafe void scan_or_throw()
        {
            if (!scan())
            {
                throw new StreamException("stream corruption: premature end of header");
            }
        }
    }
}
