///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Collections.Generic;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading;
using AMPS.Client.Fields;
using AMPS.Client.Exceptions;

namespace AMPS.Client
{

    /// <summary>
    /// LoggedBookmarkStore implements a sequentially written log of incoming and
    /// discarded messages. The store tracks every bookmark processed in a file.
    /// An application should periodically call  <see cref="AMPS.Client.LoggedBookmarkStore.prune"/>
    /// to manage the size of the file by removing outdated entries. When using LoggedBookmarkStore,
    /// it is important to note that this store is intended to provide the ability for
    /// the application to fall and resume. If your application does not require this
    /// fall-and-resume functionality, it is recommended to use a MemoryBookmarkStore.
    /// </summary>
    ///
    /// <remarks>
    /// Usage:
    /// The LoggedBookmarkStore is automatically set on the returned client when
    /// `createFileBacked` is called. This store is responsible for tracking and
    /// managing bookmarks associated with the client's state.
    /// </remarks>
    public class LoggedBookmarkStore : BookmarkStore, IDisposable
    {
        // Each entry begins with a single byte indicating the type of entry:
        // a new bookmark, or a discard of a previous one.
        const byte ENTRY_BOOKMARK = (byte)'b';
        const byte ENTRY_DISCARD =  (byte)'d';
        const byte ENTRY_PERSISTED = (byte)'p';

        // We use an in-memory array to remember the order in which bookmarks
        // arrive on a given subscription.  When the user has discard()ed all of the
        // bookmarks up to a certain bookmark B, we log B, and return it the next time
        // someone calls getMostRecent().
        class Subscription : AMPS.Client.Subscription
        {
            // This subscription's ID.
            Field _sub;

            // The last persisted bookmark
            internal BookmarkField _lastPersisted = null;

            // If this subscription uses a range, save it here
            private BookmarkRangeField _range = new BookmarkRangeField();

            // <summary>
            // The last modification time of the underlying file at the time when
            // it is going to be recovered. If this is used in the most recent
            // bookmark, it allows recovery from a very old file where the
            // bookmarks are no longer in the server journal files to recover
            // from essentially EPOCH rather than NOW, which is the default
            // for not found bookmarks.
            // It must be stored separately here because setting it as a sub's
            // last-persisted could cause us to lose a persisted ack that is earlier
            // than the backing file's last-modified time.
            // </summary>
            protected volatile string _recoveryTimestamp = null;

            // EPOCH field for ease-of-use
            public static BookmarkField EPOCH_FIELD = new BookmarkField(Client.Bookmarks.EPOCH);

            // A set of all of the undiscarded entries recovered that
            // had been received after the one returned by getMostRecent().
            // When these come in again, we won't log them a 2nd time.
            Dictionary<Field, long> _recovered = new Dictionary<Field, long>();

            // A bookmark ring buffer.
            BookmarkRingBuffer _ring = new BookmarkRingBuffer();

            // The per-subscription memory of what we've seen from publishers
            internal Dictionary<ulong, ulong> _publishers = new Dictionary<ulong, ulong>();

            // The store that we log to when the time comes.
            LoggedBookmarkStore _parent;

            private Object internalLock = new Object();

            public Subscription()
            {
                _lastPersisted = new BookmarkField();
                _lastPersisted.copyFrom(EPOCH_FIELD);
            }

            // <summary>
            // Reset the state of this subscription object such that it can be returned
            // to the pool for reuse. This method clears all internal data structures and
            // sets necessary attributes to their initial values. This helps manage file
            // size and resource usage efficiently.
            // </summary>
            public void reset()
            {
                _sub.reset();
                _lastPersisted.copyFrom(EPOCH_FIELD);
                _recoveryTimestamp = null;
                _recovered.Clear();
                _ring.reset();
                _publishers.Clear();
                _range = new BookmarkRangeField();
            }

            // <summary>
            // Initializes the subscription with the given subscription ID and the parent
            // LoggedBookmarkStore. This method sets up the initial state of the subscription,
            // including associating it with a specific store.
            // </summary>
            // <param name="subscriptionId">The ID of the subscription to initialize.</param>
            // <param name="parent">The parent LoggedBookmarkStore to associate with this subscription.</param>
            public void init(Field subscriptionId, LoggedBookmarkStore parent)
            {
                _sub = subscriptionId.copy();
                _ring.setSubId(_sub);
                _parent = parent;
            }

            // <summary>
            // Get the recovery timestamp associated with this message.
            // The recovery timestamp represents the time at which the message
            // was recovered by the client during a resubscription after
            // disconnection. This timestamp provides information about when
            // the message was originally published or received by the server.
            // </summary>
            // <returns>The recovery timestamp as a string in a format that represents
            // the time and date when the message was recovered by the client.</returns>
            public string getRecoveryTimestamp()
            {
                return _recoveryTimestamp;
            }

            // <summary>
            // Sets the recovery timestamp for the Advanced Message Processing System (AMP).
            // </summary>
            // <param name="rts">The recovery timestamp to be set.</param>
            public void setRecoveryTimestamp(string rts)
            {
                _recoveryTimestamp = rts;
            }

            // <summary>
            // This is a method the Client calls and is not for customer user. Logs a bookmark
            // into the subscription. If the bookmark is a range, it processes the range accrdingly.
            // </summary>
            // <param name="bookmark">The bookmark field to be logged.</param>
            // <returns>The sequence number associated with the logged bookmark. 0 if no bookmark
            // was logged.</returns>
            // <exception cref="CommandException"></exception>
            public override long log(BookmarkField bookmark)
            {
                lock (internalLock) {
                    if (bookmark.Equals(Client.Bookmarks.NOW)) return 0;
                    if (!bookmark.isRange())
                    {
                        // Check to see if this is a recovered bookmark.
                        // If so, re-log in the new order.
                        long seq = 0;
                        if (_recovered.TryGetValue(bookmark, out seq))
                        {
                            _recovered.Remove(bookmark);
                            return seq;
                        }

                        // Add this entry onto our list to remember the order it was seen.
                        if (!bookmark.isBookmarkList())
                        {
                            return _log(bookmark);
                        }
                        else
                        {
                            foreach (BookmarkField bm in bookmark.parseBookmarkList())
                            {
                                isDiscarded(bm);
                                seq = _log(bm);
                                if (seq != 0)
                                {
                                    _ring.discard(seq);
                                    _parent._recentChanged = true;
                                }
                            }
                            adapterUpdate((BookmarkField)_range);
                            return 0;
                        }
                    }
                    else
                    {
                        _range.copyFrom(bookmark);
                        if (!_range.isValid())
                        {
                            throw new CommandException("Invalid bookmark range specified");
                        }
                        long seq = 0;
                        if (_range.isStartExclusive())
                        {
                            BookmarkField start = _range.getStart();
                            if (!start.isBookmarkList())
                            {
                                isDiscarded(start);
                                seq = _log(start);
                                if (seq != 0)
                                {
                                    _ring.discard(seq);
                                    _parent._recentChanged = true;
                                }
                            }
                            else
                            {
                                foreach (BookmarkField bm in start.parseBookmarkList())
                                {
                                    isDiscarded(bm);
                                    seq = _log(bm);
                                    if (seq != 0)
                                    {
                                        _ring.discard(seq);
                                        _parent._recentChanged = true;
                                    }
                                }
                                seq = 0;
                            }
                        }
                        adapterUpdate((BookmarkField)_range);
                        // The only time a range is logged is when subscribing.
                        // Don't put it in the ring.
                        return seq;
                    }
                }
            }

            public long _log(BookmarkField bookmark)
            {
                long seq = 0;
                if (!bookmark.isTimestamp())
                {
                    seq = _ring.log(bookmark);
                    while (seq == 0)
                    {
                        Monitor.Exit(internalLock);
                        try { _ring.checkResize(); }
                        finally { Monitor.Enter(internalLock); }
                        seq = _ring.log(bookmark);
                    }
                }
                else
                {
                    setLastPersisted(bookmark);
                }
                return seq;
            }

            // <summary>
            // Overrides the discard method.
            // </summary>
            // <param name="index">The index of the entry to be discarded.</param>
            public override void discard(long index)
            {
                // Get the entry from the ring buffer using the specified index.
                lock (internalLock) {
                    BookmarkRingBuffer.Entry entry = _ring.getByIndex(index);

                    // Check if the entry is null or not active, and return if so.
                    if (entry == null || !entry.isActive())
                    {
                        return;
                    }
                    // If not in recovery mode, write the discard entry to the parent.
                    if (!_parent._recovering)
                        _parent.write(_sub, LoggedBookmarkStore.ENTRY_DISCARD,
                                      entry.getBookmark());

                    // Discard the entry from the ring buffer and perform additional actions
                    // if successful.
                    if (_ring.discard(index))
                    {
                        _parent._recentChanged = true;
                        _recoveryTimestamp = null;
                        adapterUpdate((BookmarkField)_getMostRecentList(false));
                    }
                }
            }

            // <summary>
            // Check to see if this message is older than the most recent one seen,
            // and if it is, check if it discarded.
            // </summary>
            // <param name="bookmark">The bookmark field to check for discard status.</param>
            // <returns><c>true</c> if the bookmark has been discarded; otherwise, <c>false</c>.</returns>
            public override bool isDiscarded(BookmarkField bookmark)
            {
                lock (internalLock) {
                    // Check if the bookmark is a predefined constant, indicating immediate discard.
                    if (bookmark.Equals(Client.Bookmarks.NOW)) return true;

                    // Attempt to retrieve the recovered index for the bookmark.
                    long recoveredIndex;
                    bool inFlight = _recovered.TryGetValue(bookmark, out recoveredIndex);

                    // Extract publisher and sequence information from the bookmark.
                    ulong publisher = bookmark.getPublisherId();
                    ulong sequence = bookmark.getSequenceNumber();

                    // Check if the publisher exists or has a sequence number greater than the current one.
                    if (!_publishers.ContainsKey(publisher) ||
                        _publishers[publisher] < sequence)
                    {
                        _publishers[publisher] = sequence;
                        // If not in flight, return false indicating not discarded.
                        if (!inFlight)
                        {
                            return false;
                        }
                    }
                    // If in flight, retrieve the entry from the ring buffer.
                    if (inFlight)
                    {
                        BookmarkRingBuffer.Entry entry = _ring.getByIndex(recoveredIndex);

                        // Check if the entry is missing and the index is below the ring's start index.
                        if (entry == null &&
                            (recoveredIndex < _ring.getStartIndex() || _ring.isEmpty()))
                        {
                            // Remove the bookmark from the recovered set and return true (discarded).
                            _recovered.Remove(bookmark);
                            return true;
                        }
                        // Return the current active state
                        return (entry == null) ? false : !entry.isActive();
                    }

                    // Return the recovery status if not in flight.
                    return !_parent._recovering;
                }
            }

            // <summary>
            // Gets the last persisted bookmark.
            // </summary>
            // <returns>The last persisted bookmark.</returns>
            public Field getLastPersisted()
            {
                lock (internalLock) {
                    return _lastPersisted;
                }
            }

            // <summary>
            // Retrieves the current bookmark range.
            // </summary>
            // <returns>The current bookmark range.</returns>
            public BookmarkRangeField getRange()
            {
                lock (internalLock) {
                    return _range;
                }
            }

            // <summary>
            // Overrides the getMostRecent method.
            // </summary>
            // <returns>The most recent bookmark.</returns>
            public override Field getMostRecent()
            {
                lock (internalLock) {
                    return getMostRecent(false);
                }
            }

            internal Field getMostRecent(bool updateRecovery_)
            {
                lock (internalLock) {
                    // when this is called, we'll take a moment to update the
                    // list of things recovered,
                    // so we don't accidentally log anything we ought not to.
                    if (updateRecovery_ && _parent._recentChanged)
                    {
                        updateRecovery();
                    }
                    return _ring.getLastDiscarded();
                }
            }

            // <summary>
            // Retrieves the most recent bookmark as a list and constructs a list of bookmarks
            // based on specified conditions, including the last discarded and last persisted
            // bookmarks, the recovery timestamp, and whether a range is valid.
            // </summary>
            // <param name="useList">A flag indicating whether to return a list of recent bookmarks.</param>
            // <returns> The most recent field or a list of recent bookmarks, based on the specified conditions.
            // </returns>
            public override Field getMostRecentList(bool useList)
            {
                lock (internalLock) {
                    return _getMostRecentList(useList);
                }
            }

            public Field _getMostRecentList(bool useList)
            {
                bool rangeIsValid = _range.isValid();
                BookmarkField lastDiscarded = (BookmarkField)_ring.getLastDiscarded();
                bool useLastDiscarded = (lastDiscarded != null &&
                                         !lastDiscarded.isNull());
                // Handle last discard as EPOCH
                if (useLastDiscarded && lastDiscarded.Equals(EPOCH_FIELD))
                {
                    if (rangeIsValid)
                    {
                        return _range.copy();
                    }
                    else if (_recoveryTimestamp != null)
                    {
                        BookmarkField ret = new BookmarkField();
                        ret.setValue(_recoveryTimestamp, Encoding.ASCII);
                        return ret;
                    }
                    // Return just EPOCH
                    return EPOCH_FIELD.copy();
                }
                ulong lastDiscardedPub = 0;
                ulong lastDiscardedSeq = 0;
                bool useLastPersisted = (_lastPersisted != null &&
                                         _lastPersisted.length > 1);
                ulong lastPersistedPub = 0;
                ulong lastPersistedSeq = 0;
                if (useLastPersisted)
                {
                    lastPersistedPub = _lastPersisted.getPublisherId();
                    lastPersistedSeq = _lastPersisted.getSequenceNumber();
                }
                // Append last discarded information to the string builder if conditions are met.
                if (useLastDiscarded)
                {
                    if (useLastPersisted && _ring.isEmpty()
                        && (!rangeIsValid || _range.getEnd().Equals(_lastPersisted)))
                    {
                        useLastDiscarded = false;
                    }
                    else
                    {
                        lastDiscardedPub = lastDiscarded.getPublisherId();
                        lastDiscardedSeq = lastDiscarded.getSequenceNumber();
                        if (useLastPersisted &&
                            (lastDiscardedPub == lastPersistedPub))
                        {
                            useLastDiscarded = (lastDiscardedSeq
                                                < lastPersistedSeq);
                            useLastPersisted = !useLastDiscarded;
                        }
                    }
                }
                StringBuilder recentStr = new StringBuilder();
                BookmarkField recentList = new BookmarkField();
                if (_recoveryTimestamp != null)
                {
                    recentStr.Append(_recoveryTimestamp);
                }
                if (useLastDiscarded)
                {
                    if (recentStr.Length > 0) recentStr.Append(',');
                    recentStr.Append(lastDiscarded.getValue(Encoding.ASCII));
                }
                // If we don't have a last persisted or a last discarded OR we are
                // expecting persisted acks but haven't received one yet, then we
                // should try to build a list of bookmarks based on publishers we
                // have seen so far, if any, or return EPOCH.
                if (useList &&
                    ((!useLastDiscarded && !useLastPersisted)
                     || (_lastPersisted != null
                         && _lastPersisted.Equals(EPOCH_FIELD))
                     ))
                {
                    if (_publishers.Count == 0 && !_range.isValid())
                    {
                        if (_lastPersisted == null)
                        {
                            _lastPersisted = new BookmarkField();
                            _lastPersisted.copyFrom(EPOCH_FIELD);
                        }
                        return _lastPersisted.copy();
                    }
                    // If an EPOCH lastDiscarded was added, remove it
                    if (useLastDiscarded && lastDiscarded.Equals(EPOCH_FIELD))
                    {
                        if (recentStr.Length <= 1) // EPOCH only
                            recentStr.Clear();
                        else // remove ,0 for EPOCH
                            recentStr.Remove(recentStr.Length-2, 2);
                    }
                    foreach (var item in _publishers)
                    {
                        if (item.Key == 0 ||
                            (useLastDiscarded && item.Key == lastDiscardedPub))
                        {
                            continue;
                        }
                        if (recentStr.Length > 0)
                            recentStr.Append(',');
                        recentStr.Append(item.Key)
                                 .Append('|')
                                 .Append(item.Value)
                                 .Append('|');
                    }
                    recentList.setValue(recentStr.ToString(), Encoding.ASCII);
                    if (rangeIsValid)
                    {
                        if (recentList.length > 1
                            && (_range.isStartInclusive()
                               || !recentList.Equals(_range.getStart())))
                        {
                            _range.replaceStart(recentList, true);
                        }
                        return _range.copy();
                    }
                    return recentList;
                }
                if (useLastPersisted)
                {
                    if (recentStr.Length > 0) recentStr.Append(',');
                    recentStr.Append(_lastPersisted.getValue(Encoding.ASCII));
                }
                recentList.setValue(recentStr.toString(), Encoding.ASCII);
                if (rangeIsValid)
                {
                    if (recentList.length > 1
                        && (_range.isStartInclusive()
                           || !recentList.Equals(_range.getStart())))
                    {
                        _range.replaceStart(recentList, true);
                    }
                    return _range.copy();
                }
                if (recentList.length == 0) return EPOCH_FIELD;
                return recentList;
            }

            private void updateRecovery()
            {
                _recovered.Clear();
                long end = _ring.getEndIndex();
                for (long index = _ring.getStartIndex(); index < end; ++index)
                {
                    if (index == BookmarkRingBuffer.UNSET_INDEX) return;
                    BookmarkRingBuffer.Entry entry = _ring.getByIndex(index);
                    if (entry != null && entry.getBookmark() != null &&
                        !entry.getBookmark().isNull())
                    {
                        _recovered[entry.getBookmark().copy()] = index;
                    }
                }
            }

            // Gets the active entries from the ring buffer.
            internal void getActiveEntries(IList<BookmarkRingBuffer.Entry> entryList_)
            {
                _ring.getActiveEntries(entryList_);
            }

            // <summary>
            // Sets the last persisted bookmark.
            // based on the provided index. Performs various checks and updates system state accordingly.
            // </summary>
            // <param name="bookmark">The index of the entry to be set as the last persisted bookmark.</param>
            public override void setLastPersisted(long bookmark)
            {
                lock (internalLock) {
                    BookmarkRingBuffer.Entry entry = _ring.getByIndex(bookmark);
                    if (entry == null)
                        return;
                    BookmarkField bookmarkField = entry.getBookmark();
                    if (bookmarkField == null || bookmarkField.isNull()) return;
                    // If not in recovery mode, write the persisted entry to the parent.
                    if (!_parent._recovering)
                        _parent.write(_sub, ENTRY_PERSISTED, bookmarkField);
                    // Update system state if necessary and set the last persisted field.
                    if (_lastPersisted == null || _ring.isEmpty())
                    {
                        _parent._recentChanged = true;
                        _recoveryTimestamp = null;
                        adapterUpdate((BookmarkField)_getMostRecentList(false));
                    }
                    if (_lastPersisted != null) _lastPersisted.reset();
                    _lastPersisted = bookmarkField.copy();
                }
            }

            // <summary>
            // Sets the last persisted bookmark using a provided BookmarkField object. Performs
            // various checks and updates system state accordingly.
            // </summary>
            // <param name="bookmark">The bookmark field to be set as the last persisted bookmark.</param>
            public override void setLastPersisted(BookmarkField bookmark)
            {
                lock (internalLock) {
                    if (bookmark == null || bookmark.isNull()
                        || bookmark.Equals(_lastPersisted)
                        || bookmark.isRange())
                        return;
                    // If the bookmark is a timestamp, update system state and set the last persisted field.
                    if (bookmark.isTimestamp())
                    {
                        _recoveryTimestamp = bookmark.ToString();
                        // If not in recovery mode, write the persisted entry to the parent.
                        if (!_parent._recovering)
                            _parent.write(_sub, ENTRY_PERSISTED, bookmark);
                        _parent._recentChanged = true;
                        adapterUpdate((BookmarkField)_getMostRecentList(false));
                        return;
                    }
                    ulong publisherId = bookmark.getPublisherId();
                    bool lastPersistedUnset = (_lastPersisted == null
                                               || _lastPersisted.Equals(EPOCH_FIELD));
                    // Check conditions for updating system state and setting the last persisted field.
                    if (!lastPersistedUnset
                        && publisherId == _lastPersisted.getPublisherId()
                        && bookmark.getSequenceNumber() <= _lastPersisted.getSequenceNumber())
                    {
                        return;
                    }
                    if (_lastPersisted != null) _lastPersisted.reset();
                    _lastPersisted = bookmark.copy();
                    BookmarkField lastDiscarded = (BookmarkField)_ring.getLastDiscarded();
                    if (lastPersistedUnset && _ring.isEmpty()
                        && lastDiscarded.Equals(EPOCH_FIELD))
                    {
                        discard(log(bookmark));
                    }

                    // If not in recovery mode, write the persisted entry to the parent.
                    if (!_parent._recovering)
                    {
                        _parent.write(_sub, ENTRY_PERSISTED, bookmark);
                    }
                    if (lastPersistedUnset || _ring.isEmpty()
                        || publisherId == lastDiscarded.getPublisherId())
                    {
                        _recoveryTimestamp = null;
                        _parent._recentChanged = true;
                        adapterUpdate((BookmarkField)_getMostRecentList(false));
                    }
                }
            }

            protected void adapterUpdate(BookmarkField bookmark_)
            {
                if (_parent._adapter != null)
                {
                    RecoveryPoint rp = _parent._factory.createRecoveryPoint(_sub,
                                                                      bookmark_);
                    Monitor.Exit(internalLock);
                    try
                    {
                        _parent._adapter.update(rp);
                    }
                    finally
                    {
                        Monitor.Enter(internalLock);
                    }
                }
            }

            // <summary>
            // Retrieves the sequence number of the oldest bookmark in the subscription's
            // bookmark ring buffer. This represents the starting point of the subscription.
            // </summary>
            // <returns>The sequence number of the oldest bookmark in the ring buffer.</returns>
            public override long getOldestBookmarkSeq()
            {
                lock (internalLock) {
                    return _ring.getStartIndex();
                }
            }

            // <summary>
            // Call on a Subscription object just after recovery is performed to
            // convert logged entries into recovery entries and set the publishers
            // cache state to the earliest sequence seen for each publisher minus
            // one.
            // </summary>
            // NOTE: If after recovery lastDiscarded for this sub is null (i.e.
            // nothing was discarded) and lastPersisted is null or EPOCH (i.e.
            // no persisted ack was recorded in the log), then we can throw away
            // this subscription state and return this Subscription object to the
            // pool.
            // <returns> Indicates whether this subscription's state was reset
            // (because it's undeeded due to the lack of any discards and persisted
            // acks) and this object should be returned to the Subscription pool. </returns>
            public bool justRecovered()
            {
                lock (internalLock)
                {
                    BookmarkField ld = _ring.getLastDiscarded();
                    if ((ld == null || ld.isNull() || ld.Equals(EPOCH_FIELD))
                            && (_lastPersisted == null || _lastPersisted.isNull()
                                    || _lastPersisted.Equals(EPOCH_FIELD))
                            && _recoveryTimestamp == null
                            && (_range.isNull() || !_range.isValid()))
                    {
                        // Reset this sub for reuse.
                        reset();
                        return true;
                    }

                    updateRecovery();
                    IList<BookmarkRingBuffer.Entry> entryList = new List<BookmarkRingBuffer.Entry>();
                    getActiveEntries(entryList);
                    setPublishersToDiscarded(entryList, _publishers);
                }

                return false;
            }

            // <summary>
            // Update the provided publishers map with the highest sequence number
            // that has been discarded for each publisher based on the given list of
            // active bookmark entries.
            // </summary>
            // <param name="active">A list of active bookmark entries to consider for
            // updating the publishers' highest discarded sequence numbers.</param>
            // <param name="publishers">A map that associates publisher IDs with their highest
            // discarded sequence numbers. This map will be updated with the latest discarded
            // sequence numbers for each publisher.</param>
            public static void setPublishersToDiscarded(
                                        IList<BookmarkRingBuffer.Entry> active,
                                        Dictionary<ulong, ulong> publishers)
            {
                if (active == null || publishers == null || publishers.Count == 0)
                {
                    return;
                }
                foreach (BookmarkRingBuffer.Entry entry in active)
                {
                    if (!entry.isActive()) continue;
                    BookmarkField bf = entry.getBookmark();
                    ulong publisher = bf.getPublisherId();
                    ulong seq = bf.getSequenceNumber();
                    if (publisher == 0 || seq == 0) continue;
                    if (publishers[publisher] >= seq) {
                        publishers[publisher] = (seq - 1);
                    }
                }
            }

            // <summary>
            // Sets the resize handler for the bookmark store. The resize handler is
            // responsible for managing file size when the bookmark store needs to be
            // resized.
            // </summary>
            // <param name="handler">The `BookmarkStoreResizeHandler` responsible for handling file
            // size management.</param>
            // <param name="store">The `BookmarkStore` associated with the resize handler.</param>
            public override void setResizeHandler(BookmarkStoreResizeHandler handler,
                                                  BookmarkStore store)
            {
                _ring.setResizeHandler(handler, store);
            }

            // <summary>
            // Lock self's internal lock. This method is used by the `LoggedBookmarkStore`
            // when gathering subscriptions to prune.
            // </summary>
            internal void _lock()
            {
                Monitor.Enter(internalLock);
            }

            // <summary>
            // Unlock self's internal lock. This method is used by the `LoggedBookmarkStore`
            // when finishing a prune operation.
            // </summary>
            internal void _unlock()
            {
                Monitor.Exit(internalLock);
            }
        }

        Dictionary<Field, Subscription> _subs = new Dictionary<Field, Subscription>();
        FileStream _fileStream = null;
        string _path;
        bool _recovering = false;
        bool _recentChanged = false;
        private byte[] _readIntBytes = new byte[4];
        int VERSION = 2;
        Pool<Subscription> _pool;
        BookmarkStoreResizeHandler _resizeHandler = null;
        private int _serverVersion = Client.MIN_MULTI_BOOKMARK_VERSION;
        private Object internalStoreLock = new Object();
        private Object subsLock = new Object();
        RecoveryPointAdapter _adapter = null;
        RecoveryPointFactory _factory = new FixedRecoveryPointFactory();

        /// <summary>
        /// This constructor is equivalent to calling `LoggedBookmarkStore(path, 1,
        /// false)`. Creates a new `LoggedBookmarkStore` instance with the specified
        /// path to the backing bookmark log file and default settings.
        /// </summary>
        /// <param name="path">The path to the backing bookmark log file.</param>
        public LoggedBookmarkStore(string path)
            : this(path, 1)
        { }

        /// <summary>
        /// This constructor is equivalent to calling `LoggedBookmarkStore(path,
        /// targetNumberOfSubscriptions, false)`. Creates a new `LoggedBookmarkStore`
        /// instance with the specified path to the backing bookmark log file and an
        /// initial capacity for subscriptions.
        /// </summary>
        /// <param name="path">The path to the backing bookmark log file.</param>
        /// <param name="targetNumberOfSubscriptions">The excpected number of subscriptions.</param>
        public LoggedBookmarkStore(string path, int targetNumberOfSubscriptions)
            : this(path, targetNumberOfSubscriptions, false)
        { }

        /// <summary>
        /// Creates a new `LoggedBookmarkStore` instance with the specified path to the backing
        /// bookmark log file and a flag indicating whether to use the last modified time.
        /// </summary>
        /// <param name="path">The path to the backing bookmark log file.</param>
        /// <param name="useLastModifiedTime">A flag indicating whether to use the last modified
        /// time of the file for the recovery timestamp.
        /// </param>
        public LoggedBookmarkStore(string path, bool useLastModifiedTime)
            : this(path, 1, useLastModifiedTime)
        { }

        /// <summary>
        /// Creates a new `LoggedBookmarkStore` instance with the specified path to the backing
        /// bookmark log file, a recovery point adapter, and an optional recovery point factory.
        /// This constructor is equivalent to calling `LoggedBookmarkStore(path, 1, false, adapter_, factory_)`.
        /// </summary>
        /// <param name="path">The path to the backing bookmark log file.</param>
        /// <param name="adapter_">The recovery point adapter used in case the file is deleted.</param>
        /// <param name="factory_">Optional: The recovery point factory. If not provided, a default factory is used.</param>
        public LoggedBookmarkStore(string path, RecoveryPointAdapter adapter_,
                                   RecoveryPointFactory factory_ = null)
            : this(path, 1, false, adapter_, factory_)
        { }

        /// <summary>
        /// A file-backed bookmark store implementation that fully supports discarding
        /// messages in an order different from the order they arrived (i.e. out-of-order
        /// discards) and fail-over to a replicated server. All messages must eventually
        /// be discarded, otherwise memory usage and disk space used will increase in
        /// proportion to the number of messages received on a bookmark subscription since
        /// the first undiscarded message. This implementation requires that the prune()
        /// method be called periodically to removed unneeded discarded bookmarks from the
        /// backing-file, otherwise it will continue to grow without bound. The prune() method
        /// is thread-safe and can be called from any thread. The store also has a backup
        /// RecoveryPointAdapter used in case the file is deleted.
        /// </summary>
        /// <param name="path">The path to the backing bookmark log file. </param>
        /// <param name="targetNumberOfSubscriptions">The initial capacity for the number of
        /// bookmark subscriptions you anticipate creating on the AMPS client instance that
        /// this bookmark store is registered on.This will grow as needed if more subscriptions
        /// are created than anticipated.</param>
        /// <param name="adapter_">The recovery point adapter used in case the file is deleted.</param>
        /// <param name="factory_">Optional: The recovery point factory. If not provided, a default
        /// factory is used.</param>
        public LoggedBookmarkStore(string path, int targetNumberOfSubscriptions,
                                   RecoveryPointAdapter adapter_,
                                   RecoveryPointFactory factory_ = null)
            : this(path, targetNumberOfSubscriptions, false, adapter_, factory_)
        { }

        /// <summary>
        /// Constructor for the `LoggedBookmarkStore` class with additional parameters for file-related
        /// settings.
        /// </summary>
        /// <param name="path">The path to the backing bookmark log file.</param>
        /// <param name="targetNumberOfSubscriptions">The initial capacity for the number of bookmark
        /// subscriptions anticipated on the AMPS client instance. It grows as needed if more subscriptions
        /// are created than anticipated.</param>
        /// <param name="useLastModifiedTime">A flag indicating whether to use the last modified time of
        /// the file for recovery timestamp.</param>
        public LoggedBookmarkStore(string path, int targetNumberOfSubscriptions,
                                   bool useLastModifiedTime)
        {
            _pool = new Pool<Subscription>(targetNumberOfSubscriptions);
            _path = path;
            string recoveryTimestamp = null;
            FileInfo fInfo = new FileInfo(path);
            if (useLastModifiedTime && fInfo.Exists)
            {
                DateTime ts = fInfo.LastWriteTimeUtc;
                recoveryTimestamp = $"{ts:yyyyMMdd}T{ts:hhmmss}Z";
            }
            fInfo = null;
            _fileStream = new FileStream(_path, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite);
            try
            {
                recover();
                if (recoveryTimestamp != null)
                {
                    foreach (var sub in _subs.Values)
                    {
                        sub.setRecoveryTimestamp(recoveryTimestamp);
                    }
                }
            }
            catch (IOException)
            {
                try
                {
                    _fileStream.Dispose();
                }
                catch (IOException)
                {
                }
                finally
                {
                    _fileStream = null;
                }
                throw;
            }
        }

        /// <summary>
        /// Constructor for the `LoggedBookmarkStore` class with additional parameters for file-related settings.
        /// </summary>
        /// <param name="path">The path to the backing bookmark log file.</param>
        /// <param name="targetNumberOfSubscriptions">The initial capacity for the number of bookmark subscriptions
        /// anticipated on the AMPS client instance. It grows as needed if more subscriptions are created than
        /// anticipated.</param>
        /// <param name="useLastModifiedTime"> A flag indicating whether to use the last modified time of the file
        /// for recovery timestamp.</param>
        /// <param name="adapter_">The recovery point adapter used in case the file is deleted.</param>
        /// <param name="factory_">Optional: The recovery point factory. If not provided, a default factory is used.</param>
        public LoggedBookmarkStore(string path, int targetNumberOfSubscriptions,
                                   bool useLastModifiedTime,
                                   RecoveryPointAdapter adapter_,
                                   RecoveryPointFactory factory_ = null)
        {
            _pool = new Pool<Subscription>(targetNumberOfSubscriptions);
            _path = path;
            string recoveryTimestamp = null;
            Message m = new JSONMessage(Encoding.UTF8, Encoding.UTF8);
            try {
                _recovering = true;
                foreach (var rp in adapter_)
                {
                    Field subId = rp.getSubId();
                    m.reset();
                    m.setSubId(subId.buffer, subId.position, subId.length);
                    BookmarkField bookmark = rp.getBookmark();
                    if (bookmark.isRange())
                    {
                        m.setBookmark(bookmark.buffer, bookmark.position,
                                      bookmark.length);
                        log(m);
                    }
                    else
                    {
                        foreach (BookmarkField bm in bookmark.parseBookmarkList())
                        {
                            if (!bm.isTimestamp())
                            {
                                m.setBookmark(bm.buffer, bm.position, bm.length);
                                isDiscarded(m);
                                log(m);
                                discard(m);
                            }
                            else
                            {
                                find(subId).setRecoveryTimestamp(bm.ToString());
                            }
                        }
                    }
                }
            }
            finally
            {
                _recovering = false;
            }
            FileInfo fInfo = new FileInfo(path);
            if (fInfo.Exists)
            {
                if (useLastModifiedTime)
                {
                    DateTime ts = fInfo.LastWriteTimeUtc;
                    recoveryTimestamp = $"{ts:yyyyMMdd}T{ts:hhmmss}Z";
                }
                // Trust file over adapter
                _purge();
            }
            _fileStream = new FileStream(_path, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite);
            try
            {
                recover();
                if (recoveryTimestamp != null)
                {
                    foreach (var sub in _subs)
                    {
                        sub.Value.setRecoveryTimestamp(recoveryTimestamp);
                        adapter_.update(_factory.createRecoveryPoint(sub.Key,
                          (BookmarkField)sub.Value.getMostRecentList(false)));
                    }
                }
            }
            catch (IOException)
            {
                try
                {
                    _fileStream.Dispose();
                }
                catch (IOException)
                {
                }
                finally
                {
                    _fileStream = null;
                }
                throw;
            }
            fInfo = null;
            // Set this after recovery is complete so no updates are sent
            // while recovering.
            _adapter = adapter_;
            if (factory_ != null) _factory = factory_;
        }

        /// <summary>
        /// Remove outdated entries in the bookmark file. This function creates
        /// a temporary file, writes active entries to that file, and then replaces
        /// the current file with the temporary file.
        /// Called on a LoggedBookmarkStore to keep the file from growing too
        /// large; see the Developer Guide for advice on stratgies for
        /// when to call this function. If this function is not called,
        /// the file can grow indefinitely until it's too large and causes
        /// an exception.
        /// </summary>
        public void prune()
        {
            string name = _path + ".tmp";
            lock (subsLock) {
                if (_fileStream == null)
                {
                    throw new StoreException("BookmarkStore file is not open");
                }
                _prune(name);
            }
        }

        /// <summary>
        /// Remove outdated entries in the bookmark file. This function creates
        /// a temporary file, tmpFileName_, writes active entries to that file,
        /// and then replaces the current file with the temporary file.
        /// Called on a LoggedBookmarkStore to keep the file from growing too
        /// large; see the Developer Guide for advice on stratgies for
        /// when to call this function. If this function is not called,
        /// the file can grow indefinitely until it's too large and causes
        /// an exception.
        /// </summary>
        /// <param name="tmpFileName_">The name of the temporary file.</param>
        ///
        public void prune(String tmpFileName_)
        {
            lock (subsLock) {
                if (_fileStream == null)
                {
                    throw new StoreException("BookmarkStore file is not open");
                }
                _prune(tmpFileName_);
            }
        }

        // Lock is acquired, check if prune is warranted, lock the subs and __prune
        public void _prune(String tmpFileName_)
        {
            if (_fileStream == null)
            {
                throw new StoreException("BookmarkStore file is not open");
            }
            if (!_recentChanged) return;
            List<KeyValuePair<Field,Subscription>> subs = null;
            try {
                subs = _lockSubs();
                _prune(tmpFileName_, subs);
                _recentChanged = false;
            }
            finally {
                if (subs != null) _unlockSubs(subs);
            }
        }

        // The real implementation of pruning the file.
        private void _prune(String tmpFileName_,
                            List<KeyValuePair<Field,Subscription>> subs)
        {
            FileStream file = new FileStream(tmpFileName_, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite);
            // Write out our version and return
            file.Write(BitConverter.GetBytes(this.VERSION), 0, 4);
            file.WriteByte((byte)'\n');
            lock (internalStoreLock)
            {
                try
                {
                    StringBuilder bookmarkBuilder = new StringBuilder(64);
                    StringField bookmark = new StringField();
                    foreach (var pairs in subs)
                    {
                        Field subId = pairs.Key;
                        Subscription sub = pairs.Value;
                        BookmarkField recent = (BookmarkField)sub.getMostRecent(false);
                        if (recent == null || recent.isNull())
                        {
                            recent.copyFrom(Subscription.EPOCH_FIELD);
                        }
                        ulong recentPub = recent.getPublisherId();
                        ulong recentSeq = recent.getSequenceNumber();
                        if (sub.getRange().isValid())
                        {
                            // Update and return the range
                            Field range = sub.getMostRecentList(true);
                            // A range is automatically marked discarded when
                            // it is logged, so no need to write the discard
                            writeBookmarkToFile(file, subId, range,
                                                ENTRY_BOOKMARK);
                            // Ignore recent after this because it's a range
                            recentPub = 0;
                            recent.reset();
                        }
                        IList<BookmarkRingBuffer.Entry> recovered = new List<BookmarkRingBuffer.Entry>();
                        sub.getActiveEntries(recovered);
                        Dictionary<ulong, ulong> publishers = new Dictionary<ulong, ulong>(sub._publishers);
                        Subscription.setPublishersToDiscarded(recovered,
                                                              publishers);

                        // First write highest discarded bookmark for each
                        // publisher id skipping the most recent.
                        foreach (var publisher in publishers)
                        {
                            ulong pubId = publisher.Key;
                            ulong pubSeq = publisher.Value;
                            if (pubId == 0 || pubSeq == 0 || pubId == recentPub)
                                continue;
                            bookmarkBuilder.Clear();
                            bookmarkBuilder.Append(pubId).Append("|")
                                           .Append(pubSeq) .Append("|");
                            bookmark.setValue(bookmarkBuilder.ToString(),
                                              System.Text.Encoding.ASCII);
                            writeBookmarkToFile(file, subId, bookmark,
                                                ENTRY_BOOKMARK);
                            writeBookmarkToFile(file, subId, bookmark,
                                                ENTRY_DISCARD);
                        }
                        // Write most recent
                        if (recent.length > 1)
                        {
                            writeBookmarkToFile(file, subId, recent,
                                                ENTRY_BOOKMARK);
                            writeBookmarkToFile(file, subId, recent,
                                                ENTRY_DISCARD);
                        }
                        // Write last persisted
                        bookmark = sub._lastPersisted;
                        if (bookmark != null && !bookmark.isNull() &&
                            bookmark.length > 1)
                        {
                            writeBookmarkToFile(file, subId, bookmark,
                                                ENTRY_PERSISTED);
                        }
                        // Finally write everything since most recent
                        foreach (var entry in recovered)
                        {
                            // Skip nulled out entries.
                            bookmark = entry.getBookmark();
                            if (bookmark == null || bookmark.isNull())
                                continue;
                            writeBookmarkToFile(file, subId, bookmark,
                                                ENTRY_BOOKMARK);
                            if (!entry.isActive())
                            {
                                writeBookmarkToFile(file, subId,
                                                    entry.getBookmark(),
                                                    ENTRY_DISCARD);
                            }
                        }
                    }
                }
                catch (IOException e)
                {
                    file.Dispose();
                    // We can ignore a failure of delete here, prune has failed
                    try
                    {
                        System.IO.File.Delete(tmpFileName_);
                    }
                    catch (IOException) { }

                    throw new StoreException("Failed attempting to prune file " + _path + " to " + tmpFileName_, e);
                }
                // Close the files, delete the original, move the pruned file
                file.Dispose();
                _fileStream.Dispose();
                _fileStream = null;
                int retries = 0;
                while (retries++ < 3)
                {
                    try
                    {
                        System.IO.File.Delete(_path);
                        break;
                    }
                    catch(Exception e)
                    {
                        if (retries >= 3)
                            throw new StoreException("Failed to delete original file " + _path + " after completing prune to " + tmpFileName_, e);
                    }
                }
                // Let file system catch up
                while (System.IO.File.Exists(_path))
                {
                    System.Threading.Thread.Sleep(100);
                }
                retries = 0;
                while (retries++ < 3)
                {
                    try
                    {
                        System.IO.File.Move(tmpFileName_, _path);
                        break;
                    }
                    catch(Exception e)
                    {
                        if (retries >= 3)
                            throw new StoreException("Failed to rename pruned file " + tmpFileName_ + " to original file name: " + _path, e);
                    }
                }
                _fileStream = new FileStream(_path, FileMode.OpenOrCreate,
                                             FileAccess.ReadWrite,
                                             FileShare.ReadWrite);
                if(_fileStream.Length > 1)
                    _fileStream.Seek(_fileStream.Length, SeekOrigin.Begin);
            }

        }

        // Convenience wrapper to write an int
        private void writeInt(FileStream file_, int val_)
        {
            file_.Write(BitConverter.GetBytes(val_), 0, 4);
        }

        // Convenience wrapper to read an int
        private int readInt(FileStream file_)
        {
            file_.Read(_readIntBytes, 0, 4);
            return BitConverter.ToInt32(_readIntBytes, 0);
        }

        // Convenience wrapper to write a bookmark
        private void writeBookmarkToFile(FileStream file_, Field subId_,
                                         Field bookmark_, byte entry_)
        {
            writeInt(file_, subId_.length);
            file_.Write(subId_.buffer, subId_.position, subId_.length);
            file_.WriteByte(entry_);
            writeInt(file_, bookmark_.length);
            file_.Write(bookmark_.buffer, bookmark_.position, bookmark_.length);
            file_.WriteByte((byte)'\n');
            file_.Flush();
        }

        // Convenience wrapper to write an entry
        void write(Field sub, byte entry, Field data)
        {
            lock (internalStoreLock) {
                if (!_recovering)
                {
                    writeBookmarkToFile(_fileStream, sub, data, entry);
                }
            }
        }

        // Recover from what's in the file, used at startup.
        private void recover()
        {
            _recovering = true;
            if (_fileStream.Length < 4)
            {
                // Write out our version and return
                _fileStream.Write(BitConverter.GetBytes(this.VERSION), 0, 4);
                _fileStream.WriteByte((byte)'\n');
                _recovering = false;
                return;
            }
            int len = readInt(_fileStream);
            // In prior versions, length for a subid and bookmark was limted
            // to 255 characters. Now, it can be up to max int. Old files also
            // didn't write a version to start the file.
            bool readInts =  (len == this.VERSION);
            // There was no version information at the start, so reset position
            if (!readInts)
            {
                _fileStream.Position = 0;
            }
            else
            {
                // Read the newline
                _fileStream.ReadByte();
                if (_fileStream.Length == 5)
                {
                    _recovering = false;
                    return;
                }
            }
            Dictionary<Field, Dictionary<BookmarkField, long>> ids = new Dictionary<Field, Dictionary<BookmarkField, long>>();
            List<BookmarkField> removeRemoveList = new List<BookmarkField>();
            int maxSubLen = 255;
            byte[] sub = new byte[maxSubLen];
            int maxBookmarkLen = 255;
            byte[] bookmark = new byte[maxBookmarkLen];
            long lastGoodPosition = _fileStream.Position;
            int subLen = 0;
            Field subId = new Field();
            BookmarkField bookmarkField = new BookmarkField();
            try
            {
                while (lastGoodPosition < _fileStream.Length)
                {
                    Field subIdCopy = null;
                    long position = _fileStream.Position;
                    subLen = readInts ? readInt(_fileStream) : (int)_fileStream.ReadByte();
                    if (subLen >= _fileStream.Length - _fileStream.Position)
                    {
                        // File is corrupt
                        throw new IOException("Invalid subId length " + subLen + " read during recovery of LoggedBookmarkStore " + _path + " at position " + _fileStream.Position);
                    }
                    if (subLen > maxSubLen)
                    {
                        do
                        {
                            maxSubLen *= 2;
                        } while (subLen > maxSubLen);
                        sub = new byte[maxSubLen];
                    }
                    _fileStream.Read(sub, 0, subLen);
                    subId.set(sub, 0, subLen);
                    Dictionary<BookmarkField, long> subscriptionMap = null;
                    if (!ids.TryGetValue(subId, out subscriptionMap))
                    {
                        subscriptionMap = new Dictionary<BookmarkField, long>();
                        subIdCopy = subId.copy();
                        ids.Add(subIdCopy, subscriptionMap);
                    }
                    Subscription subscription = find(subId);

                    switch (_fileStream.ReadByte())
                    {
                        case -1:
                            // done
                            return;
                        case ENTRY_BOOKMARK:
                            len = readInts ? readInt(_fileStream) : (int)_fileStream.ReadByte();
                            if (len == -1 ||
                                len > (_fileStream.Length - _fileStream.Position))
                            {
                                // File is corrupt
                                _fileStream.Position = position;
                                throw new IOException("Invalid bookmark length " + len + " read during recovery of LoggedBookmarkStore " + _path + " at position " + _fileStream.Position);
                            }
                            if (len > maxBookmarkLen)
                            {
                                do
                                {
                                    maxBookmarkLen *= 2;
                                } while (len > maxBookmarkLen);
                                bookmark = new byte[maxBookmarkLen];
                            }
                            _fileStream.Read(bookmark, 0, len);
                            bookmarkField.set(bookmark, 0, len);


                            if (bookmarkField.isRange())
                            {
                                subscription.log(bookmarkField);
                            }
                            else
                            {
                                if (subscriptionMap.ContainsKey(bookmarkField))
                                {
                                    subscription.getMostRecent(true);
                                    subscriptionMap.Clear();
                                }
                                if (!subscription.isDiscarded(bookmarkField))
                                {
                                    long addedIndex = subscription.log(bookmarkField);
                                    subscriptionMap.Add(bookmarkField.copy(), addedIndex);
                                }
                                else
                                {
                                    subscriptionMap.Add(bookmarkField.copy(), 0);
                                }
                            }
                            break;

                        case ENTRY_DISCARD:
                            len = readInts ? readInt(_fileStream) : (int)_fileStream.ReadByte();
                            if (len == -1 ||
                                len > (_fileStream.Length - _fileStream.Position))
                            {
                                // File is corrupt
                                _fileStream.Position = position;
                                throw new IOException("Invalid bookmark length " + len + " read during recovery of LoggedBookmarkStore " + _path + " at position " + _fileStream.Position);
                            }
                            if (len > maxBookmarkLen)
                            {
                                do
                                {
                                    maxBookmarkLen *= 2;
                                } while (len > maxBookmarkLen);
                                bookmark = new byte[maxBookmarkLen];
                            }
                            _fileStream.Read(bookmark, 0, len);
                            bookmarkField.set(bookmark, 0, len);

                            long subscriptionMapEntry = 0;
                            if (subscriptionMap.TryGetValue(bookmarkField,
                                                            out subscriptionMapEntry))
                            {
                                subscriptionMap.Remove(bookmarkField);
                                if (subscriptionMapEntry > 0)
                                    subscription.discard(subscriptionMapEntry);
                            }
                            break;

                        case ENTRY_PERSISTED:
                            len = readInts ? readInt(_fileStream) : (int)_fileStream.ReadByte();
                            if (len == -1 ||
                                len > (_fileStream.Length - _fileStream.Position))
                            {
                                // File is corrupt
                                _fileStream.Position = position;
                                throw new IOException("Invalid bookmark length " + len + " read during recovery of LoggedBookmarkStore " + _path + " at position " + _fileStream.Position);
                            }
                            if (len > maxBookmarkLen)
                            {
                                do
                                {
                                    maxBookmarkLen *= 2;
                                } while (len > maxBookmarkLen);
                                bookmark = new byte[maxBookmarkLen];
                            }
                            _fileStream.Read(bookmark, 0, len);
                            bookmarkField.set(bookmark, 0, len);

                            subscription.setLastPersisted(bookmarkField);
                            break;

                        default:
                            throw new IOException("Invalid operation read while recovering LoggedBookmarkStore " + _path + " at position " + _fileStream.Position);
                    }
                    // Read the new line
                    _fileStream.ReadByte();
                    lastGoodPosition = _fileStream.Position;
                }
            }
            catch(IOException ex)
            {
                // Check if we made progress to the last line of the file or not.
                if (lastGoodPosition > 0)
                {
                    try
                    {
                        while (_fileStream.Position < _fileStream.Length)
                        {
                            if (_fileStream.ReadByte() == (byte)'\n')
                            {
                                // There's more to the file, let's throw the exception
                                // for investigation into the corruption.
                                throw new IOException("IOException while reading " + _path + " after position " + lastGoodPosition + ". See inner exception.", ex);
                            }
                        }
                        // Only final line is corrupt, so pick up from here since it
                        // was likely caused by an untimely exit while writing.
                        _fileStream.Seek(lastGoodPosition, SeekOrigin.Begin);
                    }
                    catch (IOException exc)
                    {
                        // Something is very wrong, let's throw the first error
                        throw new IOException("IOException while reading " + _path + " after position " + lastGoodPosition + ". See inner exception.", exc);
                    }
                }
                else
                {
                    // We didn't recover anything at all, but there's something there.
                    // Let's just fail.
                    throw new IOException("IOException while reading " + _path + " at the start of the file", ex);
                }
            }
            finally
            {
                List<Field> toRemove = null;
                foreach (KeyValuePair<Field, Subscription> s in _subs)
                {
                    bool disposeSub = s.Value.justRecovered();
                    if (disposeSub)
                    {
                        if (toRemove == null) toRemove = new List<Field>();

                        toRemove.Add(s.Key);
                        _pool.returnToPool(s.Value);
                    }
                }
                if (toRemove != null)
                {
                    foreach (Field sid in toRemove)
                    {
                        _subs.Remove(sid);
                    }
                }
                _recovering = false;
                // This was an older version, so recreate the file using prune
                if (!readInts)
                {
                    prune();
                }
            }
        }

        /// <summary>
        /// This is a method the Client calls and is not for customer user. Logs a
        /// message's bookmark to the bookmark store and updates the message's bookmark
        /// sequence number.
        /// </summary>
        /// <param name="message">The message to log, containing the bookmark to be stored.</param>
        /// <returns>The index at which the bookmark is stored in the subscription.</returns>
        /// <exception cref="Exceptions.AMPSException">If an error occurs while logging to the bookmark store.</exception>
        public long log(Message message)
        {
            BookmarkField bookmark = message.getBookmarkRaw();
            if (bookmark.isTimestamp() ||
                bookmark.Equals(Subscription.EPOCH_FIELD))
            {
                return 0;
            }
            LoggedBookmarkStore.Subscription sub = (LoggedBookmarkStore.Subscription)message.getSubscription();
            Field subId = message.getSubIdRaw();
            if (subId == null || subId.isNull())
                subId = message.getSubIdsRaw();
            try
            {
                lock(internalStoreLock)
                {
                    if (_fileStream == null && !_recovering)
                    {
                        throw new StoreException("BookmarkStore file is not open");
                    }
                }
                if (sub == null)
                {
                    sub = find(subId);
                    message.setSubscription(sub);
                }
                long index = sub.log(bookmark);
                message.setBookmarkSeqNo(index);

                // Log the arrival of the bookmark
                if (!_recovering)
                {
                    write(subId, ENTRY_BOOKMARK, bookmark);
                }

                return index;
            }
            catch (IOException ioex)
            {
                throw new AMPSException("Error logging to bookmark store", ioex);
            }
        }

        /// <summary>
        /// Discards a bookmark entry associated with a specific subscription and bookmark
        /// sequence number. Calling discard with the Message can be faster.
        /// </summary>
        /// <param name="subId">The identifier of the subscription.</param>
        /// <param name="bookmarkSeqNo">The sequence number of the bookmark entry to discard.</param>
        /// <exception cref="AMPSException">If an error occurs while discarding from the bookmark
        /// store.</exception>
        public void discard(Field subId, long bookmarkSeqNo)
        {
            lock(internalStoreLock)
            {
                if (_fileStream == null && !_recovering)
                {
                    throw new StoreException("BookmarkStore file is not open");
                }
            }
            try
            {
                find(subId).discard(bookmarkSeqNo);
            }
            catch (IOException ioex)
            {
                throw new AMPSException("Error discarding from bookmark store", ioex);
            }
        }

        /// <summary>
        /// Discards a bookmark entry based on the information provided in the message.
        /// </summary>
        /// <param name="message">The message containing information about the bookmark to
        /// discard.</param>
        /// <exception cref="AMPSException">If an error occurs while discarding from the bookmark
        /// store.</exception>
        public void discard(Message message)
        {
            lock(internalStoreLock)
            {
                if (_fileStream == null && !_recovering)
                {
                    throw new StoreException("BookmarkStore file is not open");
                }
            }
            try
            {
                BookmarkField bookmark = message.getBookmarkRaw();
                if (bookmark.isTimestamp() || bookmark.isRange()
                    || bookmark.Equals(Client.Bookmarks.NOW)
                    || bookmark.Equals(Client.Bookmarks.EPOCH))
                {
                    return;
                }
                long bookmarkSeq = message.getBookmarkSeqNo();
                LoggedBookmarkStore.Subscription sub = (LoggedBookmarkStore.Subscription)message.getSubscription();
                if (sub == null)
                {
                    Field subId = message.getSubIdRaw();
                    if (subId == null || subId.isNull())
                        subId = message.getSubIdsRaw();
                    sub = find(subId);
                    message.setSubscription(sub);
                }
                sub.discard(bookmarkSeq);
            }
            catch (IOException ioex)
            {
                throw new AMPSException("Error discarding Message from LoggedBookmarkStore", ioex);
            }
        }

        /// <summary>
        /// Retrieves the most recent bookmark associated with a subscription identified by the
        /// given subId.
        /// </summary>
        /// <param name="subId">The identifier of the subscription for which to retrieve the
        /// most recent bookmark.</param>
        /// <param name="useList">A flag indicating whether to retrieve the most recent bookmark
        /// as a list or just the single bookmark that was last discarded. A
        /// list is the only way to guarantee successful failover and no missed
        /// message.</param>
        /// <returns>The most recent bookmark for the specified subscription.</returns>
        /// <exception cref="AMPSException">If an error occurs while retrieving the most recent
        /// bookmark or if the store is not open.</exception>
        public Field getMostRecent(Field subId, bool useList = true)
        {
            lock(internalStoreLock)
            {
                if (_fileStream == null && !_recovering)
                {
                    throw new StoreException("BookmarkStore file is not open");
                }
            }
            try
            {
                return find(subId).getMostRecentList(useList);
            }
            catch (IOException ioex)
            {
                throw new AMPSException("Error getting most recent from bookmark store", ioex);
            }
        }

        /// <summary>
        /// This method is called internally by the client to determine whether a message's
        /// bookmark received from the amps server has already been discarded OR already been
        /// delivered to the subscriber during this run.
        /// </summary>
        /// <param name="message">The message containing the bookmark to check for discarding.</param>
        /// <returns>True if the bookmark has been discarded; false otherwise.</returns>
        /// <exception cref="AMPSException">If an error occurs while checking if the bookmark is
        /// discarded or if the store is not open.</exception>
        public bool isDiscarded(Message message)
        {
            lock(internalStoreLock)
            {
                if (_fileStream == null && !_recovering)
                {
                    throw new StoreException("BookmarkStore file is not open");
                }
            }
            try
            {
                BookmarkField bookmark = message.getBookmarkRaw();
                if (bookmark.isTimestamp())
                    return false;
                if (bookmark.Equals(Subscription.EPOCH_FIELD))
                    return true;
                LoggedBookmarkStore.Subscription sub = (LoggedBookmarkStore.Subscription)message.getSubscription();
                if (sub == null)
                {
                    Field subId = message.getSubIdRaw();
                    if (subId == null || subId.isNull())
                        subId = message.getSubIdsRaw();
                    sub = find(subId);
                    message.setSubscription(sub);
                }
                return sub.isDiscarded(bookmark);
            }
            catch (IOException ioex)
            {
                throw new AMPSException("Error checking is discarded in bookmark store", ioex);
            }
        }

        /// <summary>
        /// Deprecated method for setting a persisted bookmark. Use
        /// <see cref="AMPS.Client.LoggedBookmarkStore.persisted(Field, BookmarkField)"/> instead.
        /// </summary>
        /// <param name="subId">The identifier of the subscription for which to set the persisted
        /// bookmark.</param>
        /// <param name="bookmark">The persisted bookmark value to set.</param>
        /// <exception cref="AMPSException">If an error occurs while setting the persisted bookmark
        /// or if the store is not open.</exception>
        public void persisted(Field subId, long bookmark)
        {
            lock(internalStoreLock)
            {
                if (_fileStream == null && !_recovering)
                {
                    throw new StoreException("BookmarkStore file is not open");
                }
            }
            try
            {
                find(subId).setLastPersisted(bookmark);
            }
            catch (IOException ioex)
            {
                throw new AMPSException("Error logging persisted to bookmark store", ioex);
            }
        }

        /// <summary>
        /// This method is called by the client and it is used to process persisted
        /// acknowledgements that track a safe recovery point in the txlog. Sets a
        /// persisted bookmark for a subscription.
        /// </summary>
        /// <param name="subId">The identifier of the subscription for which to set the persisted
        /// bookmark.</param>
        /// <param name="bookmark">The persisted bookmark value to set.</param>
        /// <exception cref="AMPSException">If an error occurs while setting the persisted bookmark
        /// or if the store is not open.</exception>
        public void persisted(Field subId, BookmarkField bookmark)
        {
            lock(internalStoreLock)
            {
                if (_fileStream == null && !_recovering)
                {
                    throw new StoreException("BookmarkStore file is not open");
                }
            }
            try
            {
                if (bookmark.Equals(Client.Bookmarks.NOW)
                    || bookmark.Equals(Subscription.EPOCH_FIELD))
                    return;
                find(subId).setLastPersisted(bookmark);
            }
            catch (IOException ioex)
            {
                throw new AMPSException("Error logging persisted  to bookmark store", ioex);
            }
        }

        /// <summary>
        /// Retrieves the oldest active bookmark sequence number that is not
        /// discarded for the given subscription.
        /// </summary>
        /// <param name="subId">The identifier of the subscription for which to retrieve the oldest
        /// bookmark sequence number.</param>
        /// <returns>The oldest bookmark sequence number for the specified subscription.</returns>
        /// <exception cref="AMPSException">If an error occurs while retrieving the oldest bookmark
        /// sequence number or if the store is not open.</exception>
        public long getOldestBookmarkSeq(Field subId)
        {
            lock(internalStoreLock)
            {
                if (_fileStream == null && !_recovering)
                {
                    throw new StoreException("BookmarkStore file is not open");
                }
            }
            long retVal = 0;
            try
            {
                // Retrieve and return the oldest bookmark sequence number for the specified
                // subscription.
                retVal = find(subId).getOldestBookmarkSeq();
            }
            catch (IOException ioex)
            {
                throw new AMPSException("Error getting oldest bookmark seq from bookmark store", ioex);
            }
            return retVal;
        }

        /// <summary>
        /// Sets a handler for bookmark store resize events.
        /// </summary>
        /// <param name="handler">The BookmarkStoreResizeHandler to set.</param>
        public void setResizeHandler(BookmarkStoreResizeHandler handler)
        {
            lock (subsLock) {
                lock (internalStoreLock) {
                    // holds the handler responsible for managing resize events in the bookmark
                    // store.
                    _resizeHandler = handler;
                    foreach (KeyValuePair<Field, Subscription> sub in _subs)
                    {
                        sub.Value.setResizeHandler(handler, this);
                    }
                }
            }
        }

        // Internal method to find a Subscription based on its subId.
        private Subscription find(Field subId)
        {
            lock (subsLock) {
                Subscription s = null;
                if (_subs.TryGetValue(subId, out s) == false)
                {
                    s = _pool.get();
                    s.init(subId, this);
                    s.setResizeHandler(_resizeHandler, this);
                    _subs.Add(subId.copy(), s);
                }
                return s;
            }
        }

        /// <summary>
        /// Purges the entire bookmark store, removing all entries. This method
        /// truncates the file and all information about any subscriptions is
        /// removed. This should not be called if the store is still attached
        /// to a connected Client with any active subscriptions.
        /// </summary>
        public void purge()
        {
            if (_fileStream == null)
            {
                throw new StoreException("BookmarkStore file is not open");
            }
            _purge();
            lock(internalStoreLock) {
                _recentChanged = true;
                try
                {
                    _fileStream.SetLength(0);
                    // Write out our version and return
                    _fileStream.Write(BitConverter.GetBytes(this.VERSION), 0, 4);
                    _fileStream.WriteByte((byte)'\n');
                }
                catch (IOException ioex)
                {
                    throw new StoreException("Error truncating file", ioex);
                }
                if (_adapter != null)
                {
                    try
                    {
                        _adapter.purge();
                    }
                    catch (AMPSException)
                    {
                        throw;
                    }
                    catch (Exception ex)
                    {
                        throw new StoreException("Exception in RecoveryPointAdapter.purge()", ex);
                    }
                }
            }
        }

        private void _purge()
        {
            lock (subsLock) {
                lock (internalStoreLock) {
                    foreach (Subscription sub in _subs.Values)
                    {
                        sub.reset();
                        _pool.returnToPool(sub);
                    }

                    _subs.Clear();
                }
            }
        }

        /// <summary>
        /// Removes all entries in the bookmark store associated with a
        /// specific subscription  ID(`subId_`). This method is used to clear
        /// messages received and discarded for a particular subscription. This
        /// should not be called if the store is still attached to a connected
        /// Client where this subId is still active.
        /// This does not reduce the underlying file size.
        /// </summary>
        /// <param name="subId_">The subscription ID for which to remove entries from the store.</param>
        public void purge(Field subId_)
        {
            lock(internalStoreLock)
            {
                if (_fileStream == null && !_recovering)
                {
                    throw new StoreException("BookmarkStore file is not open");
                }
            }
            _purge(subId_);
            if (_adapter != null)
            {
                _adapter.purge(subId_);
            }
        }

        private void _purge(Field subId_)
        {
            lock (subsLock) {
                lock (internalStoreLock) {
                    Subscription sub;
                    if (!_subs.TryGetValue(subId_, out sub)) return;
                    _subs.Remove(subId_);
                    sub.reset();
                    _pool.returnToPool(sub);
                    _recentChanged = true;
                    string name = _path + ".tmp";
                    _prune(name);
                }
            }
        }

        /// <summary>
        /// Change the RecoveryPointFactory used by this store for its adapter.
        /// </summary>
        /// <param name="factory_">The new RecoveryPointFactory.</param>
        /// <exception cref="AMPSException">If one of factory or adapter is null.</exception>
        public void setRecoveryPointFactory(RecoveryPointFactory factory_)
        {
            if (factory_ == null || _adapter == null) {
                throw new CommandException("Factory and Adapter must not be null.");
            }
            _factory = factory_;
        }

        /// <summary>
        /// Disposes of the bookmark store resources, including the adapter and file stream.
        /// </summary>
        public void Dispose()
        {
            _factory = null;
            if (_adapter != null)
            {
                _adapter.Dispose();
                _adapter = null;
            }
            if (_fileStream != null)
            {
                _fileStream.Dispose();
                _fileStream = null;
            }
        }

        /// <summary>
        /// Closes the bookmark store by disposing of the adapter and file stream.
        /// Any IOException during the close operation is caught and rethrown as
        /// a StoreException.
        /// </summary>
        /// <exception cref="StoreException">If there is an error closing the file backing in the
        /// store, or the store already closed.</exception>
        public void close()
        {
            // We have to Dispose of the adapter without the lock
            // but we'll rethrow any exception once we've closed everything
            Exception e = null;
            if (_adapter != null)
            {
                try
                {
                    _adapter.Dispose();
                }
                catch (Exception ex)
                {
                    e = ex;
                }
                finally
                {
                    _adapter = null;
                }
            }
            lock (internalStoreLock) {
                if (_fileStream == null)
                {
                    throw new StoreException("Store not open.");
                }
                try
                {
                    _fileStream.Dispose();
                    // Do we need to rethrow from adapter disposal?
                    if (e != null)
                    {
                        throw e;
                    }
                }
                catch (IOException ioex)
                {
                    throw new StoreException("Error closing file", ioex);
                }
                finally
                {
                    _fileStream = null;
                    _factory = null;
                }
            }
        }

        /// <summary>
        /// Used internally to change the version of the AMPS server that this
        /// bookmark store's client has connected to.
        /// </summary>
        /// <param name="version">An AMPS server version integer of the form 05030400 for
        /// version 5.3.4.0.</param>
        public void setServerVersion(int version)
        {
            _serverVersion = version;
        }

        /// <summary>
        /// Return the server version that was set by the Client.
        /// </summary>
        /// <returns>The server version, represented as an integer</returns>
        public int getServerVersion()
        {
            return _serverVersion;
        }

        // Lock all of the subs AFTER locking the store.
        private List<KeyValuePair<Field,Subscription>> _lockSubs()
        {
            // Must already have subsLock locked.
            List<KeyValuePair<Field,Subscription>> subs = new List<KeyValuePair<Field,Subscription>>(_subs.Count);
            foreach( var pair in _subs)
            {
                subs.Add(pair);
                pair.Value._lock();
            }
            return subs;
        }

        // Unlock all of the subs BEFORE unlocking the store.
        private void _unlockSubs(List<KeyValuePair<Field,Subscription>> subs)
        {
            foreach (var pair in subs)
            {
                pair.Value._unlock();
            }
        }

    }
}
