///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading;
using System.Collections.Generic;
using AMPS.Client.Fields;
using AMPS.Client.Exceptions;


namespace AMPS.Client
{

    /// <summary>
    /// Implements a bookmark store useful for handling server failover scenarios,
    /// but without a backing store to recover from subscriber failure.
    /// <remarks>
    /// An optional RecoveryPointAdapter can be specified at construction
    /// to persist bookmark subscription recovery state to an external store so the
    /// subscription can be resumed if the subscriber fails.
    /// </remarks>
    /// </summary>
    public class MemoryBookmarkStore : BookmarkStore
    {
        /// <summary>
        /// The Subscription object is used to represent internal bookmark state
        /// for the messages received and discarded on a specific subscription
        /// within the bookmark store.
        /// </summary>
        class Subscription : AMPS.Client.Subscription
        {
            // This subscription's ID.
            Field _sub;

            // The last persisted bookmark
            private BookmarkField _lastPersisted = null;

            // If this subscription uses a range, save it here
            private BookmarkRangeField _range = new BookmarkRangeField();

            // EPOCH field for ease-of-use
            public static BookmarkField EPOCH_FIELD = new BookmarkField(Client.Bookmarks.EPOCH);

            BookmarkRingBuffer _ring = new BookmarkRingBuffer();

            // The per-subscription memory of what we've seen from publishers
            Dictionary<ulong, ulong> _publishers = new Dictionary<ulong, ulong>();

            private Object internalLock = new Object();

            MemoryBookmarkStore _parent;

            // <summary>
            // The last-modified timestamp, if any, from a RecoveryPointAdapter
            // before recovery is initiated. When this is not null we intend to
            // include it in the list of bookmarks returned by getMostRecentList()
            // until a message is discarded on the sub or the sub state is purged.
            // </summary>
            protected volatile string _recoveryTimestamp = null;

            // <summary>
            // Constructor that creates a Subscription object with the provided
            // subscription ID and initializes its state.
            // </summary>
            // <param name="subscriptionId">The subscription ID to associate with
            // this Subscription.</param>
            public Subscription(Field subscriptionId)
            {
                _sub = subscriptionId.copy();
                _ring.setSubId(_sub);
                _lastPersisted = new BookmarkField();
                _lastPersisted.copyFrom(EPOCH_FIELD);
            }

            // <summary>
            // Default constructor that creates a Subscription object and initializes
            // its state.
            // </summary>
            public Subscription()
            {
                _lastPersisted = new BookmarkField();
                _lastPersisted.copyFrom(EPOCH_FIELD);
            }

            // <summary>
            // Sets the subscription ID for this bookmark store.
            // </summary>
            // <param name="subscriptionId">The subscription ID to set.</param>
            public void setSubscription(Field subscriptionId)
            {
                _sub = subscriptionId.copy();
                _ring.setSubId(_sub);
            }

            // <summary>
            // Sets the parent MemoryBookmarkStore for this bookmark store.
            // </summary>
            // <param name="parent">The parent MemoryBookmarkStore to set.</param>
            public void setParent(MemoryBookmarkStore parent)
            {
                _parent = parent;
            }

            // <summary>
            // The last-modified timestamp, if any, from a RecoveryPointAdapter
            // before recovery is initiated. When this is not null, we intend to
            // include it in the list of bookmarks returned by getMostRecentList()
            // until a message is discarded on the sub or the sub state is purged.
            // </summary>
            // <param name="rts">The recovery timestamp to set.</param>
            public void setRecoveryTimestamp(string rts)
            {
                _recoveryTimestamp = rts;
            }

            // <summary>
            // Logs a bookmark, either a single bookmark or a list of bookmarks and returns
            // the sequence number.
            // </summary>
            // <param name="bookmark">The bookmark to log.</param>
            // <returns>The sequence number associated with the logged bookmark.</returns>
            // <exception cref="CommandException">Thrown when a command operation fails.</exception>
            public override long log(BookmarkField bookmark)
            {
                if (bookmark.Equals(Client.Bookmarks.NOW)) return 0;
                lock (internalLock)
                {
                    if (!bookmark.isRange())
                    {
                        if (!bookmark.isBookmarkList())
                        {
                            return _log(bookmark);
                        }
                        else
                        {
                            // If we log a list, we need to mark them all
                            // as discarded.
                            long seq = 0;
                            foreach (BookmarkField bm in bookmark.parseBookmarkList())
                            {
                                isDiscarded(bm);
                                seq = _log(bm);
                                if (seq != 0)
                                {
                                    _ring.discard(seq);
                                }
                            }
                            adapterUpdate((BookmarkField)_range);
                            return 0;
                        }
                    }
                    else
                    {
                        _range.copyFrom(bookmark);
                        if (!_range.isValid())
                        {
                            _range.reset();
                            throw new CommandException("Invalid bookmark range specified");
                        }
                        long seq = 0;
                        if (_range.isStartExclusive())
                        {
                            BookmarkField start = _range.getStart();
                            if (!start.isBookmarkList())
                            {
                                isDiscarded(start);
                                seq = _log(start);
                                if (seq != 0)
                                {
                                    _ring.discard(seq);
                                }
                            }
                            else
                            {
                                foreach (BookmarkField bm in start.parseBookmarkList())
                                {
                                    isDiscarded(bm);
                                    seq = _log(bm);
                                    if (seq != 0)
                                    {
                                        _ring.discard(seq);
                                    }
                                }
                            }
                        }
                        adapterUpdate((BookmarkField)_range);
                        // The only time a range is logged is when subscribing.
                        return 0;
                    }
                }
            }

            public long _log(BookmarkField bookmark)
            {
                long seq = 0;
                if (!bookmark.isTimestamp())
                {
                    seq = _ring.log(bookmark);
                    while (seq == 0) {
                        Monitor.Exit(internalLock);
                        try {
                            _ring.checkResize();
                        }
                        finally {
                            Monitor.Enter(internalLock);
                        }
                        seq = _ring.log(bookmark);
                    }
                }
                else
                {
                    setLastPersisted(bookmark);
                }
                return seq;
            }

            // <summary>
            // Discards a message at the specified index.
            // </summary>
            // <param name="index">The index of the message to discard.</param>
            public override void discard(long index)
            {
                lock (internalLock)
                {
                    if (_ring.discard(index))
                    {
                        adapterUpdate((BookmarkField)getMostRecentList(false));
                        _recoveryTimestamp = null;
                    }
                }
            }

            // <summary>
            // Check to see if this message is older than the most recent one seen,
            // and if it is, check if it discarded.
            // </summary>
            // <param name="bookmark">The bookmark containing the publisher ID and sequence number.</param>
            // <returns>True if the message is discarded or in-flight, false otherwise.</returns>
            public override bool isDiscarded(BookmarkField bookmark)
            {
                if (bookmark.Equals(Client.Bookmarks.NOW)) return true;
                lock (internalLock)
                {
                    ulong publisher = bookmark.getPublisherId();
                    ulong sequence = bookmark.getSequenceNumber();
                    if (!_publishers.ContainsKey(publisher) ||
                        _publishers[publisher] < sequence)
                    {
                        _publishers[publisher] = sequence;
                        return false;
                    }
                    return true; // message is in flight or discarded
                }
            }

            // <summary>
            // Gets the range of bookmarks stored in this bookmark store.
            // </summary>
            // <returns>The range of bookmarks.</returns>
            public BookmarkRangeField getRange()
            {
                return _range;
            }

            // <summary>
            // Get the most recent bookmark.
            // </summary>
            // <returns>A copy of the most recent bookmark.</returns>
            public override Field getMostRecent()
            {
                lock (internalLock)
                {
                    return _ring.getLastDiscarded().copy();
                }
            }

            // <summary>
            // Get the most recent bookmark as a list of bookmarks.
            // </summary>
            // <param name="useList">Whether to return as a list or a single bookmark.</param>
            // <returns>A bookmark or a list of bookmarks depending on useList.</returns>
            public override Field getMostRecentList(bool useList)
            {
                lock (internalLock)
                {
                    bool rangeIsValid = _range.isValid();
                    BookmarkField lastDiscarded = (BookmarkField)_ring.getLastDiscarded();
                    bool useLastDiscarded = (lastDiscarded != null &&
                                            !lastDiscarded.isNull());
                    if (useLastDiscarded && lastDiscarded.Equals(EPOCH_FIELD))
                    {
                        if (rangeIsValid)
                        {
                            return _range;
                        }
                        else if (_recoveryTimestamp != null)
                        {
                            BookmarkField ret = new BookmarkField();
                            ret.setValue(_recoveryTimestamp, Encoding.ASCII);
                            return ret;
                        }
                        // Return just EPOCH
                        return EPOCH_FIELD;
                    }
                    ulong lastDiscardedPub = 0;
                    ulong lastDiscardedSeq = 0;
                    bool useLastPersisted = (_lastPersisted != null &&
                                            !_lastPersisted.isNull() &&
                                            !_lastPersisted.Equals(EPOCH_FIELD));
                    ulong lastPersistedPub = 0;
                    ulong lastPersistedSeq = 0;
                    if (useLastPersisted)
                    {
                        lastPersistedPub = _lastPersisted.getPublisherId();
                        lastPersistedSeq = _lastPersisted.getSequenceNumber();
                    }
                    if (useLastDiscarded)
                    {
                        if (_ring.isEmpty() && useLastPersisted)
                        {
                            useLastDiscarded = false;
                        }
                        else
                        {
                            lastDiscardedPub = lastDiscarded.getPublisherId();
                            lastDiscardedSeq = lastDiscarded.getSequenceNumber();
                            if (useLastPersisted &&
                                (lastDiscardedPub == lastPersistedPub))
                            {
                                useLastDiscarded = (lastDiscardedSeq < lastPersistedSeq);
                                useLastPersisted = !useLastDiscarded;
                            }
                        }
                    }
                    BookmarkField recentList = new BookmarkField();
                    StringBuilder recentStr = new StringBuilder();
                    if (_recoveryTimestamp != null)
                    {
                        recentStr.Append(_recoveryTimestamp);
                    }
                    if (useLastDiscarded)
                    {
                        if (recentStr.Length > 0) { recentStr.Append(','); }
                        recentStr.Append(lastDiscarded.getValue(System.Text.Encoding.ASCII));
                    }
                    if ((!useLastDiscarded && !useLastPersisted)
                         || (_lastPersisted != null
                             && _lastPersisted.Equals(EPOCH_FIELD))
                        && useList)
                    {
                        if (_publishers.Count == 0 && !_range.isValid())
                        {
                            if (_lastPersisted == null)
                            {
                                _lastPersisted = new BookmarkField();
                                _lastPersisted.copyFrom(EPOCH_FIELD);
                            }
                            return _lastPersisted.copy();
                        }
                        if (useLastDiscarded && lastDiscarded.Equals(EPOCH_FIELD))
                        {
                            if (recentStr.Length == 1) // EPOCH only
                                recentStr.Clear();
                            else // remove ,0 for EPOCH
                                recentStr.Remove(recentStr.Length-2, 2);
                        }
                        foreach (var item in _publishers)
                        {
                            if (item.Key == 0 ||
                                (useLastDiscarded && item.Key == lastDiscardedPub))
                            {
                                continue;
                            }
                            if (recentStr.Length > 0)
                                recentStr.Append(',');
                            recentStr.Append(item.Key)
                                     .Append('|')
                                     .Append(item.Value)
                                     .Append('|');
                        }
                        recentList.setValue(recentStr.toString(),
                                            System.Text.Encoding.ASCII);
                        if (_range.isValid())
                        {
                            if (recentList.length > 1
                                && (_range.isStartInclusive()
                                   || !recentList.Equals(_range.getStart())))
                            {
                                _range.replaceStart(recentList, true);
                            }
                            return _range.copy();
                        }
                        return recentList;
                    }
                    if (useLastPersisted)
                    {
                        if (recentStr.Length > 0) recentStr.Append(',');
                        recentStr.Append(_lastPersisted.getValue(System.Text.Encoding.ASCII));
                    }
                    recentList.setValue(recentStr.toString(), System.Text.Encoding.ASCII);
                    if (_range.isValid())
                    {
                        if (recentList.length > 1
                            && (_range.isStartInclusive()
                               || !recentList.Equals(_range.getStart())))
                        {
                            _range.replaceStart(recentList, true);
                        }
                        return _range.copy();
                    }
                    if (recentList.length == 0) return EPOCH_FIELD;
                    return recentList;
                }
            }

            // <summary>
            // Old style of setting a persisted bookmark no longer used.
            // </summary>
            // <param name="bookmark">The bookmark to be persisted.</param>
            public override void setLastPersisted(long bookmark)
            {
                lock (internalLock) {
                    BookmarkRingBuffer.Entry entry = _ring.getByIndex(bookmark);
                    if (entry == null) return;
                    if (_lastPersisted != null) _lastPersisted.reset();
                    _lastPersisted = entry.getBookmark().copy();
                    adapterUpdate((BookmarkField)getMostRecentList(false));
                    if (_parent._adapter != null)
                    {
                        _recoveryTimestamp = null;
                    }
                }
            }

            // <summary>
            // Sets the last persisted bookmark based on the provided bookmark.
            // </summary>
            // <param name="bookmark">The index of the bookmark.</param>
            public override void setLastPersisted(BookmarkField bookmark)
            {
                lock (internalLock) {
                    if (bookmark == null || bookmark.isNull()
                        || bookmark.Equals(_lastPersisted)
                        || bookmark.isRange())
                        return;
                    if (bookmark.isTimestamp())
                    {
                        _recoveryTimestamp = bookmark.ToString();
                        adapterUpdate((BookmarkField)getMostRecentList(false));
                        return;
                    }
                    ulong publisherId = bookmark.getPublisherId();
                    if (_lastPersisted != null &&
                        publisherId == _lastPersisted.getPublisherId() &&
                        bookmark.getSequenceNumber() <= _lastPersisted.getSequenceNumber())
                    {
                        return;
                    }
                    BookmarkField lastPersisted = _lastPersisted;
                    _lastPersisted = bookmark.copy();
                    _recoveryTimestamp = null;
                    if ((lastPersisted == null || lastPersisted.Equals(EPOCH_FIELD))
                        && _ring.isEmpty()
                        && _ring.getLastDiscarded().Equals(EPOCH_FIELD))
                    {
                        discard(_log(bookmark));
                    }
                    else
                    {
                        adapterUpdate((BookmarkField)getMostRecentList(false));
                    }
                    if (lastPersisted != null) lastPersisted.reset();
                }
            }

            protected void adapterUpdate(BookmarkField bookmark_)
            {
                if (_parent._adapter != null)
                {
                    RecoveryPoint rp = _parent._factory.createRecoveryPoint(_sub,
                                                                      bookmark_);
                    Monitor.Exit(internalLock);
                    try
                    {
                        _parent._adapter.update(rp);
                    }
                    finally
                    {
                        Monitor.Enter(internalLock);
                    }
                }
            }

            // <summary>
            // Gets the sequence number of the oldest active bookmark.
            // </summary>
            // <returns>The sequence number of the oldest bookmark.</returns>
            public override long getOldestBookmarkSeq()
            {
                lock (internalLock) {
                    return _ring.getStartIndex();
                }
            }

            // <summary>
            // Sets the resize handler this Subscription.
            // </summary>
            // <param name="handler">The resize handler to set.</param>
            // <param name="store">The bookmark store.</param>
            public override void setResizeHandler(BookmarkStoreResizeHandler handler,
                                                  BookmarkStore store)
            {
                lock (internalLock) {
                    _ring.setResizeHandler(handler, store);
                }
            }

        }

        Dictionary<Field, Subscription> _subs = new Dictionary<Field, Subscription>();
        BookmarkStoreResizeHandler _resizeHandler = null;
        private int _serverVersion = Client.MIN_MULTI_BOOKMARK_VERSION;

        // Subscriptions take a while to initialize, so we avoid doing this in the
        // message handler.
        // number of subscriptions we pre-initialize.
        Pool<Subscription> _pool;
        RecoveryPointAdapter _adapter = null;
        RecoveryPointFactory _factory = new FixedRecoveryPointFactory();

        private Object internalLock = new Object();

        /// <summary>
        /// Initializes a new instance of the MemoryBookmarkStore class with default settings.
        /// </summary>
        public MemoryBookmarkStore()
        {
            _pool = new Pool<Subscription>(1);
        }

        /// <summary>
        /// Initializes a new instance of the MemoryBookmarkStore class with the specified adapter and factory.
        /// </summary>
        /// <param name="adapter_">The recovery point adapter.</param>
        /// <param name="factory_">The recovery point factory.</param>
        public MemoryBookmarkStore(RecoveryPointAdapter adapter_,
                                   RecoveryPointFactory factory_ = null)
        {
            _pool = new Pool<Subscription>(1);
            Message m = new JSONMessage(Encoding.UTF8, Encoding.UTF8);
            foreach (var rp in adapter_)
            {
                Field subId = rp.getSubId();
                m.reset();
                m.setSubId(subId.buffer, subId.position, subId.length);
                BookmarkField bookmark = rp.getBookmark();
                if (bookmark.isRange())
                {
                    m.setBookmark(bookmark.buffer, bookmark.position,
                                  bookmark.length);
                    log(m);
                }
                else
                {
                    int start = bookmark.position;
                    int comma = bookmark.position;
                    for( ; comma<bookmark.position+bookmark.length; ++comma)
                    {
                        if (bookmark.buffer[comma] == (byte)',')
                        {
                            m.setBookmark(bookmark.buffer, start, comma-start);
                            isDiscarded(m);
                            log(m);
                            discard(m);
                            start = comma+1;
                        }
                    }
                    // Recovery timestamp should always be last
                    if (comma > start)
                    {
                        m.setBookmark(bookmark.buffer, start, comma-start);
                        if (!m.getBookmarkRaw().isTimestamp())
                        {
                            isDiscarded(m);
                            log(m);
                            discard(m);
                        }
                        else
                        {
                            find(subId).setRecoveryTimestamp(m.getBookmark());
                        }
                    }
                }
            }
            // Set this after recovery is complete so no updates are sent
            // while recovering.
            _adapter = adapter_;
            if (factory_ != null) _factory = factory_;
        }

        /// <summary>
        /// Initialize self with a target number of subscriptions to store.
        /// </summary>
        /// <param name="targetNumberOfSubscriptions">The number of subscriptions this store will need to track.</param>
        public MemoryBookmarkStore(int targetNumberOfSubscriptions)
        {
            _pool = new Pool<Subscription>(targetNumberOfSubscriptions);
        }

        /// <summary>
        /// Change the RecoveryPointFactory used by this store for its adaptor.
        /// </summary>
        /// <param name="factory_">The new RecoveryPointFactory</param>
        public void setRecoveryPointFactory(RecoveryPointFactory factory_)
        {
            _factory = factory_;
        }

        /// <summary>
        /// Disposes resources used by the MemoryBookmarkStore.
        /// </summary>
        public void Dispose()
        {
            try
            {
                // Dispose of the recovery point adapter, if one is set
                if (_adapter != null)
                    _adapter.Dispose();
            }
            finally
            {
                // Set the adapter reference to null to prevent further access to the disposed object
                _adapter = null;
                // Reset the recovery point factory reference to null, indicating no associated factory object
                _factory = null;
            }
        }

        /// <summary>
        /// This is a method the Client calls and is not for customer user. Logs a
        /// message's bookmark to the bookmark store and updates the message's bookmark
        /// sequence number.
        /// </summary>
        /// <param name="message">The message to log, containing the bookmark to be stored.</param>
        /// <returns>The index at which the bookmark is stored in the subscription.</returns>
        /// <exception cref="Exceptions.AMPSException">If an error occurs while logging to the bookmark store.</exception>
        public long log(Message message)
        {
            BookmarkField bookmark = (BookmarkField)message.getBookmarkRaw();
            // If the bookmark is a timestamp or equals Subscription.EPOCH_FIELD, do not log.
            if (bookmark.isTimestamp() ||
                bookmark.Equals(Subscription.EPOCH_FIELD))
            {
                return 0;
            }

            // Retrieve the subscription associated with the message.
            MemoryBookmarkStore.Subscription sub = (MemoryBookmarkStore.Subscription)message.getSubscription();

            // If no subscription is associated with the message, find it and set it.
            if (sub == null)
            {
                Field subId = message.getSubIdRaw();
                if (subId == null || subId.isNull())
                    subId = message.getSubIdsRaw();
                sub = find(subId);
                message.setSubscription(sub);
            }

            // Log the bookmark and set its sequence number in the message.
            long index = sub.log(bookmark);
            message.setBookmarkSeqNo(index);
            return index;
        }

        /// <summary>
        /// Discards a bookmark entry associated with a specific subscription and bookmark
        /// sequence number.
        /// </summary>
        /// <param name="subId">The identifier of the subscription.</param>
        /// <param name="bookmarkSeqNo">The sequence number of the bookmark entry to discard.</param>
        /// <exception cref="AMPSException">If an error occurs while discarding from the bookmark
        /// store.</exception>
        public void discard(Field subId, long bookmarkSeqNo)
        {
            find(subId).discard(bookmarkSeqNo);
        }

        /// <summary>
        /// Discards a bookmark entry based on the information provided in the message.
        /// </summary>
        /// <param name="message">The message containing information about the bookmark to
        /// discard.</param>
        /// <exception cref="AMPSException">If an error occurs while discarding from the bookmark
        /// store.</exception>
        public void discard(Message message)
        {
            BookmarkField bookmark = (BookmarkField)(message.getBookmarkRaw());
            if (bookmark.isTimestamp()
                || bookmark.Equals(Client.Bookmarks.NOW))
            {
                return;
            }
            long bookmarkSeq = message.getBookmarkSeqNo();
            MemoryBookmarkStore.Subscription sub = (MemoryBookmarkStore.Subscription)message.getSubscription();
            if (sub == null)
            {
                Field subId = message.getSubIdRaw();
                if (subId == null || subId.isNull())
                    subId = message.getSubIdsRaw();
                sub = find(subId);
                message.setSubscription(sub);
            }
            sub.discard(bookmarkSeq);
        }

        /// <summary>
        /// Retrieves the most recent bookmark associated with a subscription identified by the
        /// given subId.
        /// </summary>
        /// <param name="subId">The identifier of the subscription for which to retrieve the
        /// most recent bookmark.</param>
        /// <param name="useList">A flag indicating whether to retrieve the most recent bookmark
        /// as a list or just the single bookmark that was last discarded. A
        /// list is the only way to guarantee successful failover and no missed
        /// message.</param>
        /// <returns>The most recent bookmark for the specified subscription.</returns>
        /// <exception cref="AMPSException">If an error occurs while retrieving the most recent
        /// bookmark or if the store is not open.</exception>
        public Field getMostRecent(Field subId, bool useList = true)
        {
            return find(subId).getMostRecentList(useList);
        }

        /// <summary>
        /// This method is called internally by the client to determine whether a message's
        /// bookmark received from the amps server has already been discarded OR already been
        /// delivered to the subscriber during this run.
        /// </summary>
        /// <param name="message">The message containing the bookmark to check for discarding.</param>
        /// <returns>True if the bookmark has been discarded; false otherwise.</returns>
        /// <exception cref="AMPSException">If an error occurs while checking if the bookmark is
        /// discarded or if the store is not open.</exception>
        public bool isDiscarded(Message message)
        {
            BookmarkField bookmark = (BookmarkField)message.getBookmarkRaw();
            if (bookmark.isTimestamp())
                return false;
            if (bookmark.Equals(Subscription.EPOCH_FIELD))
                return true;
            MemoryBookmarkStore.Subscription sub = (MemoryBookmarkStore.Subscription)message.getSubscription();
            if (sub == null)
            {
                Field subId = message.getSubIdRaw();
                if (subId == null || subId.isNull())
                    subId = message.getSubIdsRaw();
                sub = find(subId);
                message.setSubscription(sub);
            }
            return sub.isDiscarded(bookmark);
        }

        /// <summary>
        /// Old style of setting a persisted bookmark no longer used.
        /// </summary>
        /// <param name="subId">The subId associated with the bookmark to be persisted.</param>
        /// <param name="bookmark">The bookmark to be persisted.</param>
        public void persisted(Field subId, long bookmark)
        {
            find(subId).setLastPersisted(bookmark);
        }

        /// <summary>
        /// This method is called by the client and it is used to process persisted
        /// acknowledgements that track a safe recovery point in the txlog. Sets a
        /// persisted bookmark for a subscription.
        /// </summary>
        /// <param name="subId">The identifier of the subscription for which to set the persisted
        /// bookmark.</param>
        /// <param name="bookmark">The persisted bookmark value to set.</param>
        /// <exception cref="AMPSException">If an error occurs while setting the persisted bookmark
        /// or if the store is not open.</exception>
        public void persisted(Field subId, BookmarkField bookmark)
        {
            if (bookmark.Equals(Client.Bookmarks.NOW)
                || bookmark.Equals(Subscription.EPOCH_FIELD))
                return;
            find(subId).setLastPersisted(bookmark);
        }

        // Internally used to get Subsription based on subId
        private Subscription find(Field subId)
        {
            lock (internalLock) {
                Subscription s = null;
                if (_subs.ContainsKey(subId) == false)
                {
                    s = _pool.get();
                    s.setSubscription(subId);
                    s.setParent(this);
                    s.setResizeHandler(_resizeHandler, this);
                    _subs.Add(subId.copy(), s);
                }
                else
                {
                    s = _subs[subId];
                }

                return s;
            }
        }

        /// <summary>
        /// Removes all entries in the bookmark store associated with a
        /// specific subscription  ID(`subId_`). This method is used to clear
        /// messages received and discarded for a particular subscription. This
        /// should not be called if the store is still attached to a connected
        /// Client where this subId is still active.
        /// </summary>
        /// <param name="subId_">The subscription ID for which to remove entries from the store.</param>
        public void purge(Field subId_)
        {
            lock (internalLock) {
                // If an adapter is available, purge the subscription.
                if (_adapter != null) _adapter.purge(subId_);
                // Remove the subscription from the internal dictionary.
                _subs.Remove(subId_);
            }
        }

        /// <summary>
        /// Purges the entire bookmark store, removing all entries. This method
        /// removes all information about any subscriptions. This should not
        /// be called if the store is still attached to a connected Client
        /// with any active subscriptions.
        /// </summary>
        public void purge()
        {
            lock (internalLock) {
                if (_adapter != null) _adapter.purge();
                _subs = new Dictionary<Field, Subscription>();
            }
        }

        /// <summary>
        /// Retrieves the oldest active bookmark sequence number that is not
        /// discarded for the given subscription.
        /// </summary>
        /// <param name="subId">The identifier of the subscription for which to retrieve the oldest
        /// bookmark sequence number.</param>
        /// <returns>The oldest bookmark sequence number for the specified subscription.</returns>
        /// <exception cref="AMPSException">If an error occurs while retrieving the oldest bookmark
        /// sequence number or if the store is not open.</exception>
        public long getOldestBookmarkSeq(Field subId)
        {
            return find(subId).getOldestBookmarkSeq();
        }

        /// <summary>
        /// Sets a handler for bookmark store resize events.
        /// </summary>
        /// <param name="handler">The BookmarkStoreResizeHandler to set.</param>
        public void setResizeHandler(BookmarkStoreResizeHandler handler)
        {
            lock (internalLock) {
                _resizeHandler = handler;
                foreach (KeyValuePair<Field, Subscription> sub in _subs)
                {
                    sub.Value.setResizeHandler(handler, this);
                }
            }
        }

        /// <summary>
        /// Used internally to change the version of the AMPS server that this
        /// bookmark store's client has connected to.
        /// </summary>
        /// <param name="version">An AMPS server version integer of the form 05030400 for
        /// version 5.3.4.0.</param>
        public void setServerVersion(int version)
        {
            _serverVersion = version;
        }

        /// <summary>
        /// Return the server version that was set by the Client.
        /// </summary>
        /// <returns>The server version, represented as an integer</returns>
        public int getServerVersion()
        {
            return _serverVersion;
        }
    }
}
