////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using AMPS.Client.Exceptions;


namespace AMPS.Client
{
    /// <summary>
    /// An in-memory publish store, used to provide republish capability when
    /// in case of failover to an AMPS secondary.
    /// <para>
    /// Publish stores hold messages until the AMPS server has acknowledged
    /// that they are persisted. When the client detects disconnection and
    /// reconnects, the client can republish any messages which have not been
    /// acknowledged by AMPS.
    /// <para>
    /// Use this PublishStore when you are concerned about
    /// server failover, only: MemoryPublishStore does not protect you in case
    /// of subscriber failure, because it has no on-disk backing store.
    /// </summary>
    public class MemoryPublishStore : BlockPublishStore
    {
        /// <summary>
        /// Constructs an in-memory public store with the specified number of blocks
        /// as the initial capacity. Creates a new MemoryPublishStore.
        /// </summary>
        /// <param name="initialCapacity"> The initial capacity (in 2k blocks) of the
        /// store. This size is also used when resizing the store: the store is resized
        /// by this value each time the store grows. A general guideline for initial
        /// capacity is to set the capacity to <b><i>messages_published_per_second</i> * int(<i>average_message_size</i> / <i>block_size</i>) + 1</b>.
        /// </param>
        /// <param name="errorOnPublishGap">Indicates if an excpetion should be thrown if discardUpTo is called with an idex below the last persisted.</param>
        public MemoryPublishStore(int initialCapacity, bool errorOnPublishGap = false)
            : base(new MemoryStoreBuffer(), initialCapacity, errorOnPublishGap)
        {
            try
            {
                if (_usedList == null)
                {
                    growFreeListIfEmpty();
                }
                recover();
            }
            catch (StoreException)
            {
                // pass
            }
        }

        /// <summary>
        /// Disposes the MemoryPublishStore instance.
        /// </summary>
        public override void Dispose()
        {
            // no-op
        }
    }
}
