////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;

namespace AMPS.Client
{

   /// <summary>
   /// MemoryStoreBuffer is a wrapper for a list that maintains position.
   /// In order to do this, MemoryStoreBuffer uses a ByteBuffer.
   /// </summary>
    public class MemoryStoreBuffer : BlockPublishStore.Buffer
    {
        protected byte[] _buffer = null;
        public long _position = 0;
        protected Store _store = null;

        internal MemoryStoreBuffer()
        {
            // MemoryPublishStore.growFreeListIfEmpty() will call setSize()
        }

        /// <summary>
        /// Constructor for MemoryStoreBuffer.
        /// </summary>
        /// <remarks>
        /// Initializes a new instance of the MemoryStoreBuffer class with the specified initial size.
        /// </remarks>
        /// <param name="initialSize">The initial size of the buffer, specified in terms of the number of blocks.</param>
        public MemoryStoreBuffer(int initialSize)
        {
            setSize(initialSize * BlockPublishStore.Block.SIZE);
        }

        /// <summary>
        /// Gets the size of the buffer.
        /// </summary>
        /// <returns> The size of the buffer </returns>
        public long getSize()
        {
            return _buffer == null ? 0 : _buffer.Length;
        }

        /// <summary>
        /// Sets the size of the buffer, growing if necessary.
        /// </summary>
        /// <param name="newSize_"> The new size to be assigned to the buffer. </param>
        public void setSize(long newSize_)
        {
            if (_buffer == null)
            {
                _buffer = new byte[newSize_];
                return;
            }
            if (newSize_ > _buffer.Length)
            {
                byte[] newBuffer = new byte[newSize_];
                Array.Copy(_buffer, newBuffer, _buffer.Length);
                _buffer = newBuffer;
            }
        }

        /// <summary>
        /// Gets the current position of the buffer.
        /// </summary>
        /// <returns> The current position of the buffer. </returns>
        public long getPosition()
        {
            return _buffer == null ? 0 : _position;
        }

        /// <summary>
        /// Sets the new current position in the buffer.
        /// </summary>
        /// <param name="position_"> The new current position to be set in the buffer. </param>
        public void setPosition(long position_)
        {
            _position = position_;
        }

        /// <summary>
        /// Writes the byte to the buffer and advances the position.
        /// </summary>
        /// <param name="b"> The byte to be written in the buffer </param>
        public void putByte(byte b)
        {
            _buffer[_position++] = b;
        }

        /// <summary>
        /// Gets the next byte from the buffer and advances the position.
        /// </summary>
        /// <returns> The byte from the buffer. </returns>
        public byte getByte()
        {
            return _buffer[_position++];
        }

        /// <summary>
        /// Writes the integer to the buffer. Advances current position.
        /// </summary>
        /// <param name="i"> The integer to be placed in the buffer. </param>
        public void putInt(int i)
        {
            Utilities.writeInt32(i, _buffer, _position);
            _position += 4;
        }

        /// <summary>
        /// Writes the integer to the buffer at position, doesn't change current position.
        /// </summary>
        /// <param name="position"> The location where the integer is placed in the buffer. </param>
        /// <param name="intValue"> The integer to bee placed in the buffer. </param>
        public void putInt(long position, int intValue)
        {
            Utilities.writeInt32(intValue, _buffer, position);
        }

        /// <summary>
        /// Gets an integer from the buffer and advances the position.
        /// </summary>
        /// <returns> The integer from the buffer. </returns>
        public int getInt()
        {
            _position += 4;
            return Utilities.readInt32(_buffer, _position - 4);
        }

        /// <summary>
        /// Gets the integer from the buffer at position and doesn't change current position.
        /// </summary>
        /// <param name="position"> The position in the buffer where the integer begins. </param>
        /// <returns> The integer from the buffer. </returns>
        public int getInt(long position)
        {
            return Utilities.readInt32(_buffer, position);
        }

        /// <summary>
        /// Writes thee long to the buffer and advances the position.
        /// </summary>
        /// <param name="l"> The long to write in the buffer. </param>
        public void putLong(long l)
        {
            Utilities.writeLong64(l, _buffer, _position);
            _position += 8;
        }

        /// <summary>
        /// Writes the long to the buffer at position and doesn't change current position.
        /// </summary>
        /// <param name="position"> The position in the buffer where the long should be placed. </param>
        /// <param name="longValue"> The long to write in the buffer. </param>
        public void putLong(long position, long longValue)
        {
            Utilities.writeLong64(longValue, _buffer, position);
        }

        /// <summary>
        /// Gets the long from the buffer and advances the position.
        /// </summary>
        /// <returns> The long from the buffer. </returns>
        public long getLong()
        {
            _position += 8;
            return Utilities.readLong64(_buffer, _position - 8);
        }

        /// <summary>
        /// Writes an unsigned long to the buffer and advances the position.
        /// </summary>
        /// <param name="l"> The unsigned long to write in the buffer. </param>
        public void putUlong(ulong l)
        {
            Utilities.writeLong64((long)l, _buffer, _position);
            _position += 8;
        }

        /// <summary>
        /// Writes an unsigned long to the buffer at position and doesn't change current position.
        /// </summary>
        /// <param name="position"> The position in the buffer where the unsigned long
        /// should be placed. </param>
        /// <param name="ulongValue"> The unsigned long to write in the buffer. </param>
        public void putUlong(long position, ulong ulongValue)
        {
            Utilities.writeLong64((long)ulongValue, _buffer, position);
        }

        /// <summary>
        /// Gets the unsigned long from the buffer and advances the position.
        /// </summary>
        /// <returns> The unsigned long from the buffer. </returns>
        public ulong getUlong()
        {
            _position += 8;
            return (ulong)Utilities.readLong64(_buffer, _position - 8);
        }

        /// <summary>
        /// Writes a series of bytes at the current position, taken from the specified
        /// BlockPublishStore.ByteSequence and advances the position.
        /// </summary>
        /// <param name="bytes"> The series of bytes to write at the current position. </param>
        public void putBytes(BlockPublishStore.ByteSequence bytes)
        {
            Array.Copy(bytes.array, bytes.offset, _buffer, _position, bytes.len);
            _position += bytes.len;
        }

        /// <summary>
        /// Gets the series of bytes at the current position by placing them into the provided
        /// BlockPublishStore.ByteSequence object and advances the position.
        /// </summary>
        /// <param name="outBytes"> The ByteSequence object to be copied. </param>
        public void getBytes(BlockPublishStore.ByteSequence outBytes)
        {
            outBytes.array = _buffer;
            outBytes.offset = _position;
            _position += outBytes.len;
        }

        /// <summary>
        /// Writes a series of bytes at the current position, taken from a byte array and advances the position.
        /// </summary>
        /// <param name="buffer"> The buffer containing a series of bytes. </param>
        /// <param name="position"> The position in the buffer where the series of bytes are located. </param>
        /// <param name="length"> The length of the byte buffer. </param>
        public void putBytes(byte[] buffer, int position, int length)
        {
            Array.Copy(buffer, position, _buffer, _position, length);
            _position += length;
        }

        /// <summary>
        /// Gets the series of bytes at the current position, this is done by copying them
        /// into a Field object and advances the position.
        /// </summary>
        /// <param name="outField"> The Field object used to copy the byte sequence. </param>
        /// <param name="length"> The length of the Field object. </param>
        public void getBytes(Fields.Field outField, int length)
        {
            outField.length = length;
            if (outField.length > 0)
            {
                outField.buffer = _buffer;
                outField.position = (int)_position;
                _position += outField.length;
            }
        }

        /// <summary>
        /// Sets length bytes starting at offset to 0.
        /// </summary>
        /// <param name="offset"> The offset at which to start </param>
        /// <param name="length"> The length of bytes to set to 0. </param>
        public void zero(long offset, int length)
        {
            for(long i = 0; i < length; ++i)
            {
                _buffer[offset + i] = 0;
            }
        }
    }
}
