////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System.Collections.Generic;
using AMPS.Client.Exceptions;
using AMPS.Client.Fields;
using System;
using System.Threading;

namespace AMPS.Client
{

    /// <summary>
    /// An in-memory implementation of SubscriptionManager. Used by HAClient.
    /// </summary>
    public class MemorySubscriptionManager : SubscriptionManager
    {
        private static long DefaultResubscriptionTimeout = 0;
        private long _resubscriptionTimeout = 0;

        class SubscriptionInfo
        {
            internal Message _message = null;
            internal MessageHandler _handler;
            public Field _subId;
            bool _hasBookmark;
            bool _paused;
            internal Message.AckTypes _ackType;
            Field _recent;
            MemorySubscriptionManager _parent;

            // <summary>
            // Constructs a SubscriptionInfo for the given handler and subscribe message.
            // </summary>
            // <param name="messageHandler"> The message handler for this subscription. </param>
            // <param name="message"> The subscribe message for this handler. </param>
            // <param name="parent">The parent MemorySubscriptionManager instance.</param>
            public SubscriptionInfo(MessageHandler messageHandler,
                                    Message message,
                                    MemorySubscriptionManager parent)
            {
                _message = message;
                _handler = messageHandler;
                _subId = _message.getSubIdRaw().copy();
                _hasBookmark = !_message.isBookmarkNull();
                _paused = false;
                _ackType = message.getAckTypeOutgoing();
                byte[] recentBuf = new byte[512];
                _recent = new Field(recentBuf, 0, 0);
                // Check for pause
                if (!_message.isOptionsNull())
                {
                    string options = _message.getOptions();
                    _paused = options.Contains("pause");

                }
                _parent = parent;
            }

            // <summary>
            // Places this subscription on the given Client.
            // </summary>
            // <param name="client"> The client to place the subscription on. </param>
            public void resubscribe(Client client)
            {
                // Don't set replace
                if (!_message.isOptionsNull())
                {
                    string options = _message.getOptions();
                    options = options.Replace("replace,?", "");
                    if (options != _message.getOptions())
                    {
                        if (options.Length > 1) _message.setOptions(options);
                        else _message.setOptions(null);
                    }
                }
                if (_hasBookmark)
                {
                    // Always resubscribe to the most recent.
                    if (_paused && _recent.length > 0)
                    {
                        _message.setBookmark(_recent.buffer, _recent.position,
                                             _recent.length);
                    }
                    else
                    {
                        Field recent = client.getBookmarkStore().getMostRecent(_subId);
                        _message.setBookmark(recent.buffer, recent.position, recent.length);
                    }
                }
                _message.setAckType(_ackType);
                client.send(_handler, _message, _parent._resubscriptionTimeout);
                _recent.length = 0;
            }

            // <summary>
            // Remove the given subscription ID from the list of sub ids on self.
            // </summary>
            // <param name="subId_"> The subscription ID to remove. </param>
            // <returns>True if the subscription ID is removed, false otherwise.</returns>
            public bool removeSubId(Field subId_)
            {
                int subIdStart = subId_.position;
                int subIdLen = subId_.length;
                while (subId_.buffer[subIdStart + subIdLen - 1] == (byte)',')
                    --subIdLen;
                while (subId_.buffer[subIdStart] == (byte)',')
                {
                    ++subIdStart; --subIdLen;
                }
                if (subIdLen > _subId.length) return _subId.isNull();
                bool match = true;
                int matchStart = 0;
                int matchCount = 0;
                for (int i = 0; i < _subId.length; ++i)
                {
                    if (_subId.buffer[_subId.position + i] == (byte)',')
                    {
                        if (matchCount == subIdLen)
                        {
                            break;
                        }
                        matchStart = i + 1;
                        matchCount = 0;
                        match = true;
                        continue;
                    }
                    else if (match)
                    {
                        if (matchCount < subIdLen &&
                            _subId.buffer[_subId.position + i] ==
                             subId_.buffer[subIdStart + matchCount])
                        {
                            ++matchCount;
                        }
                        else
                        {
                            matchCount = 0;
                            match = false;
                        }
                    }
                }
                if (match && matchCount == subIdLen)
                {
                    int len = _subId.length - matchCount;
                    if (len > 1) // More than just , left
                    {
                        byte[] buffer = new byte[len];
                        int offset = 0;
                        int newLen = 0;
                        for (int i = 0; i + offset < _subId.length; ++i)
                        {
                            if (i == matchStart)
                            {
                                offset = matchCount;
                                if (i + offset >= _subId.length) break;
                                if (_subId.buffer[_subId.position + offset + i] == (byte)',')
                                {
                                    if (++offset + i >= _subId.length) break;
                                }
                            }
                            buffer[i] = _subId.buffer[_subId.position + offset + i];
                            ++newLen;
                        }
                        if (newLen > 0)
                        {
                            _subId = new Field(buffer, 0, newLen);
                            _message.getSubIdRaw().set(buffer, 0, newLen);
                            return false;
                        }
                    }
                    _subId.reset();
                    _message.getSubIdRaw().reset();
                    return true;
                }
                return _subId.isNull();
            }

            // <summary>
            // Marks a subscription as paused.
            // </summary>
            public void pause()
            {
                if (_paused) return;
                string options = Message.Options.Pause + _message.getOptions();
                _message.setOptions(options);
                _paused = true;
            }

            // <summary>
            // Create the most recent bookmark string for this subscription.
            // </summary>
            // <param name="client"> The Client to check. </param>
            // <returns> The most recent bookmark string, used when resubscribing. </returns>
            public Field getMostRecent(Client client)
            {
                if (!_recent.isNull()) return _recent;
                Field subId = new Field();
                int start = 0;
                while (_subId.buffer[_subId.position + start] == (byte)',')
                    ++start;
                int end = start + 1;
                while (end < _subId.length)
                {
                    if (_subId.buffer[_subId.position + end] == (byte)',')
                    {
                        subId.set(_subId.buffer, _subId.position + start, end - start);
                        Field mostRecent = client.getBookmarkStore().getMostRecent(subId);
                        if (_recent.length > 0)
                        {
                            if (_recent.length == _recent.buffer.Length)
                            {
                                byte[] newRecent = new byte[_recent.buffer.Length + 512];
                                Array.Copy(_recent.buffer, 0,
                                           newRecent, 0, _recent.length);
                                _recent.set(newRecent, 0, _recent.length);
                            }
                            _recent.buffer[_recent.length++] = (byte)',';
                        }
                        if (_recent.length + mostRecent.length > _recent.buffer.Length)
                        {
                            int newSize = _recent.buffer.Length + 512;
                            while (_recent.length + mostRecent.length > newSize)
                                newSize += 512;
                            byte[] newRecent = new byte[newSize];
                            Array.Copy(_recent.buffer, 0,
                                       newRecent, 0, _recent.length);
                            _recent.set(newRecent, 0, _recent.length);
                        }
                        Array.Copy(mostRecent.buffer, mostRecent.position,
                                                             _recent.buffer, _recent.length,
                                                             mostRecent.length);
                        _recent.length += mostRecent.length;
                        while (end < _subId.length &&
                               _subId.buffer[_subId.position + end] == (byte)',')
                            ++end;
                        start = end;
                    }
                    ++end;
                }
                return _recent;
            }
        }

        private System.Object _lock = new System.Object();
        private Dictionary<Field, SubscriptionInfo> _active = new Dictionary<Field, SubscriptionInfo>();
        private Dictionary<Field, SubscriptionInfo> _resumed = new Dictionary<Field, SubscriptionInfo>();
        private List<SubscriptionInfo> _resumeList = new List<SubscriptionInfo>();
        private FailedResubscribeHandler _failedResubscribeHandler = null;
        volatile bool _resubscribing = false;

        /// <summary>
        /// Initializes a new instance of the MemorySubscriptionManager class.
        /// </summary>
        public MemorySubscriptionManager()
        {
            _resubscriptionTimeout = DefaultResubscriptionTimeout;
        }

        /// <summary>
        /// Adds a subscription to be resubscribed with the given handler based
        /// on the subscribe Message being sent to the server.
        /// </summary>
        /// <param name="messageHandler">The message handler to subscribe.</param>
        /// <param name="message">The message to subscribe to.</param>
        public void subscribe(MessageHandler messageHandler, Message message)
        {
            lock (_lock)
            {
                while (_resubscribing)
                {
                    Monitor.Wait(_lock, 100);
                }
                if (!message.isSubIdNull())
                {
                    String subId = message.getSubId();
                    if (!message.isOptionsNull())
                    {
                        String options = message.getOptions();
                        if (options.Contains("resume"))
                        {
                            SubscriptionInfo subInfo1 = new SubscriptionInfo(
                                                                messageHandler,
                                                                message, this);
                            bool saved = false;
                            foreach (String sid in subId.Split(','))
                            {
                                Field sidField1 = new Field(sid);
                                if (!_resumed.ContainsKey(sidField1))
                                {
                                    _resumed[sidField1] = subInfo1;
                                    saved = true;
                                }
                            }
                            if (saved) _resumeList.Add(subInfo1);
                            return;
                        }
                        else if (options.Contains("pause"))
                        {
                            foreach (String sid in subId.Split(','))
                            {
                                if (string.IsNullOrEmpty(sid)) continue;
                                MessageHandler handler1 = messageHandler;
                                Field sidField2 = new Field(sid);
                                SubscriptionInfo existingSubInfo1;
                                if (_resumed.TryGetValue(sidField2, out existingSubInfo1))
                                {
                                    if (existingSubInfo1.removeSubId(sidField2))
                                        _resumeList.Remove(existingSubInfo1);
                                    _resumed.Remove(sidField2);
                                }
                                if (_active.TryGetValue(sidField2, out existingSubInfo1))
                                {
                                    if (!options.Contains("replace"))
                                    {
                                        existingSubInfo1.pause();
                                        continue;
                                    }
                                    else
                                    {
                                        handler1 = existingSubInfo1._handler;
                                    }
                                }
                                Message m = message.copy();
                                m.setSubId(sid);
                                SubscriptionInfo subInfo2 =
                                                    new SubscriptionInfo(handler1,
                                                                         m, this);
                                _active[subInfo2._subId] = subInfo2;
                            }
                            return;
                        }
                    }
                    SubscriptionInfo subInfo = new SubscriptionInfo(messageHandler,
                                                                    message, this);
                    _active[subInfo._subId] = subInfo;
                }
            }
        }

        /// <summary>
        /// Removes the subscription with the specified subscription ID so it
        /// not get resubscribed in the future.
        /// </summary>
        /// <param name="subId">The subscription ID to unsubscribe from.</param>
        public void unsubscribe(CommandId subId)
        {
            lock (_lock)
            {
                while (_resubscribing)
                {
                    Monitor.Wait(_lock, 100);
                }
                StringField f = new StringField();
                f.setValue(subId);
                _active.Remove(f);
                SubscriptionInfo existingSubInfo;
                if (_resumed.TryGetValue(f, out existingSubInfo))
                {
                    if (existingSubInfo.removeSubId(f))
                        _resumeList.Remove(existingSubInfo);
                    _resumed.Remove(f);
                }
            }
        }

        /// <summary>
        /// Checks if a subscription with the specified subscription ID exists.
        /// </summary>
        /// <param name="subId">The subscription ID to check.</param>
        /// <returns>True if the subscription exists, otherwise false.</returns>
        public bool isExistingSub(Field subId)
        {
            if (subId == null || subId.isNull()) return false;

            bool existing = false;
            lock (_lock)
            {
                if (_active.ContainsKey(subId) || _resumed.ContainsKey(subId))
                {
                    existing = true;
                }
            }

            return existing;
        }

        /// <summary>
        /// Clears all subscriptions.
        /// </summary>
        public void clear()
        {
            lock(_lock)
            {
                _active.Clear();
                _resumed.Clear();
                _resumeList.Clear();
            }
        }

        /// <summary>
        /// Called by the Client after reconneting to resubscribe all
        /// subscriptions on the given client.
        /// </summary>
        /// <param name="client">The client on which to resubscribe.</param>
        public void resubscribe(Client client)
        {
            List<SubscriptionInfo> subs = null;
            try
            {
                lock(_lock)
                {
                    subs = new List<SubscriptionInfo>(_active.Count);
                    _resubscribing = true;
                    foreach (SubscriptionInfo si in _active.Values)
                    {
                        subs.Add(si);
                    }
                    foreach (SubscriptionInfo si in _resumeList)
                    {
                        subs.Add(si);
                    }
                }
                bool throwExcept = false;
                AMPSException ex = null;
                List<SubscriptionInfo> removals = null;
                foreach (SubscriptionInfo si in subs)
                {
                    try {
                        si.resubscribe(client);
                    }
                    catch (AMPSException ex_) {
                        if (_failedResubscribeHandler != null) {
                            if (_failedResubscribeHandler.failure(si._message,
                                                                  si._handler,
                                                                  (int)si._ackType,
                                                                  ex_))
                            {
                                if (removals == null) removals = new List<SubscriptionInfo>(_active.Count);
                                removals.Add(si);
                            }
                            else
                            {
                                throwExcept = true;
                                ex = ex_;
                            }
                        }
                        else {
                            throw;
                        }
                    }
                }
                if (removals != null) {
                    foreach (SubscriptionInfo removal in removals) {
                        _active.Remove(removal._subId);
                    }
                }
                if (throwExcept) {
                  throw ex;
                }
            }
            finally
            {
                lock(_lock)
                {
                    Monitor.PulseAll(_lock);
                    _resubscribing = false;
                }
            }
        }

        /// <summary>
        /// Sets the default resubscription timeout for all instances of this class in milliseconds.
        /// </summary>
        /// <param name="timeout"> The default timeout for resubscription calls. </param>
        public static void setDefaultResubscriptionTimeout(long timeout)
        {
            DefaultResubscriptionTimeout = timeout;
        }

        /// <summary>
        /// Gets the default resubscription timeout for all instances of this class in milliseconds.
        /// </summary>
        /// <returns> The default timeout for resubscription calls. </returns>
        public static long getDefaultResubscriptionTimeout()
        {
            return DefaultResubscriptionTimeout;
        }

        /// <summary>
        /// Sets the resubscription timeout in milliseconds.
        /// </summary>
        /// <param name="timeout"> The timeout for resubscription calls. </param>
        public void setResubscriptionTimeout(long timeout)
        {
            _resubscriptionTimeout = timeout;
        }

        /// <summary>
        /// Gets the resubscription timeout in milliseconds.
        /// </summary>
        /// <returns> The timeout for resubscription calls. </returns>
        public long getResubscriptionTimeout()
        {
            return _resubscriptionTimeout;
        }

        /// <summary>
        /// Sets the FailedResubscribeHandler.
        /// </summary>
        /// <param name="failureHandler"> A FailedResubscribeHandler to call if a resubscribe fails. </param>
        public void setFailedResubscribeHandler(FailedResubscribeHandler failureHandler)
        {
            _failedResubscribeHandler = failureHandler;
        }

        /// <summary>
        /// Gets the FailedResubscribeHandler.
        /// </summary>
        /// <returns> The FailedResubscribeHandler or null if not set. </returns>
        public FailedResubscribeHandler getFailedResubscribeHandler()
        {
            return _failedResubscribeHandler;
        }
    }
}
