////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Text;

using AMPS.Client.Fields;
using AMPS.Client.Exceptions;


namespace AMPS.Client
{

    /// <summary>
    /// All communication with AMPS occurs through passing Messages.
    /// The AMPS clients provide named convenience methods for core AMPS functionality.
    /// These named methods work by creating and sending Messages to AMPS.
    /// </summary>
    public abstract partial class Message
    {

        protected const int MINIMUM_SERVER_VERSION = Client.DEFAULT_SERVER_VERSION;

        /// <summary>
        /// Tells AMPS how to interpret a message. Without a command, AMPS will reject the message.
        /// The relevancy or meaning of other Message fields like options will depend upon the
        /// value of its command field.
        /// </summary>
        public enum Commands : int
        {
            Unknown = 0,
            Publish = 1,
            Subscribe = 2,
            Unsubscribe = 4,
            SOW = 8,
            Heartbeat = 16,
            SOWDelete = 32,
            DeltaPublish = 64,
            Logon = 128,
            SOWAndSubscribe = 256,
            DeltaSubscribe = 512,
            SOWAndDeltaSubscribe = 1024,
            StartTimer = 2048,
            StopTimer = 4096,
            GroupBegin = 8192,
            GroupEnd = 16384,
            OOF = 32768,
            Ack = 65536,
            Flush = 131072,
            NoDataCommands = Publish|Unsubscribe|Heartbeat|SOWDelete|DeltaPublish
                             |Logon|StartTimer|StopTimer|Flush
        }

        // These AckTypes are often combined for a single command
        // so we favor the int 'mask' versus a heavy EnumSet.
        /// <summary>
        /// Enum of maskable ints that represent the available types of Message acknowledgement.
        /// <see cref="AMPS.Client.Command#getAckType()"/>
        /// <see cref="AMPS.Client.Command#setAckType()"/>
        /// <see cref="AMPS.Client.Command#addAckType()"/>
        /// <see cref="AMPS.Client.Fields.AckTypeField"/>
        /// </summary>
        public enum AckTypes : int
        {
            None = 0,
            Received = 1,
            Parsed = 2,
            Processed = 4,
            Persisted = 8,
            Completed = 16,
            Stats = 32,
        }

        /// <summary>
        /// Enum of ints that represent the possible values for Message status.
        /// </summary>
        public enum State : int { None, Success, Failure, Retry };

        /// <summary>
        /// Enum of ints that represent the possible values for a Message reason.
        /// <see cref="AMPS.Client.Message#setReason(int)"/>
        /// <see cref="AMPS.Client.Message#getReason()"/>
        /// </summary>
        public enum Reasons : int
        {
            None = 0,
            Duplicate = 1,
            BadFilter = 2,
            BadRegexTopic = 3,
            SubscriptionAlreadyExists = 4,
            Deleted = 5,
            Expired = 6,
            Match = 7,
            InvalidTopic = 8,
            NameInUse = 9,
            AuthFailure = 10,
            NotEntitled = 11,
            AuthDisabled = 12,
            InvalidBookmark = 13,
            InvalidOrderBy = 14,
            SubidInUse = 15,
            NoTopic = 16,
            LogonFailed = 17,
            LogonRequired = 18,
            InvalidTopicOrFilter = 19,
            InvalidSubId = 20,
            NoTopicOrFilter = 21,
            NoClientName = 22,
            BadSowKey = 23,
            RegexTopicNotSupported = 24,
            SowStoreFailed = 25,
            ParseError = 26,
            NotSupported = 27,
            TxStoreFailure = 28,
            DuplicateLogon = 29,
            TxnReplayFailed = 30,
            SowCanceled = 31,
            InvalidOptions = 32,
            InvalidMessageType = 33,
            OrderByTooLarge = 34,
            RejectedByTransportFilter = 35,
            SowDeleteInvalid = 36,
            Other = 38,
        }

        /// <summary>
        /// Represents the possible values for message options in an AMPS command. Options in a message
        /// are combined in to a comma-delimited list. Each of these option values is comma-terminated,
        /// so they can be easily combined by simple string concatenation.
        ///
        /// For options that take a list of parameters, this class provides
        /// convenience constructors to format the list.
        /// <see cref="AMPS.Client.Command#getOptions()"/>
        /// <see cref="AMPS.Client.Command#setOptions(string)"/>
        /// <see cref="AMPS.Client.Message#getOptions()"/>
        /// <see cref="AMPS.Client.Message#setOptions(string)"/>
        /// </summary>
        public class Options
        {
            public const string None = null;

            /// <summary>
            /// Constant string for the <code>Live</code> option.
            /// Requests that a bookmark subscription should receive
            /// messages <em>before</em> they are written to the
            /// transaction log once replay is complete. This could
            /// potentially slightly reduce latency in some configurations,
            /// although subscriptions using this option increase the risk
            /// of inconsistent data in failover situations.
            /// </summary>
            public const string Live = "live,";

            /// <summary>
            /// Constant string for the <code>oof</code> (out of focus)
            /// option. When used with a subscription, the <code>oof</code>
            /// option requests notification messages when a message that
            /// previously matched the subscription no longer matches the
            /// subscription.
            /// </summary>
            public const string OOF = "oof,";

            /// <summary>
            /// Constant string for the <code>replace</code> option.
            /// When this option is provided on a command, AMPS updates the
            /// existing subscription specified in the <pre>SubId</pre> for
            /// the command rather than creating a new subscription.
            /// See the <em>AMPS User Guide</em> for details.
            /// </summary>
            public const string Replace = "replace,";

            /// <summary>
            /// Constant string for the <code>no_empties</code> option. When
            /// used with a delta subscription, this option specifies that
            /// AMPS will not create a message to this subscription when a
            /// publish happens that does not change the value of any field
            /// in the message.
            /// </summary>
            public const string NoEmpties = "no_empties,";

            /// <summary>
            /// Constant string for the <code>send_keys</code> option.
            /// This option is the current server default, and does not
            /// need to be specified. When set, AMPS provides the key
            /// fields for a message along with changed fields in
            /// messages to a delta subscription.
            /// </summary>
            public const string SendKeys = "send_keys,";

            /// <summary>
            /// Constant string for the <code>timestamp</code> option.
            /// When specified on a command, AMPS will include the
            /// timestamp header on data messages sent in response to
            /// that command.
            /// </summary>
            public const string Timestamp = "timestamp,";

            /// <summary>
            /// Constant string for the <code>no_sow_key</code> option.
            /// When specified on a command, AMPS will not provide the
            /// SOW key field on messages provided in response to the
            /// command.
            /// </summary>
            public const string NoSowKey = "no_sowkey,";

            /// <summary>
            /// Constant string for the <codecancel</code> option, used
            /// when acknowledging (<code>sow_delete</code> ) a leased
            /// message to return the message to the queue rather than
            /// acknowledgin the messsage as processed.
            /// </summary
            public const string Cancel = "cancel,";

            /// <summary>
            /// Constant string for the <code>expire</code> option, used
            /// when acknowledging (<code>sow_delete</code>) a leased message
            /// to immediately expire the message from the queue rather than
            /// acknowledging the message as processed.
            /// </summary>
            public const string Expire = "expire,";

            /// <summary>
            /// Constant string for the <code>resume</code> option.
            /// This option is to start a set of bookmark subscriptions
            /// that have been entered in a paused state.
            /// See the <em>AMPS User Guide</em> for details.
            /// </summary>
            public const string Resume = "resume,";

            /// <summary>Constant string for the <code>pause</code> option.
            /// This option is used to group bookmark subscriptions so
            /// that a set of independent subscriptions can be started
            /// simultaneously and proceed through replay together,
            /// typically used for testing or auditing purposes.
            /// See the <em>AMPS User Guide</em> for details.
            /// </summary>
            public const string Pause = "pause,";

            /// <summary>
            /// Constant string for the <code>fully_durable</code> option.
            /// This option specifies that a message will not be provided
            /// to a bookmark replay until it has been acknowledged as
            /// persisted by all synchronous replication destinations. The
            /// option is used to provide increased repeatability of bookmark
            /// subscriptions, although it may increase the latency for
            /// current publishes to be delivered to the subscriber.
            /// </summary>
            public const string FullyDurable = "fully_durable,";

            /// <summary>
            /// Formats the <code>conflation</code> option for a command.
            /// This option can be used with subscriptions to topics in the
            /// State-of-the-World to request that AMPS conflate publishes
            /// to the same record. This can reduce bandwidth and processing
            /// requirements for the the subscriber. If a large number of
            /// subscribers will use the same <code>conflation</code>
            /// setting for the same topic, consider configuring a
            /// <code>ConflatedTopic</code> on the server. See the
            /// <em>AMPS User Guide</em> for details.
            /// </summary>
            /// <param name="conflation"> The conflation interval. </param>
            /// <returns> The formatted value. </returns>
            public static string Conflation(String conflation)
            {
               return String.Format("conflation={0},", conflation);
            }

            /// <summary>
            /// > Formats the <code>conflation_key</code> option for a command
            /// that uses the <code>conflation</code> option. This option can be
            /// used with subscriptions to topics in the State-of-the-World to
            /// request that AMPS conflate po messages with a common value for a
            /// field or set of fields that is <em>not</em> the key for the topic.
            /// This option must be used with the <code>conflation</code> option.
            /// </summary>
            /// <param name= "conflationKey"> The field or fields to use as a conflation key. </param>
            /// <returns> The formatted value. </returns>
            public static string ConflationKey(String conflationKey)
            {
                return String.Format("conflation_key={0},", conflationKey);
            }

            /// <summary>
            /// Formats the <code>max_backlog</code> option for a command, used when
            /// subscribing to a queue. When consuming from a queue, this option sets
            /// the maximum number of messages the client is willing to have outstanding
            /// at a given time. (The actual number of messages provided is determined
            /// by both this parameter and the server configuration.) See the <em>AMPS
            /// User Guide</em> and the <em>AMPS Command Reference</em> for details.
            /// </summary>
            /// <param name= "maxBacklog"> The maximum number of messages to provide
            /// to this subscription at a given time. </param>
            /// <returns> The formatted value. </returns>
            public static string MaxBacklog(int maxBacklog)
            {
                return String.Format("max_backlog={0},",maxBacklog);
            }

            /// <summary>
            /// Formats the <code>top_n</code> option for a command. When this option is
            /// specified, AMPS will return thespecified number of messages from the results
            /// See the <em>AMPS Command Reference</em> for details.
            /// </summary>
            /// <param name="topN"> The number of messages to return from the result set. </param>
            /// <returns> The formatted value. </returns>
            public static string TopN(int topN)
            {
                return String.Format("top_n={0},",topN);
            }

            /// <summary>
            /// Formats the <code>rate</code> option for a command, used to slow down the rate
            /// of a bookmark replay. When used with a bookmark replay, this option directs
            /// AMPS to slow down the replay to approximately the rate specified. See the
            /// <em>AMPS Command Reference</em> for details.
            /// </summary>
            /// <param name="rate"> The requested rate limit for the bookmark replay. </param>
            /// <returns> The formatted value. </returns>
            public static string Rate(String rate)
            {
                return String.Format("rate={0},",rate);
            }

            /// <summary>
            /// Formats the <code>rate_max_gap</code> option for a
            /// command, which can be used when a bookmark replay specifies
            /// a relative <code>rate</code>. When used with a bookmark replay that
            /// specifies a relative rate (for example, that the replay should be
            /// slowed down to no more than <code>2X</code> the original publish
            /// rate), this option sets the maximum amount of time for the replay
            /// to go without producing a message. See the <em>AMPS User Guide</em>
            /// for details.
            /// </summary>
            /// <param name="rateMaxGap"> The maximum amount of time without producing
            /// a message. </param>
            /// <returns> The formatted value. </returns>
            public static string RateMaxGap(String rateMaxGap)
            {
                return String.Format("rate_max_gap={0},",rateMaxGap);
            }

            /// <summary>
            /// Formats the <code>select</code> option for a subscription or SOW query, which
            /// specifies the fields produced in the delivered messages. This overload of the
            /// method expects a string that contains a comma-delimited list of field directives.
            /// See the <em>AMPS User Guide</em> for details.
            /// </summary>
            /// <param name="selectList"> The list of fields to produce. </param>
            /// <returns> The formatted value. </returns>
            public static string Select(String selectList)
            {
                return String.Format("select=[{0}],",selectList);
            }

            /// <summary>
            /// Formats the <code>select</code>
            /// <br>
            /// <br>
            /// option for a subscription or SOW query, which
            /// specifies the fields produced in the delivered messages. This overload of the
            /// method expects a <see cref="System.Collections.Generic.IEnumerable{T}"/> that
            /// contains individual field directives. The method formats the field directives
            /// into a comma-delimited list.
            /// See the <em>AMPS User Guide</em> for details.
            /// </summary>
            /// <param name="selectList"> A list of fields to produce. </param>
            /// <returns> The formatted value. </returns>
            public static string Select(IEnumerable<String> selectList)
            {
                StringBuilder builder = new StringBuilder();
                builder.Append("select=[");
                foreach (String s in selectList)
                {
                    builder.Append(s).Append(',');
                }
                builder.Insert(builder.Length-1, ']');
                return builder.ToString();
            }

            /// <summary>
            /// Formats the <code>skip_n</code>
            /// <br>
            /// <br>
            /// option for a <code>sow</code> or
            /// <code>sow_and_subscribe</code> command. When this option is provided,
            /// AMPS will skip the specified number of messages before returning the
            /// <code>top_n</code> records specified. See the <em>AMPS Command Reference</em>
            /// and the <em>Developer Guide</em> for details.
            /// </summary>
            /// <param name="skipN"> The number of results to skip before beginning to return
            /// results </param>
            /// <returns> The formatted value. </returns>
            public static string SkipN(int skipN)
            {
                return String.Format("skip_n={0},",skipN);
            }

            /// <summary>
            /// Formats the <code>projection</code> option for an aggregated subscription, which
            /// specifies the fields produced by the aggregation. When this option is provided, a
            /// <code>grouping</code> option must also be provided. This overload of the method
            /// expects a string that contains a comma-delimited list of projections.
            /// See the <em>AMPS User Guide</em> for details.
            /// </summary>
            /// <param name="projection"> The fields to produce. </param>
            /// <returns> The formatted value. </returns>
            public static string Projection(String projection)
            {
                return String.Format("projection=[{0}],",projection);
            }

            /// <summary>
            /// Formats the <code>projection</code> option for an aggregated subscription, which
            /// specifies the fields produced by the aggregation. When this option is provided,
            /// a <code>grouping</code> option must also be provided. This overload of the method
            /// expects a <see cref="System.Collections.Generic.IEnumerable{T}"/> that contains individual field projections.
            /// The method formats the field projections into a comma-delimited list.
            /// See the <em>AMPS User Guide</em> for details.
            /// </summary>
            /// <param name="projections"> A list of fields to produce </param>
            /// <returns> The formatted value. </returns>
            public static string Projection(IEnumerable<String> projections)
            {
                StringBuilder builder = new StringBuilder();
                builder.Append("projection=[");
                foreach (String column in projections)
                {
                    builder.Append(column).Append(',');
                }
                builder.Insert(builder.Length-1, ']');
                return builder.ToString();
            }

            /// <summary>
            /// Formats the <code>grouping</code> option for an aggregated subscription, which
            /// specifies how to determine which records to aggregate together. When this option
            /// is provided, a <code>projection</code> option must also be provided. This overload
            /// of the method expects a <code>String</code> that contains a comma-delimited list
            /// of fields to group by. See the <em>AMPS User Guide</em> for details.
            /// </summary>
            /// <param name="grouping"> The fields to group by. </param>
            /// <returns> The formatted value. </returns>
            public static string Grouping(String grouping)
            {
                return String.Format("grouping=[{0}],",grouping);
            }

            /// <summary>
            /// Formats the <code>grouping</code> option for an aggregated subscription, which
            /// specifies how to determine which records to aggregate together. When this option
            /// is provided, a <code>projection</code> option must also be provided. This overload
            /// of the method expects a <see cref="System.Collections.Generic.IEnumerable{T}"/>
            /// that contains individual fields. This method formats the contents of the
            /// <code>Iterable</code> into a comma-delimited list. See the <em>AMPS User Guide</em>
            /// for details.
            /// </summary>
            /// <param name="groupings"> The list of fields to group by. </param>
            /// <returns> The formatted value. </returns>
            public static string Grouping(IEnumerable<String> groupings)
            {
                StringBuilder builder = new StringBuilder();
                builder.Append("grouping=[");
                foreach (String column in groupings)
                {
                    builder.Append(column).Append(',');
                }
                builder.Insert(builder.Length-1, ']');
                return builder.ToString();
            }

            /// <summary>
            /// Formats the <code>ack_conflation</code> interval for a logon command. This option can
            /// be used to request that, for this connection, the server return persisted acknowledgements
            /// for publishes more frequently than the default <code>1s</code> interval. Notice that,
            /// if the server replicates messages to other destinations, the effective conflation interval
            /// may depend on the replication configuration of this server and the replication configuration
            /// of the  destinations that this server replicates to: AMPS will not provide a persisted
            /// acknowledgment for a message that has not been properly acknowledged by the applicable
            /// replication destinations. This option is only valid on a <code>logon</code> command.
            /// </summary>
            /// <param name="interval"> The interval at which to conflate persisted publish acknowledgements
            /// for this connection. </param>
            /// <returns> The formatted value. <returns>
            public static string AckConflationInterval(String interval)
            {
                return String.Format("ack_conflation={0},",interval);
            }
        }

        public Message(Encoding encoder, Encoding decoder)
        {
            this.encoder = encoder;
            this.decoder = decoder;
        }

        public Encoding encoder = null;
        public Encoding decoder = null;
        long _bookmarkSeqNo = 0;

        internal bool _ignoreAutoAck = false;

        /// <summary>
        /// This field is only used internally by bookmark store implementations to cache a tracking
        /// index for undiscarded messages, so they can be look up quickly when discarded for a
        /// bookmark subscription. There is generally no reason for end-user applications to use this.
        /// Sets the bookmark sequence number for this message.
        /// </summary>
        /// <param name="val"> The value to set as the new BookmarkSeqNo for this message. </param>
        public void setBookmarkSeqNo(long val)
        {
            _bookmarkSeqNo = val;
        }

        /// <summary>
        /// Returns the bookmark sequence number for this message.
        /// This field is only used internally by bookmark store implementations to cache a tracking
        /// index for undiscarded messages, so they can be looked up quickly when discarded for a
        /// bookmark subscription. There is generally no reason for end-user applications to use this.
        /// </summary>
        /// <returns> The bookmark sequence number for this message. </returns>
        public long getBookmarkSeqNo()
        {
            return _bookmarkSeqNo;
        }

        /// <summary>
        /// Initially, _subscription does not reference any object, it is empty
        /// </summary>
        AMPS.Client.Subscription _subscription = null;

        /// <summary>
        /// Sets the subscription for this Message.
        /// This field is only used internally by bookmark store implementations to establish a link
        /// between a message and its corresponding subscription, enabling the tracking of messages
        /// for delivery and acknowledgement.
        /// </summary>
        /// <param name="val"> The subscription to set. </param>
        internal void setSubscription(AMPS.Client.Subscription val)
        {
            _subscription = val;
        }

        /// <summary>
        /// Gets the subscription for this Message. This method is only used internally by bookmark
        /// store implementations to access the subscription object for the message being tracked.
        /// This method help identify which subscriber acknowleged the message.
        /// </summary>
        /// <returns> The subscription used for this Message. </returns>
        internal AMPS.Client.Subscription getSubscription()
        {
            return _subscription;
        }

        /// <summary>
        /// Returned by calls to <see cref="serialize(ByteBuffer)"/>
        /// </summary>
        public enum SerializationResult { OK, BufferTooSmall }

        /// <summary>
        /// Serializes this message into the provided byte buffer at its current position.
        /// </summary>
        /// <param name="buffer"> The byte buffer to write into </param>
        /// <returns> A result enum that either indicates success or that the  provided
        /// buffer doesn't have enough remaining space to hold the serialized
        /// representation of the message. </returns>
        public abstract SerializationResult serialize(System.IO.MemoryStream buffer);

        /// <summary>
        /// Resets all of the Field members to null.
        /// </summary>
        /// <returns> The Message instance. </returns>
        public Message reset()
        {
            _AckType.reset();
            _BatchSize.reset();
            _Bookmark.reset();
            _ClientName.reset();
            _CommandId.reset();
            _Command.reset();
            _CorrelationId.reset();
            _Data.reset();
            _Expiration.reset();
            _Filter.reset();
            _OrderBy.reset();
            _GroupSeqNo.reset();
            _Length.reset();
            _LeasePeriod.reset();
            _Matches.reset();
            _MessageType.reset();
            _Options.reset();
            _QueryId.reset();
            _Reason.reset();
            _RecordsInserted.reset();
            _RecordsUpdated.reset();
            _RecordsDeleted.reset();
            _RecordsReturned.reset();
            _Sequence.reset();
            _SowKey.reset();
            _SowKeys.reset();
            _Status.reset();
            _SubId.reset();
            _SubIds.reset();
            _Timestamp.reset();
            _TopN.reset();
            _Topic.reset();
            _TopicMatches.reset();
            _UserId.reset();
            _Version.reset();
            _Password.reset();
            _bookmarkSeqNo = 0;
            _subscription = null;
            _ignoreAutoAck = false;
            return this;
        }

        internal AckTypeField _AckType = new AckTypeField();

        /// <summary>
        /// Returns the type of acknowledgement for an acknowledgement Message.
        /// </summary>
        /// <returns> The AckType for this acknowledgement message. </returns>
        public Message.AckTypes getAckType()
        {
            return _AckType.getValue();
        }

        /// <summary>
        /// Returns the type of acknowledgement for an outgoing Message.
        /// </summary>
        /// <returns> The AckType for an outgoing acknowledgement message. </returns>
        public Message.AckTypes getAckTypeOutgoing()
        {
            if (_AckType.isNull()) return AckTypes.None;
            AckTypes ackType = AckTypes.None;
            int len = 0;
            for (int start = _AckType.position;
                 start <= _AckType.position + _AckType.length;
                 ++start)
            {
                if (start == _AckType.position + _AckType.length ||
                    _AckType.buffer[start] == (byte)',')
                {
                    ackType |= AckTypeField.decodeAckType(_AckType.buffer,
                                                          start - len, len);
                    len = 0;
                }
                else ++len;
            }
            return ackType;
        }

        /// <summary>
        /// Sets the type of acknowledgement for an acknowledgement message.
        /// </summary>
        /// <param name="v"> The value to set as the AckType for the acknowledgement message. </param>
        public Message setAckType(Message.AckTypes v)
        {
            _AckType.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks whether the AckType field is null.
        /// </summary>
        /// <returns> True if AckType field is null, false if not. </returns>
        public bool isAckTypeNull()
        {
            return _AckType.isNull();
        }

        internal IntegerField _BatchSize = new IntegerField();

        /// <summary>
        /// Returns the batch size of this message.
        /// </summary>
        /// <returns> The batch size of this message. </returns>
        public int getBatchSize()
        {
            return _BatchSize.getValue();
        }

        /// <summary>
        /// Sets the batch size for this message.
        /// </summary>
        /// <param name="v"> The batch size to set for the message. </param>
        public Message setBatchSize(int v)
        {
            _BatchSize.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks to see if the batch size is null.
        /// </summary>
        /// <returns> True if batch size is null, false if not. </returns>
        public bool isBatchSizeNull()
        {
            return _BatchSize.isNull();
        }

        internal BookmarkField _Bookmark = new BookmarkField();

        /// <summary>
        /// Returns the AMPS bookmark for this message. The bookmark is an identifier
        /// assigned to AMPS to locate a Message in the transaction log.
        /// </summary>
        /// <returns> The AMPS bookmark for this message. </returns>
        public String getBookmark()
        {
            return _Bookmark.getValue(decoder);
        }

        /// <summary>
        /// Gets the raw value for the AMPS bookmark. The bookmark is an identifier assigned to
        /// AMPS to locate a Message in the transaction log.
        /// </summary>
        /// <remarks>
        /// The Raw variation of this getter allows for finer-grained performance tuning. It provides
        /// direct access to the Field that encapsulates the data.
        /// It does not incur any additional allocation or unnecessary decoding of the byte[] in which the
        /// data is stored. The lifetime of the underlying buffer for the field is the lifetime of the Message.
        /// If you need the data to persist longer than the lifetime of the Message, you must
        /// copy the Field.
        /// </remarks>
        /// <returns> The raw value for the AMPS bookmark. </returns>
        public BookmarkField getBookmarkRaw()
        {
            return _Bookmark;
        }

        /// <summary>
        /// Sets the value of the AMPS bookmark. The bookmark is an identifier assigned by AMPS to locate
        /// a message in the transaction log. Setting the bookmark is used for commands that locate messages
        /// in the log, such as bookmark subscribe or historical SOW query. Setting the bookmark on a publish
        /// message has no effect.
        /// </summary>
        /// <param name="v"> The string used to set the value of the AMPS bookmark. </param>
        public Message setBookmark(String v)
        {
            _Bookmark.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Sets the value of the AMPS bookmark. The bookmark is an identifier assigned by AMPS to locate a message
        /// in the transaction log. Setting the bookmark is used for commands that locate messages in the log,
        /// such as bookmark subscribe or historical SOW query. Setting the bookmark on a publish message has no
        /// effect. </summary>
        /// <param name="buffer"> The byte buffer containing the value for the bookmark. </param>
        /// <param name="offset"> The location at which the value begins. </param>
        /// <param name="length"> The length of the value in the buffer. </param>
        public Message setBookmark(byte[] buffer, int offset, int length)
        {
            _Bookmark.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks to see if the bookmark field is null.
        /// </summary>
        /// <returns> True if the bookmark field is null, false if not. </returns>
        public bool isBookmarkNull()
        {
            return _Bookmark.isNull();
        }

        internal StringField _ClientName = new StringField();

        /// <summary>
        /// Gets the ClientName on this Message.
        /// </summary>
        /// <returns> The client name of the message. </returns>
        public String getClientName()
        {
            return _ClientName.getValue(decoder);
        }

        /// <summary>
        /// Gets the ClientName on this Message.
        /// </summary>
        /// <remarks>
        /// The RAW variation of this getter allows for finer-grained performance tuning.
        /// It provides direct access to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding of the byte[]
        /// in which the data is stored. The lifetime of the underlying buffer for the field is the lifetime
        /// of the Message. If you need the data to persist longer than the lifetime of the Message, you must
        /// copy the Field.
        /// </remarks>
        /// <returns> The raw ClientName on this Message. </returns>
        public Field getClientNameRaw()
        {
            return _ClientName;
        }

        /// <summary>
        /// Sets the name of the Client sending the Message.
        /// </summary>
        /// <param name="v"> The String used to set ClientName. </param>
        /// <returns> Returns this instance so that various operations can be chained together. </returns>
        public Message setClientName(String v)
        {
            _ClientName.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Sets the name of the Client sending the Message.
        /// </summary>
        /// <param name="buffer"> The buffer containing the ClientName. </param>
        /// <param name="offset"> The location at which ClientName begins in the buffer. </param>
        /// <param name="length"> The length of the ClientName. </param>
        public Message setClientName(byte[] buffer, int offset, int length)
        {
            _ClientName.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks whether the ClientName field is null.
        /// </summary>
        /// <returns> true if the ClientName field is null, false if not. </returns>
        public bool isClientNameNull()
        {
            return _ClientName.isNull();
        }

        internal StringField _CommandId = new StringField();

        /// <summary>
        /// Gets the CommandId for this Message. CommandId is an identifier set by the client that is used
        /// to correlate later Messages. For example, the client sets a CommandId on a subscription request to
        /// AMPS, and can later use that CommandId to unsubscribe.
        /// </summary>
        /// <returns> The decoded CommandId for this Message. </returns>
        public String getCommandId()
        {
            return _CommandId.getValue(decoder);
        }

        /// <summary>
        /// Gets the CommandId for this Message. CommandId is an identifier set by the client that is used
        /// to correlate later Messages. For example, the client sets a CommandId on a subscription request to
        /// AMPS, and can later use that CommandId to unsubscribe.
        /// </summary>
        /// <remarks>
        /// The RAW variation of this getter allows for finer-grained performance tuning. It provides direct
        /// access to the Field that encapsulates the data. This does not incur any additional allocation or
        /// unnecessary decoding of the byte[] in which the data is stored. The lifetime of the underlying buffer
        /// for the field is the lifetime of the Message. If you need the data to persist longer than the lifetime
        /// of the Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw CommandId on this Message. </returns>
        public Field getCommandIdRaw()
        {
            return _CommandId;
        }

        /// <summary>
        /// Gets the CommandId on this Message by copying it into the provided CommandId object. CommandId is
        /// an identifier set by the client that is used to correlate later Messages. For example, the client
        /// sets a CommandId on a subscription request to AMPS, and can later use that CommandId to unsubscribe.
        /// The CommandId is returned on ack messages in response to the command.
        /// </summary>
        /// <param name="v"> The object to hold the value of the CommandId object. </param>
        /// <returns> A copy of the Message in the provided CommandId object. </returns>
        public bool getCommandId(CommandId v)
        {
            return _CommandId.getValue(v);
        }

        /// <summary>
        /// Sets the CommandId for this Message. CommandId is an identifier set by the client that is used to
        /// correlate later Messages and Commands. For example, the client sets a CommandId on a subscription request
        /// to AMPS, and can later use that CommandId to unsubscribe. The CommandId is returned on ack messages
        /// in response to the command.
        /// </summary>
        /// <param name="v"> The String CommandId to set on this Message. </param>
        public Message setCommandId(String v)
        {
            _CommandId.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Sets the CommandId for this Message. CommandID is an identifier set by the client that is used to
        /// correlate later Messages and Commands. For example, the client sets a CommandId on a subscription request
        /// to AMPS, and can later use that CommandId to unsubscribe. The CommandId is returned on ack messages
        /// in response to the command.
        /// </summary>
        /// <param name="v"> The CommandId object to set on this Message. </param>
        /// <returns> The CommandId object. </returns>
        public Message setCommandId(CommandId v)
        {
            _CommandId.setValue(v);
            return this;
        }

        /// <summary>
        /// Sets the CommandId for this Message. CommandId is an identifier set by the client that is used to
        /// correlate later Messages and Commands. For example, the client sets a CommandId on a subscription request
        /// to AMPS, and can later use that CommandId to unsubscribe. The CommandId is returned on ack messages
        /// in response to the command.
        /// </summary>
        /// <param name="buffer"> The buffer containing the CommandId on this Message. </param>
        /// <param name="offset"> The location at which the CommandId begins. </param>
        /// <param name="length"> The length of the CommandId. </param>
        /// <returns> Returns this instance </returns>
        public Message setCommandId(byte[] buffer, int offset, int length)
        {
            _CommandId.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks to see if the CommandId is null.
        /// </summary>
        /// <returns> True if the CommandId is null, false if not. </returns>
        public bool isCommandIdNull()
        {
            return _CommandId.isNull();
        }

        internal CommandField _Command = new CommandField();

        /// <summary>
        /// Returns the Command for this Message, indicating the type of message this is. The fields that are
        /// supported by this message, and how they are set and interpreted, depend on the type of message.
        /// See the AMPS Command reference for details.
        /// </summary>
        /// <returns> The Command for this Message. </returns>
        public Commands getCommand()
        {
            return _Command.getValue();
        }

        /// <summary>
        /// Sets the Command for this Message, indicating the type of messagethis is. The fields that are supported
        /// by this message, and how they are set and interpreted, depend on the type of message.
        /// See the AMPS Command reference for details.
        /// </summary>
        /// <param name="v"> The Command object used to set the Command. </param>
        /// <returns> The Command that was set. </returns>
        public Message setCommand(Commands v)
        {
            _Command.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks to see if the Command is null.
        /// </summary>
        /// <returns> True if the Command is null, false if not. </returns>
        public bool isCommandNull()
        {
            return _Command.isNull();
        }

        internal StringField _Data = new StringField();

        /// <summary>
        /// Returns the payload of the Message.
        /// </summary>
        /// <returns> The decoded payload of this Message. </returns>
        public virtual String getData()
        {
            return _Data.getValue(decoder);
        }

        /// <summary>
        /// Returns the payload of the Message.
        /// </summary>
        /// <remarks>
        /// The RAW variation of this getter allows for finer-grained performance tuning. It provides direct access
        /// to the Field that encapsulates the data. This does not incur any additional allocation or unnecessary
        /// decoding of the byte[] in which the data is stored. The lifetime of the underlying buffer of the field
        /// is the lifetime of the Message. If you need the data to persist longer than the lifetime of the Message,
        /// you must copy the <see cref="AMPS.Client.Field"/>
        /// </remarks>
        /// <returns> The raw payload of this Message. </returns>
        public Field getDataRaw()
        {
            return _Data;
        }

        /// <summary>
        /// Sets the payload of the Message.
        /// </summary>
        /// <param name="v"> The string to set as the payload of this Message. </param>
        public Message setData(String v)
        {
            _Data.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Sets the payload of the Message.
        /// </summary>
        /// <param name="buffer"> The buffer containing the data. </param>
        /// <param name="offset"> The location at which the data begins. </param>
        /// <param name="length"> The length of the data. </param>
        public Message setData(byte[] buffer, int offset, int length)
        {
            _Data.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks to see if the data field is null.
        /// </summary>
        /// <returns> True if the data field is null, false if not. </returns>
        public bool isDataNull()
        {
            return _Data.isNull();
        }

        internal IntegerField _Expiration = new IntegerField();

        /// <summary>
        /// Returns the Expiration field set for this Message.
        /// </summary>
        /// <returns> The Expiration for this Message. </returns>
        public IntegerField getExpirationRaw()
        {
            return _Expiration;
        }

        /// <summary>
        /// Returns the Expiration field for this Message.
        /// </summary>
        /// <remarks>
        /// The RAW variation of this getter allows for finer-grained performance tuning. It provides direct
        /// access to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding of the byte[] in which the data
        /// is stored. The lifetime of the underlying buffer for the field is the lifetime of the Message.
        /// If you need the data to persist longer than the lifetime of the Message, then you must copy the Field.
        /// </remarks>
        /// <returns> The raw Expiration for this Message. </returns>
        public int getExpiration()
        {
            return _Expiration.getValue();
        }

        /// <summary>
        /// Sets the Expiration for a Message.
        /// </summary>
        /// <remarks>
        /// The expiration is used on a publish command to set the lifetime of a message. For the lifetime to be
        /// processed by AMPS, the message must be published to a SOW topic that supports message expiration. See
        /// the AMPS user guide for details.
        /// </remarks>
        /// <param name="v"> The lifetime of the Message. </param>
        public Message setExpiration(int v)
        {
            _Expiration.setValue(v);
            return this;
        }

        /// <summary>
        /// Checked whether the Expiration is field is null.
        /// </summary>
        /// <returns> True if the Expiration field is null, false if not. </returns>
        public bool isExpirationNull()
        {
            return _Expiration.isNull();
        }

        internal StringField _Filter = new StringField();

        /// <summary>
        /// Returns the Filter field for this Message.
        /// </summary>
        /// <remarks>
        /// Filters are used for commands that provide content filtering, such as SOW queries, subscriptions, and SOW
        /// delete.
        /// </remarks>
        /// <returns> The decoded Filter on this Message. </returns>
        public String getFilter()
        {
            return _Filter.getValue(decoder);
        }

        /// <summary>
        /// Returns the Filter field on this Message.
        /// </summary>
        /// <remarks>
        /// Filters are used for commands that provide content filtering, such as SOW queries, subscriptions, and SOW
        /// delete. The RAW variation of this getter allows for finer-grained performance tuning. It provides direct
        /// access to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding of the byte[] in which the data is
        /// stored. The lifetime of the underlying buffer for the field is the lifetime of the Message. If you need
        /// the data to persist longer than the lifetime of the Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw Filter on this Message. </returns>
        public Field getFilterRaw()
        {
            return _Filter;
        }

        /// <summary>
        /// Sets the Filter field on this Message.
        /// </summary>
        /// <remarks>
        /// Filters are used for commands that provide content filtering, such as SOW queries, subscriptions, and SOW
        /// delete.
        /// </remarks>
        /// <param name="v"> The string to set as the Filter on this Message. </param>
        public Message setFilter(String v)
        {
            _Filter.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Sets the Filter field on this Message.
        /// </summary>
        /// <remarks>
        /// Filters are used for commands that provide content filtering, such as SOW queries, subscriptions, and SOW
        /// delete.
        /// </remarks>
        /// <param name="buffer"> The buffer containing the Filter for this Message. </param>
        /// <param name="offset"> The location at which the Filter begins. </param>
        /// <param name="length"> The length of the Filter. </param>
        public Message setFilter(byte[] buffer, int offset, int length)
        {
            _Filter.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks to see if the Filter is null on this Message.
        /// </summary>
        /// <returns> True if the Filter is null, false if not. </returns>
        public bool isFilterNull()
        {
            return _Filter.isNull();
        }

        internal StringField _OrderBy = new StringField();

        /// <summary>
        /// Get the OrderBy parameter for this Message.
        /// For commands that support ordering, this parameter provides the order in which AMPS returns results.
        /// </summary>
        /// <returns> The decoded OrderBy parameter for this Message. </returns>
        public String getOrderBy()
        {
            return _OrderBy.getValue(decoder);
        }

        /// <summary>
        /// Gets the OrderBy parameter for this Message.
        /// For commands that support ordering, this parameter provides the order in which AMPS returns results.
        /// </summary>
        /// <remarks>
        /// The RAW variation of this getter allows for finer-grained performance tuning. It provides direct access
        /// to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding of the byte[] in  which the data is
        /// stored. The lifetime of the underlying buffer for the field is the lifetime of the Message. If you need
        /// the data to persist longer than the lifetime of the Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw OrderBy parameter for this Message. </returns>
        public Field getOrderByRaw()
        {
            return _OrderBy;
        }

        /// <summary>
        /// Set the OrderBy parameter for this Message.
        /// For commands that support ordering, this parameter provides the order in which AMPS returns results.
        /// See the AMPS User Guide for details.
        /// </summary>
        /// <param name="v"> The ordering expression to set for this Message. </param>
        public Message setOrderBy(String v)
        {
            _OrderBy.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Set the OrderBy parameter for this Message.
        /// For commands that support ordering, this parameter provies the order in which AMPS returns results.
        /// See the AMPS User Guide for details.
        /// </summary>
        /// <param name="buffer"> The buffer that contains the expression. </param>
        /// <param name="offset"> The location at which the expression begins. </param>
        /// <param name="length"> The length of the expresson. </param>
        public Message setOrderBy(byte[] buffer,int offset,int length)
        {
            _OrderBy.setValue(buffer,offset,length);
            return this;
        }

        /// <summary>
        /// Checks to see if the OrderBy field is null.
        /// </summary>
        /// <returns> True if the OrderBy parameter is null, false if not. </returns>
        public bool isOrderByNull()
        {
            return _OrderBy.isNull();
        }

        internal LongField _GroupSeqNo = new LongField();

        /// <summary>
        /// Get the GroupSeqNo. The group sequence number is the number of the batch within the SOW response,
        /// and is set on incoming messages by the AMPS client.
        /// </summary>
        /// <returns> The GroupSeqNo for this Message. </returns>
        public long getGroupSeqNo()
        {
            return _GroupSeqNo.getValue();
        }

        /// <summary>
        /// Set the GroupSeqNo. The group sequence number is the number of the batch within the SOW response, and
        /// is set on incoming messages by the AMPS client.
        /// </summary>
        /// <param name="v"> The number to set as the GroupSeqNo. </param>
        public Message setGroupSeqNo(long v)
        {
            _GroupSeqNo.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks whether the GroupSeqNo field is null.
        /// </summary>
        /// <returns> True if the GroupSeqNo field is null, false if not. </returns>
        public bool isGroupSeqNoNull()
        {
            return _GroupSeqNo.isNull();
        }

        internal LongField _Matches = new LongField();

        /// <summary>
        /// Returns the number of matches in the command this Message acknowledges. This header is provided on an
        /// acknowledgement message that includes information on the number of matches. See the AMPS Command Reference
        /// for details.
        /// </summary>
        /// <returns> The number of matches in the command that this Message acknowledges. </returns>
        public long getMatches()
        {
            return _Matches.getValue();
        }

        /// <summary>
        /// Sets the matches field on this Message. This header is typically provided by AMPS.
        /// </summary>
        /// <param name="v"> The number of matches in the command this Message acknowledges. </param>
        public Message setMatches(long v)
        {
            _Matches.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks whether the Matches field is null.
        /// </summary>
        /// <returns> True if the Matches field is null, false if not. </returns>
        public bool isMatchesNull()
        {
            return _Matches.isNull();
        }

        internal StringField _MessageType = new StringField();

        /// <summary>
        /// Returns the MessageType on this client.
        /// MessageType is used to specify the message type given in the URI.
        /// </summary>
        /// <returns> The MessageType on this client. </returns>
        public String getMessageType()
        {
            return _MessageType.getValue(decoder);
        }

        /// <summary>
        /// Sets the MessageType on this client.
        /// MessageType is used to specify the message type given in the URI.
        /// </summary>
        /// <returns> The MessageType of this client. </returns>
        public Message setMessageType(String v)
        {
            _MessageType.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Checks whether the MessageType field is null.
        /// </summary>
        /// <returns> True if MessageType field is null, false if not. </returns>
        public bool isMessageTypeNull()
        {
            return _MessageType.isNull();
        }

        internal StringField _Password = new StringField();

        /// <summary>
        /// Returns the Password set on this Message.
        /// This field is typically used during the logon sequence.
        /// </summary>
        /// <returns> The decoded Password set on the specified Message. </returns>
        public string getPassword()
        {
            return _Password.getValue(decoder);
        }

        /// <summary>
        /// Returns the Password set on this Message.
        /// This field is typically used during the logon sequence.
        /// </summary>
        /// <remarks>
        /// The RAW variation of this getter allows for finer-grained performance tuning. It provides direct access
        /// to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding of the byte[] in which the data
        /// is stored. The lifetime of the underlying buffer for the field is the lifetime of the Message.
        /// If you need the data to persist longer than the lifetime of the Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw Password set on this Message. </returns>
        public Field getPasswordRaw()
        {
            return _Password;
        }

        /// <summary>
        /// Set the Password on the Message.
        /// This field is typically used during the logon sequence.
        /// </summary>
        /// <param name="data"> The buffer that contains the Password. </param>
        /// <param name="offset"> The location where the Password starts. </param>
        /// <param name="length"> The length of the Password. </param>
        public Message setPassword(byte[] data, int offset, int length)
        {
            _Password.setValue(data, offset, length);
            return this;
        }

        /// <summary>
        /// Set the password on the Message.
        /// This field is typically used during the logon sequence.
        /// </summary>
        /// <param name="password"> The string Password to set on the Message. </param>
        public Message setPassword(string password)
        {
            _Password.setValue(password, encoder);
            return this;
        }
        internal IntegerField _Length = new IntegerField();

        /// <summary>
        /// Returns the Length of the Message.
        /// </summary>
        /// <returns> The Length of the message. </returns>
        public int getLength()
        {
            return _Length.getValue();
        }

        /// <summary>
        /// Sets the Length of the Message.
        /// This field is typically set by the AMPS Client API rather than by application code.
        /// </summary>
        /// <param name="v"> The value to be set as the Length of the Message. </param>
        public Message setLength(int v)
        {
            _Length.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks whether the Password field is null.
        /// </summary>
        /// <returns> True if the Password field is null, false if not. </returns>
        public bool isPasswordNull()
        {
            return _Password.isNull();
        }

        internal OptionsField _Options = new OptionsField();

        /// <summary>
        /// Returns the Options to be set on this Message.
        /// </summary>
        /// <remarks>
        /// The RAW variation of this getter allows for finer-grained performance tuning. It provides direct access
        /// to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding of the byte[] in which the data is
        /// stored. The lifetime of the underlying buffer for the field is the lifetime of the Message. If you need
        /// the data to persist longer than the lifetime of the Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw Options for this Message. </returns>
        public Field getOptionsRaw()
        {
            return _Options;
        }

        /// <summary>
        /// Returns the Options to be set on this Message.
        /// </summary>
        /// <returns> The options to be set on this Message. </returns>
        public string getOptions()
        {
            return _Options.getValue(decoder);
        }

        /// <summary>
        /// Sets the Options for this Message.
        /// Options are a comma-delimited list of parameters. The values accepted for options depend on the command:
        /// see the AMPS Command Reference for details.
        /// </summary>
        /// <param name="options"> The string to be used as the Options for this Message. </param>
        public Message setOptions(string options)
        {
            _Options.setValue(options, encoder);
            return this;
        }

        /// <summary>
        /// Checks to see if the Options field is null.
        /// </summary>
        /// <returns> True if null, false if not. </returns>
        public bool isOptionsNull()
        {
            return _Options.isNull();
        }

        internal StringField _QueryId = new StringField();

        /// <summary>
        /// Returns the QueryId for this Message.
        /// </summary>
        /// <remarks>
        /// The QueryId returned on a message is the CommandId of the command that ran the query. For example,
        /// when sending a SOW command to AMPS, messages returned in response to that command will have the QueryId
        /// set to the CommandId of the SOW command.
        /// </remarks>
        /// <returns> The decoded QueryId for this Message. </returns>
        public String getQueryId()
        {
            return _QueryId.getValue(decoder);
        }

        /// <summary>
        /// Returns the QueryId for this Message.
        /// </summary>
        /// <remarks>
        /// The QueryId returned on a message is the CommandId of the command that ran the query. For example, when
        /// sending a SOW command to AMPS, messages returned in response to that command will have the QueryId set to
        /// the CommandId of the SOW command.
        /// The RAW variation of this getter allows for finer-grained performance tuning. It provides direct access
        /// to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding of the byte[] in which the data
        /// is stored. The lifetime of the underlying buffer is the lifetime of the Message. If you need the data
        /// to persist longer than the lifetime of the Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw QueryId for this Message. </returns>
        public Field getQueryIdRaw()
        {
            return _QueryId;
        }

        /// <summary>
        /// Checks whether the QueryId field is null.
        /// </summary>
        /// <returns> True if the QueryId field is null, false if not. </returns>
        public bool isQueryIdNull()
        {
            return _QueryId.isNull();
        }

        /// <summary>
        /// Returns the QueryId for this Message by copying it into the provided CommandId.
        /// </summary>
        /// <remarks>
        /// The QueryId returned on a message is the CommandId of the command that ran the query. For example,
        /// when sending a SOW command to AMPS, messages returned in response to that command will have the QueryId
        /// set to the CommandId of the SOW command.
        /// </remarks>
        /// <param name="v"> The CommandId to be used to copy the QueryId. </param>
        public bool getQueryId(CommandId v)
        {
            return _QueryId.getValue(v);
        }

        /// <summary>
        /// Sets the QueryId field for this Message.
        /// Typically, the QueryId is set by the AMPS client on incoming messages.
        /// </summary>
        /// <param name="v"> The string to be set as the QueryId for this Message. </param>
        public Message setQueryId(String v)
        {
            _QueryId.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Sets the QueryId for this Message by copying it from a CommandId object.
        /// Typically, the QueryId is set for incoming messages by the AMPS Client.
        /// </summary>
        /// <param name="v"> The CommandId to be set as the QueeryId for this Message. </param>
        public Message setQueryId(CommandId v)
        {
            _QueryId.setValue(v);
            return this;
        }

        /// <summary>
        /// Set the QueryId for this Message.
        /// Typically, the QueryId is set for incoming messages by the AMPS Client.
        /// </summary>
        /// <param name="buffer"> The buffer containing the QueryId. </param>
        /// <param name="offset"> The location in the buffer where the QueryId begins. </param>
        /// <param name="position"> The length of the QueryId to be set. </param>
        public Message setQueryId(byte[] buffer, int offset, int position)
        {
            _QueryId.setValue(buffer, offset, position);
            return this;
        }

        internal ReasonField _Reason = new ReasonField();

        /// <summary>
        /// Returns the Reason value for this Message.
        /// The Reason is set on acknowledgement messages to provide more information about the acknowledgement.
        /// </summary>
        /// <returns> The value in the Reason field of this Message. </returns>
        public Reasons getReason()
        {
            return _Reason.getValue();
        }

        /// <summary>
        /// Returns the Reason value for this Message.
        /// Typically, the Reason is set for incoming messages by the AMPS Client.
        /// </summary>
        /// <returns> The value in the Reason field of this Message. </returns>
        public string getReasonText()
        {
            return _Reason.getText();
        }

        /// <summary>
        /// Set the Reason value for this Message.
        /// Typically, the Reason is set for incoming messages by the AMPS Client.
        /// </summary>
        /// <param name="v"> The Reason value to set for this Message. </param>
        public Message setReason(Reasons v)
        {
            _Reason.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks whether the Reason is null.
        /// </summary>
        /// <returns> True if the Reason field is null, false if not. </returns>
        public bool isReasonNull()
        {
            return _Reason.isNull();
        }

        internal LongField _RecordsDeleted = new LongField();

        /// <summary>
        /// Returns the number of records deleted for the command that this Message was produced in response to.
        /// This header is provided on an acknowledgement message. See the AMPS Command Reference for more details.
        /// </summary>
        /// <returns> The number of records deleted for the command that this Message was produced for. </returns>
        public long getRecordsDeleted()
        {
            return _RecordsDeleted.getValue();
        }

        /// <summary>
        /// Sets the records deleted header field.
        /// This header is provided on an acknowledgement message. An application does not typically need to set
        /// this field. See the AMPS Command Reference for details.
        /// </summary>
        /// <param name="v"> The value to set as the RecordsDeleted field. </param>
        public Message setRecordsDeleted(long v)
        {
            _RecordsDeleted.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks whether the RecordsDeleted field is null.
        /// </summary>
        /// <returns> True if the RecordsDeleted field is null, false if not. </returns>
        public bool isRecordsDeletedNull()
        {
            return _RecordsDeleted.isNull();
        }

        internal LongField _RecordsInserted = new LongField();

        /// <summary>
        /// Returns the number of records inserted for the command that this Message was produced for.
        /// This header is provided on an acknowledgement message that includes information on the number
        /// of records inserted. See the AMPS Command Reference for details.
        /// </summary>
        /// <returns> The number of records inserted for the command that produced this Message. </returns>
        public long getRecordsInserted()
        {
            return _RecordsInserted.getValue();
        }

        /// <summary>
        /// Sets the Records Inserted field. This header is provided on an acknowledgement message. An application
        /// typically does not need to set this field. See the AMPS Command Reference for details.
        /// </summary>
        /// <param name="v"> The value to set as the RecordsInserted field of the Message. </param>
        public Message setRecordsInserted(long v)
        {
            _RecordsInserted.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks whether the RecordsInserted field is null.
        /// </summary>
        /// <returns> True if the RecordsInserted field is null, false if not. </returns>
        public bool isRecordsInsertedNull()
        {
            return _RecordsInserted.isNull();
        }

        internal LongField _RecordsUpdated = new LongField();

        /// <summary>
        /// Returns the number of records updated for the command that this Message was produced in response to.
        /// </summary>
        /// <remarks>
        /// This header is provided on an acknowledgement message. See the AMPS Command Reference for details.
        /// </remarks>
        /// <returns> The number of records updated for the command that this Message was produced for. </returns>
        public long getRecordsUpdated()
        {
            return _RecordsUpdated.getValue();
        }

        /// <summary>
        /// Sets the RecordsUpdated header field.
        /// </summary>
        /// <remarks>
        /// This header is provided on an acknowledgement message. An application does not typically need to set
        /// this field. See the AMPS Command Reference for details.
        /// </remarks>
        /// <param name="v"> The value to set the RecordsUpdated field to. </param>
        public Message setRecordsUpdated(long v)
        {
            _RecordsUpdated.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks whether the RecordsUpdated field is null.
        /// </summary>
        /// <returns> True if the RecordsUpdated field is null, false if not. </returns>
        public bool isRecordsUpdatedNull()
        {
            return _RecordsUpdated.isNull();
        }

        internal LongField _RecordsReturned = new LongField();

        /// <summary>
        /// Returns the number of records returned for the command that this Message was produced in response to.
        /// </summary>
        /// <remarks>
        /// This header is provided on an acknowledgement message. See the AMPS Command Reference for details.
        /// </remarks>
        /// <returns> The number of RecordsReturned for the command that this Message was produced for. </returns>
        public long getRecordsReturned()
        {
            return _RecordsReturned.getValue();
        }

        /// <summary>
        /// Sets the RecordsReturned header field.
        /// </summary>
        /// <remarks>
        /// This header is provided on acknowledgement message. An application does not typically need to set this
        /// field. See the AMPS Command Reference for details.
        /// </remarks>
        /// <param name="v"> The value to set as the RecordsReturned field. </param>
        public Message setRecordsReturned(long v)
        {
            _RecordsReturned.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks whether the RecordsReturned field is null.
        /// </summary>
        /// <returns> True if the RecordsReturned field is null, false if not. </returns>
        public bool isRecordsReturnedNull()
        {
            return _RecordsReturned.isNull();
        }

        internal UlongField _Sequence = new UlongField();

        /// <summary>
        /// Returns the sequence number for this Message.
        /// </summary>
        /// <remarks>
        /// On outgoing messages, this field is typically set by the AMPS client.
        /// The sequence number is used by AMPS for duplicate detection. Each combination of client name and
        /// sequence number should be a unique message. In response to a login, AMPS may return the last sequence
        /// number received from a client to assist in recovery.
        /// The Raw variation of this getter allows for finer-grained performance tuning. It provides direct access
        /// to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding of the byte[] in which the data
        /// is stored. The lifetime of the underlying buffer for the field is the lifetime of the Message.
        /// If you need the data to persist longer than the lifetime of the Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw sequence number for this Message. </returns>
        public Field getSequenceRaw()
        {
            return _Sequence;
        }

        /// <summary>
        /// Returns the sequence number for this Message.
        /// </summary>
        /// <remarks>
        /// On outgoing messages, this field is typically set by the AMPS client. The sequence number is used by
        /// AMPS for duplicate detection. Each combination of client name and sequence number should be a unique
        /// message. In response to a logon, AMPS may return the last sequence number received from a client to
        /// assist in recovery. </remarks>
        /// <returns> The decoded sequence number for this Message, in the form of an unsigned long. </returns>
        public ulong getSequence()
        {
            return _Sequence.getValue();
        }

        /// <summary>
        /// Sets the sequence number for this Message.
        /// </summary>
        /// <remarks>
        /// This field is typically set by the AMPS client.
        /// The AMPS client uses the sequence number in outgoing messages. The sequence number is used by AMPS for
        /// duplicate detection. Each combination of client name and sequence number should be a unique message.
        /// </remarks>
        /// <param name="v"> The value to set as the sequence number for this Message, in the form of an unsigned
        /// long. </param>
        public Message setSequence(ulong v)
        {
            _Sequence.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks whether the Sequence field is null.
        /// </summary>
        /// <returns> true if the Sequence field is null, false if not. </returns>
        public bool isSequenceNull()
        {
            return _Sequence.isNull();
        }

        internal StringField _SowKey = new StringField();

        /// <summary>
        /// Returns the SowKey for this Message.
        /// </summary>
        /// <remarks>
        /// The SowKey is an opaque identifier used to uniquely identify a SOW record within AMPS. For messages
        /// received from a SOW, AMPS provides the SowKey on each message.
        /// </remarks>
        /// <returns> The decoded SowKey for this Message. </returns>
        public String getSowKey()
        {
            return _SowKey.getValue(decoder);
        }

        /// <summary>
        /// Returns the SowKey for this Message.
        /// </summary>
        /// <remarks>
        /// The SowKey is an opaque identifier used to uniquely identify a SOW record within AMPS> For messages
        /// received from a SOW, AMPS provides the SowKey on each message.
        /// The Raw variation of this getter allows for finer-grained performance tuning. It provides direct access
        /// to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding of the byte[] in which the data is
        /// stored. The lifetime of the underlying buffer is the lifetime of the Message. If you need the data
        /// to persist longer than the lifetime of the Message, you must copy the Field.
        /// </remarks>
        /// <returns> The Raw SowKey for this Message. </returns>
        public Field getSowKeyRaw()
        {
            return _SowKey;
        }

        /// <summary>
        /// Sets the SowKey for this Message.
        /// </summary>
        /// <remarks>
        /// The SowKey is an opaque identifier used to uniquely identify a SOW record within AMPS. For messages
        /// received from a SOW, AMPS provides the SOW key on each message.
        /// </remarks>
        /// <param name="v"> The string to be used for the SowKey. </param>
        public Message setSowKey(String v)
        {
            _SowKey.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Sets the SowKey for this Message.
        /// </summary>
        /// <remarks>
        /// The SowKey is an opaque identifier used to uniquely identify a SOW record within AMPS. For messages
        /// received from a SOW, AMPS provides the SOW key on each message.
        /// </remarks>
        /// <param name="buffer"> The buffer that contains the SOW key. </param>
        /// <param name="offset"> The location within the buffer where the SOW Key begins. </param>
        /// <param name="length"> The length of the SOW Key. </param>
        public Message setSowKey(byte[] buffer, int offset, int length)
        {
            _SowKey.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks whether the SowKey field is null.
        /// </summary>
        /// <returns> True if the SowKey field is null, false if not. </returns>
        public bool isSowKeyNull()
        {
            return _SowKey.isNull();
        }

        internal StringField _SowKeys = new StringField();

        /// <summary>
        /// Gets the set of SowKeys this message applies to.
        /// </summary>
        /// <remarks>
        /// The SowKey is an opaque identifier used to uniquely identify a SOW record within AMPS. For messages
        /// received from a SOW, AMPS provides the SOW key on each message.
        /// </remarks>
        /// <returns> The decoded set of SowKeys this message applies to. </returns>
        public String getSowKeys()
        {
            return _SowKeys.getValue(decoder);
        }

         /// <summary>
        /// Returns the set of SowKeys this message applies to.
        /// </summary>
        /// <remarks>
        /// The SowKey is an opaque identifier used to uniquely identify a SOW record within AMPS> For messages
        /// received from a SOW, AMPS provides the SowKey on each message.
        /// The Raw variation of this getter allows for finer-grained performance tuning. It provides direct access
        /// to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding of the byte[] in which the data is
        /// stored. The lifetime of the underlying buffer is the lifetime of the Message. If you need the data
        /// to persist longer than the lifetime of the Message, you must copy the Field.
        /// </remarks>
        /// <returns> The Raw SowKeys this message applies to. </returns>
        public Field getSowKeysRaw()
        {
            return _SowKeys;
        }

        /// <summary>
        /// Sets the set of SowKeys this message applies to, as a comma-delimited list of identifiers. This can
        /// be useful for commands that operate on multiple SOW records, such as a sow_delete that specifies
        /// a set of keys to remove.
        /// </summary>
        /// <remarks>
        /// The SowKey is an opaque identifier used to uniquely identify a SOW record within AMPS> For messages
        /// received from a SOW, AMPS provides the SOW key on each message.
        /// </remarks>
        /// <param name="v"> A comma-delimited list of identifiers to be set as the SowKeys that this message
        /// applies to. </param>
        public Message setSowKeys(String v)
        {
            _SowKeys.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Sets the set of SowKeys this message applies to, as a comma-delimited list of identifiers. This can
        /// be useful for commands that operate on multiple SOW records, such as a sow_delete that specifies
        /// a set of keys to remove.
        /// </summary>
        /// <remarks>
        /// The SowKey is an opaque identifier used to uniquely identify a SOW record within AMPS> For messages
        /// received from a SOW, AMPS provides the SOW key on each message.
        /// </remarks>
        /// <param name="buffer"> The buffer that contains the SOW Key. </param>
        /// <param name="offset"> The location within the buffer where the SOW Key begins. </param>
        /// <param name="length"> The length of the SOW Key. </param>
        public Message setSowKeys(byte[] buffer, int offset, int length)
        {
            _SowKeys.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks whether the SowKeys field is null.
        /// </summary>
        /// <returns> True if the SowKeys field is null, false if not. </returns>
        public bool isSowKeysNull()
        {
            return _SowKeys.isNull();
        }

        internal StatusField _Status = new StatusField();

        /// <summary>
        /// Returns the Status of this Message.
        /// </summary>
        /// <remarks>
        /// This field is provided by AMPS on acknowledgements to indicate the success or failure of the command
        /// being acknowledged.
        /// </remarks>
        /// <returns> The Status of this Message, which will be None, Success, Failure or Retry. </returns>
        public State getStatus()
        {
            return _Status.getValue();
        }

        /// <summary>
        /// Sets the Status of this Message.
        /// </summary>
        /// <remarks>
        /// This field is provided by AMPS on acknowledgements to indicate the
        /// success or failure of the command being acknowledged.
        /// </remarks>
        /// <param name="v"> The value to be set as the Status of the Message. </param>
        public Message setStatus(State v)
        {
            _Status.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks to see whether the Status field is null.
        /// </summary>
        /// <returns> True if the Status field is null, false if not. </returns>
        public bool isStatusNull()
        {
            return _Status.isNull();
        }

        internal StringField _SubId = new StringField();

        /// <summary>
        /// Gets the SubId of this Message.
        /// </summary>
        /// <remarks>
        /// The SubId is a reference to the CommandId of the command that created the subscription.
        /// </remarks>
        /// <returns> The decoded SubId of this Message. </returns>
        public String getSubId()
        {
            return _SubId.getValue(decoder);
        }

        /// <summary>
        /// Gets the SubId of this Message.
        /// </summary>
        /// <remarks>
        /// The SubId is a reference to the CommandId of the command that created the subscription.
        /// The Raw variation of this getter allows for finer-grained performance tuning.
        /// It provides direct access to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding of the byte[] in which
        /// the data is stored. The lifetime of the underlying buffer for the field is the lifetime of the
        /// Message. If you need the data to persist longer than the lifetime of the Message, you must
        /// copy the Field.
        /// </remarks>
        /// <returns> The raw SubId of this Message. </returns>
        public Field getSubIdRaw()
        {
            return _SubId;
        }

        /// <summary>
        /// Gets the SubId of this Message by copying it into the provided CommandId.
        /// </summary>
        /// <remarks>
        /// The SubId is a reference to the CommandId of the command that created the subscription.
        /// </remarks>
        /// <returns> The CommandId containing the SubId. </returns>
        /// <param name="v"> The CommandId object to hold the SubId. </param>
        public bool getSubId(CommandId v)
        {
            return _SubId.getValue(v);
        }

        /// <summary>
        /// Sets the SubId of this Message.
        /// </summary>
        /// <remarks>
        /// The SubId is a reference to the CommandId of the command that created the subscription.
        /// </remarks>
        /// <param name="v"> The string used to set the SubId of this Message. </param>
        public Message setSubId(String v)
        {
            _SubId.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Sets the SubId of this Message by copying it from a CommandId object.
        /// </summary>
        /// <remarks>
        /// The SubId is the CommandId of the command that created the subscription. For example, to replace
        /// a subscription, you provide the SubId of the subscription to be replaced with the command that
        /// replaces the subscription.
        /// </remarks>
        /// <param name="v"> The CommandId to be set as the SubId for this Message. </param>
        public Message setSubId(CommandId v)
        {
            _SubId.setValue(v);
            return this;
        }

        /// <summary>
        /// Sets the SubId of this Message.
        /// </summary>
        /// <remarks>
        /// The SubId is a reference to the CommandId of the command that created the subscription. For example,
        /// messages returned for a subscription contain the SubId of the subscription.
        /// </remarks>
        /// <param name="buffer"> The buffer that contains the SubId key. </param>
        /// <param name="offset"> The location within the buffer where the SubId begins. </param>
        /// <param name="length"> The length of the SubId. </param>
        public Message setSubId(byte[] buffer, int offset, int length)
        {
            _SubId.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks whether the SubId field is null.
        /// </summary>
        /// <returns> true if the SubId is null, false if not. </returns>
        public bool isSubIdNull()
        {
            return _SubId.isNull();
        }

        internal StringField _SubIds = new StringField();

        /// <summary>
        /// The set of SubIds for this Message. AMPS returns a SubId for each
        /// subscription that matches the message as a comma-delimited list.
        /// </summary>
        /// <remarks>
        /// The SubId is the identifier provided when the subscription was
        /// registered with AMPS.
        /// </remarks>
        /// <returns> The decoded set of SubIds for this Message. </returns>
        public String getSubIds()
        {
            return _SubIds.getValue(decoder);
        }

        /// <summary>
        /// The set of SubIds for this message. AMPS returns a SubId for each
        /// subscription that matches the message as a comma-delimited list.
        /// </summary>
        /// <remarks>
        /// The SubId is the identifier provided when the subscription was
        /// registered with AMPS.
        /// The Raw variation of this getter allows for finer-grained performance tuning.
        /// It provides direct access to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding
        /// of the byte[] in which the data is stored. The lifetime of the
        /// underlying buffer for the field is the lifetime of the Message.
        /// If you need the data to persist longer than the lifetime of the
        /// Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw set of SubIds for this message. </returns>
        public Field getSubIdsRaw()
        {
            return _SubIds;
        }

        /// <summary>
        /// The set of SubIds for this message. AMPS returns a SubId for each
        /// subscription that matches the message as a comma-delimited list.
        /// </summary>
        /// <remarks>
        /// The SubId is the identifier provided when the subscription was
        /// registered with AMPS.
        /// </remarks>
        /// <param name="v"> The comma-delimited list to be set as the SubIds for this message. </param>
        public Message setSubIds(String v)
        {
            _SubIds.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// The set of SubIds for this message. AMPS returns a SubId for each
        /// subscription that matches the message.
        /// </summary>
        /// <remarks>
        /// The SubId is the identifier provided when the subscription was
        /// registered with AMPS.
        /// </remarks>
        /// <param name="buffer"> The buffer containing the SubIds for this message. </param>
        /// <param name="offset"> The location of where the SubIds begin. </param>
        /// <param name="length"> The length of the list of SubIds. </param>
        public Message setSubIds(byte[] buffer, int offset, int length)
        {
            _SubIds.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks whether the SubIds field is null.
        /// </summary>
        /// <returns> True if null, false if not. </returns>
        public bool isSubIdsNull()
        {
            return _SubIds.isNull();
        }

        internal StringField _Timestamp = new StringField();

        /// <summary>
        /// Returns the timestamp for this message, an ISO-8601 formatted string.
        /// The timestamp is set by AMPS at the time that AMPS processes the message.
        /// </summary>
        /// <returns> The Timestamp for this Message as an ISO-8601 formatted string. </returns>
        public String getTimestamp()
        {
            return _Timestamp.getValue(decoder);
        }

        /// <summary>
        /// Returns the timestamp for this message, an ISO-8601 formatted string.
        /// The timestamp is set by AMPS  at the time that AMPS processes
        /// the message.
        /// </summary>
        /// <remarks>
        /// The Raw variation of this getter allows for finer-grained performance tuning.
        /// It provides direct access to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding
        /// of the byte[] in which the data is stored. The lifetime of the
        /// underlying buffer for the field is the lifetime of the Message.
        /// If you need the data to persist longer than the lifetime of the
        /// Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw Timestamp for this Message. </returns>
        public Field getTimestampRaw()
        {
            return _Timestamp;
        }

        /// <summary>
        /// Sets the Timestamp for this Message. The timestamp is set by AMPS
        /// at the time that AMPS processes the message. There is generally no
        /// need to set this field in your application.
        /// </summary>
        /// <param name="v"> The string to set as the Timestamp for this Message. </param>
        public Message setTimestamp(String v)
        {
            _Timestamp.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Sets the timestamp for this message. The timestamp is set by AMPS
        /// at the time that AMPS processes the message. There is generally no
        /// need to set this field in your application.
        /// </summary>
        /// <param name="buffer"> The buffer that contains the timestamp </param>
        /// <param name="offset"> The location within the buffer where the timestamp begins </param>
        /// <param name="length"> The length of the timestamp </param>
        public Message setTimestamp(byte[] buffer, int offset, int length)
        {
            _Timestamp.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks to see if the Timestamp field is null.
        /// </summary>
        /// <returns> True if the Timestamp field is null, false if not. </returns>
        public bool isTimestampNull()
        {
            return _Timestamp.isNull();
        }

        internal LongField _TopN = new LongField();

        /// <summary>
        /// Get the TopN parameter for this message.
        /// For commands that support limiting the number of messages in the result
        /// set, this parameter provides the limit to AMPS.
        /// </summary>
        /// <returns> The TopN parameter for this Message. </returns>
        public long getTopN()
        {
            return _TopN.getValue();
        }

        /// <summary>
        /// Set the TopN parameter for this message.
        /// For commands that support limiting the number of messages in the result
        /// set, this parameter provides the limit to AMPS.
        /// </summary>
        /// <param name="v"> The value to be set as the TopN parameter. </param>
        public Message setTopN(long v)
        {
            _TopN.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks to see whether the TopN parameter for this message is null.
        /// For commands that support limiting the number of messages in the result
        /// set, this parameter provides the limit to AMPS.
        /// </summary>
        /// <returns> True if the TopN parameter is null, false if not. </returns>
        public bool isTopNNull()
        {
            return _TopN.isNull();
        }

        internal StringField _Topic = new StringField();

        /// <summary>
        /// Return the topic that the message applies to.
        /// </summary>
        /// <returns> The decoded Topic that the Message applies to. </returns>
        public String getTopic()
        {
            return _Topic.getValue(encoder);
        }

        /// <summary>
        /// Return the topic that the message applies to.
        /// </summary>
        /// <remarks>
        /// The Raw variation of this getter allows for finer-grained performance tuning.
        /// It provides direct access to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding
        /// of the byte[] in which the data is stored. The lifetime of the
        /// underlying buffer for the field is the lifetime of the Message.
        /// If you need the data to persist longer than the lifetime of the
        /// Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw Topic that the Message applies to. </returns>
        public Field getTopicRaw()
        {
            return _Topic;
        }

        /// <summary>
        /// Set the topic that the message applies to.
        /// </summary>
        /// <param name="v"> The string to be set as the Topic that the Message applies to. </param>
        /// <returns> Returns this instance so that various operations can be chained together. </returns>
        public Message setTopic(String v)
        {
            _Topic.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Set the topic that the message applies to.
        /// </summary>
        /// <param name="buffer"> The buffer that contains the Topic </param>
        /// <param name="offset"> The location within the buffer where the Topic begins </param>
        /// <param name="length"> The length of the Topic </param>
        /// <returns> Returns this instance so that various operations can be chained together. </returns>
        public Message setTopic(byte[] buffer, int offset, int length)
        {
            _Topic.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks whether the Topic field is null.
        /// </summary>
        /// <returns> true if the Topic field is null, false if not </returns>
        public bool isTopicNull()
        {
            return _Topic.isNull();
        }

        internal LongField _TopicMatches = new LongField();

        /// <summary>
        /// Returns the number of matching topics for the command that this message
        /// was produced in response to.
        /// </summary>
        /// <remarks>
        /// This header is provided on an acknowledgement message that includes
        /// information on the number of records inserted. See the AMPS Command
        /// Reference for details.
        /// </remarks>
        /// <returns> The number of matching topics for the command that this Message was produced in response to. </returns>
        public long getTopicMatches()
        {
            return _TopicMatches.getValue();
        }

        /// <summary>
        /// Sets the topic matches field on this message. This header is typically
        /// provided by AMPS.
        /// </summary>
        /// <param name="v"> The value to set as the TopicMatches field on this Message. </param>
        /// <returns> Returns this instance so that various operations can be chained together. </returns>
        public Message setTopicMatches(long v)
        {
            _TopicMatches.setValue(v);
            return this;
        }

        /// <summary>
        /// Checks to see if the topic matches field on this message is null.
        /// </summary>
        /// <returns> True if the topic matches field is null, false if not. </returns>
        public bool isTopicMatchesNull()
        {
            return _TopicMatches.isNull();
        }

        internal StringField _UserId = new StringField();

        /// <summary>
        /// Get the UserId for this message. This field is typically set during
        /// the logon sequence. The field may also be provided on published
        /// messages, depending on the authentication used in AMPS. When provided
        /// on a published message, AMPS provides the identity of the connection
        /// that sent the message, not the value provided with the message
        /// as published.
        /// </summary>
        /// <returns> The decoded UserId for this message. </returns>
        public String getUserId()
        {
            return _UserId.getValue(decoder);
        }

        /// <summary>
        /// Get the UserId for this message. This field is typically set during
        /// the logon sequence. The field may also be provided on published
        /// messages, depending on the authentication used in AMPS. When provided
        /// on a published message, AMPS provides the identity of the connection
        /// that sent the message, not the value provided with the message
        /// as published.
        /// </summary>
        /// <remarks>
        /// The Raw variation of this getter allows for finer-grained performance tuning.
        /// It provides direct access to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding
        /// of the byte[] in which the data is stored. The lifetime of the
        /// underlying buffer for the field is the lifetime of the Message.
        /// If you need the data to persist longer than the lifetime of the
        /// Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw UserId for this Message. </returns>
        public Field getUserIdRaw()
        {
            return _UserId;
        }

        /// <summary>
        /// Set the UserId for this message. This field is typically set during
        /// the logon sequence. The field may also be provided on published
        /// messages, depending on the authentication used in AMPS. When provided
        /// on a published message, AMPS provides the identity of the connection
        /// that sent the message, not the value provided with the message
        /// as published. Therefore, applications generally only set this
        /// value during logon.
        /// </summary>
        /// <param name="v"> The string to be set as the UserId for this Message. </param>
        /// <returns> Returns this instance so that various operations can be chained together. </returns>
        public Message setUserId(String v)
        {
            _UserId.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Set the UserId for this message. This field is typically set during
        /// the logon sequence. The field may also be provided on published
        /// messages, depending on the authentication used in AMPS. When provided
        /// on a published message, AMPS provides the identity of the connection
        /// that sent the message, not the value provided with the message
        /// as published. Therefore, applications generally only set this
        /// value during logon.
        /// </summary>
        /// <param name="buffer"> The buffer that contains the UserId </param>
        /// <param name="offset"> The position where the UserId begins </param>
        /// <param name="length"> The length of the UserId </param>
        /// <returns> Returns this instance so that various operations can be chained together. </returns>
        public Message setUserId(byte[] buffer, int offset, int length)
        {
            _UserId.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks to see if the UserId field is null.
        /// </summary>
        /// <returns> True if UserId is null, false if not. </returns>
        public bool isUserIdNull()
        {
            return _UserId.isNull();
        }

        public StringField _Version = new StringField();

        /// <summary>
        /// Returns the version of the AMPS server, provided in response to a logon.
        /// </summary>
        /// <returns> The decoded version of the AMPS server. </returns>
        public String getVersion()
        {
            return _Version.getValue(decoder);
        }

        /// <summary>
        /// Returns the version of the AMPS server, provided in response to a logon.
        /// </summary>
        /// <remarks>
        /// The Raw variation of this getter allows for finer-grained performance tuning.
        /// It provides direct access to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding
        /// of the byte[] in which the data is stored. The lifetime of the
        /// underlying buffer for the field is the lifetime of the Message.
        /// If you need the data to persist longer than the lifetime of the
        /// Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw version of the AMPS server. </returns>
        public Field getVersionRaw()
        {
            return _Version;
        }

        /// <summary>
        /// Returns the version of the AMPS server, provided in response to a logon.
        /// </summary>
        /// <returns> The version of AMPS server. </returns>
        public int getVersionAsInt()
        {
            int version = 0;
            try
            {
                version = Client.getVersionAsInt(_Version.getValue(decoder));
            }
            catch (CommandException ) {;}
            if (version == 0)
                version = MINIMUM_SERVER_VERSION;

            return version;
        }

        /// <summary>
        /// Sets the version field of this message. This is typically done by
        /// the AMPS client API when a message with this field set is received.
        /// </summary>
        /// <param name="v"> The string used to set the Version field of this Message. </param>
        /// <returns> Returns this instance so that various operations can be chained together. </returns>
        public Message setVersion(String v)
        {
            _Version.setValue(v, encoder);
            return this;
        }

        /// <summary>
        /// Sets the version field of this message. This is typically done by
        /// the AMPS client API when a message with this field set is received.
        /// </summary>
        /// <param name="buffer"> The buffer that contains the version string </param>
        /// <param name="offset"> The position where the version string begins </param>
        /// <param name="length"> The length of the version string </param>
        /// <returns> Returns this instance so that various operations can be chained together. </returns>
        public Message setVersion(byte[] buffer, int offset, int length)
        {
            _Version.setValue(buffer, offset, length);
            return this;
        }

        /// <summary>
        /// Checks to see if the Version field is null.
        /// </summary>
        /// <returns> True if the Version field is null, false if not. </returns>
        public bool isVersionNull()
        {
            return _Version.isNull();
        }

        internal StringField _CorrelationId = new StringField();

        /// <summary>
        /// Get the CorrelationId on this message. The CorrelationId
        /// is an opaque identifier provided by the publisher of a message.
        /// AMPS provides the identifier to subscribers of the message without
        /// interpreting or changing the identifier.
        /// </summary>
        /// <returns> The decoded version of the CorrelationId on this Message. </returns>
        public String getCorrelationId()
        {
            return _CorrelationId.getValue(decoder);
        }

        /// <summary>
        /// Set the CorrelationId on this message. The CorrelationId
        /// is an opaque identifier provided by the publisher of a message.
        /// AMPS provides the identifier to subscribers of the message without
        /// interpreting or changing the identifier.
        /// </summary>
        /// <param name="correlationId_"> The string used to set the CorrelationId on this message. </param>
        /// <returns> Returns this instance so that various operations can be chained together. </returns>
        public Message setCorrelationId(String correlationId_)
        {
            _CorrelationId.setValue(correlationId_, encoder);
            return this;
        }

        /// <summary>
        /// Get the CorrelationId on this message. The CorrelationId
        /// is an opaque identifier provided by the publisher of a message.
        /// AMPS provides the identifier to subscribers of the message without
        /// interpreting or changing the identifier.
        /// </summary>
        /// <remarks>
        /// The Raw variation of this getter allows for finer-grained performance tuning.
        /// It provides direct access to the Field that encapsulates the data.
        /// This does not incur any additional allocation or unnecessary decoding
        /// of the byte[] in which the data is stored. The lifetime of the
        /// underlying buffer for the field is the lifetime of the Message.
        /// If you need the data to persist longer than the lifetime of the
        /// Message, you must copy the Field.
        /// </remarks>
        /// <returns> The raw CorrelationId on this Message. </returns>
        public StringField getCorrelationIdRaw()
        {
            return _CorrelationId;
        }

        /// <summary>
        /// Checks to see if the CorrelationId field is null.
        /// </summary>
        /// <returns> true if null, false if not. </returns>
        public bool isCorrelationIdNull()
        {
            return _CorrelationId.isNull();
        }

        /// <summary>
        /// Set by the server messages on messages received
        /// from a queue and indicates the point in time at
        /// which the server will break the lease on the
        /// message if the message is still leased to the
        /// subscription. During this time frame, the subscriber
        /// is granted exclusive access to the message, allowing
        /// it to be processed or acknowledged. If the message
        /// is not acknowledged within the LeasePeriod, AMPS may
        /// take predefined actions, such as re-delivery or error
        /// handling, to ensure message reliability and resource
        /// efficiency.
        /// </summary>
        internal StringField _LeasePeriod = new StringField();
        /// <summary>
        /// Returns true if the LeasePeriod field is null
        /// </summary>
        public bool isLeasePeriodNull()
        {
            return _LeasePeriod.isNull();
        }
        /// <summary>
        /// Returns the value of the LeasePeriod field
        /// </summary>
        public String getLeasePeriod()
        {
            return _LeasePeriod.getValue(decoder);
        }

        /// <summary>
        /// Returns the raw value of the LeasePeriod field
        /// </summary>
        /// <returns>A <see cref="StringField"/> pointing to the raw message's LeasePeriod.</returns>
        public StringField getLeasePeriodRaw()
        {
            return _LeasePeriod;
        }
        /// <summary>
        /// Returns the value of the LeasePeriod field
        /// </summary>
        public String LeasePeriod
        {
            get { return _LeasePeriod.getValue(decoder);  }
        }
        public override string ToString()
        {
            String tmp = "";
            StringBuilder s = new StringBuilder(1024);
            s.append("Message{");

            s.append("Command=").append(CommandField.encodeCommand(getCommand())).append("; ");
            try
            {
                // IDEA:
                //    Use try/catch coupled with a string to fail absolutely and not
                //    append anything if the Field throws a NullPointerEx when evaluating
                //    if the buffer is null.
                tmp = "AckType=" + AckTypeField.encodeAckType(getAckType());
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}
            try
            {
                tmp = "BatchSize=" + getBatchSize();
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            try
            {
                tmp = "Bookmark=" + getBookmark();
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            if (getClientName() != null)
                s.append("ClientName=").append(getClientName()).append("; ");

            if (getCommandId() != null)
                s.append("CommandId=").append(getCommandId()).append("; ");

            try
            {
                tmp = "CorrelationId=" + getCorrelationId();
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            try
            {
                tmp = "Expiration=" + getExpiration();
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            if (getFilter() != null)
                s.append("Filter=").append(getFilter()).append("; ");

            try
            {
                tmp = "GroupSeqNo=" + Long.toString(getGroupSeqNo());
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            try
            {
                tmp = "LeasePeriod=" + getLeasePeriod();
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            try
            {
                tmp = "Length=" + Integer.toString(getLength());
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            try
            {
                tmp = "Matches=" + Long.toString(getMatches());
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            if ( getMessageType() != null)
                s.append("MessageType=").append(getMessageType()).append("; ");

            try
            {
                Field tmpField = getOptionsRaw();
                tmp = "Options=" + new string(Encoding.ASCII.GetChars(tmpField.buffer, tmpField.position, tmpField.length), tmpField.position, tmpField.length);
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}
            catch (ArgumentException) { ; }

            if ( getOrderBy() != null)
                s.append("OrderBy=").append(getOrderBy()).append("; ");

            if (getPassword() != null)
                s.append("Password=").append(getPassword()).append("; ");

            if (getQueryId() != null)
                s.append("QueryId=").append(getQueryId()).append("; ");

            try
            {
                tmp = "Reason=" + ReasonField.encodeReason(getReason());
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            try
            {
                tmp = "RecordsInserted=" + Long.toString(getRecordsInserted());
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            try
            {
                tmp = "RecordsUpdated=" + Long.toString(getRecordsUpdated());
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            try
            {
                tmp = "RecordsReturned=" + Long.toString(getRecordsReturned());
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            try
            {
                tmp = "RecordsDeleted=" + Long.toString(getRecordsDeleted());
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            try
            {
                tmp = "Sequence=" + getSequence();
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            if (getSowKey() != null)
                s.append("SowKey=").append(getSowKey()).append("; ");

            if (getSowKeys() != null)
                s.append("SowKeys=").append(getSowKeys()).append("; ");

            try
            {
                tmp = "Status=" + StatusField.encodeStatus(getStatus());
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            if (getSubId() != null)
                s.append("SubId=").append(getSubId()).append("; ");

            if (getSubIds() != null)
                s.append("SubIds=").append(getSubIds()).append("; ");

            if (getTimestamp() != null)
                s.append("Timestamp=").append(getTimestamp()).append("; ");

            try
            {
                tmp = "TopN=" + Long.toString(getTopN());
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            if (getTopic() != null)
                s.append("Topic=").append(getTopic()).append("; ");

            try
            {
                tmp = "TopicMatches=" + Long.toString(getTopicMatches());
                s.append(tmp + "; ");
            }
            catch (NullReferenceException) { ;}

            if (getUserId() != null)
                s.append("UserId=").append(getUserId().toString()).append("; ");

            if (getData() != null)
                s.append("Data=").append(getData()).append("; ");

            s.append("}");

            return s.toString();
        }
        public abstract Message copy();
        private byte[] buffer = null;

        /// <Summary>
        ///  Sets the byte buffer used by the message. It's not common for
        ///  applications to use this directly. </summary>
        ///  <param name="buffer"> The byte buffer to use for the message. </param>
        public void setBuffer(byte[] buffer)
        {
            this.buffer = buffer;
        }

        /// <summary>
        /// Gets the byte buffer used by the message.
        /// </summary>
        /// <returns> the raw byte buffer backing this message object. </returns>
        public byte[] getBuffer()
        {
            return this.buffer;
        }

        private int _rawBufferOffset = 0;

        /// <summary>
        /// Sets the offset into the byte buffer where this message begins.
        /// </summary>
        /// <param name= "offset"> offset of first byte of message within the buffer. </param>>
        public void setRawBufferOffset(int offset)
        {
            this._rawBufferOffset = offset;
        }

        /// <summary>
        /// Gets the offset into the byte buffer where this message begins.
        /// </summary>
        /// <returns> offset of first byte of message within the buffer. </returns>
        public int getRawBufferOffset()
        {
            return this._rawBufferOffset;
        }

        private int _rawBufferLength = 0;

        /// <summary>
        /// Sets the length of the message within the byte buffer.
        /// </summary>
        /// <param name="length"> of message within the buffer. </param)
        public void setRawBufferLength(int length)
        {
            this._rawBufferLength = length;
        }

        /// <summary>
        /// Gets the length of the message within the byte buffer.
        /// </summary>
        /// <returns> length of message within the buffer. </returns>
        public int getRawBufferLength()
        {
            return this._rawBufferLength;
        }

        internal Client _client = null;

        /// <summary>
        /// Returns the Client who allocated this message.
        /// </summary>
        /// <returns>The <see cref="Client"/>.</returns>
        public Client getClient()
        {
            return _client;
        }

        /// <summary>
        /// The <see cref="Client"/> who allocated this message.
        /// </summary>
        public Client Client
        {
            get { return _client; }
        }

        /// <summary>
        /// Acks the message to the server if it was a Queue message.
        /// </summary>
        /// <param name="options"> The optional options to use for the ack, such as "cancel". </param>
        public void ack(String options = null)
        {
            if (_client != null && !_LeasePeriod.isNull() &&
                (!_client.getAutoAck() || _ignoreAutoAck))
            {
                if (options == null)
                    _client._ack(_Topic, _Bookmark);
                else
                    _client._ack(_Topic, _Bookmark, new Field(options));
            }
        }

        /// <summary>
        /// Acks the message to the server if it was a Queue message.
        /// </summary>
        /// <param name="options"> The optional options to use for the ack, such as "cancel". </param>
        public void ack(Field options)
        {
            if (_client != null && !_LeasePeriod.isNull() &&
                (!_client.getAutoAck() || _ignoreAutoAck))
            {
                _client._ack(_Topic, _Bookmark, options);
            }
        }

        /// <summary>
        /// Used to check if automatic acking of this Message by the Client should be skipped.
        /// </summary>
        /// <returns> If automatic acking is ignored for this Message. </returns>
        public bool isIgnoreAutoAck()
        {
            return _ignoreAutoAck;
        }

        /// <summary>
        /// Used to disable automatic acking of this Message by the Client after the
        /// (check this again, wether the path of the link is correct or not)
        /// <see cref="AMPS.Client.MessageHandler"/> completes.
        /// </summary>
        public void setIgnoreAutoAck()
        {
            _ignoreAutoAck = true;
        }

        internal const int Ready = 0;
        internal const int Update = 1;
        internal const int Frozen = 2;
        internal int _updateState = 0;


        /// <summary>
        /// Copies the content of the current message to the specified destination message.
        /// This method is used to create a copy of the current message and transfer its
        /// content to a destination message. It handles copying various fields and properties,
        /// including a buffer containing raw data and other relevant AMPS-specific attributes.
        /// </summary>
        /// <param name= "destination"> The message to copy the content to. </param>
        internal void _copyTo(Message destination)
        {
            destination._ignoreAutoAck = _ignoreAutoAck;
            if (this.buffer != null)
            {
                ///<summary>
                /// If buffer is not null, create a new buffer in the destination and copy content
                ///</summary>
                destination.buffer = new byte[_rawBufferLength];
                destination._rawBufferLength = _rawBufferLength;
                destination._rawBufferOffset = 0;
                Buffer.BlockCopy(buffer, _rawBufferOffset, destination.buffer, 0, _rawBufferLength);
                destination._AckType.set(destination.buffer, _AckType.position-_rawBufferOffset, _AckType.length);
                destination._BatchSize.set(destination.buffer, _BatchSize.position-_rawBufferOffset, _BatchSize.length);
                destination._Bookmark.set(destination.buffer, _Bookmark.position-_rawBufferOffset, _Bookmark.length);
                destination._ClientName.set(destination.buffer, _ClientName.position-_rawBufferOffset, _ClientName.length);
                destination._Command.set(destination.buffer, _Command.position-_rawBufferOffset, _Command.length);
                destination._CommandId.set(destination.buffer, _CommandId.position-_rawBufferOffset, _CommandId.length);
                destination._CorrelationId.set(destination.buffer, _CorrelationId.position-_rawBufferOffset, _CorrelationId.length);
                destination._Data.set(destination.buffer, _Data.position-_rawBufferOffset, _Data.length);
                destination._Expiration.set(destination.buffer, _Expiration.position-_rawBufferOffset, _Expiration.length);
                destination._Filter.set(destination.buffer, _Filter.position-_rawBufferOffset, _Filter.length);
                destination._OrderBy.set(destination.buffer, _OrderBy.position-_rawBufferOffset, _OrderBy.length);
                destination._GroupSeqNo.set(destination.buffer, _GroupSeqNo.position-_rawBufferOffset, _GroupSeqNo.length);
                destination._Length.set(destination.buffer, _Length.position-_rawBufferOffset, _Length.length);
                destination._LeasePeriod.set(destination.buffer, _LeasePeriod.position-_rawBufferOffset, _LeasePeriod.length);
                destination._Matches.set(destination.buffer, _Matches.position-_rawBufferOffset, _Matches.length);
                destination._MessageType.set(destination.buffer, _MessageType.position-_rawBufferOffset, _MessageType.length);
                destination._Options.set(destination.buffer, _Options.position-_rawBufferOffset, _Options.length);
                destination._Password.set(destination.buffer, _Password.position-_rawBufferOffset, _Password.length);
                destination._QueryId.set(destination.buffer, _QueryId.position-_rawBufferOffset, _QueryId.length);
                destination._Reason.set(destination.buffer, _Reason.position-_rawBufferOffset, _Reason.length);
                destination._RecordsInserted.set(destination.buffer, _RecordsInserted.position-_rawBufferOffset, _RecordsInserted.length);
                destination._RecordsUpdated.set(destination.buffer, _RecordsUpdated.position-_rawBufferOffset, _RecordsUpdated.length);
                destination._RecordsDeleted.set(destination.buffer, _RecordsDeleted.position-_rawBufferOffset, _RecordsDeleted.length);
                destination._RecordsReturned.set(destination.buffer, _RecordsReturned.position-_rawBufferOffset, _RecordsReturned.length);
                destination._Sequence.set(destination.buffer, _Sequence.position-_rawBufferOffset, _Sequence.length);
                destination._SowKey.set(destination.buffer, _SowKey.position-_rawBufferOffset, _SowKey.length);
                destination._SowKeys.set(destination.buffer, _SowKeys.position-_rawBufferOffset, _SowKeys.length);
                destination._Status.set(destination.buffer, _Status.position-_rawBufferOffset, _Status.length);
                destination._SubId.set(destination.buffer, _SubId.position-_rawBufferOffset, _SubId.length);
                destination._SubIds.set(destination.buffer, _SubIds.position-_rawBufferOffset, _SubIds.length);
                destination._Timestamp.set(destination.buffer, _Timestamp.position-_rawBufferOffset, _Timestamp.length);
                destination._Topic.set(destination.buffer, _Topic.position-_rawBufferOffset, _Topic.length);
                destination._TopicMatches.set(destination.buffer, _TopicMatches.position-_rawBufferOffset, _TopicMatches.length);
                destination._TopN.set(destination.buffer, _TopN.position-_rawBufferOffset, _TopN.length);
                destination._UserId.set(destination.buffer, _UserId.position-_rawBufferOffset, _UserId.length);
                destination._Version.set(destination.buffer, _Version.position-_rawBufferOffset, _Version.length);
            }
            else // Copy each Field
            {
                destination.buffer = null;
                destination._rawBufferLength = 0;
                destination._rawBufferOffset = 0;
                Field f = _AckType.copy();
                destination._AckType.set(f.buffer, f.position, f.length);
                f = _BatchSize.copy();
                destination._BatchSize.set(f.buffer, f.position, f.length);
                f = _Bookmark.copy();
                destination._Bookmark.set(f.buffer, f.position, f.length);
                f = _ClientName.copy();
                destination._ClientName.set(f.buffer, f.position, f.length);
                f = _Command.copy();
                destination._Command.set(f.buffer, f.position, f.length);
                f = _CommandId.copy();
                destination._CommandId.set(f.buffer, f.position, f.length);
                f = _CorrelationId.copy();
                destination._CorrelationId.set(f.buffer, f.position, f.length);
                f = _Data.copy();
                destination._Data.set(f.buffer, f.position, f.length);
                f = _Expiration.copy();
                destination._Expiration.set(f.buffer, f.position, f.length);
                f = _Filter.copy();
                destination._Filter.set(f.buffer, f.position, f.length);
                f = _OrderBy.copy();
                destination._OrderBy.set(f.buffer, f.position, f.length);
                f = _GroupSeqNo.copy();
                destination._GroupSeqNo.set(f.buffer, f.position, f.length);
                f = _Length.copy();
                destination._Length.set(f.buffer, f.position, f.length);
                f = _LeasePeriod.copy();
                destination._LeasePeriod.set(f.buffer, f.position, f.length);
                f = _Matches.copy();
                destination._Matches.set(f.buffer, f.position, f.length);
                f = _MessageType.copy();
                destination._MessageType.set(f.buffer, f.position, f.length);
                f = _Options.copy();
                destination._Options.set(f.buffer, f.position, f.length);
                f = _Password.copy();
                destination._Password.set(f.buffer, f.position, f.length);
                f = _QueryId.copy();
                destination._QueryId.set(f.buffer, f.position, f.length);
                f = _Reason.copy();
                destination._Reason.set(f.buffer, f.position, f.length);
                f = _RecordsInserted.copy();
                destination._RecordsInserted.set(f.buffer, f.position, f.length);
                f = _RecordsUpdated.copy();
                destination._RecordsUpdated.set(f.buffer, f.position, f.length);
                f = _RecordsDeleted.copy();
                destination._RecordsDeleted.set(f.buffer, f.position, f.length);
                f = _RecordsReturned.copy();
                destination._RecordsReturned.set(f.buffer, f.position, f.length);
                f = _Sequence.copy();
                destination._Sequence.set(f.buffer, f.position, f.length);
                f = _SowKey.copy();
                destination._SowKey.set(f.buffer, f.position, f.length);
                f = _SowKeys.copy();
                destination._SowKeys.set(f.buffer, f.position, f.length);
                f = _Status.copy();
                destination._Status.set(f.buffer, f.position, f.length);
                f = _SubId.copy();
                destination._SubId.set(f.buffer, f.position, f.length);
                f = _SubIds.copy();
                destination._SubIds.set(f.buffer, f.position, f.length);
                f = _Timestamp.copy();
                destination._Timestamp.set(f.buffer, f.position, f.length);
                f = _Topic.copy();
                destination._Topic.set(f.buffer, f.position, f.length);
                f = _TopicMatches.copy();
                destination._TopicMatches.set(f.buffer, f.position, f.length);
                f = _TopN.copy();
                destination._TopN.set(f.buffer, f.position, f.length);
                f = _UserId.copy();
                destination._UserId.set(f.buffer, f.position, f.length);
                f = _Version.copy();
                destination._Version.set(f.buffer, f.position, f.length);
            }
            destination._bookmarkSeqNo = _bookmarkSeqNo;
            destination._subscription = _subscription;
            destination._client = _client;
        }
    }
}
