////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Diagnostics;
using System.Collections.Concurrent;
using System.Collections.Generic;

namespace AMPS.Client
{
/// <summary>
/// MessageRouter is used to register and manage a list of handler objects for messages,
/// and then to route messages to those handlers as messages arrive. MessageRouter also
/// "knows" about the meaning of AMPS acks and can use them to automatically clean up
/// routes as acks arrive.
/// </summary>
public class MessageRouter
{
    /// <summary>
    /// Adds a route to self (MessageRouter).
    /// </summary>
    /// <param name= "commandId_"> The command, query, or subid used for this route. </param>
    /// <param name= "messageHandler_"> The message handler to route to. </param>
    /// <param name= "requestedAcks_"> The acks requested by the AMPS client infrastructure. </param>
    /// <param name= "systemAcks_"> The acks requested by AMPS client infrastructure.
    /// <br> These will not be delivered to the message handler, but are still processed for
    /// auto-removal. </br></param>
    /// <param name= "isSubscribe_"> True if this route is for an ongoing subscription. </param>
    public void addRoute(CommandId commandId_, MessageHandler messageHandler_,
                         Message.AckTypes requestedAcks_,
                         Message.AckTypes systemAcks_, bool isSubscribe_)
    {
        MessageRoute newRoute = new MessageRoute(messageHandler_, (int)requestedAcks_,
                                                 (int)systemAcks_, isSubscribe_);
        _routes.AddOrUpdate(commandId_, newRoute, (key,oldValue)=>newRoute);
    }

    /// <summary>
    /// Remove a route from MessageRouter.
    /// </summary>
    /// <param name= "commandId_"> The route to remove. </param>
    /// <returns> True if the route was removed. </returns>
    public bool removeRoute(CommandId commandId_)
    {
        MessageRoute route = null;
        _routes.TryRemove(commandId_, out route);
        return route != null;
    }

    /// <summary>
    /// Return whether or not a command id has a route.
    /// </summary>
    /// <param name= "commandId_"> The command id for this route. </param>
    /// <returns> Whether or not the route exists. </returns>
    public bool hasRoute(CommandId commandId_)
    {
        MessageRoute route = null; _routes.TryGetValue(commandId_, out route);
        return (route != null);
    }

    /// <summary>
    /// Find and return a route.
    /// </summary>
    /// <param name="commandId_"> The command id for this route. </param>
    /// <returns> The MessageHandler registered for this route, or null if none is registered. </returns>
    public MessageHandler findRoute(CommandId commandId_)
    {
        MessageRoute route = null; _routes.TryGetValue(commandId_, out route);
        if(route != null)
            return route.getMessageHandler();
        else
            return null;
        }

    /// <summary>
    /// Removes all routes from self (MessageRouter).
    /// </summary>
    public void clear()
    {
        _routes.Clear();
    }

    /// <summary>
    /// Removes all subscriptions.
    /// </summary>
    public void unsubscribeAll()
    {
        List<CommandId> toRemove = new List<CommandId>(_routes.Count);
        foreach (var pair in _routes)
        {
            if (pair.Value.isTerminationAck(0))
            {
                toRemove.Add(pair.Key);
            }
        }
        foreach (var id in toRemove)
        {
            removeRoute(id);
        }
    }

    /// <summary>
    /// Deliver a message that is known already to be an Ack. Coordinates the
    /// removal of routes based on the ack received and the original message type.
    /// </summary>
    /// <param name="ackMessage_"> The Message to deliver. </param>
    /// <param name="ackType_"> The ack type from that message. </param>
    /// <returns> The number of message deliveries that occurred. </returns>
    public int deliverAck(Message ackMessage_, Message.AckTypes ackType_)
    {
        Debug.Assert(ackMessage_.getCommand() == Message.Commands.Ack);
        Debug.Assert(ackType_ != (int)Message.AckTypes.None);
        int messagesDelivered = 0;
        if(ackMessage_.getCommandId(_key))
        {
            messagesDelivered += _deliverAck(ackMessage_, (int)ackType_, _key);
        }
        if(ackMessage_.getQueryId(_key))
        {
            if(messagesDelivered==0)
            {
                messagesDelivered += _deliverAck(ackMessage_, (int)ackType_, _key);
            } else {
                _processAckForRemoval((int)ackType_, _key);
            }
        }
        if(ackMessage_.getSubId(_key))
        {
            if(messagesDelivered == 0)
            {
                messagesDelivered += _deliverAck(ackMessage_, (int)ackType_, _key);
            } else {
                _processAckForRemoval((int)ackType_, _key);
            }
        }
        return messagesDelivered;
    }

    /// <summary>
    /// Delivers a data message (not an Ack) to the registered route.
    /// Uses the commandID, subID, and queryID to deliver find a route and
    /// deliver to the first one found. This method is optimized for speed
    /// and does not attempt to examine ack types for removal of routes.
    /// </summary>
    /// <param name="dataMessage_"> The non-ack message to deliver. </param>
    /// <returns> The number of deliveries performed. </returns>
    public int deliverData(Message dataMessage_)
    {
        Debug.Assert(dataMessage_.getCommand() != Message.Commands.Ack);
        int messagesDelivered = 0;
        if(dataMessage_.getQueryId(_key))
        {
            messagesDelivered += deliverData(dataMessage_, _key);
        }
        if(messagesDelivered==0 && dataMessage_.getCommandId(_key))
        {
            messagesDelivered += deliverData(dataMessage_, _key);
        }
        if(messagesDelivered==0 && dataMessage_.getSubId(_key))
        {
            messagesDelivered += deliverData(dataMessage_, _key);
        }
        return messagesDelivered;
    }

    /// <summary>
    /// Delivers a data message(not an Ack) using a specific command ID from the message.
    /// Optimized for speed and does not attempt to examine the message for auto-removal of routes
    /// </summary>
    /// <param name="dataMessage_"> The message to deliver. </param>
    /// <param name="commandId_"> The command ID which will be used to lookup the delivery route. </param>
    /// <returns> The number of deliveries performed. </returns>
    public int deliverData(Message dataMessage_, CommandId commandId_)
    {
        Debug.Assert(dataMessage_.getCommand() != Message.Commands.Ack);
        int messagesDelivered = 0;
        MessageRoute route = null; _routes.TryGetValue(commandId_, out route);
        if(route!=null)
        {
            messagesDelivered += route.deliverData(dataMessage_);
        }
        return messagesDelivered;
    }
    //////////////// INTERNAL METHODS AND CLASSES ///////////////////

    // MessageRoute is an internal class representing a route in the MessageRouter.
    // It holds information about the associated message handler, acks, and termination
    // acks.
    private class MessageRoute
    {
        MessageHandler _messageHandler;
        int _requestedAcks, _terminationAck = 0;
        public MessageRoute(MessageHandler messageHandler_, int requestedAcks_,
            int systemAcks_, bool isSubscribe_)
        {
            _messageHandler = messageHandler_;
            _requestedAcks = requestedAcks_;
            // For non-subscriptions we autoremove a handler when the termination ack comes in.
            if(!isSubscribe_)
            {
                if (messageHandler_ is MessageStream)
                {
                    // If the message handler is a MessageStream, then
                    // we delegate all handling of automatic route removal
                    // to the MessageStream for completed non-subscription
                    // commands. We do this by specifying no termination ack
                    // here so the message router doesn't try to remove it.
                    _terminationAck = (int) Message.AckTypes.None;
                }
                else // Otherwise the message router handles route removal...
                {
                    // compute the termination ack we're looking for
                    // make terminationAck the highest bit set in all acks
                    int bitCounter = requestedAcks_ | systemAcks_;
                    while(bitCounter > 0) { bitCounter>>=1; _terminationAck = _terminationAck>0?2*_terminationAck:1; }
                }
            }
        }

        public int deliverAck(Message message_, int ackType_)
        {
            // If it was not a requested ack, do not deliver.
            if( (_requestedAcks & ackType_) == 0) return 0;
            _messageHandler.invoke(message_);
            return 1;
        }

        /// <summary>
        /// Checks if the given acknowledgment type is a termination acknowledgment.
        /// </summary>
        /// <param name="ackType_">The acknowledgment type to check.</param>
        /// <returns>True if the acknowledgment type is a termination acknowledgment; otherwise, false.</returns>
        public bool isTerminationAck(int ackType_)
        {
            return ackType_ == _terminationAck;
        }

        /// <summary>
        /// Delivers a data message to the associated message handler.
        /// </summary>
        /// <param name="message_">The data message.</param>
        /// <returns>The number of messages delivered (always 1).</returns>
        public int deliverData(Message message_)
        {
            _messageHandler.invoke(message_);
            return 1;
        }

        /// <summary>
        /// Gets the associated message handler.
        /// </summary>
        /// <returns>The associated message handler.</returns>
        public MessageHandler getMessageHandler()
        {
            return _messageHandler;
        }
    }

    // Delivers an acknowledgment message to the appropriate route.
    // param name="ackMessage_": The acknowledgment message.
    // param name="ackType_": The type of acknowledgment.
    // param name="commandId_": The command ID associated with the acknowledgment.
    // returns: The number of messages delivered.
    private int _deliverAck(Message ackMessage_, int ackType_, CommandId commandId_)
    {
        int messagesDelivered = 0;
        MessageRoute route = null; _routes.TryGetValue(commandId_, out route);
        if(route!=null)
        {
            messagesDelivered += route.deliverAck(ackMessage_, ackType_);
            if(route.isTerminationAck(ackType_))
            {
                _routes.TryRemove(commandId_, out route);
            }
        }
        return messagesDelivered;
    }

    // Internal method to process an acknowledgement for route removal of routes.
    // Checks if the ack type matches the termination ack and remove the associated
    // route.
    // param name="ackType_":The type of acknowledgment.
    // param name="commandId_":The command ID associated with the acknowledgment.
    private void _processAckForRemoval(int ackType_, CommandId commandId_)
    {
        MessageRoute route = null;
        if (_routes.TryGetValue(commandId_, out route) && route.isTerminationAck(ackType_))
        {
            _routes.TryRemove(commandId_, out route);
        }
    }

    /////////// MEMBER VARIABLES FOR MessageRouter /////////////
    // Storage for current routes.
    ConcurrentDictionary<CommandId, MessageRoute> _routes = new ConcurrentDictionary<CommandId, MessageRouter.MessageRoute>();

    // Cached CommandId object used for retrieval of command Ids.
    CommandId _key = new CommandId();

}
} // namespace AMPS.Client
