////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2012-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.Concurrent;
using System.Threading;
using AMPS.Client.Exceptions;
using AMPS.Client.Fields;

namespace AMPS.Client
{
    /// <summary>
    /// MessageStream provides an iteration abstraction over the results of an AMPS
    /// command such as a subscribe, a SOW query, or SOW delete. MessageStream is produced when
    /// calling Client.execute() and continues iterating over the results until the connection
    /// is closed, or the iterator is explicitly closed, or when the SOW query is ended.
    /// You can use a MessageStream as you would other Iterators, for example, using a for loop (Java 1.7):
    /// </summary>
    /// <pre>
    /// MessageStream stream = Client.execute(new Command("sow").setTopic("/orders")).timeout(1000);
    /// foreach(Message message in stream)
    /// {
    ///   ...
    /// }
    /// </pre>
    public class MessageStream : IEnumerator<Message>, IEnumerable<Message>, MessageHandler
    {
        private Object internalLock = new Object();
        EventHandler<Client.ConnectionStateChangedEventArgs> _eventHandler = null;
        CommandId _commandId;
        CommandId _queryId;
        Client _client;
        ConcurrentQueue<Message> _q = new ConcurrentQueue<Message>();
        volatile ConcurrentDictionary<string, Message> _sowKeyMap = null;
        Message _current = null;
        enum State { Unset = 0, Reading = 0x10, Subscribed = 0x12, SOWOnly = 0x14, StatsOnly=0x18, AcksOnly=0x18, Disconnected = 0x01, Complete = 0x02 };
        volatile State _state = State.Unset;
        int _timeout = 0;
        int _maxQueueDepth = Int32.MaxValue;
        internal Message.AckTypes _requestedAcks = Message.AckTypes.None;
        internal Field _previousTopic = new Field();
        internal Field _previousBookmark = new Field();
        private static MessageStream _emptyMessageStream = new MessageStream();

        private MessageStream()
        {
            _client = null;
            _commandId = null;
            _queryId = null;
            _state = State.Complete;
            _requestedAcks = Message.AckTypes.None;
        }


        /// <summary>
        /// Method to retrieve the empty MessageStream.
        /// </summary>
        /// <returns> The static, empty MessageStream </returns>
        internal static MessageStream getEmptyMessageStream()
        {
            return _emptyMessageStream;
        }

        /// <summary>
        /// Constructs an instance of message stream with the given client. The
        /// message stream will hold a reference to the provided client, and
        /// registers itself as a connection state listener for the client.
        /// </summary>
        /// <param name="client"> The client to use for this message stream. </param>
        internal MessageStream(Client client)
        {
            _client = client;
            _eventHandler = new EventHandler<Client.ConnectionStateChangedEventArgs>(_client_ConnectionStateChanged);
            _client.ConnectionStateChanged += _eventHandler;
        }

        /// <summary>
        /// When this message stream is the result of a subscription command, this gets
        /// called with the command ID.
        /// </summary>
        /// <param name= "commandId_"> the CommandId to set as the commandId of this command. </param>
        internal void setSubscription(CommandId commandId_)
        {
            _commandId = commandId_;

            lock (internalLock)
            {
                if (_state != State.Disconnected) _state = State.Subscribed;
            }
        }

        /// <summary>
        /// When this message stream is the result of a form of SOW command, this gets
        /// called with the query ID of the command.
        /// </summary>
        /// <param name="queryId_"> The CommandId to set as the QueryId of this command. </param>
        internal void setQueryId(CommandId queryId_)
        {
            if (queryId_ != null &&
                (_commandId == null || !_commandId.Equals(queryId_)))
            {
                _queryId = queryId_;
            }
        }

        /// <summary>
        /// This function is used to set the internal state to SOWOnly.
        /// It should be called in a thread-safe manner using the internal lock to avoid
        /// race conditions. The state is only updated if the current state is not Disconnected.
        /// </summary>
        /// <seealso cref="State"/>
        /// <seealso cref="internalLock"/>
        /// <seealso cref="State.Disconnected"/>
        /// <seealso cref="State.SOWOnly"/>
        /// <example>
        /// <code>
        /// // Acquire the internal lock before calling setSOWOnly
        /// lock(internalLock);
        /// setSOWOnly();
        /// // Release the internal lock after calling setSOWOnly
        /// unlock(internalLock);
        /// </code>
        /// </example>
        internal void setSOWOnly()
        {
            lock(internalLock)
            {
                if (_state != State.Disconnected) _state = State.SOWOnly;
            }
        }

        /// <summary>
        /// This function is used to set the internal state of the AMPS to StatsOnly and configure the requested acknowledgments.
        /// It should be called in a thread-safe manner using the internal lock to avoid race conditions.
        /// The state is only updated if the current state is not Disconnected.
        /// </summary>
        /// <seealso cref="State"/>
        /// <seealso cref="internalLock"/>
        /// <seealso cref="State.Disconnected"/>
        /// <seealso cref="State.StatsOnly"/>
        /// <seealso cref="Message.AckTypes.Stats"/>
        /// <example>
        /// <code>
        /// // Acquire the internal lock before calling setStatsOnly
        /// lock(internalLock);
        /// setStatsOnly();
        /// // Release the internal lock after calling setStatsOnly
        /// unlock(internalLock);
        /// </code>
        /// </example>
        internal void setStatsOnly()
        {
            lock(internalLock)
            {
                if (_state != State.Disconnected)
                {
                    _state = State.StatsOnly;
                    _requestedAcks = Message.AckTypes.Stats;
                }
            }
        }

        /// <summary>
        /// Sets the Advanced Message Processing System (AMPS) state to AcksOnly and configures the command ID and requested acknowledgments.
        /// </summary>
        /// <param name="commandId_"> The command ID to be set. </param>
        /// <param name="ackTypes_"> The requested acknowledgment types to be set. </param>
        /// <remarks>
        /// This function is used to set the internal state of the AMPS to AcksOnly and configure the command ID and requested acknowledgments.
        /// It should be called in a thread-safe manner using the internal lock to avoid race conditions.
        /// The state is only updated if the current state is not Disconnected.
        /// </remarks>
        /// <seealso cref="AMPS.Client.MessageStream.State"/>
        /// <seealso cref="internalLock"/>
        /// <seealso cref="AMPS.Client.MessageStream.State.Disconnected"/>
        /// <seealso cref="AMPS.Client.MessageStream.State.AcksOnly"/>
        /// <seealso cref="Message.AckTypes"/>
        /// <example>
        /// <code>
        /// // Acquire the internal lock before calling setAcksOnly
        /// lock(internalLock);
        /// setAcksOnly(commandId, Message.AckTypes.SomeType);
        /// // Release the internal lock after calling setAcksOnly
        /// unlock(internalLock);
        /// </code>
        /// </example>
        internal void setAcksOnly(CommandId commandId_, Message.AckTypes ackTypes_)
        {
            lock(internalLock)
            {
                if (_state != State.Disconnected)
                {
                    _state = State.AcksOnly;
                    _commandId = commandId_;
                    _requestedAcks = ackTypes_;
                }
            }
        }

        /// <sumaary>
        /// This function is used to set the internal state of the AMPS to Reading.
        /// It should be called in a thread-safe manner using the internal lock to
        /// <br> avoid race conditions. </br>
        /// The state is only updated if the current state is not Disconnected.
        /// </summary>
        /// <seealso cref="AMPS.Client.MessageStream.State"/>
        /// <seealso cref="internalLock"/>
        /// <seealso cref="AMPS.Client.MessageStream.State.Disconnected"/>
        /// <seealso cref="AMPS.Client.MessageStream.State.Reading"/>
        internal void setRunning()
        {
            lock(internalLock)
            {
                if (_state != State.Disconnected) _state = State.Reading;
            }
        }

        /// <summary>
        /// Sets a timeout on this message stream. If no message is received with this timeout,
        /// forward iteration returns null for the next message.
        /// </summary>
        /// <param name="timeout_">The timeout value in milliseconds.</param>
        /// <returns>This instance.</returns>
        public MessageStream timeout(int timeout_)
        {
            _timeout = timeout_;
            return this;
        }

        /// <summary>
        /// Causes messages on a SOW topic to be conflated by their SOW key, so that when messages are consumed
        /// the latest value for each distinct key is returned instead of every message.
        /// </summary>
        /// <returns>This instance</returns>
        public MessageStream conflate()
        {
            if(_sowKeyMap == null) _sowKeyMap = new ConcurrentDictionary<string, Message>();
            return this;
        }

        /// <summary>
        /// Bounds the internal queue of this stream to a specified count of messages.
        /// if the consumer of this stream does not process messages in a timely fashion and
        /// this internal queue limit is reached, the AMPS client will not respond to incoming
        /// messages including acknowledgements and may become disconnected.
        /// </summary>
        /// <param name="maxQueueDepth_"> The internal queue depth limit (default: unlimited) </param>
        /// <returns>This instance</returns>
        public MessageStream maxDepth(int maxQueueDepth_)
        {
            _maxQueueDepth = maxQueueDepth_;
            return this;
        }

        /// <summary>
        /// Returns the current maximum queue depth before pushing back on the receive thread.
        /// </summary>
        /// <returns>The internal queue depth limit, default is In32.MaxValue. </returns>
        public int getMaxDepth()
        {
            return _maxQueueDepth;
        }

        /// <summary>
        /// Returns the current count of messages queued in the MessageStream.
        /// </summary>
        /// <returns>The current number of messages queued in the stream. </returns>
        public int getDepth()
        {
            return _q.Count;
        }

        /// <summary>
        /// Event handler for the connection state changes of the client.
        /// </summary>
        /// <param name="sender"> The object that triggered the event. </param>
        /// <param name="e"> Event arguments containing information about the connection state change. </param>
        void _client_ConnectionStateChanged(object sender, Client.ConnectionStateChangedEventArgs e)
        {
            /// <summary>
            /// Handles the event when the connection state changes to Disconnected.
            /// </summary>
            if (e.NewState == Client.ConnectionStateChangedEventArgs.ConnectionState.Disconnected)
            {
                lock(internalLock)
                {
                    // Update the internal state to indicate a Disconnected state.
                    _state |= State.Disconnected;
                    // If the client is null, reset associated variables and return.
                    if (_client == null)
                    {
                        _eventHandler = null;
                        _commandId = null;
                        _queryId = null;
                        return;
                    }
                    // Unsubscribe from the ConnectionStateChanged event.
                    if (_eventHandler != null)
                    {
                        _client.ConnectionStateChanged -= _eventHandler;
                        _eventHandler = null;
                    }
                    // Remove message handler for the command ID.
                    if (_commandId != null)
                    {
                        _client.removeMessageHandler(_commandId);
                        _commandId = null;
                    }
                    // Remove message handler for the query ID.
                    if (_queryId != null)
                    {
                        _client.removeMessageHandler(_queryId);
                        _queryId = null;
                    }
                }
            }
        }

        /// <summary>
        /// Handles the invocation of the MessageHandler with a given Message.
        /// </summary>
        /// <param name="message"> The Message to be processed by the MessageHandler. </param>
        void MessageHandler.invoke(Message message)
        {
            try
            {
                /// <summary>
                /// Processes the incoming Message based on certain conditions.
                /// </summary>
                if (_sowKeyMap == null || message.getSowKeyRaw().length == 0)
                {
                    /// <summary>
                    /// Enqueues the Message if SOW key information is unavailable.
                    /// </summary>
                    if (_q.Count > _maxQueueDepth)
                    {
                        throw new MessageStreamFullException("MessageStream is full");
                    }
                    _q.Enqueue(message.copy());

                    /// <summary>
                    /// Adjusts the Message if it is a Publish command and specific conditions such as
                    /// are met.
                    /// </summary>
                    if (message.getCommand() == Message.Commands.Publish
                        && _client != null && _client.getAutoAck()
                        && !message.isBookmarkNull() && !message.isLeasePeriodNull())
                    {
                        message.setIgnoreAutoAck();
                    }
                }
                else
                {
                    Message currentMessage;
                    string sowKey = message.getSowKey();

                    /// <summary>
                    /// Checks if the SOW key is present in the map and updates the existing Message.
                    /// </summary>
                    if (_sowKeyMap.TryGetValue(sowKey, out currentMessage))
                    {
                        if (Interlocked.CompareExchange(ref currentMessage._updateState, Message.Update, Message.Ready) == Message.Ready)
                        {
                            try
                            {
                                message._copyTo(currentMessage);
                            }
                            finally
                            {
                                currentMessage._updateState = Message.Ready;
                            }
                            return;
                        }
                        // if the cas failed, that means a consumer has also removed this key from the map
                    }
                    /// <summary>
                    /// Creates a new Message if the SOW key is not found in the map.
                    /// </summary>
                    if (_q.Count > _maxQueueDepth)
                    {
                        throw new MessageStreamFullException("MessageStream is full");
                    }
                    currentMessage = message.copy();
                    _sowKeyMap[sowKey] = currentMessage;
                    _q.Enqueue(currentMessage);

                    /// <summary>
                    /// Adjusts the Message if it is a Publish command and specific conditions are met.
                    /// </summary>
                    if (message.getCommand() == Message.Commands.Publish
                        && _client != null && _client.getAutoAck()
                        && !message.isBookmarkNull() && !message.isLeasePeriodNull())
                    {
                        message.setIgnoreAutoAck();
                    }
                }
            }
            catch (MessageStreamFullException)
            {
                throw;
            }
            catch (Exception)
            {
                /// <summary>
                /// Handles exceptions by updating the internal state and cleaning up resources.
                /// </summary>
                lock (internalLock)
                {
                    _state |= State.Disconnected;
                    if (_client == null)
                    {
                        _eventHandler = null;
                        _commandId = null;
                        _queryId = null;
                        return;
                    }
                    if (_eventHandler != null)
                    {
                        _client.ConnectionStateChanged -= _eventHandler;
                        _eventHandler = null;
                    }
                    if (_commandId != null)
                    {
                        _client.removeMessageHandler(_commandId);
                        _commandId = null;
                    }
                    if (_queryId != null)
                    {
                        _client.removeMessageHandler(_queryId);
                        _queryId = null;
                    }
                }
            }
        }

        // Retrieves the current message in the stream and performs necessary actions.
        // returns: The current message or null based on stream state and conditions.
        private Message current()
        {
            if (!_previousTopic.isNull() && !_previousBookmark.isNull())
            {
                try
                {
                    /// <summary>
                    /// Acknowledges the previous topic and bookmark if available.
                    /// </summary>
                    if (_client != null)
                    {
                        _client._ack(_previousTopic, _previousBookmark);
                    }
                    _previousTopic.length = 0;
                    _previousBookmark.length = 0;
                }
                catch (Exception)
                {
                    /// <summary>
                    /// Resets state and objects in case of an exception.
                    /// </summary>
                    _previousTopic.reset();
                    _previousBookmark.reset();
                    _current = null;
                    return null;
                }
            }

            if (_current == null) return _current;
            Message.Commands cmd = _current.getCommand();
            // This stream automatically completes if either of these two cases:
            //  (a) we are a "sow" query (no subscribe) and we encounter a group_end message; or
            //  (b) we are any message with a "stats" ack requested, and we receive the stats ack.
            if (((_state&State.SOWOnly) == State.SOWOnly && cmd == Message.Commands.GroupEnd)
                || (_state&State.AcksOnly) == State.AcksOnly)
            {
                /// <summary>
                /// Checks and handles completion conditions of the stream.
                /// </summary>
                if ((_state&State.AcksOnly) == State.AcksOnly)
                {
                    _requestedAcks &= ~(_current.getAckType());
                    if (_requestedAcks != Message.AckTypes.None)
                    {
                        return _current;
                    }
                }
                _state = State.Complete | (_state&State.Disconnected);
                if (_client == null)
                {
                    _eventHandler = null;
                    _commandId = null;
                    _queryId = null;
                    return _current;
                }
                if (_eventHandler != null)
                {
                    _client.ConnectionStateChanged -= _eventHandler;
                    _eventHandler = null;
                }
                if (_commandId != null)
                {
                    _client.removeMessageHandler(_commandId);
                    _commandId = null;
                }
                if (_queryId != null)
                {
                    _client.removeMessageHandler(_queryId);
                    _queryId = null;
                }
            }
            /// <summary>
            /// Adjusts previous topic and bookmark if the current message is a Publish command
            /// with specific conditions met.
            /// </summary>
            if (cmd == Message.Commands.Publish && _client != null &&
                _client.getAutoAck() && !_current.isLeasePeriodNull() &&
                !_current.isBookmarkNull())
            {
                _previousTopic.copyFrom(_current.getTopicRaw());
                _previousBookmark.copyFrom(_current.getBookmarkRaw());
            }
            return _current;
        }

        /// <summary>
        /// Gets the current message in the stream through the Current property.
        /// </summary>
        /// <remarks>
        /// This property returns the current message in the stream by invoking the internal
        /// current() method. It is typically used in the context of iterating over a message
        /// stream, allowing access to the current message without explicitly invoking the
        /// internal method.
        /// </remarks>
        public Message Current
        {
            get
            {
                return current();
            }
        }

        /// <summary>
        /// Disposes of the resources associated with the MessageStream.
        /// </summary>
        /// <remarks>
        /// This method is responsible for cleaning up resources, including unsubscribing
        /// from the client, removing event handlers, and setting the internal state to Complete.
        /// It is typically called when the MessageStream is no longer needed or is being closed.
        /// </remarks>
        public void Dispose()
        {
            try
            {
                if (_client != null)
                {
                    if (_state == State.Subscribed && _commandId != null)
                    {
                        _client.unsubscribe(_commandId);
                        _commandId = null;
                    }
                    if (_eventHandler != null)
                    {
                        _client.ConnectionStateChanged -= _eventHandler;
                        _eventHandler = null;
                    }
                    if (_commandId != null)
                    {
                        _client.removeMessageHandler(_commandId);
                        _commandId = null;
                    }
                    if (_queryId != null)
                    {
                        _client.removeMessageHandler(_queryId);
                        _queryId = null;
                    }
                    _client = null;
                }
            }
            catch (Exception) { }
            _state = State.Complete;
        }

        /// <summary>
        /// Closes the MessageStream by invoking the Dispose method.
        /// </summary>
        /// <remarks>
        /// This method serves as an alias for the Dispose method and is provided for
        /// convenience. It is used to close the MessageStream and release associated resources.
        /// </remarks>
        public void close()
        {
            Dispose();
        }

        /// <summary>
        /// Gets the current object in the enumerator.
        /// </summary>
        /// <returns>The current Message in the stream.</returns>
        object System.Collections.IEnumerator.Current
        {
            get { return current(); }
        }

        // Attempts to dequeue from the internal queue, and Yields while no messages are available but we're still connected.
        /// <summary>
        /// Advances the MessageStream to the next message in the sequence.
        /// </summary>
        /// <returns>
        /// True if the MessageStream successfully moves to the next message; otherwise, false.
        /// </returns>
        /// <remarks>
        /// This method is responsible for advancing the MessageStream to the next message in the
        /// sequence. It employs a forward-only, one-time iteration model. It supports a timeout
        /// mechanism to wait for the next message to be available and handles update completion
        /// in the case of messages with SOW (State of the World) keys.
        /// </remarks>
        public bool MoveNext()
        {
            _current = null;
            bool isTimeout = false;
            if (_timeout > 0)
            {
                int start = Environment.TickCount;
                while ((_state & State.Reading) > 0 && !_q.TryDequeue(out _current) && !isTimeout)
                {
                    System.Threading.Thread.Sleep(1);
                    isTimeout = Environment.TickCount - start > _timeout;
                }
            }
            else
            {
                while ((_state & State.Reading) > 0 && !_q.TryDequeue(out _current))
                {
                    if ((_state&State.Disconnected) == State.Disconnected) break;
                    System.Threading.Thread.Sleep(1);
                }
            }
            if (_current == null)
            {
                if ((_state&State.Disconnected) == State.Disconnected)
                {
                    _state = State.Disconnected;
                    throw new Exceptions.DisconnectedException("Disconnected.");
                }
                // If isTimeout, then we have a "next" but it is null.
                return isTimeout && (_state != State.Complete);
            }
            else if(_sowKeyMap != null && _current.getSowKeyRaw().length > 0)
            {
                Message _;
                string sowKey = _current.getSowKey();
                _sowKeyMap.TryRemove(sowKey, out _);
                // tight spin to wait for update completion -- rare, and nearly instant
                while(Interlocked.CompareExchange(ref _current._updateState, Message.Frozen, Message.Ready) == Message.Update)
                {
                    System.Threading.Thread.Yield();
                }
            }
            return true;
        }

        /// <summary>
        /// Throws a NotImplementedException as the MessageStream is a forward-only, one-time iterator and cannot be reset.
        /// </summary>
        /// <exception cref="NotImplementedException">Thrown to indicate that resetting the iterator is not supported.</exception>
        public void Reset()
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Returns the current enumerator instance for the MessageStream.
        /// </summary>
        /// <returns> The current enumerator instance. </returns>
        public IEnumerator<Message> GetEnumerator()
        {
            return this;
        }

        /// <summary>
        /// Returns the current enumerator instance for the MessageStream.
        /// </summary>
        /// <returns>The current enumerator instance as a non-generic IEnumerator.</returns>
        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return this;
        }
    }
}
