////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Text;

using AMPS.Client.Exceptions;

namespace AMPS.Client
{
    /// <summary>
    /// Used to build up message strings for NVFIX.
    /// </summary>
    public class NVFIXBuilder
    {
        private byte[] buffer = null;
        private int size = 0;
        private int capacity = 1024;

        private byte fieldSeparator;

        public NVFIXBuilder(int capacity, byte fieldSeparator)
        {
            this.fieldSeparator = fieldSeparator;
            this.capacity = capacity < 8 ? 8 : capacity;
            this.buffer = new byte[capacity];
        }

        /// <summary>
        /// Returns the number of bytes in the array containing the NVFIX message.
        /// </summary>
        /// <returns> The number of valid bytes in teh byte array. </returns>
        public int getSize()
        {
            return this.size;
        }

        /// <summary>
        /// Returns the byte array containing the NVFIX message. The number of valid bytes
        /// within the buffer is returned from getSize().
        /// </summary>
        /// <returns> The byte array containing the NVFIX message. </returns>
        public byte[] getBytes()
        {
            return buffer;
        }

        private void checkCapacity(int bytesNeeded)
        {
            if (capacity - size < bytesNeeded)
            {
                // Not enough capacity left, must resize
                while (capacity - size < bytesNeeded)
                {
                    capacity = capacity * 2;
                }

                byte[] newbuf = new byte[capacity];
                Array.Copy(buffer, 0, newbuf, 0, size);
                buffer = newbuf;
            }
        }

        /// <summary>
        /// Appends a tag and value pair contained within a byte buffer to the NVFIX message.
        /// </summary>
        /// <param name="tag">The NVFIX tag to append</param>
        /// <param name="value">The byte buffer containing the NVFIX value to append</param>
        /// <param name="offset">The starting location of the value inside the byte buffer</param>
        /// <param name="length">The length of the value inside the byte buffer</param>
        /// <returns>A reference to this object</returns>
        /// <exception cref="CommandException">Thrown if tag cannot be converted to ISO-8859-1.</exception>
        public NVFIXBuilder append(String tag, byte[] value, int offset, int length)
        {
            byte[] tagBytes;

            try
            {
                tagBytes = Encoding.GetEncoding("ISO-8859-1").GetBytes(tag);
            }
            catch (EncoderFallbackException ioex)
            {
                throw new CommandException("Unable to encode tag '" + tag
                      + "' to a byte array in FIXBuilder.\n" + ioex.ToString());
            }
            int sizeNeeded = tagBytes.Length + length + 2;
            checkCapacity(sizeNeeded);

            // Tag
            int writeIndex = size;

            Array.Copy(tagBytes, 0, buffer, writeIndex, tagBytes.Length);
            writeIndex += tagBytes.Length;

            // Equal sign
            buffer[writeIndex] = (byte)'=';
            ++writeIndex;

            // Value
            Array.Copy(value, offset, buffer, writeIndex, length);
            writeIndex += length;

            // Field separator
            buffer[writeIndex] = fieldSeparator;
            size += sizeNeeded;

            return this;
        }

        /// <summary>
        /// Appends a tag and value pair to the NVFIX message.
        /// </summary>
        /// <param name="tag">the NVFIX tag to append</param>
        /// <param name="value">the NVFIX value for the tag to append. The value will be converted
        /// to an ISO-8859-1 byte array for writing.</param>
        /// <returns>a reference to this object.</returns>
        /// <exception cref="CommandException">If the tag or value argument is not convertible to ISO-8859-1</exception>
        public NVFIXBuilder append(String tag, String value)
        {
            byte[] tagBytes;
            byte[] valueBytes;
            try
            {
                tagBytes = Encoding.GetEncoding("ISO-8859-1").GetBytes(tag);
            }
            catch (EncoderFallbackException ioex)
            {
                throw new CommandException("Unable to encode tag '" + tag
                      + "' to a byte array in FIXBuilder.\n" + ioex.ToString());
            }
            try
            {
                valueBytes = Encoding.GetEncoding("ISO-8859-1").GetBytes(value);
            }
            catch (EncoderFallbackException ioex)
            {
                throw new CommandException("Unable to encode value '" + value
                      + "' to a byte array in FIXBuilder.\n" + ioex.ToString());
            }
            int sizeNeeded = tagBytes.Length + valueBytes.Length + 2;
            checkCapacity(sizeNeeded);

            // Tag
            int writeIndex = size;

            Array.Copy(tagBytes, 0, buffer, writeIndex, tagBytes.Length);

            writeIndex += tagBytes.Length;

            // Equal sign
            buffer[writeIndex] = (byte)'=';
            ++writeIndex;

            // Value
            Array.Copy(valueBytes, 0, buffer, writeIndex, valueBytes.Length);
            writeIndex += valueBytes.Length;

            // Field separator
            buffer[writeIndex] = fieldSeparator;
            size += sizeNeeded;

            return this;
        }

        /// <summary>
        /// Clears the current buffer in the Builder so it can be reused. Capacity is unaffected.
        /// </summary>
        /// <returns>a reference to this object.</returns>
        public NVFIXBuilder clear()
        {
            size = 0;
            return this;
        }

    }
}
