///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System.Collections.Generic;
namespace AMPS.Client
{

    /// <summary>
    /// A simple generic instance pool for objects that take a while to construct.
    /// </summary>
    /// <typeparam name="T">The pooled type</typeparam>
    public class Pool<T> where T:new()
    {
        int _initialSize = 0;
        Queue<T> _freePool = new Queue<T>();

        /// <summary>
        /// Initializes a new instance of the Pool class with the specified initial size.
        /// </summary>
        /// <param name="initialSize">The initial size of the pool.</param>
        public Pool(int initialSize)
        {
            if(initialSize < 1) initialSize = 1;
            _initialSize = initialSize;
            grow();
        }

        private void grow()
        {
            for(int i = 0; i < _initialSize; ++i)
            {
                _freePool.Enqueue(new T());
            }
        }

        /// <summary>
        /// Returns an object from the pool, growing itself if necessary (by initialSize).
        /// If T cannot be instantiated, returns null.
        /// </summary>
        /// <returns>A pooled instance.</returns>
        public T get()
        {
            lock(_freePool)
            {
                if(_freePool.Count < 1)
                {
                    grow();
                }
                return _freePool.Dequeue();
            }
        }

        /// <summary>
        /// Return an object to the pool. Note, you're responsible for clear/reset of objects being returned.
        /// </summary>
        /// <param name="theInstance">The T to return to the pool.</param>
        public void returnToPool(T theInstance)
        {
            lock (_freePool)
            {
                // The Instance must be reset prior to return.
                _freePool.Enqueue(theInstance);
            }
        }
    }
}