////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System.Collections.Generic;
using System;
using System.Reflection;

using AMPS.Client.Exceptions;

namespace AMPS.Client
{

    /// <summary>
    /// Class which creates Protocols as well as registering and unregistering them from a Dictionary.
    /// </summary>
    public class ProtocolFactory
    {

        private static Dictionary<string, string> registry = new Dictionary<string, string>();

        static ProtocolFactory()
        {
            registry["amps"] = "AMPS.Client.JSONProtocol";
        }

        /// <summary>
        /// Method to register the name of the message and the message type.
        /// </summary>
        /// <param name="name"> A string containing the name of the message. </param>
        /// <param name="msgType"> A string containing the type of the message. </param>
        /// <exception cref="ProtocolException"> The name is already registered </exception>
        public static void register(string name, string msgType)
        {
            if (!registry.ContainsKey(name))
            {
                registry[name] = msgType;
            }
            else
            {
                throw new ProtocolException("'" + name + "' is already registered.");
            }
        }

        /// <summary>
        /// Method to unregister a message.
        /// </summary>
        /// <param name="name"> A string containing the name of the message. </param>
        /// <exception cref="ProtocolException"> The provided name is not registered. </exception>
        public static void unregister(string name)
        {
            if (registry.ContainsKey(name))
            {
                registry.Remove(name);
            }
            else
            {
                throw new ProtocolException("'" + name
                      + "' is not a known Protocol, can not unregister.");
            }
        }

        /// <summary>
        /// Creates a protocol.
        /// </summary>
        /// <param name="name"> A string which holds the name of the protocol. </param>
        /// <param name="props"> A Properties object which holds the properties of the protocol. </param>
        /// <exception cref="ProtocolException"> The provided name is not a valid protocol. </exception>
        /// <returns> This instance so that various operations can be chained together. </returns>
        public static Protocol createProtocol(
              string name,
              Properties props)
        {
            // Guess at the protocol name
            string mtName = "AMPS.Client." + name.ToUpperInvariant() + "Protocol";
            if (registry.ContainsKey(name))
            {
                mtName = registry[name];
            }
            try
            {
                return (Protocol)Assembly.GetExecutingAssembly().GetType(mtName)  // find the protocol class
                                    .GetConstructor(new Type[] { typeof(Properties) })   // find a ctor that takes a single Properties arg
                                    .Invoke(new object[] { props });  // invoke that baby

            }
            catch (Exception)
            {
                throw new ProtocolException("'" + name + "' is not a valid Protocol.");
            }
        }

    }
}
