////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using AMPS.Client.Exceptions;
using System;
using System.IO;
using System.IO.MemoryMappedFiles;
using System.Runtime.InteropServices;

namespace AMPS.Client
{
    /// <summary>
    /// PublishStore is a memory-mapped file-backed store for storing outgoing
    /// messages. As messages are stored, space is allocated from a pre-created flat
    /// buffer on disk. As messages are discarded, space in that buffer is marked
    /// "free" for future store operations. If messages are stored faster than they
    /// are published, the buffer is re-sized to create more capacity. PublishStore
    /// helps in keeping track of unacknowledged messages. If messages are published
    /// at a rate exceeding the storage capacity, the buffer is resized to accommodate
    /// more messages.
    /// </summary>
    /// <remarks>
    /// The PublishStore class provides a robust foundation for implementing features
    /// such as tracking unacknowledged messages and ensuring message persistence.
    /// </remarks>
    public class PublishStore : BlockPublishStore
    {
        const int DEFAULT_INITIAL_BLOCKS = 10000;
        private int _initialCapacity;
        private bool _truncateOnClose = false;
        private string _path;

        /// <summary>
        /// Represents the buffer used for memory-mapped file storage.
        /// </summary>
        class MMapStoreBuffer : BlockPublishStore.Buffer
        {
            Guid _guid;
            string _path;
            MemoryMappedFile _file = null;
            FileStream _file_stream = null;
            MemoryMappedViewAccessor _file_accessor = null;
            long _file_position = 0;
            int _initialCapacity;

            /// <summary>
            /// Initializes a new instance of the MMapStoreBuffer class with the specified path and initial
            /// capacity.
            /// </summary>
            /// <param name="path">The path of the memory-mapped file.</param>
            /// <param name="initialCapacity">The initial capacity of the buffer.</param>
            public MMapStoreBuffer(string path, int initialCapacity = DEFAULT_INITIAL_BLOCKS)
            {
                _guid = Guid.NewGuid();
                _path = path;
                _file = null;
                _file_stream = null;
                _file_accessor = null;
                _initialCapacity = initialCapacity;
                open();
            }
            ~MMapStoreBuffer()
            {
                close();
            }

            /// <summary>
            /// Opens the memory-mapped file and initializes the buffer.
            /// </summary>
            public void open()
            {
                if (_file_stream != null || _file != null || _file_accessor != null)
                    return;
                //check if the file exists so we can find the size
                FileInfo fi1 = new FileInfo(_path);
                bool exists = fi1.Exists;

                //if the file doesn't exist or it exists but is size 0, initialize the size
                long filesize = 0;
                if (exists == false)
                {
                    filesize = BlockPublishStore.Block.SIZE * _initialCapacity;
                }
                else if (fi1.Length == 0)
                {
                    filesize = BlockPublishStore.Block.SIZE * _initialCapacity;
                 }


                _file_stream = new FileStream(_path, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite);
                if(filesize != 0)
                {
                  _file_stream.SetLength(filesize);
                }

#if (NET46_OR_GREATER || NETSTANDARD || NET)
                String mapname = null; // Named maps aren't allowed on linux
                if (System.Runtime.InteropServices.RuntimeInformation.IsOSPlatform(OSPlatform.Windows))
                {
                  mapname = "MMapStoreBuffer" + _guid.ToString();
                }
                _file = MemoryMappedFile.CreateFromFile(_file_stream, mapname, filesize, MemoryMappedFileAccess.ReadWrite, HandleInheritability.Inheritable, false);
#else
                _file = MemoryMappedFile.CreateFromFile(_file_stream, "MMapStoreBuffer" + _guid.ToString(), filesize, MemoryMappedFileAccess.ReadWrite, null, HandleInheritability.Inheritable, false);
#endif
                _file_accessor = _file.CreateViewAccessor();
            }

            /// <summary>
            /// Gets the size of the memory-mapped file.
            /// </summary>
            /// <returns></returns>
            public long getSize()
            {
                return _file_accessor.Capacity;
            }

            /// <summary>
            /// Sets the size of the memory-mapped file to the specified new size.
            /// </summary>
            /// <param name="newSize">The new size to set for the memory-mapped file.</param>
            /// <exception cref="IOException">Thrown when the store is closed.</exception>
            public void setSize(long newSize)
            {
                if (_file_accessor == null)
                {
                    throw new IOException("The store is closed.");
                }

                //if a smaller size is requested, ignore the resize
                if (_file_accessor.Capacity > newSize)
                    return;

                //close the file and reopen w/ a larger size
                this.close();
                _file_stream = new FileStream(_path, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite);
#if (NET46_OR_GREATER || NETSTANDARD || NET)
                String mapname = null; // Named maps aren't allowed on linux
                if (System.Runtime.InteropServices.RuntimeInformation.IsOSPlatform(OSPlatform.Windows))
                {
                  mapname = "MMapStoreBuffer" + _guid.ToString();
                }
                _file = MemoryMappedFile.CreateFromFile(_file_stream, mapname, newSize, MemoryMappedFileAccess.ReadWrite, HandleInheritability.Inheritable, false);
#else
                _file = MemoryMappedFile.CreateFromFile(_file_stream, "MMapStoreBuffer" + _guid.ToString(), newSize, MemoryMappedFileAccess.ReadWrite, null, HandleInheritability.Inheritable, false);
#endif
                _file_accessor = _file.CreateViewAccessor();
            }

            /// <summary>
            /// Gets the current position within the memory-mapped file.
            /// </summary>
            /// <returns>The current position within the memory-mapped file.</returns>
            public long getPosition()
            {
                return _file_position;
            }

            /// <summary>
            /// Sets the position within the memory-mapped file to the specified position.
            /// </summary>
            /// <param name="position">The position to set within the memory-mapped file.</param>
            public void setPosition(long position)
            {
                _file_position = position;
            }

            /// <summary>
            /// Writes a byte to the current position within the memory-mapped file and advances
            /// the position by one byte.
            /// </summary>
            /// <param name="b">The byte to write.</param>
            public void putByte(byte b)
            {
                _file_accessor.Write(_file_position, b);
                _file_position += 1;
            }

            /// <summary>
            /// Reads a byte from the current position within the memory-mapped file and advances
            /// the position by one byte.
            /// </summary>
            /// <returns>The byte read from the current position within the memory-mapped file.</returns>
            public byte getByte()
            {
                _file_position += 1;
                return _file_accessor.ReadByte(_file_position - 1);
            }

            /// <summary>
            /// Writes an integer value to the current position within the memory-mapped file and
            /// advances the position by four bytes.
            /// </summary>
            /// <param name="i">The integer value to write.</param>
            public void putInt(int i)
            {
                _file_accessor.Write(_file_position, i);
                _file_position += 4;
            }

            /// <summary>
            /// Writes an integer value to the specified position within the memory-mapped file.
            /// </summary>
            /// <param name="position">The position to write the integer value.</param>
            /// <param name="intValue">The integer value to write.</param>
            public void putInt(long position, int intValue)
            {
                _file_accessor.Write(position, intValue);
            }

            /// <summary>
            /// Reads an integer value from the current position within the memory-mapped file and
            /// advances the position by four bytes.
            /// </summary>
            /// <returns>The integer value read from the current position within the memory-mapped
            /// file.</returns>
            public int getInt()
            {
                int retval = _file_accessor.ReadInt32(_file_position);
                _file_position += 4;
                return retval;
            }

            /// <summary>
            /// Reads an integer value from the specified position within the memory-mapped file.
            /// </summary>
            /// <param name="position">The position from which to read the integer value.</param>
            /// <returns>The integer value read from the specified position within the memory-mapped
            /// file.</returns>
            public int getInt(long position)
            {
                int retval =_file_accessor.ReadInt32(position);
                return retval;
            }

            /// <summary>
            /// Writes a long integer value to the current position within the memory-mapped file and
            /// advances the position by eight bytes.
            /// </summary>
            /// <param name="l">The long integer value to write.</param>
            public void putLong(long l)
            {
                _file_accessor.Write(_file_position, l);
                _file_position += 8;
            }

            /// <summary>
            /// Writes a long integer value to the specified position within the memory-mapped file.
            /// </summary>
            /// <param name="position">The position to write the long integer value.</param>
            /// <param name="longValue">The long integer value to write.</param>
            public void putLong(long position, long longValue)
            {
                _file_accessor.Write(position, longValue);
            }

            /// <summary>
            /// Reads a long integer value from the current position within the memory-mapped file and
            /// advances the position by eight bytes.
            /// </summary>
            /// <returns>The long integer value read from the current position within the memory-mapped
            /// file.</returns>
            public long getLong()
            {
                _file_position += 8;
                return _file_accessor.ReadInt64(_file_position - sizeof(long));
            }

            /// <summary>
            /// Writes an unsigned long integer value to the current position within the memory-mapped
            /// file and advances the position by eight bytes.
            /// </summary>
            /// <param name="l">The unsigned long integer value to write.</param>
            public void putUlong(ulong l)
            {
                _file_accessor.Write(_file_position, l);
                _file_position += 8;
            }

            /// <summary>
            /// Writes an unsigned long integer value to the specified position within the memory-mapped
            /// file.
            /// </summary>
            /// <param name="position">The position to write the unsigned long integer value.</param>
            /// <param name="ulongValue">The unsigned long integer value to write.</param>
            public void putUlong(long position, ulong ulongValue)
            {
                _file_accessor.Write(position, ulongValue);
            }

            /// <summary>
            /// Reads an unsigned long integer value from the current position within the memory-mapped
            /// file and advances the position by eight bytes.
            /// </summary>
            /// <returns>The unsigned long integer value read from the current position within the
            /// memory-mapped file.</returns>
            public ulong getUlong()
            {
                _file_position += 8;
                return _file_accessor.ReadUInt64(_file_position - sizeof(ulong));
            }

            /// <summary>
            /// Writes a byte sequence to the current position within the memory-mapped file and advances
            /// the position by the length of the byte sequence.
            /// </summary>
            /// <param name="bytes">The byte sequence to write.</param>
            public void putBytes(ByteSequence bytes)
            {
                _file_accessor.WriteArray<byte>(_file_position, bytes.array, (int)bytes.offset, (int)bytes.len);
                _file_position += bytes.len;
            }

            /// <summary>
            /// Reads a byte sequence from the current position within the memory-mapped file and advances
            /// the position by the length of the byte sequence.
            /// </summary>
            /// <param name="outBytes">The byte sequence to store the read data.</param>
            public void getBytes(ByteSequence outBytes)
            {
                if (outBytes.array == null || outBytes.array.Length < outBytes.len)
                {
                    outBytes.array = new byte[outBytes.len];
                }
                outBytes.offset = 0;
                _file_accessor.ReadArray<byte>(_file_position, outBytes.array, 0, (int)outBytes.len);
                _file_position += outBytes.len;
            }

            /// <summary>
            /// Writes a portion of a byte array to the current position within the memory-mapped file and
            /// advances the position by the specified length.
            /// </summary>
            /// <param name="buffer">The byte array containing the data to write.</param>
            /// <param name="position">The starting position in the byte array from which to write.</param>
            /// <param name="length">The number of bytes to write.</param>
            public void putBytes(byte[] buffer, int position, int length)
            {
                _file_accessor.WriteArray<byte>(_file_position, buffer,
                                                position, length);
                _file_position += length;
            }

            /// <summary>
            /// Reads a byte sequence from the current position within the memory-mapped file and advances
            /// the position by the specified length.
            /// </summary>
            /// <param name="outField">The field to store the read data.</param>
            /// <param name="length">The length of the byte sequence to read.</param>
            public void getBytes(Fields.Field outField, int length)
            {
                outField.reset();
                outField.length = length;
                if (outField.length > 0)
                {
                    outField.buffer = new byte[outField.length];
                    outField.position = 0;
                    _file_accessor.ReadArray<byte>(_file_position, outField.buffer, 0, outField.length);
                    _file_position += outField.length;
                }
            }

            /// <summary>
            /// Sets a range of bytes within the memory-mapped file to zero.
            /// </summary>
            /// <param name="offset">The starting offset within the memory-mapped file.</param>
            /// <param name="length">The number of bytes to set to zero.</param>
            public void zero(long offset, int length)
            {
                for (long i = 0; i < length; i += 8)
                {
                    _file_accessor.Write(offset + i, (long)0);
                }
            }

            /// <summary>
            /// Disposes of the current instance, closing the memory-mapped file.
            /// </summary>
            public void Dispose()
            {
                close();
            }

            /// <summary>
            /// Close the file access, making this store unusable.
            /// </summary>
            public void close()
            {
                if (_file_accessor != null)
                {
                    _file_accessor.Flush();
                    _file_accessor.Dispose();
                    _file_accessor = null;
                }
                if (_file != null)
                {
                    _file.Dispose();
                    _file = null;
                }
                if (_file_stream != null)
                {
                    _file_stream.Close();
                    _file_stream = null;
                }
            }

            /// <summary>
            /// Flushes any buffered data to the memory-mapped file and underlying file
            /// stream, ensuring that all changes are persisted.
            /// </summary>
            public void sync()
            {
                _file_accessor.Flush();
                _file_stream.Flush();
            }
        }

        /// <summary>
        /// Create a PublishStore using a file at path
        /// </summary>
        /// <param name="path">The file to use for the store</param>
        /// <param name="errorOnPublishGap_">Indicates if an excpetion should be thrown if discardUpTo is called with an idex below the last persisted.</param>
        public PublishStore(string path, bool errorOnPublishGap = false)
            : base(new MMapStoreBuffer(path), true, DEFAULT_INITIAL_BLOCKS, errorOnPublishGap)
        {
            _initialCapacity = DEFAULT_INITIAL_BLOCKS;
            _path = path;
            try
            {
                recover();
            }
            catch (StoreException ex)
            {
                if (ex.InnerException != null && ex.InnerException is IOException)
                {
                    throw (IOException)ex.InnerException;
                }
                else
                {
                    throw new IOException(ex.Message, ex);
                }
            }
            if (_usedList == null) growFreeListIfEmpty();
        }

        /// <summary>
        /// Create a PublishStore using a file at path with an initial size and
        /// resize value of blocksPerRealloc.
        /// </summary>
        /// <param name="path">The file to use for the store.</param>
        /// <param name="blocksPerRealloc">The number of 2k blocks to use as initial size and resize increment.</param>
        /// <param name="errorOnPublishGap_">Indicates if an excpetion should be thrown if discardUpTo is called with an idex below the last persisted.</param>
        public PublishStore(string path, int blocksPerRealloc, bool errorOnPublishGap = false)
            : base(new MMapStoreBuffer(path, blocksPerRealloc), true, blocksPerRealloc, errorOnPublishGap)
        {
            _initialCapacity = blocksPerRealloc;
            _path = path;
            try
            {
                recover();
            }
            catch (StoreException ex)
            {
                if (ex.InnerException != null && ex.InnerException is IOException)
                {
                    throw (IOException)ex.InnerException;
                }
                else
                {
                    throw new IOException(ex.Message, ex);
                }
            }
            if (_usedList == null) growFreeListIfEmpty();
        }

        /// <summary>
        /// Set an implementation of CRC32 to be used by the Store to prevent message corruption
        /// Default implementation is AMPS.Client.Utilities.DefaultCRC32
        /// Passing in null will mean no crc calculation and no checking for file corruption.
        /// </summary>
        /// <param name="crc_">The CRC32 checksum to set.</param>
        public void setCrc(Utilities.CRC32 crc_)
        {
            base._crc = crc_;
        }

        /// <summary>
        /// Opens the memory-mapped file buffer.
        /// </summary>
        public void open()
        {
            ((MMapStoreBuffer)_buffer).open();
        }

        /// <summary>
        /// Sets whether to truncate the file on close.
        /// </summary>
        /// <param name="truncate">A boolean indicating whether to truncate the file on close.</param>
        public void truncateOnClose(bool truncate)
        {
            _truncateOnClose = truncate;
        }

        /// <summary>
        /// Disposes of the current instance, closing the memory-mapped file and performing any necessary
        /// cleanup.
        /// </summary>
        public override void Dispose()
        {
            close();
        }

        /// <summary>
        /// Closes the memory-mapped file and performs any necessary cleanup.
        /// </summary>
        public void close()
        {
            if (_buffer == null) return;
            long unpersisted = this.unpersistedCount();
            ((MMapStoreBuffer)_buffer).close();
            if (_truncateOnClose && unpersisted == 0)
            {
                using (FileStream fs = new FileStream(_path,
                                                      FileMode.Open,
                                                      FileAccess.ReadWrite,
                                                      FileShare.ReadWrite))
                {
                    fs.SetLength(_initialCapacity * BlockPublishStore.Block.SIZE);
                }
            }
        }

        /// <summary>
        /// Flushes any buffered data to the memory-mapped file and underlying file stream, ensuring
        /// that all changes are persisted.
        /// </summary>
        public void sync()
        {
            ((MMapStoreBuffer)_buffer).sync();
        }

    }
}
