///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading;
using System.Collections.Generic;
using AMPS.Client.Fields;
using AMPS.Client.Exceptions;

/// <summary>
/// Interface to provide a recovery point for a bookmark subscription. Used in
/// the RecoveryPointAdapter interface for updates and recovery.
/// </summary>
namespace AMPS.Client
{

    /// <summary>
    /// Interface used by RecoveryPointAdapter's to represent a bookmark
    /// subscription's recovery state, to allow that subscription to be resumed
    /// in a reasonable place when the subscriber application exits and restarts.
    /// This is typically a representation for one or more bookmarks, similar to
    /// the MOST_RECENT state of a bookmark store that's used for resuming replays.
    /// </summart>
    /// <remarks>
    /// Implementations of this interface may contain a snapshot of recovery state
    /// or may be lazily evaluated, such that they return the associated bookmark
    /// store's current recovery state for the specified subscription id.
    /// </remarks>
    public interface RecoveryPoint
    {
        /// <summary>
        /// Return the subscription id for this recovery point
        /// </summary>
        Field getSubId();

        /// <summary>
        /// Return the bookmark for this recovery point
        /// </summary>
        BookmarkField getBookmark();

        /// <summary>
        /// Return a deep copy of self
        /// </summary>
        RecoveryPoint copy();

    }

    /// <summary>
    /// Interface for creating recovery points.
    /// </summary>
    public interface RecoveryPointFactory
    {

        /// <summary>
        /// Creates a recovery point with the specified subscription id and bookmark.
        /// </summary>
        /// <param name="subId_">The subscription id.</param>
        /// <param name="bookmark_">The bookmark.</param>
        /// <returns>The created recovery point.</returns>
        RecoveryPoint createRecoveryPoint(Field subId_,
                                          BookmarkField bookmark_);
    }

}


