////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Runtime.CompilerServices;
using System.IO;
using System.IO.MemoryMappedFiles;
using AMPS.Client.Exceptions;
using AMPS.Client.Fields;

/**
 * RingBookmarkStore is an implementation of BookmarkStore that stores state on disk and in memory.
 * For each subscription, RingBookmarkStore keeps the most recent bookmark B for which all
 * messages prior to B have been discard()ed by the subscriber.  As messages are logged and
 * discarded, an in-memory array is used to track when a new bookmark can be logged.
 */
namespace AMPS.Client
{
    public class RingBookmarkStore : BookmarkStore, IDisposable
    {
        // Manages the portion of the store for a single subscription.
        // RingBookmarkStore uses a memory mapped file for persistent storage.
        // Each subscription has four elements: one subId of size BYTES_SUBID,
        // and three bookmarks of size BYTES_BOOKMARK.
        // On disk, each bookmark entry begins with one byte designating the next bookmark entry
        // to be written.  The value '*', aka the Cursor, in this position indicates this is the oldest bookmark
        // stored for the subscription, and is the next to be written.
        // Writing always follows a predictable sequence: the '*' for the next entry is written,
        // the bookmark id is written to the current entry, and then the cursor is erased for the current entry.
        // On recovery, the entry *before* the earliest one with the '*' is the most recently
        // written: in case of crash during write, there may be two entries with a '*', and the third one is the one to recover.

        // In memory, each Subscription has an array of entry's.  Each entry holds a bookmark and whether or not it has been discard()ed.
        // Whenever the 0th entry in the array is discarded, we know it's time to write a bookmark to disk.
        // We truncate the front of the array up through the last one that is inactive, and then write the next one to disk.

        class Subscription : AMPS.Client.Subscription
        {
            // The buffer holding the memory-mapped file of all subscriptions
            MemoryMappedViewAccessor _buffer;
            // The offset into _buffer where our subscription begins.
            internal int _offset;
            // The disk entry to be written next.
            short _currentDiskPosition = 0;
            // Don't write while recovering
            volatile bool _recovering = false;
            BookmarkRingBuffer _ring = new BookmarkRingBuffer();
            // If this subscription is a range, we save it here
            BookmarkRangeField _range = new BookmarkRangeField();

            // Initial number of in-memory entries to create.  Resize takes some time.
            public const int INITIAL_SIZE = 100000;
            // On-disk bytes for a bookmark entry.
            public const int BYTES_BOOKMARK = BookmarkField.MAX_BOOKMARK_LENGTH * 6 + 8;
            // On-disk bytes for each entry
            public const int BYTES_ENTRY = 1024;
            // Number of bookmarks stored per subscription.  Must be at least 3 for recovery guarantee.
            public const short POSITIONS = 3;
            // On-disk bytes for the subscription ID.
            public const int BYTES_SUBID = BYTES_ENTRY - (BYTES_BOOKMARK * POSITIONS);

            private Object internalLock = new Object();

            public void init(MemoryMappedViewAccessor buffer, int offset)
            {
                _buffer = buffer;
                _offset = offset;

                recover();
            }

            /// <summary>
            /// Logs a bookmark to the subscription.
            /// </summary>
            /// <remarks> A new bookmark has arrived for this subscription.  Just remember it:
            /// nothing to write about until the message is discarded. </remarks>
            /// <param name="bookmark">The bookmark to log.</param>
            /// <returns>The sequence number of the logged bookmark.</returns>
            public override long log(BookmarkField bookmark)
            {
                lock (internalLock) {
                    if (!bookmark.isRange())
                    {
                        long seq = _ring.log(bookmark);
                        while (seq == 0) {
                            Monitor.Exit(internalLock);
                            try {
                                _ring.checkResize();
                            }
                            finally {
                                Monitor.Enter(internalLock);
                            }
                            seq = _ring.log(bookmark);
                        }
                        return seq;
                    }
                    else
                    {
                        _range.copyFrom(bookmark);
                        if (!_range.isValid())
                        {
                            throw new CommandException($"Invalid bookmark range specified {bookmark}");
                        }
                        long seq = 0;
                        if (!_range.isStartInclusive()) {
                            seq = _ring.log(_range.getStart());
                            _ring.discard(seq);
                        }
                        write(_range);
                        return seq;
                    }
                }
            }


            /// <summary>
            /// Determines whether a bookmark is discarded.
            /// </summary>
            /// <remarks>Since we don't persist the discard()ed message that came
            /// after the most recently logged bookmark, we always assume any
            /// message the user is testing for should be delivered.</remarks>
            /// <param name="bookmark">The bookmark to check.</param>
            /// <returns>Always returns false.</returns>
            public override bool isDiscarded(BookmarkField bookmark)
            {
                return false;
            }

            /// <summary>
            /// User is finished with a bookmark. Marks a bookmark entry as inactive and, if it's the oldest entry,
            /// forwards-truncates and logs the new most recent.</summary>
            /// <param name="bookmark">The sequence number of the bookmark to discard.</param>
            public override void discard(long bookmark)
            {
                lock (internalLock) {
                    // If discard changes most recent, save updated recent
                    if (_ring.discard(bookmark))
                    {
                        // If we have a range, we'll log the range
                        if (_range.isValid())
                        {
                            BookmarkField recent = _ring.getLastDiscarded();
                            // Replace the start if it's changed
                            if (_range.isStartInclusive()
                               || !recent.Equals(_range.getStart()))
                            {
                                _range.replaceStart(recent, true);
                                write(_range);
                            }
                        }
                        else // Simply write most recent
                        {
                            write(_ring.getLastDiscarded());
                        }
                    }
                }
            }

            /// <summary>
            /// Retrieves the most recent bookmark.
            /// </summary>
            /// <returns>The most recent bookmark.</returns>
            public override Field getMostRecent()
            {
                lock (internalLock) {
                    return _ring.getLastDiscarded();
                }
            }

            /// <summary>
            /// Retrieves the most recent bookmark, optionally returning a copy of
            /// the bookmark as a list.
            /// </summary>
            /// <param name="useList">Specifies whether to return a copy of the
            /// bookmark as a list.</param>
            /// <returns>The most recent bookmark.</returns>
            public override Field getMostRecentList(bool useList)
            {
                lock (internalLock) {
                    if (_range.isValid())
                    {
                        return _range.copy();
                    }
                    else
                    {
                        return _ring.getLastDiscarded().copy();
                    }
                }
            }

            private void write(Field bookmark)
            {
                if (_recovering) return;
                lock (internalLock) {
                    // We want to write to _currentDiskPosition.
                    short nextDiskPosition = (short)((_currentDiskPosition + 1) % POSITIONS);

                    // Mark the next position with the 'cursor'
                    _buffer.Write(_offset + BYTES_SUBID + (BYTES_BOOKMARK * nextDiskPosition), (byte)'*');

                    // write the current position and validate it
                    for (int i = 0; i < bookmark.length; i++)
                    {
                        _buffer.Write(_offset + BYTES_SUBID + (BYTES_BOOKMARK * _currentDiskPosition) + 1 + i, bookmark.byteAt(i));
                    }
                    for (int i = bookmark.length; i < BYTES_BOOKMARK - 1; i++)
                    {
                        _buffer.Write(_offset + BYTES_SUBID + (BYTES_BOOKMARK * _currentDiskPosition) + 1 + i, (byte)0);
                    }
                    _buffer.Write(_offset + BYTES_SUBID + (BYTES_BOOKMARK * _currentDiskPosition), (byte)'+');

                    // advance _currentDiskPosition
                    _currentDiskPosition = nextDiskPosition;
                }
            }

            /// <summary>
            /// Recovers the subscription state after a crash.
            /// </summary>
            public void recover()
            {
                _recovering = true;
                try
                {
                    // find the first cursor
                    short foundCursor = 0;
                    for (; foundCursor < POSITIONS; foundCursor++)
                    {
                        byte b = _buffer.ReadByte(_offset + BYTES_SUBID + (BYTES_BOOKMARK * foundCursor));
                        if (b == (byte)'*') break;
                    }
                    if (foundCursor == 0)
                    {
                        byte b = _buffer.ReadByte(_offset + BYTES_SUBID + (BYTES_BOOKMARK * (POSITIONS - 1)));
                        if (b == (byte)'*')
                        {
                            foundCursor = POSITIONS - 1;
                        }
                    }
                    if (foundCursor < POSITIONS)
                    {
                        int mostRecentValid = (foundCursor==0 ? POSITIONS-1 : foundCursor-1);
                        byte b = _buffer.ReadByte(_offset + BYTES_SUBID + (BYTES_BOOKMARK * mostRecentValid));
                        if (b != (byte)'+')
                        {
                            _currentDiskPosition = 0;
                            return;
                        }
                        // Found an existing "cursor": start the writing there.
                        _currentDiskPosition = foundCursor;
                        byte[] buf = new byte[BYTES_BOOKMARK - 1];
                        _buffer.ReadArray<byte>(_offset + BYTES_SUBID + (BYTES_BOOKMARK * mostRecentValid) + 1, buf, 0, BYTES_BOOKMARK - 1);

                        int bookmarkLength = 0;
                        for (; bookmarkLength < buf.Length && buf[bookmarkLength] != 0; bookmarkLength++) ;

                        BookmarkField f = new BookmarkField();
                        f.set(buf, 0, bookmarkLength);
                        // discard and log to make this the "starting point" for the subscription.
                        if (f.isRange()) log(f);
                        else discard(log(f));
                    }
                    else
                    {
                        _currentDiskPosition = 0;
                    }
                }
                finally
                {
                    _recovering = false;
                }
            }

            /// <summary>
            /// Sets the last persisted bookmark field (no operation).
            /// </summary>
            /// <param name="bookmark">The last persisted bookmark field.</param>
            public override void setLastPersisted(BookmarkField bookmark)
            {
                // no-op
            }

            /// <summary>
            /// Sets the last persisted bookmark sequence (no operation).
            /// </summary>
            /// <param name="bookmark">The last persisted bookmark sequence.</param>
            public override void setLastPersisted(long bookmark)
            {
                // no-op
            }

            /// <summary>
            /// Retrieves the sequence number of the oldest bookmark.
            /// </summary>
            /// <returns>The sequence number of the oldest bookmark.</returns>
            public override long getOldestBookmarkSeq()
            {
                lock (internalLock) {
                    return _ring.getStartIndex();
                }
            }

            /// <summary>
            /// Sets the resize handler for the bookmark store.
            /// </summary>
            /// <param name="handler">The resize handler.</param>
            /// <param name="store">The bookmark store.</param>s
            public override void setResizeHandler(BookmarkStoreResizeHandler handler, BookmarkStore store)
            {
                lock (internalLock) {
                    _ring.setResizeHandler(handler, store);
                }
            }
        }

        MemoryMappedFile _file;
        MemoryMappedViewAccessor _buffer;
        private const int ENTRIES = 16384;
        private const int ENTRY_SIZE = Subscription.BYTES_SUBID + (Subscription.POSITIONS * Subscription.BYTES_BOOKMARK);
        private Dictionary<Field, Subscription> _map;
        // Initialize to ENTRIES, so that there is no free space in the log until recover() has succeeded.
        private int _free = ENTRIES;
        string _path;
        BookmarkStoreResizeHandler _resizeHandler = null;
        private int _serverVersion = Client.DEFAULT_SERVER_VERSION;
        private BookmarkRangeField _range = new BookmarkRangeField();
        private Object internalLock = new Object();

        Pool<Subscription> _pool;

        /// <summary>
        /// Initializes a new instance of the RingBookmarkStore class with the specified path.
        /// </summary>
        /// <param name="path">The path to the file.</param>
        public RingBookmarkStore(string path)
            : this(path, 1)
        {
        }

        /// <summary>
        /// Initializes a new instance of the RingBookmarkStore class with the specified path
        /// and target number of subscriptions. </summary>
        /// <param name="path">The path to the file.</param>
        /// <param name="targetNumberOfSubscriptions">The target number of subscriptions.</param>
        public RingBookmarkStore(string path, int targetNumberOfSubscriptions)
        {

            try
            {
                _path = path;
                FileInfo fi1 = new FileInfo(_path);
                long initSize = (fi1.Exists && fi1.Length > 0) ? 0 : ENTRIES * ENTRY_SIZE;
                _file = MemoryMappedFile.CreateFromFile(_path, FileMode.OpenOrCreate, null, initSize, MemoryMappedFileAccess.ReadWrite);
                _buffer = _file.CreateViewAccessor();
                _pool = new Pool<Subscription>(targetNumberOfSubscriptions);
            }
            catch (IOException ioex)
            {
                throw new AMPSException("I/O Error initializing file " + path, ioex);
            }
            init();
        }

        /// <summary>
        /// Logs a message and returns the bookmark sequence number.
        /// </summary>
        /// <param name="message">The message to log.</param>
        /// <returns>The bookmark sequence number.</returns>
        public long log(Message message)
        {
            BookmarkField bookmark = (BookmarkField)message.getBookmarkRaw();
            RingBookmarkStore.Subscription sub = (RingBookmarkStore.Subscription)message.getSubscription();
            if (sub == null)
            {
                Field subId = message.getSubIdRaw();
                if (subId == null || subId.isNull())
                    subId = message.getSubIdsRaw();
                sub = find(subId);
                message.setSubscription(sub);
            }
            long index = sub.log(bookmark);
            message.setBookmarkSeqNo(index);
            return index;
        }

        /// <summary>
        /// Discards a message with the specified subId and bookmark sequence number.
        /// </summary>
        /// <param name="subId">The subscription ID.</param>
        /// <param name="bookmarkSeqNo">The bookmark sequence number.</param>
        public void discard(Field subId, long bookmarkSeqNo)
        {
            find(subId).discard(bookmarkSeqNo);
        }

        /// <summary>
        /// Discards the message.
        /// </summary>
        /// <param name="message">The message to discard.</param>
        public void discard(Message message)
        {
            long bookmark = message.getBookmarkSeqNo();
            RingBookmarkStore.Subscription sub = (RingBookmarkStore.Subscription)message.getSubscription();
            if (sub == null)
            {
                Field subId = message.getSubIdRaw();
                if (subId == null || subId.isNull())
                    subId = message.getSubIdsRaw();
                sub = find(subId);
                message.setSubscription(sub);
            }
            sub.discard(bookmark);
        }

        /// <summary>
        /// Retrieves the most recent message for the specified subId.
        /// </summary>
        /// <param name="subId">The subscription ID.</param>
        /// <param name="useList">Flag indicating whether to use list.</param>
        /// <returns>The most recent message.</returns>
        public Field getMostRecent(Field subId, bool useList = true)
        {
            return find(subId).getMostRecentList(useList);
        }

        /// <summary>
        /// Checks if the message is discarded.
        /// </summary>
        /// <param name="message">The message to check.</param>
        /// <returns>True if the message is discarded; otherwise, false.</returns>
        public bool isDiscarded(Message message)
        {
            BookmarkField bookmark = (BookmarkField)message.getBookmarkRaw();
            RingBookmarkStore.Subscription sub = (RingBookmarkStore.Subscription)message.getSubscription();
            if (sub == null)
            {
                Field subId = message.getSubIdRaw();
                if (subId == null || subId.isNull())
                    subId = message.getSubIdsRaw();
                sub = find(subId);
                message.setSubscription(sub);
            }
            return sub.isDiscarded(bookmark);
        }

        private void recover()
        {
            int currentEntry = 0;
            for (; currentEntry < ENTRIES; currentEntry++)
            {
                // is it active?
                byte firstByte = _buffer.ReadByte(currentEntry * ENTRY_SIZE);
                if (firstByte != 0)
                {
                    // read it out
                    byte[] id = new byte[Subscription.BYTES_SUBID - 1];
                    _buffer.ReadArray<byte>(currentEntry * ENTRY_SIZE, id, 0, Subscription.BYTES_SUBID - 1);
                    int idLength = 0;
                    for (; idLength < id.Length && id[idLength] != 0; idLength++) ;

                    Field f = new Field(id, 0, idLength);
                    Subscription s = _pool.get();
                    s.init(_buffer, currentEntry * ENTRY_SIZE);
                    try
                    {
                        _map.Add(f, s);
                    }
                    catch (Exception e)
                    {
                        throw new AMPSException("Bookmark store corrupted.", e);
                    }
                    // This is handled in s.init()
                    // s.recover();
                }
                else break;
            }
            if (currentEntry == ENTRIES)
            {
                // todo: resize here
                throw new AMPSException("Unable to allocate space in this bookmark store.");
            }
            _free = currentEntry;
        }

        private Subscription find(Field subId)
        {
            lock (internalLock) {
                if (_map.ContainsKey(subId))
                {
                    return _map[subId];
                }

                if (_free >= ENTRIES)
                {
                    throw new AMPSException("Unable to allocate space in this bookmark store.");
                }
                int pos = _free++;

                Subscription s = _pool.get();
                s.init(_buffer, pos * ENTRY_SIZE);
                s.setResizeHandler(_resizeHandler, this);
                _map.Add(subId.copy(), s);
                for (int i = 0; i < subId.length; i++)
                {
                    _buffer.Write((pos * ENTRY_SIZE) + i, subId.buffer[subId.position + i]);
                }
                return s;
            }
        }

        /// <summary>
        /// Marks the specified bookmark as persisted.
        /// </summary>
        /// <param name="subId">The subscription ID.</param>
        /// <param name="bookmark">The bookmark to mark as persisted.</param>
        public void persisted(Field subId, BookmarkField bookmark)
        {
            // no-op
        }

        /// <summary>
        /// Marks the specified bookmark as persisted.
        /// </summary>
        /// <param name="subId">The subscription ID.</param>
        /// <param name="bookmark">The bookmark to mark as persisted.</param>
        public void persisted(Field subId, long bookmark)
        {
            // no-op
        }

        /// <summary>
        /// Retrieves the oldest bookmark sequence number for the specified
        /// subscription ID.</summary>
        /// <param name="subId">The subscription ID.</param>
        /// <returns>The oldest bookmark sequence number.</returns>
        public long getOldestBookmarkSeq(Field subId)
        {
            return find(subId).getOldestBookmarkSeq();
        }

        /// <summary>
        /// Sets the resize handler for the bookmark store.
        /// </summary>
        /// <param name="handler">The resize handler to set.</param>
        public void setResizeHandler(BookmarkStoreResizeHandler handler)
        {
            lock (internalLock) {
                _resizeHandler = handler;
                foreach (KeyValuePair<Field, Subscription> fieldSub in _map)
                {
                    fieldSub.Value.setResizeHandler(handler, this);
                }
            }
        }

        /// <summary>
        /// Purges the bookmark store.
        /// </summary>
        public void purge()
        {
            lock (internalLock) {
                // This is the quickest way I've found, so far, to zero out the file.
                int longs = (int)_buffer.Capacity / 8;
                for (int i = 0; i < longs; i++)
                {
                    _buffer.Write(i * 8, (long)0);
                }

                _buffer.Flush();
                _map = new Dictionary<Field, Subscription>();
                _free = 0;
                recover();
            }
        }

        /// <summary>
        /// Purges the specified subscription from the bookmark store.
        /// </summary>
        /// <param name="subId_">The subscription ID to purge.</param>
        public void purge(Field subId_)
        {
            lock (internalLock) {
                if (!_map.ContainsKey(subId_)) return;
                Subscription sub = _map[subId_];
                int pos = sub._offset/ENTRY_SIZE;
                // NULL out the subs data
                byte[] buf = new byte[ENTRY_SIZE];
                _buffer.WriteArray<byte>(sub._offset, buf, 0, ENTRY_SIZE);
                // Move all subs back one to fill the hole
                --_free;
                Field subId = new Field();
                for (int j = pos; j < _free; ++j)
                {
                    int i = j * ENTRY_SIZE;
                    // Read the next entry
                    _buffer.ReadArray<byte>((j+1)*ENTRY_SIZE, buf, 0, ENTRY_SIZE);
                    // Write it to current location
                    _buffer.WriteArray<byte>(i, buf, 0, ENTRY_SIZE);
                    int idLength = 0;
                    for (; idLength < Subscription.BYTES_SUBID && buf[idLength] != 0;
                         idLength++);
                    subId.set(buf, 0, idLength);
                    _map[subId]._offset = i;
                }
                // NULL out the end and reset the end
                for (int i = _free*ENTRY_SIZE; i < (_free+1)*ENTRY_SIZE; ++i)
                {
                    _buffer.Write(i, (byte)0);
                }

                _map.Remove(subId_);
            }
        }

        private void init()
        {
            _map = new Dictionary<Field, Subscription>();
            _free = 0;
            recover();
        }

        /// <summary>
        /// Disposes of the resources used by the <see cref="RingBookmarkStore"/>.
        /// </summary>
        public void Dispose()
        {
            _buffer.Dispose();
            _file.Dispose();
        }

        /// <summary>
        /// Sets the server version for the bookmark store.
        /// </summary>
        /// <param name="version">The server version to set.</param>
        public void setServerVersion(int version)
        {
            lock (internalLock) {
                _serverVersion = version;
            }
        }
    }
}
