///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading;
using AMPS.Client.Fields;
using AMPS.Client.Exceptions;

/** Interface to provide a recovery point for a bookmark subscription. Used in
 * the RecoveryPointAdapter interface for updates and recovery.
 */
namespace AMPS.Client
{
    /// <summary>
    /// SOWRecoveryPointAdapter provides functionality for interacting with a SOW (State of the World) topic
    /// to facilitate recovery point adaptation.
    /// </summary>
    public class SOWRecoveryPointAdapter : RecoveryPointAdapter
    {
        // The name of client we're tracking
        private string _trackedName = "";
        // The SOW topic where information is stored
        private string _topic = "/ADMIN/bookmark_store";
        // The name of the SOW topic field for client names
        private string _clientNameField = "clientName";
        // The name of the SOW topic field for sud ids
        private string _subIdField = "subId";
        // The name of the SOW topic field for bookmarks
        private string _bookmarkField = "bookmark";
        // The client used for interacting with the SOW topic
        // This MUST NOT be the same as the tracked client
        private Client _client = null;
        // The MessageStream used during recovery iteration
        private MessageStream _stream = null;
        // The reusable publish Command
        private Command _cmd = new Command("publish");
        // A place to report swallowed exceptions if not thrown
        public delegate void ExceptionListener(Exception e);
        private ExceptionListener _exceptionListener = null;
        // A buffer to use for conversion between SOW data and
        // RecoveryPoint objects
        private Field _serializeField = new Field();
        // The maximum time to wait for a processed ack when querying the
        // SOW during recovery in milliseconds
        private long _timeoutMillis = 0;
        // Indicates if the internal client should be closed during Dispose
        private bool _closeClient = false;
        // Indicates if the recovered bookmarks should have their timestamp
        // included.
        private bool _useTimestamp = false;
        // Indicates if we should throw exceptions rather than report them to
        // the ExceptionListener
        private bool _throwNotListen = false;
        // Make sure updates don't arrive out of order
        private Dictionary<Field, ulong> _sequences = new Dictionary<Field, ulong>();
        private Object _sequencesLock = new Object();

        // A couple constants useful for json de/serialization
        private const int JSON_START = 11; // {, 7", 2:, 1,
        private const int JSON_END = 8; // }, 5", 1:, 1,
        private const int JSON_LEN = JSON_START + JSON_END;
        private const int SUBID_LEN = 64; // rough typical max
        // Allow for a range with 2 ',' 1 ':' and begin/end markers
        private const int JSON_INIT_LEN = JSON_LEN + SUBID_LEN + 5
                                       + (4*BookmarkField.MAX_BOOKMARK_LENGTH);
        private string _initStr = null;
        private int _serializeStart = 0;
        private RecoveryPoint _current = new FixedRecoveryPoint();
        private Field EPOCH_FIELD = new Field("0");

        /// <summary>
        /// Creates a new SOWRecoveryPointAdapter.
        /// </summary>
        /// <param name="client_">The Client or HAClient connected to the AMPS
        /// server containing the SOW where bookmarks are tracked. This client
        /// MUST NOT be the same as the client being tracked.</param>
        /// <param name="trackedClientName_">The name of the client whose
        /// bookmarks are being tracked. </param>
        /// <param name="timeoutMillis_">The maximum amount of time in ms to
        /// wait for sow query completion to occur during recovery and for
        /// publishFlush to complete during Dispose. Use 0 for no timeout.</param>
        /// <param name="closeClient_">If true, then client_ will be closed
        /// during Dispose.</param>
        /// <param name="useTimestamp_">If true, then recovery bookmarks will
        /// also include the timestamp when the last update for each subId
        /// was stored in the SOW.</param>
        /// <param name="throwExceptions_">If true, the exceptions will be
        /// thrown out of this class instead of sending them to any set
        /// ExceptionListener.</param>
        /// <param name="topic_">Sets the name of the bookmark store SOW topic
        /// to be used. It defaults to '/ADMIN/bookmark_store' if null.</param>
        /// <param name="clientNameField_">Sets the name of the field for the
        /// trackedClientName in the bookmark store SOW topic. It defaults to
        /// 'clientName' if null.</param>
        /// <param name="bookmarkField_">Sets the name of the field for the
        /// bookmark in the bookmark store SOW topic. It defaults to
        /// 'bookmark' if null.</param>
        public SOWRecoveryPointAdapter(Client client_,
                                       string trackedClientName_,
                                       bool closeClient_,
                                       bool useTimestamp_,
                                       bool throwExceptions_,
                                       long timeoutMillis_ = 0,
                                       string topic_ = "/ADMIN/bookmark_store",
                                       string clientNameField_ = "clientName",
                                       string subIdField_ = "subId",
                                       string bookmarkField_ = "bookmark")
        {
            if (client_.getName().Equals(trackedClientName_))
            {
                throw new StoreException("SOWRecoveryPointAdapter can't track a client using that same client");
            }
            _client = client_;
            _trackedName = trackedClientName_;
            _timeoutMillis = timeoutMillis_;
            _closeClient = closeClient_;
            _useTimestamp = useTimestamp_;
            _throwNotListen = throwExceptions_;
            _topic = topic_;
            _clientNameField = clientNameField_;
            _bookmarkField = bookmarkField_;
            _cmd.setTopic(_topic);
        }

        /// <summary>
        /// Creates a new SOWRecoveryPointAdapter with default properties for
        /// NOT closing the provided client, using the ExceptionListener,
        /// no timeout for sow query of publshFlush, and a SOW topic of
        /// /ADMIN/bookmark_store with fields clientName, subId, and bookmark.
        /// </summary>
        /// <param name="client_">The Client or HAClient connected to the AMPS
        /// server containing the SOW where bookmarks are tracked. This client
        /// MUST NOT be the same as the client being tracked.</param>
        /// <param name="trackedClientName_">The name of the client whose
        /// bookmarks are being tracked.</param>
        public SOWRecoveryPointAdapter(Client client_,
                                       string trackedClientName_)
        {
            _client = client_;
            _trackedName = trackedClientName_;
        }

        /// <summary>
        /// Sets the ExceptionListener instance used for communicating absorbed exceptions.
        /// </summary>
        /// <param name="exceptionListener">The exception listener instance to invoke for exceptions.</param>
        public void setExceptionListener(ExceptionListener exceptionListener)
        {
            _exceptionListener = exceptionListener;
        }

        /** Process an update to the newest recovery point.
         */
        public void update(RecoveryPoint recoveryPoint_)
        {
            if (recoveryPoint_ is SequencedRecoveryPoint)
            {
                ulong currentSequence = 0;
                lock(_sequencesLock)
                {
                    if (_sequences.TryGetValue(recoveryPoint_.getSubId(),
                                               out currentSequence))
                    {
                        if (currentSequence
                            > ((SequencedRecoveryPoint)recoveryPoint_).getSequence())
                        {
                            return;
                        }
                    }
                    _sequences[recoveryPoint_.getSubId()]
                        = ((SequencedRecoveryPoint)recoveryPoint_).getSequence();
                }
            }
            try
            {
                if (!serialize(recoveryPoint_)) return;
                _cmd.setData(_serializeField.buffer, _serializeField.position,
                             _serializeField.length);
                _client.execute(_cmd);
            }
            catch (Exception e)
            {
                if (_throwNotListen)
                {
                    throw new StoreException($"SOW update Exception {e}", e);
                }
                else
                {
                    absorbedException(new StoreException($"SOW update Exception {e}", e));
                }
            }
        }

        /** Remove all recovery point information.
         */
        public void purge()
        {
            try
            {
                _client.sowDelete(_topic,
                        $"\"/{_clientNameField}\"=\"{_trackedName}\"");
            }
            catch (Exception e)
            {
                if (_throwNotListen)
                {
                    throw;
                }
                else
                {
                    absorbedException(new StoreException($"SOW purge Exception {e}", e));
                }
            }
        }

        /** Remove all recovery point information for a given sub id.
         */
        public void purge(Field subId_)
        {
            try
            {
                _client.sowDelete(_topic,
                        $"\"/{_clientNameField}\"=\"{_trackedName}\" and \"/{_subIdField}\"=\"{subId_}\"");
            }
            catch (Exception e)
            {
                if (_throwNotListen)
                {
                    throw;
                }
                else
                {
                    absorbedException(new StoreException($"SOW purge(subId) Exception {e}", e));
                }
            }
        }

        /** Take any necessary actions to close the underlying storage.
         */
        public void close()
        {
            try
            {
                if (_client != null) _client.publishFlush((int)_timeoutMillis);
            }
            catch (DisconnectedException)
            {
            }
            catch (Exception e)
            {
                if (_throwNotListen)
                {
                    throw;
                }
                else
                {
                    absorbedException(new StoreException($"SOW close Exception {e}", e));
                }
            }
            finally
            {
                if (_closeClient && _client != null)
                {
                    _client.close();
                    _client = null;
                }
            }
        }

        class Enumerator : IEnumerator<RecoveryPoint>, IDisposable,
                                  System.Collections.IEnumerator
        {
            private SOWRecoveryPointAdapter _adapter = null;

            /// <summary>
            /// Constructor for Enumerator class.
            /// </summary>
            /// <param name="adapter_">The SOWRecoveryPointAdapter instance to be used by the enumerator.</param>
            public Enumerator(SOWRecoveryPointAdapter adapter_)
            {
                _adapter = adapter_;
            }

            /// <summary>
            /// Moves the enumerator to the next recovery point.
            /// </summary>
            /// <returns>True if the enumerator was successfully moved to the next recovery point; false if the
            /// enumerator has passed the end of the collection.</returns>
            public bool MoveNext()
            {
                return _adapter.MoveNext();
            }

            /// <summary>
            /// Sets the enumerator to its initial position, which is before the first element in the collection.
            /// </summary>
            public void Reset()
            {
                throw new NotImplementedException();
            }

            /// <summary>
            /// Gets the current recovery point in the collection.
            /// </summary>
            public RecoveryPoint Current
            {
                get { return _adapter.Current; }
            }

            /// <summary>
            /// Gets the current recovery point in the collection.
            /// </summary>
            /// <returns>The current recovery point in the collection.</returns>
            object System.Collections.IEnumerator.Current
            {
                get { return Current; }
            }

            /// <summary>
            /// Performs application-defined tasks associated with freeing, releasing, or resetting
            /// unmanaged resources.
            /// </summary>
            void IDisposable.Dispose()
            {
            }
        }

        /** Return an enumerator over recovered RecoveryPoints.
         */
        public IEnumerator<RecoveryPoint> GetEnumerator()
        {
            Command cmd = new Command("sow");
            cmd.setTopic(_topic).setTimeout(_timeoutMillis)
               .setFilter($"/{_clientNameField} = \"{_trackedName}\"");
            if (_useTimestamp)
            {
                cmd.setOptions($"select=[-/,+/{_subIdField},+/{_bookmarkField}],timestamp");
            }
            else
            {
                cmd.setOptions($"select=[-/,+/{_subIdField},+/{_bookmarkField}]");
            }
            try
            {
                _stream = _client.execute(cmd);
            }
            catch (Exception e)
            {
                if (_throwNotListen)
                {
                    throw new StoreException($"SOW IEnumerator<> Exception {e}", e);
                }
                else
                {
                    absorbedException(new StoreException($"SOW IEnumerator Exception {e}", e));
                }
            }
            return new Enumerator(this);
        }

        /** IEnumerable version of get enumerator.
         */
        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            Command cmd = new Command("sow");
            cmd.setTopic(_topic).setTimeout(_timeoutMillis)
               .setFilter($"/{_clientNameField} = {_trackedName}");
            if (_useTimestamp)
            {
                cmd.setOptions($"select=[-/,+/{_subIdField},+/{_bookmarkField}],timestamp");
            }
            else
            {
                cmd.setOptions($"select=[-/,+/{_subIdField},+/{_bookmarkField}]");
            }
            try
            {
                _stream = _client.execute(cmd);
            }
            catch (Exception e)
            {
                if (_throwNotListen)
                {
                    throw;
                }
                else
                {
                    absorbedException(new StoreException($"SOW IEnumerator Exception {e}", e));
                }
            }
            return new Enumerator(this);
        }

        /** IEnumerator move to next item.
         */
        public bool MoveNext()
        {
            try
            {
                if (!_stream.MoveNext())
                {
                    // Recovery complete, prepare to send publish commands
                    initSerialization();
                    _current = null;
                    return false;
                }
                Message m = _stream.Current;
                if (m != null && m.getCommand() == Message.Commands.GroupBegin)
                {
                    return MoveNext();
                }
                if (m == null || m.getCommand() == Message.Commands.GroupEnd
                    || m.getCommand() == Message.Commands.Ack)
                {
                    // Recovery complete, prepare to send publish commands
                    initSerialization();
                    _current = null;
                    return false;
                }
                if (!deserialize(m))
                {
                    return MoveNext();
                }
                return true;
            }
            catch (Exception e)
            {
                if (_throwNotListen)
                {
                    // Recovery complete, prepare to send publish commands
                    initSerialization();
                    _current = null;
                    throw new StoreException($"SOW MoveNext Exception {e}", e);
                }
                else
                {
                    absorbedException(new StoreException($"SOW MoveNext Exception {e}", e));
                }
                // Recovery complete, prepare to send publish commands
                initSerialization();
                _current = null;
                return false;
            }
        }

        /** IEnumerator reset - not implemented.
         */
        public void Reset()
        {
            throw new NotImplementedException();
        }

        /** Dispose of the adapter. Make sure everything is saved and close the
         *  client connection if requested during construction.
         */
        void IDisposable.Dispose()
        {
            close();
            _client = null;
        }

        /** The current RecoveryPoint from SOW recovery for the client.
         */
        public RecoveryPoint Current
        {
            get { return _current; }
        }

        /** IEnumerator version of current.
         */
        object System.Collections.IEnumerator.Current
        {
            get { return Current; }
        }

        protected virtual bool serialize(RecoveryPoint recoveryPoint_)
        {
            Field subId = recoveryPoint_.getSubId();
            BookmarkField bookmark = recoveryPoint_.getBookmark();
            if (bookmark.Equals(EPOCH_FIELD))
            {
                return false;
            }
            int endLength = subId.length + _bookmarkField.Length
                          + bookmark.length + JSON_END;
            int length = _clientNameField.Length + _trackedName.Length
                         + _subIdField.Length + JSON_START + endLength;
            if (_serializeField.buffer == null
                || _serializeField.buffer.Length < length)
            {
                // Need a new buffer
                int newLength = length + (128-(length%128));
                byte[] buffer = new byte[newLength];
                _serializeField.set(buffer, 0, length);
                initSerialization();
            }
            Array.Copy(subId.buffer, subId.position,
                       _serializeField.buffer, _serializeStart, subId.length);
            Array.Copy($"\",\"{_bookmarkField}\":\"{bookmark.ToString()}\"}}".getBytes(),
                       0, _serializeField.buffer,
                       _serializeStart + subId.length, endLength-subId.length);
            _serializeField.length=length;
            return true;
        }

        protected virtual bool deserialize(Message m_)
        {
            int subIdIdx = 0;
            int bookmarkIdx = 0;
            string subId = "";
            string bookmark = "";
            bool nextIsSubId = false;
            bool nextIsBookmark = false;
            char[] delims = { '{', '"' };
            string data = m_.getData();
            string[] values = data.Split(delims);
            int index = 0;
            foreach (string val in values)
            {
                ++index;
                val.Trim();
                if (val.Length == 0)
                {
                    continue;
                }
                // Both : and , are used in json and bookmarks
                else if ((val.Contains(":") || val.Contains(","))
                         && val.Length == 1)
                {
                    // Just a : or , and not a range
                    continue;
                }
                else if (nextIsSubId)
                {
                    subIdIdx = index-1;
                    subId = val;
                    nextIsSubId = false;
                }
                else if (nextIsBookmark)
                {
                    bookmarkIdx = index-1;
                    bookmark = val;
                    nextIsBookmark = false;
                }
                else if (val.Equals(_subIdField))
                {
                    nextIsSubId = true;
                }
                else if (val.Equals(_bookmarkField))
                {
                    nextIsBookmark = true;
                }
            }
            if (_useTimestamp)
            {
                if ((bookmark[0] != '[' && bookmark[0] != '(')
                    && m_.getTimestampRaw() != null
                    && !m_.getTimestampRaw().isNull())
                {
                    bookmark = $"{bookmark},{m_.getTimestamp()}";
                }
            }
            int len = subId.Length + bookmark.Length;
            if (_serializeField.buffer == null
                || _serializeField.buffer.Length < len)
            {
                byte[] buffer = new byte[len];
                _serializeField.set(buffer, 0, len);
            }
            Array.Copy($"{subId}{bookmark}".getBytes(),
                       0, _serializeField.buffer, 0, len);
            _current.getSubId().set(_serializeField.buffer, 0,
                                    subId.Length);
            _current.getBookmark().set(_serializeField.buffer, subId.Length,
                                       bookmark.Length);
            return true;
        }

        protected virtual void initSerialization()
        {
            if (_serializeField.buffer == null
                || _serializeField.buffer.Length < JSON_INIT_LEN)
            {
                byte[] buffer = new byte[JSON_INIT_LEN];
                _serializeField.set(buffer, 0, JSON_INIT_LEN);
            }
            if (_initStr == null)
            {
                _initStr = $"{{\"{_clientNameField}\":\"{_trackedName}\",\"{_subIdField}\":\"";
                _serializeStart = _initStr.Length;
            }
            Array.Copy(_initStr.getBytes(), 0,
                       _serializeField.buffer, 0, _initStr.Length);
        }

        private void absorbedException(Exception e)
        {
            if (_exceptionListener != null)
            {
                try
                {
                    _exceptionListener(e);
                }
                catch(Exception) { }
            }
        }
    }
}