////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

namespace AMPS.Client
{

    /**
    * Replay the messages saved in a store. Provided to the
    * <see cref="Store.replay"/> or <see cref="Store.replaySingle"/> methods
    * to replay stored operations. The StoreReplayer is responsible for
    * publishing the operations to AMPS.
    */
    public interface StoreReplayer
    {

        /**
         * Replay the specified message.
         * <param name="message">the Message to replay</param>
         */

        void execute(Message message);
    }

    /// <summary>
    /// SOWDeleteTypes provides constants representing different types of delete
    /// operations in a SOW (State of the World) topic.
    /// </summary>
    public sealed class SOWDeleteTypes
    {
        /// <summary>
        /// Delete operation by data.
        /// </summary>
        public const int ByData = 1;

        /// <summary>
        /// Delete operation by filter.
        /// </summary>
        public const int ByFilter = 2;

        /// <summary>
        /// Delete operation by keys.
        /// </summary>
        public const int ByKeys = 4;

        /// <summary>
        /// Delete operation by bookmark.
        /// </summary>
        public const int ByBookmark = 8;

        /// <summary>
        /// Cancel delete operation by bookmark.
        /// </summary>
        public const int ByBookmarkCancel = 16;
    }

    /**
     * Represents a message store. The AMPS client uses message stores for
     * recovery purposes. The store is responsible for maintaining the state
     * of published messages and recovering that state in the event of a
     * disconnection. Optionally, the store may persist message state and
     * recover that state if the application restarts.
     *
     */
    public interface Store : System.IDisposable
    {

        /**
         * Store the provided message. The message sequence is the key
         * the client can later use to replay the operation or remove the
         * operation from the store. Implementations may assume that the index
         * increases monotonically.
         * <param name="message">The message to store.</param>
         */
        void store(Message message);

        /**
         * Discard messages in the store, up to and including the message
         * specified by the passed in sequnence number. This is called by the
         * client when it's notified a message is persisted on the server.
         * <param name="sequence">The highest sequence number of the messages
         *     that can be safely discarded. The specified message and all prior
         *     messages are removed from the store.</param>
         */
        void discardUpTo(ulong sequence);

        /**
         * Replay all unpersisted messages in the store using the provided StoreReplayer.
         * <param name="replayer">The StoreReplayer used to replay undiscarded messages.</param>
         */
        void replay(StoreReplayer replayer);

        /**
         * Replay the unpersisted message with the specified sequence.
         * <param name="replayer">The StoreReplayer used to replay the specified undiscarded message.</param>
         * <param name="sequence">The sequence number of the undiscarded message to replay.</param>
         */
        bool replaySingle(StoreReplayer replayer, ulong sequence);

        /**
         * Return the number of unpersisted messages in the store.
         */
        long unpersistedCount();

        /**
         * Return the sequence of the last persisted message.
         */
        ulong getLastPersisted();

        /**
         * Wait for the store to discard everything that has been stored up
         * to the point in time when flush is called, then return.
         */
        void flush();

        /**
         * Wait for the store to discard everything that has been stored up
         * to the point in time when flush is called, then return. Throw an
         * exception if this is not completed in the number of milliseconds
         * specified by the timeout.
         * <param name="timeout">the number of milliseconds to wait for the flush</param>
         * <exception cref="TimedOutException">If the timeout period passes without the
         *     messages being discarded.</exception>
         */
        void flush(int timeout);

        /**
         * Used internally by the Client to give the store a Message
         * implementing the correct protocol for the connection. This
         * Message instance is also used for replay operations.
         */
        void setMessage(Message message);

        /**
         * Set the resize handler for the publish store. An implementation can
         * call the resize handler in the event that it needs to grow the
         * size of the publish store.
         */
        void setResizeHandler(PublishStoreResizeHandler handler);

        /**
         * Set if a PublishStoreGapException should be thrown by the Store if
         * there is an attempt to call `discardUpTo` with a sequence number
         * lower than the last persisted sequence number. This could occur
         * during logon after a failover event if the new server was not
         * current on messages due to being an async replication destination.
         * <param name="errorOnPublishGap_">Indicates if the exception should be thrown</param>
         */
        void setErrorOnPublishGap(bool errorOnPublishGap_);

        /**
         * Get if a PublishStoreGapException can be thrown by the Store if
         * there is an attempt to call `discardUpTo` with a sequence number
         * lower than the last persisted sequence number. This could occur
         * during logon after a failover event if the new server was not
         * current on messages due to being an async replication destination.
         * <returns>true if the exception will be thrown</returns>
         */
        bool getErrorOnPublishGap();
    }
}
