﻿////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2012-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.Security;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Text;

namespace AMPS.Client
{
    /// <summary>
    /// Implements a secure transport for communicating with AMPS
    /// over a SslStream.
    /// <see cref="System.Net.Security.SslStream"/>
    /// </summary>
    public class TCPSTransport : TCPTransport
    {
        /// <summary>
        /// Construct a TCPSTransport object for the given Protocol.
        /// </summary>
        /// <param name="protocol">The protocol for communcating with AMPS.</param>
        /// <param name="impl">An instance of TCPSTransportImpl</param>
        public TCPSTransport(Protocol protocol, TCPSTransportImpl impl)
            : base(protocol, impl)
        {
        }

        /// <summary>
        /// Construct a TCPSTransport object for the given Protocol.
        /// </summary>
        /// <param name="protocol">The protocol for communicating with AMPS.</param>
        /// <param name="properties">Custom properties used by the underlying
        ///  TCP transport.</param>
        public TCPSTransport(Protocol protocol, Properties properties)
            : base(protocol, properties)
        {
        }

        /// <summary>
        /// Gets or sets the callback delegate used by SslStream to validate
        /// the remote certificate from AMPS.
        /// <seealso cref="SslStream.RemoteCertificate"/>
        /// </summary>
        public static RemoteCertificateValidationCallback RemoteCertificateValidationCallback
        {
            get; set;
        }

        // NOTE: Per Microsoft API docs, selecting None allows the OS to pick properly:
        // https://docs.microsoft.com/en-us/dotnet/api/system.security.authentication.sslprotocols?view=netframework-4.7
        static SslProtocols _sslProtocols = SslProtocols.None;

        /// <summary>
        /// Gets or sets the enabled SSL protocols used for connecting to AMPS.
        /// <seealso cref="SslStream.SslProtocol"/>
        /// </summary>
        public static SslProtocols EnabledSslProtocols
        {
            get
            {
                return _sslProtocols;
            }
            set
            {
                _sslProtocols = value;
            }
        }

        /// <summary>
        /// Gets or sets whether certificate revocation should be checked.
        /// <seealso cref="SslStream.CheckCertRevocationStatus"/>
        /// </summary>
        public static bool CheckCertificateRevocation
        {
            get; set;
        }

        /// <summary>
        /// Creates the underlying implementation of the TCPSTransport class.
        /// </summary>
        /// <remarks>
        /// This method initializes the internal implementation (<see cref="TCPSTransportImpl"/>)
        /// based on the specified protocol and properties.
        /// </remarks>
        protected override void createImpl()
        {
            this.impl = new TCPSTransportImpl(this.protocol, this.properties);
        }

    }
}
