////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2012-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.IO;
using System.Net;
using System.Net.Sockets;

using AMPS.Client.Exceptions;

namespace AMPS.Client
{
    /// <summary>
    /// Implements an AMPS transport over standard TCP/IP.
    /// Use this transport by connecting to URIs beginning with {@code tcp://}.
    /// </summary>
    public class TCPTransport : Transport
    {
        protected TCPTransportImpl impl = null;
        protected Protocol protocol = null;
        protected Uri uri = null;
        protected object sendLock = new object();
        protected MemoryStream sendBuffer = new MemoryStream(4096);
        protected Properties properties = null;
        protected static int _defaultConnectTimeout = 0;

        /// <summary>
        /// Constructs a new instance of TCPTransport with the specified message type and properties.
        /// </summary>
        /// <param name="msgType">The protocol message type for the transport.</param>
        /// <param name="properties">The properties to configure the transport.</param>
        public TCPTransport(Protocol msgType, Properties properties)
        {
            this.properties = properties;
            this.protocol = msgType;
            createImpl();
        }

        /// <summary>
        /// Constructs a new instance of TCPTransport with the specified message type and existing implementation.
        /// </summary>
        /// <param name="msgType">The protocol message type for the transport.</param>
        /// <param name="impl">An existing implementation of TCPTransportImpl.</param>
        public TCPTransport(Protocol msgType, TCPTransportImpl impl)
            : this(msgType, new Properties())
        {
            this.impl = impl;
        }

        /// <summary>
        /// Creates the underlying implementation of the transport.
        /// </summary>
        protected virtual void createImpl()
        {
            this.impl = new TCPTransportImpl(this.protocol, this.properties);
        }

        /// <summary>
        /// Factory method for creating a new instance of TCPTransport with a specific implementation.
        /// </summary>
        /// <param name="messageType">The protocol message type for the transport.</param>
        /// <param name="impl">An existing implementation of TCPTransportImpl.</param>
        /// <returns>A new instance of TCPTransport with the specified protocol and implementation.</returns>
        public static TCPTransport createTransport(Protocol messageType, TCPTransportImpl impl)
        {
            return new TCPTransport(messageType, impl);
        }

        /// <summary>
        /// Sets the default connect timeout used for new connections.
        /// </summary>
        /// <param name="defaultConnectTimeoutMillis_">The default connect timeout, in milliseconds.
        /// A 0 value means no timeout is specified, i.e. wait until the OS returns an error.</param>
        public static void setDefaultConnectTimeout(int defaultConnectTimeoutMillis_)
        {
            TCPTransport._defaultConnectTimeout = defaultConnectTimeoutMillis_;
        }

        /// <summary>
        /// Returns the default connect timeout used for new connections.
        /// </summary>
        /// <returns>
        /// The default connect timeout in milliseconds. A 0 value means
        /// no timeout is specified, i.e. wait until the OS returns an error.
        /// </returns>
        public static int getDefaultConnectTimeout()
        {
            return TCPTransport._defaultConnectTimeout;
        }

        /// <summary>
        /// Gets the action to be performed during idle periods.
        /// </summary>
        /// <returns>The action to be performed during idle periods.</returns>
        public Action getIdleAction()
        {
            return impl.getIdleAction();
        }

        /// Sets the action to be performed during idle periods.
        /// </summary>
        /// <param name="idleAction">The action to be performed during idle periods.</param>
        public void setIdleAction(Action idleAction)
        {
            impl.setIdleAction(idleAction);
        }

        /// <summary>
        /// Sets the message handler for the transport.
        /// </summary>
        /// <param name="ml">The message handler to set.</param>
        public void setMessageHandler(MessageHandler ml)
        {
            this.impl.setMessageHandler(ml);
        }

        /// <summary>
        /// Sets the handler for disconnect events.
        /// </summary>
        /// <param name="dh">The disconnect handler to set.</param>
        public void setDisconnectHandler(TransportDisconnectHandler dh)
        {
            this.impl.setDisconnectHandler(dh);
        }

        /// <summary>
        /// Sets the listener for exception events.
        /// </summary>
        /// <param name="exceptionListener">The exception listener to set.</param>
        public void setExceptionListener(TCPTransportImpl.ExceptionListener exceptionListener)
        {
            this.impl.setExceptionListener(exceptionListener);
        }

        /// <summary>
        /// Sets the handler for newly created threads.
        /// </summary>
        /// <param name="handler">The thread created handler to set.</param>
        public void setThreadCreatedHandler(ThreadCreatedHandler handler)
        {
            impl.setThreadCreatedHandler(handler);
        }

        /// <summary>
        /// Sets the transport filter.
        /// </summary>
        /// <param name="filter_">The transport filter to set.</param>
        public void setTransportFilter(TransportFilter filter_)
        {
            if (filter_ == null) {
                impl.setTransportFilter(null);
            }
            else {
                impl.setTransportFilter(new TransportFilterWrapper(filter_));
            }
        }

        /// <summary>
        /// Sets the transport filter.
        /// </summary>
        /// <param name="filter_">The modifiable transport filter to set.</param>
        public void setTransportFilter(TransportFilterModifiable filter_)
        {
            impl.setTransportFilter(filter_);
        }

        /// <summary>
        /// Sets the read timeout for the transport.
        /// </summary>
        /// <param name="readTimeoutMillis_">The read timeout in milliseconds.</param>
        public void setReadTimeout(int readTimeoutMillis_)
        {
            this.impl.setReadTimeout(readTimeoutMillis_);
        }

        /// <summary>
        /// Connects to the specified URI using the underlying implementation.
        /// </summary>
        /// <param name="uri">The URI to connect to.</param>
        public void connect(Uri uri)
        {
            this.uri = uri;
            this.impl.connect(uri);
        }

        /// <summary>
        /// Closes the transport, disconnecting it from the server.
        /// </summary>
        public void close()
        {
            this.impl.disconnect();
        }

        /// <summary>
        /// Disconnects the transport from the server.
        /// </summary>
        public void disconnect()
        {
            this.impl.disconnect();
        }

        /// <summary>
        /// Handles a close event, notifying the underlying implementation.
        /// </summary>
        /// <param name="failedVersion">The failed version.</param>
        /// <param name="message">The message associated with the close event.</param>
        /// <param name="exception">The exception associated with the close event.</param>
        public void handleCloseEvent(int failedVersion, String message, Exception exception)
        {
            this.impl.handleCloseEvent(failedVersion, message, exception);
        }

        /// <summary>
        /// Sends a message without retrying, handling the serialization process.
        /// </summary>
        /// <param name="message">The message to send.</param>
        public void sendWithoutRetry(Message message)
        {
            lock (sendLock)
            {
                // Reserve space for a 4-byte int at the
                // beginning for the size of the message
                sendBuffer.SetLength(4);
                sendBuffer.Position = 4;

                Message.SerializationResult sr = message.serialize(sendBuffer);
                if (sr == Message.SerializationResult.BufferTooSmall)
                {
                    // Buffer was too small, let's resize and retry
                    sendBuffer = new MemoryStream(2 * sendBuffer.Capacity);
                    sendWithoutRetry(message);
                    return;
                }
                else if (sr == Message.SerializationResult.OK)
                {
                    int length = (int)sendBuffer.Position - 4;
                    Utilities.writeInt32(length, sendBuffer.GetBuffer(), 0);
                    sendBuffer.Position = 0;
                    // Write buffer to socket
                    this.impl.send(sendBuffer);
                }
            }
        }

        /// <summary>
        /// Sends the specified message. This method handles synchronization, retries,
        /// and exception handling.
        /// </summary>
        /// <param name="message">The message to send.</param>
        public void send(Message message)
        {
            lock (sendLock)
            {
                while (true)
                {
                    try
                    {
                        sendWithoutRetry(message);
                        return;
                    }
                    catch (DisconnectedException exception)
                    {
                        try
                        {
                            handleCloseEvent(getVersion(), "Disconnected", exception);
                        }
                        catch (RetryOperationException)
                        {
                            ;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Allocates a new message using the underlying protocol.
        /// </summary>
        /// <returns>The allocated message.</returns>
        public Message allocateMessage()
        {
            return this.protocol.allocateMessage();
        }

        /// <summary>
        /// Returns the size of the write queue.
        /// </summary>
        /// <returns>The size of the write queue.</returns>
        /// <exception cref="DisconnectedException">If the transport is disconnected.</exception>
        public long writeQueueSize()
        {
            try
            {
                return this.impl.writeQueueSize();
            }
            catch (NullReferenceException)
            {
                throw new DisconnectedException("not connected");
            }
        }

        /// <summary>
        /// Returns the size of the read queue.
        /// </summary>
        /// <returns>The size of the read queue.</returns>
        /// <exception cref="DisconnectedException">If the transport is disconnected.</exception>
        public long readQueueSize()
        {
            try
            {
                return this.impl.writeQueueSize();
            }
            catch (NullReferenceException)
            {
                throw new DisconnectedException("not connected");
            }
        }

        /// <summary>
        /// Flushes the transport's write queue.
        /// </summary>
        /// <returns>The number of messages flushed.</returns>
        /// <exception cref="DisconnectedException">If the transport is disconnected.</exception>
        public long flush()
        {
            try
            {
                return this.impl.flush();
            }
            catch (NullReferenceException)
            {
                throw new DisconnectedException("not connected");
            }
        }

        /// <summary>
        /// Flushes the transport's write queue with a specified timeout.
        /// </summary>
        /// <param name="timeout">The maximum time to wait for flushing, in milliseconds.</param>
        /// <returns>The number of messages flushed.</returns>
        /// <exception cref="DisconnectedException">If the transport is disconnected.</exception>
        public long flush(long timeout)
        {
            try
            {
                return this.impl.flush(timeout);
            }
            catch (NullReferenceException)
            {
                throw new DisconnectedException("not connected");
            }
        }

        /// <summary>
        /// When this method is called, it internally invokes  the 'socket()'
        /// method of TCPTransportImpl class (the underlying implementation).
        /// This method in TCPTransportImpl is responsible for providing access
        /// to the actual 'Socket' instance used for communication.
        /// </summary>
        /// <returns>The underlying socket.</returns>
        public Socket socket()
        {
            try
            {
                return this.impl.socket();
            }
            catch (Exception)
            {
                return null;
            }
        }

        /// <summary>
        /// Returns the version of the transport.
        /// </summary>
        /// <returns>The version of the transport.</returns>
        public int getVersion()
        {
            return this.impl._connectionVersion;
        }

        /// <summary>
        /// Initializes the transport from the provided client.
        /// </summary>
        /// <param name="client_">The client to initialize from.</param>
        public void initFromClient(Client client_)
        {
            this.impl.initFromClient(client_);
        }
    }
}
