////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;

using AMPS.Client.Exceptions;

namespace AMPS.Client
{
    /// <summary>
    /// Socket implementation for {@link TCPTransport}. TCP Transport delegates to an instance
    /// of this class. This is used internally by the client's transport. There is usually no
    /// reason to make direct use of it. Its functionality is best accessed via the client instance.
    /// </summary>
    public class TCPTransportImpl
    {
        protected Client _client = null;
        protected Uri _addr;
        protected IPAddress _resolvedAddr;
        protected Stream _socketStream = null;
        protected Socket _socket = null;
        public readonly object _lock = new object();
        public volatile int _connectionVersion = 0;
        private volatile bool _disconnecting = false;
        private Protocol _messageType = null;
        private MessageHandler _onMessage = DefaultMessageHandler.instance;
        private TransportDisconnectHandler _onDisconnect = DefaultDisconnectHandler.instance;
        private TCPReaderThread _readerThread = null;
        public delegate void ExceptionListener(Exception e);
        private ExceptionListener _exceptionListener = null;
        private Properties _properties;
        private int _readTimeout = 0;
        private int _connectTimeout = 0;
        private int _idleReadTimeout = 1000; // 1 second default
        private TransportFilterModifiable _filter = null;
        private volatile Action _idleAction = null;
        private ThreadCreatedHandler _threadCreatedHandler = null;
        protected bool _httpPreflight = false;
        protected bool _handshakeSetStream = false;
        private List<String> _httpPreflightHeaders = null;

        /// <summary>
        /// Initializes a new instance of the <see cref="TCPTransportImpl"/> class with
        /// the specified message type and properties.
        /// </summary>
        /// <param name="messageType_">The message type for the transport.</param>
        /// <param name="properties_">The properties for configuring the transport.</param>
        public TCPTransportImpl(Protocol messageType_, Properties properties_)
        {
            this._messageType = messageType_;
            this._properties = properties_;
            this._connectTimeout = TCPTransport.getDefaultConnectTimeout();
        }

        /// <summary>
        /// Sets the transport filter for the TCP transport.
        /// </summary>
        /// <param name="filter_">The transport filter to set.</param>
        public void setTransportFilter(TransportFilterModifiable filter_)
        {
            _filter = filter_;
        }

        /// <summary>
        /// Sets the read timeout for the TCP transport.
        /// </summary>
        /// <param name="readTimeoutMillis_">The read timeout in milliseconds.</param>
        public void setReadTimeout(int readTimeoutMillis_)
        {
            _readTimeout = readTimeoutMillis_;
        }

        /// <summary>
        /// Sets the message handler for the TCP transport.
        /// </summary>
        /// <param name="h">The message handler to set.</param>
        public void setMessageHandler(MessageHandler h)
        {
            this._onMessage = h;
        }

        /// <summary>
        /// Sets the disconnect handler for the TCP transport.
        /// </summary>
        /// <param name="h">The disconnect handler to set.</param>
        public void setDisconnectHandler(TransportDisconnectHandler h)
        {
            this._onDisconnect = h;
        }

        /// <summary>
        /// Sets the exception listener for the TCP transport.
        /// </summary>
        /// <param name="e">The exception listener to set.</param>
        public void setExceptionListener(ExceptionListener e)
        {
            _exceptionListener = e;
        }

        /// <summary>
        /// Broadcasts an exception to the exception listener if available.
        /// </summary>
        /// <param name="e">The exception to broadcast.</param>
        protected void broadcastException(Exception e)
        {
            if (_exceptionListener != null)
            {
                try
                {
                    _exceptionListener.Invoke(e);
                }
                catch (Exception) { /* ignore */ }
            }
        }

        /// <summary>
        /// Sets the thread created handler for the TCP transport.
        /// </summary>
        /// <param name="handler">The thread created handler to set.</param>
        public void setThreadCreatedHandler(ThreadCreatedHandler handler)
        {
            _threadCreatedHandler = handler;
        }

        /// <summary>
        /// Gets the idle action for the TCP transport.
        /// </summary>
        /// <returns>The action to be performed during idle periods.</returns>
        public Action getIdleAction()
        {
            return _idleAction;
        }

        /// Sets the idle action for the TCP transport.
        /// </summary>
        /// <param name="idleAction">The idle action to set.</param>
        public void setIdleAction(Action idleAction)
        {
            _idleAction = idleAction;
        }

        /// <summary>
        /// Establishes a connection to the specified address.
        /// </summary>
        /// <param name="addr">The address to connect to.</param>
        /// <exception cref="AlreadyConnectedException"></exception>
        /// <exception cref="InvalidURIException"></exception>
        /// <exception cref="ConnectionRefusedException"></exception>
        public void connect(Uri addr)
        {
            lock (_lock)
            {
                _disconnecting = false;
                try
                {
                    if (this._addr != null)
                    {
                        throw new AlreadyConnectedException("Already connected to AMPS at " +
                           this._addr.Host + ":" + this._addr.Port + "\n");
                    }
                    _addr = addr;
                    createSocket();
                    URIProperties p = new URIProperties(addr);
                    applySocketProperties(_properties != null ? p.Union(_properties.Where(k => !p.ContainsKey(k.Key))) : p);
                    connectSocket(addr);
                    handshake();
                    _socket.ReceiveTimeout = _idleReadTimeout;
                    createStream(addr);

                    string compressionType;
                    if (p.TryGetValue("compression", out compressionType))
                    {
                        switch (compressionType)
                        {
                            case "zlib":
                                _socketStream = new ZlibIOStream(_socketStream, -15);
                                break;
                            default:
                                throw new InvalidURIException("Compression type '" + compressionType + "' is unsupported.");
                        }
                    }
                    _connectionVersion++;
                    _readerThread = new TCPReaderThread(this, this._messageType);
                }
                catch (AlreadyConnectedException)
                {
                    throw;
                }
                catch (InvalidURIException)
                {
                    _addr = null;
                    throw;
                }
                catch (ThreadInterruptedException e)
                {
                    _addr = null;
                    throw new ConnectionRefusedException("Interrupted, but please try again.", e);
                }
                catch (Exception e)
                {
                    _addr = null;
                    throw new ConnectionRefusedException("Unable to connect to AMPS at " +
                          addr.Host + ":" + addr.Port, e);
                }
            }

        }

        /// <summary>
        /// Resolves the IP address for the provided URI.
        /// </summary>
        /// <param name="addr">The URI to resolve.</param>
        /// <returns>The resolved IP address.</returns>
        /// <exception cref="InvalidURIException">Thrown when the URI hostname cannot be resolved.</exception>
        protected virtual IPAddress resolveAddress(Uri addr)
        {
            if (_resolvedAddr == null || addr.DnsSafeHost != _addr.DnsSafeHost)
            {
                #if AMPS_DEFAULT_IPV6
                    String ip_proto_prefer = "ipv6";
                #else
                    String ip_proto_prefer = "ipv4";
                #endif
                URIProperties p = new URIProperties(addr);
                if (p.ContainsKey("ip_protocol_prefer"))
                {
                    ip_proto_prefer = p["ip_protocol_prefer"];
                }
                else if (_properties.ContainsKey("ip_protocol_prefer"))
                {
                    ip_proto_prefer = _properties["ip_protocol_prefer"];
                }
                if(!(ip_proto_prefer == "ipv4" || ip_proto_prefer == "ipv6"))
                {
                    throw new InvalidURIException("Invalid value '" + ip_proto_prefer + "' for ip_protocol_prefer URI parameter");
                }
                IPAddress resolvedAddr = null;
                // https://github.com/dotnet/corefx/issues/26840
                IPAddress[] addresses = Dns.GetHostAddresses(addr.DnsSafeHost);
                if (addresses.Length > 0)
                {
                    var preferredFam = ip_proto_prefer == "ipv4" ? AddressFamily.InterNetwork : AddressFamily.InterNetworkV6;
                    foreach (var candidate in addresses)
                    {
                      if (candidate.AddressFamily == preferredFam)
                      {
                        resolvedAddr = candidate; break;
                      }
                      else if (resolvedAddr == null)
                      {
                        resolvedAddr = candidate; // but keep looking for preferred fam
                      }
                    }
                }
                if(resolvedAddr == null)
                {
                    throw new InvalidURIException("Hostname " + _addr.DnsSafeHost + " is unable to be resolved");
                }
                _resolvedAddr = resolvedAddr;
            }
            return _resolvedAddr;
        }

        /// <summary>
        /// Creates a socket based on the resolved address.
        /// </summary>
        protected virtual void createSocket()
        {
            IPAddress resolvedAddr = resolveAddress(_addr);
            _socket = new Socket(resolvedAddr.AddressFamily, SocketType.Stream, ProtocolType.Tcp);
        }

        /// <summary>
        /// Creates a stream for the provided URI and connects the socket.
        /// </summary>
        /// <param name="uri">The URI to create the stream for.</param>
        protected virtual void createStream(Uri uri)
        {
            if (!_handshakeSetStream)
            {
                _socketStream = new NetworkStream(_socket);
            }
            else
            {
                _handshakeSetStream = false;
            }
        }

        /// <summary>
        /// Connects the socket to the specified URI.
        /// </summary>
        /// <param name="addr">The URI to connect to.</param>
        /// <exception cref="SocketException">Thrown when an error occurs while connecting the socket.</exception>
        protected virtual void connectSocket(Uri addr)
        {
            if(addr.DnsSafeHost != _addr.DnsSafeHost)
            {
                _addr = addr;
                createSocket();
                URIProperties p = new URIProperties(addr);
                applySocketProperties(_properties != null ? p.Union(_properties.Where(k => !p.ContainsKey(k.Key))) : p);
            }
            // Set to a default socket read timeout of 1000, until after handshake,
            // when any user-specified idle read timeout can override this.
            _socket.ReceiveTimeout = 1000;
            IPAddress resolvedAddr = resolveAddress(addr);
            // If a connection timeout is set, use it.
            if (_connectTimeout > 0)
            {
                IAsyncResult ar = _socket.BeginConnect(resolvedAddr, addr.Port, null, null);
                using (ar.AsyncWaitHandle)
                {
                    bool isTimeout = !ar.AsyncWaitHandle.WaitOne(_connectTimeout);
                    if (isTimeout && !ar.IsCompleted)
                    {
                        _socket.Close();
                        throw new SocketException((int) SocketError.TimedOut);
                    }
                    _socket.EndConnect(ar);
                }
            }
            else
            {
                // https://github.com/dotnet/corefx/issues/5829
                _socket.Connect(resolvedAddr, addr.Port);
            }
        }

        /// <summary>
        /// Initiates the handshake process for the transport.
        /// </summary>
        protected void handshake()
        {
            if (_httpPreflight)
            {
                createStream(_addr);
                _handshakeSetStream = true;
                StringBuilder getMsg = new StringBuilder($"GET {_addr.AbsolutePath} HTTP/1.1\r\nHost: {_addr.Host}\r\nConnection: upgrade\r\nUpgrade: {_addr.Scheme}\r\n");
                if (_httpPreflightHeaders != null)
                {
                    foreach (string header in _httpPreflightHeaders)
                    {
                        getMsg.Append(header).Append("\r\n");
                    }
                }
                getMsg.Append("\r\n");
                byte[] getMsgBytes = Encoding.UTF8.GetBytes(getMsg.ToString());
                _socketStream.Write(getMsgBytes, 0, getMsgBytes.Length);
                const int maxReadSize = 2048;
                byte[] getResponse = new byte[maxReadSize];
                int readBytes = _socketStream.Read(getResponse, 0, maxReadSize);
                int totalBytes = readBytes;
                // Make sure we have enough of the response
                while (totalBytes < 15)
                {
                    if (readBytes == -1 || readBytes == 0)
                    {
                        throw new DisconnectedException("The remote server has closed the upgrade connection");
                    }
                    readBytes = _socketStream.Read(getResponse, readBytes, maxReadSize-readBytes);
                    totalBytes += readBytes;
                }
                if (getResponse[9] != (byte)'1' || getResponse[10] != (byte)'0' || getResponse[11] != (byte)'1')
                {
                    throw new ConnectionRefusedException($"Failed to upgrade connection, failed response code {Encoding.UTF8.GetString(getResponse)}");
                }
                // Remove any remaining response bytes
                try
                {
                    while (readBytes > 0)
                    {
                        readBytes = _socketStream.Read(getResponse, 0, maxReadSize);
                    }
                }
                catch (IOException ex)
                {
                    SocketException sockEx = ex.InnerException as SocketException;
                    if (sockEx != null)
                    {
                        switch (sockEx.SocketErrorCode)
                        {
                            case SocketError.TimedOut:
                                break;
                            case SocketError.Interrupted:
                                // Send unexpected exception to exception listener
                                broadcastException(ex);
                                break;
                            case SocketError.IOPending:
                                handleCloseEvent(_connectionVersion,
                                    "Unexpected async I/O socket error during http upgrade negotiation", ex);
                                return;
                            default:
                                handleCloseEvent(_connectionVersion, "Socket failure during http upgrade negotiation", ex);
                                return;
                        }
                    }
                    else
                    {
                        handleCloseEvent(_connectionVersion, "IO failure during http upgrade negotiation", ex);
                        return;
                    }
                }
                catch (Exception ex)
                {
                    handleCloseEvent(_connectionVersion, "Failure during http upgrade negotiation", ex);
                    return;
                }
            }
        }
        private void applySocketOptionBool(SocketOptionLevel level, SocketOptionName name, string v)
        {
            try
            {
                switch (v)
                {
                    case "true":
                        _socket.SetSocketOption(level, name, true);
                        return;

                    case "false":
                        _socket.SetSocketOption(level, name, false);
                        return;
                }
            }
            catch (Exception e)
            {
                throw new InvalidURIException("Error setting socket properties.", e);
            }
            throw new InvalidURIException("Invalid socket option value `" + v + "'");
        }
        private void applySocketOptionInt(SocketOptionLevel level, SocketOptionName name, string v)
        {
            try
            {
                int val = Convert.ToInt32(v);
                _socket.SetSocketOption(level, name, val);
            }
            catch (Exception e)
            {
                throw new InvalidURIException("Invalid integer value in URI.", e);
            }
        }

        private void applySocketProperties(IEnumerable<KeyValuePair<string, string>> properties_)
        {
            _socket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.KeepAlive, true);

            if (properties_ == null) return;
            foreach (var pair in properties_)
            {
                applySocketProperty(pair.Key, pair.Value);
            }
        }

        /// <summary>
        /// Applies a socket property to self. Override this to add custom properties to your
        /// transport or intercept the stock ones.
        /// </summary>
        /// <param name="propertyName">The property name from the URI.</param>
        /// <param name="propertyValue">The property value.</param>
        protected virtual void applySocketProperty(String propertyName, String propertyValue)
        {
            switch (propertyName)
            {
                case "bind":
                    int portStart = propertyValue.IndexOf(':');
                    if (portStart == -1)
                    {
                        _socket.Bind(new IPEndPoint(IPAddress.Parse(propertyValue), 0));
                    }
                    else
                    {
                        string addr = propertyValue.Substring(0, portStart);
                        string portString = propertyValue.Substring(++portStart);
                        int port = Convert.ToInt32(portString);
                        _socket.Bind(new IPEndPoint(IPAddress.Parse(addr), port));
                    }
                    break;
                case "tcp_keepalive":
                    applySocketOptionBool(SocketOptionLevel.Socket, SocketOptionName.KeepAlive, propertyValue);
                    break;
                case "tcp_sndbuf":
                    applySocketOptionInt(SocketOptionLevel.Socket, SocketOptionName.SendBuffer, propertyValue);
                    break;
                case "tcp_rcvbuf":
                    applySocketOptionInt(SocketOptionLevel.Socket, SocketOptionName.ReceiveBuffer, propertyValue);
                    break;
                case "tcp_linger":
                    applySocketOptionInt(SocketOptionLevel.Socket, SocketOptionName.Linger, propertyValue);
                    break;
                case "tcp_nodelay":
                    applySocketOptionBool(SocketOptionLevel.Tcp, SocketOptionName.NoDelay, propertyValue);
                    break;
                case "tcp_connecttimeout":
                    try
                    {
                        _connectTimeout = Convert.ToInt32(propertyValue);
                    }
                    catch (Exception ex)
                    {
                        throw new InvalidURIException("Invalid value for tcp_connecttimeout.", ex);
                    }
                    break;
                case "tcp_idlereadtimeout":
                    try
                    {
                        int idleReadTimeout = Convert.ToInt32(propertyValue);
                        if (idleReadTimeout <= 0 || idleReadTimeout > 1000) {
                            throw new InvalidURIException("Invalid value for "
                                + "tcp_idlereadtimeout. Must be between 1 and 1000"
                                + " ms inclusive: " + propertyValue);
                        }
                        _idleReadTimeout = idleReadTimeout;
                    }
                    catch (Exception ex)
                    {
                        throw new InvalidURIException("Invalid value for tcp_connecttimeout.", ex);
                    }
                    break;
                case "http_preflight":
                    _httpPreflight = propertyValue.Equals("true");
                    break; // handled at handshake
                case "compression":
                    break; // handled at stream creation
                case "pretty":
                    break; // handled by the Client
                case "ip_protocol_prefer":
                    break; // handled by the connect method
                case "sni":
                    break; // handled in TCPSTransportImpl
                default:
                    throw new InvalidURIException("Unrecognized URI Parameter `" +
                        propertyName + "'");
            }
        }

        private void _disconnect()
        {
            try
            {
                if (_addr != null)
                {
                    if(_readerThread != null) _readerThread.stopThread();
                    if(_socket != null)
                    {
                      _socket.Shutdown(SocketShutdown.Send);
                      _socket.Close();
                    }
                    // we'd like the next blocking operation in the
                    // reader thread to throw so it can end.  Unless,
                    // of course, *we* are the reader thread.
                    if (_readerThread != null
                        && !_readerThread._thread.Equals(Thread.CurrentThread)
                        && _readerThread._thread.IsAlive)
                    {
                        _readerThread._thread.Interrupt();
                        Thread.Sleep(5);
                        while (_readerThread._thread.IsAlive)
                        {
                            _readerThread._thread.Interrupt();
                            if (!_readerThread._thread.Join(5000)
                                && _exceptionListener != null)
                            {
                                try
                                {
                                    _exceptionListener.Invoke(
                                        new TimedOutException(
                                          "Client reader thread "
                                          + $"{_readerThread._thread.Name} is "
                                          + "taking a long time to exit. "
                                          + "Current state is "
                                          + $"{_readerThread._thread.ThreadState}"));
                                }
                                catch (Exception)
                                {
                                }
                            }
                        }
                    }
                }
            }
            catch (Exception)
            {
                ;
            }
            this._resolvedAddr = null;
            this._addr = null;
        }

        /// <summary>
        /// Disconnects the TCP transport.
        /// </summary>
        public void disconnect()
        {
            lock (_lock)
            {
                _disconnecting = true;
                this._disconnect();
                _disconnecting = false;
            }
        }

        /// <summary>
        /// Sends the data contained in the provided memory stream over the socket.
        /// </summary>
        /// <param name="buf">The memory stream containing the data to send.</param>
        /// <exception cref="DisconnectedException">Thrown when the socket is closed or encounters an error while sending the message.</exception>
        public void send(MemoryStream buf)
        {
            try
            {
                byte[] bytes = buf.GetBuffer();
                if (_filter != null) _filter.outgoing(ref bytes, 4, (int)buf.Length-4);
                _socketStream.Write(bytes, 0, (int)buf.Length);
            }
            catch (ObjectDisposedException)
            {
                throw new DisconnectedException("Socket was closed while sending message.");
            }
            catch (SocketException)
            {
                throw new DisconnectedException("Socket error while sending message.");
            }
            catch (IOException ioex)
            {
                throw new DisconnectedException("IO error while sending message.", ioex);
            }
            // See .NET core issue (fixed in 3.0) https://github.com/dotnet/runtime/issues/29006
            catch (NotSupportedException)
            {
                throw new DisconnectedException("Socket was previously closed before sending message.");
            }
        }

        /// <summary>
        /// Returns the socket associated with this TCP transport.
        /// </summary>
        /// <returns>The socket associated with this TCP transport.</returns>
        public Socket socket()
        {
            try
            {
                return _socket;
            }
            catch (Exception)
            {
                return null;
            }
        }

        /// <summary>
        /// Returns the size of the write queue.
        /// </summary>
        /// <returns>The size of the write queue.</returns>
        public long writeQueueSize()
        {
            return 0;
        }

        /// <summary>
        /// Returns the size of the read queue.
        /// </summary>
        /// <returns>The size of the read queue.</returns>
        public long readQueueSize()
        {
            return 0;
        }

        /// <summary>
        /// Flushes any pending data in the TCP transport.
        /// </summary>
        /// <returns>Always returns 0.</returns>
        public long flush()
        {
            // Having the lock means we've sent all data in this
            //   TCP implementation.
            return 0;
        }

        /// <summary>
        /// Flushes any pending data in the TCP transport with a specified timeout.
        /// </summary>
        /// <param name="timeout">The timeout period.</param>
        /// <returns>Always returns 0.</returns>
        public long flush(long timeout)
        {
            // Having the lock means we've sent all data in this
            // TCP implementation.
            return 0;
        }

        /// <summary>
        /// Creates a new TCP transport based on the provided protocol and implementation.
        /// </summary>
        /// <param name="messageType">The protocol for the transport.</param>
        /// <param name="impl">The implementation of the TCP transport.</param>
        /// <returns>A new TCP transport instance.</returns>
        public virtual TCPTransport createTransport(Protocol messageType, TCPTransportImpl impl)
        {
            return TCPTransport.createTransport(messageType, impl);
        }

        /// <summary>
        /// Handles a close event, such as a disconnection or error, by invoking the appropriate disconnect handler.
        /// </summary>
        /// <param name="failedVersion_">The version of the failed connection.</param>
        /// <param name="message">The error message.</param>
        /// <param name="exception">The exception that caused the disconnection.</param>
        /// <exception cref="DisconnectedException">Thrown when a disconnect occurs.</exception>
        /// <exception cref="RetryOperationException">Thrown when a reconnect is attempted.</exception>
        public virtual void handleCloseEvent(int failedVersion_, String message, Exception exception)
        {
            broadcastException(exception);
            _onDisconnect.preInvoke(failedVersion_);
            if (_readerThread == null || !_readerThread._thread.Equals(Thread.CurrentThread))
            {
                Monitor.Enter(_lock);
            }
            else
            {
                // The reader thread can cause a deadlock if send thread first grabs
                // the lock and then reader thread blocks trying to acquire it before
                // getting interrupted by the send thread.
                try
                {
                    while (!Monitor.TryEnter(_lock,100))
                    {
                        // Check if we're disconnecting
                        if (_disconnecting)
                        {
                            throw new DisconnectedException("Disconnect in progress.");
                        }
                        if (_readerThread == null || _readerThread.stopped)
                        {
                            throw new DisconnectedException("Reconnect is in progress in send thread.");
                        }
                    }
                }
                catch (ThreadInterruptedException)
                {
                    // Check if we're disconnecting
                    if (_disconnecting)
                    {
                        throw new DisconnectedException("Disconnect in progress.");
                    }
                    throw new DisconnectedException("Reconnect already in progress in send thread.");
                }
            }
            try
            {
                if (_disconnecting)
                {
                    throw new DisconnectedException("Disconnect in progress.");
                }
                // If there's a new version of the connection available, you should use that.
                if (failedVersion_ != _connectionVersion)
                {
                    throw new RetryOperationException("A new connection is available.");
                }

                //OK, our thread is in charge of disconnecting and reconnecting. Let's do it.
                try
                {
                    _disconnect();
                }
                catch (Exception)
                {
                }


                // Create a completely new TCPTransport
                TCPTransport t = createTransport(this._messageType, this);
                try
                {
                    this._onDisconnect.invoke(t, new DisconnectedException(message, exception));
                }
                catch (Exception aex)
                {
                    throw new DisconnectedException("User's DisconnectHandler threw an exception.", aex);
                }
            }
            finally
            {
                Monitor.Exit(_lock);
            }
            // This doesn't need to be locked, because we only care if the connection version is
            // something other than what we arrived with.
            if (_connectionVersion == failedVersion_)
            {
                throw new DisconnectedException("A disconnect occurred, and no disconnect handler successfully reconnected.");
            }
            else
            {
                throw new RetryOperationException("Reconnect successful; retry the operation.");
            }
        }

        // Handles an external timeout event by invoking the appropriate action or idle action.
        internal void handleExternalTimeoutEvent()
        {
            if (_readerThread.handleTimeoutEvent())
            {
                return;
            }
            if (_idleAction != null) {
                _idleAction.Invoke();
            }
        }

        /// <summary>
        /// Initializes the transport configuration from the provided client.
        /// </summary>
        /// <param name="client">The client from which to initialize the transport configuration.</param>
        public void initFromClient(Client client)
        {
            _client = client;
            setMessageHandler(client.getTransportMessageHandler());
            setDisconnectHandler(client.getTransportDisconnectHandler());
            IEnumerable<string> preflightHeaders = client.getHttpPreflightHeaders();
            if (preflightHeaders != null)
            {
                _httpPreflightHeaders = new List<string>(client.getHttpPreflightHeaders());
            }
            else
            {
                _httpPreflightHeaders = null;
            }
            setThreadCreatedHandler(client.getThreadCreatedHandler());
            Client.ExceptionListener listener = client.getExceptionListener();
            if (listener != null)
            {
                setExceptionListener(new TCPTransportImpl.ExceptionListener(listener));
            }
            setTransportFilter(client.getTransportFilter());
        }

        // Thread responsible for reading data from the TCP transport.
        internal class TCPReaderThread
        {
            internal TCPTransportImpl transport = null;
            internal Protocol messageType = null;
            internal volatile bool stopped = false;
            internal Thread _thread;
            internal int _currentVersion;
            internal DateTime lastRead = DateTime.Now;
            internal Barrier barrier = new Barrier(2);
            internal bool msgReceived = false;
            internal MessageHandler onMessage;

            class MsgReceivedHandler : MessageHandler
            {
                TCPReaderThread _thread;
                public MsgReceivedHandler(TCPReaderThread thread)
                {
                    _thread = thread;
                }

                public void invoke(Message msg)
                {
                    _thread.msgReceived = true;
                    _thread.transport._onMessage.invoke(msg);
                }
            }

            internal TCPReaderThread(TCPTransportImpl transport, Protocol messageType)
            {
                this.transport = transport;
                this.messageType = messageType;
                this.stopped = false;
                _thread = new Thread(_ => run());
                _thread.IsBackground = true;
                _thread.Start();
                // We use the barrier to allow the callback to set the thread name
                this.barrier.SignalAndWait();
                if (_thread.Name == null)
                {
                    _thread.Name = "AMPS C# Client Background Reader Thread";
                }
                onMessage = new MsgReceivedHandler(this);
            }

            /// <summary>
            /// Stops the thread.
            /// </summary>
            public void stopThread()
            {
                this.stopped = true;
            }

            /// <summary>
            /// Handles a close event by invoking the appropriate close handler.
            /// </summary>
            /// <param name="e">The exception associated with the close event.</param>
            void handleCloseEvent(Exception e)
            {
                if(!stopped)
                {
                    try
                    {
                        transport.handleCloseEvent(_currentVersion, e.Message, e);
                    }
                    catch (Exception ex)
                    {
                        transport.broadcastException(ex);
                    }
                }
            }

            // Handles a timeout event by checking if there has been no activity on the socket.
            // returns: True if a timeout event occurred and was handled; otherwise, false.
            internal bool handleTimeoutEvent()
            {
                if (!stopped && transport._readTimeout > 0
                    && (DateTime.Now - lastRead).TotalMilliseconds
                        > transport._readTimeout)
                {
                    try
                    {
                        const string message = "No activity on socket or socket closed.";
                        transport.handleCloseEvent(_currentVersion, message, new DisconnectedException(message));
                    }
                    catch(Exception ex)
                    {
                        transport.broadcastException(ex);
                    }
                    return true;
                }
                else
                {
                    return false;
                }
            }

            // Handles the end of the stream event by invoking the appropriate close handler.
            void handleEndOfStream()
            {
                if (!stopped)
                {
                    try
                    {
                        const string message = "No activity on socket or socket closed.";
                        transport.handleCloseEvent(_currentVersion, message, new DisconnectedException(message));
                    }
                    catch (Exception ex)
                    {
                        transport.broadcastException(ex);
                    }
                }
            }
            void handleStreamCorruption()
            {
                if (!stopped)
                {
                    try
                    {
                        const string message = "Corruption of stream detected.";
                        transport.handleCloseEvent(_currentVersion, message, new DisconnectedException(message));
                    }
                    catch (Exception ex)
                    {
                        transport.broadcastException(ex);
                    }
                }
            }

            /// <summary>
            /// Runs the TCP reader thread.
            /// </summary>
            public void run()
            {
                if (transport._threadCreatedHandler != null)
                {
                    try {
                        transport._threadCreatedHandler.invoke();
                    }
                    catch (Exception e)
                    {
                        transport.broadcastException(e);
                    }
                }
                // We use the barrier to allow the callback to set the thread name
                this.barrier.SignalAndWait();
                byte[] buf = new byte[16 * 1024];
                SimpleReadBuffer readBuffer = new SimpleReadBuffer();
                ProtocolParser messageStream = this.messageType.getMessageStream();
                int readPoint = 0, parsePoint = 0;
                lastRead = DateTime.Now;
                while (true)
                {
                    if (this.stopped)
                        return;
                    _currentVersion = this.transport._connectionVersion;
                    int bytesRead = 0;
                    try
                    {
                        try {
                            // Only call these after the first message (logon ack)
                            // has been received for this thread.
                            if (msgReceived) {
                                // Send a heartbeat message, if its time has arrived.
                                transport._client.checkAndSendHeartbeat(false);
                                // Do we need to do any idle-time processing (flushing acks)
                                if (transport._idleAction != null) transport._idleAction.Invoke();
                            }
                        }
                        catch (Exception e)
                        {
                            transport.broadcastException(e);
                        }
                        bytesRead = transport._socketStream.
                            Read(buf, readPoint, buf.Length - readPoint);
                        if (bytesRead == 0)
                        {
                            handleEndOfStream();
                            return;
                        }
                        else
                        {
                            lastRead = DateTime.Now;
                        }
                    }
                    catch (IOException ex)
                    {
                        SocketException sockEx = ex.InnerException as SocketException;
                        if (sockEx != null)
                        {
                            switch (sockEx.SocketErrorCode)
                            {
                                case SocketError.TimedOut:
                                    if(handleTimeoutEvent())
                                    {
                                        return;
                                    }
                                    break;
                                case SocketError.Interrupted:
                                    // Send unexpected exception to exception listener
                                    transport.broadcastException(ex);
                                    if(handleTimeoutEvent())
                                    {
                                        return;
                                    }
                                    break;
                                case SocketError.IOPending:
                                    Exception knownIssue = new AMPSException(
                                        "A known issue has been observed in .Net Framework 4.x when "
                                        + "performing high-volume reading and writing to a socket's network "
                                        + "stream while a socket ReceiveTimeout is set. Rarely a SocketException "
                                        + "is thrown with SocketError.IOPending (Overlapped I/O operation is in "
                                        + "progress). Windows is throwing this error for blocking Read()/Write() "
                                        + "calls, even though this error should only be thrown for async I/O "
                                        + "calls. You can work-around this issue by using a separate AMPS "
                                        + "client instance for publishing. The AMPS Client's only option when "
                                        + "catching this error is to disconnect since this error is known to "
                                        + "corrupt further reads from the stream.", ex);
                                    handleCloseEvent(knownIssue);
                                    return;
                                default:
                                    handleCloseEvent(ex);
                                    return;
                            }
                        }
                        else
                        {
                            handleCloseEvent(ex);
                            return;
                        }
                    }
                    catch (Exception ex)
                    {
                        handleCloseEvent(ex);
                        return;
                    }

                    readPoint += bytesRead;
                    while(readPoint >= parsePoint+4)
                    {
                        int availableBytes = readPoint - parsePoint;
                        int messageSize = Utilities.readInt32(buf, parsePoint);
                        int sizeWithFrame = 4 + messageSize;
                        int messageStartPoint = parsePoint + 4;
                        if (sizeWithFrame <= availableBytes)
                        {
                            try
                            {
                                if (transport._filter != null) transport._filter.incoming(ref buf, messageStartPoint, messageSize);
                                readBuffer.Reset(buf, messageStartPoint, messageSize);
                                messageStream.process(readBuffer, messageSize, this.onMessage);
                            }
                            catch (Exception e)
                            {
                                transport.broadcastException(e);
                            }
                            parsePoint += sizeWithFrame;
                        }
                        else
                        {
                            // enough space remaining to hold the whole message?
                            if (sizeWithFrame > (buf.Length - parsePoint))
                            {
                                // do we need to resize the buffer or just move the data
                                if (sizeWithFrame > buf.Length)
                                {
                                    // Full resize needed
                                    if (buf.Length > (Int32.MaxValue / 2))
                                    {
                                        handleStreamCorruption();
                                        return;
                                    }
                                    byte[] newBuffer = new byte[2 * buf.Length];
                                    Array.Copy(buf, parsePoint, newBuffer, 0, readPoint - parsePoint);
                                    buf = newBuffer;
                                }
                                else
                                {
                                    // just move the data down to the beginning of the buffer
                                    // MSDN documents Array.Copy as safe for overlapping regions.
                                    Array.Copy(buf, parsePoint, buf, 0, readPoint - parsePoint);
                                }
                                readPoint -= parsePoint;
                                parsePoint = 0;
                            }
                            // now we need to get more.
                            break;
                        }
                    }// end while(readpoint > parsePoint + 4)
                    // if the buffer contains no pending data to parse, start reading at 0
                    if (readPoint == parsePoint)
                    {
                        readPoint = 0;
                        parsePoint = 0;
                    }
                    else if (buf.Length - parsePoint < 32)
                    {
                        Array.Copy(buf, parsePoint, buf, 0, readPoint - parsePoint);
                        readPoint -= parsePoint;
                        parsePoint = 0;
                    }
                }// while(true)
            }//public void run()
        }
    }
}