////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Net;
using AMPS.Client.Exceptions;

namespace AMPS.Client
{
    /// <summary>
    /// Interface that is used to manage and configure connections to AMPS. Typically
    /// a transport is created and used internally by Client instance. There
    /// is usually no reason to make direct use of it. Its functionality is best accessed
    /// via the client instance.
    /// </summary>
    public interface Transport
    {
        /// <summary>
        /// Connect to an AMPS server.
        /// </summary>
        /// <param name="uri">Specifies the protocol and address of the server.</param>
        /// <exception cref="ConnectionRefusedException"/>Thrown when the connection is refused.</exception>
        /// <exception cref="AlreadyConnectedException"/>Thrown when already connected.</exception>
        void connect(Uri uri);

        /// <summary>
        /// Closes the connection.
        /// </summary>
        void close();

        /// <summary>
        /// Disconnects from the AMPS server.
        /// </summary>
        void disconnect();

        /// <summary>
        /// Sets the message handler for processing incoming messages.
        /// </summary>
        /// <param name="ml">The message handler.</param>
        void setMessageHandler(MessageHandler ml);

        /// <summary>
        /// Sets the disconnect handler for handling disconnection events.
        /// </summary>
        /// <param name="dh">The disconnect handler.</param>
        void setDisconnectHandler(TransportDisconnectHandler dh);

        /// <summary>
        /// Sets the exception listener for handling exceptions.
        /// </summary>
        /// <param name="exceptionListener">The exception listener.</param>
        void setExceptionListener(TCPTransportImpl.ExceptionListener exceptionListener);

        /// <summary>
        /// Sets the handler for thread creation events.
        /// </summary>
        /// <param name="handler">The thread created handler.</param>
        void setThreadCreatedHandler(ThreadCreatedHandler handler);

        /// <summary>
        /// Sends a message to the AMPS server.
        /// </summary>
        /// <param name="message">The message to send.</param>
        /// <exception cref="DisconnectedException">Thrown when disconnected.</exception>
        void send(Message message);

        /// <summary>
        /// Sets the read timeout in milliseconds.
        /// </summary>
        /// <param name="readTimeoutMillis_">The read timeout in milliseconds.</param>
        void setReadTimeout(int readTimeoutMillis_);

        /// <summary>
        /// Sends a message to the AMPS server without retrying on failure.
        /// </summary>
        /// <param name="message">The message to send.</param>
        void sendWithoutRetry(Message message);

        /// <summary>
        /// Allocates a new message.
        /// </summary>
        /// <returns>The allocated message.</returns>
        Message allocateMessage();

        /// <summary>
        /// Returns the size of the write queue.
        /// </summary>
        /// <returns>The size of the write queue.</returns>
        /// <exception cref="DisconnectedException">Thrown when disconnected.</exception>
        long writeQueueSize();

        /// <summary>
        /// Returns the size of the read queue.
        /// </summary>
        /// <returns>The size of the read queue.</returns>
        /// <exception cref="DisconnectedException">Thrown when disconnected.</exception>
        long readQueueSize();

        /// <summary>
        /// Flushes the transport.
        /// </summary>
        /// <returns>The number of bytes flushed.</returns>
        /// <exception cref="DisconnectedException">Thrown when disconnected.</exception>
        long flush();

        /// <summary>
        /// Flushes the transport with a specified timeout.
        /// </summary>
        /// <param name="timeout">The timeout value in milliseconds.</param>
        /// <returns>The number of bytes flushed.</returns>
        /// <exception cref="DisconnectedException">Thrown when disconnected.</exception>
        long flush(long timeout);

        /// <summary>
        /// Handles the close event with the specified parameters.
        /// </summary>
        /// <param name="failedVersion">The failed version.</param>
        /// <param name="message">The message.</param>
        /// <param name="exception">The exception.</param>
        void handleCloseEvent(int failedVersion, String message, Exception exception);

        /// <summary>
        /// Returns the version of the transport.
        /// </summary>
        /// <returns>The version of the transport.</returns>
        int getVersion();

        /// <summary>
        /// Sets a TransportFilter that filters raw bytes before send and after receive.
        /// </summary>
        /// <param name="filter_">A TransportFilter such as AMPS.Client.TransportTraceFilter.</param>
        void setTransportFilter(TransportFilter filter_);

        /// <summary>
        /// Sets a {@link TransportFilterModifiable} that filters and possibly modifies
        /// raw bytes before send and after receive.
        /// </summary>
        /// <param name="filter_">A TransportFilterModifiable.</param>
        void setTransportFilter(TransportFilterModifiable filter_);

        /// <summary>
        /// Gets the action to be performed during idle periods.
        /// </summary>
        /// <returns>The action to be performed during idle periods.</returns>
        Action getIdleAction();

        /// Sets a {@link AMPSRunnable} that is invoked during idle-time processing for this connection.
        /// </summary>
        /// <param name="runnable">A {@link AMPSRunnable} instance whose run() method is invoked.</param>
        void setIdleAction(Action runnable);

        /// <summary>
        /// Called by Client before calling connect to give Transport access to
        /// any information in the Client needed by the Transport, such as
        /// message handler, exception listener, etc.
        /// </summary>
        /// <param name="client">A Client to initialize from.</param>
        void initFromClient(Client client);
    }

}
