////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Reflection;
using System.Collections.Generic;

using AMPS.Client.Exceptions;
namespace AMPS.Client
{

/// <summary>
/// Factory class that allows you to register and unregister handler classes
/// for various connection URI prefixes. Handler class is expected to create
/// a transport when asked. Built-in transports such as "tcp" cannot be
/// unregistered.
/// </summary>
public class TransportFactory {


	private struct TransportRegistryEntry
	{
	  public string _name;
	  public Assembly _assembly;
	}
	private static Dictionary<string, TransportRegistryEntry> registry= new Dictionary<string, TransportRegistryEntry>();

   /// <summary>
   /// Registers a new Transport with the specified name, class name,
   /// and assembly.
   /// </summary>
   /// <param name="name">The URI prefix to register for this transport.</param>
   /// <param name="className">The name of the Java class to instantiate.</param>
   /// <param name="assembly">The assembly containing the transport type (defaults
   /// to the executing assembly).</param>
   /// <exception cref="TransportTypeException">Thrown when this transport type
   /// is already registered.</exception>
   public static void register(
         string name,
         string className,
         Assembly assembly=null)

   {
      if (assembly == null)
      {
	        assembly = Assembly.GetExecutingAssembly();
      }
      if(!registry.ContainsKey(name)) {
         TransportRegistryEntry e;
         e._name = className;
         e._assembly = assembly;
         registry[name] = e;
      } else
			throw new TransportTypeException("'" + name + "' is already registered");
	}

    /// <summary>
    /// Unregisters a Transport from the AMPS client.
    /// <b>Note</b>: built-in transport types (e.g. 'tcp') cannot be unregistered.
    /// </summary>
    /// <param name="name">The URI prefix to unregister.</param>
    /// <exception cref="TransportTypeException">Thrown when this transport type has
    /// not been registered.</exception>
	public static void unregister(string name)
	{
		if (registry.ContainsKey(name)) {
			registry.Remove(name);
		} else {
			throw new TransportTypeException("The TransportType '"  + name + "' is not a known TransportType");
		}
	}

    /// <summary>
    /// Creates a {@link Transport} using the parameters provided.
    /// </summary>
    /// <param name="name">A string which contains the name of the future Transport</param>
    /// <param name="messageType">A parameter of type Protocol which contains the classification
    /// of the message.</param>
    /// <param name="props">A parameter of type Properties which contains the information about
    /// the message to be created.</param>
    /// <returns>The newly-created Transport.</returns>
    /// <exception cref="TransportTypeException">Thrown when the requested transport type cannot
    /// be constructed.</exception>
	public static Transport createTransport(
			string name,
			Protocol messageType,
			Properties props)
	{
        // Whoops, we don't have a Transport by this name in the registry,
        // let's guess at the correct type by the name.
        string tName = "AMPS.Client." + name.ToUpperInvariant() + "Transport";
        Assembly tAssembly = Assembly.GetExecutingAssembly();
		if (TransportFactory.registry.ContainsKey(name)) {
			tName = TransportFactory.registry[name]._name;
			tAssembly = TransportFactory.registry[name]._assembly;
		}
        try {
           return (Transport)tAssembly.GetType(tName) // get the transport name
                .GetConstructor(new Type[] { typeof(Protocol), typeof(Properties) }) // find the right ctor
                .Invoke(new object[] { messageType, props }); // invoke with the right args
        } catch(Exception) {
           throw new TransportTypeException("The TransportType '" + name + "' could not be found.");
        }
	}

}
}
