////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

namespace AMPS.Client
{
    public class Utilities
    {
        /// <summary>
        /// Reads a 32-bit signed integer from the specified byte array starting at the specified byte offset.
        /// </summary>
        /// <param name="buf">The byte array to read from.</param>
        /// <param name="ofs">The byte offset at which to begin reading.</param>
        /// <returns>The 32-bit signed integer read from the byte array.</returns>
        public static int readInt32(byte[] buf, long ofs)
        {
            int x = buf[ofs] << 24;
            x |= buf[ofs + 1] << 16;
            x |= buf[ofs + 2] << 8;
            x |= buf[ofs + 3];
            return x;
        }

        /// <summary>
        /// Writes a 32-bit signed integer to the specified byte array starting at the specified
        /// byte offset.
        /// </summary>
        /// <param name="val">The 32-bit signed integer to write.</param>
        /// <param name="buf">The byte array to write to.</param>
        /// <param name="ofs">The byte offset at which to begin writing.</param>
        public static void writeInt32(int val, byte[] buf, long ofs)
        {
            buf[ofs] = (byte)((val & 0xff000000) >> 24);
            buf[ofs+1] = (byte)((val & 0x00ff0000) >> 16);
            buf[ofs+2] = (byte)((val & 0x0000ff00) >> 8);
            buf[ofs+3] = (byte)((val & 0x000000ff));
        }

        /// <summary>
        /// Reads a 64-bit signed integer from the specified byte array starting at the specified
        /// byte offset.
        /// </summary>
        /// <param name="buf">The byte array to read from.</param>
        /// <param name="ofs">The byte offset at which to begin reading.</param>
        /// <returns>The 64-bit signed integer read from the byte array.</returns>
        public static long readLong64(byte[] buf, long ofs)
        {
            long x = (long)buf[ofs] << 56;
            x |= (long)buf[ofs + 1] << 48;
            x |= (long)buf[ofs + 2] << 40;
            x |= (long)buf[ofs + 3] << 32;
            x |= (long)buf[ofs + 4] << 24;
            x |= (long)buf[ofs + 5] << 16;
            x |= (long)buf[ofs + 6] << 8;
            x |= buf[ofs + 7];
            return x;
        }

        /// <summary>
        /// Writes a 64-bit signed integer to the specified byte array starting at the specified
        /// byte offset.
        /// </summary>
        /// <param name="val">The 64-bit signed integer to write.</param>
        /// <param name="buf">The byte array to write to.</param>
        /// <param name="ofs">The byte offset at which to begin writing.</param>
        public static void writeLong64(long val, byte[] buf, long ofs)
        {
            buf[ofs]     = (byte)(((ulong)val & 0xff00000000000000) >> 56);
            buf[ofs + 1] = (byte)((val & 0x00ff000000000000) >> 48);
            buf[ofs + 2] = (byte)((val & 0x0000ff0000000000) >> 40);
            buf[ofs + 3] = (byte)((val & 0x000000ff00000000) >> 32);
            buf[ofs + 4] = (byte)((val & 0x00000000ff000000) >> 24);
            buf[ofs + 5] = (byte)((val & 0x0000000000ff0000) >> 16);
            buf[ofs + 6] = (byte)((val & 0x000000000000ff00) >> 8);
            buf[ofs + 7] = (byte)((val & 0x00000000000000ff));
        }

        /// <summary>
        /// Represents a CRC32 calculator.
        /// </summary>
        public interface CRC32
        {
            /// <summary>
            /// Initializes the CRC32 calculator.
            /// </summary>
            /// <returns>The initial value of the CRC32 calculator.</returns>
            long init();

            /// <summary>
            /// Gets the CRC32 value.
            /// </summary>
            /// <param name="result_">The current value of the CRC32 calculator.</param>
            /// <returns>The CRC32 value.</returns>
            long getValue(long result_);

            /// <summary>
            /// Updates the CRC32 calculator with the specified byte array.
            /// </summary>
            /// <param name="result_">The current value of the CRC32 calculator.</param>
            /// <param name="buf">The byte array to update the CRC32 calculator with.</param>
            /// <param name="offset">The zero-based byte offset in <paramref name="buf"/> at which
            /// to begin reading bytes to use in the CRC32 calculation.</param>
            /// <param name="len">The number of bytes to read from <paramref name="buf"/> for the
            /// CRC32 calculation.</param>
            /// <returns>The updated value of the CRC32 calculator.</returns>
            long update(long result_, byte[] buf, int offset, int len);
        }

        /// <summary>
        /// Represents a default implementation of the CRC32 calculator.
        /// </summary>
        public class DefaultCRC32 : CRC32
        {
            // Initializes a new instance of the DefaultCRC32 class.
            private uint[] _table = new uint[256];

            /// <summary>
            /// Initializes a new instance of the DefaultCRC32 class.
            /// </summary>
            public DefaultCRC32()
            {
                // Populate the table
                for (uint i = 0; i < 256; i++)
                {
                    uint c = i;
                    for (int j = 0; j < 8; j++)
                    {
                        c = (c & 0x01) != 0 ? (0xEDB88320 ^ (c >> 1)) : (c >> 1);
                    }
                    _table[i] = c;
                }
            }

            long CRC32.init()
            {
                return 0xFFFFFFFF;
            }
            long CRC32.getValue(long result_)
            {
                return result_ ^ 0xFFFFFFFF;
            }


            long CRC32.update(long result_, byte[] buf, int offset, int len)
            {
                if (buf.Length < offset + len)
                {
                    throw new System.IndexOutOfRangeException("Trying to calculate CRC on bytes beyond the length of the buffer");
                }
                unsafe
                {
                    fixed (byte* b = buf)
                    {
                        for (uint i = 0; i < len; i++)
                        {
                            result_ = _table[(result_ ^ b[i+offset]) & 0xFF] ^ (result_ >> 8);
                        }
                    }
                }
                return result_;
            }

        }
    }
}
