////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2012-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using AMPS.Client.Exceptions;

namespace AMPS.Client
{
    /// <summary>
    /// An IComparable representation of the server's version.
    /// </summary>
    public class VersionInfo : IComparable<VersionInfo>, IEquatable<VersionInfo>
    {
        /// <summary>
        /// Gets or sets the version string.
        /// </summary>
        public string VersionString {get; set;}
        private ulong _versionNum;
        private static ulong DEFAULT_VERSION_NUM = 9900990009900099UL;
        private static int MAXVALUES = 4;
        private static int[] MAXDIGITS = new int[] { 4, 4, 5, 5 };

        /// <summary>
        /// Initializes a new instance of the VersionInfo class with the
        /// default version string.
        /// </summary>
        public VersionInfo()
        {
            VersionString = "default";
            _versionNum = 0;
        }

        /// <summary>
        /// Initializes a new instance of the VersionInfo class with the
        /// specified version string.
        /// </summary>
        /// <param name="version">The version string.</param>
        public VersionInfo(string version)
        {
            if (!string.IsNullOrEmpty(version))
            {
                VersionString = version;
            }
            else
            {
                VersionString = "default";
            }
            _versionNum = 0;
        }

        /// <summary>
        /// Gets the old-style version as an integer.
        /// </summary>
        /// <returns>The old-style version as an integer.</returns>
        public int getOldStyleVersion()
        {
            try
            {
                return Client.getVersionAsInt(this.VersionString);
            }
            catch (CommandException)
            {
                return Client.DEFAULT_SERVER_VERSION;
            }
        }

        /// <summary>
        /// Gets the version number.
        /// </summary>
        /// <returns>The version number.</returns>
        public ulong getVersion()
        {
            if (_versionNum == 0)
            {
                parseVersion();
            }
            return _versionNum;
        }

        private void parseVersion()
        {
            // If it's not numeric, return default
            if (!Char.IsDigit(VersionString, 0))
            {
                _versionNum = DEFAULT_VERSION_NUM;
                return;
            }
            int digits = 0;
            int values = 0;
            ulong current = 0;
            char c;
            for (int i=0; i<VersionString.Length; ++i)
            {
                c = VersionString[i];
                if (Char.IsDigit(c))
                {
                    // If we exceed allowed digits, return default
                    if (++digits > MAXDIGITS[values])
                    {
                        _versionNum = DEFAULT_VERSION_NUM;
                        return;
                    }
                    // Just x10 and add next digit
                    current *= (ulong)10;
                    current += (ulong)(c - '0');
                }
                else
                {
                    if (c == '.')
                    {
                        // Value complete, add current to _versionNum
                        _versionNum += current;
                        // That's all the values we collect, we're done
                        if (++values >= MAXVALUES)
                            return;
                        // More to collect, make room for next value
                        for (int j=0; j < MAXDIGITS[values]; ++j)
                        {
                            _versionNum *= (ulong)10;
                        }
                        digits = 0;
                        current = 0;
                    }
                    else // Got odd text, return default
                    {
                        _versionNum = DEFAULT_VERSION_NUM;
                        return;
                    }
                }
            }
            _versionNum += current;
            while (++values < MAXVALUES)
            {
                for (int i=0; i < MAXDIGITS[values]; ++i)
                {
                    _versionNum *= (ulong)10;
                }
            }
            return;
        }

        /// <summary>
        /// Compares the current instance with another VersionInfo object and
        /// returns an integer that indicates whether the current instance precedes,
        /// follows, or occurs in the same position in the sort order as the other
        /// VersionInfo.
        /// </summary>
        /// <param name="info">A VersionInfo object to compare with this instance.</param>
        /// <returns>A value that indicates the relative order of the objects being compared.
        /// The return value has the following meanings: Less than zero: This instance precedes
        /// <paramref name="info"/> in the sort order. Zero: This instance occurs in the same
        /// position in the sort order as <paramref name="info"/>. Greater than zero: This instance
        /// follows <paramref name="info"/> in the sort order.</returns>
        public int CompareTo(VersionInfo info)
        {
            if ((Object)info == null)
                return 1;
            return getVersion().CompareTo(info.getVersion());
        }

        /// <summary>
        /// Determines whether the specified VersionInfo object is equal to the current VersionInfo object.
        /// </summary>
        /// <param name="info">The VersionInfo object to compare with the current VersionInfo object.</param>
        /// <returns>True if the specified VersionInfo is equal to the current VersionInfo; otherwise,
        /// false.</returns>
        public bool Equals(VersionInfo info)
        {
            if ((Object)info == null)
                return false;
            return getVersion().Equals(info.getVersion());
        }

        /// <summary>
        /// Determines whether the specified object is equal to the current object.
        /// </summary>
        /// <param name="obj">The object to compare with the current object.</param>
        /// <returns>True if the specified object is equal to the current object; otherwise,
        /// false.</returns>
        override public bool Equals(Object obj)
        {
            if (obj == null)
                return false;
            VersionInfo info = obj as VersionInfo;
            if (info == null)
                return false;
            else
                return Equals(info);
        }

        /// <summary>
        /// Serves as the default hash function.
        /// </summary>
        /// <returns>A hash code for the current object.</returns>
        override public int GetHashCode()
        {
            return VersionString.GetHashCode();
        }

        /// <summary>
        /// Determines whether two specified VersionInfo objects have the same value.
        /// </summary>
        /// <param name="left">The first VersionInfo to compare.</param>
        /// <param name="right">The second VersionInfo to compare.</param>
        /// <returns>True if the value of <paramref name="left"/> is the same as the
        /// value of <paramref name="right"/>; otherwise, false.</returns>
        public static bool operator > (VersionInfo left, VersionInfo right)
        {
            return left.CompareTo(right) == 1;
        }

        /// <summary>
        /// Determines whether two specified VersionInfo objects have the same value.
        /// </summary>
        /// <param name="left">The first VersionInfo to compare.</param>
        /// <param name="right">The second VersionInfo to compare.</param>
        /// <returns>True if the value of <paramref name="left"/> is the same as the
        /// value of <paramref name="right"/>; otherwise, false.</returns>
        public static bool operator >= (VersionInfo left, VersionInfo right)
        {
            return left.CompareTo(right) >= 0;
        }

        /// <summary>
        /// Determines whether two specified VersionInfo objects have the same value
        /// </summary>
        /// <param name="left">The first VersionInfo to compare.</param>
        /// <param name="right">The second VersionInfo to compare.</param>
        /// <returns>true if the value of <paramref name="left"/> is the same as the
        /// value of <paramref name="right"/>; otherwise, false.</returns>
        public static bool operator < (VersionInfo left, VersionInfo right)
        {
            return left.CompareTo(right) == -1;
        }

        /// <summary>
        /// Determines whether two specified VersionInfo objects have the same value.
        /// </summary>
        /// <param name="left">The first VersionInfo to compare.</param>
        /// <param name="right">The second VersionInfo to compare.</param>
        /// <returns>true if the value of <paramref name="left"/> is the same as the
        /// value of <paramref name="right"/>; otherwise, false.</returns>
        public static bool operator <= (VersionInfo left, VersionInfo right)
        {
            return left.CompareTo(right) <= 0;
        }

        /// <summary>
        /// Determines whether two specified VersionInfo objects have the same value.
        /// </summary>
        /// <param name="left">The first VersionInfo to compare.</param>
        /// <param name="right">The second VersionInfo to compare.</param>
        /// <returns>true if the value of <paramref name="left"/> is the same as the
        /// value of <paramref name="right"/>; otherwise, false.</returns>
        public static bool operator == (VersionInfo left, VersionInfo right)
        {
            return left.Equals(right);
        }

        /// <summary>
        /// Determines whether two specified VersionInfo objects have different values.
        /// </summary>
        /// <param name="left">The first VersionInfo to compare.</param>
        /// <param name="right">The second VersionInfo to compare.</param>
        /// <returns>true if the value of <paramref name="left"/> is different from the
        /// value of <paramref name="right"/>; otherwise, false.</returns>
        public static bool operator != (VersionInfo left, VersionInfo right)
        {
            return !left.Equals(right);
        }

    } // VersionInfo
} // namespace
