////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.IO;
using System.Text;
using AMPS.Client.Exceptions;

namespace AMPS.Client
{
    /// <summary>
    /// Implements the ProtocolParser interface. Parses a byte stream
    /// using XML protocol.
    /// </summary>
    public class XMLProtocolParser : ProtocolParser
    {
        private const String LATIN1 = "ISO-8859-1";
        private const byte LATIN1_ZERO = 48;
        private static byte[] HEADER = null;
        private static byte[] BODY = null;
        private static byte[] BODY_TERMINAL = null;
        private static byte[] SOW_MSGID_X = null;
        private static byte[] SOW_KEYID = null;
        private static byte[] SOW_LEN = null;
        private static byte[] SOW_TS = null;
        private static byte[] SOW_X = null;
        private static byte DBL_QUOTE = 0;
        private static byte GREATER_THAN = 0;

        static XMLProtocolParser()
        {
            HEADER = "<SOAP-ENV:Header>".getBytes(LATIN1);
            BODY = "<SOAP-ENV:Body>".getBytes(LATIN1);
            BODY_TERMINAL = "</SOAP-ENV:Body>".getBytes(LATIN1);
            SOW_MSGID_X = "</Msg>".getBytes(LATIN1);
            SOW_KEYID = " key=\"".getBytes(LATIN1);
            SOW_LEN = " len=\"".getBytes(LATIN1);
            SOW_TS = " ts=\"".getBytes(LATIN1);
            SOW_X = " x=\"".getBytes(LATIN1);
            DBL_QUOTE = "\"".getBytes(LATIN1)[0];
            GREATER_THAN = ">".getBytes(LATIN1)[0];
        }

        private XMLMessage message = null;
        private SimpleReadBuffer buffer = null;
        private int remainingBytes = 0;
        private int bufferEnd = 0;
        private enum StreamState
        {
            start, in_sow, end
        }
        private StreamState state;

        /// <summary>
        /// Initializes a new instance of the XMLProtocolParser class with the
        /// specified XML protocol message type.
        /// </summary>
        /// <param name="messageType">The XML protocol message type.</param>
        public XMLProtocolParser(XMLProtocol messageType)
        {
            this.message = (XMLMessage)messageType.allocateMessage();
        }

        /// <summary>
        /// Processes the incoming data.
        /// </summary>
        /// <param name="buffer">The read buffer containing the data.</param>
        /// <param name="remainingBytes">The number of remaining bytes to process.</param>
        /// <param name="listener">The message handler to invoke when messages are parsed.</param>
        public void process(
                SimpleReadBuffer buffer,
                int remainingBytes,
                MessageHandler listener)
        {
            this.buffer = buffer;
            this.remainingBytes = remainingBytes;
            this.state = StreamState.start;

            message.reset();
            message.setBuffer(buffer.Buffer);
            bufferEnd = buffer.Position + buffer.Length;
            //this.dumpBuffer("buffer = ");

            while (read(message))
            {
                listener.invoke(message);
            }
        }

        private enum HeaderField
        {
            AckTyp, BkMrk, BtchSz, ClntName, Cmd, CmdId,
            Expn, Fltr, GrpSqNum, Hrtbt, Matches,
            MsgLn, Options, QId, Reason, RecordsInserted,
            RecordsUpdated, RecordsDeleted, RecordsReturned, Seq, SubIds,
            SowKey, SowKeys, Status, SubId,
            TxmTm, Tpc, TopicMatches, TopN, UsrId, Version, Pw, CrlId, LeasePeriod,
            UNKNOWN
        }

        private HeaderField extractHeaderField()
        {
            if (remainingBytes < 5 || getByte() != '<')
            {
                // Not enough room for the shortest tag '<Tpc>'
                return HeaderField.UNKNOWN;
            }

            // Find the end of the tag, so that we can optimize the field lookup
            int length = 0;
            int start = buffer.Position;
            while (remainingBytes > 0 && getByte() != '>')
            {
                ++length;
            }

            switch (peekByte(start))
            {
                case (byte)'A':
                    if (length == 6)
                    {
                        return HeaderField.AckTyp;
                    }
                    else return HeaderField.UNKNOWN;
                case (byte)'B':
                    if (length == 5) return HeaderField.BkMrk;
                    if (length != 6) return HeaderField.UNKNOWN;
                    if (peekByte(start + 4) == (byte)'S')
                    {
                        return HeaderField.BtchSz;
                    }
                    return HeaderField.UNKNOWN;
                case (byte)'C':
                    switch (length)
                    {
                        case 3:  // Cmd
                            return HeaderField.Cmd;
                        case 5:  // CmdId
                            switch(peekByte(start+1))
                            {
                                case (byte)'m':return HeaderField.CmdId;
                                case (byte)'r':return HeaderField.CrlId;
                            }
                            return HeaderField.UNKNOWN;
                        case 8:  // ClntName
                            return HeaderField.ClntName;
                        default:
                            return HeaderField.UNKNOWN;
                    }
                case (byte)'E':
                    if (length == 4)
                    {
                        advance(length);
                        return HeaderField.Expn;
                    }
                    else return HeaderField.UNKNOWN;
                case (byte)'F':
                    if (length == 4)
                    {
                        return HeaderField.Fltr;
                    }
                    else return HeaderField.UNKNOWN;
                case (byte)'G':
                    switch (length)
                    {
                        case 8:  // GrpSqNum
                            return HeaderField.GrpSqNum;
                        default:
                            return HeaderField.UNKNOWN;
                    }
                case (byte)'H':
                    if (length == 4)
                    {
                        return HeaderField.Hrtbt;
                    }
                    else return HeaderField.UNKNOWN;
                case (byte)'L':
                    if (length == 11)
                    {
                        return HeaderField.LeasePeriod;
                    }
                    else return HeaderField.UNKNOWN;
                case (byte)'M':
                    switch (length)
                    {
                        case 5:  // MsgLn
                            return HeaderField.MsgLn;
                        case 7:  // Matches
                            return HeaderField.Matches;
                        default:
                            return HeaderField.UNKNOWN;
                    }

                case (byte)'O':
                    if (length == 4)
                        return HeaderField.Options;
                    else
                        return HeaderField.UNKNOWN;
                case (byte)'P':
                    if (length == 2)
                        return HeaderField.Pw;
                    else
                        return HeaderField.UNKNOWN;

                case (byte)'Q':
                    if (length == 3)
                        return HeaderField.QId;
                    else
                        return HeaderField.UNKNOWN;
                case (byte)'R':
                    switch (length)
                    {
                        case 6:  // Reason
                            return HeaderField.Reason;
                        case 14:
                            switch (peekByte(start+7))
                            {
                                case (byte)'D':
                                    return HeaderField.RecordsDeleted;
                                case (byte)'U':
                                    return HeaderField.RecordsUpdated;
                                default:
                                    return HeaderField.UNKNOWN;
                            }
                        case 15:
                            switch (peekByte(start + 7))
                            {
                                case (byte)'I':
                                    return HeaderField.RecordsInserted;
                                case (byte)'R':
                                    return HeaderField.RecordsReturned;
                                default:
                                    return HeaderField.UNKNOWN;
                            }
                        default:
                            return HeaderField.UNKNOWN;
                    }
                case (byte)'S':
                    switch (length)
                    {
                        case 3:  // Seq
                            return HeaderField.Seq;
                        case 5:  // SubId
                            return HeaderField.SubId;
                        case 6:  // SubIds,SndOOF,SowKey,Status
                            switch (peekByte(start + 1))
                            {
                                case (byte)'u': // SubIds
                                    return HeaderField.SubIds;
                                case (byte)'o': // SowKey
                                    return HeaderField.SowKey;
                                case (byte)'t': // Status
                                    return HeaderField.Status;
                                default:
                                    return HeaderField.UNKNOWN;
                            }
                        case 7:  // SowKeys
                            return HeaderField.SowKeys;
                        default:
                            return HeaderField.UNKNOWN;
                    }
                case (byte)'T':
                    switch (length)
                    {
                        case 3:  // Tpc
                            return HeaderField.Tpc;
                        case 4:  // TopN
                            return HeaderField.TopN;
                        case 5:  // TxmTm
                            return HeaderField.TxmTm;
                        case 12:  // TopicMatches
                            return HeaderField.TopicMatches;
                        default:
                            return HeaderField.UNKNOWN;
                    }
                case (byte)'U':
                    if (length != 5) return HeaderField.UNKNOWN;
                    return HeaderField.UsrId;
                case (byte)'V':
                    if (length != 7) return HeaderField.UNKNOWN;
                    return HeaderField.Version;
                default: return HeaderField.UNKNOWN;
            }
        }

        private byte getUnescapedChar(int startPos)
        {
            switch (peekByte(startPos + 1))
            {
                case (byte)'a':
                    switch (peekByte(startPos + 2))
                    {
                        case (byte)'m':                 //amp;
                            return (byte)'&';
                        case (byte)'p':                //apos;
                            return (byte)'\'';
                    }
                    break;
                case (byte)'l':
                    return (byte)'<';
                case (byte)'g':
                    return (byte)'>';
                case (byte)'q':
                    return (byte)'\"';
                default:
                    return 0x00;
            }

            return 0x00;
        }

        private int processEscapeField()
        {
            int b = buffer.Position;
            int valueLength = 0;
            while (peekByte(buffer.Position) != '<')
            {
                if (peekByte(buffer.Position) == '&')
                {
                    this.buffer.put(b++, getUnescapedChar(buffer.Position));
                    advancePast((byte)';');
                }
                else
                    this.buffer.put(b++, getByte());

                ++valueLength;
            }
            return valueLength;
        }

        private void extractFieldValues(XMLMessage message)
        {
            HeaderField field = extractHeaderField();
            while (field != HeaderField.UNKNOWN)
            {
                // Have a header field, let's extract the value
                int valueStart = buffer.Position;
                int valueLength = 0;

                Boolean needsEscape = (field == HeaderField.Tpc
                                       || field == HeaderField.UsrId
                                       || field == HeaderField.Pw
                                       || field == HeaderField.ClntName
                                       || field == HeaderField.CrlId
                                       || field == HeaderField.SubId
                                       || field == HeaderField.SubIds);
                if(needsEscape)
                    valueLength = processEscapeField();
                else
                {
                    while (remainingBytes > 0 && getByte() != '<')
                        ++valueLength;
                }
                switch (field)
                {
                    case HeaderField.AckTyp: message._AckType.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.BtchSz: message._BatchSize.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.BkMrk: message._Bookmark.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.ClntName: message._ClientName.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.Cmd: message._Command.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.CmdId: message._CommandId.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.CrlId: message._CorrelationId.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.Expn: message._Expiration.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.Fltr: message._Filter.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.GrpSqNum: message._GroupSeqNo.set(this.buffer.Buffer, valueStart, valueLength); break;
                    //          case HeaderField.Hrtbt:           message._Heartbeat.set(this.buffer.Buffer,valueStart,valueLength); break;
                    case HeaderField.LeasePeriod: message._LeasePeriod.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.Matches: message._Matches.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.MsgLn: message._Length.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.Options: message._Options.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.Pw: message._Password.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.QId: message._QueryId.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.Reason: message._Reason.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.RecordsInserted: message._RecordsInserted.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.RecordsUpdated: message._RecordsUpdated.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.RecordsDeleted: message._RecordsDeleted.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.RecordsReturned: message._RecordsReturned.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.Seq: message._Sequence.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.SubIds: message._SubIds.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.SowKey: message._SowKey.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.SowKeys: message._SowKeys.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.Status: message._Status.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.SubId: message._SubId.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.TxmTm: message._Timestamp.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.Tpc: message._Topic.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.TopicMatches: message._TopicMatches.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.TopN: message._TopN.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.UsrId: message._UserId.set(this.buffer.Buffer, valueStart, valueLength); break;
                    case HeaderField.Version: message._Version.set(this.buffer.Buffer, valueStart, valueLength); break;
                    default: break;
                }

                // Strip to end of this tag
                while (remainingBytes > 0 && getByte() != '>')
                {
                    ;  // Just advancing past the tag
                }
                field = extractHeaderField();
            }
        }

        private bool extractSOWHeaderFields(XMLMessage message)
        {
            int position = advancePast(SOW_KEYID);
            if (position < 0)
            {
                // No SowKey
                return false;
            }

            // Read Key
            if (advancePast(DBL_QUOTE) < 0)
            {
                // broken record
                return false;
            }
            message._SowKey.set(buffer.Buffer, position, buffer.Position - position - 1);

            // We've had the length in the past, let's assume we always have it
            position = advancePast(SOW_LEN);
            if (position < 0)
            {
                // No len, but had one in the past?
                return false;
            }
            // Read Key
            if (advancePast(DBL_QUOTE) < 0)
            {
                // broken record
                return false;
            }

            message._Length.set(buffer.Buffer, position, buffer.Position - position - 1);
            position = buffer.Position;
            if(peekByte(position+1)=='t' && peekByte(position+2)=='s')
            {
                position=advancePast(SOW_TS);
                advancePast(DBL_QUOTE);
                message._Timestamp.set(buffer.Buffer,position,buffer.Position-position-1);
            }
            position=buffer.Position;
            if(peekByte(position+1)=='x' && peekByte(position+2)=='=')
            {
                position=advancePast(SOW_X);
                advancePast(DBL_QUOTE);
                message._CorrelationId.set(buffer.Buffer,position,buffer.Position-position-1);
            }

            // Advance to the start of the sow record body
            if (advancePast(GREATER_THAN) < 0)
            {
                return false;
            }
            // Body starts at buffer.position and the message length exists
            position = buffer.Position;
            int len = message.getLength();
            message.setRawBufferLength(position + len - message.getRawBufferOffset());
            message._Data.set(buffer.Buffer, position, len);
            return true;
        }

        // Returns position after sequence or -1 if not found
        private int advancePast(byte[] t)
        {
            for (int position = buffer.Position; position < buffer.Position + remainingBytes - t.Length; ++position)
            {
                bool found = true;
                for (int i = 0; i < t.Length; ++i)
                {
                    if (peekByte(position + i) != t[i])
                    {
                        found = false;
                        break;
                    }
                }
                if (found)
                {
                    position += t.Length;
                    remainingBytes -= position - buffer.Position;
                    buffer.Position = position;
                    return position;
                }
            }
            return -1;
        }
        private int advancePast(byte t)
        {
            int c = remainingBytes;
            while (c-- > 0)
            {
                if (getByte() == t)
                {
                    remainingBytes = c;
                    return buffer.Position;
                }
            }
            return -1;
        }
        private void advance(int bytes)
        {
            remainingBytes -= bytes;
            buffer.Position += bytes;
        }
        private int findBytes(byte[] t)
        {
            for (int position = buffer.Position; position < bufferEnd; ++position)
            {
                bool found = true;
                for (int i = 0; i < t.Length; ++i)
                {
                    if (peekByte(position + i) != t[i])
                    {
                        found = false;
                        break;
                    }
                }
                if (found) return position;
            }
            return -1;
        }
        private byte getByte()
        {
            --remainingBytes;
            return (byte)buffer.ReadByte();
        }

        // Peek forward by offset bytes
        private byte peekByte(int offset)
        {
            return buffer.PeekByte(offset);
        }

        private bool read(XMLMessage m)
        {
            if (remainingBytes <= 0)
            {
                // No more messages in this stream
                return false;
            }
            int position;
            if (state == StreamState.start)
            {
                // First time through, let's setup the header
                m.setRawBufferOffset(buffer.Position);
                position = advancePast(HEADER);
                if (position < 0)
                {
                    throw new StreamException("stream corruption: couldn't locate XML SOAP header.");
                }
                extractFieldValues(m);
                // The first message in a SOW message sets all parts of the header
                if (m.getCommand() == Message.Commands.SOW)
                {
                    if (advancePast(BODY) < 0)
                    {
                        throw new StreamException("stream corruption: couldn't locate XML SOAP body.");
                    }
                    if (!extractSOWHeaderFields(m))
                    {
                        throw new StreamException("stream corruption: couldn't locate message segment within XML SOAP body.");
                    }

                    state = StreamState.in_sow;
                }
                else
                {   // Not a SOW message, so it's just a standard message
                    int bodyStart;
                    if (advancePast(BODY) < 0)
                    {
                        // <SOAP-ENV:Body> not found, let's assume a naked body
                        bodyStart = buffer.Position;
                        position = bodyStart;
                    }
                    else
                    {
                        bodyStart = buffer.Position;
                        position = findBytes(BODY_TERMINAL);
                        if (position < 0)
                        {
                            throw new StreamException("stream corruption: couldn't locate XML SOAP body end.");
                        }
                    }

                    m.setRawBufferLength(position-m.getRawBufferOffset());
                    m._Data.set(this.buffer.Buffer, bodyStart, position - bodyStart);
                    state = StreamState.end;
                    advance(remainingBytes);
                }
            }
            else if (state == StreamState.in_sow)
            {
                if (!extractSOWHeaderFields(m))
                {
                    state = StreamState.end;
                    advance(remainingBytes);
                    return false;
                }
            }
            else return false;

            return true;
        }

        private void dumpBuffer(String prefix)
        {
            Console.Error.Write(prefix);
            for (int j = buffer.Position; j < buffer.Length; ++j)
            {
                try
                {
                    Console.Error.Write(Encoding.Default.GetString(buffer.Buffer, j, 1));
                }
                catch (Exception)
                {
                    Console.Error.Write("{error}");
                }

            }
            Console.Error.WriteLine();
        }

    }
}