////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

using ComponentAce.Compression.Libs.zlib;
using System;
using System.IO;
using AMPS.Client.Exceptions;
namespace AMPS.Client
{
    /// <summary>
    /// A System.IO.Stream wrapper around the ComponentAce zlib implementation
    /// which works bidirectionally and eliminates a data copy.
    /// </summary>
    public class ZlibIOStream : System.IO.Stream
    {
        public ComponentAce.Compression.Libs.zlib.ZStream _deflaterZStream = new ZStream();
        public ComponentAce.Compression.Libs.zlib.ZStream _inflaterZStream = new ZStream();
        public Stream _underlying;
        public int _deflateMarker = 0;
        public byte[] _deflateBuffer = new byte[16384];

        /// <summary>
        /// Gets a value indicating whether the stream supports reading.
        /// </summary>
        public override bool CanRead
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the stream supports seeking.
        /// </summary>
        public override bool CanSeek
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the stream supports writing.
        /// </summary>
        public override bool CanWrite
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets the length of the stream.
        /// </summary>
        public override long Length
        {
            get
            {
                throw new NotImplementedException();
            }
        }

        /// <summary>
        /// Gets or sets the position within the current stream.
        /// </summary>
        public override long Position
        {
            get
            {
                throw new NotImplementedException();
            }
            set
            {
                throw new NotImplementedException();
            }
        }

        /// <summary>
        /// Initializes a new instance of the ZlibIOStream class with the specified
        /// underlying stream and window size.
        /// </summary>
        /// <param name="underlying">The underlying stream.</param>
        /// <param name="windowSize">The window size.</param>
        public ZlibIOStream(Stream underlying, int windowSize)
        {
            _underlying = underlying;
            _deflaterZStream.deflateInit(zlibConst.Z_BEST_SPEED, windowSize);
            _deflaterZStream.next_out = _deflateBuffer;

            _inflaterZStream.inflateInit(windowSize);
            _inflaterZStream.next_in = new byte[16384];
        }

        /// <summary>
        /// Writes any data buffered in the internal buffer to the underlying device.
        /// </summary>
        public override void Flush()
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Reads a sequence of bytes from the current stream and advances the position
        /// within the stream by the number of bytes read.
        /// </summary>
        /// <param name="buffer">An array of bytes. When this method returns, the buffer
        /// contains the specified byte array with the values between offset and
        /// (offset + count - 1) replaced by the bytes read from the current source.</param>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin
        /// storing the data read from the current stream.</param>
        /// <param name="count">The maximum number of bytes to be read from the current
        /// stream.</param>
        /// <returns>The total number of bytes read into the buffer. This can be less than
        /// the number of bytes requested if that many bytes are not currently available, or
        /// zero if the end of the stream is reached before any bytes are read.</returns>
        public override int Read(byte[] buffer, int offset, int count)
        {
            int bytesRead = 0;
            if (count == 0)
            {
                return 0;
            }
            try
            {
                //Console.WriteLine("BEGIN READ {0} {1}", offset, count);
                while (bytesRead == 0)
                {
                    if (_inflaterZStream.avail_in == 0)
                    {
                        int result = _underlying.Read(_inflaterZStream.next_in, 0, _inflaterZStream.next_in.Length);
                        _inflaterZStream.next_in_index = 0;
                        if (result <= 0)
                        {
                            return result;
                        }
                        _inflaterZStream.avail_in = result;
                    }
                    int code = zlibConst.Z_OK;
                    while (_inflaterZStream.avail_in > 0 && bytesRead < count
                           && code == zlibConst.Z_OK)
                    {
                        _inflaterZStream.next_out = buffer;
                        _inflaterZStream.next_out_index = offset + bytesRead;
                        _inflaterZStream.avail_out = count - bytesRead;

                        code = _inflaterZStream.inflate(zlibConst.Z_NO_FLUSH);
                        if (code != zlibConst.Z_OK && code != zlibConst.Z_STREAM_END)
                        {
                            if (code == zlibConst.Z_DATA_ERROR)
                            {
                              throw new StreamException($"Error inflating compressed data: received result {code} while inflating: {_inflaterZStream.msg} ");
                            }
                            else
                            {
                              throw new StreamException($"Error inflating compressed data: received result {code} while inflating");
                            }

                        }
                        int inflated = count - _inflaterZStream.avail_out;
                        bytesRead += inflated;
                    }
                }
            }
            catch (Exception)
            {
                throw;
            }
            return bytesRead;
        }

        /// <summary>
        /// Sets the position within the current stream.
        /// </summary>
        /// <param name="offset">The byte offset to seek.</param>
        /// <param name="origin">The reference point from which to obtain
        /// the new position.</param>
        /// <returns>The new position within the current stream.</returns>
        public override long Seek(long offset, SeekOrigin origin)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Sets the length of the current stream.
        /// </summary>
        /// <param name="value">The desired length of the current stream
        /// in bytes.</param>
        public override void SetLength(long value)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Writes a sequence of bytes to the current stream and advances
        /// the position within the stream by the number of bytes written.
        /// </summary>
        /// <param name="buffer">The byte array containing the data to write.</param>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin
        /// copying bytes to the current stream.</param>
        /// <param name="count">The number of bytes to be written to the current stream.</param>
        public override void Write(byte[] buffer, int offset, int count)
        {
            _deflaterZStream.next_in = buffer;
            _deflaterZStream.next_in_index = offset;
            _deflaterZStream.avail_in = count;

            while (_deflaterZStream.avail_in > 0)
            {
                _deflaterZStream.next_out_index = 0;
                _deflaterZStream.avail_out = _deflateBuffer.Length;
                _deflaterZStream.deflate(zlibConst.Z_SYNC_FLUSH);
                _underlying.Write(_deflateBuffer, 0, _deflateBuffer.Length - _deflaterZStream.avail_out);
            }
        }
    }
}