﻿using System;
using System.Threading;

using AMPS.Client;
using AMPS.Client.Exceptions;
using AMPS.Client.Fields;

// AMPSSOWAndSubscribeWithOOF
//
// This sample retrieves messages from a state-of-the-world database. The program flow is simple:
//
// * Connect to AMPS
// * Logon
// * Get the state-of-the-world for the "messages-sow" topic, filtered
// messages with a message number less than 10.
// * Output all messages received to the console
//
// For demonstration purposes, this sample uses the asynchronous method
// of executing a command. In this sample, there's no advantage to using
// this method.
//
// This sample doesn't include robust error handling.
//
// This file is a part of the AMPS Evaluation Kit.
////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2013-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

namespace AMPSSOWSubscribeWithOOF
{
    class AMPSSOWSubscribeWithOOF
    {
        private static string uri_ = "tcp://127.0.0.1:9027/amps/json";

        static void Main(string[] args)
        {
            using(HAClient client = new HAClient("SOWandSubscribeConsoleSubscribeWithOOF"))
            {
                DefaultServerChooser sc = new DefaultServerChooser();
                sc.add(uri_);
                client.setServerChooser(sc);

                try {
                    client.connectAndLogon();

                    // create an object to process the messages.

                    Action<Message> mh = (message) =>
                    {
                        if (message.getCommand() == Message.Commands.OOF)
                        {
                            Console.WriteLine("Message no longer in focus because : " +
                                ReasonField.encodeReason(message.getReason()) +
                                " : " + message.getData() );
                            return;
                        }
                        if (message.Command == Message.Commands.GroupBegin)
                        {
                            Console.WriteLine("Receiving messages from the SOW.");
                        }
                        Console.WriteLine(message.Data);

                        // when the GroupEnd message arrives, the SOW
                        // query is complete.
                        if (message.Command == Message.Commands.GroupEnd)
                        {
                            Console.WriteLine("Done receiving messages from the SOW.");
                        }
                    };

                    // request messages from the messages-sow topic where
                    // the message number is less than 10. Retrieve in batches
                    // of 5 messages at a time.
                    // for sample purposes, demonstrate executeAsync
                    System.Console.WriteLine("... entering subscription ...");

                    client.executeAsync(
                        new Command(Message.Commands.SOWAndSubscribe)
                            .setTopic("oof-messages-sow")
                            .setFilter("/messageNumber % 10 = 0 AND " +
                                    "(/OptionalField IS NULL OR /OptionalField <> 'ignore_me')")
                            .setBatchSize(5)
                            .setOptions(Message.Options.OOF)
                            .setTimeout(5000),
                        new ActionMessageHandler(mh));

                    System.Console.WriteLine("Requested SOW messages.");

                    // the results of the sow query arrive asynchronously,
                    // so the sample sleeps to let the messages arrive.
                    // this program uses this construct for sample purposes.
                    // generally speaking, the program would use the results
                    // of the query as they arrive.
                    while (true)
                    {
                        Thread.Sleep(100);
                    }
                }
                catch (AMPSException e)
                {
                    System.Console.WriteLine(e.Message);
                    System.Console.WriteLine(e.StackTrace);
                }
                catch (Exception e)
                {
                    System.Console.WriteLine(e.Message);
                    System.Console.WriteLine(e.StackTrace);
                }
            }
        }
    }
}
