﻿using System;

using AMPS.Client;
using AMPS.Client.Exceptions;

// AMPSSOWUpdateForOOF
//
// This sample publishes messages to a topic in AMPS that
// maintains a state-of-the-world (SOW) database.
// The program flow is simple:
//
// * Connect to AMPS
// * Logon
// * Publish a single message that matches subscriber filter, with a
//   3 second custom expiration
// * Publish update to above message, still matching subscriber filter
// * Publish series of messages potentially matching subscriber filter
// * Publish updates to above messages, excluding them from subscriber filter
//   so subscriber receives out-of-focus (OOF) messages for previously
//   matching messages.
// * Publish another series of messages potentially matching subscriber filter
// * Delete the initially published message from the SOW using a filter
//   specifying its key
// * Wait up to 4 seconds for all published messages to be flushed to AMPS
//
// The "oof-messages-sow" topic is configured in config/sample.xml to
// maintain a SOW database, where each messageNumber is a unique
// message.
//
// This sample doesn't include robust error handling.
//
// This file is a part of the AMPS Evaluation Kit.
////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2013-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

namespace AMPSSOWUpdateForOOF
{
    class AMPSSOWUpdateForOOF
    {
        private static string uri_ = "tcp://127.0.0.1:9027/amps/json";

        static void Main(string[] args)
        {
            using(HAClient client = new HAClient("exampleSOWUpdateForOOF"))
            {
                try {
                    DefaultServerChooser sc = new DefaultServerChooser();
                    sc.add(uri_);

                    client.setServerChooser(sc);

                    client.connectAndLogon();
                    Console.WriteLine("SOWUpdateForOOF Publisher Connected");

                    // Publish a message with expiration set
                    client.publish("oof-messages-sow",
                        "{ \"messageNumber\":50000, \"message\":\"Here and then gone...\"}",
                        3);

                    // Another message with the same messageNumber (50000) is published shortly after,
                    // effectively updating the previous message.
                    client.publish("oof-messages-sow",
                        "{ \"messageNumber\":50000, " +
                            "\"message\":\"I've got a bad feeling about this...\"}");

                    // Publish two more sets of messages, the first one to match the
                    // subscriber filter, the next one to make messages no longer
                    // match the subscriber filter.
                    for (int number = 0 ; number < 10000; number += 1250)
                    {
                        client.publish("oof-messages-sow",
                            "{ \"messageNumber\":" + number + ", \"message\":\"Hello, World!\"}");
                    }

                    for (int number = 0 ; number < 10000; number += 1250)
                    {
                        client.publish("oof-messages-sow",
                            "{ \"messageNumber\":" + number + ", " +
                                "\"OptionalField\":\"ignore_me\", " +
                                    "\"message\":\"Updated, world!\"}");
                    }

                    // Publish 100 messages with unique message numbers to test out the
                    // multiples of 10 in the subscriber filter.
                    for (int number = 0; number < 100; ++number)
                    {
                        client.publish("oof-messages-sow",
                            "{ \"text\" : \"Hello, world!\", " +
                                " \"messageNumber\" : " + number + "}");
                    }

                    // Delete the first published message by filter. The subscriber should
                    // receive an OOF message for it.
                    client.sowDelete(new DefaultMessageHandler(),
                        "oof-messages-sow", "/messageNumber = 50000", 0);

                    // Wait up to 4 seconds for all messages to be published
                    client.publishFlush(4000);
                    System.Console.WriteLine("... SOWUpdateForOOF Publisher done ...");

                }
                catch (AMPSException e)
                {
                    System.Console.WriteLine(e.Message);
                    System.Console.WriteLine(e.StackTrace);
                }
            }
        }
    }
}
