////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2012-2025 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////
using System;
using System.Threading;
using System.Diagnostics;
using System.Reflection;
using System.Text;

using AMPS.Client.Exceptions;
using AMPS.Client.Fields;
using System.Collections.Generic;


namespace AMPS.Client
{
    /// <summary>
    /// The base AMPS client object used for AMPS applications that do not need a
    /// resilient connection to AMPS. Most production applications use the HAClient
    /// rather than Client. Each client object manages a single connection to AMPS.
    /// Each AMPS connection has a name, a specific transport (such as tcp), a protocol
    /// (used for framing messages to AMPS), and a message type(such as FIX or JSON).
    /// <p>
    /// The Client object creates and manages background threads. The object provides
    /// both a synchronous interface for processing messages on the calling thread and
    /// an asynchronous interface suitable for populating a queue to be processed by
    /// worker threads.
    /// The Client provides named convenience methods for many AMPS commands. These
    /// methods provide control over commonly used parameters and support most programs.
    /// For full control over AMPS, you build a {@link AMPS.Client.Command}
    /// object and use {@link Client#execute } or {@link Client#executeAsync} to run the
    /// command.
    /// AMPS uses the client name to detect duplicate messages, so the name of each
    /// instance of the client application should be unique.
    /// <p>
    /// </summary>
    /// An example of a simple C# program publishing the JSON message
    /// <code>{ "message" : "Hello, World!"}</code> is listed below.
    ///
    /// <pre>
    /// <code>
    /// class ConnectToAMPS
    ///    {
    ///         static void Main(string[] args)
    ///         {
    ///             using(Client client = new Client("exampleClient"))
    ///             {
    ///                 try
    ///                 {
    ///                     client.connect("tcp://192.168.1.3:9007/amps");
    ///                     client.logon();
    ///                     client.publish("messages", @"{ ""message"" : ""Hello, World!"" }");
    ///                 }
    ///                 catch (AMPSException e)
    ///                 {
    ///                     Console.Error.WriteLine(e);
    ///                 }
    ///              }
    ///           }
    ///     }
    ///  </code>
    ///  </pre>
    ///
    public class Client : IDisposable
    {
        private volatile String name = null;
        private String nameHash = null;
        private volatile String logonCorrelationData = null;
        private Transport transport = null;
        private ulong lastSentHaSequenceNumber = 0;
        private Store publishStore = null;
        private volatile bool connected = false;
        private bool _isRetryOnDisconnect = true;
        private volatile Uri uri = null;
        private ClientDisconnectHandler disconnectHandler = new DefaultDisconnectHandler();
        public delegate void ExceptionListener(Exception e);
        protected ExceptionListener _exceptionListener = null;
        private Message message = null;
        private Message beatMessage = null;
        protected Object internalLock = new Object();
        private Stopwatch heartbeatStopwatch = new Stopwatch();
        private int _heartbeatInterval = 0;
        private int _readTimeout = 0;
        private int _defaultMaxDepth = Int32.MaxValue;
        private Command _command = new Command();
        private ClientStoreReplayer replayer = null;
        private ThreadLocal<Command> _publishCommand = new ThreadLocal<Command>(() =>
                                                            { return new Command(); });

        private const int AMPS_3_8 = 3080000;  // Server version that supports persisted acks for bookmarks.
        private const int MIN_FLUSH_VERSION = 4000000;  // Server version that supports flush command
        public const int MIN_MULTI_BOOKMARK_VERSION = 4000000;  // Server version that supports multiple bookmarks in bookmark subscribe request
        public const int MIN_QUEUE_VERSION = 5000000; // Server version that supports queues
        internal const int DEFAULT_SERVER_VERSION = 99999999;
        private ThreadCreatedHandler _threadCreatedHandler = null;
        private TransportFilterModifiable _transportFilter = null;
        private List<string> _httpPreflightHeaders = null;

        private object acksLock = new object();
        protected volatile int _lastFailedConnectionVersion = -1;

        class AckResponse
        {
            public int ConnectionVersion { get; set; }
            public VersionInfo ServerVersion { get; set; }
            public bool Responded { get; set; }
            public bool Abandoned { get; set; }
            public Message.State State { get; set; }
            public Message.Reasons Reason { get; set; }
            public string ReasonText { get; set; }
            public string Username { get; set; }
            public string Password { get; set; }
            public ulong Sequence { get; set; }
            public string LeasePeriod { get; set; }
            public string Options { get; set; }
            public string Bookmark { get; set; }
        }
        private Dictionary<CommandId, AckResponse> _acks = new Dictionary<CommandId, AckResponse>();
        private MessageRouter _routes = new MessageRouter();
        private ClientHandler handler;
        private int version = Version.AMPS_3;
        private BookmarkStore bookmarkStore = new DefaultBookmarkStore();

        private volatile String username;
        private volatile SubscriptionManager subscriptionManager = new DefaultSubscriptionManager();
        private volatile bool _logonInProgress = false;
        private volatile bool _reconnectingSubscriptionManager = false;
        private bool _defaultSubscriptionManager = true;
        private volatile FailedWriteHandler _failedWriteHandler = null;
        private VersionInfo serverVersion = new VersionInfo("default");

        /// <summary>
        /// Used to specify what type of command should be directed to a global
        /// handler when it is set. All messages of this type will be sent to
        /// the handler in addition to any handlers set for the specific
        /// command, subscription, or query id in the message.
        /// </summary>
        public enum GlobalCommandTypeHandlers : int
        {
            Publish = 0,
            SOW = 1,
            GroupBegin = 2,
            GroupEnd = 3,
            Heartbeat = 4,
            OOF = 5,
            Ack = 6,
            LastChance = 7,
            DuplicateMessage = 8,
            COUNT = 9
        }
        private List<MessageHandler> _globalCommandTypeHandlers = new List<MessageHandler>((int)GlobalCommandTypeHandlers.COUNT);

        /// <summary>
        /// Provides data for the {@link ConnectionStateChanged} event.
        /// </summary>
        public class ConnectionStateChangedEventArgs : EventArgs
        {
            /// <summary>
            /// Constants to represent the state of a connection.
            /// </summary>
            public enum ConnectionState
            {
                Disconnected,
                Shutdown,
                Connected,
                LoggedOn,
                PublishReplayed,
                HeartbeatInitiated,
                Resubscribed,
                UNKNOWN
            };

            /// <summary>
            /// Property to hold the connection state for this event.
            /// </summary>
            public ConnectionState NewState { get; set; }

            /// <summary>
            /// Initializes a new instance of the ConnectionStateChangedEventArgs class
            /// using the specified connection state.
            /// </summary>
            public ConnectionStateChangedEventArgs(ConnectionState newState)
            {
                NewState = newState;
            }
        }

        /// <summary>
        /// Represents the method to be called when the connection state for this client
        /// changes. Notice that this event is raised as soon as the client detects the
        /// change of state. This means that, when the event is raised because the client
        /// is connected, the client has not yet logged on or completed any recovery (such
        /// as replaying the publish store, re-entering subscriptions,and so on). The event
        /// handler should not issue commands on the client while recovery is in progress.
        /// </summary>
        public event EventHandler<ConnectionStateChangedEventArgs> ConnectionStateChanged;

        /// <summary>
        /// Broadcasts the connection state changed event to registered subscribers.
        /// </summary>
        /// <param name="args">
        protected void broadcastConnectionStateChangedEvent(ConnectionStateChangedEventArgs args)
        {
            if (ConnectionStateChanged != null) ConnectionStateChanged(this, args);
        }

        // Queue Acking members //
        private long _queueAckTimeout  = 0;
        private bool _isAutoAckEnabled = false;
        private int _ackBatchSize = 0;
        private CommandId _queueAckCommandId = CommandId.nextIdentifier();
        private ThreadLocal<byte[]> _bookmarksData = new ThreadLocal<byte[]>(() =>
                                                            { return new byte[128]; });
        private static OptionsField CANCEL_FIELD = new OptionsField(Message.Options.Cancel);
        private class QueueBookmarks
        {
            internal byte[] _topic;
            internal byte[] _data;
            internal DateTime _oldestTime;
            internal volatile int _bookmarkCount;
            internal int _length;
        }
        private Dictionary<long, QueueBookmarks> _topicHashMap = new Dictionary<long, QueueBookmarks>();
        private volatile int _topicHashMapSize = 0;
        /// <summary>
        /// Static class containing version constants for AMPS.
        /// </summary>
        public static class Version
        {
            public static int AMPS_1 = 0x00010000;
            public static int AMPS_2 = 0x00020000;
            public static int AMPS_3 = 0x00030000;
        }

        /// <summary>
        /// Functions that take a bookmark (String) such as bookmarkSubscribe()
        /// can be passed a literal bookmark ID, or one of these special values.
        /// </summary>
        public sealed class Bookmarks
        {
            /// <summary>
            /// Indicates that the most recent bookmark logged should be used.
            /// </summary>
            public const String MOST_RECENT = "recent";

            /// <summary>
            /// Indicates that the subscription should begin at the beginning.
            /// </summary>
            public const String EPOCH = "0";

            /// <summary>
            /// Indicates that the subscription should begin "now", without replay.
            /// </summary>
            public const String NOW = "0|1|";
        }

        /// <summary>
        /// Creates a new AMPS client
        /// </summary>
        /// <param name="name">A name to identify this client.
        /// This name is used for duplicate message detection and should be
        /// unique. AMPS does not enforce specific restrictions on the
        /// character set used, however some protocols (for example, XML) may
        /// not allow specific characters. 60East recommends that the client
        /// name be meaningful, short, human readable, and avoid using control
        /// characters, newline characters, or square brackets</param>
        public Client(String name)
            : this(name, null, null)
        {
        }

        /// <summary>
        /// Creates a new client
        /// </summary>
        /// <param name="name">A name to identify this client.
        /// This name is used for duplicate message detection and should be
        /// unique. AMPS does not enforce specific restrictions on the
        /// character set used, however some protocols (for example, XML) may
        /// not allow specific characters. 60East recommends that the client
        /// name be meaningful, short, human readable, and avoid using control
        /// characters, newline characters, or square brackets</param>
        /// <param name="version">The AMPS server version to connect to</param>
        public Client(String name, int version)
            : this(name, null, null, version)
        {
        }

        /// <summary>
        /// Creates a client with a persistent store.
        /// </summary>
        /// <param name="name">A name to identify this client.
        /// This name is used for duplicate message detection and should be
        /// unique. AMPS does not enforce specific restrictions on the
        /// character set used, however some protocols (for example, XML) may
        /// not allow specific characters. 60East recommends that the client
        /// name be meaningful, short, human readable, and avoid using control
        /// characters, newline characters, or square brackets</param>
        /// <param name="store">Store to use for the client</param>
        public Client(String name, Store store)
            : this(name, null, store)
        {
        }

        /// <summary>
        /// Creates a client with a persistent store.
        /// </summary>
        /// <param name="name">A name to identify this client.
        /// This name is used for duplicate message detection and should be
        /// unique. AMPS does not enforce specific restrictions on the
        /// character set used, however some protocols (for example, XML) may
        /// not allow specific characters. 60East recommends that the client
        /// name be meaningful, short, human readable, and avoid using control
        /// characters, newline characters, or square brackets</param>
        /// <param name="store">Store to use for the client</param>
        /// <param name="version">Server version connecting to</param>
        ///
        public Client(String name, Store store, int version)
            : this(name, null, store, version)
        {
        }


        /// <summary>
        /// Creates a client with a transport.
        /// </summary>
        /// <param name="name">A name to identify this client.
        /// This name is used for duplicate message detection and should be
        /// unique. AMPS does not enforce specific restrictions on the
        /// character set used, however some protocols (for example, XML) may
        /// not allow specific characters. 60East recommends that the client
        /// name be meaningful, short, human readable, and avoid using control
        /// characters, newline characters, or square brackets</param>
        /// <param name="transport">Transport to use with the client</param>
        ///
        public Client(String name, Transport transport)
            : this(name, transport, null)
        {
        }

        /// <summary>
        /// Creates a client with a transport.
        /// </summary>
        /// <param name="name">A name to identify this client.
        /// This name is used for duplicate message detection and should be
        /// unique. AMPS does not enforce specific restrictions on the
        /// character set used, however some protocols (for example, XML) may
        /// not allow specific characters. 60East recommends that the client
        /// name be meaningful, short, human readable, and avoid using control
        /// characters, newline characters, or square brackets</param>
        /// <param name="transport">Transport to use with the client</param>
        /// <param name="version">Server version connecting to</param>
        ///
        public Client(String name, Transport transport, int version)
            : this(name, transport, null, version)
        {
        }

        /// <summary>
        /// Creates a client with a transport and persistent store.
        /// </summary>
        /// <param name="name">A name to identify this client.
        /// This name is used for duplicate message detection and should be
        /// unique. AMPS does not enforce specific restrictions on the
        /// character set used, however some protocols (for example, XML) may
        /// not allow specific characters. 60East recommends that the client
        /// name be meaningful, short, human readable, and avoid using control
        /// characters, newline characters, or square brackets.
        /// <param name="transport">Transport to use with the client</param>
        /// <param name="store">Store to use for the client</param>
        ///
        public Client(String name, Transport transport, Store store)
            : this(name, transport, store, Version.AMPS_3)
        {
        }

        /// <summary>
        /// Creates a client with a transport and persistent store.
        /// </summary>
        /// <param name="name">A name to identify this client.
        /// This name is used for duplicate message detection and should be
        /// unique. AMPS does not enforce specific restrictions on the
        /// character set used, however some protocols (for example, XML) may
        /// not allow specific characters. 60East recommends that the client
        /// name be meaningful, short, human readable, and avoid using control
        /// characters, newline characters, or square brackets.
        /// <param name="transport">Transport to use with the client</param>
        /// <param name="store">Store to use for the client</param>
        /// <param name="version">Server version connecting to</param>
        ///
        public Client(String name, Transport transport, Store store, int version)
        {
            this.name = name;
            this.handler = new ClientHandler(this);
            if (transport != null)
            {
                this.transport = transport;
                this.transport.setMessageHandler(getTransportMessageHandler());
                this.transport.setDisconnectHandler(getTransportDisconnectHandler());
                this.transport.setIdleAction(new Action(_ackFlusher));
                this.transport.initFromClient(this);
                this.message = this.transport.allocateMessage();
                this.beatMessage = this.transport.allocateMessage();
                this.beatMessage.Command = Message.Commands.Heartbeat;
                this.beatMessage.setOptions("beat");
            }
            this.publishStore = store;
            this.version = version;
            this.replayer = new ClientStoreReplayer(this);
            for (int i=0; i<(int)GlobalCommandTypeHandlers.COUNT; ++i)
            {
                _globalCommandTypeHandlers.Add(new DefaultMessageHandler());
            }
        }

        /// <summary>
        /// Return the name of the Client.
        /// </summary>
        /// <returns>String Client name</returns>
        public String getName()
        {
            return this.name;
        }

        /// <summary>
        /// Return the name hash of the Client as returned by the server after logon.
        /// </summary>
        /// <returns>String Client name hash</returns>
        public String getNameHash()
        {
            return this.nameHash;
        }

        /// <summary>
        /// Return the name hash of the Client as a ulong as returned by the server after logon.
        /// </summary>
        /// <returns>String Client name hash</returns>
        public ulong getNameHashValue()
        {
            if (this.nameHash == null)
            {
                return 0;
            }
            return System.Convert.ToUInt64(this.nameHash);
        }

        /// <summary>
        /// Return the server version retrieved during logon. If the client has
        /// not logged on or is connected to a server whose version is less than
        /// 3.8.0.0 this function will return 0.
        /// The version uses 2 digits each for major minor maintenance and hotfix
        /// i.e., 3.8.1.5 will return 3080105
        /// Versions of AMPS prior to 3.8.0.0 did not return the server version
        /// to the client in the logon ack.
        /// </summary>
        /// <returns> The version of the server as a long.</returns>
        ///
        public int getServerVersion()
        {
            try
            {
                return getVersionAsInt(this.serverVersion.VersionString);
            }
            catch (CommandException)
            {
                return DEFAULT_SERVER_VERSION;
            }
        }

        /// <summary>
        /// Return the server version retrieved during logon. If the client has
        /// not logged on or is connected to a server whose version is less than
        /// 3.8.0.0 this function will return the latest release version.
        /// The version is returned as a VersionInfo.
        /// Versions of AMPS prior to 3.8.0.0 did not return the server version
        /// to the client in the logon ack.
        /// </summary>
        /// <returns><typeparamref name="VersionInfo"/>the version of the server.</returns>
        ///
        public VersionInfo getServerVersionInfo()
        {
            return this.serverVersion;
        }

        /// <summary>
        /// Return the numeric value for the given version string with the pattern:
        /// Major.Minor.Maintenance.Hotfix
        /// The version uses 2 digits each for major minor maintenance and hotfix
        /// i.e., 3.8.1.5 will return 3080105
        /// Version strings passed in can be shortened to not include all levels
        /// so 3.8 will return 3080000.
        /// </summary>
        /// <param name="version"> The version string to convert</param>
        /// <returns> The version as a int.</returns>
        /// <exception cref="CommandException">The string doesn't represent a valid version.</exception>
        static public int getVersionAsInt(String version)
        {
            if (String.IsNullOrEmpty(version))
                return 0;

            int retVersion = 0;
            int dots = 0;
            int lastDot = -1;
            char c;
            for (int i=0; i<version.Length && dots<4; ++i)
            {
                c = version[i];
                if (c == '.')
                {
                    ++dots;
                    retVersion *= 10;
                    if (i-lastDot > 5)
                    {
                        throw new CommandException("Too many digits between dots found translating version string.");
                    }
                    else if (i-lastDot >= 4)
                    {
                        retVersion *= 10;
                        retVersion += 99;
                    }
                    else
                    {
                        if (i-lastDot == 3)
                        {
                            retVersion += version[i-2] - '0';
                        }
                        retVersion *= 10;
                        retVersion += version[i-1] - '0';
                    }
                    lastDot = i;
                }
                else if (c < '0' || c > '9')
                {
                    throw new CommandException("Invalid character found in version string.");
                }
                else if (i == version.Length-1)
                {
                    ++dots;
                    retVersion *= 10;
                    if (i-lastDot > 4)
                    {
                        throw new CommandException("Too many digits between dots found translating version string.");
                    }
                    else if (i-lastDot >= 3)
                    {
                        retVersion *= 10;
                        retVersion += 99;
                    }
                    else
                    {
                        if (i-lastDot == 2)
                        {
                            retVersion += version[i-1] - '0';
                        }
                        retVersion *= 10;
                        retVersion += version[i] - '0';
                    }
                }
            }
            for (; dots<4; ++dots) retVersion *= 100;
            return retVersion;
        }

        /// <summary>Returns the URI the client is connected to.</summary>
        /// <returns><typeparamref name="Uri"/>the URI the client is configured
        /// to connect to</returns>
        public Uri getURI()
        {
            return this.uri;
        }

        /// <summary>
        /// Return the underlying transport.
        /// For advanced users, this method provides direct access to the transport.
        /// Care needs to be taken to not modify the underlying transport in ways that
        /// are incompatible with the Client.
        /// </summary>
        /// <returns>Transport Underlying transport instance</returns>
        public Transport getTransport()
        {
            lock (internalLock)
            {
                return this.transport;
            }
        }

        /// <summary>
        /// Sets the underlying bookmark store, which is used to track which messages the
        /// client has received and which messages have been processed by the program. This
        /// method replaces any existing bookmark store for the client, without transferring
        /// the state of that store.
        /// </summary>
        /// <param name="val">The new bookmark store</param>
        /// <exception cref="AlreadyConnectedException">The store cannot be set while the client
        /// is connected. </exception>
        public void setBookmarkStore(BookmarkStore val)
        {
            lock (internalLock)
            {
                if (this.connected)
                    throw new AlreadyConnectedException("Setting a bookmark store on a connected client is undefined behavior.");
            }
            if (val != null)
            {
                this.bookmarkStore = val;
            }
            else
            {
                this.bookmarkStore = new DefaultBookmarkStore();
            }
        }

        /// <summary>
        /// Returns the underlying bookmark store for this client. The bookmark store is used
        /// to track which messages the client has received and which messages have been processed
        /// by the program.
        /// </summary>
        /// <returns>The bookmark store for this client.</returns>
        public BookmarkStore getBookmarkStore()
        {
            return this.bookmarkStore;
        }

        /// <summary>
        /// Sets the underlying publish store, which is used to store published messages
        /// until the AMPS instance has acknowledged those messages. This method replaces
        /// any existing publish store for the client, without transferring the state of
        /// that store.
        /// </summary>
        /// <param name="store">
        /// <exception cref="AlreadyConnectedException"></exception>
        public void setPublishStore(Store store)
        {
            lock (internalLock)
            {
                if (this.connected)
                    throw new AlreadyConnectedException("Setting a publish store on a connected client is undefined behavior.");
            }
            this.publishStore = store;
        }

        /// <summary>
        /// Returns the underlying publish store for this client. The publish store is used to
        /// store published messages until the AMPS instance has acknowledged those messages.
        /// </summary>
        /// <returns>The publish store for this client.</returns>
        public Store getPublishStore()
        {
            return this.publishStore;
        }

        /// <summary>
        /// Connects to the AMPS instance through the provided URI.
        /// <pre>
        /// <code>
        /// namespace AMPSConnectExample
        /// {
        ///     class AMPSConnectExample
        ///     {
        ///         private static string uri_ = "tcp://127.0.0.1:9027/amps/json";
        ///         static void Main(string[] args)
        ///         {
        ///             Client client = new Client("SOWConsolePublisher");
        ///             try
        ///             {
        ///                 System.Console.WriteLine("... starting ...");
        ///                 // connect and logon
        ///                 client.connect(uri_);
        ///                 client.logon();
        ///             }
        ///             catch (AMPSException e)
        ///             {
        ///                 System.Console.WriteLine(e.Message);
        ///                 System.Console.WriteLine(e.StackTrace);
        ///             }
        ///         }
        ///     }
        /// }
        /// </code>
        /// </pre>
        /// The URI is a <typeparamref name="String"/> with the format:
        /// <code>transport://user:password@host:port/protocol</code>
        /// </summary>
        /// <p>The elements in the URI string are as follows:
        /// <ul>
        /// <li><i>transport</i> -- the network transport used for the connection</li>
        /// <li><i>user</i> -- <i>(optional)</i> the userId used to authenticate the connection</li>
        /// <li><i>password</i> -- <i>(optional)</i> the password used to authenticate the connection</li>
        /// <li><i>host</i> -- the hostname or IP address of the host where AMPS is running</li>
        /// <li><i>port</i> -- the port to connect to
        /// <li><i>protocol</i> -- the protocol used for the connection
        /// </ul></p>
        /// <param name="uri">The URI string to connect to
        /// <exception cref="ConnectionRefusedException">The connection could not be established</exception>
        /// <exception cref="AlreadyConnectedException">The connection is already connected</exception>
        /// <exception cref="InvalidURIException">The specified URI is invalid</exception>
        /// <exception cref="ProtocolException">The protocol is invalid</exception>
        /// <exception cref="TransportTypeException">The transport type is invalid</exception>
        public virtual void connect(string uri)
        {
            lock (internalLock)
            {
                Uri connectUri = null;
                try
                {
                    connectUri = new Uri(uri);
                }
                catch (UriFormatException urisex)
                {
                    throw new InvalidURIException(urisex);
                }
                String[] sPath = connectUri.AbsolutePath.Substring(1, connectUri.AbsolutePath.Length - 1).Split('/');
                String sProtocol = sPath[sPath.Length-1];
                if (sPath.Length > 1)
                {
                    if (!sProtocol.Equals("amps"))
                    {
                        sProtocol = sPath[sPath.Length-2];
                    }
                    if (!sProtocol.Equals("amps"))
                    {
                        throw new InvalidURIException("Specification of message type requires amps protocol");
                    }
                }
                if (this.transport == null)
                {
                    this.uri = connectUri;
                    Properties props = new URIProperties(connectUri);
                    Protocol protocol = ProtocolFactory.createProtocol(sProtocol, props);
                    this.transport = TransportFactory
                          .createTransport(connectUri.Scheme, protocol, props);
                    this.message = this.transport.allocateMessage();
                    this.beatMessage = this.transport.allocateMessage();
                    this.beatMessage.Command = Message.Commands.Heartbeat;
                    this.beatMessage.setOptions("beat");
                    transport.setIdleAction(new Action(_ackFlusher));
                    if (this.publishStore != null)
                        this.publishStore.setMessage(this.transport.allocateMessage());
                    this.transport.setMessageHandler(getTransportMessageHandler());
                    this.transport.setDisconnectHandler(getTransportDisconnectHandler());
                    this.transport.initFromClient(this);
                }
                else if (this.uri == null)
                {
                    this.uri = connectUri;
                }
                else
                {
                    String[] oldPath = this.uri.AbsolutePath.Substring(1, this.uri.AbsolutePath.Length - 1).Split('/');
                    String oldProtocol = oldPath[oldPath.Length-1];
                    if (oldPath.Length > 1 && !oldProtocol.Equals("amps"))
                    {
                        oldProtocol = oldPath[oldPath.Length-2];
                    }
                    if (!this.uri.Scheme.Equals(connectUri.Scheme))
                    {
                        throw new ConnectionException($"Client cannot change transport type from {this.uri.Scheme} to {connectUri.Scheme}");
                    }
                    if (!oldProtocol.Equals(sProtocol))
                    {
                        throw new ConnectionException($"Client cannot change protocol type from {oldProtocol} to {sProtocol}");
                    }
                    this.uri = connectUri;
                }
                this.transport.connect(this.uri);
                if(ConnectionStateChanged!=null) ConnectionStateChanged(this, new ConnectionStateChangedEventArgs(ConnectionStateChangedEventArgs.ConnectionState.Connected));
                this.connected = true;
            }
        }

        /// <summary>
        /// Sets the <typeparamref name="ClientDisconnectHandler"/>. In the event that the Client is unintentionally
        /// disconnected from AMPS, the <code>invoke</code> method from the
        /// ClientDisconnectHandler will be invoked.
        /// </summary>
        /// <param name="disconnectHandler">The disconnect handler</param>
        public void setDisconnectHandler(ClientDisconnectHandler disconnectHandler)
        {
            if (disconnectHandler != null)
            {
                this.disconnectHandler = disconnectHandler;
            }
            else
            {
                this.disconnectHandler = new DefaultDisconnectHandler();
            }
        }

        /// <summary>
        /// Sets the lambda function called in the event of client disconnect
        /// </summary>
        /// <param name="disconnectHandler">
        public void setDisconnectHandler(Action<Client> disconnectHandler)
        {
            if (disconnectHandler != null)
            {
                this.disconnectHandler = new ActionDisconnectHandler(disconnectHandler);
            }
            else
            {
                this.disconnectHandler = new DefaultDisconnectHandler();
            }
        }

        /// <summary>
        /// Gets the currently configured <typeparamref name="ClientDisconnectHandler"/>
        /// </summary>
        /// <returns>returns the currently configured <typeparamref name="ClientDisconnectHandler"/></returns>
        public ClientDisconnectHandler getDisconnectHandler()
        {
            return this.disconnectHandler;
        }

        /// <summary>
        /// Sets the <typeparamref name="MessageHandler"/> instance used when no other handler matches.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke when no other handler matches</param>
        public void setLastChanceMessageHandler(MessageHandler messageHandler)
        {
            if (messageHandler != null)
                this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.LastChance] = messageHandler;
            else
                this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.LastChance] = new DefaultMessageHandler();
        }

        /// <summary>
        /// Sets the <typeparamref name="MessageHandler"/> instance used when no other handler matches.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke when no other handler matches</param>
        public void setLastChanceMessageHandler(Action<Message> messageHandler)
        {
            if (messageHandler != null)
                this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.LastChance] = new ActionMessageHandler(messageHandler);
            else
                this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.LastChance] = new DefaultMessageHandler();
        }

        /// <summary>
        /// Sets the <typeparamref name="MessageHandler"/> instance used when no other handler matches.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke when no other handler matches</param>
        [Obsolete("Use setLastChanceMessageHandler instead.")]
        public void setUnhandledMessageHandler(MessageHandler messageHandler)
        {
            setLastChanceMessageHandler(messageHandler);
        }

        /// <summary>
        /// Sets the lambda function used when no other handler matches.
        /// </summary>
        /// <param name="messageHandler">The function to invoke when no other handler matches</param>
        [Obsolete("Use setLastChanceMessageHandler instead.")]
        public void setUnhandledMessageHandler(Action<Message> messageHandler)
        {
            setLastChanceMessageHandler(messageHandler);
        }

        /// <summary>
        /// Sets the ExceptionListener instance used for communicating absorbed exceptions.
        /// </summary>
        /// <param name="exceptionListener">The exception listener instance to invoke for exceptions</param>
        public void setExceptionListener(ExceptionListener exceptionListener)
        {
            _exceptionListener = exceptionListener;
            if (_exceptionListener != null && this.transport != null)
            {
                this.transport.setExceptionListener(new TCPTransportImpl.ExceptionListener(_exceptionListener));
            }
        }

        /// <summary>
        /// Gets the currently configured <typeparamref name="ExceptionListener"/>
        /// </summary>
        /// <returns>returns the currently configured <typeparamref name="ExceptionListener"/></returns>
        public ExceptionListener getExceptionListener()
        {
            return _exceptionListener;
        }

        /// <summary>
        /// Sets the <typeparamref name="SubscriptionManager"/> instance used
        /// for recording active subscriptions.
        /// </summary>
        /// <param name="subscriptionManager">The subscription manager invoked
        /// when there is a change to the subscriptions</param>
        public void setSubscriptionManager(SubscriptionManager subscriptionManager)
        {
            if (subscriptionManager == null)
            {
                this.subscriptionManager = new DefaultSubscriptionManager();
                _defaultSubscriptionManager = true;
            }
            else
            {
                this.subscriptionManager = subscriptionManager;
                _defaultSubscriptionManager = subscriptionManager is DefaultSubscriptionManager;
            }
        }

        /// <summary>
        /// Gets the currently configured <typeparamref name="SubscriptionManager"/>
        /// </summary>
        /// <returns>returns the currently configured <typeparamref name="SubscriptionManager"/></returns>
        public SubscriptionManager getSubscriptionManager()
        {
            return this.subscriptionManager;
        }

        /// Handles exceptions by invoking the registered exception listener.
        protected void absorbedException(Exception e)
        {
            try
            {
                if (_exceptionListener != null)
                {
                    _exceptionListener(e);
                }
            }
            catch (Exception)
            {
                // just absorb any exceptions from the exception listener.
            }
        }

        /// <summary>
        /// Sets the <typeparamref name="MessageHandler"/> that will be called in the
        /// event of a message of the given command type.
        /// </summary>
        /// <param name="command_"> The type of command to send to messageHandler_</param>
        /// <param name="messageHandler"><typeparamref name="MessageHandler"/>
        /// to invoke</param>
        public void setGlobalCommandTypeMessageHandler(string command_, MessageHandler messageHandler_)
        {
            MessageHandler messageHandler = messageHandler_;
            if (messageHandler == null)
                messageHandler = new DefaultMessageHandler();
            switch (command_[0])
            {
                /* Not currently allowing some types to avoid extra branching
                 * in message delivery.
                case 'p':
                    this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.Publish] = messageHandler;
                    break;
                case 's':
                    this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.SOW] = messageHandler;
                    break;
                case 'g':
                    if (command_[6] == 'b'
                    {
                        this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.GroupBegin] = messageHandler;
                    }
                    else if (command_[6] == 'e'
                    {
                        this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.GroupEnd] = messageHandler;
                    }
                    else
                    {
                        throw new CommandException("Invalid command '" + command_ + "' passed to setGlobalCommandTypeMessageHandler");
                    }
                    break;
                case 'o':
                    this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.OOF] = messageHandler;
                    break;
                */
                case 'h':
                    this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.Heartbeat] = messageHandler;
                    break;
                case 'a':
                    this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.Ack] = messageHandler;
                    break;
                case 'l':
                case 'L':
                    this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.LastChance] = messageHandler;
                    break;
                case 'd':
                case 'D':
                    this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.DuplicateMessage] = messageHandler;
                    break;
                default:
                    throw new CommandException("Invalid command '" + command_ + "' passed to setGlobalCommandTypeMessageHandler");
            }
        }

        /// <summary>
        /// Sets the <typeparamref name="MessageHandler"/> that will be called in the
        /// event of a message of the given command type.
        /// </summary>
        /// <param name="command_"><typeparamref name="Message.Commands"/> type of command
        /// to send to messageHandler_</param>
        /// <param name="messageHandler"><typeparamref name="MessageHandler"/>
        /// to invoke</param>
        public void setGlobalCommandTypeMessageHandler(Message.Commands command_, MessageHandler messageHandler_)
        {
            MessageHandler messageHandler = messageHandler_;
            if (messageHandler == null)
                messageHandler = new DefaultMessageHandler();
            switch (command_)
            {
                /* Not currently allowing some types to avoid extra branching
                 * in message delivery.
                case Message.Commands.Publish:
                    this._globalCommandTypeHandlers[GlobalCommandTypeHandlers.Publish] = messageHandler;
                    break;
                case Message.Commands.SOW:
                    this._globalCommandTypeHandlers[GlobalCommandTypeHandlers.SOW] = messageHandler;
                    break;
                case Message.Commands.GroupBegin:
                    this._globalCommandTypeHandlers[GlobalCommandTypeHandlers.GroupBegin] = messageHandler;
                    break;
                case Message.Commands.GroupEnd:
                    this._globalCommandTypeHandlers[GlobalCommandTypeHandlers.GroupEnd] = messageHandler;
                    break;
                case Message.Commands.OOF:
                    this._globalCommandTypeHandlers[GlobalCommandTypeHandlers.OOF] = messageHandler;
                    break;
                */
                case Message.Commands.Heartbeat:
                    this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.Heartbeat] = messageHandler;
                    break;
                case Message.Commands.Ack:
                    this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.Ack] = messageHandler;
                    break;
                default:
                    throw new CommandException("Invalid command '" + CommandField.encodeCommand(command_) + "' passed to setGlobalCommandTypeMessageHandler");
            }
        }

        /// <summary>
        /// Sets the <typeparamref name="MessageHandler"/> that will be called in the
        /// event of a message of the given command type.
        /// </summary>
        /// <param name="command_"><typeparamref name="GlobalCommandTypeHandlers"/> type of command
        /// to send to messageHandler_</param>
        /// <param name="messageHandler"><typeparamref name="MessageHandler"/>
        /// to invoke</param>
        public void setGlobalCommandTypeMessageHandler(GlobalCommandTypeHandlers command_,
                                                       MessageHandler messageHandler_)
        {
            this._globalCommandTypeHandlers[(int)command_] = messageHandler_;
        }

        /// <summary>
        /// Sets the <typeparamref name="MessageHandler"/> that will be called in the
        /// event of a duplicate message
        /// </summary>
        /// <param name="messageHandler"><typeparamref name="MessageHandler"/>
        /// to invoke in the event of duplicate messages</param>
        public void setDuplicateMessageHandler(MessageHandler messageHandler)
        {
            if (messageHandler != null)
            {
                this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.DuplicateMessage] = messageHandler;
            }
            else
            {
                this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.DuplicateMessage] = new DefaultMessageHandler();
            }
        }

        /// <summary>
        /// Returns the <typeparamref name="MessageHandler"/> that will be called in the
        /// event of a duplicate message
        /// </summary>
        /// <returns name="messageHandler"><typeparamref name="MessageHandler"/>
        /// to invoke in the event of duplicate messages
        public MessageHandler getDuplicateMessageHandler()
        {
            return this._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.DuplicateMessage];
        }

        /// <summary>Sets the <typeparamref name="FailedWriteHandler"/>
        /// instance that will be called in the event that messages fail to be written</summary>
        /// <param name="handler_"><typeparamref name="FailedWriteHandler"/> to invoke for failed writes</param>
        public void setFailedWriteHandler(FailedWriteHandler handler_)
        {
            this._failedWriteHandler = handler_;
        }

        /// <summary>Sets the <typeparamref name="FailedWriteHandlerV4"/>
        /// instance that will be called in the event that messages fail to be written</summary>
        /// <param name="handler_"><typeparamref name="FailedWriteHandlerV4"/> to invoke for failed writes</param>
        public void setFailedWriteHandler(FailedWriteHandlerV4 handler_)
        {
            this._failedWriteHandler = new FailedWriteHandlerV4Compat(handler_);
        }

        /// <summary>Returns the <typeparamref name="FailedWriteHandler"/>
        /// instance that will be called in the event that messages fail to be written</summary>
        /// <returns><typeparamref name="FailedWriteHandler"/> to invoke for failed writes.
        public FailedWriteHandler getFailedWriteHandler()
        {
            return this._failedWriteHandler;
        }

        /// <summary>
        /// Disconnect from the AMPS server.
        /// </summary>
        public virtual void disconnect()
        {
            if (this.transport != null)
            {
                this.transport.setIdleAction(null);
            }
            _disconnect(false);
            if (ConnectionStateChanged != null)
            {
                ConnectionStateChanged(this, new ConnectionStateChangedEventArgs(ConnectionStateChangedEventArgs.ConnectionState.Shutdown));
            }
        }

        /// <summary>
        /// Disconnect from the AMPS server.
        /// <pre>
        /// <code>
        /// namespace AMPSDisconnectExample
        /// {
        ///     class AMPSDisconnectExample
        ///     {
        ///         private static string uri_ = "tcp://127.0.0.1:9027/amps/json";
        ///         static void Main(string[] args)
        ///         {
        ///             Client client = new Client("SOWConsolePublisher");
        ///             try
        ///             {
        ///                 System.Console.WriteLine("... starting ...");
        ///                 // connect and logon
        ///                 client.connect(uri_);
        ///                 client.logon();
        ///                 client.publish("messages-sow",
        ///                     "{ \"messageNumber\":50000, \"message\":\"Here and then gone...\"}",
        ///                     3);
        ///                 // wait up to 2 seconds for all messages to be published
        ///                 client.flush(2000);
        ///                 client.disconnect();
        ///                 System.Console.WriteLine("... done ...");
        ///             }
        ///             catch (AMPSException e)
        ///             {
        ///                 System.Console.WriteLine(e.Message);
        ///             }
        ///         }
        ///     }
        /// }
        /// </code>
        /// </pre>
        /// </summary>
        internal virtual void _disconnect(bool reconnecting_)
        {
            Transport currentTransport = null;
            if (!reconnecting_)
            {
                flushAcks();
            }
            lock (internalLock)
            {
                currentTransport = this.transport;
                heartbeatStopwatch = new Stopwatch();
                if (ConnectionStateChanged != null && this.connected)
                {
                    ConnectionStateChanged(this, new ConnectionStateChangedEventArgs(ConnectionStateChangedEventArgs.ConnectionState.Disconnected));
                }
                this.connected = false;
            }

            // only disconnect if a transport has been created
            if (currentTransport != null)
            {
                currentTransport.disconnect();
            }

            lock (internalLock)
            {
                _routes.clear();
                // Signal and then clear the ack info
                cancelSynchronousWaiters(Int32.MaxValue);
            }
        }

        /// <summary>
        /// Disconnect from the AMPS server.
        /// </summary>
        public void close()
        {
            disconnect();
        }

        /// <summary>
        /// Creates a new <typeparamref name="Message"/> appropriate for this client. This function should be
        /// called rarely, since it does allocate a handful of small objects.  Users sensitive
        /// to garbage collection delays should cache the message object for later usage.
        /// </summary>
        /// <returns>A new <typeparamref name="Message"/> instance</returns>
        /// </summary>
        public Message allocateMessage()
        {
            return this.transport.allocateMessage();
        }

        /// <summary>
        /// Sets a heartbeat on this connection.  When a heartbeat is set, the client
        /// periodically sends a heartbeat message to AMPS. This helps the client
        /// more quickly detect disconnection from AMPS.
        /// </summary>
        /// <param name="heartbeatInterval_">the interval (in seconds) at which to send a heartbeat message</param>
        /// <param name="readTimeout_">the length of time (in seconds) to wait for a response to a heartbeat message</param>
        public virtual void setHeartbeat(int heartbeatInterval_, int readTimeout_)
        {
            _heartbeatInterval = heartbeatInterval_;
            _readTimeout = Math.Max(readTimeout_, heartbeatInterval_);
            _sendHeartbeat();
        }

        /// <summary>
        /// Sets a heartbeat on this connection.  When a heartbeat is set, the client
        /// periodically sends a heartbeat message to AMPS. This helps the client
        /// more quickly detect disconnection from AMPS.
        /// </summary>
        /// <param name="heartbeatInterval_">the interval (in seconds) at which to send a heartbeat message</param>
        public void setHeartbeat(int heartbeatInterval_)
        {
            setHeartbeat(heartbeatInterval_, heartbeatInterval_ * 2);
        }

        /// <summary>
        /// Sends the configured heartbeat to the server if a connection is
        /// available.
        /// </summary>
        void _sendHeartbeat()
        {
            lock (this.internalLock)
            {
                if (this.transport != null && _heartbeatInterval > 0)
                {
                    message.reset();
                    message.setCommand(Message.Commands.Heartbeat);
                    message.setOptions(String.Format("start,{0}", _heartbeatInterval));
                    this.heartbeatStopwatch = Stopwatch.StartNew();
                    this.transport.sendWithoutRetry(message);
                    this.transport.setReadTimeout(_readTimeout * 1000);
                    if (ConnectionStateChanged != null)
                        ConnectionStateChanged(this, new ConnectionStateChangedEventArgs(ConnectionStateChangedEventArgs.ConnectionState.HeartbeatInitiated));
                }
            }
        }

        /// <summary>
        /// Send a <typeparamref name="Message"/> to AMPS via the <typeparamref name="Transport"/> used in the Client.
        /// </summary>
        /// <param name="m">The message to send</param>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        public void send(Message m)
        {
            lock (this.internalLock)
            {
                sendInternal(m);
            }
        }

        /// <summary>
        /// Adds a MessageHandler for a given CommandId to self.
        /// </summary>
        /// <param name="commandId_">The command, query, or subid used for this route</param>
        /// <param name="messageHandler_">The message handler to route to</param>
        /// <param name="requestedAcks_">The acks requested to be sent to the handler</param>
        /// <param name="isSubscribe_">True if this route is for an ongoing subscription</param>
        public void addMessageHandler(CommandId commandId_,
                                      MessageHandler messageHandler_,
                                      Message.AckTypes requestedAcks_,
                                      bool isSubscribe_)
        {
            _routes.addRoute(commandId_, messageHandler_,
                             requestedAcks_,
                             Message.AckTypes.None,
                             isSubscribe_);
        }

        /// Remove a route from self.
        /// <param name="commandId_">The route to remove</param>
        /// <returns>True if the route was removed.</returns>
        public bool removeMessageHandler(CommandId commandId_)
        {
            return _routes.removeRoute(commandId_);
        }

        private int sendWithoutRetry(Message message)
        {
            if (this.transport != null)
            {
                if (_logonInProgress && message.getCommand() != Message.Commands.Logon)
                {
                    throw new DisconnectedException("The client has been disconnected.");
                }
                int version = transport.getVersion();
                transport.sendWithoutRetry(message);
                return version;
            }
            else
            {
                throw new DisconnectedException("There is no current connection.");
            }
        }

        // Used internally to send a message, handling potential exceptions and disconnections.
        // parameter message is the message to be sent
        // parameter haSeq is the high-availability sequence number
        // parameter isSubscribe is the Flag indicating if it is a subscription message.
        private int sendInternal(Message message, ulong haSeq = 0, bool isSubscribe = false)
        {
            // We may need to make a local copy of message, if we
            // need to let go of our lock.
            bool isLocalMessage = false;
            while (true)
            {
                if (haSeq != 0 && _logonInProgress)
                {
                    if (!_isRetryOnDisconnect)
                        throw new DisconnectedException();
                    if (!isLocalMessage)
                    {
                        message = message.copy();
                        isLocalMessage = true;
                    }

                    Monitor.Wait(internalLock, 100);
                    continue;
                }
                // this.transport is only safe to access while in the lock.
                Transport currentTransport = this.transport;
                if (currentTransport == null)
                    throw new DisconnectedException("The client is not connected.");
                int version = currentTransport.getVersion();
                if ((haSeq != 0 && (haSeq <= lastSentHaSequenceNumber
                                    || lastSentHaSequenceNumber == 0))
                    || (isSubscribe && _reconnectingSubscriptionManager))
                {
                    return -1;
                }
                // It's possible to get here out of sequence number order,
                // try sending lower sequence number messages first.
                while (haSeq > (lastSentHaSequenceNumber + 1))
                {
                    try
                    {
                        // This increments lastSentSequenceNumber
                        if (!publishStore.replaySingle(replayer,
                                                  lastSentHaSequenceNumber + 1))
                        {
                            Monitor.Wait(internalLock, 1000);
                        }
                    }
                    catch (AMPSException)
                    {
                        return version;
                    }
                }
                try
                {
                    bool isLogonCmd = message.getCommand() == Message.Commands.Logon;
                    while (_logonInProgress && !isLogonCmd)
                    {
                        Monitor.Wait(internalLock, 1000);
                    }
                    currentTransport.sendWithoutRetry(message);
                    if (haSeq != 0)
                    {
                        ++lastSentHaSequenceNumber;
                        Monitor.PulseAll(internalLock);
                    }
                    return version;
                }
                catch (DisconnectedException ex)
                {
                    if (_isRetryOnDisconnect && message.getCommand() != Message.Commands.Logon)
                    {
                        // attempt to reconnect.  If we get anything other than
                        // a Retry(meaning it succeeded), let the exception
                        // flow upwards.
                        if (!isLocalMessage)
                        {
                            message = message.copy();
                            isLocalMessage = true;
                        }

                        Monitor.Exit(internalLock);
                        try
                        {
                            currentTransport.handleCloseEvent(currentTransport.getVersion(), "Disconnected", ex);
                        }
                        catch (RetryOperationException)
                        {
                            if (isSubscribe || haSeq != 0) return version;
                        }
                        finally
                        {
                            Monitor.Enter(internalLock);
                        }
                    }
                    else
                    {
                        throw;
                    }
                }
            }
        }

        // Registers a message handler for a specific command, subscription, or message ID.
        // param name="qid": Command ID for the query</param>
        // param name="subid">Command ID for the subscription</param>
        // param name="cid">Command ID for the message</param>
        // param name="messageHandler">Message handler to invoke</param>
        // param name="requestedAcks">Requested acknowledgment types</param>
        // param name="systemAddedAcks">System-added acknowledgment types</param>
        // param name="isSubscribe">Flag indicating if it is a subscription</param>
        // returns: True if the handler is successfully registered.
        private bool _registerHandler(CommandId qid, CommandId subid,
                                     CommandId cid,
                                     MessageHandler messageHandler,
                                     Message.AckTypes requestedAcks,
                                     Message.AckTypes systemAddedAcks,
                                     bool isSubscribe)
        {
            bool added = (qid != null && qid.id != null) ||
                         (subid != null && subid.id != null) ||
                         (cid != null && cid.id != null);
            if(qid != null && qid.id != null && _routes.findRoute(qid) == null)
            {
                _routes.addRoute(qid, messageHandler, requestedAcks,
                                 systemAddedAcks, isSubscribe);
                added = true;
            }
            if(subid != null && subid.id != null &&
               _routes.findRoute(subid) == null)
            {
                if(qid == null || qid.id == null || !subid.Equals(qid)) {
                    _routes.addRoute(subid, messageHandler, requestedAcks,
                                     systemAddedAcks, isSubscribe);
                }
                if (requestedAcks != Message.AckTypes.None && !cid.Equals(subid))
                    _routes.addRoute(cid, messageHandler, requestedAcks,
                                     systemAddedAcks, isSubscribe);
                added = true;
            }
            else if (cid != null && cid.id != null &&
                     (qid == null || qid.id == null || !cid.Equals(qid)) &&
                     _routes.findRoute(cid) == null)
            {
                _routes.addRoute(cid, messageHandler, requestedAcks,
                                 systemAddedAcks, isSubscribe);
                added = true;
            }
            return added;
        }

        /// <summary>
        /// Set whether or not messages being sent to the server should retry if the
        /// client is disconnected. This is most useful if you are publishing data that
        /// has a very short lifetime and so is no longer relevant after the time it
        /// takes to reconnect.
        /// </summary>
        /// <param name="isRetryOnDisconnect">False to disable default behavior of automatic retry</param>
        public void setRetryOnDisconnect(bool isRetryOnDisconnect)
        {
            _isRetryOnDisconnect = isRetryOnDisconnect;
        }

        /// <summary>
        /// Return whether or not messages being sent to the server will retry if the
        /// client is disconnected.
        /// </summary>
        /// <returns>true if it will retry, false if not.</returns>
        public bool getRetryOnDisconnect()
        {
            return _isRetryOnDisconnect;
        }

        /// <summary>
        /// Set the default max depth to be used on all MessageStream results.
        /// </summary>
        /// <param name="defaultMaxDepth">The new default value for max depth</param>
        public void setDefaultMaxDepth(int defaultMaxDepth)
        {
            _defaultMaxDepth = defaultMaxDepth;
        }

        /// <summary>
        /// Return the current default max depth value set on MessageStream results.
        /// </summary>
        /// <returns>The current default max depth value set on MessageStream results.</returns>
        public int getDefaultMaxDepth()
        {
            return _defaultMaxDepth;
        }

        /// <summary>
        /// Send a Message to AMPS and register the messageHandler for any messages resulting from
        /// command execution.
        /// </summary>
        /// <param name="messageHandler">The message handler that'll receive messages for this command</param>
        /// <param name="message"> The message to send</param>
        /// <param name="timeout">The number of milliseconds to wait for command acknowledgment, where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        /// <exception cref="SubscriptionAlreadyExistsException">The specified subscription ID already exists.</exception>
        public CommandId send(MessageHandler messageHandler, Message message, long timeout)
        {
            CommandId id = null;
            String strId = message.getCommandId();
            if (!String.IsNullOrEmpty(strId))
            {
                id = new CommandId(strId);
            }
            Message.AckTypes requestedAcks = message.getAckTypeOutgoing();
            Message.AckTypes systemAddedAcks = 0;
            bool isSubscribe = false;
            switch(message.getCommand())
            {
                // Subscriptions need to have a sub id
            case Message.Commands.Subscribe:
            case Message.Commands.DeltaSubscribe:
                if (!message.getBookmarkRaw().isNull() &&
                    this.bookmarkStore != null)
                {
                    systemAddedAcks |= Message.AckTypes.Persisted;
                }
                // fall through
                goto case Message.Commands.SOWAndSubscribe;
            case Message.Commands.SOWAndSubscribe:
            case Message.Commands.SOWAndDeltaSubscribe:
                if (id == null)
                {
                    id = CommandId.nextIdentifier();
                    message.setCommandId(id);
                }
                if (String.IsNullOrEmpty(message.getSubId()))
                {
                    message.setSubId(id);
                }
                // SOW queries just need a QueryId.
                // fall through.
                isSubscribe = true;
                goto case Message.Commands.SOW;
            case Message.Commands.SOW:
                if(id == null)
                {
                    id = CommandId.nextIdentifier();
                    message.setCommandId(id);
                }
                if (String.IsNullOrEmpty(message.getQueryId()))
                {
                    message.setQueryId(id);
                }
                systemAddedAcks |= Message.AckTypes.Processed;
                // Don't add completed ack as a system ack for a SOW command
                // processed using a MessageStream handler because the message
                // stream will handle removing the route on group_end if a
                // completed or stats ack isn't explicitly requested by the user.
                // But an async message handler will still need automatic route
                // removal handled by the router when a SOW command is completed.
                if (!isSubscribe && messageHandler != null
                        && !(messageHandler is MessageStream)) {
                    systemAddedAcks |= Message.AckTypes.Completed;
                }
                message.setAckType(requestedAcks|systemAddedAcks);
                if (messageHandler != null)
                {
                    strId = message.getSubId();
                    CommandId subCmdId = String.IsNullOrEmpty(strId) ?
                                                null : new CommandId(strId);
                    strId = message.getQueryId();
                    CommandId qid =  String.IsNullOrEmpty(strId) ?
                                                null : new CommandId(strId);
                    // This will never return false.
                    _registerHandler(qid, subCmdId, id, messageHandler,
                                     requestedAcks, systemAddedAcks,
                                     isSubscribe);
                }
                lock (internalLock)
                {
                    try
                    {
                        // We aren't adding to subscription manager, so this isn't an HA subscribe.
                        this.syncAckProcessing(id, message, timeout,0,false);
                    }
                    catch (TimedOutException)
                    {
                        _routes.removeRoute(id);
                        throw;
                    }
                }
                break;
                // These are valid commands that are used as-is
            case Message.Commands.StartTimer:
                if (this.serverVersion.getVersion() >= 500030000200000UL)
                {
                    throw new CommandException("The start_timer command has been deprecated starting in AMPS server 5.3.2.0");
                }
                goto case Message.Commands.SOWDelete;
            case Message.Commands.StopTimer:
                if (this.serverVersion.getVersion() >= 500030000200000UL)
                {
                    throw new CommandException("The stop_timer command has been deprecated starting in AMPS server 5.3.2.0");
                }
                goto case Message.Commands.SOWDelete;
            case Message.Commands.Unsubscribe:
            case Message.Commands.Heartbeat:
            case Message.Commands.Logon:
            case Message.Commands.DeltaPublish:
            case Message.Commands.Publish:
            case Message.Commands.SOWDelete:
                // if an ack is requested, it'll need a command ID.
                if(requestedAcks != Message.AckTypes.None)
                {
                    if (id == null)
                    {
                        id = CommandId.nextIdentifier();
                        message.setCommandId(id);
                    }
                    if(messageHandler != null)
                    {
                        _routes.addRoute(id, messageHandler, requestedAcks,
                                         Message.AckTypes.None, false);
                    }
                }
                this.send(message);
                break;
                // These are things that shouldn't be sent (not meaningful)
            case Message.Commands.GroupBegin:
            case Message.Commands.GroupEnd:
            case Message.Commands.OOF:
            case Message.Commands.Ack:
            case Message.Commands.Unknown:
            default:
                throw new CommandException("Command type can not be sent directly to AMPS");
            }
            return id;
        }

        /// <summary>
        /// Send a Message to AMPS and register the messageHandler for any messages resulting from
        /// command execution.
        /// </summary>
        /// <param name="messageHandler">The message handler that'll receive messages for this command</param>
        /// <param name="message"> The message to send</param>
        /// <param name="timeout">The number of milliseconds to wait for command acknowledgment, where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        /// <exception cref="SubscriptionAlreadyExistsException">The specified subscription ID already exists.</exception>
        public CommandId send(Action<Message> messageHandler, Message message, long timeout)
        {
            return send(new ActionMessageHandler(messageHandler), message, timeout);
        }


        /// <summary>
        /// Publish a message to an AMPS topic. If the client has a publish store set, then the client will store
        /// the message before forwarding the message to AMPS. This method does not wait for a response from the AMPS
        /// server. To detect failure, install a failed write handler. If a DisconnectException occurs, the message is still
        /// stored in the publish store.
        /// <pre>
        /// <code>
        /// namespace AMPSPublishExample
        /// {
        ///     class AMPSPublishExample
        ///     {
        ///         private static string uri_ = "tcp://127.0.0.1:9027/amps/json";
        ///
        ///         static void Main(string[] args)
        ///         {
        ///             Client client = new Client("SOWConsolePublisher");
        ///             try
        ///             {
        ///                 System.Console.WriteLine("... starting ...");
        ///                 // connect and logon
        ///                 client.connect(uri_);
        ///                 client.logon();
        ///
        ///                 // publish a message with expiration set
        ///                 client.publish("messages-sow",
        ///                          "{ \"messageNumber\":50000, \"message\":\"Here and then gone...\"}",
        ///                           3);
        ///             }
        ///             catch (AMPSException e)
        ///             {
        ///                 System.Console.WriteLine(e.Message);
        ///                 System.Console.WriteLine(e.StackTrace);
        ///             }
        ///         }
        ///     }
        /// }
        /// </code>
        /// </pre>
        /// </summary>
        /// <param name="topic">array containing the topic to publish the data to</param>
        /// <param name="topicOffset">offset into topic array where the topic name begins</param>
        /// <param name="topicLength">length of topic in <paramref name="topic"/> array</param>
        /// <param name="data">array containing the data to publish</param>
        /// <param name="dataOffset">offset into the data array where the data begins</param>
        /// <param name="dataLength">length of the data in <paramref name="data"/> array
        /// <param name="expiration">the number of seconds until the message expires</param>
        /// <returns>The sequence number for the message when a persisten store is configured for the client, 0 if no store is
        /// configured.</returns>
        /// <exception cref="DisconnectedException">The client is not connected; the program needs to call connect()</exception>
        /// <exception cref="StoreException">An error occurred writing to the local HA store</exception>
        public ulong publish(byte[] topic, int topicOffset, int topicLength,
                            byte[] data, int dataOffset, int dataLength, int expiration)
        {
            lock (internalLock)
            {
                _publishCommand.Value.reset(Message.Commands.Publish)
                              .setTopic(topic, topicOffset, topicLength)
                              .setData(data, dataOffset, dataLength)
                              .setExpiration(expiration);
                executeAsyncNoLock(_publishCommand.Value, null);
                // The next line is "safe" even though command may have been
                //  reused. See the comment in publish(byte[]...)
                return _publishCommand.Value.getClientSequenceNumber();
            }
        }

        /// <summary>
        /// Publish a message to an AMPS topic. If the client has a publish store set, then the client will store the message before
        /// forwarding the message to AMPS. This method does not wait for a response from the AMPS server. To detect failure, install
        /// a failed write handler. If a DisconnectException occurs, the message is still stored in the publish store.
        /// </summary>
        /// <param name="topic">array containing the topic to publish the data to</param>
        /// <param name="topicOffset">offset into topic array where the topic name begins</param>
        /// <param name="topicLength">length of topic in <paramref name="topic"/> array
        /// <param name="data">array containing the data to publish</param>
        /// <param name="dataOffset">offset into the data array where the data begins</param>
        /// <param name="dataLength">length of the data in <paramref name="data"/> array</param>
        /// <returns>The sequence number for the message when a persisten store is configured for the client, 0 if no store is
        /// configured.</returns>
        /// <exception cref="DisconnectedException">The client is not connected; the program needs to call connect()</exception>
        /// <exception cref="StoreException">An error occurred writing to the local HA store</exception>
        public ulong publish(byte[] topic, int topicOffset, int topicLength,
                            byte[] data, int dataOffset, int dataLength)
        {
            lock (internalLock)
            {
                _publishCommand.Value.reset(Message.Commands.Publish)
                        .setTopic(topic, topicOffset, topicLength)
                        .setData(data, dataOffset, dataLength);
                executeAsyncNoLock(_publishCommand.Value, null);
                // Generally, the shared "command" is no longer safe after a call
                // to executeAsyncNoLock because another thread may have used it.
                // getClientSequenceNumber is an exception to this rule, because
                // we explicitly set it with the correct value after we've
                // re-acquired the lock in executeAsyncNoLock.
                return _publishCommand.Value.getClientSequenceNumber();
            }
        }

        /// <summary>
        /// Publish a message to an AMPS topic. If the client has a publish store set, then the client will store the message
        /// before forwarding the message to AMPS. This method does not wait for a response from the AMPS server. To detect
        /// failure, install a failed write handler. If a DisconnectException occurs, the message is still stored in the publish store.
        /// </summary>
        /// <param name="topic">Topic to publish the data to</param>
        /// <param name="data">Data to publish to the topic</param>
        /// <returns>The sequence number for the message when a persisten store is configured for the client, 0 if no store is configured.</returns>
        /// <exception cref="DisconnectedException">The client is disconnected; the program needs to call connect()</exception>
        public ulong publish(
                String topic,
                String data)
        {
            lock (internalLock)
            {
                _publishCommand.Value.reset(Message.Commands.Publish).setTopic(topic)
                              .setData(data);
                executeAsyncNoLock(_publishCommand.Value, null);
                // The next line is "safe" even though command may have been
                //  reused. See the comment in publish(byte[]...)
                return _publishCommand.Value.getClientSequenceNumber();
            }
        }

        /// <summary>
        /// Publish a message to an AMPS topic. If the client has a publish store set, then the client will store the message before
        /// forwarding the message to AMPS. This method does not wait for a response from the AMPS server. To detect failure, install a
        /// failed write handler. If a DisconnectException occurs, the message is still stored in the publish store.
        /// </summary>
        /// <param name="topic">Topic to publish the data to</param>
        /// <param name="data">Data to publish to the topic</param>
        /// <param name="expiration">The number of seconds until the message expires</param>
        /// <returns>The sequence number for the message when a persisten store is configured for the client, 0 if no store is configured.</returns>
        /// <exception cref="DisconnectedException">The client is not connected; the program needs to call connect()</exception>
        public ulong publish(
                String topic,
                String data, int expiration)
        {
            lock (internalLock)
            {
                _publishCommand.Value.reset(Message.Commands.Publish).setTopic(topic)
                    .setData(data).setExpiration(expiration);
                executeAsyncNoLock(_publishCommand.Value, null);
                // The next line is "safe" even though command may have been
                //  reused. See the comment in publish(byte[]...)
                return _publishCommand.Value.getClientSequenceNumber();
            }
        }

        /// <summary>
        /// Delta publish a message to an AMPS topic. If the client has a publish store set,
        /// then the client will store the message before forwarding the message to AMPS.
        /// This method does not wait for a response from the AMPS server. To detect failure,
        /// install a failed write handler. If a DisconnectException occurs, the message is still
        ///  stored in the publish store.
        /// </summary>
        /// <param name="topic">array containing the topic to publish the data to</param>
        /// <param name="topicOffset">offset into topic array where the topic name begins</param>
        /// <param name="topicLength">length of topic in <paramref name="topic"/> array</param>
        /// <param name="data">array containing the data to publish</param>
        /// <param name="dataOffset">offset into the data array where the data begins</param>
        /// <param name="dataLength">length of the data in <paramref name="data"/> array</param>
        /// <returns>The sequence number for the message when a persisten store is configured for the client, 0 if no store is configured.</returns>
        public ulong deltaPublish(
            byte[] topic, int topicOffset, int topicLength,
            byte[] data, int dataOffset, int dataLength)
        {
            lock (internalLock)
            {
                _publishCommand.Value.reset(Message.Commands.DeltaPublish)
                              .setTopic(topic, topicOffset, topicLength)
                              .setData(data, dataOffset, dataLength);
                executeAsyncNoLock(_publishCommand.Value, null);
                return _publishCommand.Value.getClientSequenceNumber();
            }
        }

        /// <summary>
        /// Delta publish a message to an AMPS topic. If the client has a publish store set,
        /// then the client will store the message before forwarding the message to AMPS.
        /// This method does not wait for a response from the AMPS server. To detect failure,
        /// install a failed write handler. If a DisconnectException occurs, the message is still
        /// stored in the publish store.
        /// </summary>
        /// <param name="topic">array containing the topic to publish the data to</param>
        /// <param name="topicOffset">offset into topic array where the topic name begins</param>
        /// <param name="topicLength">length of topic in <paramref name="topic"/> array</param>
        /// <param name="data">array containing the data to publish</param>
        /// <param name="dataOffset">offset into the data array where the data begins</param>
        /// <param name="expiration">The number of seconds until the message expires</param>
        /// <param name="dataLength">length of the data in <paramref name="data"/> array</param>
        /// <returns>The sequence number for the message when a persisten store is configured for the client, 0 if no store is configured.</returns>
        public ulong deltaPublish(
            byte[] topic, int topicOffset, int topicLength,
            byte[] data, int dataOffset, int dataLength, int expiration)
        {
            lock (internalLock)
            {
                _publishCommand.Value.reset(Message.Commands.DeltaPublish)
                              .setTopic(topic, topicOffset, topicLength)
                              .setData(data, dataOffset, dataLength)
                              .setExpiration(expiration);
                executeAsyncNoLock(_publishCommand.Value, null);
                return _publishCommand.Value.getClientSequenceNumber();
            }
        }

        /// <summary>
        /// Delta publish a message to an AMPS topic. If the client has a publish store set,
        /// then the client will store the message before forwarding the message to AMPS.
        /// This method does not wait for a response from the AMPS server. To detect failure,
        /// install a failed write handler. If a DisconnectException occurs, the message is still
        /// stored in the publish store.
        /// </summary>
        /// <param name="topic">Topic to publish the data to</param>
        /// <param name="data">Data to publish to the topic</param>
        /// <returns>The sequence number for the message when a persisten store is configured for the client, 0 if no store is configured.</returns>
        /// <exception cref="DisconnectedException">The client was disconnected at time of publish</exception>
        public ulong deltaPublish(
                String topic,
                String data)
        {
            lock (internalLock)
            {
                _publishCommand.Value.reset(Message.Commands.DeltaPublish)
                              .setTopic(topic)
                              .setData(data);
                executeAsyncNoLock(_publishCommand.Value, null);
                return _publishCommand.Value.getClientSequenceNumber();
            }
        }

        /// <summary>
        /// Delta publish a message to an AMPS topic. If the client has a publish store set,
        /// then the client will store the message before forwarding the message to AMPS.
        /// This method does not wait for a response from the AMPS server. To detect failure,
        /// install a failed write handler. If a DisconnectException occurs, the message is still
        /// stored in the publish store.
        /// </summary>
        /// <param name="topic">Topic to publish the data to</param>
        /// <param name="data">Data to publish to the topic</param>
        /// <param name="expiration">The number of seconds until the message expires</param>
        /// <returns>The sequence number for the message when a persisten store is configured for the client, 0 if no store is configured.</returns>
        /// <exception cref="DisconnectedException">The client was disconnected at time of publish</exception>
        public ulong deltaPublish(
                String topic,
                String data, int expiration)
        {
            lock (internalLock)
            {
                _publishCommand.Value.reset(Message.Commands.DeltaPublish)
                              .setTopic(topic)
                              .setData(data).setExpiration(expiration);
                executeAsyncNoLock(_publishCommand.Value, null);
                return _publishCommand.Value.getClientSequenceNumber();
            }
        }

        /// <summary>
        /// Logs into AMPS with the parameters provided in the connect method and the logon correlation data if set.
        /// </summary>
        /// <param name="timeout">The number of milliseconds to wait for the command to execute, where 0 indicates no timeout</param>
        /// <returns>The command identifier</returns>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        public virtual CommandId logon(long timeout)
        {
            return logon(timeout, new DefaultAuthenticator(), null);
        }

        /// <summary>
        /// Logs into AMPS with the parameters provided in the connect method and the logon correlation data if set.
        /// </summary>
        /// <param name="options">The options string to pass on to the server during logon</param>
        /// <returns>The command identifier</returns>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        public virtual CommandId logon(string options)
        {
            return logon(0, new DefaultAuthenticator(), options);
        }

        /// <summary>
        /// Logs into AMPS with the parameters provided in the connect method and the logon correlation data if set.
        /// </summary>
        /// <param name="timeout">The number of milliseconds to wait for the command to execute, where 0 indicates no timeout</param>
        /// <param name="authenticator">The authenticator to use for logging in</param>
        /// <returns>The command identifier</returns>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        public virtual CommandId logon(long timeout, Authenticator authenticator)
        {
            return logon(timeout, authenticator, null);
        }

        /// <summary>
        /// Logs into AMPS with the parameters provided in the connect method and the logon correlation data if set.
        /// </summary>
        /// <param name="timeout">The number of milliseconds to wait for the command to execute, where 0 indicates no timeout</param>
        /// <param name="options">The options string to pass on to the server during logon</param>
        /// <returns>The command identifier</returns>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        public virtual CommandId logon(long timeout, string options)
        {
            return logon(timeout, new DefaultAuthenticator(), options);
        }

        /// <summary>
        /// Logs into AMPS with the parameters provided in the connect method and the logon correlation data if set.
        /// </summary>
        /// <param name="timeout">The number of milliseconds to wait for the command to execute, where 0 indicates no timeout</param>
        /// <param name="authenticator">The authenticator to use for logging in</param>
        /// <param name="options">The options string to pass on to the server during logon</param>
        /// <returns>The command identifier</returns>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        public virtual CommandId logon(long timeout, Authenticator authenticator, string options)
        {
            if (uri == null)
                throw new DisconnectedException("Not connected");

            if (authenticator == null)
            {
                authenticator = new DefaultAuthenticator();
            }

            CommandId id = CommandId.nextIdentifier();
            lock (internalLock)
            {
                try
                {
                    this._logonInProgress = true;
                    this.message.reset();
                    this.message.setCommand(Message.Commands.Logon);
                    this.message.setCommandId(id);
                    this.message.setClientName(this.name);
                    if (!string.IsNullOrEmpty(options))
                    {
                        this.message.setOptions(options);
                    }
                    if (!string.IsNullOrEmpty(this.uri.UserInfo))
                    {
                        int upSep = this.uri.UserInfo.IndexOf(':');
                        if (upSep < 0)
                        {
                            message.UserId = Uri.UnescapeDataString(uri.UserInfo);
                        }
                        else
                        {
                            message.UserId = Uri.UnescapeDataString(uri.UserInfo.Substring(0, upSep));
                            if (uri.UserInfo.Length > upSep + 1)
                                message.Password = Uri.UnescapeDataString(uri.UserInfo.Substring(upSep + 1));
                        }
                        message.Password = authenticator.authenticate(message.UserId, message.Password);
                        username = message.UserId;
                    }

                    String[] sPath = uri.AbsolutePath.Substring(1, uri.AbsolutePath.Length - 1).Split('/');
                    if (sPath.Length > 1)
                    {
                        this.message.setMessageType(sPath[sPath.Length-1]);
                    }
                    // Versions before 3.0 did not support the Processed ack
                    //   for a logon command.
                    if (this.version >= Version.AMPS_3)
                    {
                        this.message.setAckType(Message.AckTypes.Processed);
                        this.message.setVersion(Client.getVersion() + ":c#");
                        var properties = new URIProperties(uri);
                        if(properties.ContainsKey("pretty"))
                        {
                            String pretty = properties["pretty"];
                            if(pretty == "true" || pretty == "True" || pretty=="t" || pretty=="T" || pretty == "1")
                            {
                                message.setOptions("pretty");
                            }
                        }
                        if (logonCorrelationData != null) this.message.setCorrelationId(logonCorrelationData);
                        AckResponse response = null;
                        while (true)
                        {
                            response = this.syncAckProcessing(id, this.message, timeout);
                            if (response.State == Message.State.Retry)
                            {
                                this.message.Password = authenticator.retry(response.Username, response.Password);
                                this.username = response.Username;
                                this.message.setUserId(this.username);
                            }
                            else
                            {
                                break;
                            }
                        }
                        if(ConnectionStateChanged!=null)
                            ConnectionStateChanged(this, new ConnectionStateChangedEventArgs(ConnectionStateChangedEventArgs.ConnectionState.LoggedOn));
                        this.serverVersion = response.ServerVersion;
                        this.nameHash = response.Bookmark.Substring(0,
                                                         response.Bookmark.IndexOf('|'));
                        if (this.bookmarkStore != null)
                            this.bookmarkStore.setServerVersion(this.serverVersion.getOldStyleVersion());

                        authenticator.completed(response.Username, response.Password, response.Reason);
                        // Send configured heartbeat to the server.
                        _sendHeartbeat();
                    }
                    else
                    {
                        sendInternal(this.message);
                        if(ConnectionStateChanged!=null)
                            ConnectionStateChanged(this, new ConnectionStateChangedEventArgs(ConnectionStateChangedEventArgs.ConnectionState.LoggedOn));
                    }
                    if (this.publishStore != null)
                    {
                        this._logonInProgress = false;
                        this.publishStore.replay(this.replayer);
                        Monitor.PulseAll(internalLock);
                        if(ConnectionStateChanged!=null)
                            ConnectionStateChanged(this, new ConnectionStateChangedEventArgs(ConnectionStateChangedEventArgs.ConnectionState.PublishReplayed));
                    }
                }
                catch (StoreException storeException)
                {
                    throw new ConnectionException($"A local store exception occurred while logging on: {storeException.Message}", storeException);
                }
                finally
                {
                    this._logonInProgress = false;
                }
                return id;
            }
        }

        /// <summary>
        /// Logs into AMPS with the parameters provided in the connect method.
        /// </summary>
        /// <returns>The command identifier</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        public virtual CommandId logon()
        {
            return logon(0);
        }

        /// <summary>
        /// Places a bookmark subscription with AMPS.
        /// Starts replay at the most recent message reported by the BookmarkStore.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">Only messages with data that match this filter will be received</param>
        /// <param name="subId"> You may optionally provide a subscription ID
        /// to ease recovery scenarios, instead of having the system automatically
        /// generate one for you. When used with the 'replace' option, this
        /// is the subscription to be replaced. With a bookmark store,
        /// this is the subscription ID used for recovery. So, when
        /// using a persistent bookmark store, provide an explicit
        /// subscription ID that is consistent across application restarts</param>
        /// <param name="bookmark">A bookmark identifer, one of the constants from <see cref="Client.Bookmarks"/>, or a timestamp of the form <i>YYYYmmddTHHMMSS</i>,
        /// as described in the AMPS User's Guide</param>
        /// <param name="options">A AMPS.Client.Message.Options value indicating desired options for this subscription. Use Message.Options.None if no options are desired</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The universally unique <typeparamref name="CommandId"/> assigned to this command</returns>
        /// <exception cref="SubscriptionAlreadyExistsException">A subscription with this <paramref name="subId"/> is already registered</exception>
        /// <exception cref="BadFilterException">The provided filter is invalid</exception>
        /// <exception cref="BadRegexTopicException">The topic specified was an invalid regular expression</exception>
        /// <exception cref="TimedOutException">The operation took longer than the timeout to execute</exception>
        /// <exception cref="DisconnectedException">The client wasn't connected when the operation was executed</exception>
        public CommandId bookmarkSubscribe(
            MessageHandler messageHandler,
            String topic,
            String filter = null,
            CommandId subId = null,
            String bookmark = null,
            string options = null,
            long timeout = 0)
        {
            lock (internalLock)
            {
                _command.reset(Message.Commands.Subscribe)
                              .setTopic(topic).setFilter(filter)
                              .setBookmark(bookmark)
                              .setOptions(options).setTimeout(timeout);
                if (subId != null) _command.setSubId(subId);
                return executeAsyncNoLock(_command, messageHandler);
            }
        }

        /// <summary>
        /// Places a subscription with AMPS.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        /// <exception cref="SubscriptionAlreadyExistsException">The specified subscription ID already exists.</exception>
        public CommandId subscribe(
              MessageHandler messageHandler,
              String topic,
              String filter,
              long timeout)
        {
            return subscribe(messageHandler, topic, filter, Message.Options.None, timeout);
        }


        /// <summary>
        /// Places a subscription with AMPS.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        /// <exception cref="SubscriptionAlreadyExistsException">The specified subscription ID already exists.</exception>
        public CommandId subscribe(
              Action<Message> messageHandler,
              String topic,
              String filter,
              long timeout)
        {
            return subscribe(new ActionMessageHandler(messageHandler), topic, filter, Message.Options.None, timeout);
        }

        /// <summary>Places a subscription with AMPS.</summary>
        ///
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="options">A value from Message.Options indicating additional processing options</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <param name="subId">The subscription id to use for the subscription</param>
        /// <returns>A MessageStream to iterate over</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        /// <exception cref="SubscriptionAlreadyExistsException">The specified subscription ID already exists.</exception>
        public MessageStream subscribe(string topic, string filter=null, string options=null, long timeout=0, string subid=null)
        {
            lock(internalLock)
            {
                return _execute(_command.reset(Message.Commands.Subscribe)
                                              .setTopic(topic)
                                              .setFilter(filter)
                                              .setOptions(options)
                                              .setTimeout(timeout)
                                              .setSubId(subid), false);
            }
        }


        /// <summary>Places a subscription with AMPS.</summary>
        ///
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="options">A value from Message.Options indicating additional processing options</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <param name="subId">The subscription id to use for the subscription</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        /// <exception cref="SubscriptionAlreadyExistsException">The specified subscription ID already exists.</exception>
        public CommandId subscribe(MessageHandler messageHandler,
                String topic,
                String filter = null,
                string options = null,
                long timeout = 0,
                String subId=null)
        {
            lock (internalLock)
            {
                return executeAsyncNoLock(_command
                                            .reset(Message.Commands.Subscribe)
                                            .setTopic(topic).setFilter(filter)
                                            .setOptions(options)
                                            .setTimeout(timeout)
                                            .setSubId(subId), messageHandler);
            }
        }

        /// <summary>
        /// Places a subscription with AMPS.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="options">A value from <typeparamref name="Message.Options"/> indicating additional processing options</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId subscribe(
              Action<Message> messageHandler,
              String topic,
              String filter = null,
              String options = null,
              long timeout = 0,
              string subid = null)
        {
            return subscribe(new ActionMessageHandler(messageHandler), topic, filter, options, timeout, subid);
        }

        /// <summary>
        /// Places a subscription with AMPS.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        /// <exception cref="SubscriptionAlreadyExistsException">The specified subscription ID already exists.</exception>
        public CommandId subscribe(
              MessageHandler messageHandler,
              String topic,
              long timeout)
        {
            return subscribe(messageHandler, topic, null, timeout);
        }


        /// <summary>
        /// Places a subscription with AMPS.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        /// <exception cref="SubscriptionAlreadyExistsException">The specified subscription ID already exists.</exception>
        public CommandId subscribe(
              Action<Message> messageHandler,
              String topic,
              long timeout)
        {
            return subscribe(new ActionMessageHandler(messageHandler), topic, null, timeout);
        }

        /// <summary>
        /// Places a delta subscription with AMPS.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        /// <exception cref="SubscriptionAlreadyExistsException">The specified subscription ID already exists.</exception>
        public CommandId deltaSubscribe(
              MessageHandler messageHandler,
              String topic,
              String filter=null,
              String options=null,
              long timeout=0,
              String subId=null)
        {
            lock (internalLock)
            {
                _command.reset(Message.Commands.DeltaSubscribe)
                              .setTopic(topic)
                              .setFilter(filter)
                              .setOptions(options)
                              .setTimeout(timeout);
                if (subId != null) _command.setSubId(subId);
                return executeAsyncNoLock(_command, messageHandler);
            }
        }

        /// <summary>
        /// Places a delta subscription with AMPS.
        /// </summary>
        /// <param name="messageHandler">The lambda function to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        /// <exception cref="SubscriptionAlreadyExistsException">The specified subscription ID already exists.</exception>
        public CommandId deltaSubscribe(
              Action<Message> messageHandler,
              String topic,
              String filter=null,
              String options=null,
              long timeout=0,
              String subId=null)
        {
            return deltaSubscribe(new ActionMessageHandler(messageHandler), topic, filter, options, timeout, subId);
        }


        /// <summary>
        /// Remove a subscription from AMPS.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier to remove</param>
        /// <exception cref="DisconnectedException">The client was disconnected at the time of execution.</exception>
        public void unsubscribe(CommandId subscriptionId)
        {
            if (subscriptionId != null)
            {
                this.subscriptionManager.unsubscribe(subscriptionId);
            }
            lock(internalLock)
            {
                unsubscribeInternal(subscriptionId);
            }
        }

        /// <summary>
        /// Remove all of the client's subscriptions from AMPS.
        /// </summary>
        /// <exception cref="DisconnectedException">The client was disconnected at the time of execution</exception>
        public void unsubscribe()
        {
            this.subscriptionManager.clear();
            CommandId id = CommandId.nextIdentifier();
            lock(internalLock)
            {
                _routes.unsubscribeAll();
                if (this.transport != null)
                {
                    this.message.reset();
                    this.message.setCommand(Message.Commands.Unsubscribe);
                    this.message.setCommandId(id);
                    this.message.setSubId("all");
                    this.transport.sendWithoutRetry(this.message);
                }
                else
                {
                    throw new DisconnectedException("There is no current connection.");
                }
            }
        }

        // Internal method to handle the removal of a specific subscription from AMPS.
        // param name="subscriptionId": The subscription identifier to remove</param>
        // This method is used internally by the "unsubscribe" method to handle the removal
        // of a specific subscription. It unsubscribes the subscription from the subscription manager,
        // removes the subscription route, and sends an unsubscribe message to the AMPS server if there
        // is an active connection.
        // exception ="DisconnectedException": The client was disconnected at the time of execution.
        public void unsubscribeInternal(CommandId subscriptionId)
        {
            if (subscriptionId != null)
            {
                _routes.removeRoute(subscriptionId);
                if (this.transport != null)
                {
                    this.message.reset();
                    this.message.setCommand(Message.Commands.Unsubscribe);
                    this.message.setCommandId(CommandId.nextIdentifier());
                    this.message.setSubId(subscriptionId);
                    this.transport.sendWithoutRetry(this.message);
                }
                else
                {
                    throw new DisconnectedException("There is no current connection.");
                }
            }
        }

        /// <summary>
        /// Executes a SOW query.
        /// </summary>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="orderBy">The ordering property</param>
        /// <param name="bookmark">A bookmark identifer, one of the constants from <see cref="Client.Bookmarks"/>, or a timestamp of the form <i>YYYYmmddTHHMMSS</i>,
        /// as described in the AMPS User's Guide. This the point at which AMPS executes the historical query</param>
        /// <param name="options">A AMPS.Client.Message.Options value indicating desired options for this query.  Use Message.Options.None if no options are desired</param>
        /// <param name="batchSize">The batching parameter to use for the results</param>
        /// <param name="topN">The maximum number of records the server will return (default is all that match)</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>A MessageStream to iterate over</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public MessageStream sow(String topic,
            String filter=null,
            String orderBy=null,
            String bookmark=null,
            String options=null,
            int batchSize=10,
            int topN=-1,
            long timeout=0)
        {
            lock (internalLock)
            {
                return _execute(_command.reset(Message.Commands.SOW)
                                 .setTopic(topic).setFilter(filter)
                                 .setOrderBy(orderBy).setBookmark(bookmark)
                                 .setOptions(options).setBatchSize(batchSize)
                                 .setTopN(topN).setTimeout(timeout), false);
            }
        }

        /// <summary>
        /// Executes a SOW query.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="orderBy">The ordering property</param>
        /// <param name="bookmark">A bookmark identifer, one of the constants from <see cref="Client.Bookmarks"/>, or a timestamp of the form <i>YYYYmmddTHHMMSS</i>,
        /// as described in the AMPS User's Guide. This the point at which AMPS executes the historical query</param>
        /// <param name="options">A AMPS.Client.Message.Options} value indicating desired options for this query.  Use Message.Options.None if no options are desired</param>
        /// <param name="batchSize">The batching parameter to use for the results</param>
        /// <param name="topN">The maximum number of records the server will return (default is all that match)</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sow(MessageHandler messageHandler,
              String topic,
              String filter=null,
              String orderBy=null,
              String bookmark=null,
              String options=null,
              int batchSize=10,
              int topN=-1,
              long timeout=0)
        {
            lock (internalLock)
            {
                return executeAsyncNoLock(_command.reset(Message.Commands.SOW)
                                                  .setTopic(topic)
                                                  .setFilter(filter)
                                                  .setOrderBy(orderBy)
                                                  .setBookmark(bookmark)
                                                  .setOptions(options)
                                                  .setBatchSize(batchSize)
                                                  .setTopN(topN)
                                                  .setTimeout(timeout),
                                                              messageHandler);
            }
        }

        /// <summary>
        /// Executes a SOW query.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="orderBy">The ordering property</param>
        /// <param name="bookmark">A bookmark identifer, one of the constants from <see cref="Client.Bookmarks"/>, or a timestamp of the form <i>YYYYmmddTHHMMSS</i>,
        /// as described in the AMPS User's Guide. This the point at which AMPS executes the historical query</param>
        /// <param name="options">A AMPS.Client.Message.Options value indicating desired options for this query.  Use Message.Options.None if no options are desired</param>
        /// <param name="batchSize">The batching parameter to use for the results</param>
        /// <param name="topN">The maximum number of records the server will return (default is all that match)</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sow(Action<Message> messageHandler,
          String topic,
          String filter=null,
          String orderBy=null,
          String bookmark=null,
          String options=null,
          int batchSize=10,
          int topN=-1,
          long timeout=0)
        {
            return sow(new ActionMessageHandler(messageHandler), topic, filter, orderBy, bookmark, options, batchSize, topN, timeout);
        }

        /// <summary>
        /// Executes a SOW query.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="batchSize">The batching parameter to use for the results</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sow(MessageHandler messageHandler,
              String topic,
              String filter,
              int batchSize,
              long timeout)
        {
            return sow(messageHandler, topic, filter, null, null,null,batchSize, -1, timeout);
        }

        /// <summary>
        /// Executes a SOW query.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="batchSize">The batching parameter to use for the results</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns></returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sow(MessageHandler messageHandler,
              String topic,
              int batchSize,
              long timeout)
        {
            return sow(messageHandler, topic, null, batchSize, timeout);
        }

        /// <summary>
        /// Executes a SOW query.
        /// </summary>
        /// <param name="messageHandler">The function to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="batchSize">The batching parameter to use for the results</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns></returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sow(Action<Message> messageHandler,
              String topic,
              int batchSize,
              long timeout)
        {
            return sow(new ActionMessageHandler(messageHandler), topic, null, batchSize, timeout);
        }

        /// <summary>
        /// Executes a SOW query.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns></returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sow(MessageHandler messageHandler,
              String topic,
              long timeout)
        {
            // Default batchsize to 10 for good default performance
            return sow(messageHandler, topic, null, 10, timeout);
        }

        /// <summary>
        /// Executes a SOW query.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns></returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sow(Action<Message> messageHandler,
              String topic,
              long timeout)
        {
            // Default batchsize to 10 for good default performance
            return sow(new ActionMessageHandler(messageHandler), topic, null, 10, timeout);
        }

        /// <summary>
        /// Executes a SOW query and places a subscription.
        /// </summary>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="orderBy">The ordering property</param>
        /// <param name="bookmark">A bookmark identifer, one of the constants from <see cref="Client.Bookmarks"/>, or a timestamp of the form <i>YYYYmmddTHHMMSS</i>,
        /// as described in the AMPS User's Guide. This the point at which AMPS executes the historical query</param>
        /// <param name="options">A AMPS.Client.Message.Options value indicating desired options for this query. Use Message.Options.None if no options are desired</param>
        /// <param name="batchSize">The batching parameter to use for the results</param>
        /// <param name="topN">The maximum number of records the server will return (default is all that match)<i/param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>A MessageStream to iterate over.</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public MessageStream sowAndSubscribe(String topic,
              String filter = null,
              String orderBy = null,
              String bookmark = null,
              String options = null,
              int batchSize = 10,
              int topN = -1,
              long timeout = 0)
        {
            lock (internalLock)
            {
                return _execute(_command.reset(Message.Commands.SOWAndSubscribe)
                                        .setTopic(topic).setFilter(filter)
                                        .setOrderBy(orderBy)
                                        .setBookmark(bookmark)
                                        .setBatchSize(batchSize)
                                        .setTopN(topN)
                                        .setTimeout(timeout)
                                        .setOptions(options), false);
            }
        }
        /// <summary>
        /// Executes a SOW query and places a subscription.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="orderBy">The ordering property</param>
        /// <param name="bookmark">A bookmark identifer, one of the constants from <see cref="Client.Bookmarks"/>, or a timestamp of the form <i>YYYYmmddTHHMMSS</i>,
        /// as described in the AMPS User's Guide. This the point at which AMPS executes the historical query</param>
        /// <param name="options">A AMPS.Client.Message.Options value indicating desired options for this query. Use Message.Options.
        /// None if no options are desired</param>
        /// <param name="batchSize">The batching parameter to use for the results</param>
        /// <param name="topN">The maximum number of records the server will return (default is all that match)</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndSubscribe(
              MessageHandler messageHandler,
              String topic,
              String filter = null,
              String orderBy = null,
              String bookmark = null,
              String options = null,
              int batchSize = 10,
              int topN = -1,
              long timeout = 0)
        {
            lock (internalLock)
            {
                return executeAsyncNoLock(_command
                                      .reset(Message.Commands.SOWAndSubscribe)
                                      .setTopic(topic).setFilter(filter)
                                      .setOrderBy(orderBy)
                                      .setBookmark(bookmark)
                                      .setBatchSize(batchSize)
                                      .setTopN(topN).setTimeout(timeout)
                                      .setOptions(options), messageHandler);
            }
        }

        /// <summary>
        /// Executes a SOW query and places a subscription.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="orderBy">The ordering property</param>
        /// <param name="bookmark">A bookmark identifer, one of the constants from <see cref="Client.Bookmarks"/>, or a timestamp of the form <i>YYYYmmddTHHMMSS</i>,
        /// as described in the AMPS User's Guide. This the point at which AMPS executes the historical query</param>
        /// <param name="options">A AMPS.Client.Message.Options value indicating desired options for this</param>
        /// query. Use Message.Options.None if no options are desired</param>
        /// <param name="batchSize">The batching parameter to use for the results</param>
        /// <param name="topN">The maximum number of records the server will return (default is all that match)</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid</exception>
        public CommandId sowAndSubscribe(
              Action<Message> messageHandler,
              String topic,
              String filter = null,
              String orderBy = null,
              String bookmark = null,
              String options = null,
              int batchSize = 10,
              int topN = -1,
              long timeout = 0)
        {
            lock (internalLock)
            {
                return executeAsyncNoLock(_command
                                      .reset(Message.Commands.SOWAndSubscribe)
                                      .setTopic(topic).setFilter(filter)
                                      .setOrderBy(orderBy)
                                      .setBookmark(bookmark)
                                      .setBatchSize(batchSize)
                                      .setTopN(topN)
                                      .setTimeout(timeout)
                                      .setOptions(options),
                                      new ActionMessageHandler(messageHandler));
            }
        }
        /// <summary>
        /// Executes a SOW query and places a subscription.
        /// <pre>
        /// <code>
        /// namespace AMPSSOWandSubscribeConsoleSubscriber
        /// {
        ///     class FilteredSowAndSubscribeExample
        ///     {
        ///         private static string uri_ = "tcp://127.0.0.1:9027/amps/json";
        ///         public static void AttemptReconnection(Client client)
        ///         {
        ///             Thread.Sleep(250);
        ///             client.connect(uri_);
        ///         }
        ///         static void Main(string[] args)
        ///         {
        ///             using (Client client = new Client("filteredSowAndSubscribeClient"))
        ///             {
        ///                 try
        ///                 {
        ///                     client.setDisconnectHandler(AttemptReconnection);
        ///                     client.connect(uri_);
        ///                     client.logon();
        ///                     // Subscribe with filter
        ///                     string filter = "status = 'active'";
        ///                     foreach (var message in client.execute(new Command(Message.Commands.SOWAndSubscribe)
        ///                                                             .setTopic("example_topic")
        ///                                                             .setFilter(filter)
        ///                                                             .setOrderBy("timestamp")
        ///                                                             .setBookmark("20241111T120000")))
        ///                     {
        ///                         if (message.Command == Message.Commands.GroupBegin)
        ///                         {
        ///                             Console.WriteLine("Receiving filtered messages from the SOW.");
        ///                             continue;
        ///                         }
        ///                         if (message.Command == Message.Commands.GroupEnd)
        ///                         {
        ///                             Console.WriteLine("Done receiving filtered messages.");
        ///                             continue;
        ///                         }
        ///                         Console.WriteLine("Filtered Message: " + message.Data);
        ///                     }
        ///                 }
        ///                 catch (AMPSException e)
        ///                 {
        ///                     Console.Error.WriteLine(e.Message);
        ///                 }
        ///             }
        ///         }
        ///     }
        /// }
        /// </code>
        /// </pre>
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="batchSize">The batching parameter to use for the SOW query results</param>
        /// <param name="oofEnabled">True if out-of-focus ("OOF") messages are desired.</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndSubscribe(
              MessageHandler messageHandler,
              String topic,
              String filter,
              int batchSize,
              bool oofEnabled,
              long timeout)
        {
            return sowAndSubscribe(messageHandler, topic, filter, null, null, oofEnabled?"oof":null, batchSize, -1, timeout);
        }

        /// <summary>
        /// Executes a SOW query and places a subscription.
        /// </summary>
        /// <param name="a">The function to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="batchSize">The batching parameter to use for the SOW query results</param>
        /// <param name="sendOOF">True if out-of-focus ("OOF") messages are desired.</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndSubscribe(
              Action<Message> a,
              String topic,
              String filter,
              int batchSize,
              bool sendOOF,
              long timeout)
        {
            return sowAndSubscribe(new ActionMessageHandler(a), topic, filter, batchSize, sendOOF, timeout);
        }

        /// <summary>
        /// Executes a SOW query and places a subscription.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="batchSize">The batching parameter to use for the SOW query results</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndSubscribe(
              MessageHandler messageHandler,
              String topic,
              int batchSize,
              long timeout)
        {
            return sowAndSubscribe(messageHandler, topic, null, batchSize, false, timeout);
        }

        /// <summary>
        /// Executes a SOW query and places a subscription.
        /// </summary>
        /// <param name="a">The function to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="batchSize">The batching parameter to use for the SOW query results</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndSubscribe(
              Action<Message> a,
              String topic,
              int batchSize,
              long timeout)
        {
            return sowAndSubscribe(new ActionMessageHandler(a), topic, batchSize, timeout);
        }

        /// <summary>
        /// Executes a SOW query and places a subscription.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndSubscribe(
              MessageHandler messageHandler,
              String topic,
              long timeout)
        {
            return sowAndSubscribe(messageHandler, topic, null, 10, false, timeout);
        }

        /// <summary>
        /// Executes a SOW query and places a subscription.
        /// </summary>
        /// <param name="a">The function to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndSubscribe(
              Action<Message> a,
              String topic,
              long timeout)
        {
            return sowAndSubscribe(new ActionMessageHandler(a), topic, timeout);
        }

        /// <summary>
        /// Executes a SOW query and places a delta subscription.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="orderBy">The ordering property</param>
        /// <param name="bookmark">A bookmark identifer, one of the constants from <see cref="Client.Bookmarks"/>, or a timestamp of the form <i>YYYYmmddTHHMMSS</i>,
        /// as described in the AMPS User's Guide. This the point at which AMPS executes the historical query</param>
        /// <param name="options">A AMPS.Client.Message.Options value indicating desired options for this
        /// query. Use Message.Options.None if no options are desired</param>
        /// <param name="batchSize">The batching parameter to use for the results</param>
        /// <param name="topN">The maximum number of records the server will return (default is all that match)</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndDeltaSubscribe(
              MessageHandler messageHandler,
              String topic,
              String filter=null,
              String orderBy=null,
              String bookmark=null,
              String options=null,
              int batchSize=10,
              int topN=-1,
              long timeout=0)
        {
            lock (internalLock)
            {
                return executeAsyncNoLock(_command
                                  .reset(Message.Commands.SOWAndDeltaSubscribe)
                                  .setTopic(topic).setFilter(filter)
                                  .setOrderBy(orderBy).setBookmark(bookmark)
                                  .setBatchSize(batchSize).setTopN(topN)
                                  .setTimeout(timeout).setOptions(options),
                                  messageHandler);
            }
        }

        /// <summary>
        /// Executes a SOW query and places a delta subscription.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="batchSize">The batching parameter to use for the SOW query results</param>
        /// <param name="oofEnabled">Specifies whether or not Out-of-Focus processing is enabled</param>
        /// <param name="sendEmpties">Specifies whether or not unchanged records are received on the delta subscription</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndDeltaSubscribe(
              MessageHandler messageHandler,
              String topic,
              String filter,
              int batchSize,
              bool oofEnabled,
              bool sendEmpties,
              long timeout)
        {
            return sowAndDeltaSubscribe(messageHandler, topic, filter, null, null,addOptions(oofEnabled, sendEmpties), batchSize, -1, timeout);
        }

        /// <summary>
        /// Executes a SOW query and places a delta subscription.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter</param>
        /// <param name="batchSize">The batching parameter to use for the SOW query results</param>
        /// <param name="oofEnabled">Specifies whether or not Out-of-Focus processing is enabled</param>
        /// <param name="sendEmpties">Specifies whether or not unchanged records are received on the delta subscription</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndDeltaSubscribe(
              Action<Message> messageHandler,
              String topic,
              String filter,
              int batchSize,
              bool oofEnabled,
              bool sendEmpties,
              long timeout)
        {
            return sowAndDeltaSubscribe(new ActionMessageHandler(messageHandler), topic,
                filter, batchSize, oofEnabled, sendEmpties, timeout);
        }
        /// <summary>
        /// Executes a SOW query and places a delta subscription.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="batchSize">The batching parameter to use for the SOW query results</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndDeltaSubscribe(
              MessageHandler messageHandler,
              String topic,
              int batchSize,
              long timeout)
        {
            return sowAndDeltaSubscribe(messageHandler, topic, null, batchSize, false, true, timeout);
        }

        /// <summary>
        /// Executes a SOW query and places a delta subscription.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="batchSize">The batching parameter to use for the SOW query results
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndDeltaSubscribe(
              Action<Message> messageHandler,
              String topic,
              int batchSize,
              long timeout)
        {
            return sowAndDeltaSubscribe(new ActionMessageHandler(messageHandler), topic, batchSize, timeout);
        }

        /// <summary>
        /// Executes a SOW query and places a delta subscription.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndDeltaSubscribe(
              MessageHandler messageHandler,
              String topic,
              long timeout)
        {
            return sowAndDeltaSubscribe(messageHandler, topic, null, 10, false, true, timeout);
        }
        /// <summary>
        /// Executes a SOW query and places a delta subscription.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with matching messages</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="timeout">The maximum time to wait for the subscription to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadRegexTopicException">The specified regex topic was invalid.</exception>
        public CommandId sowAndDeltaSubscribe(
              Action<Message> messageHandler,
              String topic,
              long timeout)
        {
            return sowAndDeltaSubscribe(new ActionMessageHandler(messageHandler), topic, timeout);
        }


        /// <summary>
        /// Executes a SOW delete with filter.
        /// <pre>
        /// <code>
        /// namespace AMPSSOWDeleteExample
        /// {
        ///     class AMPSSOWDeleteExample
        ///     {
        ///        private static string uri_ = "tcp://127.0.0.1:9027/amps/json";
        ///         static void Main(string[] args)
        ///         {
        ///             Client client = new Client("SOWConsolePublisher");
        ///             try
        ///             {
        ///                 System.Console.WriteLine("... starting ...");
        ///                 // connect and logon
        ///                 client.connect(uri_);
        ///                 client.logon();
        ///                 for (int number = 0; number < 10000; number += 1250)
        ///                 {
        ///                     client.publish("messages-sow",
        ///                         "{ \"messageNumber\":" + number + ", " +
        ///                         "\"OptionalField\":\"ignore_me\", " +
        ///                         "\"message\":\"Updated, world!\"}");
        ///                 }
        ///                 client.sowDelete(
        ///                     new DefaultMessageHandler(),
        ///                     "messages-sow", "/messageNumber = 500", 0);
        ///             }
        ///             catch (AMPSException e)
        ///             {
        ///                 System.Console.WriteLine(e.Message);
        ///                 System.Console.WriteLine(e.StackTrace);
        ///             }
        ///         }
        ///     }
        /// }
        /// </code>
        /// </pre>
        /// </summary>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter. To delete all records, set a filter that is always true ('1=1')</param>
        /// <param name="timeout">The maximum time to wait for the SOW delete to be started to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The stats message returned from the sow_delete.</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        public Message sowDelete(String topic,
                                 String filter,
                                 long timeout=0)
        {
            MessageStream ms = null;
            lock (internalLock)
            {
                ms = _execute(_publishCommand.Value.reset(Message.Commands.SOWDelete)
                                       .addAckType(Message.AckTypes.Stats)
                                       .setTopic(topic)
                                       .setFilter(filter).setTimeout(timeout), false);
            }
            if (ms == null) return null;
            ms.MoveNext();
            return ms.Current;
        }
        /// <summary>
        /// Executes a SOW delete with filter.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with stats and completed acknowledgements</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter. To delete all records, set a filter that is always true ('1=1')</param>
        /// <param name="timeout">The maximum time to wait for the SOW delete to be started to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        public CommandId sowDelete(
                      MessageHandler messageHandler,
                      String topic,
                      String filter,
                      long timeout=0)
        {
            lock (internalLock)
            {
                return executeAsyncNoLock(_publishCommand.Value
                                          .reset(Message.Commands.SOWDelete)
                                          .addAckType(Message.AckTypes.Stats)
                                          .setTopic(topic).setFilter(filter)
                                          .setTimeout(timeout),
                                          messageHandler);
            }
        }

        /// <summary>
        /// Executes a SOW delete with filter.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with stats and completed acknowledgements</param>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="filter">The filter. To delete all records, set a filter that is always true ('1=1')</param>
        /// <param name="timeout">The maximum time to wait for the SOW delete to be started to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="AuthenticationException">An authentication exception occurred while logging on.</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        /// <exception cref="BadFilterException">The specified filter was invalid.</exception>
        public CommandId sowDelete(
                      Action<Message> messageHandler,
                      String topic,
                      String filter,
                      long timeout=0)
        {
            return sowDelete(new ActionMessageHandler(messageHandler), topic, filter, timeout);
        }

        /// <summary>
        /// Deprecated method to start a timer. Deprecated in AMPS server versions starting with 5.3.2.0.
        /// </summary>
        [Obsolete("startTimer() is deprecated in AMPS server versions starting with 5.3.2.0.")]
        public void startTimer()
        {
            if (this.serverVersion.getVersion() >= 500030000200000UL)
            {
                throw new CommandException("The start_timer command has been deprecated starting in AMPS server 5.3.2.0");
            }
            lock (internalLock)
            {
                executeAsyncNoLock(_command.reset(Message.Commands.StartTimer), null);
            }
        }

        /// <summary>
        /// Deprecated method to stop a timer. Deprecated in AMPS server versions starting with 5.3.2.0.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with acknowledgment</param>
        /// <param name="timeout">The maximum time to wait for the stop timer to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command.</returns>
        [Obsolete("stopTimer() is deprecated in AMPS server versions starting with 5.3.2.0.")]
        public CommandId stopTimer(MessageHandler messageHandler, long timeout)
        {
            if (this.serverVersion.getVersion() >= 500030000200000UL)
            {
                throw new CommandException("The stop_timer command has been deprecated starting in AMPS server 5.3.2.0");
            }
            lock (internalLock)
            {
                return executeAsyncNoLock(_command.reset(Message.Commands.StopTimer)
                                          .addAckType(Message.AckTypes.Completed),
                                          messageHandler);
            }
        }

        /// <summary>
        /// Gets the uninterpreted logon correlation information currently set.
        /// </summary>
        /// <param name="correlationData">The base64 data to be used for logon correlation</param>
        public void setLogonCorrelationData(String correlationData)
        {
            logonCorrelationData = correlationData;
        }


        /// <summary>
        /// Gets the uninterpreted logon correlation information currently set.
        /// </summary>
        /// <returns>The base64 data to be used for logon correlation.</returns>
        public String getLogonCorrelationData()
        {
            return logonCorrelationData;
        }

        /// <summary>
        /// Executes a SOW delete by sow key.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with stats and completed acknowledgements</param>
        /// <param name="topic">The topic to delete from</param>
        /// <param name="keys">The keys to delete</param>
        /// <param name="timeout">The maximum time to wait for the SOW delete to be started to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        public CommandId sowDeleteByKeys(
            MessageHandler messageHandler,
            String topic,
            String keys,
            long timeout=0)
        {
            lock (internalLock)
            {
                return executeAsyncNoLock(_publishCommand.Value
                                          .reset(Message.Commands.SOWDelete)
                                          .addAckType(Message.AckTypes.Stats)
                                          .setTopic(topic).setSowKeys(keys)
                                          .setTimeout(timeout),
                                          messageHandler);
            }
        }

        /// <summary>
        /// Executes a SOW delete by sow key.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with stats and completed acknowledgements</param>
        /// <param name="topic">The topic to delete from</param>
        /// <param name="keys">The keys to delete</param>
        /// <param name="timeout">The maximum time to wait for the SOW delete to be started to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        public CommandId sowDeleteByKeys(
            Action<Message> messageHandler,
            String topic,
            String keys,
            long timeout=0)
        {
            return sowDeleteByKeys(new ActionMessageHandler(messageHandler), topic, keys, timeout);
        }

        /// <summary>
        /// Executes a SOW delete by data. AMPS uses key fields in the data to find and delete a message with the same keys.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with stats and completed acknowledgements</param>
        /// <param name="topic">The topic to delete from</param>
        /// <param name="data">The message to match and delete in the SOW cache</param>
        /// <param name="timeout">The maximum time to wait for the SOW delete to be started to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        public CommandId sowDeleteByData(
            MessageHandler messageHandler,
            String topic,
            String data,
            long timeout=0)
        {
            lock (internalLock)
            {
                return executeAsyncNoLock(_publishCommand.Value
                                          .reset(Message.Commands.SOWDelete)
                                          .addAckType(Message.AckTypes.Stats)
                                          .setTopic(topic).setData(data)
                                          .setTimeout(timeout),
                                          messageHandler);
            }
        }

        /// <summary>
        /// Executes a SOW delete by data. AMPS uses key fields in the data to find and delete a message with the same keys.
        /// </summary>
        /// <param name="messageHandler">The message handler to invoke with stats and completed acknowledgements</param>
        /// <param name="topic">The topic to delete from</param>
        /// <param name="data">The message to match and delete in the SOW cache</param>
        /// <param name="timeout">The maximum time to wait for the SOW delete to be started to be placed (milliseconds), where 0 indicates no timeout</param>
        /// <returns>The command identifier assigned to this command</returns>
        /// <exception cref="TimedOutException">The command execution exceeded the timeout value</exception>
        /// <exception cref="DisconnectedException">The connection was disconnected at time of send</exception>
        /// <exception cref="NotEntitledException">An entitlement error occurred on this topic.</exception>
        /// <exception cref="InvalidTopicException">The specified topic was invalid.</exception>
        public CommandId sowDeleteByData(
            Action<Message> messageHandler,
            String topic,
            String data,
            long timeout=0)
        {
            return sowDeleteByData(new ActionMessageHandler(messageHandler), topic, data, timeout);
        }

        // Handles acknowledgment for flush operation.
        private class FlushAckHandler : MessageHandler
        {

            /// <summary>
            /// Enumeration representing the status of the flush acknowledgment handler.
            /// </summary>
            public enum Status : int
            {
                // The handler is waiting for acknowledgment.
                Waiting = 0,
                // The acknowledgment has been received.
                Acked = 1,
                // The handler is disconnected.
                Disconnected = 2
            }

            public volatile Status _status = Status.Waiting;

            /// <summary>
            /// Initializes a new instance of the FlushAckHandler class.
            /// </summary>
            /// <param name="client_">The AMPS client instance</param>
            public FlushAckHandler(Client client_)
            {
                client_.ConnectionStateChanged += (caller, args) => { this.invoke(args); };
            }

            /// <summary>
            /// Handles acknowledgment for a flush message.
            /// </summary>
            /// <param name="m">The acknowledgment message</param>
            public void invoke(Message m)
            {
                _status = Status.Acked;
            }

            /// <summary>
            /// Handles connection state changes for the client.
            /// </summary>
            /// <param name="args">The event arguments containing the new connection state</param>
            public void invoke(Client.ConnectionStateChangedEventArgs args)
            {
                if (Client.ConnectionStateChangedEventArgs.ConnectionState.Disconnected == args.NewState
                    || Client.ConnectionStateChangedEventArgs.ConnectionState.Shutdown == args.NewState)
                {
                    _status = Status.Disconnected;
                }
            }
        }

        /// <summary>Ensures that AMPS messages are sent and have been processed
        /// by the AMPS server. The client issues a flush command and waits for
        /// an acknowledgement. This method blocks until messages have been
        /// processed and is most useful when the application reaches a point at
        /// which it is acceptable to block to ensure that messages are
        /// delivered to the AMPS server. For example, an application might call
        /// publishFlush before exiting. One thing to note is that if AMPS is
        /// unavailable (HA Client), publishFlush needs to wait for a connection
        /// to come back up which may look like it's hanging.
        public void publishFlush()
        {
            if (publishStore == null)
            {
                try
                {
                    lock (this.internalLock)
                    {
                        this.message.reset();
                        CommandId id = CommandId.nextIdentifier();
                        this.message.setCommandId(id);
                        this.message.setAckType(Message.AckTypes.Processed);
                        this.message.setCommand(Message.Commands.Flush);
                        this.syncAckProcessing(id, message, 0);
                        return;
                    }
                }
                catch (NullReferenceException)
                {
                    throw new DisconnectedException("not connected");
                }

            }
            else
            {
                publishStore.flush();
            }
        }

        /// <summary>Ensures that AMPS messages are sent and have been processed
        /// by the AMPS server. The client issues a flush command and waits for
        /// an acknowledgement or until the timeout expires. This method blocks
        /// until messages have been processed and is most useful when the
        /// application reaches a point at which it is acceptable to block to
        /// ensure that messages are  delivered to the AMPS server. For example,
        /// an application might call publishFlush before exiting. One thing to
        /// note is that if AMPS is  unavailable (HA Client), publishFlush needs
        /// to wait for a connection to come back up which may look like
        /// it's hanging.
        /// <param name="timeout"> Number of milliseconds to wait for flush, where 0 indicates no timeout</param>
        public void publishFlush(int timeout)
        {
            if (publishStore == null)
            {
                try
                {
                    lock (this.internalLock)
                    {
                        this.message.reset();
                        CommandId id = CommandId.nextIdentifier();
                        this.message.setCommandId(id);
                        this.message.setAckType(Message.AckTypes.Processed);
                        this.message.setCommand(Message.Commands.Flush);
                        this.syncAckProcessing(id, message, timeout);
                        return;
                    }
                }
                catch (NullReferenceException)
                {
                    throw new DisconnectedException("not connected");
                }
            }
            else
            {
                publishStore.flush(timeout);
            }
        }

        private static VersionInfo minPersistedFlushVersion = new VersionInfo("5.3.3");

        /// <summary>Ensures that AMPS messages are sent and have been processed
        /// by the AMPS server. The client issues a flush command and waits for
        /// an acknowledgement. This method blocks until messages have been
        /// processed and is most useful when the application reaches a point at
        /// which it is acceptable to block to ensure that messages are
        /// delivered to the AMPS server. For example, an application might call
        /// publishFlush before exiting. One thing to note is that if AMPS is
        /// unavailable (HA Client), publishFlush needs to wait for a connection
        /// to come back up which may look like it's hanging.
        /// <param name="ackType_"> The acknowledgement type that the flush command should wait to receive, see {@link Message.AckType}.
        /// 'processed' matches the timeout only version of this function. `persisted` ensures all previous messages have been persisted</param>
        ///
        public void publishFlush(string ackType_)
        {
            Message.AckTypes ackType = 0;
            if ("persisted" == ackType_)
            {
                if (this.serverVersion < minPersistedFlushVersion)
                {
                    ackType = Message.AckTypes.Processed;
                }
                else
                {
                    ackType = Message.AckTypes.Persisted;
                }
            }
            else if ("processed" == ackType_)
            {
                ackType = Message.AckTypes.Processed;
            }
            else
            {
                throw new CommandException("publishFlush only accepts 'processed' or 'persisted' ack types");
            }
            CommandId id = CommandId.nextIdentifier();
            FlushAckHandler handler = new FlushAckHandler(this);
            try
            {
                lock (this.internalLock)
                {
                    if (!this.connected)
                    {
                        throw new DisconnectedException("Not connected");
                    }
                    this.message.reset();
                    this.message.setCommandId(id).setCommand(Message.Commands.Flush).setAckType(ackType);
                    this.addMessageHandler(id, handler, ackType, false);
                    this.send(this.message);
                }
            }
            catch (NullReferenceException)
            {
                throw new DisconnectedException("not connected");
            }
            try
            {
                if (publishStore != null)
                {
                    publishStore.flush();
                }
                while (handler._status == FlushAckHandler.Status.Waiting)
                {
                    System.Threading.Thread.Sleep(10);
                }
                if (handler._status == FlushAckHandler.Status.Disconnected
                    && publishStore == null)
                {
                    throw new DisconnectedException("Client disconnected waiting for flush acknowledgement");
                }
            }
            finally
            {
                removeMessageHandler(id);
                this.ConnectionStateChanged -= (caller, args) => { handler.invoke(args); };
            }
        }

        /// <summary>Ensures that AMPS messages are sent and have been processed
        /// by the AMPS server. The client issues a flush command and waits for
        /// an acknowledgement or until the timeout expires. This method blocks
        /// until messages have been processed and is most useful when the
        /// application reaches a point at which it is acceptable to block to
        /// ensure that messages are  delivered to the AMPS server. For example,
        /// an application might call publishFlush before exiting. One thing to
        /// note is that if AMPS is  unavailable (HA Client), publishFlush needs
        /// to wait for a connection to come back up which may look like
        /// it's hanging.
        /// <param name="ackType_"> The acknowledgement type that the flush command should wait to receive, see {@link Message.AckType}.
        /// 'processed' matches the timeout only version of this function. `persisted` ensures all previous messages have been persisted</param>
        /// <param name="timeout"> Number of milliseconds to wait for flush, where 0 indicates no timeout</param>
        public void publishFlush(string ackType_, int timeout)
        {
            Message.AckTypes ackType = 0;
            if ("persisted" == ackType_)
            {
                if (this.serverVersion < minPersistedFlushVersion)
                {
                    ackType = Message.AckTypes.Processed;
                }
                else
                {
                    ackType = Message.AckTypes.Persisted;
                }
            }
            else if ("processed" == ackType_)
            {
                ackType = Message.AckTypes.Processed;
            }
            else
            {
                throw new CommandException("publishFlush only accepts 'processed' or 'persisted' ack types");
            }
            CommandId id = CommandId.nextIdentifier();
            FlushAckHandler handler = new FlushAckHandler(this);
            try
            {
                lock (this.internalLock)
                {
                    if (!this.connected)
                    {
                        throw new DisconnectedException("Not connected");
                    }
                    this.message.reset();
                    this.message.setCommandId(id).setCommand(Message.Commands.Flush).setAckType(ackType);
                    this.addMessageHandler(id, handler, ackType, false);
                    this.send(this.message);
                }
            }
            catch (NullReferenceException)
            {
                throw new DisconnectedException("not connected");
            }
            Stopwatch timer = timeout > 0 ? Stopwatch.StartNew() : null;
            try
            {
                if (publishStore != null)
                {
                    publishStore.flush();
                }
                while (handler._status == FlushAckHandler.Status.Waiting
                       && (timer == null || timer.ElapsedMilliseconds <= timeout))
                {
                    System.Threading.Thread.Sleep(10);
                }
                if (handler._status == FlushAckHandler.Status.Disconnected
                    && publishStore == null)
                {
                    throw new DisconnectedException("Client disconnected waiting for flush acknowledgement");
                }
                if (handler._status != FlushAckHandler.Status.Acked
                    && publishStore == null)
                {
                    throw new TimedOutException("Timed out waiting for flush acknowledgement");
                }
            }
            finally
            {
                removeMessageHandler(id);
                this.ConnectionStateChanged -= (caller, args) => { handler.invoke(args); };
            }
        }

        /// <summary>
        /// Clear the queued messages which may be waiting in the transport
        /// </summary>
        /// <returns> long Number of messages or bytes queued after flush is completed or timed out. </returns>
        [Obsolete("flush() is deprecated and had no functionality, please use publishFlush instead if you wish to wait for all publishes to complete.")]
        public long flush()
        {
            try
            {
                lock (this.internalLock)
                {
                    return this.transport.flush();
                }
            }
            catch (NullReferenceException)
            {
                throw new DisconnectedException("not connected");
            }
        }

        /// <summary>
        /// Clear the queued messages which may be waiting in the transport
        /// </summary>
        /// <param name="timeout">Number of milliseconds to wait for flush, where 0 indicates no timeout</param>
        /// <returns>long Number of messages or bytes queued after flush is completed or timed out</returns>
        [Obsolete("flush() is deprecated and had no functionality, please use publishFlush instead if you wish to wait for all publishes to complete.")]
        public long flush(long timeout)
        {
            try
            {
                return this.transport.flush(timeout);
            }
            catch (NullReferenceException)
            {
                throw new DisconnectedException("not connected");
            }
        }

        /// <summary>
        /// Return the build number for the client that is stored in the Manifest.mf of the jar file.
        /// </summary>
        /// <returns>String Build version number.</returns>

        public static String getVersion()
        {
            object[] attributes = typeof(Client).Assembly.GetCustomAttributes(
                typeof(AssemblyInformationalVersionAttribute), false);
            if (attributes != null && attributes.Length > 0)
            {
                AssemblyInformationalVersionAttribute attribute = (AssemblyInformationalVersionAttribute) attributes[0];
                if (attribute != null)
                    return attribute.InformationalVersion.ToString();
            }
            return typeof(Client).Assembly.GetName().Version.ToString();
        }

        /// <summary>
        /// Set a handler that is invoked immediately by any thread created by the
        /// transport.
        /// </summary>
        /// <param><typeparamref name="ThreadCreatedHandler"/> to be invoked by any
        /// new transport threads</param>
        public void setThreadCreatedHandler(ThreadCreatedHandler handler_)
        {
            _threadCreatedHandler = handler_;
            if (transport != null)
            {
                transport.setThreadCreatedHandler(_threadCreatedHandler);
            }
        }

        /// <summary>
        /// Set a handler that is invoked immediately by any thread created by the
        /// transport.
        /// </summary>
        /// <param><typeparamref name="Action"/> to be invoked by any
        /// new transport threads</param>
        public void setThreadCreatedHandler(Action handler_)
        {
            _threadCreatedHandler = new ActionThreadCreatedHandler(handler_);
            if (transport != null)
            {
                transport.setThreadCreatedHandler(_threadCreatedHandler);
            }
        }

        /// <summary>
        /// Get the handler that is invoked immediately by any thread created by the
        /// transport.
        /// </summary>
        /// <returns><typeparamref name="ThreadCreatedHandler"/> to be invoked by any
        /// new transport threads.</returns>
        public ThreadCreatedHandler getThreadCreatedHandler()
        {
            return _threadCreatedHandler;
        }

        /// <summary>
        /// Sets a TransportFilter that filters raw bytes before send and after receive.
        /// </summary>
        /// <param name="filter_">A TransportFilter such as AMPS.Client.TransportTraceFilter.
        public void setTransportFilter(TransportFilter filter_)
        {
            _transportFilter = filter_ == null ? null : new TransportFilterWrapper(filter_);
            if (transport != null)
            {
                transport.setTransportFilter(_transportFilter);
            }
        }

        /// <summary>
        /// Sets a TransportFilterModifiable that filters and possibly modifies raw bytes
        /// before send and after receive.
        /// </summary>
        /// <param name="filter_">A TransportFilterModifiable.
        public void setTransportFilter(TransportFilterModifiable filter_)
        {
            _transportFilter = filter_;
            if (transport != null)
            {
                transport.setTransportFilter(_transportFilter);
            }
        }

        /// <summary>
        /// Gets the TransportFilterModifiable that filters and possibly modifies raw bytes
        /// before send and after receive.
        /// </summary>
        /// <returns>A TransportFilterModifiable.</returns>
        public TransportFilterModifiable getTransportFilter()
        {
            return _transportFilter;
        }

        /// <summary>
        /// Set the list of HTTP preflight headers, replacing any previously set.
        /// </summary>
        /// <param name="headers_">The new list of each line of headers for an
        /// HTTP connect request</param>
        public void setHttpPreflightHeaders(IEnumerable<string> headers_)
        {
            _httpPreflightHeaders = new List<string>(headers_);
        }

        /// <summary>
        /// Add a new header line to the list of HTTP preflight headers.
        /// </summary>
        /// <param name="header_">The header line to add.
        /// HTTP connect request</param>
        public void addHttpPreflightHeader(string header_)
        {
            if (_httpPreflightHeaders == null)
            {
                _httpPreflightHeaders = new List<string>();
            }
            _httpPreflightHeaders.Add(header_);
        }

        /// <summary>
        /// Add a new key/value pair to the list of HTTP preflight headers. It
        /// will be added as key: value
        /// </summary>
        /// <param name="key_">The key to add</param>
        /// <param name="value_">The value for the given key.
        /// HTTP connect request</param>
        public void addHttpPreflightHeader(string key_, string value_)
        {
            if (_httpPreflightHeaders == null)
            {
                _httpPreflightHeaders = new List<string>();
            }
            _httpPreflightHeaders.Add($"{key_}: {value_}");
        }

        /// <summary>
        /// Get the list of HTTP preflight headers.
        /// </summary>
        /// <returns>The list of each line of headers for an
        /// HTTP connect request.</returns>
        public IEnumerable<string> getHttpPreflightHeaders()
        {
            return _httpPreflightHeaders;
        }

        protected internal MessageHandler getTransportMessageHandler()
        {
            return this.handler;
        }

        protected internal TransportDisconnectHandler getTransportDisconnectHandler()
        {
            return this.handler;
        }


         /// <summary>
         /// Assembles a new ConnectionInfo with the state of this client and
         /// associated classes.
         /// </summary>
         /// <returns>A ConnectionInfo object.</returns>
        public virtual ConnectionInfo getConnectionInfo()
        {
            ConnectionInfo ci = new ConnectionInfo();
            ci["client.uri"] = this.uri == null ? null : this.uri.ToString();
            ci["client.name"] = this.name;
            ci["client.username"] = this.username;
            if (this.publishStore != null)
            {
                ci["publishStore.unpersistedCount"] = this.publishStore.unpersistedCount();
            }
            return ci;
        }

        // Computes the hash value for a given byte array representing a topic.
        private long _hashTopic(byte[] topic, int pos, int len)
        {
            long result = 0;
            unsafe
            {
                fixed(byte* topicBytes = &topic[pos])
                {
                    byte* b = topicBytes;
                    byte* be = b + len;
                    while (b != be)
                    {
                        result = (result * 63) ^ *b;
                        ++b;
                    }
                    return result;
                }
            }
        }

        // Processes acknowledgments for message queue messages.
        private int _ack(QueueBookmarks bookmarks)
        {
            // called with lock taken
            if(bookmarks._bookmarkCount > 0)
            {
                if (bookmarks._length == 0 || bookmarks._data == null)
                {
                    bookmarks._bookmarkCount = 0;
                    bookmarks._length = 0;
                    return 0;
                }
                // Because of the unlock to call store(), we need to copy data
                // from bookmarks while the lock is taken. Other bookmarks
                // could get added while the lock is released, or another
                // thread could call _ack and cause a double-send. We'll reset
                // length and count to 0 and get a copy of the buffer before we
                // release the lock.
                int bookmarksLength = bookmarks._length;
                bookmarks._length = 0;
                bookmarks._bookmarkCount = 0;
                if (_publishCommand.Value._message == null)
                    _publishCommand.Value._message = allocateMessage();
                _publishCommand.Value._message.reset();
                _publishCommand.Value._message.setCommand(Message.Commands.SOWDelete)
                                              .setCommandId(_queueAckCommandId);
                _publishCommand.Value._message.getTopicRaw().set(bookmarks._topic, 0,
                                                                 bookmarks._topic.Length);
                if (_bookmarksData.Value.Length < bookmarksLength)
                {
                    _bookmarksData.Value = new byte[bookmarksLength*2];
                }
                Array.Copy(bookmarks._data, 0, _bookmarksData.Value, 0, bookmarksLength);
                _publishCommand.Value._message.getBookmarkRaw().set(_bookmarksData.Value,
                                                                    0, bookmarksLength);
                ulong sequence = 0;
                if (publishStore != null) {
                    _publishCommand.Value._message.setAckType(Message.AckTypes.Persisted);
                    try {
                        Monitor.Exit(internalLock);
                        this.publishStore.store(_publishCommand.Value._message);
                        if (!_publishCommand.Value._message.isSequenceNull())
                            sequence = _publishCommand.Value._message.getSequence();
                    }
                    finally
                    {
                        Monitor.Enter(internalLock);
                    }
                }
                bookmarks._length = 0;
                bookmarks._bookmarkCount = 0;
                sendInternal(_publishCommand.Value._message, sequence);
                return 1;
            }
            return 0;
        }

        /// <summary>
        /// ACKs a message queue message.
        /// </summary>
        /// <param name="topic">The message topic</param>
        /// <param name="topicPos">The starting index of the topic name</param>
        /// <param name="topicLen">The length of the topic name</param>
        /// <param name="bookmark">The message bookmark</param>
        /// <param name="bookmarkPos">The starting index of the message bookmark</param>
        /// <param name="bookmarkLen">The length of the message bookmark</param>
        /// <param name="options">An options string for the ack, such as 'cancel'</param>
        /// <param name="optionsPos">The starting index of the options</param>
        /// <param name="optionsLen">The length of the options</param>
        public void ack(byte[] topic, int topicPos, int topicLen,
          byte[] bookmark, int bookmarkPos, int bookmarkLen,
          byte[] options = null, int optionsPos = 0, int optionsLen = 0)
        {
            if (_isAutoAckEnabled) return;
            if (_ackBatchSize < 2 || (options != null && optionsLen > 0))
            {
                _ackImmediate(topic, topicPos, topicLen, bookmark, bookmarkPos,
                              bookmarkLen, options, optionsPos, optionsLen);
            }
            else
            {
                _ackMany(topic, topicPos, topicLen, bookmark, bookmarkPos,
                         bookmarkLen, options, optionsPos, optionsLen);
            }
        }

        private void _ackOne(byte[] topic, int topicPos, int topicLen,
          byte[] bookmark, int bookmarkPos, int bookmarkLen,
          byte[] options = null, int optionsPos = 0, int optionsLen = 0)
        {
            lock (internalLock)
            {
                long topicHash = _hashTopic(topic, topicPos, topicLen);
                QueueBookmarks bookmarks = null;
                if (!_topicHashMap.TryGetValue(topicHash, out bookmarks))
                {
                    bookmarks = new QueueBookmarks();
                    bookmarks._topic = new byte[topicLen];
                    Array.Copy(topic, topicPos, bookmarks._topic, 0, topicLen);
                    bookmarks._data = new byte[512];
                    _topicHashMap[topicHash] = bookmarks;
                    ++_topicHashMapSize;
                }
                int bookmarksCapacity = bookmarks._data.Length;
                int newLength = bookmarks._length + bookmarkLen + 1;
                if (bookmarksCapacity < newLength)
                {
                    bookmarksCapacity = 512 * (1+(newLength/512));
                    byte[] newData = new byte[bookmarksCapacity];
                    Array.Copy(bookmarks._data, 0, newData, 0, bookmarks._length);
                    bookmarks._data = newData;
                }
                if (bookmarks._length > 0)
                {
                    bookmarks._data[bookmarks._length] = (byte)',';
                    ++bookmarks._length;
                }
                else
                {
                    bookmarks._oldestTime = DateTime.Now;
                }
                Array.Copy(bookmark, bookmarkPos, bookmarks._data, bookmarks._length, bookmarkLen);
                bookmarks._length += bookmarkLen;
                if (++bookmarks._bookmarkCount >= _ackBatchSize) _ack(bookmarks);
            }
        }

        private void _ackMany(byte[] topic, int topicPos, int topicLen,
          byte[] bookmark, int bookmarkPos, int bookmarkLen,
          byte[] options = null, int optionsPos = 0, int optionsLen = 0)
        {
            lock (internalLock)
            {
                long topicHash = _hashTopic(topic, topicPos, topicLen);
                QueueBookmarks bookmarks = null;
                if (!_topicHashMap.TryGetValue(topicHash, out bookmarks))
                {
                    bookmarks = new QueueBookmarks();
                    bookmarks._topic = new byte[topicLen];
                    Array.Copy(topic, topicPos, bookmarks._topic, 0, topicLen);
                    bookmarks._data = new byte[512];
                    _topicHashMap[topicHash] = bookmarks;
                    ++_topicHashMapSize;
                }
                int start = bookmarkPos;
                int end = start+1;
                for (; end < bookmarkPos + bookmarkLen; ++end)
                {
                    if (bookmark[end] == (byte)',')
                    {
                        if (++bookmarks._bookmarkCount >= _ackBatchSize)
                        {
                            int bookmarksCapacity = bookmarks._data.Length;
                            int newLength = bookmarks._length + end - start + 1;
                            if (bookmarksCapacity < newLength)
                            {
                                bookmarksCapacity = 512 * (1+(newLength/512));
                                byte[] newData = new byte[bookmarksCapacity];
                                Array.Copy(bookmarks._data, 0, newData, 0, bookmarks._length);
                                bookmarks._data = newData;
                            }
                            if (bookmarks._length > 0)
                            {
                                bookmarks._data[bookmarks._length] = (byte)',';
                                ++bookmarks._length;
                            }
                            else
                            {
                                bookmarks._oldestTime = DateTime.Now;
                            }
                            Array.Copy(bookmark, start, bookmarks._data, bookmarks._length, end-start);
                            bookmarks._length += end-start;
                            _ack(bookmarks);
                            start = ++end;
                        }
                    }
                }
                if (end - start >= 4)
                {
                    int bookmarksCapacity = bookmarks._data.Length;
                    int newLength = bookmarks._length + end - start + 1;
                    if (bookmarksCapacity < newLength)
                    {
                        bookmarksCapacity = 512 * (1+(newLength/512));
                        byte[] newData = new byte[bookmarksCapacity];
                        Array.Copy(bookmarks._data, 0, newData, 0, bookmarks._length);
                        bookmarks._data = newData;
                    }
                    if (bookmarks._length > 0)
                    {
                        bookmarks._data[bookmarks._length] = (byte)',';
                        ++bookmarks._length;
                    }
                    else
                    {
                        bookmarks._oldestTime = DateTime.Now;
                    }
                    Array.Copy(bookmark, bookmarkPos, bookmarks._data, bookmarks._length, end-start);
                    bookmarks._length += end-start;
                    if (++bookmarks._bookmarkCount >= _ackBatchSize) _ack(bookmarks);
                }
            }
        }

        private void _ackImmediate(byte[] topic, int topicPos, int topicLen,
          byte[] bookmark, int bookmarkPos, int bookmarkLen,
          byte[] options = null, int optionsPos = 0, int optionsLen = 0)
        {
            if (bookmarkLen == 0 || bookmark == null) return;
            if (_publishCommand.Value._message == null)
                _publishCommand.Value._message = allocateMessage();
            _publishCommand.Value._message.reset();
            _publishCommand.Value._message.setCommand(Message.Commands.SOWDelete)
                                          .setCommandId(_queueAckCommandId);
            _publishCommand.Value._message.getTopicRaw().set(topic,
                                                             topicPos,
                                                             topicLen);
            _publishCommand.Value._message.getBookmarkRaw().set(bookmark,
                                                                bookmarkPos,
                                                                bookmarkLen);
            if (options != null && optionsLen > 0) {
                _publishCommand.Value._message.getOptionsRaw().set(options,
                                                                    optionsPos,
                                                                    optionsLen);
            }
            ulong sequence = 0;
            if (publishStore != null) {
                _publishCommand.Value._message.setAckType(Message.AckTypes.Persisted);
                this.publishStore.store(_publishCommand.Value._message);
                if (!_publishCommand.Value._message.isSequenceNull())
                    sequence = _publishCommand.Value._message.getSequence();
            }
            if (sequence > 0 || options == null ||
                !CANCEL_FIELD.Equals(_publishCommand.Value._message.getOptionsRaw()))
            {
                lock (internalLock)
                {
                    sendInternal(_publishCommand.Value._message, sequence);
                }
            }
            else
            {
                lock (internalLock)
                {
                    sendWithoutRetry(_publishCommand.Value._message);
                }
            }
            return;
        }

        /// <summary>
        /// ACKs a message queue message.
        /// </summary>
        /// <param name="topic">The topic of the message
        /// <param name="bookmark">The bookmark of the message
        /// <param name="options">The options for the ack, such as 'cancel'</param>
        public void ack(Field topic, Field bookmark, Field options = null)
        {
            if (_isAutoAckEnabled) return;
            if (options == null)
                ack(topic.buffer, topic.position, topic.length,
                    bookmark.buffer, bookmark.position, bookmark.length);
            else
                ack(topic.buffer, topic.position, topic.length,
                    bookmark.buffer, bookmark.position, bookmark.length,
                    options.buffer, options.position, options.length);
        }

        internal void _ack(Field topic, Field bookmark, Field options = null)
        {
            if (_ackBatchSize < 2 || (options != null && !options.isNull()))
            {
                if (options == null)
                {
                    _ackImmediate(topic.buffer, topic.position, topic.length,
                         bookmark.buffer, bookmark.position, bookmark.length);
                }
                else
                {
                    _ackImmediate(topic.buffer, topic.position, topic.length,
                         bookmark.buffer, bookmark.position, bookmark.length,
                         options.buffer, options.position, options.length);
                }
            }
            else
            {
                if (options == null)
                {
                    _ackOne(topic.buffer, topic.position, topic.length,
                         bookmark.buffer, bookmark.position, bookmark.length);
                }
                else
                {
                    _ackOne(topic.buffer, topic.position, topic.length,
                         bookmark.buffer, bookmark.position, bookmark.length,
                         options.buffer, options.position, options.length);
                }
            }
        }

        /// <summary>
        /// ACKS a message queue message.
        /// </summary>
        /// <param name="topic">The topic of the message</param>
        /// <param name="bookmark">The bookmark of the message</param>
        public void ack(String topic, String bookmark, String options = null)
        {
            if (bookmark == null || bookmark.Length == 0 || _isAutoAckEnabled)
            {
                return;
            }
            byte[] topicBytes = Encoding.UTF8.GetBytes(topic);
            byte[] bookmarkBytes = Encoding.UTF8.GetBytes(bookmark);
            if (options == null)
            {
                ack(topicBytes, 0, topicBytes.Length,
                    bookmarkBytes, 0, bookmarkBytes.Length);
            }
            else
            {
                byte[] optionsBytes = Encoding.UTF8.GetBytes(options);
                ack(topicBytes, 0, topicBytes.Length,
                    bookmarkBytes, 0, bookmarkBytes.Length,
                    optionsBytes, 0, optionsBytes.Length);
            }
        }

        /// <summary>
        /// Enables or disables auto-acking. When auto-acking is enabled any
        /// successful return from a message handler function will result in
        /// an ACK (sow_delete) message sent to the server. Any thrown
        /// exception will result in a "cancel" ACK sent to the server.
        /// </summary>
        /// <param name="isAutoAckEnabled">true to enable, false to disable</param>
        public void setAutoAck(bool isAutoAckEnabled)
        {
            _isAutoAckEnabled = isAutoAckEnabled;
        }

        /// <summary>
        /// Returns the current setting of auto-acking.
        /// </summary>
        /// <returns>true if enabled, false if disabled</returns>
        public bool getAutoAck()
        {
            return _isAutoAckEnabled;
        }

        /// <summary>
        /// The AutoAck property enables or disables auto-acking.
        /// When auto-acking is enabled any
        /// successful return from a message handler function will result in
        /// an ACK (sow_delete) message sent to the server. Any thrown
        /// exception will result in a "cancel" ACK sent to the server.
        /// </summary>
        public bool AutoAck
        {
            get { return _isAutoAckEnabled; }
            set { _isAutoAckEnabled = value; }
        }

        /// <summary>
        /// Sets the current ACK batch size.The ACK batch size controls how many successful ACKs
        /// (which are sow_delete messages) are batched together before sending to the server.When
        /// combined with the "max_backlog" and MaxPerSubscriptionBacklog server configuration parameter,
        /// greater network efficiency can be achieved when using message queues.Setting this parameter
        /// causes calls to "ack" and successful auto-acks to be batched; unsuccessful/cancel acks
        /// are sent immediately.
        /// </summary>
        /// <param name="batchSize">The batch size to use (default:1)</param>
        public void setAckBatchSize(int batchSize)
        {
            _ackBatchSize = batchSize;
        }

        /// <summary>
        /// Returns the current ACK batch size. See the documentation for setAckBatchSize(int).
        /// </summary>
        /// <returns>The current batch size.</returns>
        public int getAckBatchSize()
        {
            return _ackBatchSize;
        }

        /// <summary>
        /// The ACK batch size controls how many successful ACKs
        /// (which are sow_delete messages) are batched together before sending to the server.When
        /// combined with the "max_backlog" and MaxPerSubscriptionBacklog server configuration parameter,
        /// greater network efficiency can be achieved when using message queues.Setting this parameter
        /// causes calls to "ack" and successful auto-acks to be batched; unsuccessful/cancel acks
        /// are sent immediately.
        /// </summary>
        public int AckBatchSize
        {
            get { return _ackBatchSize; }
            set { _ackBatchSize = value; if (value > 0 && _queueAckTimeout == 0) _queueAckTimeout = 1000; }
        }

        /// <summary>
        /// Sets the ack timeout -- the maximum time to let a success ack be cached before sending.
        /// If 0, there is no timeout. The 0 value is only allowed if ack batch size is 1.
        /// </summary>
        /// <param name="ackTimeout">the ack timeout, in milliseconds</param>
        public void setAckTimeout(long ackTimeout)
        {
            if (ackTimeout > 0 || (ackTimeout == 0 && _ackBatchSize == 1))
            {
                _queueAckTimeout = ackTimeout;
            }
        }

        /// <summary>
        /// Returns the queue ack timeout.
        /// </summary>
        /// <returns>the ack timeout in milliseconds</returns>
        public long getAckTimeout()
        {
            return _queueAckTimeout;
        }

        /// <summary>
        /// Sends any pending queue ACKs to the server.
        /// </summary>
        public void flushAcks()
        {
            int sendCount = 0;
            lock(this.internalLock)
            {
                foreach (var queueBookmarks in _topicHashMap.Values)
                {
                    sendCount += _ack(queueBookmarks);
                }
            }
            if (sendCount > 0) publishFlush();
        }

        /// <summary>
        ///  Execute the provided command and process responses on the
        ///  client receive thread using the provided handler.
        ///
        ///
        ///  This method creates a message based on the provided {@link Command},
        ///  sends the message to AMPS, and invokes the provided
        ///  {@link MessageHandler } to process messages returned in response
        ///  to the command. Rather than providing messages on the calling
        ///  thread, the Client runs the message handler directly on the
        ///  client receive thread.
        ///
        ///  When the provided MessageHandler is not null, this method blocks
        ///  until AMPS acknowledges that the command has been processed. The
        ///  results of the command after that acknowledgement are provided to
        ///  the MessageHandler.
        /// </summary>
        ///  <param name="command"> The Command object containing the command to send to AMPS</param>
        ///  <param name="handler"> The MessageHandler to invoke to process messages received</param>
        ///  <returns> the CommandId for the command</returns>
        ///

        public CommandId executeAsync(Command        command,
                                      MessageHandler handler=null)
        {
            lock(this.internalLock)
            {
                return executeAsyncNoLock(command, handler);
            }
        }


        /// <summary>
        ///  Execute the provided command on a background thread and provide
        ///  messages received in response to the handler provided.
        ///
        ///  This method creates a message based on the provided {@link Command},
        ///  sends the message to AMPS, and invokes the provided
        ///  {@link Action } to process messages returned in response
        ///  to the command. AMPS sends the message and receives the results
        ///  on a background thread. The MessageHandler runs on the background thread.
        /// </summary>
        ///  <param name="command"> The Command object containing the command to send to AMPS</param>
        ///  <param name="action"> The Action to invoke to process messages received</param>
        ///  <returns> the CommandId for the command</returns>
        ///
        public CommandId executeAsync(Command         command,
                                      Action<Message> action)
        {
            lock(this.internalLock)
            {
                return executeAsyncNoLock(command,
                    new ActionMessageHandler(action));
            }
        }

        /// <summary>
        /// Execute the provided command and return messages received in response in a MessageStream.
        ///
        /// This method creates a message based on the provided Command,
        /// sends the message to AMPS, and receives the results. AMPS sends the
        /// message and receives the results on a background thread. That thread
        /// populates the MessageStream returned by this method.
        /// </summary>
        ///
        /// <param name="command">The Command object containing the command to send to AMPS</param>
        /// <returns>A MessageStream that provides message received in response to the command</returns>
        public MessageStream execute(Command command)
        {
            return _execute(command, true);
        }

        private MessageStream _execute(Command command, bool acquireLock)
        {
            // If the command is sow and has a sub_id, OR
            // if the command has a replace option, return the existing
            // messagestream, don't create a new one.
            CommandId cid = null;
            Message.Commands commandEnum = command.getCommand();
            Message.AckTypes ackType = command.getAckType();
            bool useExistingHandler = command.getSubId() != null && ((command.getOptions() != null && command.getOptions().Contains("replace")) || commandEnum == Message.Commands.SOW);
            if(useExistingHandler)
            {
                // Try to find the exist message handler.
                byte[] subId = command.getSubId().id;
                if(subId != null)
                {
                    MessageHandler existingHandler = _routes.findRoute(command.getSubId());
                    if (existingHandler != null)
                    {
                        // we found an existing handler. It might not be a message stream, but that's okay.
                        if (acquireLock)
                            executeAsync(command, existingHandler);
                        else
                            executeAsyncNoLock(command, existingHandler);
                        // This will return null if handler was not a messagestream.
                        return existingHandler as MessageStream;
                    }
                }
                // fall through; we'll a new handler altogether.
            }
            else if((commandEnum & Message.Commands.NoDataCommands)
                     != Message.Commands.Unknown
                    && (ackType == Message.AckTypes.None ||
                        ackType == Message.AckTypes.Persisted))
            {
                if (acquireLock)
                    executeAsync(command);
                else
                    executeAsyncNoLock(command, null);
                return MessageStream.getEmptyMessageStream();
            }
            MessageStream result = new MessageStream(this).maxDepth(_defaultMaxDepth);
            try
            {
                if (acquireLock)
                    cid = executeAsync(command, result);
                else
                    cid = executeAsyncNoLock(command, result);
            }
            finally
            {
                if (cid == null)
                {
                    result.close();
                    result = null;
                }
            }
            bool statsAck = ((command._message.getAckTypeOutgoing() & Message.AckTypes.Stats) != 0
                             && commandEnum != Message.Commands.Subscribe
                             && commandEnum != Message.Commands.DeltaSubscribe
                             && commandEnum != Message.Commands.SOWAndSubscribe
                             && commandEnum != Message.Commands.SOWAndDeltaSubscribe);

            if (result != null)
            {
                if(statsAck)
                {
                    result.setStatsOnly();
                    result.setQueryId(command.getQueryId());
                }
                else if (commandEnum == Message.Commands.SOW)
                {
                    result.setQueryId(command.getQueryId());
                    if ((ackType & Message.AckTypes.Completed) > 0) {
                        result.setAcksOnly(cid, ackType);
                    } else {
                        result.setSOWOnly();
                    }
                }
                else if(cid!=null)
                {
                    if (command.isSubscribe())
                    {
                        result.setSubscription(cid);
                        if (commandEnum == Message.Commands.SOWAndSubscribe ||
                            commandEnum == Message.Commands.SOWAndDeltaSubscribe)
                        {
                            result.setQueryId(command.getQueryId());
                        }
                    }
                    else if (commandEnum == Message.Commands.Publish ||
                             commandEnum == Message.Commands.DeltaPublish ||
                             commandEnum == Message.Commands.SOWDelete)
                    {
                        result.setAcksOnly(cid,
                                           command._message.getAckTypeOutgoing()
                                            & ~Message.AckTypes.Persisted);
                    }
                    else
                    {
                        result.setAcksOnly(cid,
                                         command._message.getAckTypeOutgoing());
                    }
                }
                else
                {
                    result.close();
                    result = null;
                }
            }
            return result;
        }

        // This is a bit complicated, but we're basically doing all the stuff we
        // need to do here to get the Client's state to reflect the command
        // we're about to send -- for example, record the subscription, set a
        // client sequence number, add a handler, and determine whether or not
        // to wait for a processed ack.
        private CommandId executeAsyncNoLock(Command command,
                                             MessageHandler messageHandler)
        {
            CommandId cid = command.prepare(this);
            if (cid != null) cid = cid.copy();
            CommandId subid = command.getSubId();
            if (subid != null) subid = subid.copy();
            CommandId qid = command.getQueryId();
            if (qid != null) qid = qid.copy();
            Message.Commands commandEnum = command.getCommand();
            Message.AckTypes requestedAcks = command.getAckType();
            Message.AckTypes systemAddedAcks = (messageHandler == null)
                                               ? Message.AckTypes.None
                                               : Message.AckTypes.Processed;
            bool isPublishStore = this.publishStore != null
                                  && command.needsSequenceNumber();
            // Don't add completed ack as a system ack for a SOW command
            // processed using a MessageStream handler because the message
            // stream will handle removing the route on group_end if a
            // completed or stats ack isn't explicitly requested by the user.
            // But an async message handler will still need automatic route
            // removal handled by the router when a SOW command is completed.
            if (commandEnum == Message.Commands.SOW && messageHandler != null
                    && !(messageHandler is MessageStream)) {
                systemAddedAcks |= Message.AckTypes.Completed;
            }
            else if (commandEnum == Message.Commands.Unsubscribe)
            {
                if (subid.ToString().Equals("all"))
                {
                    try
                    {
                        Monitor.Exit(internalLock);
                        this.subscriptionManager.clear();
                    }
                    finally
                    {
                        Monitor.Enter(internalLock);
                    }
                    _routes.unsubscribeAll();
                }
                else
                {
                    try
                    {
                        Monitor.Exit(internalLock);
                        this.subscriptionManager.unsubscribe(subid);
                    }
                    finally
                    {
                        Monitor.Enter(internalLock);
                    }
                    _routes.removeRoute(subid);
                }
            }

            // fixup bookmark and apply ack types as necessary.
            string bookmark = command.getBookmark();
            if(bookmark != null)
            {
                if (this.bookmarkStore != null)
                {
                    if(command.isSubscribe())
                    {
                        systemAddedAcks |= Message.AckTypes.Persisted;
                    }
                    if(Client.Bookmarks.MOST_RECENT.Equals(bookmark))
                    {
                        Field mostRecent = bookmarkStore.getMostRecent(command._message.getSubIdRaw());
                        command._message.setBookmark(mostRecent.buffer,
                                                     mostRecent.position,
                                                     mostRecent.length);
                    }
                    else if (bookmark.Length >= 5 || bookmark[0] != '0')
                    {
                        command._message.setBookmark(bookmark);
                        if (!bookmarkStore.isDiscarded(command._message))
                        {
                            bookmarkStore.log(command._message);
                            bookmarkStore.discard(command._message);
                            bookmarkStore.persisted(command._message.getSubIdRaw(),
                                                    command._message.getBookmarkRaw());
                        }
                    }
                }
                else if(Client.Bookmarks.MOST_RECENT.Equals(bookmark))
                {
                    command._message.setBookmark(Bookmarks.EPOCH);
                }
            }
            // If this is destined for a publish store, make sure to request a persisted ack before we add routes
            if (isPublishStore)
            {
                systemAddedAcks |= Message.AckTypes.Persisted;
            }
            bool isSubscribe = command.isSubscribe();
            // Register any message handler supplied if not a subscribe command.
            // For subscribe commands, we register later.
            bool isRegistered = false;
            if (messageHandler != null && cid == null && subid == null
                && qid == null
                && (commandEnum & Message.Commands.NoDataCommands) == Message.Commands.Unknown)
            {
                throw new ArgumentException("To use a messagehandler, you must also supply a command, query, or subscription ID.");
            }

            // if there's a command id & a processed ack set, use syncAckProcessing.
            bool useSyncSend = cid != null
                  && ((systemAddedAcks|requestedAcks) & Message.AckTypes.Processed) != 0;
            try
            {
                if(isPublishStore)
                {
                    command._message.setAckType(systemAddedAcks|requestedAcks);
                    while (true)
                    {
                        try
                        {
                            Monitor.Exit(internalLock);
                            this.publishStore.store(command._message);
                        }
                        finally
                        {
                            Monitor.Enter(internalLock);
                        }
                        // Register any message handler supplied
                        if (messageHandler != null
                            && !_registerHandler(qid, subid, cid,
                                                messageHandler, requestedAcks,
                                                systemAddedAcks, isSubscribe))
                        {
                            if ((commandEnum & Message.Commands.NoDataCommands) != Message.Commands.Unknown)
                            {
                                cid = CommandId.nextIdentifier();
                                command.setCommandId(cid);
                                command._message.setCommandId(cid);
                                _routes.addRoute(cid, messageHandler,
                                                 requestedAcks, systemAddedAcks,
                                                 isSubscribe);
                                useSyncSend = ((systemAddedAcks|requestedAcks)
                                                & Message.AckTypes.Processed) != 0;
                            }
                            isRegistered = true;
                        }
                        command.setClientSequenceNumber(command._message.getSequence());
                        if (useSyncSend)
                            syncAckProcessing(cid, command._message,
                                              command.getTimeout(),
                                              command._message.getSequence(), false);
                        else
                            sendInternal(command._message, command._message.getSequence());
                        break;
                    }
                }
                else
                {
                    if (isSubscribe)
                    {
                        // Cache values we'll need, since the command could be
                        // invalid after syncAckProcessing.
                        subid = new CommandId(subid.id, 0, subid.id.Length);
                        if (!_defaultSubscriptionManager && !(messageHandler is MessageStream))
                        {
                            bool existingSub = _routes.hasRoute(subid);
                            try
                            {
                                Message message = command._message.copy();
                                if (cid != null) message.getCommandId(cid);
                                command._message.setAckType(requestedAcks);
                                Monitor.Exit(internalLock);
                                try
                                {
                                    // Register sub in subscription manager.
                                    this.subscriptionManager.subscribe(messageHandler,
                                                                       message);
                                }
                                finally
                                {
                                    Monitor.Enter(internalLock);
                                }
                                if (_reconnectingSubscriptionManager) return subid;
                                if (cid != null) message.setCommandId(cid);
                                message.setAckType(systemAddedAcks | requestedAcks);
                                if (useSyncSend)
                                {
                                    if (!_registerHandler(qid, subid, cid,
                                                          messageHandler,
                                                          requestedAcks,
                                                          systemAddedAcks,
                                                          isSubscribe))
                                    {
                                        throw new ArgumentException("To use a messagehandler, you must also supply a command, query, or subscription ID.");
                                    }
                                    isRegistered = !existingSub;
                                    syncAckProcessing(cid, message,
                                                      command.getTimeout(),
                                                      0, !(messageHandler is MessageStream));
                                }
                                else
                                {
                                    sendInternal(message, 0, !(messageHandler is MessageStream));
                                }
                            }
                            catch (AMPSException)
                            {
                                // If a new subscription failed with an exception, be sure to
                                // remove it from the subscription manager so that it isn't
                                // retried in the event of a disconnect. If this is an
                                // existing subscription, then leave it alone.
                                if (!existingSub)
                                {
                                    Monitor.Exit(internalLock);
                                    try
                                    {
                                        this.subscriptionManager.unsubscribe(subid);
                                    }
                                    finally
                                    {
                                        Monitor.Enter(internalLock);
                                    }
                                }

                                // Rethrow in any case so the caller knows this failed.
                                throw;
                            }
                        }
                        else
                        {
                            command._message.setAckType(systemAddedAcks | requestedAcks);
                            if (useSyncSend)
                            {
                                if (!_registerHandler(qid, subid, cid,
                                                      messageHandler,
                                                      requestedAcks,
                                                      systemAddedAcks,
                                                      isSubscribe))
                                {
                                    command._message.setAckType(requestedAcks);
                                    throw new ArgumentException("To use a messagehandler, you must also supply a command, query, or subscription ID.");
                                }
                                isRegistered = true;
                                syncAckProcessing(cid, command._message,
                                                  command.getTimeout(),
                                                  0, false);
                            }
                            else
                            {
                                sendInternal(command._message);
                            }
                        }
                    }
                    else
                    {
                        // Register any message handler supplied
                        if (messageHandler != null)
                        {
                            if (!_registerHandler(qid, subid, cid,
                                                 messageHandler, requestedAcks,
                                                 systemAddedAcks, isSubscribe))
                            {
                                if ((commandEnum & Message.Commands.NoDataCommands) != Message.Commands.Unknown)
                                {
                                    cid = CommandId.nextIdentifier();
                                    command.setCommandId(cid);
                                    command._message.setCommandId(cid);
                                    _routes.addRoute(cid, messageHandler,
                                                     requestedAcks, systemAddedAcks,
                                                     isSubscribe);
                                    useSyncSend = ((systemAddedAcks|requestedAcks)
                                                    & Message.AckTypes.Processed) != 0;
                                    isRegistered = true;
                                }
                            }
                            else
                            {
                                isRegistered = true;
                            }
                        }
                        command._message.setAckType(systemAddedAcks|requestedAcks);
                        if (useSyncSend)
                        {
                            syncAckProcessing(cid, command._message, command.getTimeout());
                        }
                        else
                        {
                            sendInternal(command._message);
                        }
                    }
                }
            }
            catch (TimedOutException)
            {
                // Need to remove any registered handlers due to failure
                // Unless it's saved in the publish store and will be
                // resent after a reconnect without restoring the handler.
                if (isRegistered && !isPublishStore)
                {
                    try
                    {
                        unsubscribeInternal(cid);
                    }
                    catch (DisconnectedException) // ignore
                    {
                    }
                    if (qid != null && !cid.Equals(qid))
                    {
                        try
                        {
                            unsubscribeInternal(qid);
                        }
                        catch (DisconnectedException) // ignore
                        {
                        }
                    }
                    if (subid != null && !cid.Equals(subid))
                    {
                        try
                        {
                            unsubscribeInternal(subid);
                        }
                        catch (DisconnectedException) // ignore
                        {
                        }
                    }
                }
                command._message.setAckType(requestedAcks);
                throw;
            }
            catch (AMPSException)
            {
                // Need to remove any registered handlers due to failure
                // Unless it's saved in the publish store and will be
                // resent after a reconnect without restoring the handler.
                if (isRegistered && !isPublishStore)
                {
                    _routes.removeRoute(cid);
                    if (qid != null && !cid.Equals(qid))
                    {
                        _routes.removeRoute(qid);
                    }
                    if (subid != null && !cid.Equals(subid))
                    {
                        _routes.removeRoute(subid);
                    }
                }
                command._message.setAckType(requestedAcks);
                throw;
            }
            command._message.setAckType(requestedAcks);
            if (subid != null) return subid;
            if (qid != null) return qid;
            return cid;
        }

        private string addOptions(bool oofEnabled, bool sendEmpties = true, string options = null)
        {
            string result = options;
            if (oofEnabled)
            {
                if (result == null) result = "oof"; else result = result + ",oof";
            }
            if (!sendEmpties)
            {
                if (result == null) result = "no_empties"; else result = result + ",no_empties";
            }
            return result;
        }

        // This is a helper function used to encapsulate the logic for waiting on a
        // sent message so that async functions like subscribe and sowcan have an
        // easier time using an exception interface for failed command execution.
        // param name="id": universally unique id associated with the paramref name="m"
        // param name="m":typeparamref name="Message" to be sent
        // param name="timeout": timeout in ms to wait for ack, where 0 indicates no timeout
        // param name="haSeq": The HA sequence number of the command
        // param name="isSubscribe": true if the command is an HA subscribe variant
        private AckResponse syncAckProcessing(CommandId id, Message m,
                                              long timeout, ulong haSeq=0, bool isSubscribe=false)
        {
            // NOTE: This is called with lock already held
            // Make an empty ack response and stick it in the map.
            // The reader thread takes care of setting it, removing it from the map,
            // and signaling everyone when an ack comes in
            AckResponse ackResponse = new AckResponse();
            ackResponse.Responded = false;
            // This is the only place that should have both locks, but can't
            // release the client lock until after send to protect the message
            // from being changed
            lock(acksLock)
            {
                _acks.Add(id, ackResponse);
            }
            ackResponse.ConnectionVersion = sendInternal(m, haSeq, isSubscribe);
            if (ackResponse.ConnectionVersion == -1)
            {
                lock(acksLock)
                {
                    _acks.Remove(id);
                }
                return ackResponse;
            }
            // Timeout enforcement is our job since we may need to wait multiple times.
            // We stay in this loop until we see a response or we've run out of time.
            Monitor.Exit(internalLock);
            try
            {
                Stopwatch timer = timeout > 0 ? Stopwatch.StartNew() : null;
                lock(acksLock)
                {
                    // First check to see that our connection version hasn't
                    // already failed / been canceled by a client background
                    // thread that detected a disconnect.
                    if (!ackResponse.Abandoned
                        && ackResponse.ConnectionVersion <= _lastFailedConnectionVersion)
                    {
                        ackResponse.Abandoned = true;
                        _acks.Remove(id);
                    }

                    while (!ackResponse.Abandoned && !ackResponse.Responded &&
                          (timer == null || timer.ElapsedMilliseconds <= timeout))
                    {
                        try
                        {
                            if (timeout > 0)
                            {
                                int remainingTime = (int)timeout - (int)timer.ElapsedMilliseconds;
                                if (remainingTime > 0)
                                {
                                    Monitor.Wait(acksLock, remainingTime);
                                }
                            }
                            else
                            {
                                Monitor.Wait(acksLock);
                            }
                        }
                        catch (ThreadInterruptedException)
                        {
                            ;  // It's OK to be interrupted, let's just continue on...
                        }
                    }
                }
            }
            finally
            {
                Monitor.Enter(internalLock);
            }

            if (ackResponse.Responded)
            {
                // We found the ack in the callback, therefore we have
                //  the status and failure reason when they're set.
                if (ackResponse.State != Message.State.Failure)
                {
                    if (m.getCommand() == Message.Commands.Logon)
                    {
                        // if we're a 'logon' command we need to extract the seq number,
                        // if available
                        if (this.publishStore != null)
                        {
                            this.publishStore.discardUpTo(ackResponse.Sequence);
                            // On a fresh client, we need to adopt the sequence
                            // number the server has for us or lastPersisted from our store.
                            if (lastSentHaSequenceNumber < ackResponse.Sequence)
                            {
                                this.lastSentHaSequenceNumber = ackResponse.Sequence;
                            }
                            if (publishStore.getLastPersisted() != ulong.MaxValue &&
                                lastSentHaSequenceNumber <= publishStore.getLastPersisted() )
                            {
                                this.lastSentHaSequenceNumber = publishStore.getLastPersisted();
                            }
                        }
                    }
                    if(_ackBatchSize > 1 && ackResponse.Options != null && ackResponse.Options.Length > 0)
                    {
                        int maxBacklogReturned = extractMaxBacklog(ackResponse.Options);
                        if(maxBacklogReturned > 0 && maxBacklogReturned < _ackBatchSize)
                        {
                            _ackBatchSize = maxBacklogReturned;
                        }
                    }
                    // all is good, return
                    return ackResponse;
                }
                switch (ackResponse.Reason)
                {
                    case Message.Reasons.None:
                        return ackResponse;
                    case Message.Reasons.BadFilter:
                        throw new BadFilterException("Filter '" + m.getFilter() + "' is invalid.");
                    case Message.Reasons.BadRegexTopic:
                        throw new BadRegexTopicException("Regular Expression Topic '" + m.getTopic()
                            + "' is invalid.");
                    case Message.Reasons.InvalidTopic:
                        throw new InvalidTopicException("Topic '" + m.getTopic() + "' is invalid.");
                    case Message.Reasons.SubscriptionAlreadyExists:
                        throw new SubscriptionAlreadyExistsException("Subscription for command '"
                            + m.getCommandId() + "' already exists.");
                    case Message.Reasons.SubidInUse:
                        throw new SubidInUseException("Subscription with subscription id '"
                            + m.getSubId() + "' already exists.");
                    case Message.Reasons.NameInUse:
                        throw new NameInUseException("Client name \"" + m.getClientName() + "\" is already in use.");
                    case Message.Reasons.AuthFailure:
                        throw new AuthenticationException("Logon failed for user \"" + m.getUserId() + "\"");
                    case Message.Reasons.NotEntitled:
                        if (m.getCommand() == Message.Commands.Logon)
                            throw new NotEntitledException("User \"" + username + "\" not entitled to logon.");
                        else
                            throw new NotEntitledException("User \"" + username + "\" not entitled to topic \""
                            + m.Topic + "\".");
                    case Message.Reasons.InvalidBookmark:
                        throw new InvalidBookmarkException("Bookmark '" + m.getBookmark() + "' is not valid.");
                    case Message.Reasons.InvalidOrderBy:
                        throw new InvalidOrderByException("OrderBy '" + m.getOrderBy() + "' is not valid.");
                    case Message.Reasons.LogonRequired:
                        throw new LogonRequiredException();
                    case Message.Reasons.InvalidTopicOrFilter:
                        throw new InvalidTopicException("Invalid topic '" + m.getTopic() + "' or filter '" + m.getFilter() + "'.");
                    case Message.Reasons.InvalidSubId:
                        throw new InvalidSubIdException("SubId '" + m.getSubId() + "' is not valid.");
                    case Message.Reasons.BadSowKey:
                        if (m.isSowKeyNull())
                            throw new BadSowKeyException("SowKeys '" + m.getSowKeys() + "' is not valid.");
                        else
                            throw new BadSowKeyException("SowKey '" + m.getSowKey() + "' is not valid.");
                    case Message.Reasons.DuplicateLogon:
                        throw new DuplicateLogonException();
                    case Message.Reasons.InvalidOptions:
                        throw new InvalidOptionsException("Options '" + m.getOptions() + "' is not valid.");
                    case Message.Reasons.OrderByTooLarge:
                        throw new InvalidOrderByException("OrderBy '" + m.getOrderBy() + "' is too large.");
                    default:
                        // if we've made it here, we're not successful, we're not a known
                        //    reason code either...
                        throw new UnknownException("Unable to process this result. Reason: '" +
                                                ackResponse.ReasonText + "'.");
                }
            }
            else
            {
                // No ack found, must've timed out or disconnected waiting.
                if (!ackResponse.Abandoned)
                {
                    throw new TimedOutException("timed out waiting for operation");
                }
                else
                {
                    throw new DisconnectedException("Connection closed while waiting for operation.");
                }
            }
        }

        private int extractMaxBacklog(String optionsString)
        {
            // Parse options correctly even though, at the moment, only max_backlog is returned.
            int optionsLength = optionsString.Length;
            int keyStart = 0;
            int data = 0;
            bool inData = false;
            for (int i = 0; i < optionsLength; ++i)
            {
                char c = optionsString[i];
                switch (c)
                {
                    case '=':
                        if (!inData)
                        {
                            data = 0;
                            inData = true;
                        }
                        break;
                    case ' ':
                        if (!inData) ++keyStart;
                        break;
                    case ',':
                        if (optionsString.Substring(keyStart).StartsWith("max_backlog="))
                        {
                            return data;
                        }
                        data = 0;
                        keyStart = i + 1;
                        inData = false;
                        break;
                    default:
                        if (inData)
                        {
                            data = (data * 10) + ((int)c - 48);
                        }
                        break;
                }
            }
            if (optionsString.Substring(keyStart).StartsWith("max_backlog="))
            {
                return data;
            }
            return 0;
        }

        /// <summary>
        /// Handles incoming messages and disconnections for a client in AMPS.
        /// </summary>
        class ClientHandler : MessageHandler, TransportDisconnectHandler
        {
            Client client = null;
            CommandId key = new CommandId();

            /// <summary>
            /// Initializes a new instance of the ClientHandler class.
            /// </summary>
            /// <param name="client">
            public ClientHandler(Client client)
            {
                this.client = client;
            }

            /// <summary>
            /// Invoked before processing a message in case of a failed version.
            /// </summary>
            /// <param name="failedVersion">The failed version number.
            public void preInvoke(int failedVersion)
            {
                this.client.cancelSynchronousWaiters(failedVersion);
            }

            /// <summary>
            /// Invoked upon receiving a message.
            /// </summary>
            /// <param name="m">The received message.
            public void invoke(Message m)
            {
                const Message.Commands SOWMask = Message.Commands.SOW | Message.Commands.GroupBegin | Message.Commands.GroupEnd;
                const Message.Commands PublishMask = Message.Commands.OOF | Message.Commands.Publish | Message.Commands.DeltaPublish;
                try
                {
                    Message.Commands commandType = m.getCommand();

                    if ((commandType & SOWMask) > 0)
                    {
                        /* Not currently implemented to avoid an extra branch in delivery.
                        // A small cheat here using knowledge of Commands values for
                        // SOW (8), GroupBegin (8192), and GroupEnd (16384)
                        // and their GlobalCommandTypeHandlers values of 1, 2, 3
                        client._globalCommandTypeHandlers[(int)commandType/8192+1].invoke(m);
                        */
                        m.getQueryId(key);
                        do
                        {
                            try
                            {
                                client._routes.deliverData(m, key);
                                break;
                            }
                            catch (MessageStreamFullException)
                            {
                                checkAndSendHeartbeat(false);
                            }
                        } while (client.connected);
                    }
                    else if ((commandType & PublishMask) > 0)
                    {
                        /* Not currently implemented to avoid an extra branch in delivery.
                        client._globalCommandTypeHandlers[(int)(commandType==Message.Commands.OOF?GlobalCommandTypeHandlers.OOF:GlobalCommandTypeHandlers.Publish)].invoke(m);
                        */
                        Field subIds = m.getSubIdsRaw();
                        BookmarkField bookmark = m.getBookmarkRaw();
                        // Publish messages coming through on a subscription
                        int index = 0;
                        while( index < subIds.length )
                        {
                            int end = index;
                            for (; end < subIds.length && subIds.buffer[subIds.position + end] != (byte)','; ++end) ;
                            key.set(subIds.buffer, subIds.position + index, end-index);

                            MessageHandler handler = client._routes.findRoute(key);
                            bool isMessageQueue = !m.getLeasePeriodRaw().isNull();
                            if(handler != null)
                            {
                                // Make sure to refer to message's buffer, not
                                // the key's buffer since only the message
                                // buffer gets copied if the message is copied.
                                m.setSubId(subIds.buffer,
                                           subIds.position + index,
                                           end-index);
                                m.setSubscription(null);
                                if(!bookmark.isNull() && !isMessageQueue)
                                {
                                    // send it on to the BookmarkStore for logging,
                                    // only log those messages we're really delivering.
                                    // check to see if the log says its a duplicate
                                    if(client.bookmarkStore.isDiscarded(m))
                                    {
                                        // yes it is: don't send it
                                        try { this.client._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.DuplicateMessage].invoke(m); } catch (Exception e) { client.absorbedException(e); }
                                    }
                                    else  // not a dupe, log it.
                                    {
                                        client.bookmarkStore.log(m);
                                        try
                                        {
                                            do
                                            {
                                                try
                                                {
                                                    handler.invoke(m);
                                                    break;
                                                }
                                                catch (MessageStreamFullException)
                                                {
                                                    checkAndSendHeartbeat(false);
                                                }
                                            } while (client.connected);
                                        } catch (Exception e) { client.absorbedException(e); }
                                    }
                                }
                                else
                                {
                                    if (isMessageQueue) m._client = client;
                                    if (isMessageQueue && client._isAutoAckEnabled)
                                    {
                                        // message queue auto-acking code
                                        try
                                        {
                                            do
                                            {
                                                try
                                                {
                                                    handler.invoke(m);
                                                    break;
                                                }
                                                catch (MessageStreamFullException)
                                                {
                                                    checkAndSendHeartbeat(false);
                                                }
                                            } while (client.connected);
                                            if (!m.isIgnoreAutoAck())
                                            {
                                                client._ack(m.getTopicRaw(),
                                                            m.getBookmarkRaw());
                                            }
                                        }
                                        catch (Exception e)
                                        {
                                            client.absorbedException(e);
                                            try
                                            {
                                                client._ack(m.getTopicRaw(),
                                                            m.getBookmarkRaw(),
                                                            CANCEL_FIELD);
                                            }
                                            catch (Exception eInner)
                                            {
                                                client.absorbedException(eInner);
                                            }
                                        }
                                    }
                                    else
                                    {
                                        // just send to handler
                                        try
                                        {
                                            do
                                            {
                                                try
                                                {
                                                    handler.invoke(m);
                                                    break;
                                                }
                                                catch (MessageStreamFullException)
                                                {
                                                    checkAndSendHeartbeat(false);
                                                }
                                            } while (client.connected);
                                        } catch (Exception e) { client.absorbedException(e); }
                                    }
                                } // if !bookmark.isNull()
                            }// if handler!=null
                            index = end + 1;
                        } // while index < subIds.length
                    }
                    else if (commandType == Message.Commands.Ack)
                    {
                        client._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.Ack].invoke(m);
                        Message.AckTypes ackType = m.getAckType();

                        // call any handlers registered for this ack
                        do
                        {
                            try
                            {
                                client._routes.deliverAck(m, ackType);
                                break;
                            }
                            catch (MessageStreamFullException)
                            {
                                checkAndSendHeartbeat(false);
                            }
                        } while (client.connected);
                        switch(ackType)
                        {
                        // Persisted acks are handled internally.
                        case Message.AckTypes.Persisted:
                            persistedAck(m);
                            return;
                        // Processed acks are used to respond to waiting synchronous commands.
                        case Message.AckTypes.Processed:
                            processedAck(m);
                            return;
                        default:
                            break;
                        }
                    } else if (m.getCommand() == Message.Commands.Heartbeat)
                    {
                        client._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.Heartbeat].invoke(m);
                        if (client._heartbeatInterval != 0)
                        {
                            checkAndSendHeartbeat(true);
                        }
                        else
                        {
                            client._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.LastChance].invoke(m);
                        }
                        return;
                    }
                    else
                    {
                        int deliveries = 0;
                        do
                        {
                            try
                            {
                                deliveries = client._routes.deliverData(m, key);
                                break;
                            }
                            catch (MessageStreamFullException)
                            {
                                checkAndSendHeartbeat(false);
                            }
                        } while (client.connected);
                        if (deliveries == 0)
                        {
                            client._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.LastChance].invoke(m);
                        }
                    }
                    checkAndSendHeartbeat(false);
                }
                catch(Exception e)
                {
                    this.client.absorbedException(e);
                }
            }

            internal void checkAndSendHeartbeat(bool force)
            {
                if (force || client.heartbeatStopwatch.ElapsedMilliseconds > client._heartbeatInterval*1000)
                {
                    try
                    {
                        client.heartbeatStopwatch.Restart();
                        client.sendWithoutRetry(client.beatMessage);
                    }
                    catch (Exception)
                    {
                        // we'll pass on handling the disconnect here
                    }
                }
            }

            private void processedAck(Message message)
            {
                if(message.getCommandId(key))
                {
                    // If this command is awaiting a Processed ack, let's notify it
                    // that it's arrived
                    lock (client.acksLock)
                    {
                        try
                        {
                            AckResponse response = null;
                            if (client._acks.TryGetValue(key, out response))
                            {
                                response.State = message.getStatus();
                                response.Reason = message.getReason();
                                response.ReasonText = message.getReasonText();
                                response.Username = message.getUserId();
                                response.Password = message.getPassword();
                                response.ServerVersion = new VersionInfo(message.getVersion());
                                response.Sequence = message.getSequenceRaw().isNull() ? 0 : message.getSequence();
                                response.Responded = true;  // Reset after setting Status/Reason
                                response.LeasePeriod = message.getLeasePeriod();
                                response.Options = message.getOptions();
                                response.Bookmark = message.getBookmark();
                                client._acks.Remove(key);
                                Monitor.PulseAll(client.acksLock);
                                return;
                            }
                        }
                        catch (Exception e)
                        {
                            this.client.absorbedException(e);
                        }
                    }

                    try
                    {
                        client._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.LastChance].invoke(message);
                    }
                    catch (Exception e)
                    {
                        this.client.absorbedException(e);
                    }
                }
            }

            private void persistedAck(Message message)
            {

                // Best Practice:  If you don't care about the dupe acks that occur during
                // failover or rapid disconnect/reconnect, then just ignore them.
                // We could discard each duplicate from the persisted store, but the
                // storage costs of doing 1 record discards is heavy.  In most scenarios
                // we'll just quickly blow through the duplicates and get back to
                // processing the non-dupes.
                bool handled = false;
                Message.Reasons reason = message.getReason();
                if (!message.isSequenceNull())
                {
                    if (client.publishStore != null)
                    {
                        handled = true;
                        UlongField sequenceField = (UlongField)message.getSequenceRaw();
                        ulong sequence = sequenceField.getValue();
                        if ((message.getStatus() == Message.State.Failure ||
                             reason == Message.Reasons.Duplicate ||
                             reason == Message.Reasons.NotEntitled) &&
                            client._failedWriteHandler != null)
                        {
                            try
                            {
                                FailedWriteReplayer replayer = new FailedWriteReplayer(client, reason);
                                client.publishStore.replaySingle(replayer, sequence);
                            }
                            catch (Exception e)
                            {
                                this.client.absorbedException(e);
                            }
                        }
                        try
                        {
                            client.publishStore.discardUpTo(sequence);
                        }
                        catch (Exception e)
                        {
                            this.client.absorbedException(e);
                        }
                    }
                    else if ((message.getStatus() == Message.State.Failure ||
                             reason == Message.Reasons.Duplicate ||
                             reason == Message.Reasons.NotEntitled) &&
                            client._failedWriteHandler != null)
                    {
                        handled = true;
                        try
                        {
                            UlongField sequenceField = (UlongField)message.getSequenceRaw();
                            ulong sequence = sequenceField.getValue();
                            message.reset();
                            message.setSequence(sequence);
                            client._failedWriteHandler.failedWrite(message, reason);
                        }
                        catch (Exception e)
                        {
                            this.client.absorbedException(e);
                        }
                    }
                }
                if (client.bookmarkStore != null && !handled)
                {
                    BookmarkField bookmark = message.getBookmarkRaw();
                    if (bookmark != null && bookmark.length > 1 &&
                        message.isSequenceNull())
                    {
                        try
                        {
                            Field subId = message.getSubIdRaw();
                            if (client._routes.hasRoute(new CommandId(subId.buffer,
                                                                      subId.position,
                                                                      subId.length)))
                            {
                                handled = true;
                                client.bookmarkStore.persisted(message.getSubIdRaw(), bookmark);
                            }
                        }
                        catch (AMPSException e)
                        {
                            this.client.absorbedException(e);
                        }
                    }
                }
                if(!handled)
                {
                  try
                  {
                      client._globalCommandTypeHandlers[(int)GlobalCommandTypeHandlers.LastChance].invoke(message);
                  }
                  catch (Exception e)
                  {
                      this.client.absorbedException(e);
                  }
                }
            }


            // This is our way of "spying" on the activity of the user's disconnect handler.
            // We need to know if the user successfully connected to the underlying at least once,
            // to tell if they've "given up" on reconnecting or not.
            class TransportConnectionMeasurer : Transport
            {
                Transport _t;
                bool _successfullyConnected = false;

                /// <summary>
                /// Initializes a new instance of the TransportConnectionMeasurer class.
                /// </summary>
                /// <param name="underlying">The underlying transport to measure.
                public TransportConnectionMeasurer(Transport underlying)
                {
                    _t = underlying;
                }

                /// <summary>
                /// Clears the connection state measurement.
                /// </summary>
                public void clear()
                {
                    _successfullyConnected = false;
                }

                /// <summary>
                /// Indicates whether the transport has successfully connected.
                /// </summary>
                /// <returns>True if the transport has successfully connected, otherwise false.</returns>
                public bool successfullyConnected()
                {
                    return _successfullyConnected;
                }

                /// <inheritdoc/>
                public void connect(Uri uri)
                {
                    _t.connect(uri);
                    _successfullyConnected = true;
                }

                /// <inheritdoc/>
                public void close()
                {
                    _t.close();
                }

                /// <inheritdoc/>
                public void disconnect()
                {
                    _t.disconnect();
                }

                /// <inheritdoc/>
                public void setMessageHandler(MessageHandler ml)
                {
                    _t.setMessageHandler(ml);
                }

                /// <inheritdoc/>
                public void setDisconnectHandler(TransportDisconnectHandler dh)
                {
                    _t.setDisconnectHandler(dh);
                }

                /// <inheritdoc/>
                public void setExceptionListener(TCPTransportImpl.ExceptionListener exceptionListener)
                {
                    _t.setExceptionListener(exceptionListener);
                }

                /// <inheritdoc/>
                public void setThreadCreatedHandler(ThreadCreatedHandler handler)
                {
                    _t.setThreadCreatedHandler(handler);
                }

                /// <inheritdoc/>
                public void send(Message message)
                {
                    _t.send(message);
                }

                /// <inheritdoc/>
                public void sendWithoutRetry(Message message)
                {
                    _t.sendWithoutRetry(message);
                }

                /// <inheritdoc/>
                public Message allocateMessage()
                {
                    return _t.allocateMessage();
                }

                /// <inheritdoc/>
                public long writeQueueSize()
                {
                    return _t.writeQueueSize();
                }

                /// <inheritdoc/>
                public long flush()
                {
                    return _t.flush();
                }

                /// <inheritdoc/>
                public long readQueueSize()
                {
                    return _t.readQueueSize();
                }

                /// <inheritdoc/>
                public long flush(long timeout)
                {
                    return _t.flush(timeout);
                }

                /// <inheritdoc/>
                public void handleCloseEvent(int failedVersion, String message, Exception exception)
                {
                    _t.handleCloseEvent(failedVersion, message, exception);
                }

                /// <inheritdoc/>
                public void setReadTimeout(int readTimeout_)
                {
                    _t.setReadTimeout(readTimeout_);
                }

                /// <inheritdoc/>
                public int getVersion()
                {
                    return _t.getVersion();
                }

                /// <inheritdoc/>
                public void setTransportFilter(TransportFilter filter_)
                {
                    _t.setTransportFilter(filter_);
                }

                /// <inheritdoc/>
                public void setTransportFilter(TransportFilterModifiable filter_)
                {
                    _t.setTransportFilter(filter_);
                }

                public Action getIdleAction()
                {
                    return _t.getIdleAction();
                }

                /// <inheritdoc/>
                public void setIdleAction(Action idleAction)
                {
                    _t.setIdleAction(idleAction);
                }

                /// <inheritdoc/>
                public void initFromClient(Client client_)
                {
                    _t.initFromClient(client_);
                }
            }

            /// <summary>
            /// The TransportDisconnectHandler implementation.
            /// </summary>
            /// <param name="newTransport">New transport that will be opened during reconnect
            public void invoke(Transport newTransport, Exception exception)
            {
                Transport oldTransport = null;
                try
                {
                    lock (client.internalLock)
                    {
                        if(client.ConnectionStateChanged!=null && client.connected) client.ConnectionStateChanged(client, new ConnectionStateChangedEventArgs(ConnectionStateChangedEventArgs.ConnectionState.Disconnected));
                        oldTransport = client.transport;
                        TransportConnectionMeasurer measurer = new TransportConnectionMeasurer(newTransport);
                        client.transport = measurer;
                        while (true)
                        {
                            bool connectComplete = false;
                            try
                            {
                                // Subscriptions aren't saved by default sub manager
                                // Only set this to true if we have a real sub manager
                                client._reconnectingSubscriptionManager =
                                                    !client._defaultSubscriptionManager;
                                measurer.clear();
                                try
                                {
                                    Monitor.Exit(client.internalLock);
                                    client.disconnectHandler.invoke(client);
                                }
                                finally
                                {
                                    Monitor.Enter(client.internalLock);
                                }
                                connectComplete = true;
                            }
                            finally
                            {
                                if (!connectComplete)
                                {
                                    client._reconnectingSubscriptionManager = false;
                                }
                                Monitor.PulseAll(client.internalLock);
                            }
                            // If the disconnect handler successfully connected to something,
                            // we keep going, even if the connection isn't alive now.
                            if (!measurer.successfullyConnected())
                            {
                                if (client.ConnectionStateChanged != null)
                                    client.ConnectionStateChanged(client, new ConnectionStateChangedEventArgs(ConnectionStateChangedEventArgs.ConnectionState.Shutdown));
                                client.connected = false;
                                client.absorbedException(new DisconnectedException("reconnect failed."));
                                client._reconnectingSubscriptionManager = false;
                                Monitor.PulseAll(client.internalLock);
                                return;
                            }
                            try
                            {
                                Monitor.Exit(client.internalLock);
                                client.subscriptionManager.resubscribe(client);
                                if (client.ConnectionStateChanged != null)
                                    client.ConnectionStateChanged(client, new ConnectionStateChangedEventArgs(ConnectionStateChangedEventArgs.ConnectionState.Resubscribed));
                                break;
                            }
                            catch (AMPSException ex)
                            {
                                // exception while attempting to resubscribe.
                                this.client.absorbedException(ex);
                                this.client._disconnect(true);
                            }
                            finally
                            {
                                Monitor.Enter(client.internalLock);
                                client._reconnectingSubscriptionManager = false;
                                Monitor.PulseAll(client.internalLock);
                            }
                        }
                    }
                }
                catch (Exception e)
                {
                    if (client.ConnectionStateChanged != null)
                        client.ConnectionStateChanged(client, new ConnectionStateChangedEventArgs(ConnectionStateChangedEventArgs.ConnectionState.Shutdown));
                    client.connected = false;
                    this.client.absorbedException(e);
                }
                finally
                {
                    client._reconnectingSubscriptionManager = false;
                    client.transport = oldTransport;
                }

            }

        }

        /// <summary>
        /// Internal-access helper method that allows other parts of the client API
        /// to make sure that heartbeat messages are sent in a timely manner.
        /// </summary>
        internal void checkAndSendHeartbeat(bool force) {
            handler.checkAndSendHeartbeat(force);
        }


        /// <summary>
        /// Implementation of <typeparamref name="StoreReplayer"/> to replay messages which were
        /// recorded as published, but did not receive an ack. This is performed as
        /// part of the logon function by default.
        /// </summary>
        class ClientStoreReplayer : StoreReplayer
        {
            Client client;

            /// <summary>
            /// Initializes a new instance of the ClientStoreReplayer class.
            /// </summary>
            /// <param name="c">The client associated with the store replayer.
            public ClientStoreReplayer(Client c)
            {
                this.client = c;
            }

            /// <summary>
            /// Executes the store replayer with the provided message.
            /// </summary>
            /// <param name="message">The message to replay.
            public void execute(Message message)
            {
                // Null command should simply reset last sent
                // Cancel acks shouldn't be resent since disconnect caused a cancel anyway
                if (!message.isCommandNull() &&
                    (message.isOptionsNull() || !client._logonInProgress
                     || !message.getOptions().Contains("cancel")))
                {
                    message.setAckType(message.getAckTypeOutgoing()|
                                       Message.AckTypes.Persisted);
                    client.transport.sendWithoutRetry(message);
                }

                if (message.getSequence() >= client.lastSentHaSequenceNumber)
                {
                    client.lastSentHaSequenceNumber = message.getSequence();
                }
            }
        }

        /// <summary>
        /// Adapter class for FailedWriteHandlerV4 to make it compatible with V4.
        /// </summary>
        class FailedWriteHandlerV4Compat : FailedWriteHandler
        {
            FailedWriteHandlerV4 _handler;

            /// <summary>
            /// Initializes a new instance of the FailedWriteHandlerV4Compat class
            /// </summary>
            /// <param name="handler_">The original V4 handler.
            public FailedWriteHandlerV4Compat(FailedWriteHandlerV4 handler_)
            {
                this._handler = handler_;
            }

            /// <summary>
            /// Handles a failed write for the given message and reason.
            /// </summary>
            /// <param name="message_">The message that failed to write.
            /// <param name="reason_">The reason for the failure.
            public void failedWrite(Message message_, Message.Reasons reason_)
            {
                Message.Commands command = message_.getCommand();
                if (command == Message.Commands.Unknown) // no store
                {
                    _handler.failedWrite(0, command, null, 0, 0, null, 0, 0,
                                         null, 0, 0, reason_);
                    return;
                }
                Field topic = message_.getTopicRaw();
                Field corId = message_.getCorrelationIdRaw();
                Field data = null;
                if (command == Message.Commands.Publish ||
                    command == Message.Commands.DeltaPublish ||
                    (command == Message.Commands.SOWDelete &&
                     !message_.isDataNull()))
                {
                    data = message_.getDataRaw();
                }
                else if (!message_.isFilterNull())
                {
                    data = message_.getFilterRaw();
                }
                else
                {
                    data = message_.getSowKeysRaw();
                }
                _handler.failedWrite(message_.getSequence(), command,
                                     topic.buffer, topic.position, topic.length,
                                     data.buffer, data.position, data.length,
                                     corId.buffer, corId.position, corId.length,
                                     reason_);
            }
        }

        // Implementation of StoreReplayer to replay failed writes.
        class FailedWriteReplayer : StoreReplayer
        {
            Client _client;
            Message.Reasons _reason;
            int _replayCount;

            /// <summary>
            /// Initializes a new instance of the FailedWriteReplayer class.
            /// </summary>
            /// <param name="c">The client associated with the store replayer.
            /// <param name="reason">The reason for the failed write.
            public FailedWriteReplayer(Client c, Message.Reasons reason)
            {
                this._client = c;
                this._reason = reason;
            }

            /// <summary>
            /// Executes the store replayer with the provided message.
            /// </summary>
            /// <param name="message">
            public void execute(Message message)
            {
                ++_replayCount;
                _client._failedWriteHandler.failedWrite(message, _reason);
            }

            /// <summary>
            /// Gets the count of replays performed.
            /// </summary>
            /// <returns>The count of replays.</returns>
            public int replayCount() { return _replayCount; }

        }

        // Cancels synchronous waiters up to the specified failed version.
        private void cancelSynchronousWaiters(int failedVersion)
        {
            var removeList = new List<CommandId>();
            lock (acksLock)
            {
                // _disconnect() calls this with Int32.MaxValue, so don't save that
                if (failedVersion < Int32.MaxValue)
                {
                    _lastFailedConnectionVersion = failedVersion;
                }
                foreach (var entry in _acks)
                {
                    if (entry.Value.ConnectionVersion <= failedVersion)
                    {
                        entry.Value.Abandoned = true;
                        removeList.Add(entry.Key);
                    }
                }
                foreach (CommandId cid in removeList)
                {
                    _acks.Remove(cid);
                }
                Monitor.PulseAll(acksLock);
            }
        }

        /// <summary>
        /// Disposes of the client resources.
        /// </summary>
        public void Dispose()
        {
            try
            {
                this.disconnect();
                if (publishStore != null)
                {
                    publishStore.Dispose();
                    publishStore = null;
                }
                if (bookmarkStore != null)
                {
                    bookmarkStore.Dispose();
                    bookmarkStore = null;
                }
            }
            catch (AMPSException)
            {
                // just eat these, here.
            }
        }

        // Flushes acknowledgments for queued bookmarks that have timed out.
        private void _ackFlusher()
        {
            if(_queueAckTimeout > 0 && _topicHashMapSize > 0
               && this.transport != null)
            {
                bool locked = false;
                try
                {
                    locked = Monitor.TryEnter(internalLock, 50);
                    if (locked)
                    {
                        foreach(var queueAckBookmarks in _topicHashMap.Values)
                        {
                            if(queueAckBookmarks._bookmarkCount>0 && (DateTime.Now - queueAckBookmarks._oldestTime).TotalMilliseconds > _queueAckTimeout)
                            {
                                _ack(queueAckBookmarks);
                            }
                        }
                    }
                }
                finally
                {
                    if (locked)
                    {
                        Monitor.Exit(internalLock);
                    }
                }
            }
        }
    }
}
